package appl.ext;

import java.awt.Component;
import java.util.Set;
import java.util.Vector;

import javax.swing.DefaultListModel;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JTable;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableModel;

import appl.util.NonEditableTableModel;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.gui.XuluInternalFrame;
import edu.bonn.xulu.plugin.appl.AbstractMenuPlugin;

/**
 * Handles the {@link ConfigurationEditorGUI}. The prefixes of the
 * {@link XuluConfig} are added to the {@link JList} and the suffixes to the
 * {@link JTable}. For the documentation column for every key with a name lets
 * say 'key' the value of "key.doc" is looked up in the configuration. The
 * prefixes which are values of the key "Filter.prefixes" in the
 * configuration are ignored. Also all keys which show up in "Filter.keys"
 * 
 * @see XuluConfig
 * @author Dominik Appl
 */
public class ConfigurationEditorEngine implements ListSelectionListener,
		TableModelListener {

	ConfigurationEditorGUI GUI = null;

	private DefaultTableModel tableModel;

	private JTable entryTable;

	private JList categoryList;

	private XuluConfig config;

	private String currentCategory;

	/**
	 * creates a new Editor. Use {@link #getGUI()} to access the GUI.
	 */
	public ConfigurationEditorEngine() {
		config = XuluConfig.getXuluConfig();
		GUI = new ConfigurationEditorGUI();
		initGUI();
	}

	/**
	 * inits the GUI
	 */
	private void initGUI() {

		entryTable = GUI.entryTable;
		categoryList = GUI.keyList;
		initList();
		initTable();
	}

	/**
	 * initializes the table
	 */
	private void initTable() {
		// only the 2nd column is editable
		tableModel = new NonEditableTableModel(
				new Object[] { "Name", "Entry" },
				new boolean[] { false, true }, 0);
		entryTable.setModel(tableModel);
		entryTable.setDefaultRenderer(Object.class, new ToolTipRenderer());
		entryTable.getModel().addTableModelListener(this);
		entryTable.getSelectionModel().addListSelectionListener(this);
		String firstElement = (String) categoryList.getModel().getElementAt(0);
		if (firstElement != null)
			initTableData(firstElement);
	}

	/**
	 * The table is initalized with the given category
	 * 
	 * @param firstElement
	 */
	private void initTableData(String categoryName) {
		// get ignored keys
		String[] filteredKeys = config.getMultiProperty("Filter.keys");
		Vector<String> filter = new Vector<String>();
		for (String string : filteredKeys) {
			filter.add(string);
		}
		Set<String> suffixes = config.getSuffixesOfPrefix(categoryName);
		// clear table
		tableModel.setRowCount(0);
		for (String suffix : suffixes) {
			// do not enter the documentation as seperate entries
			if (suffix.endsWith(".doc"))
				continue;
			String key = categoryName + "." + suffix;
			// do not enter filtered keys
			if (filter.contains(key))
				continue;
			String value = config.getProperty(key);
			tableModel.addRow(new Object[] { suffix, value });
		}
		currentCategory = categoryName;
	}

	public ConfigurationEditorGUI getGUI() {
		return GUI;
	}

	/**
	 * 
	 */
	private void initList() {
		// get ignored prefixes
		String[] filteredPrefixes = config.getMultiProperty("Filter.prefixes");
		Vector<String> filter = new Vector<String>();
		for (String string : filteredPrefixes) {
			filter.add(string);
		}
		DefaultListModel listModel = new DefaultListModel();
		Object[] prefixes = config.getPrefixes().toArray();
		for (Object object : prefixes) {
			if (!filter.contains(object))
				listModel.addElement(object);
		}
		this.GUI.keyList.setModel(listModel);
		// select first element
		categoryList.setSelectedIndex(0);
		categoryList.addListSelectionListener(this);
	}

	/**
	 * Listens for selection changes in list and table
	 * 
	 * @see javax.swing.event.ListSelectionListener#valueChanged(javax.swing.event.ListSelectionEvent)
	 */
	public void valueChanged(ListSelectionEvent e) {
		if (e.getSource() == categoryList) {
			String value = (String) categoryList.getSelectedValue();
			if (value != null)
				initTableData(value);
			// clear documentation field
			GUI.docArea
					.setText("<select a table value to see its documentation>");
		}
		// update documentation
		else if (e.getSource() == entryTable.getSelectionModel()) {
			int selRow = entryTable.getSelectedRow();
			// if a row is selected:
			if (selRow >= 0) {
				String value = (String) entryTable.getModel().getValueAt(
						selRow, 0);
				String docKey = currentCategory + "." + value + ".doc";
				String doc = config.getProperty(docKey);
				if (doc == null)
					doc = "<no documentation found for this entry>";
				GUI.docArea.setText(doc);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see javax.swing.event.TableModelListener#tableChanged(javax.swing.event.TableModelEvent)
	 */
	public void tableChanged(TableModelEvent e) {
		// handle the editing the user does
		if (e.getType() == TableModelEvent.UPDATE) {
			int changedRow = e.getFirstRow();
			// only the 2nd col is editable
			String value = tableModel.getValueAt(changedRow, 1).toString();
			String suffix = tableModel.getValueAt(changedRow, 0).toString();
			String prefix = categoryList.getSelectedValue().toString();
			String key = prefix + "." + suffix;
			config.setProperty(key, value);

		}
	}

	/**
	 * just for the tooltips
	 * 
	 * @author Dominik Appl
	 */
	private class ToolTipRenderer extends DefaultTableCellRenderer {

		public ToolTipRenderer() {
			setOpaque(true);
		}

		public Component getTableCellRendererComponent(JTable table,
				Object color, boolean isSelected, boolean hasFocus, int row,
				int column) {
			Object value = table.getModel().getValueAt(row, column);
			if (value != null)
				setToolTipText(value.toString());
			return super.getTableCellRendererComponent(table, color,
					isSelected, hasFocus, row, column);
		}

	}
}