package appl.ext;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Properties;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import org.geotools.io.DefaultFileFilter;

/**
 * Simple Property store, which is loaded at each Xulu start and saved at each
 * Xulu exit. Every class or plugin can store and retrieve configuration
 * information easy through a static method. To keep the property file readable
 * and to avoid conflicts with other plugins please prefix the key-name with a
 * unique name, which identifies your class, followed by a dot and the property.
 * This is important for correct display in a User-GUI (which could be
 * implemented). <br>
 * <br>
 * Keyformat: <code>domain.property</code> <br>
 * Example key: <code>DiscoveryService.className</code> <br>
 * <br>
 * There could be no double entries.But you can save multiple values to one key
 * using the {@link #setMultiProperty(String, String[]) setMultiProperty}
 * method. <br>
 * <b>These values are internally separated with ';', so see that you don't use
 * ';' in your values of {@link #setMultiProperty(String, String[]) Multi
 * Properties}</b> <br>
 * <br>
 * First the settings of the file "DefaultProperties" are loaded. After that the
 * values are possibly overwritten by the loading of the file "XuluProperties"
 * in which is intended to store a actual user configuration.
 * <b>"DefaultPropoerties" should never be modified by the user.</b> The
 * programmer should store all default values there. <br>
 * <br>
 * The user may change values through the {@link ConfigurationEditorGUI}.
 *
 * @see ConfigurationEditorGUI
 * @see ConfigurationEditorEngine
 * @author Dominik Appl
 */
public class XuluConfig {

	private final String delimiter = ";";

	// File for default values
	File defaultFile = new File("DefaultProperties");

	// File in which the props are stored
	File file = new File("XuluProperties");

	Properties propStore = null;

	private static XuluConfig instance = null;

        private XuluConfig() {
          this(null);
        }

	private XuluConfig(File rootDir) {
                if ( rootDir != null && rootDir.isDirectory() ) {
                  defaultFile = new File(rootDir, defaultFile.getName());
                  file        = new File(rootDir, file.getName());
                }
		// try to read defaults
		Properties defaults = new Properties();
		try {
			defaults.load(new FileInputStream(defaultFile));
		} catch (Exception e) {
			System.out
					.println("Warning: defaults could not be loaded from file '"
							+ defaultFile + "'");
			System.out.println(e.getMessage());
		}
		propStore = new Properties(defaults);
		if (file.exists())
			load();
		else
			store();
	}

        public static XuluConfig getXuluConfig() {
          return getXuluConfig(null);
        }

	public static XuluConfig getXuluConfig(File rootDir) {
		if (instance == null)
			instance = new XuluConfig(rootDir);
		return instance;
	}

	/**
	 * Saves the configuration to the file system
	 */
	public void store() {
		try {
			propStore.store(new FileOutputStream(file),
					"Stored properties of the Xulu modelling platform");
		} catch (FileNotFoundException e) {
			System.err.println("Could not write PropertyStore!!!");
			e.printStackTrace();
		} catch (IOException e) {
			System.err.println("Could not write PropertyStore!!!");
			e.printStackTrace();
		}

	}

	/**
	 * Sets the Property file. Notice that this has only an effect if you call
	 * load afterwards.
	 *
	 * @param file
	 */
	public void setPropertyFile(File file) {
		this.file = file;
	}

	/**
	 * loads the Properties from the Propertyfile. No loaded values are deleted -
	 * only existing ones are overwritten;
	 */
	public void load() {
		try {
			propStore.load(new FileInputStream(file));
		} catch (FileNotFoundException e) {
			System.err.println("Property file with name '" + file
					+ "' not found");
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * inserts the property value with the specified key into the
	 * {@link XuluConfig}
	 *
	 * @param key
	 *            the key
	 * @param value
	 *            the value associated with the key
	 */
	public void setProperty(String key, String value) {
		if (!key.contains("."))
			System.out
					.println("WARNING for key "
							+ key
							+ ": Propertykeys in XuluConfig should have at least one '.' ! ");
		propStore.setProperty(key, value);
	}

	/***************************************************************************
	 * @return the property associated with the provided key. If not found the
	 *         defaults are searched. If still not found, null is returned.
	 *         Gives out a warning to console if the key was not found
	 **************************************************************************/
	public String getProperty(String key) {
		String value = propStore.getProperty(key);
		if (value == null)
			System.err.println("Warning! Value for property '" + key
					+ "' was not found in property file!");
		return value;
	}

	/***************************************************************************
	 * Same as {@link #getProperty(String)}, but you can disable the warning
	 * message
	 *
	 * @param giveWarning
	 *            if true, a warning is given to the console if the specified
	 *            entry was not found
	 * @return the property associated with the provided key. If not found the
	 *         defaults are searched. If still not found, null is returned.
	 *         Gives out a warning to console if the key was not found
	 **************************************************************************/
	public String getProperty(String key, boolean giveWarning) {
		String value = propStore.getProperty(key);
		if (value == null && giveWarning)
			System.err.println("Warning! Value for property '" + key
					+ "' was not found in property file!");
		return value;
	}

	/***************************************************************************
	 *
	 * @return the property associated with the provided key as integer value.
	 *         If not found the defaults are searched. If not found or the value
	 *         is in a wrong format 0 is returned and a warning is given to the
	 *         console
	 **************************************************************************/
	public int getIntProperty(String key) {
		String value = propStore.getProperty(key);
		if (value == null) {
			System.err.println("Error! Value for property '" + key
					+ "' was not found in property file!");
			return 0;
		}
		int returnvalue = 0;
		try {
			returnvalue = Integer.valueOf(value);
		} catch (java.lang.NumberFormatException e) {
			System.err
					.println("Error! Value for property '"
							+ key
							+ "' was not in the right format. Should be a Integer value! - Returning 0");
		}
		return returnvalue;
	}

	/***************************************************************************
	 * @return the property associated with the provided key as double value. If
	 *         not found the defaults are searched. If not found or the value is
	 *         in a wrong format 0 is returned and a warning is given to the
	 *         console
	 **************************************************************************/
	public double getDoubleProperty(String key) {
		String value = propStore.getProperty(key);
		if (value == null) {
			System.err.println("Error! Value for property '" + key
					+ "' was not found in property file!");
			return 0;
		}
		double returnvalue = 0;
		try {
			returnvalue = Double.valueOf(value);
		} catch (java.lang.NumberFormatException e) {
			System.err
					.println("Error! Value for property '"
							+ key
							+ "' was not in the right format. Should be a Integer value! - Returning 0");
		}
		return returnvalue;
	}

	/**
	 * inserts the property value with the specified key into the
	 * {@link XuluConfig}
	 *
	 * @param key
	 *            the key
	 * @param value
	 *            the value associated with the key
	 */
	public void setBooleanProperty(String key, boolean value) {
		this.setProperty(key, String.valueOf(value));
	}

	/**
	 * inserts the property value with the specified key into the
	 * {@link XuluConfig}
	 *
	 * @param key
	 *            the key
	 * @param value
	 *            the value associated with the key
	 */
	public void setIntProperty(String key, int value) {
		this.setProperty(key, String.valueOf(value));
	}

	/***************************************************************************
	 * @return the property associated with the provided key as boolean value.
	 *         If not found the defaults are searched. If still not found or
	 *         null then false is returned. Gives out a warning to console if
	 *         the key was not found
	 **************************************************************************/
	public boolean getBooleanProperty(String key) {
		String value = propStore.getProperty(key);
		if (value == null) {
			System.err.println("Error! Value for property '" + key
					+ "' was not found in property file!");
			return false;
		}
		if(value.equals("1"))
			return true;
		boolean bool = false;
		try {
			bool = Boolean.valueOf(value);
		} catch (Exception e) {
			System.err.println("Error! Value for property '" + key
					+ "' was not a valid boolean value");
		}
		return bool;
	}

	/**
	 * A multiproperty stores multiple values for one key. These values are
	 * internaly seperated with spaces, so see that you dont use spaces in your
	 * values.
	 *
	 * @param key
	 *            the key
	 * @param values
	 *            array of the values which should be associated with the key
	 */
	public void setMultiProperty(String key, String values[]) {
		if (!key.contains("."))
			System.out
					.println("WARNING for key "
							+ key
							+ ": Propertykeys in XuluConfig should have at least one '.' ! ");
		String value = "";
		if (values != null)
			for (String string : values) {
				value += (string + delimiter);
			}
		propStore.setProperty(key, value);
	}

	/***************************************************************************
	 * A multiproperty stores multiple values for one key. These values are
	 * internaly seperated with ';', so do not use ';' in your values. if the
	 * key was not found it gives out a warning to console and a empty array is
	 * retured.
	 *
	 * @return the multiProperty associated with the provided key, which is in
	 *         this case just a String array
	 **************************************************************************/

	public String[] getMultiProperty(String key) {
		String value = propStore.getProperty(key);
		if (value != null)
			return value.split(delimiter);
		System.err.println("Error! Value for property '" + key
				+ "' was not found in property file!");
		return new String[0];
	}

	/**
	 * removes the Property from the list
	 *
	 * @param key
	 */
	public void removeProperty(String key) {
		propStore.remove(key);
	}

	/**
	 * removes all keys starting with the value given. USE WITH CAUTION! If you
	 * give a String "a" all entrys starting with "a" will be removed from the
	 * registry
	 *
	 * @param key
	 */
	public void removeAll(String key) {
		if (key == null)
			return;

		Enumeration keys = propStore.propertyNames();
		while (keys.hasMoreElements()) {
			String propKey = (String) keys.nextElement();
			if (propKey == null)
				continue;
			if (propKey.startsWith(key))
				propStore.remove(propKey);
		}
	}

	/**
	 * Gets all prefixes. A Prefix is the string of all characters up to the
	 * first dot. A Suffix is the string of all characters following this dot.
	 *
	 * @return the prefixes
	 *
	 */
	public Set<String> getPrefixes() {
		TreeSet set = new TreeSet();
		Enumeration keys = propStore.propertyNames();
		while (keys.hasMoreElements()) {
			String propKey = (String) keys.nextElement();
			String[] strings = propKey.split("\\.");
			set.add(strings[0]);
		}
		return set;
	}

	/**
	 * Gets all suffixes for one prefix. A Prefix is the string of all
	 * characters up to the first dot. A Suffix is the string of all characters
	 * following this dot. The Strings are ordered alphabeticly
	 *
	 * @param prefix
	 *            the prefix.
	 */
	public Set<String> getSuffixesOfPrefix(String prefix) {
		TreeSet set = new TreeSet();
		Enumeration keys = propStore.propertyNames();
		while (keys.hasMoreElements()) {
			String propKey = (String) keys.nextElement();
			String[] strings = propKey.split("\\.");
			if (strings.length > 1 && strings[0].equals(prefix))
				set.add(propKey.substring(strings[0].length() + 1));
		}
		return set;
	}

	/**
	 * Gets all entries starting with the given {@link String}
	 *
	 * @param startingString
	 *            the prefixstring
	 * @return all keys starting the given prefix
	 */
	public Set<String> getKeysStartingWith(String startingString) {
		TreeSet set = new TreeSet();
		Enumeration keys = propStore.propertyNames();
		while (keys.hasMoreElements()) {
			String propKey = (String) keys.nextElement();
			if (propKey.startsWith(startingString))
				set.add(propKey);
		}
		return set;
	}

	/**
	 * Just for testing.
	 *
	 * @deprecated
	 */
	public static void main(String[] args) {
		XuluConfig config = XuluConfig.getXuluConfig();
		config.setProperty("atest2", "test3");
		config.setProperty("test4.testproperty", "test");
		String[] multiTest = { "value1", "value2", "value3" };
		config.setMultiProperty("multiproperty.1", multiTest);
		String[] multioutput = config.getMultiProperty("multiproperty.1");
		for (String string : multioutput) {
			System.out.println("Multivalue: " + string);
		}
		config.store();
		config.setPropertyFile(new File("XuluTestfile"));
		config.load();
		String[] multiTest2 = { "value1", "value2", "value3" };
		config.setMultiProperty("multiproperty.2", multiTest);
		config.removeAll("multiproperty");
		config.removeProperty("atest2");
		config.store();
	}
}
