package appl.parallel.client;

import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import appl.parallel.event.CommEvent;
import appl.parallel.event.CommEventSink;
import appl.parallel.event.RemoteEvent;
import appl.parallel.event.RemoteEventSink;
import appl.parallel.event.TimeEvent;
import appl.parallel.event.TimeMonitor;
import appl.parallel.event.TransferEvent;
import appl.parallel.event.TransferMonitor;
import appl.parallel.services.Service;
import appl.parallel.util.Helper;

/**
 * Starts a Service for receiving local and remote events. Monitors can register
 * to this service. All events are forwarded the registered listeners. The
 * service tries to get info for registry port from property
 * <code>XuluClient.registryport</code> in {@link XuluConfig}.
 * 
 * @author Dominik Appl
 */
public class RemoteEventHandler extends UnicastRemoteObject implements Service,
		CommEventSink {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	// defaults
	private String bindingName = "RemoteEventReceiver";

	private int registryPort = 1099;

	private Vector<TimeMonitor> timeMonitors = new Vector<TimeMonitor>();

	private Vector<TransferMonitor> transferMonitors = new Vector<TransferMonitor>();

	private boolean timeMonitoring = false;

	private boolean transferMonitoring = false;

	private boolean isRunning;

	/**
	 * @throws RemoteException
	 * @throws RemoteException
	 *             if binding fails
	 */
	public RemoteEventHandler() throws RemoteException {
		super();
		isRunning = false;
	}

	/**
	 * tries to bind to a running registry or creates one.
	 * 
	 * @throws RemoteException
	 *             if binding fails
	 */
	private void bind() throws RemoteException {
		// try to get port from registry
		int port = XuluConfig.getXuluConfig().getIntProperty(
				"XuluClient.registryport");
		if (port != 0)
			registryPort = port;
		// bind it
		CommEventSink sink = this;
		Helper.bind(bindingName, sink, registryPort);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#isRunning()
	 */
	public boolean isRunning() {
		return isRunning;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#startService()
	 */
	public void startService() {
		try {
			bind();
		} catch (RemoteException e) {
			LOG.error("Could not bind " + bindingName
					+ " to the registry! Eventsystem will fail.");
		}
		isRunning = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.services.Service#stopService()
	 */
	public void stopService() {
		unbind();
		isRunning = false;
	}

	/**
	 * unbinds the Service
	 */
	private void unbind() {
		Helper.unbind(bindingName);
	}

	public void fireRemoteEvent(TimeEvent t) throws RemoteException {
		for (TimeMonitor timeMonitor : timeMonitors) {
			timeMonitor.receiveTimeEvent(t);
		}
	}


	public void fireRemoteEvent(TransferEvent t) throws RemoteException {
		for (TransferMonitor transferMonitor : transferMonitors) {
			transferMonitor.receiveTransferEvent(t);
		}
	}

	

	/**
	 * Adds a new time monitor
	 * 
	 * @param monitor
	 *            a time monitor
	 */
	public void addTimeEventListener(TimeMonitor monitor) {
		if (monitor != null) {
			timeMonitors.add(monitor);
			timeMonitoring = true;
		}
	}

	/**
	 * Adds a new {@link TransferMonitor}
	 * 
	 * @param monitor
	 *            a time monitor
	 */
	public void addTransferEventListener(TransferMonitor monitor) {
		if (monitor != null) {
			transferMonitors.add(monitor);
			transferMonitoring = true;
		}
	}

	/**
	 * removes the given monitor
	 * 
	 * @param monitor
	 */
	public void removeTimeMonitor(TimeMonitor monitor) {
		timeMonitors.remove(monitor);
		if (timeMonitors.size() == 0)
			timeMonitoring = false;
	}

	/**
	 * adds the given Monitor
	 * 
	 * @param monitor
	 */
	public void removeTransferMonitor(TransferMonitor monitor) {
		transferMonitors.remove(monitor);
		if (transferMonitors.size() == 0)
			transferMonitoring = false;

	}

	/**
	 * @return true if the service is running and time monitoring is enabled
	 */
	public boolean isTimeMonitoringEnabled() throws RemoteException {
		return timeMonitoring;
	}

	/**
	 * @return true if the service is running and transfer monitoring is enabled
	 */
	public boolean isTransferMonitoringEnabled() throws RemoteException {
		return transferMonitoring;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.RemoteEventSink#fireRemoteEvent(appl.parallel.event.RemoteEvent)
	 */
	public void fireRemoteEvent(RemoteEvent e) throws RemoteException {
		if (e instanceof TimeEvent)
			fireRemoteEvent((TimeEvent) e);
		else if (e instanceof TransferEvent)
			fireRemoteEvent((TransferEvent) e);
		else
			// All events should be handled properly
			System.err.println("Received unknown event of class "
					+ e.getClass().getName());

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.RemoteEventSink#fireRemoteEvents(appl.parallel.event.RemoteEvent[])
	 */
	public void fireRemoteEvents(RemoteEvent[] e) throws RemoteException {
		for (RemoteEvent event : e) {
			fireRemoteEvent(event);
		}
	}

}
