package appl.parallel.data;

import java.awt.Rectangle;
import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.Naming;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.data.LoadingException;
import appl.parallel.client.DataServer;
import appl.parallel.client.ClientDataServer;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.spmd.split.SplittableResource;

/**
 * This class is used as a base class for loaders. It is especially intended for
 * use with the SPMD-Paradigm, but may be used with other loaders, too. It
 * implements the reusable stuff of the {@link PartitionDataHandler} Interface.
 * Notice that each loader is responsible for exactly one data element. It must
 * be {@link Cloneable}, so that multi-data elements can add another element.
 * 
 * @author Dominik Appl
 */
public abstract class AbstractDataHandler implements PartitionDataHandler {

	protected transient Logger LOG = LogManager.getLogger(this.getClass()
			.getName());

	/**
	 * The id of the currently handled {@link SplittableResource}
	 */
	protected int rootID;

	/**
	 * The partition bounds (the whole bounds) of the current
	 * {@link SplittableResource}
	 */
	protected Rectangle partitionBounds;

	/**
	 * the currently handled data (must be initialized before calling
	 * {@link #unload()}
	 */
	protected transient DataPartition data;

	/**
	 * The unload bounds differ may differ from the partitionBounds. If there is
	 * a neighborhood region it is important, that only the core-region (without
	 * the neighborhood) is unloaded.
	 */
	protected Rectangle unloadBounds;

	/**
	 * A {@link ClientDataServer} may be used locally for faster access.
	 */
	protected transient ClientDataServer spmdClient;

	/**
	 * Constructs a new {@link AbstractDataHandler}. The local IP address at
	 * the time of construction is later used on serverside for communication
	 * (e.g. unloading).
	 * 
	 * @param rootID
	 *            the id of the data
	 * @param client
	 *            a {@link ClientDataServer} which may be used to get meta
	 *            information on client side before transferring the loader to
	 *            its destination
	 * @param partitionBounds
	 *            the bounds of the partition to be retrieved on server side
	 * @param unloadBounds
	 *            the bounds of the partition which is to be uploaded to the
	 *            client after calculation (may only be the calculation area)
	 */
	public AbstractDataHandler(int rootID, ClientDataServer client,
			Rectangle partitionBounds, Rectangle unloadBounds) {
		this.spmdClient = client;
		this.rootID = rootID;
		this.partitionBounds = partitionBounds;
		this.unloadBounds = unloadBounds;
	}

	/**
	 * sets a local {@link ClientDataServer} which MAY be used by the handler
	 * for local access
	 * 
	 * @param localSPMDClient
	 */
	public void setSPMDClient(ClientDataServer localSPMDClient) {
		this.spmdClient = localSPMDClient;
	}

	/**
	 * (empty constructors are important for deserialization)
	 * 
	 * @see #AbstractDataHandler(int, ClientDataServer, Rectangle, Rectangle)
	 */
	public AbstractDataHandler() {
		this(-1, null, new Rectangle(0, 0, 0, 0), new Rectangle(0, 0, 0, 0));
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#getLoadInfo()
	 */
	public abstract String getLoadInfo();

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#getLoadInfo()
	 */
	public abstract String getUnloadInfo();

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#load()
	 */
	public abstract DataPartition load() throws LoadingException;

	/**
	 * Initializes the {@link Logger} after deserialization (used by
	 * {@link java.io.Serializable})
	 * 
	 * @throws IOException
	 * @see Serializable
	 */
	private synchronized void readObject(java.io.ObjectInputStream s)
			throws IOException, ClassNotFoundException {
		// reading standard fields
		s.defaultReadObject();
		// initialize Logger
		LOG = LogManager.getLogger(this.getClass().getName());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.PartitionDataHandler#setUnloadPartition(java.awt.Rectangle)
	 */
	public void setUnloadBounds(Rectangle unloadBounds) {
		this.unloadBounds = unloadBounds;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.PartitionDataHandler#setBasePartition(appl.parallel.spmd.split.DataPartition)
	 */
	public void setBasePartition(DataPartition data) {
		this.data = data;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataUnloader#unload()
	 */
	public abstract void unload();

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.PartitionDataHandler#setRootID(int)
	 */
	public void setRootID(int rootID) {
		this.rootID = rootID;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#clone()
	 */
	public PartitionDataHandler clone() {
		return newInstance(rootID, spmdClient, this.unloadBounds,
				this.partitionBounds);
	}

}
