package appl.parallel.data;

import java.awt.Rectangle;
import java.awt.image.DataBuffer;

import appl.data.DataProxy;
import appl.parallel.server.XuluServer;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.spmd.split.WritableGridPartition;
import appl.parallel.util.PartitionUtil;
import appl.util.RasterMetaData;
import schmitzm.data.WritableGrid;
import schmitzm.data.WritableGridArray;

/** 
 * A WritableGridArrayPartition extends the {@link WritableGridArray} so that
 * it can be used in parallel computation. Every partition represents a 
 * rectangle-shaped part of a father-grid. It has information about its bounds
 * and about its father-grid. The partition may be used for data transfer
 * between computing resources. It is very fast, because it is not proxyied
 * by a {@link DataProxy}, but uses the {@link WritableGridArray} directly. 
 * 
 * @see WritableGridArrayPartition
 * 
 * @author Dominik Appl
 */
public abstract class WritableGridArrayPartition {

	/**
	 * For Float data types
	 * 
	 * @see WritableGridArray.Float
	 * @author Dominik Appl
	 */
	public static class Float extends WritableGridArray.Float implements
			WritableGridPartition {

		private final int rootID;

		private final Rectangle thisPartition;

		private final RasterMetaData metaData;

		public Float(RasterMetaData metaData, int rootID,
				Rectangle thisPartition) {
			super(metaData.getMinX(), metaData.getMinY(), metaData.getWidth(),
					metaData.getHeight(), metaData.getX(), metaData.getY(),
					metaData.getRealWidth(), metaData.getRealHeight(), metaData
							.getCoordinateReferenceSystem(), null);
			this.metaData = metaData;
			this.rootID = rootID;
			this.thisPartition = thisPartition;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getBounds(java.awt.Rectangle)
		 */
		public Rectangle getPartitionBounds() {
			return thisPartition;

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getPartition(java.awt.Rectangle)
		 */
		public DataPartition getPartition(Rectangle partitionBounds) {
			return PartitionUtil.getPartitialGrid2D(this, partitionBounds,
					rootID);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getRootID()
		 */
		public int getRootID() {
			return rootID;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#setPartition(appl.parallel.spmd.split.DataPartition,
		 *      java.awt.Rectangle)
		 */
		public void setPartition(DataPartition partition,
				Rectangle partitionBounds) {
			if (!(partition instanceof WritableGrid))
				throw new UnsupportedOperationException(
						"The partition must be an instance of WritableGrid!");
			PartitionUtil.setPartition(this, (WritableGrid) partition,
					partitionBounds);

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getEmpty(int id)
		 */
		public DataPartition getEmpty(int id) {
			return new WritableGridArrayPartition.Float(this.metaData, id,
					thisPartition);
		}
	}
	
	/**
	 * For Double data types
	 * 
	 * @see WritableGridArray.Double
	 * @author Dominik Appl
	 */
	public static class Double extends WritableGridArray.Float implements
			WritableGridPartition {

		private final int rootID;

		private final Rectangle thisPartition;

		private final RasterMetaData metaData;

		
		public Double(RasterMetaData metaData, int rootID,
				Rectangle thisPartition) {
			super(metaData.getMinX(), metaData.getMinY(), metaData.getWidth(),
					metaData.getHeight(), metaData.getX(), metaData.getY(),
					metaData.getRealWidth(), metaData.getRealHeight(), metaData
							.getCoordinateReferenceSystem(), null);
			this.metaData = metaData;
			this.rootID = rootID;
			this.thisPartition = thisPartition;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getBounds(java.awt.Rectangle)
		 */
		public Rectangle getPartitionBounds() {
			return thisPartition;

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getPartition(java.awt.Rectangle)
		 */
		public DataPartition getPartition(Rectangle partitionBounds) {
			return PartitionUtil.getPartitialGrid2D(this, partitionBounds,
					rootID);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getRootID()
		 */
		public int getRootID() {
			return rootID;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#setPartition(appl.parallel.spmd.split.DataPartition,
		 *      java.awt.Rectangle)
		 */
		public void setPartition(DataPartition partition,
				Rectangle partitionBounds) {
			if (!(partition instanceof WritableGrid))
				throw new UnsupportedOperationException(
						"The partition must be an instance of WritableGrid!");
			PartitionUtil.setPartition(this, (WritableGrid) partition,
					partitionBounds);

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getEmpty(int id)
		 */
		public DataPartition getEmpty(int id) {
			return new WritableGridArrayPartition.Double(this.metaData, id,
					thisPartition);
		}
	}
	
	/**
	 * For Integer data types
	 * 
	 * @see WritableGridArray.Integer
	 * @author Dominik Appl
	 */
	public static class Integer extends WritableGridArray.Integer implements
			WritableGridPartition {

		private final int rootID;

		private final Rectangle thisPartition;

		private final RasterMetaData metaData;

		public Integer(RasterMetaData metaData, int rootID,
				Rectangle thisPartition) {
			super(metaData.getMinX(), metaData.getMinY(), metaData.getWidth(),
					metaData.getHeight(), metaData.getX(), metaData.getY(),
					metaData.getRealWidth(), metaData.getRealHeight(), metaData
							.getCoordinateReferenceSystem(), null);
			this.metaData = metaData;
			this.rootID = rootID;
			this.thisPartition = thisPartition;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getBounds(java.awt.Rectangle)
		 */
		public Rectangle getPartitionBounds() {
			return thisPartition;

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getPartition(java.awt.Rectangle)
		 */
		public DataPartition getPartition(Rectangle partitionBounds) {
			return PartitionUtil.getPartitialGrid2D(this, partitionBounds,
					rootID);
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getRootID()
		 */
		public int getRootID() {
			return rootID;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#setPartition(appl.parallel.spmd.split.DataPartition,
		 *      java.awt.Rectangle)
		 */
		public void setPartition(DataPartition partition,
				Rectangle partitionBounds) {
			if (!(partition instanceof WritableGrid))
				throw new UnsupportedOperationException(
						"The partition must be an instance of WritableGrid!");
			PartitionUtil.setPartition(this, (WritableGrid) partition,
					partitionBounds);

		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see appl.parallel.spmd.split.DataPartition#getEmpty(int id)
		 */
		public DataPartition getEmpty(int id) {
			return new WritableGridArrayPartition.Integer(this.metaData, id,
					thisPartition);
		}
	}

}
