package appl.parallel.data;

import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.io.IOException;
import java.io.Serializable;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.data.LoadingException;
import appl.parallel.client.ClientDataServer;
import appl.parallel.client.DataServer;
import appl.parallel.server.XuluServer;
import appl.parallel.spmd.split.DataPartition;

import schmitzm.data.WritableGrid;

/**
 * Loads partitioned data from a {@link ClientDataServer}. And unloads the data
 * 
 * @author Dominik Appl
 */
public class XuluClientLoader extends AbstractDataHandler implements
		PartitionDataHandler {

	protected InetAddress address;

	protected String namingString;

	private String unloadTarget = null;

	/**
	 * constructs a new {@link XuluClientLoader}. The local IP adress at the
	 * time of construction is later used on serverside for communication (e.g.
	 * unloading).
	 * 
	 * @param rootID
	 *            the id of the data
	 * @param partitionBounds
	 *            the bounds of the partition to be retrieved on server side
	 * @param unloadBounds
	 *            the bounds of the partition which is to be uploaded to the
	 *            client after calculation
	 */
	public XuluClientLoader(int rootID, ClientDataServer client,
			Rectangle partitionBounds, Rectangle unloadBounds) {
		super(rootID, client, partitionBounds, unloadBounds);
		address = null;
		namingString = null;
		try {
			this.address = InetAddress.getLocalHost();
			unloadTarget = address.getHostAddress();
			namingString = "rmi://" + unloadTarget + "/DataServer";
		} catch (UnknownHostException e) {
			LOG.error("The client IP could not be determined!!", e);
			e.printStackTrace();
		}
	}

	/**
	 * constructs a new {@link XuluClientLoader}. The unloading is done to the
	 * {@link DataServer} at the machine given with the parameter unloadTarget
	 * (IP/Adress or Hostname).
	 * 
	 * @param rootID
	 *            the id of the data
	 * @param partitionBounds
	 *            the bounds of the partition to be retrieved on server side
	 * @param unloadBounds
	 *            the bounds of the partition which is to be uploaded to the
	 *            client after calculation
	 */
	public XuluClientLoader(int rootID, ClientDataServer client,
			Rectangle partitionBounds, Rectangle unloadBounds,
			String unloadTarget) {
		spmdClient = client;
		this.unloadTarget = unloadTarget;
		address = null;
		namingString = null;
		try {
			this.address = InetAddress.getLocalHost();
			namingString = "rmi://" + unloadTarget + "/DataServer";
		} catch (UnknownHostException e) {
			LOG.error("The client IP could not be determined!!", e);
			e.printStackTrace();
		}
		this.rootID = rootID;
		this.partitionBounds = partitionBounds;
		this.unloadBounds = unloadBounds;
	}

	public XuluClientLoader newInstance(int rootID, ClientDataServer client,
			Rectangle partitionBounds, Rectangle unloadBounds) {
		return new XuluClientLoader(rootID, client, partitionBounds,
				unloadBounds);
	}

	/**
	 * Used for deserialization. Do not use for other purposes
	 */
	public XuluClientLoader() {
		super();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#getLoadInfo()
	 */
	public String getLoadInfo() {
		return "XuluClientLoader tries to load data with rootID " + rootID
				+ " from IP: " + address.getHostAddress() + " with naming: "
				+ namingString;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#getLoadInfo()
	 */
	public String getUnloadInfo() {
		return "XuluClientLoader tries to unload data with rootID " + rootID
				+ " to IP: " + address.getHostAddress() + " with naming: "
				+ namingString;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataLoader#load()
	 */
	public DataPartition load() throws LoadingException {
		try {
			LOG.debug(getLoadInfo());
			DataServer client = (DataServer) Naming.lookup(namingString);
			data = client.getPartition(rootID, partitionBounds);
			LOG.debug("Partition with id  " + rootID
					+ " successfully retrieved from client");
			return data;
		} catch (Exception e) {
			LOG.error("Partition retrieval failed: " + e.getMessage());
		}
		return null;
	}

	/**
	 * Initializes the {@link Logger} after deserialization (used by
	 * {@link java.io.Serializable}
	 * 
	 * @author Dominik Appl
	 * @param s
	 * @throws IOException
	 * @see Serializable
	 */
	private synchronized void readObject(java.io.ObjectInputStream s)
			throws IOException, ClassNotFoundException {
		// reading standard fields
		s.defaultReadObject();
		// initialize Logger
		LOG = LogManager.getLogger(this.getClass().getName());
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.data.DataUnloader#unload()
	 */
	public void unload() {
		try {
			if (LOG.isDebugEnabled())
				LOG.debug(getUnloadInfo());
			if (data == null) {
				if (LOG.isDebugEnabled())
					LOG.debug("No data loaded -> nothing to unload");
				return;
			}
			// construct return grid only if necessary
			DataPartition returnPartition = null;
			if (unloadBounds.equals(data.getPartitionBounds()))
				returnPartition = data;
			else
				returnPartition = data.getPartition(unloadBounds);
			// if there is a local spmdClient try to use it
			if (spmdClient != null)
				spmdClient.updatePartition(rootID, returnPartition,
						unloadBounds);
			// else connect remotely to client
			else {
				DataServer client = (DataServer) Naming.lookup(namingString);
				client.updatePartition(rootID, returnPartition, unloadBounds);
			}

			LOG.debug("Partition with id  " + rootID
					+ " was transfered back to client!");
		} catch (Exception e) {
			LOG.error("Backtransfer of data to client failed!", e);
		}
	}

}
