package appl.parallel.data.splittable;

import java.io.File;

import appl.data.DataProxy;
import appl.data.WritableGridLLProxy;
import appl.ext.XuluConfig;
import appl.util.RasterUtil;

import schmitzm.io.IOUtil;
import schmitzm.data.WritableGridArray;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyReadAccess;

import edu.bonn.xulu.io.Factory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.appl.XuluRegistry;

import edu.bonn.xulu.plugin.data.grid.GridList;
import edu.bonn.xulu.plugin.io.grid.array.WritableGridArrayFactory_ArcInfoAsciiGrid;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link GridList} aus/in Dateien im ArcInfo-ASCII-Grid-Format.
 * Beim Import liefert die Factory eine Standard-Arrays basierendes Raster-Liste
 * ({@link WritableGridArray}).
 * 
 * 
 * @author Dominik Appl
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridListFactory_ArcInfoAsciiGrid extends AbstractFactory implements ImportFactory,ExportFactory {
 private static final ImportFactory importFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();
 private static final ExportFactory exportFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link GridList GridList.class}
   */
  public Class getImportType() {
    return GridList.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File[].class}
   */
  public Class getImportSourceType() {
     return File[].class;
  }

  /**
   * Importiert Raster aus Dateien im ArcInfoAsciiGrid-Format
   * und fuegt diese in einer {@link GridList} zusammen.
   * @param input Eingabe-Quelle (muss ein {@link File File[]} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link GridList} erzeugt, in
   *              die das Objekt importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Dateien angegeben werden
   */
  public GridList importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    GridList sg = (GridList)instFac.newInstance(false);
    // Raster importieren und in GridList einfuegen
    for (int i=0; i<((Object[])input).length; i++){
    	SplittableLLProxyGrid grid =  new SplittableLLProxyGrid(importFac,
        		RasterUtil.getRasterMetaData_from_ArcGridASCII_File(((File[])input)[i]), 
        		((File[])input)[i], 
        		reg );
    	//read and apply settings from the registry
        String unloadFolder = XuluConfig.getXuluConfig().getProperty("Datatypes.proxy.unloadFolder");
        boolean unloading = XuluConfig.getXuluConfig().getBooleanProperty("Datatypes.proxy.enableUnloading");
        if(unloadFolder!=null)
        	grid.setUnloadDir(unloadFolder);
        grid.setUnloading(unloading);
    	sg.addGrid(  grid  );
    }
//    // Dateinamen in die Bezeichnung aufnehmen
//    sg.setDescription(sg.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return sg;
  }
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um eine {@link GridList} handelt, die nur Instanzen
   * von {@link WritableGridArray} als Elemente enthaelt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    if ( obj==null || !(obj instanceof GridList) )
      return false;
    GridList list = (GridList)obj;
    for (int i=0; i<list.getGridCount(); i++)
      if ( !(list.getGrid(i) instanceof SplittableLLProxyGrid) )
        return false;
    return true;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link GridList} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return GridList.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt. Dabei handelt es sich um eine <b>einzelne Datei</b>, da
   * die die letztendlichen Export-Dateien automatisch durchnummeriert werden!!
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link GridList} in Dateien des
   * ArcInfoAsciiGrid-Formats.<br>
   * <b>Beachte:</b><br>
   * Als Ausgabe-Objekt ist eine <b>einzelne</b> Datei anzugeben, kein
   * Array! Die Dateien werden ausgehend von dem angegebenen Dateinamen
   * durchnummeriert!!
   * @param object zu exportiertende Raster (muss eine {@link GridList} sein
   * @param output Export-Ziel (muss ein einzelner {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            einzelne Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Objekt
   *            nicht um eine {@link GridList} handelt
   */
  public void exportObject(Object object, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(object) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(object.getClass().getName()));

    // Dateinamen-Teile ermitteln
    String fileName = IOUtil.getBaseFilePath((File)output);
    String fileExt  = IOUtil.getFileExt((File)output);

    // Grids in der Liste einzeln exportieren
    ListPropertyReadAccess listAccess = ((ListProperty) (((GridList)object).getProperty(GridList.PROP_GRIDS))).getReadAccess(this);
    for (int i=0; i<listAccess.getCount(); i++) {
      File outFile = new File( fileName.concat("_"+i).concat(fileExt) );
      exportFac.exportObject( ((DataProxy)listAccess.getValue(i)).getProxiedObject() ,outFile );
    }
    listAccess.release();

  }
}
