package appl.parallel.data.splittable;

import java.io.File;

import appl.data.DataProxy;
import appl.data.WritableGridLLProxy;
import appl.ext.XuluConfig;
import appl.util.RasterMetaData;
import appl.util.RasterUtil;

import schmitzm.data.WritableGridArray;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.appl.XuluRegistry;

import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import edu.bonn.xulu.plugin.io.grid.array.WritableGridArrayFactory_ArcInfoAsciiGrid;

// nur fuer Doku

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link SingleGrid} aus/in das ArcInfo-ASCII-Grid-Format.
 * Beim Import liefert die Factory ein auf Standard-Arrays basierendes Raster
 * ({@link WritableGridArray}).
 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @author Dominik Appl
 * @version 1.0
 */
public class SingleGridFactory_ArcInfoAsciiGrid extends AbstractFactory implements ImportFactory, ExportFactory {
  private static final ImportFactory importFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();
  private static final ExportFactory exportFac = new WritableGridArrayFactory_ArcInfoAsciiGrid();
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link SingleGrid SingleGrid.class}
   */
  public Class getImportType() {
    return SingleGrid.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link SingleGrid} aus einer Datei im
   * ArcInfoAsciiGrid-Format.
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link SingleGrid} erzeugt, in
   *              die das Raster importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   */
  public SingleGrid importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    // Aus Registry die passende Factory fuer SingleGrid suchen
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    // Raster importieren und in ein SingleGrid einfuegen
    SingleGrid sg = (SingleGrid)instFac.newInstance(false);
    SplittableLLProxyGrid newGrid = new SplittableLLProxyGrid(importFac, 
            RasterUtil.getRasterMetaData_from_ArcGridASCII_File((File) input),
            (File)input,reg);
    //read and apply settings from the registry
    String unloadFolder = XuluConfig.getXuluConfig().getProperty("Datatypes.proxy.unloadFolder");
    boolean unloading = XuluConfig.getXuluConfig().getBooleanProperty("Datatypes.proxy.enableUnloading");
    if(unloadFolder!=null)
    	newGrid.setUnloadDir(unloadFolder);
    newGrid.setUnloading(unloading);
    sg.setGrid(newGrid);
    // Dateinamen in die Bezeichnung aufnehmen
    sg.setDescription(sg.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return sg;
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link SingleGrid} handelt, das ein
   * {@link WritableGridArray} in seiner Raster-Eigenschaft besitzt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null &&
           obj instanceof SingleGrid &&
           ((SingleGrid)obj).getGrid() instanceof SplittableLLProxyGrid;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link SingleGrid} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return SingleGrid.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link SingleGrid} in eine ArcInfoAsciiGrid-Datei.
   * @param grid zu exportiertendes Raster (muss ein {@link SingleGrid} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Raster
   *            nicht um ein {@link SingleGrid} handelt
   */
  public void exportObject(Object grid, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(grid) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(grid.getClass().getName()));
    exportFac.exportObject( ((DataProxy)((SingleGrid)grid).getGrid()).getProxiedObject(),(File)output );
  }

}
