package appl.parallel.plugin.event;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Calendar;
import java.util.Comparator;
import java.util.GregorianCalendar;

import javax.swing.JTable;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableModel;
import javax.swing.table.TableRowSorter;

import appl.ext.XuluConfig;
import appl.parallel.client.RemoteEventHandler;
import appl.parallel.event.TimeEvent;
import appl.parallel.event.TimeMonitor;
import appl.parallel.event.TransferEvent;
import appl.parallel.event.TransferMonitor;

/**
 * Controls the {@link SimpleCommEventMonitorPlugin}
 * 
 * @author Dominik Appl
 */
public class SimpleCommEventMonitorEngine implements TimeMonitor,
		TransferMonitor, ActionListener {

	private XuluConfig config;

	private boolean timeMonitoring;

	private boolean transferMonitoring;

	private SimpleCommEventMonitor monitor;

	private final RemoteEventHandler eventProxy;

	private DefaultTableModel timeEventModel;

	private DefaultTableModel transferEventModel;

	public SimpleCommEventMonitorEngine(RemoteEventHandler eventProxy) {
		this.eventProxy = eventProxy;
		config = XuluConfig.getXuluConfig();

	}

	/**
	 * @return the GUI for the monitor
	 */
	public SimpleCommEventMonitor getMonitor() {
		if (monitor == null) {
			monitor = new SimpleCommEventMonitor(this);
			initGUI();
		}
		return monitor;
	}

	/**
	 * inits the {@link SimpleCommEventMonitor}
	 */
	public void initGUI() {
		// get Data from XuluConfig:
		timeMonitoring = config
				.getBooleanProperty("SimpleCommEventMonitor.timemonitoring");
		transferMonitoring = config
				.getBooleanProperty("SimpleCommEventMonitor.transfermonitoring");
		// registers itself as timeEvent receiver
		if (timeMonitoring)
			eventProxy.addTimeEventListener(this);
		if (transferMonitoring)
			eventProxy.addTransferEventListener(this);
		initTimeEventTable();
		initTransferEventTable();
		monitor.timeEventEnabled.setSelected(timeMonitoring);
		monitor.transferEventEnabled.setSelected(transferMonitoring);
	}

	/**
	 * 
	 */
	private void initTimeEventTable() {
		JTable table = monitor.timeEventTable;

		timeEventModel = new DefaultTableModel(
				new Object[] { "Receive Time", "Event type", "Event Source",
						"Event target", "Exec time (ms)" }, 0);
		TableRowSorter sorter = new TableRowSorter<TableModel>(timeEventModel);
		sorter.setComparator(4, LONG_COMPARATOR);
		table.setModel(timeEventModel);
		table.setRowSorter(sorter);

	}

	private void initTransferEventTable() {
		JTable table = monitor.transferEventTable;
		transferEventModel = new DefaultTableModel(new Object[] {
				"Receive Time", "Event type", "Event Source", "Event target",
				"Size (kb)" }, 0);
		TableRowSorter sorter = new TableRowSorter(transferEventModel);
		sorter.setComparator(4, LONG_COMPARATOR);
		table.setModel(transferEventModel);
		table.setRowSorter(sorter);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.TimeMonitor#receiveTimeEvent(appl.parallel.event.TimeEvent)
	 */
	public synchronized void receiveTimeEvent(TimeEvent t) {
		timeEventModel.addRow(new Object[] { getCurrentTimeString(),
				t.getType(), t.getSrc(), t.getTarget(),
				(long) (t.getTime() / 1000000) });
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.event.TransferMonitor#receiveTransferEvent(appl.parallel.event.TransferEvent)
	 */
	public void receiveTransferEvent(TransferEvent t) {
		transferEventModel.addRow(new Object[] { getCurrentTimeString(),
				t.getType(), t.getSrc(), t.getTarget(),
				(long) (t.getSize() / 1024) });
	}

	public String getCurrentTimeString() {
		Calendar cal = new GregorianCalendar();
		String hour = String.valueOf(cal.get(Calendar.HOUR_OF_DAY));
		while (hour.length() < 2)
			hour = "0" + hour;
		String min = String.valueOf(cal.get(Calendar.MINUTE));
		while (min.length() < 2)
			min = "0" + min;
		String sec = String.valueOf(cal.get(Calendar.SECOND));
		while (sec.length() < 2)
			sec = "0" + sec;
		String ms = String.valueOf(cal.get(Calendar.MILLISECOND));
		while (ms.length() < 3)
			ms = "0" + ms;
		String time = hour + ":" + min + ":" + sec + ":" + ms;
		return time;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		Object src = e.getSource();
		if (src == monitor.transferEventEnabled) {
			eventProxy.removeTransferMonitor(this);
			if (monitor.transferEventEnabled.isSelected()) {
				eventProxy.addTransferEventListener(this);
				config.setBooleanProperty(
						"SimpleCommEventMonitor.transfermonitoring", true);
			} else
				config.setBooleanProperty(
						"SimpleCommEventMonitor.transfermonitoring", false);
		}
		if (src == monitor.timeEventEnabled) {
			eventProxy.removeTimeMonitor(this);
			if (monitor.timeEventEnabled.isSelected()) {
				eventProxy.addTimeEventListener(this);
				config.setBooleanProperty(
						"SimpleCommEventMonitor.timemonitoring", true);
			} else
				config.setBooleanProperty(
						"SimpleCommEventMonitor.timemonitoring", false);
		}
		if (src == monitor.clearButton) {
			initTimeEventTable();
			initTransferEventTable();
		}
	}

	/**
	 * stops the execution
	 */
	public void stop() {
		eventProxy.removeTimeMonitor(this);
		eventProxy.removeTransferMonitor(this);
	}

	/**
	 * The comparator is used for sorting the values in the {@link JTable}
	 */
	public static final Comparator LONG_COMPARATOR = new Comparator() {
		public int compare(Object o1, Object o2) {
			return ((Long) o1).compareTo((Long) o2);
		}
	};

}
