package appl.parallel.server;

import java.awt.Rectangle;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Vector;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.data.LoadingException;
import appl.parallel.client.DataServer;
import appl.parallel.client.ClientDataServer;
import appl.parallel.data.PartitionDataHandler;
import appl.parallel.event.CommEventSink;
import appl.parallel.event.TimeEvent;
import appl.parallel.event.TransferEvent;
import appl.parallel.event.CommEvent.CommType;
import appl.parallel.services.RemoteEventProxy;
import appl.parallel.spmd.MultiDataInfo;
import appl.parallel.spmd.MultiDataPartitionObject;
import appl.parallel.spmd.SPMDClientController;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.spmd.split.PartitionInfo;
import appl.parallel.spmd.split.SinglePartitionInfo;
import appl.parallel.spmd.split.SplitMap;
import appl.parallel.thread.DataServerThread;

/**
 * Manages the all data for the {@link XuluServer}. This includes retrieval of
 * partitions from connected resources like the {@link DataServer} or, as part
 * of the update of neighborhood regions, from other
 * {@link XuluServer XuluServers}.
 *
 * @author Dominik Appl
 */
public class PartitionDataManager extends UnicastRemoteObject implements
		PartitionDataServer {

	/**
	 * The Thread is used for updating the partitions neighborhood region from
	 * the other servers.
	 *
	 * @author Dominik Appl
	 */
	class UpdateThread extends DataServerThread {

		private final DataPartition thisPartition;

		private final Rectangle updateBounds;

		private final int id;

		private final String hostname;

		public UpdateThread(PartitionDataServer dataServer,
				String ipOrHostname, DataPartition thisPartition,
				Rectangle updateBounds, int id, CommEventSink sink) {
			super(dataServer, null, null, CommType.REMOTE_UPDATE, sink);
			hostname = ipOrHostname;
			// TODO Auto-generated constructor stub
			this.thisPartition = thisPartition;
			this.updateBounds = updateBounds;
			this.id = id;
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see appl.parallel.thread.ExecutionThread#fireTimeEvents(long,
		 *      java.lang.Object)
		 */
		@Override
		protected void fireTimeEvents(long execTime, Object result) {
			try {
				eventSink.fireRemoteEvent(new TimeEvent(execTime,
						dataServerName, hostname, commType));
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see appl.parallel.thread.ExecutionThread#fireTransferEvent(java.lang.Object)
		 */
		@Override
		protected void fireTransferEvent(Object result) {
			try {
				eventSink.fireRemoteEvent(new TransferEvent(dataServerName,
						hostname, commType, new Object[] { result }));
			} catch (RemoteException e) {
				e.printStackTrace();
			}
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see appl.parallel.thread.ExecutionThread#run()
		 */
		@Override
		protected Object run() throws Exception {
			try {
				DataPartition updateData = getServer().getPartition(id,
						updateBounds);
				// apply update:
				this.thisPartition.setPartition(updateData, updateBounds);
				return updateData;
			} catch (Exception e) {
				LOG.error("Error while retrieving partition " + updateBounds
						+ "from PartitionDataServer with ip " + hostname);
			}
			return null;
		}
	}

	private final String bindingName = "PartitionDataServer";

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	/**
	 * all partition data. The key is the id. Is an Hashtable instead of
	 * HashMasp for thread-safty.
	 */
	private Hashtable<Integer, DataPartition> data = new Hashtable<Integer, DataPartition>(
			15);

	/**
	 * the names of the data objects
	 */
	private HashMap<String, Integer> NameToIDMapping = new HashMap<String, Integer>(
			15);

	private HashMap<Integer, SinglePartitionInfo> infos = new HashMap<Integer, SinglePartitionInfo>(
			10);

	private HashMap<String, Object> baseParameters = new HashMap<String, Object>(
			10);

	private final PartitionDataServer[] dataServers;

	private final ExecutorService executor;

	private final String[] IPs;

	private boolean dataServersInitialized = false;

	private HashMap<String, MultiDataInfo> multiDataInfos = new HashMap<String, MultiDataInfo>(
			10);

	private final String dataServerName;

	private final CommEventSink remoteEventReceiver;

	private String[] hostnames;

	private ClientDataServer localSPMDClient;

	private final int registryPort;

	/**
	 * Creates and inits a new PartitionDataserver.
	 *
	 * @param IPs
	 *            the IPs of participating Servers for neighborhood updates(may
	 *            include port specification)
	 * @param remoteEventReceiver
	 *            generated events are forwarded to the receiver
	 * @param registryPort
	 *            port of the registry to which the {@link PartitionDataManager}
	 *            should be bound
	 * @throws RemoteException
	 *             if a connection to a participating server fails
	 */
	public PartitionDataManager(String[] IPs,
			CommEventSink remoteEventReceiver, int registryPort)
			throws RemoteException {
		// create NameToID and IdToInfo-Mapping
		this.IPs = IPs;
		this.remoteEventReceiver = remoteEventReceiver;
		this.registryPort = registryPort;

		dataServers = new PartitionDataServer[IPs.length];
		startTheServer();

		// start Thread service
		executor = Executors.newCachedThreadPool();
		String name = "unknown PartitionDataServer";
		try {
			name = InetAddress.getLocalHost().getHostName();
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		dataServerName = name;
		// try to get the hostnames
		hostnames = new String[IPs.length];
		for (int i = 0; i < IPs.length; i++) {
			// try {
			hostnames[i] = IPs[i];
			// hostnames[i] = InetAddress.getByName(IPs[i]).getHostName();
			// } catch (UnknownHostException e) {
			// LOG.warn("Exception while retrivieving hostname of address: " +
			// IPs[i]);
			// }
		}
	}

	/**
	 * Same as {@link #PartitionDataManager(String[], CommEventSink, int)}, but uses
	 * a local SPMDClient. <br>
	 * This is used when a Server is running inside the Xulu-Client. It has
	 * performance advantages (direct access - no TCP/IP)
	 *
	 * @param IPs
	 *            the IPs of participating Servers for neighborhood updates(may
	 *            include port specification)
	 * @param remoteEventReceiver
	 *            generated events are forwarded to the receiver
	 * @param registryport
	 *            port of the registry to which the {@link PartitionDataManager}
	 *            should be bound
	 * @param localSPMDClient
	 *            a {@link ClientDataServer} for fast local access
	 * @throws RemoteException
	 */
	public PartitionDataManager(String[] IPs,
			RemoteEventProxy remoteEventReceiver, int registryport,
			ClientDataServer localSPMDClient) throws RemoteException {
		this(IPs, remoteEventReceiver, registryport);
		this.localSPMDClient = localSPMDClient;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.client.DataServer#addData(appl.parallel.spmd.split.DataPartition)
	 */
	public void addData(DataPartition partition) throws RemoteException {
		if (partition == null) {
			LOG.warn("To add partition was null - no data added");
			return;
		}
		data.put(partition.getRootID(), partition);

	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#addMultiDataInfos(java.util.HashMap)
	 */
	public void addMultiDataInfos(
			HashMap<String, MultiDataInfo> newMultiDataInfos)
			throws RemoteException {
		this.multiDataInfos.putAll(newMultiDataInfos);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#addPartitionInfos(java.util.Vector)
	 */
	public void addPartitionInfos(
			Vector<SinglePartitionInfo> singlePartitionInfos)
			throws RemoteException {
		for (SinglePartitionInfo singlePartitionInfo : singlePartitionInfos) {
			NameToIDMapping.put(singlePartitionInfo.getBaseResourceName(),
					singlePartitionInfo.getBaseResourceID());
			infos.put(singlePartitionInfo.getBaseResourceID(),
					singlePartitionInfo);
		}
	}

	/**
	 * @param name
	 * @param idx
	 */
	public void destroyMultiPartition(String name, int idx) {

	}

	/**
	 * returns the baseParameter by name
	 *
	 * @param name
	 *            the name of the parameter
	 * @return the parameter
	 */
	public Object getBaseParameter(String name) {
		return baseParameters.get(name);
	}

	/**
	 * Returns a a (remote) PartitionDataServer for an IP-Address. Looks up the
	 * PartitionDataServer in the registry
	 */
	private PartitionDataServer getConnection(String ip) {
		PartitionDataServer server;
		// lookup server
		try {
			server = (PartitionDataServer) Naming.lookup("rmi://" + ip + "/"
					+ bindingName);
		} catch (Exception e) {
			LOG.error("bindig of " + ip + " failed" + e.getMessage(), e);
			e.printStackTrace();
			return null;
		}
		return server;
	}

	/**
	 * Returns the {@link DataPartition} associated with the given ID
	 *
	 * @param id
	 *            the id of the data
	 * @return the data
	 */
	public synchronized DataPartition getData(int id) throws RemoteException {
		// check if in hashtable already
		DataPartition partition = data.get(id);
		if (partition != null)
			return partition;
		// if not in hashtable:
		// load the data and put it into hash
		// find the element with the ID:
		SinglePartitionInfo info = infos.get(id);

		if (info == null) {
			LOG.warn("No partition with id " + id + " found!");
			return null;
		}
		// retrieve data from client or other location
		try {
			// if there is a local server try to get the local partition
			PartitionDataHandler partitionDataHandler = info
					.getPartitionDataHandler();
			if (localSPMDClient != null)
				partitionDataHandler.setSPMDClient(localSPMDClient);
			partition = partitionDataHandler.load();
		} catch (LoadingException e) {
			// TODO Auto-generated catch block
			LOG.error("Loading of partition with id '" + id
					+ "' failed. Null returned!");
			e.printStackTrace();
			return null;
		}
		if (partition == null) {
			LOG.error("Loading of partition with id '" + id
					+ "' failed. Null returned!");
			return null;
		}

		// enter data in partition table
		data.put(id, partition);
		return partition;
	}

	/**
	 * Gets a partition by name
	 *
	 * @param name
	 *            the name of the resource (probably given by the programmer)
	 * @return the partition or null if not found
	 */
	public DataPartition getData(String name) {
		Integer id = NameToIDMapping.get(name);
		if (id == null) {
			LOG.warn("No data for name " + name + " found! Returning null");
			return null;
		}
		try {
			return getData(id);
		} catch (RemoteException e) {
			// should never be reached (because the method call is local
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * Returns the partition info with the specified id
	 *
	 * @param id
	 *            the id of the partition
	 * @return the info
	 */
	public PartitionInfo getInfo(int id) {
		return infos.get(id);
	}

	/**
	 * Gets a partition of a multi-data element. NOTE THAT ALL PARTITIONS OF THE
	 * MULTIGRID WILL BE LOADED (watch out for loading performance!).
	 *
	 * @param name
	 *            the name of the resource (probably given by the programmer)
	 * @return the partition
	 * @see SPMDClientController#addToMultiDataSplitControl(Object[], String)
	 */
	public DataPartition[] getMultiData(String name) {
		// fill an array with the partitions
		DataPartition[] partitions = new DataPartition[multiDataInfos.get(name)
				.getCount()];
		for (int i = 0; i < partitions.length; i++) {
			partitions[i] = getMultiData(name, i);
		}
		return partitions;
	}

	/**
	 * Gets a partition of a multi-data element. Only the resource at the given
	 * position will be loaded.
	 *
	 * @param name
	 *            the name of the resource (probably given by the programmer)
	 * @param pos
	 *            the index of the requested element
	 * @return the partition
	 * @see SPMDClientController#addToMultiDataSplitControl(Object[], String)
	 */
	public DataPartition getMultiData(String name, int pos) {
		try {
			return getData(multiDataInfos.get(name).getMultiID(pos));
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#getMultiDataInfo(java.lang.String)
	 */
	public MultiDataInfo getMultiDataInfo(String name) throws RemoteException {
		return multiDataInfos.get(name);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#getMultiDataObject(java.lang.String)
	 */
	public MultiDataPartitionObject getMultiDataObject(String name)
			throws RemoteException {
		return new MultiDataPartitionObject(multiDataInfos.get(name), this);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#getPartition(int,
	 *      java.awt.Rectangle)
	 */
	public synchronized DataPartition getPartition(int id, Rectangle bounds)
			throws RemoteException {
		DataPartition thisPartition = (DataPartition) data.get(id);
		if (thisPartition == null && LOG.isDebugEnabled()) {
			LOG.error("Partition with id " + id + " and bounds " + bounds
					+ "was requested but not found on DataManager");
			return null;
		}

		if (LOG.isDebugEnabled())
			LOG.debug("Data retrieved from DataManager: ID:" + id
					+ " partition: " + bounds);
		return thisPartition.getPartition(bounds);
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#getPartitionInfo(int)
	 */
	public SinglePartitionInfo getPartitionInfo(int rootID)
			throws RemoteException {
		return infos.get(rootID);
	}

	/**
	 * other dataservers may not be initialized on creation of this object.
	 * Therefore the dataServers must be looked up later
	 */
	private void initializeDataServers() {
		if (dataServersInitialized == true)
			return;
		for (int i = 0; i < dataServers.length; i++) {
			dataServers[i] = getConnection(IPs[i]);
		}
		dataServersInitialized = true;
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.client.DataServer#removeData(int)
	 */
	public void removeData(int id) {
		data.remove(id);

	}

	/**
	 * Removes the partition from the database
	 *
	 * @param name
	 *            the name of the partition to remove
	 */
	public void removeData(String name) {
		data.remove(NameToIDMapping.get(name));
	}

	/**
	 * Removes the partition from the database
	 *
	 * @param partition
	 *            the Partition to remove
	 */
	public void removePartition(DataPartition partition) {
		data.remove(partition.getRootID());
	}

	/**
	 * binds the server to the registry
	 */
	private void startTheServer() throws RemoteException {
		Registry registry = LocateRegistry.getRegistry(registryPort);
		registry.rebind(bindingName, (PartitionDataServer) this);
		LOG.info("PartitionDataServer bound to port " + registryPort
				+ " and running....");
	}

	/**
	 * unbinds the server from the registry, removes all data and runs the
	 * garbage collector
	 */
	public void stop() {
		data.clear();
		NameToIDMapping.clear();
		infos.clear();

		Registry registry;
		try {
			registry = LocateRegistry.getRegistry();
			registry.unbind(bindingName);
			LOG.debug("Sucessfully removed " + bindingName + " from reggie");
		} catch (RemoteException e) {
			LOG.warn("tried to unbind " + bindingName
					+ " from reggie, but an exception occured: "
					+ e.getMessage());
		} catch (NotBoundException e) {
			LOG.warn("tried to unbind " + bindingName
					+ " from reggie, but the name was not bound "
					+ e.getMessage());
		}
		System.gc();
	}

	/**
	 * @param rootID
	 */
	public void unloadToSource(int rootID) throws RemoteException {
		SinglePartitionInfo info = infos.get(rootID);
		if (info == null) {
			LOG.error("Could not unload partition with id " + rootID
					+ ". The partitioninfo was not found!");
			return;
		}
		// start unloading
		info.getPartitionDataHandler().setBasePartition(this.getData(rootID));
		info.getPartitionDataHandler().unload();
		// // remove data from memory and run garbage collection
		// data.remove(rootID);
		// System.gc();
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#updateBaseParameter(java.lang.Object[],
	 *      java.lang.String[])
	 */
	public void updateBaseParameter(HashMap<String, Object> newParameters)
			throws RemoteException {
		// update the mapping. Old values are replaced automaticly
		baseParameters.putAll(newParameters);
	}

	/**
	 * Updates the specified partition. For this the the neighborhood partitions
	 * are updated from remote DataServers using the {@link SplitMap} of the
	 * {@link SinglePartitionInfo}.
	 *
	 * @param id
	 */
	public void updateFromNeighbors(int id) {
		initializeDataServers();
		// get the data to be updated:
		DataPartition thisPartition = data.get(id);
		PartitionInfo info = infos.get(id);
		if (thisPartition == null || info == null) {
			LOG.error("Update with id " + id
					+ " failed, because no data was found with this id");
			return;
		}

		// get some other needed infos
		SplitMap map = info.getSplitMap();
		int mapPosition = info.getSplitMapPos();
		// check if something to do
		if (map.getNeighborhoodRange() == 0) {
			LOG.debug("Update has nothing to do (neighborhoodrange is 0)");
			return;
		}

		// calculate (for each neighbor) the needed partitions and get
		// the updates (implemented using Threads)
		int[] neighbors = map.getNeighborsForPosition(mapPosition);
		Future futures[] = new Future[neighbors.length];
		for (int j = 0; j < neighbors.length; j++) {
			int neighborIdx = neighbors[j];
			// get the overlap of the neighborhood area of the current partion
			// with the calcultation area of the remote partition
			Rectangle updateBounds = map.getPartitionNeighborhoodBounds(
					mapPosition).intersection(
					map.getPartitionCalculationBounds(neighborIdx));
			// Get the partition from remote server:
			DataPartition updateData;
			futures[j] = executor.submit(new UpdateThread(
					dataServers[neighborIdx], hostnames[neighborIdx],
					thisPartition, updateBounds, id, remoteEventReceiver));
		}

		for (Future future : futures) {
			try {
				future.get();
			} catch (InterruptedException e) {
				e.printStackTrace();
			} catch (ExecutionException e) {
				LOG.error("The Remote update of a partition failed", e);
			}
		}

	}

	/**
	 * Does the same as {@link #updateFromNeighbors(int)}
	 *
	 * @param name
	 *            the name of the resource
	 */
	public void updateFromNeighbors(String name) {
		// get by ID
		updateFromNeighbors(NameToIDMapping.get(name));
	}

	/*
	 * (non-Javadoc)
	 *
	 * @see appl.parallel.server.PartitionDataServer#setPartition(int,
	 *      appl.parallel.spmd.split.DataPartition, java.awt.Rectangle)
	 */
	public synchronized void updatePartition(int id, DataPartition partition,
			Rectangle bounds) throws RemoteException {
		DataPartition thisPartition = (DataPartition) data.get(id);
		if (thisPartition == null && LOG.isDebugEnabled()) {
			LOG
					.error("Partition of data with id " + id + " and bounds "
							+ bounds
							+ "should be set but was not found in DataManager");
			return;
		}
		thisPartition.setPartition(partition, bounds);
		if (LOG.isDebugEnabled())
			LOG.debug("Partition of data set on DataManager with id " + id
					+ " and partition bounds: " + bounds);

	}

}
