package appl.parallel.services;

import java.net.MalformedURLException;
import java.rmi.ConnectException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.RemoteException;
import java.rmi.UnknownHostException;
import java.util.Iterator;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.ext.XuluConfig;
import appl.parallel.ComputingResource;
import appl.parallel.ComputingResourceContainer;
import appl.parallel.starter.Starter;
import appl.parallel.starter.client.StarterContainer;
import appl.parallel.starter.server.XuluServerStarter;
import appl.parallel.test.PingTestObject;

/**
 * Very simple Discovery Service that simply looks up hosts from Property <br>
 * <br>
 * <code>DiscoveryServices.hostname.hosts</code> <br>
 * <br>
 * of the {@link XuluConfig} <br>
 * <br>
 * Refreshes the hostlist on every {@link #getResources()}, if <br>
 * <br>
 * <code> DiscoveryServices.hostname.refresh</code> <br>
 * <br>
 * is set to 1, it gets the ips from Property <br>
 * <br>
 * DiscoveryService.hostname.hosts <br>
 * <br>
 * in {@link XuluConfig}
 *
 * @author Dominik Appl
 */
public class HostnameDiscoveryService implements DiscoveryService {
	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	Vector<ComputingResource> res = new Vector<ComputingResource>();

	ExecutorService executor = Executors.newCachedThreadPool();

	private String[] lookupStrings;

	private int timeout;

	private boolean firstRefresh = true;

	/** @return true
	 * @see appl.parallel.services.Service#isRunning()
	 */
	public boolean isRunning() {
		return true;
	}

	/**
	 * @see appl.parallel.services.Service#startService()
	 */
	public void startService() {
		executor = Executors.newCachedThreadPool();
	}

	/**
	 * stops the service
	 */
	public void stopService() {
		executor.shutdownNow();
	}

	/**
	 * tries to (re-)discover every resource in Property <code>HostnameDiscoveryService.hosts</code>
	 * of {@link XuluConfig}. After the time in ms specified in DiscoveryServices.timeout
	 *
	 */
	@SuppressWarnings("unchecked")
	public synchronized void refresh() {
		//get timeout
		initVariables();
		res.addAll(getResourcesForLookupStrings());
	}

	private void initVariables() {
		timeout = XuluConfig.getXuluConfig().getIntProperty(
				"DiscoveryServices.timeout");
		if (timeout == 0) {
			LOG
					.warn("DiscoveryServices.timeout was not set or was 0! Setting connect timeout to 500...");
			timeout = 500;
		}
		//get IPs
		res.clear();
		String[] ips = XuluConfig.getXuluConfig().getMultiProperty(
				"DiscoveryServices.hostname.hosts");
		if (ips.length == 0 || (ips.length == 1 && ips[0].equals(""))) {
			LOG.warn("No Resources found in DiscoveryServices.hostname.hosts");
			ips = new String[0];
		}
		//generate lookupStrings:
		lookupStrings = ips;
	}

	/** gets resources, but pings them first to see if they are alive. Not
	 * 	reachable resouces are removed.
	 *
	 * @see appl.parallel.services.DiscoveryService#getResources()
	 */
	public Vector<ComputingResourceContainer> getResources() {
		if (firstRefresh
				|| XuluConfig.getXuluConfig().getBooleanProperty(
						"DiscoveryServices.hostname.refresh") == true)
			refresh();
		firstRefresh = false;
		return appl.parallel.util.Helper.getResourceContainersForVector(res);
	}

	/**
	 * @return starterContainers for {@link XuluServerStarter}s
	 */
	public Vector<StarterContainer> getStarterContainers() {
		initVariables();
		String[] lookupNames = new String[lookupStrings.length];
		for (int i = 0; i < lookupStrings.length; i++)
			lookupNames[i] = "rmi://" + lookupStrings[i] + "/XuluServerStarter";

		Vector<Starter> discoveredStarters = new Vector<Starter>();
		Vector<StarterContainer> containers = new Vector<StarterContainer>();
		// make a new Thread for each resouce
		Future[] futures = new Future[lookupStrings.length];
		for (int i = 0; i < lookupStrings.length; i++) {
			futures[i] = executor.submit(new LookupCallable(lookupNames[i]));
		}

		// the time at which the result must be there:
		long timeoutTime = System.currentTimeMillis() + timeout;
		for (int i = 0; i < futures.length; i++) {
			// servers previous in iterations may have caused a delay
			// so the remaining time is calculated
			long remainingTime = Math.max(10, timeoutTime
					- System.currentTimeMillis());

			try {
				Starter starter = (Starter) futures[i].get(remainingTime,
						TimeUnit.MILLISECONDS);
				if (starter != null) {
					discoveredStarters.add(starter);
					containers.add(new StarterContainer(starter,
							lookupStrings[i]));
					LOG.debug("Discovered Starter: " + lookupNames[i]);
				}
			} catch (InterruptedException e) {
				LOG.error(e);
				e.printStackTrace();
			} catch (ExecutionException e) {
				LOG.error("Error while connecting to " + lookupNames[i]
						+ " (Connect exception)");
			} catch (TimeoutException e) {
				LOG.warn("Could not find Starter " + lookupNames[i]
						+ ". Timed out after " + timeout + "ms");
				futures[i].cancel(true);
			}
		}
		return containers;
	}

	private Vector<ComputingResource> getResourcesForLookupStrings() {
		initVariables();
		String[] lookupNames = new String[lookupStrings.length];
		for (int i = 0; i < lookupStrings.length; i++)
			lookupNames[i] = "rmi://" + lookupStrings[i] + "/XuluServer";

		Vector<ComputingResource> discoveredResources = new Vector<ComputingResource>(
				lookupNames.length);
		// make a new Thread for each resource
		Future[] futures = new Future[lookupNames.length];
		for (int i = 0; i < lookupNames.length; i++) {
			futures[i] = executor.submit(new LookupCallable(lookupNames[i]));
		}

		// the time at which the result must be there:
		long timeoutTime = System.currentTimeMillis() + timeout;
		for (int i = 0; i < futures.length; i++) {
			// servers previous in iterations may have caused a delay
			// so the remaining time is calculated
			long remainingTime = Math.max(0, timeoutTime
					- System.currentTimeMillis());

			try {
				ComputingResource server = (ComputingResource) futures[i].get(
						remainingTime, TimeUnit.MILLISECONDS);
				discoveredResources.add(server);
				LOG.debug("Discovered resource: " + lookupNames[i]);
			} catch (InterruptedException e) {
				LOG.error(e);
				e.printStackTrace();
			} catch (ExecutionException e) {
				LOG.error("Error while connecting to " + lookupNames[i], e);
			} catch (TimeoutException e) {
				LOG.warn("Could not add server " + lookupNames[i]
						+ ". Timed out after " + timeout + "ms");
				futures[i].cancel(true);
			}
		}
		return discoveredResources;
	}

	/**
	 * This class is used for lookup of remote resources. It handles all errors
	 * and submits them to the LOG
	 *
	 * @author Dominik Appl
	 */
	private class LookupCallable implements Callable {

		private final String lookupString;

		private ComputingResource server = null;

		public LookupCallable(String lookupString) {
			this.lookupString = lookupString;

		}

		/* (non-Javadoc)
		 * @see java.util.concurrent.Callable#call()
		 */
		public Object call() throws Exception {
			try {
				return Naming.lookup(lookupString);
			} catch (UnknownHostException e) {
				LOG.warn("(Unknown Host) Could not find host " + lookupString);
			} catch (java.rmi.NotBoundException e) {
				LOG.warn("(Not Bound) Could not find host " + lookupString);
			} catch (java.rmi.ConnectException e) {
				LOG.warn("(No Connection) Could not connect to host "
						+ lookupString);
			}
			return null;
		}
	}

}
