package appl.parallel.spmd;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.parallel.ComputingResourceContainer;
import appl.parallel.client.ClientDataServer;
import appl.parallel.event.CommEventSink;
import appl.parallel.thread.OneMethodThread;

/**
 * Performance optimizations on client side can be made with this class. Allows
 * heavy multithreading.<br>
 * <br>
 * Merging can be done in a separate thread (parallel to further computation)<br>
 * 
 * @see SPMDClientController for details on the usage of a client controller
 * 
 * @author Dominik Appl
 */
public class AdvancedSPMDClientController extends SPMDClientController
		implements AdvancedSPMDClientInterface {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	// used for synchronization
	private static final Object syncOperationMontitor = new Object();

	/**
	 * For each {@link SyncPoint} a vector of threads associated with the point
	 * is created
	 */
	HashMap<SyncPoint, Vector<OneMethodThread>> vectorMap = new HashMap<SyncPoint, Vector<OneMethodThread>>();

	/**
	 * Empty thread which is used as indicator for a already synchronized sync
	 * point
	 */
	private static final Vector<OneMethodThread> alreadySynchronized = new Vector<OneMethodThread>(
			1);

	/**
	 * same parameters as superclass
	 * 
	 * @see SPMDClientController#SPMDClientController(Vector, double[],
	 *      ClientDataServer, CommEventSink)
	 */
	public AdvancedSPMDClientController(
			Vector<ComputingResourceContainer> computingResources,
			double[] weights, ClientDataServer spmdClient,
			CommEventSink eventProxy) {
		super(computingResources, weights, spmdClient, eventProxy);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergePartition(java.lang.Object,
	 *      appl.parallel.spmd.SyncPoint)
	 */
	public void mergePartition(final Object partition, SyncPoint s,
			final PrintStream stream, final String message) {
		checkSplittable(partition);
		// create a new thread
		OneMethodThread newThread = new OneMethodThread("mergeThread"
				+ s.getId(), s.getPriority()) {
			@Override
			public void run() {
				mergePartition(partition);
				if (stream != null) {
					stream.println(message);
				}
			}
		};
		mapThreadToSyncPoint(newThread, s);
		newThread.start();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergePartition(java.lang.Object,
	 *      appl.parallel.spmd.SyncPoint)
	 */
	public synchronized void mergePartition(Object partition, SyncPoint s) {
		mergePartition(partition, s, null, null);
	}

	/**
	 * Puts a Thread/Syncpoint pair into the Hashmap/Vector structure
	 */
	private void mapThreadToSyncPoint(OneMethodThread thread, SyncPoint s) {
		// access to the syncpoints must be synchronized when working with
		// multithreading
		synchronized (syncOperationMontitor) {
			Vector<OneMethodThread> threads = vectorMap.get(s);
			// if this is the first thread for the syncPoint: create a new
			// Vector
			if (threads == null) {
				threads = new Vector<OneMethodThread>(5);
				vectorMap.put(s, threads);
			}
			threads.add(thread);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergeMultiData(appl.parallel.spmd.MultiDataObject,
	 *      int, appl.parallel.spmd.SyncPoint, java.io.PrintStream,
	 *      java.lang.String)
	 */
	public synchronized void mergeMultiData(MultiDataObject multidata, int idx,
			SyncPoint s, PrintStream stream, String message) {
		// create a new thread
		OneMethodThread newThread = new OneMethodThread("mergeThread-"
				+ s.getId(), s.getPriority(), multidata, idx, stream, message) {
			@Override
			public void run() {
				mergeMultiData((MultiDataObject) getParameter(0),
						(Integer) getParameter(1));
				if (getParameter(2) != null) {
					PrintStream stream = (PrintStream) getParameter(2);
					stream.println(getParameter(3));
				}
			}
		};
		mapThreadToSyncPoint(newThread, s);
		newThread.start();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergeMultiData(appl.parallel.spmd.MultiDataObject,
	 *      int, appl.parallel.spmd.SyncPoint)
	 */
	public void mergeMultiData(MultiDataObject multidata, int idx, SyncPoint s) {
		mergeMultiData(multidata, idx, s, null, null);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#synchronizeToSyncPoint(appl.parallel.spmd.SyncPoint)
	 */
	public void synchronizeToSyncPoint(SyncPoint s) {
		// access to the syncpoints must be synchronized when working with
		// multithreading
		synchronized (syncOperationMontitor) {
			// join all threads associated with the given syncPoint
			Vector<OneMethodThread> threads = vectorMap.get(s);
			// Multithreading: check if another thread has already synchronized
			// to
			// the given point
			if (threads == alreadySynchronized) {
				if (LOG.isDebugEnabled())
					LOG.debug("SyncPoint " + s.getId()
							+ " was already synchronized");
				return;
			}
			// if no thread was found: Some error occured.
			if (threads == null) {
				UnsupportedOperationException e = new UnsupportedOperationException(
						"The Thread with the SyncPoint '" + s.getId()
								+ "' does not exist. Synchronisation failed!!");
				LOG.error("The Thread with the SyncPoint '" + s.getId()
						+ "' was not found. Synchronisation failed!!", e);
				throw e;
			}
			for (Iterator iter = threads.iterator(); iter.hasNext();) {
				OneMethodThread thread = (OneMethodThread) iter.next();
				thread.join();
			}
			// indicate as synchronized
			vectorMap.put(s, alreadySynchronized);
		}
	}
}
