package appl.parallel.spmd;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.parallel.ComputingResourceContainer;
import appl.parallel.client.DataServer;
import appl.parallel.client.ClientDataServer;
import appl.parallel.event.CommEventSink;
import appl.parallel.server.PartitionDataManager;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.thread.OneMethodThread;

/**
 * Performance optimizations can be made with this class. Allows heavy
 * multithreading. Because multiple threads may use this class concurrently
 * nearly all additional methods of this class are internally synchronized using a static variable.
 * Also allows preloading of partitions using {@link SyncPoint}s.
 * 
 * @author Dominik Appl
 */
public class AdvancedSPMDServerController extends SPMDServerController
		implements AdvancedSPMDServerInterface {

	private final Logger LOG = LogManager.getLogger(this.getClass().getName());

	private static final HashMap<String, String> partitionPreloadStatus = 
		                                                    new HashMap<String, String>();

	// private static final HashMap<Integer,String> multiPartitionPreloadStatus
	// = new HashMap<Integer, String>();
	/**
	 * For each {@link SyncPoint} a vector of threads associated with the id of
	 * a SyncPoint is created. The {@link HashMap} is static so all possible threads use
	 * the same map. Access to this map must be synchronized
	 */
	private static final HashMap<Integer, Vector<OneMethodThread>> vectors =
		                                              new HashMap<Integer, Vector<OneMethodThread>>();

	private boolean isMaster = true;

	/**
	 * same parameters as superclass
	 * 
	 * @see SPMDServerController#SPMDServerController(PartitionDataManager, int)
	 */
	public AdvancedSPMDServerController(PartitionDataManager dataManager,
			int referenceResouceID) {
		super(dataManager, referenceResouceID);
		// TODO Auto-generated constructor stub
	}

	/**
	 * The partition is loaded in a separate thread into the local
	 * {@link PartitionDataManager}<br>
	 * You must specify the partition name and a {@link SyncPoint}.<br>
	 * You can specify a priority for the thread in the {@link SyncPoint}.
	 * <br>
	 * <br>
	 * <b>Preloading will only be done one time independent of the number of 
	 * parallel threads calling this method. (synchronized)<br>
	 * </b>
	 * 
	 * @param partition
	 *            the name of the partition for identification
	 * @param s
	 *            the associated {@link SyncPoint}
	 * @see SPMDServerController#getPartition(String)
	 * @see #isMasterThread()
	 */
	public void preloadPartition(String partition, SyncPoint s) {
		// create a new thread only if some other thread has not already
		// triggered the preloading
		synchronized (vectors) {
			if (partitionPreloadStatus.get(partition) == null) {
				partitionPreloadStatus.put(partition, "inProgress");
				OneMethodThread newThread = new OneMethodThread(
						"getPartitionThread " + s.getId(), s.getPriority(),
						partition) {
					@Override
					public void run() {
						// retrieving the partition will as a sideeffect trigger
						// loading the
						// partition into the PartitionDataManager
						getPartition((String) getParameter(0));
					}
				};
				mapThreadToSyncPoint(newThread, s);
				newThread.start();
			}
		}
	}

	/**
	 * The partition is loaded in a separate thread into the local
	 * {@link PartitionDataManager} You must specify the partition name and a
	 * {@link SyncPoint}. You can specify a priority for the thread in the
	 * {@link SyncPoint} <br>
	 * <b>Preloading will only be done one time independent of the number of 
	 * parallel threads calling this method. (synchronized)<br>
	 * </b>
	 * 
	 * @param multiPartition
	 *            the name of the partition for identification
	 * @param idx
	 *            the index of the partition in the multidata object
	 * @param s
	 *            the associated {@link SyncPoint}
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergePartition(java.lang.Object,
	 *      appl.parallel.spmd.SyncPoint)
	 * @see SPMDServerController#getPartition(String)
	 */
	public void preloadMultiPartition(String multiPartition, int idx,
			SyncPoint s) {
		synchronized (vectors) {
			// create a new thread only if some other thread has not already
			// triggered the preloading
			if (partitionPreloadStatus.get(multiPartition
					+ MultiDataInfo.MULTISPLITCONSTANT + idx) == null) {
				partitionPreloadStatus.put(multiPartition
						+ MultiDataInfo.MULTISPLITCONSTANT + idx, "inProgress");

				OneMethodThread newThread = new OneMethodThread(
						"Preload thread - getMultiPartition SID:" + s.getId(),
						s.getPriority(), multiPartition, idx) {
					@Override
					public void run() {
						// retrieving the partition will also trigger loading
						// the
						// partition into the PartitionDataManager
						getMultiPartition((String) getParameter(0),
								(Integer) getParameter(1));
					}
				};
				mapThreadToSyncPoint(newThread, s);
				newThread.start();
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDServerInterface#synchronizeToSyncPoint(appl.parallel.spmd.SyncPoint)
	 */
	public void synchronizeToSyncPoint(SyncPoint s) {
		// join all threads associated with the given syncPoint
		synchronized (vectors) {
			Vector<OneMethodThread> threads = vectors.get(s.getId());
			if (threads == null) {
				UnsupportedOperationException e = new UnsupportedOperationException(
						"The Thread with the SyncPoint '"
								+ s.getId()
								+ "' was not found or was earlier synchronized. Synchronisation failed!!");
				LOG.error("The Thread with the SyncPoint '" + s.getId()
						+ "' was not found. Synchronisation failed!!", e);
				throw e;
			}
			for (Iterator iter = threads.iterator(); iter.hasNext();) {
				OneMethodThread thread = (OneMethodThread) iter.next();
				thread.join();

			}
		}
	}

	/**
	 * You should call this method from time to time to clear all syncpoints and
	 * the associated threads out of the memory
	 */
	public void clearSyncData() {
		synchronized (vectors) {
			vectors.clear();
			partitionPreloadStatus.clear();
		}
	}

	/**
	 * Puts a thread/Syncpoint pair into the Hashmap/Vector structure
	 */
	private static void mapThreadToSyncPoint(OneMethodThread thread, SyncPoint s) {
		Vector<OneMethodThread> threads = vectors.get(s.getId());
		// if this is the first thread for the syncPoint: create a new
		// Vector
		if (threads == null) {
			threads = new Vector<OneMethodThread>(5);
			vectors.put(s.getId(), threads);
		}
		threads.add(thread);

	}

	/**
	 * If there are multiple processors on the machine and the current task
	 * {@link SPMDTask#supportsMultiThreading() supports multithreading} there
	 * are several threads created. Each thread uses a different
	 * {@link #getLocalBounds(DataPartition) calculation area}. Now there might
	 * be parts of the task code were multithreading is undesired, e.g. when
	 * operations affect the whole partition. In this case you should use this
	 * method to guarantee that only one thread has access to a code block.
	 * 
	 * @return whether this thread is the "master" thread, which means whether
	 *         this thread refers to the first task created.
	 */
	public boolean isMasterThread() {
		return isMaster;

	}

	/**
	 * @param isMaster
	 *            true, if this should be the master thread
	 * @see #isMasterThread()
	 */
	public void setMasterThread(boolean isMaster) {
		this.isMaster = isMaster;
	}
}
