package appl.parallel.spmd;

import java.util.Iterator;
import java.util.Vector;

import appl.parallel.server.PartitionDataManager;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.thread.OneMethodThread;

/**
 * Performance optimizations can be made with this type of controller. It allows heavy
 * multithreading. Also allows preloading of partitions using SyncPoints.
 * 
 * @author Dominik Appl
 */
public interface AdvancedSPMDServerInterface extends SPMDServerInterface {
	/**
	 * The partition is loaded in a separate thread into the local {@link PartitionDataManager}
	 * You must specify the partitionname and a {@link SyncPoint}.
	 * You can specifiy a priority for the thread in the {@link SyncPoint}
	 * 
	 * @param partition the name of the partition for identification
	 * @param s the associated {@link SyncPoint}
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergePartition(java.lang.Object, appl.parallel.spmd.SyncPoint)
	 * @see SPMDServerController#getPartition(String)
	 * 
	 */
	public void preloadPartition(String partition, SyncPoint s);
	
	/**
	 * The partition is loaded in a separate thread into the local {@link PartitionDataManager}
	 * You must specify the partitionname and a {@link SyncPoint}.
	 * You can specifiy a priority for the thread in the {@link SyncPoint}
	 * 
	 * @param partition the name of the partition for identification
	 * @param idx the index of the partition in the multidata object
	 * @param s the associated {@link SyncPoint}
	 * 
	 * @see appl.parallel.spmd.AdvancedSPMDClientInterface#mergePartition(java.lang.Object, appl.parallel.spmd.SyncPoint)
	 * @see SPMDServerController#getPartition(String)
	 * 
	 */
	public void preloadMultiPartition(String partition, int idx, SyncPoint s);
	
	
	/**
	 * Waits until the Thread with the specified SyncPoint finishes. After that
	 * the Thread is removed. A second call will cause a exeception.
	 * 
	 * @param s
	 *            the {@link SyncPoint}
	 * @throws UnsupportedOperationException
	 *             if a thread with the specified SyncPoint was not found or was
	 *             already synchronized!
	 */
	public void synchronizeToSyncPoint(SyncPoint s); 
	
	/**
	 * If there are multiple processors on the machine and the current task
	 * supports
	 * {@link SPMDTask#supportsMultiThreading() supports multithreading} there
	 * are several threads created. Each thread uses a different
	 * {@link SPMDServerInterface#getLocalCalculationBounds(DataPartition) calculation area}. Now there might
	 * be parts of the taskcode were multithreading is undesired, e.g. when
	 * operations affect the whole partition. In this case you should use this
	 * method to guarantee that only one thread has access to a code block.
	 *  
	 * @return whether this thread is the "master" thread, which means whether this thread
	 * 			refers to the first task created.
	 */
	public boolean isMasterThread();
	
	/**
	 * You should call this method from time to time to clear all syncpoints and the associated
	 * threads out of the memory
	 */
	public void clearSyncData();
	
	
}
