package appl.parallel.spmd;

import appl.parallel.ComputingResource;
import appl.parallel.spmd.split.AbstractSplitMap;
import appl.parallel.spmd.split.SplitMap;
import appl.parallel.spmd.split.SplittableResource;
import appl.parallel.spmd.split.AbstractSplitMap.NeighborhoodBoxingMode;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;

/**
 * Using this interface the programmer can access the parallel functionality.
 *
 * @author Dominik Appl
 */
public interface SPMDClientInterface {

	/**
	 * <code>Base Parameters</code> are Objects which are used during the
	 * execution of multiple {@link SPMDTask SPMD-Tasks}. They are transfered
	 * to each of the participating
	 * {@link ComputingResource computing resources} but are not synchronized
	 * with the client and of course in no way splitted. <br>
	 * <br>
	 * <b>Notice:</b> For efficient transport it is recommended that multiple
	 * parameters are transfered at once. To do this use the
	 * {@link #addBaseParameters(Object[], String[])} method.
	 *
	 * @param parameter
	 * @param parameterName
	 * @see SPMDServerController
	 */
	public void addBaseParameter(Object parameter, String parameterName);

	/**
	 * Adds multiple Base-parameters at once. Notice that this is only for
	 * convenience and has no performance advantages over
	 * {@link #addBaseParameter(Object, String)}.
	 *
	 * @param parameters
	 *            an array of parameters
	 * @param parameterNames
	 *            an array of names. The indexes of parameters and parameter
	 *            names must of course match
	 */
	public void addBaseParameters(Object[] parameters, String[] parameterNames);

	/**
	 * All resources of the array are submitted to split-control. This means
	 * they are (virtually) split and send (virtually) to the participating
	 * computing units. The partitions can be requested on server side using the
	 * {@link SPMDServerController#getMultiData(String)} method. It is assumed,
	 * that all resources have the same Dimension. The same {@link SplitMap} and
	 * splitmap Position is used for all elements. used for every element.
	 *
	 * @param splittableResources
	 *            the resource to be splitted
	 * @param name
	 *            the name of the resource
	 * @return the multi-data object can be used for adding elements to the
	 *         multi data
	 * @see SPMDServerController
	 * @throws UnsupportedOperationException
	 *             if splitHeight and splitWidth of the resouces does not match
	 */
	public MultiDataObject addToMultiDataSplitControl(
			Object splittableResources[], String name);

	/**
	 * This method is specialized on handling MultiGrids. In principle the
	 * {@link #addToMultiDataSplitControl(Object[], String)} could be used, but
	 * using this method won't handle updates to the MultiGrid required for
	 * correct display in the XuluModellingPlatform. <p/>
	 * The elements of the Multigrid are split and send to the participating
	 * computing units. The partitions can be requested on server side using the
	 * {@link SPMDServerController#getMultiData(String)} method. It is assumed,
	 * that all resources have the same Dimension. The same {@link SplitMap} and
	 * splitmap position is used for all elements.
	 * 
	 * @param name
	 *            the name of the resource
	 * @return the multi-data object can be used for adding elements to the
	 *         multi data
	 * @see SPMDServerController
	 * @throws UnsupportedOperationException
	 *             if splitHeight and splitWidth of the resources does not match
	 */
	public MultiDataObject addToMultiDataSplitControl(MultiGrid multiGrid,
			String name);

	/**
	 * The resource is submitted to splitControl. This means it is virtually
	 * split and send virtually to the participating computing units. The
	 * partitions can be requested on server side using the
	 * {@link SPMDServerController#getPartition(String)} method. <br>
	 * <br>
	 * <i>What does "virtually" mean? </i>Well what really happens is that the
	 * metadata is stored. When
	 * {@link #runSPMDModelTask(SPMDTask, Object[]) runSPMDModelTask} is run
	 * next time, the <b>metadata</b> (and all
	 * {@link #addBaseParameter(Object, String)  base Parameters} btw.) are
	 * Transfered to the servers. <br>
	 * The "real" splitting will happen implicit, i.e. the participating servers
	 * will request the correct partition.
	 *
	 * @param splittableResource
	 *            the resource to be splitted
	 * @param name
	 *            a name for the resource (you will be able to request the
	 *            resource on server side using this name)
	 * @see SPMDServerController
	 */
	public void addToSplitControl(Object splittableResource, String name);

	/**
	 * Merges all the partitions from the server into the sources, from which
	 * the objects were loaded (which may be the Xulu-Client, but may also be a
	 * database or an other location if this feature is implemented).
	 *
	 * @see #mergeAllPartitions()
	 * @see #addToSplitControl(Object, String)
	 */
	public void mergeAllPartitions();

	/**
	 * Merges all elements of the specified multidata object
	 *
	 * @param multidata
	 *            the multidata to merge
	 */
	public void mergeMultiData(MultiDataObject multidata);

	/**
	 * Merges the element at the specified index
	 *
	 * @param multidata
	 *            the multidata object containing the element to merge
	 * @param idx
	 *            the position of the element to merge
	 */
	public void mergeMultiData(MultiDataObject multidata, int idx);

	/**
	 * The counterpart of {@link #addToSplitControl(Object, String)}.
	 * Merges the partitions from the server into the sources, from which the
	 * objects were loaded (which will be the Xulu-Client in most cases, but may
	 * also be a database at an other location if this feature is implemented).
	 * Uses multithreading.
	 *
	 * @param rootID
	 *            the ID of the source to be merged.
	 * @see #mergeAllPartitions()
	 * @see #addToSplitControl(Object, String)
	 */
	public void mergePartition(int rootID);

	/**
	 * The counterpart of {@link #addToSplitControl(Object, String)}.
	 * Merges the partitions from the server into the sources, from which the
	 * objects were loaded (which will be the Xulu-Client in most cases, but may
	 * also be a database at an other location if this feature is implemented).
	 * Use {@link #mergeAllPartitions()} to merge all partitions at once.
	 *
	 * @param splittableResource
	 *            the resource to be merged. (must be an instance of
	 *            {@link SplittableResource}
	 * @see #mergeAllPartitions()
	 * @see #addToSplitControl(Object, String)
	 */
	public void mergePartition(Object splittableResource);

	/**
	 * Run the given task on with the given parameters. The result is a Object
	 * array with results from all participating servers. The number of elements
	 * can therefore vary in different executions (using different numbers of
	 * servers)
	 *
	 * @param task
	 *            the task to be submitted. There may be certain restrictions of
	 *            the task to be submitted.
	 * @param parameters
	 * @return All results are combined in one array. Should a server not return
	 *         a result, the parallel programmer is responsible for handling
	 *         <code>null</code> values inside the array.
	 * @throws Throwable
	 *             the Exceptions thrown on serverside!
	 */
	public Object[] runSPMDModelTask(SPMDTask task, Object... parameters)
			throws Throwable;

	/**
	 * Here you can set the boxing mode. See {@link NeighborhoodBoxingMode}
	 *
	 * @param boxingMode
	 *            the boxing mode
	 */
	public void setBoxingMode(AbstractSplitMap.NeighborhoodBoxingMode boxingMode);

	/**
	 * Sets the neighborhood range. From now on every splitting is done with the
	 * given range. In most cases it should be called before any resources are
	 * add to split control.
	 *
	 * @param neighborhoodRange
	 *            the width of the neighborhood range (in cells)
	 */
	public void setNeighborhoodRange(int neighborhoodRange);

	/**
	 * All local bounds are calculated on server side using the reference
	 * resource. If {@link SplittableResource splittable resources} of different
	 * sizes or with different {@link SplitMap SplitMaps} are used, it is
	 * possible to get the local bounds of an special resource using the
	 * {@link SPMDServerController#getLocalCalculationBounds(appl.parallel.spmd.split.DataPartition)}
	 * method of the SPMDServerController. <br>
	 * <br>
	 * If no reference resource is set, than the first call to
	 * {@link #addToSplitControl(Object, String)} will set the reference.
	 *
	 * @param splittableResource
	 *            the new reference resource (must be an instance of
	 *            {@link SplittableResource})
	 */
	public void setReferenceResource(Object splittableResource);

	/**
	 * Signals to all participating servers to update the neighborhood and waits
	 * until the update is finished.
	 *
	 * @param splittableResource
	 *            (must be an instance of {@link SplittableResource}
	 */
	public void updateNeighborhood(Object splittableResource);

	/**
	 * Signals to all participating servers to update the neighborhood and waits
	 * until the update is finished.
	 *
	 * @param multiDataObject
	 *            the {@link MultiDataObject} to be updated
	 * @param index
	 *            the index of the partition to be updated
	 */
	public void updateNeighborhood(MultiDataObject multiDataObject, int index);

}
