package appl.parallel.spmd;

import java.awt.Rectangle;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.Vector;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import appl.parallel.ComputingResource;
import appl.parallel.client.DataServer;
import appl.parallel.data.WritableGridArrayPartition;
import appl.parallel.model.AbstractParallelStepModel;
import appl.parallel.server.PartitionDataManager;
import appl.parallel.spmd.split.DataPartition;
import appl.parallel.spmd.split.PartitionInfo;
import appl.parallel.spmd.split.SplitMap;
import appl.parallel.spmd.split.SplittableResource;

/**
 * This class controls all the parallelization action on the server side and is
 * the counterpart to {@link SPMDClientController}. It is accessed by the model
 * developer by retrieving the {@link SPMDServerInterface} from a
 * {@link SPMDTask}.
 * 
 * @author Dominik Appl
 */
public class SPMDServerController implements SPMDServerInterface {

	private Logger LOG = LogManager.getLogger(this.getClass().getName());

	PartitionDataManager dataManager;
	
	private Rectangle localCalculationBounds;

	private final SplitMap referenceSplitMap;

	/**
	 * the partition position (used for splitmaps)
	 */
	protected final int mapPosition;
	
	/**
	 * Creates a new instance 
	 * @param dataManager the partition data server responsible for managing the data on the server
	 * @param referenceResouceID the id of the reference resource
	 */
	public SPMDServerController(PartitionDataManager dataManager,
			int referenceResouceID) {
		this.dataManager = dataManager;
		// if there are any partitions the referenceID should be != -1
		if (referenceResouceID != -1) {
			referenceSplitMap = dataManager.getInfo(referenceResouceID)
					.getSplitMap();
			mapPosition = dataManager.getInfo(referenceResouceID)
					.getSplitMapPos();
		} else {
			referenceSplitMap = null;
			mapPosition = 0;
		}
		//setting the calculation bounds to the default area
		localCalculationBounds=referenceSplitMap.getLocalCalculationBounds(mapPosition);
	}
	
	/**
	 * This method sets new calculation bounds. This is used to divide the rasters for 
	 * parallel computation on one machine (using multiple cpu's or cpu cores)
	 * 
	 * @param newBounds the new calculation bounds
	 */
	public void setLocalCalculationBounds(Rectangle newBounds){
		localCalculationBounds = newBounds;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getBaseParameter(java.lang.String)
	 */
	public Object getBaseParameter(String parameterName) {
		Object returnValue =  dataManager.getBaseParameter(parameterName);
		 if(returnValue==null)
			 LOG.warn("No value for parameter '" + parameterName + "' found or value was null" );
		 return returnValue;
	}

		

	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalculationBounds(appl.parallel.spmd.split.DataPartition)
	 */
	public Rectangle getLocalCalculationBounds(DataPartition partition) {
		return dataManager.getInfo(partition.getRootID()).getSplitMap()
				.getLocalCalculationBounds(mapPosition);
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalBounds(appl.parallel.spmd.split.DataPartition)
	 */
	public Rectangle getLocalBounds(DataPartition partition) {
		PartitionInfo info = dataManager.getInfo(partition.getRootID());
		return info.getSplitMap().getLocalBounds(mapPosition);
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getGlobalBounds(appl.parallel.spmd.split.DataPartition)
	 */
	public Rectangle getGlobalBounds(DataPartition partition) {
		PartitionInfo info = dataManager.getInfo(partition.getRootID());
		return info.getSplitMap().getGlobalBounds();
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalculationBounds()
	 */
	public Rectangle getLocalCalculationBounds() {
		return localCalculationBounds;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalBounds()
	 */
	public Rectangle getLocalBounds() {
		return referenceSplitMap.getLocalBounds(mapPosition);
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getPartition(java.lang.String)
	 */
	public DataPartition getPartition(String name) {
		DataPartition partition = dataManager.getData(name);
		if (partition == null)
			LOG.error("ERROR: The Partition with the name '" + name
							+ "' was not found or was null");
		return partition;
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getMultiPartition(java.lang.String)
	 */
	public DataPartition[] getMultiPartition(String name) {
		DataPartition[] partitions = dataManager.getMultiData(name);
		if (partitions == null)
			LOG.error("ERROR: The Partition with the name '" + name
							+ "' was not found or was null");
//		if(partitions.length>0)
//		  if(partitions[0] instanceof WritableGridArrayPartition.Float){
//			WritableGridArrayPartition[] wgPartitions = new WritableGridArrayPartition[partitions.length];
//			System.arraycopy(partitions, 0, wgPartitions, 0, partitions.length);
//			return (DataPartition[]) wgPartitions;
//		  }
		return partitions;
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getMultiData(java.lang.String)
	 */
	public MultiDataObject getMultiData(String name){
		MultiDataObject multiDataObject=null;
		try {
			multiDataObject = dataManager.getMultiDataObject(name);
			if(multiDataObject==null)
				LOG.warn("No Multidataobject with name '"  + name + "' found. Returning null!");
		} catch (RemoteException e) {
			// Should never be be called (local call)
			e.printStackTrace();
		}
		return multiDataObject;
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getMultiPartition(java.lang.String, int)
	 */
	public DataPartition getMultiPartition(String name, int idx) {
		DataPartition partition = dataManager.getMultiData(name,idx);
		if (partition == null)
			LOG.error("ERROR: The Partition with the name '" + name
							+ "' was not found or was null");
		return partition;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalcMinX()
	 */
	public int getLocalCalcMinX() {
		return (int) getLocalCalculationBounds().getX();
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalcMinY()
	 */
	public int getLocalCalcMinY() {
		return (int) getLocalCalculationBounds().getY();
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalcMaxX()
	 */
	public int getLocalCalcMaxX() {
		return (int) (getLocalCalculationBounds().getX()
				+ getLocalCalculationBounds().getWidth() - 1);
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#getLocalCalcMaxY()
	 */
	public int getLocalCalcMaxY() {
		return (int) (getLocalCalculationBounds().getY()
				+ getLocalCalculationBounds().getHeight() - 1);
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.SPMDServerInterface#destroyData(java.lang.Object)
	 */
	public void destroyPartition(String name) {
		dataManager.removeData(name);
	}
	
	public void destroyMultiPartition(String name,int idx)
	{
		MultiDataObject multiData = this.getMultiData(name);
		multiData.deleteElement(idx);
	}

}
