package appl.parallel.spmd.split;

import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;

/**
 * This class can be used for the implementation of {@link SplitMap}s. It already
 * provides most of the functions. Normally subclasses only have to implement the methods
 * {@link SplitMap#makeMap()} and {@link SplitMap#getNeighborsForPosition(int)}.
 *
 * @see SplitMap
 *
 * @author Dominik Appl
 */
public abstract class AbstractSplitMap implements SplitMap {

	public enum NeighborhoodBoxingMode {
		inBoxing, outBoxing
	}

	protected NeighborhoodBoxingMode boxingMode;
	protected int noOfPartitions;
	protected int globalHeight;
	protected int globalWidth;

	/**
	 * a value of weights determines how big the share of the
	 * partition with this position is. The sum of all values
	 * should be 1 or near to 1.
	 */
	protected double[] weights;
	protected int neighborhoodRange;

	/** the Rectangle of the whole Grid **/
	protected Rectangle globalBounds;
	/** the calculation area of each partition */
	protected Rectangle[] partitionCalculationBounds;
	/** the bounds of the neighborhood data of the partition */
	protected Rectangle[] partitionNeighborhoodBounds;
    protected String[] IPs;

	public AbstractSplitMap(SplittableResource splittable, int neighborhoodRange,
			int noOfPartitions,
			NeighborhoodBoxingMode boxingMode) {
		this(splittable.getSplitWidth(),splittable.getSplitHeight(),neighborhoodRange, noOfPartitions, boxingMode);

	}

	public AbstractSplitMap() {
		this.globalWidth = 1;
		this.globalHeight = 1;
		this.neighborhoodRange = 0;
		this.noOfPartitions = 1;
		this.boxingMode = NeighborhoodBoxingMode.inBoxing;
		init();
		makeMap();
	}

	 /**
	 * @param width
	 * @param height
	 * @param neighborhoodRange
	 * @param noOfPartitions
	 * @param boxingMode
	 */
	public AbstractSplitMap(int width, int height, int neighborhoodRange, int noOfPartitions, NeighborhoodBoxingMode boxingMode) {
		this.globalWidth = width;
		this.globalHeight = height;
		this.neighborhoodRange = neighborhoodRange;
		this.noOfPartitions = noOfPartitions;
		this.boxingMode = boxingMode;
		init();
		makeMap();
	}

	/* non java-doc
	 * @see appl.parallel.spmd.split.SplitMap#setParameters(SplittableResource, int, int, appl.parallel.spmd.split.AbstractSplitMap.NeighborhoodBoxingMode)
	 */
	public void setParameters(SplittableResource splittable, int neighborhoodRange, int noOfPartitions, NeighborhoodBoxingMode boxingMode) {
		this.globalWidth = splittable.getSplitWidth();
		this.globalHeight = splittable.getSplitHeight();
		this.neighborhoodRange = neighborhoodRange;
		this.noOfPartitions = noOfPartitions;
		this.boxingMode = boxingMode;
		init();
	}

	/**
	 * Initializes the bounds of the partitions
	 */
	protected void init() {
		if (noOfPartitions < 1)
			throw new UnsupportedOperationException(
					"Number of Partions must be > 0");

		//Initialize the weight for equal split (should not matter if not excactly 100%)
		weights = new double[noOfPartitions];
		for (int i = 0; i < noOfPartitions; i++)
			weights[i] = (1.0 / noOfPartitions);
		globalBounds = new Rectangle(0, 0, globalWidth, globalHeight);
		partitionCalculationBounds = new Rectangle[noOfPartitions];
		partitionNeighborhoodBounds = new Rectangle[noOfPartitions];

	}



	public void setWeights(double... weights) {
		if (weights.length != noOfPartitions)
			throw new UnsupportedOperationException("The number of weights ("
					+ weights.length
					+ ") does not match the number of partitions ("
					+ noOfPartitions + ")");
		//check if the weights are valid
		for (double weight : weights) {
			if(weight <=0 || weight > 1)
				throw new UnsupportedOperationException("A weight was '" + weight + "' but must be between 0 and 1");
		}
		this.weights = weights;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getDescription()
	 */
	public abstract String getDescription();

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getNeighborsForPosition(int)
	 */
	public abstract int[] getNeighborsForPosition(int pos);

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#makeMap()
	 */
	public abstract void makeMap();


	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getGlobalCalculationBounds()
	 */
	public Rectangle getGlobalCalculationBounds() {
		return globalBounds;
	}



	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getLocalCalculationArea(int)
	 */
	public Rectangle getLocalCalculationBounds(int pos){
		Rectangle localBounds = (Rectangle) partitionCalculationBounds[pos].clone();
		//move relative to the partition bounds
		double xdiff=partitionCalculationBounds[pos].getX()-partitionNeighborhoodBounds[pos].getX();
		double ydiff=partitionCalculationBounds[pos].getY()-partitionNeighborhoodBounds[pos].getY();
		//set to the local coordinates
		localBounds.setLocation((int) xdiff, (int) ydiff);
		return localBounds;
	}


	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getLocalNeighborhoodArea(int)
	 */
	public Rectangle getLocalNeighborhoodBounds(int pos) {
		Rectangle localBounds = (Rectangle) partitionNeighborhoodBounds[pos]
				.clone();
		//move to (0,0)
		localBounds.setLocation(0, 0);
		return localBounds;
	}

	/* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getPartitionCalculationArea(int)
	 */
	public Rectangle getPartitionCalculationBounds(int pos) {
		return partitionCalculationBounds[pos];
	}

	public Rectangle getPartitionNeighborhoodBounds(int pos) {
		return partitionNeighborhoodBounds[pos];
	}


	 /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getGlobalBounds()
	 */
	public Rectangle getGlobalBounds() {
		return globalBounds;
	}

    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getPartitionBounds(int)
	 */
	public Rectangle getPartitionBounds(int pos) {
		return partitionNeighborhoodBounds[pos];
	}

	public Rectangle getLocalBounds(int pos)
	{
		return getLocalNeighborhoodBounds(pos);
	}


    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#getIP(int)
	 */
	public String getIP(int pos) {
	  if(IPs==null)
		  return null;
	  if(IPs.length!=noOfPartitions)
		  return null;
	  return IPs[pos];
	}

    /* (non-Javadoc)
	 * @see appl.parallel.spmd.split.SplitMap#setIPs(java.lang.String[])
	 */
	public void setIPs(String[] IPs) {
       this.IPs = IPs;
	}

    /**
	 * @return the neighborhoodRange
	 */
	public int getNeighborhoodRange() {
		return neighborhoodRange;
	}

	public int getCount(){
		return noOfPartitions;
	}
}
