package appl.parallel.spmd.split;

import java.awt.Rectangle;
import java.io.Serializable;
import java.net.InetAddress;

import appl.parallel.spmd.split.SplittableResource;
import appl.parallel.spmd.split.AbstractSplitMap.NeighborhoodBoxingMode;

/**
 * Defines a split map over a {@link SplittableResource}. A split map is a
 * layout of an actual partitioning. Each partition is associated with a number,
 * ranging from 0 to noOfPartitions-1.
 * <br>
 * The partitions are layouted as rectangles, where each rectangle represents a partition
 * or a neighborhood around a partition. Every partition or neighborhood can be queried.
 * IP-Addresses can be associated with partitions to identify the server on which a partition is located.
 * <br>
 * The interface is largely implementation independent. Multiple implementations exist.
 *
 * You can query the SplitMap for neighbors of a partition.
 *
 * @see SplitMap1DHorizontal
 * @see SplitMap1DVertical
 * @see SplitMap2D
 * @see AbstractSplitMap
 *
 * @author Dominik Appl
 */
public interface SplitMap extends Serializable{

	/**
	 * returns an Array of neighbors for that position in the splitmap.
	 *
	 * @param pos position in splitmap
	 * @return the array
	 *
	 */
	int[] getNeighborsForPosition(int pos);


	/**
	 * With this setter, you can assign each partition a weight, so that
	 * the data you split is not splitted equally, but arcording to
	 * these values. A value of weights specifies how big the share of the
	 * partition with this position is. The sum of all values
	 * should be 1 or near to 1.
	 *
	 * @param weights
	 * @throws UnsupportedOperationException if the number of weights does
	 * not match the number of partitions
	 */
	public void setWeights(double... weights);

	/**
	 * @return a short description of the type of the Splitting
	 */
	public String getDescription();

	/**
	 * @param pos the partition number
	 * @return the bounds of the partiton calculation area in global coordinates
	 */
	public Rectangle getPartitionCalculationBounds(int pos);

	/**
	 * @param pos the partition number
	 * @return the bounds of the partiton neighborhood area in global coordinates
	 */
	public Rectangle getPartitionNeighborhoodBounds(int pos);

	/**
	 * @param pos the partition number
	 * @return  the bounds of the partiton in global coordinates,
	 * 			which should be the same as {@link #getPartitionNeighborhoodBounds(int)}
	 *
	 */
	public Rectangle getPartitionBounds(int pos);


	/**
	 * @return {@link #getGlobalBounds()}
	 */
	public Rectangle getGlobalCalculationBounds();

	/**
	 * @return the bounds of the whole area starting with 0,0 in the topleft corner
	 */
	public Rectangle getGlobalBounds();

	/**
	 * Gets the whole partition bounds (including available neighborhood)
	 * @param pos the partition number
	 * @return the bounds of the partition using local coordinates. (0,0) is the topleft corner.
	 */
	public Rectangle getLocalBounds(int pos);
	/**
	 * @param pos the partition number
	 * @return the calculation bounds of the partition using local coordinates.
	 */
	public Rectangle getLocalCalculationBounds(int pos);
	/**
	 * @param pos the partition number
	 * @return the bounds of the partition using local coordinates. (0,0) is the topleft corner.
	 */
	public Rectangle getLocalNeighborhoodBounds(int pos);


	/**
	 * Sets the given values. Call makeMap() for this changes to take effekt.
	 *
	 * @param splittable the resource on which the splitmap is calculated
	 * @param neighborhoodRange the neighborhoodrange
	 * @param noOfPartitions the number of partitions to be generated
	 * @param boxingMode the boxing mode
	 *
	 */
	public void setParameters(SplittableResource splittable, int neighborhoodRange,
			int noOfPartitions,
			NeighborhoodBoxingMode boxingMode);

	/**
	 * @param IPs a String Array of IPs which are mapped to the partition in the given order.
	 *
	 */
	public void setIPs(String IPs[]);
	/**
	 * @param pos position for this splitmap
	 * @return the IP associated with this position
	 */
	public String getIP(int pos);

	 /**
	 * @return the neighborhoodRange
	 */
	public int getNeighborhoodRange();

	/**
	 * Creates the map
	 */
	public void makeMap();

	/**
	 * @return the number of partitions
	 */
	public int getCount();
}
