package appl.parallel.spmd.split;

import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Rectangle2D;
import java.util.Vector;

import appl.parallel.client.RemoteExecutionController;
import appl.parallel.spmd.split.AbstractSplitMap.NeighborhoodBoxingMode;
import appl.parallel.util.Helper;
import appl.util.RasterMetaData;
import schmitzm.data.WritableGrid;

/**
 * Responsible for splitting a 2D Area (e.g a {@link WritableGrid}) in a 2D
 * fashion. The splitting is irregular, meaning that the rectangles are not
 * equally sized, but partitioned according to the given weights. 
 * Splitting is as follows: <br><br>
 * 1 to 4 Partitions: 1 row <br>
 * 4 to 8 Partitions: 2 rows <br>
 * 9 to 15 Partitions: 3 rows <br>
 * and so on (depending on the square root of the partititons)<br>
 * <br>
 * For partitioning the this class strongly depends on {@link SplitMap1DHorizontal}
 * and {@link SplitMap1DVertical}. First the squareroot of the number of
 * participating resources is taken to determine how many rows should be
 * created. After this the height of the rows is weighted by adding the weight
 * of the resources in that horizontal partition. The Grid is then splitted
 * horizontal. Finally the horizontal partitions are splitted vertical according
 * to the relative weight to each other. This is only a a virtual split (a map
 * of a split). <br>
 * A neighborhood range can be specified to create partitions that overlap each
 * other. This overlapping can be done in two different ways: <br>
 * <br>
 * <b>Inboxing</b>(default):<br>
 * <br>
 * Inboxing means, that the neighborhood area is not part of the calculation
 * area. <br>
 * <br>
 * <b>Outboxing:</b><br>
 * Outboxing means that the neighborhood area is part of the calculation area.
 * <br>
 * <br>
 * Note that this may also be applied to three dimensional data structures (a 2D
 * splitting of a 3D datatype).
 * 
 * @author Dominik Appl
 */

public class SplitMap2D extends AbstractSplitMap implements SplitMap {

	private SplitMap1DHorizontal horizontalMap;

	private SplitMap1DVertical[] verticalMaps;

	public SplitMap2D(int width, int height, int neighborhoodRange,
			int noOfPartitions, NeighborhoodBoxingMode boxingMode) {
		super(width, height, neighborhoodRange, noOfPartitions, boxingMode);
	}

	/**
	 * needed for serialization
	 */
	public SplitMap2D() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.split.AbstractSplitMap#makeMap()
	 */
	public void makeMap() {
		// take the squareroot of the number of participating resources to
		// determine the number of rows:
		int noOfRows = (int) Math.sqrt(noOfPartitions);
		// calculate weights for the rows (sum of the weights of the partitions
		// inside that row)
		double rowWeights[] = new double[noOfRows];
		int noOfCols[] = new int[noOfRows]; // noOfCols per row
		for (int row = 0; row < noOfRows; row++) {
			// calculate no. of cols in the row:
			noOfCols[row] = (noOfPartitions / noOfRows);
			// the last row may contain a greater number of cols
			if (row == noOfRows - 1)
				noOfCols[row] = noOfPartitions
						- ((noOfRows - 1) * noOfCols[row]);
			for (int col = 0; col < noOfCols[row]; col++)
				rowWeights[row] += weights[row * col + col];
		}

		// create the horizontal Splitmap
		horizontalMap = new SplitMap1DHorizontal(globalWidth, globalHeight,
				neighborhoodRange, noOfRows, boxingMode);
		horizontalMap.setWeights(rowWeights);
		horizontalMap.makeMap();

		// for each horizontal partition create a vertical split
		verticalMaps = new SplitMap1DVertical[noOfRows];
		for (int row = 0; row < noOfRows; row++) {
			//notice: all splitmaps start at (0,0), later a conversion must be made
			verticalMaps[row] = new SplitMap1DVertical((int) horizontalMap
					.getGlobalBounds().getWidth(), (int) horizontalMap
					.getPartitionCalculationBounds(row).getHeight(), neighborhoodRange,
					noOfCols[row], boxingMode);
			int ratings[] = new int[noOfCols[row]];
			for (int col = 0; col < noOfCols[row]; col++) {
				{
					// the rating of the col is the weight of the col * 100000
					// (ratings must be positive integers)
					ratings[col] = (int) (weights[row * col + col] * 1000000);

				}

				verticalMaps[row].setWeights(Helper.calculateWeights(ratings));
				verticalMaps[row].makeMap();
			}
		}
		// now we have partitioned the grid. Lets assign the calculation and
		// neighborhoodareas:
		for (int i = 0; i < noOfPartitions; i++) {
			int row = getRowForIdx(i);
			int col = getColForIdx(i);
			partitionCalculationBounds[i] = verticalMaps[row]
					.getPartitionCalculationBounds(col);
			//because all splitmaps start with (0,0) we must move 
			//the rectangles to the right position
			int x = (int) partitionCalculationBounds[i].getX(); //the x value remains unchanged
			partitionCalculationBounds[i].setLocation(
					new Point(x,(int) horizontalMap.getPartitionCalculationBounds(row).getY()));
			// if there is only one partition there are no explicit neighborhood
			// bounds
			if (noOfPartitions == 1)
				partitionNeighborhoodBounds[i] = partitionCalculationBounds[i];
			// else simply extend the Rectangle with the Neighborhoodbounds
			// created calculation Area (@see AbstractSplitMap)
			else
				partitionNeighborhoodBounds[i] = new Rectangle(
						(int) partitionCalculationBounds[i].getX()
								- neighborhoodRange,
						(int) partitionCalculationBounds[i].getY()
								- neighborhoodRange,
						(int) (partitionCalculationBounds[i].getWidth() + 2 * neighborhoodRange),
						(int) (partitionCalculationBounds[i].getHeight() + 2 * neighborhoodRange));

			// cut of the obverlapping sections which are out of the grid:
			partitionNeighborhoodBounds[i] = partitionNeighborhoodBounds[i]
					.intersection(globalBounds);

			// the partitionCalculation bounds were created for inboxing:
			if (boxingMode == NeighborhoodBoxingMode.outBoxing)
				partitionCalculationBounds[i] = partitionNeighborhoodBounds[i];
		}
	}

	/**
	 * @param i
	 * @return
	 */
	private int getColForIdx(int index) {
		int noOfRows = (int) Math.sqrt(noOfPartitions);
		int avgNoOfCols = noOfPartitions / noOfRows;
		int currentRow = getRowForIdx(index);
		int returnValue = (index - currentRow*avgNoOfCols);
		return returnValue;
	}

	private int getRowForIdx(int index) {
		int noOfRows = (int) Math.sqrt(noOfPartitions);
		int avgNoOfCols = noOfPartitions / noOfRows;
		int returnValue = index / avgNoOfCols;
		//special handling for the last row (contains more partitions)
		if(noOfRows==returnValue)
			return noOfRows-1;
		else return returnValue;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.split.SplitMap#getNeighborsForPosition(int)
	 */
	public int[] getNeighborsForPosition(int pos) {
		Vector neighbors = new Vector();
		// simply intersect with all partitions to find the neighbors:
		for (int i = 0; i < noOfPartitions; i++) {
			if (i == pos)
				continue;
			if (partitionNeighborhoodBounds[i]
					.intersects(partitionNeighborhoodBounds[pos]))
				neighbors.add(pos);
		}
		int results[] = new int[neighbors.size()];
		// copy into array
		for (int i = 0; i < results.length; i++) {
			results[i] = (Integer) neighbors.get(i);
		}
		return results;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.spmd.split.AbstractSplitMap#getDescription()
	 */
	@Override
	public String getDescription() {
		return "2D-irregular Splitmap";
	}

}
