package appl.parallel.starter.server;

import java.awt.AWTException;
import java.awt.Image;
import java.awt.MenuItem;
import java.awt.PopupMenu;
import java.awt.SystemTray;
import java.awt.Toolkit;
import java.awt.TrayIcon;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.SequenceInputStream;
import java.rmi.AccessException;
import java.rmi.Naming;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;

import javax.swing.JOptionPane;
import javax.swing.JTextArea;

import schmitzm.swing.TextAreaPrintStream;
import schmitzm.swing.event.PopupMenuListener;

import appl.ext.XuluConfig;
import appl.parallel.ComputingResource;
import appl.parallel.server.XuluServer;
import appl.parallel.starter.Starter;
import appl.parallel.util.Helper;
import appl.util.GeneralUtil;

/**
 * A running instance of the XuluServerStarter waits for a signal to start a
 * {@link XuluServer}. When {@link #start()} is called the Server is started.
 * All output from the new Serverprocess are forwarded to the actual screen (if
 * any). The Server can also be {@link #restart() restarted} or
 * {@link #stop() stopped}. <br>
 * <br>
 * For configuration of the XuluServer the {@link XuluConfig} is used. The
 * following configuration entries are supported <TABLE border="1"
 * <tr>
 * <th>Entry</th>
 * <th>Default</th>
 * <th>Description</th>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.port</td>
 * <td>1099</td>
 * <td> The port which is used for the registry creation (if none is running)</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.javaprogram</td>
 * <td>java</td>
 * <td> The path to the java program</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.codebasedir</td>
 * <td></td>
 * <td> The absolute path to the codebase directory (should be the path to the
 * binaries)</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.securitypolicy</td>
 * <td></td>
 * <td> The path to the security policy</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.furtherjavaarguments</td>
 * <td></td>
 * <td> further JVM arguments</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.classpath</td>
 * <td></td>
 * <td> The libaries needed</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.memorymax</td>
 * <td>64</td>
 * <td> The maximum memory the XuluSever attempts to use (-xmx)</td>
 * </tr>
 * <tr>
 * <td>XuluServerStarter.memorymin</td>
 * <td>4</td>
 * <td> The inital memory the XuluServer uses (-xms)</td>
 * </tr>
 * </TABLE>
 * 
 * @author Dominik Appl
 */
public class XuluServerStarter extends UnicastRemoteObject implements Starter,
		ActionListener {

	private enum Status {
		/** server is not started */
		serverOff,
		/** * Server is started */
		serverStarted,
		/** A client is connected */
		serverActive
	}

	private String bindingName = "XuluServerStarter";

	private int port = 1099;

	private String javaprogram = "java";

	private String codebase = "";

	private String securitypolicy = "";

	private String furtherjavaarguments = "";

	private String xuluserverarguments = "";

	private String classpath = "";

	private String XuluServerString = "appl.parallel.server.XuluServer";

	private Process p;

	// private BufferedReader input;

	private SimpleConsoleOutputThread thread;

	private int xms = 64;

	private int xmx = 128;

	private SequenceInputStream twoStreams;

	private XuluStarterServerGUI starterServerGUI;

	private TextAreaPrintStream textAreaOutput;

	private MenuItem menu_exit;

	private MenuItem menu_start;

	private MenuItem menu_stop;

	private MenuItem menu_showConsole;

	private TrayIcon trayIcon;

	private ComputingResource xuluServer;

	private Image redIcon;

	private Image greenIcon;

	private Image yellowIcon;

	private String cmd;

	/**
	 * @param cmd
	 *            the command to be executed by the starter. If null, the
	 *            XuluConfig is used to create the commandline
	 * @param showGUI
	 *            if false, no GUI/Tray is shown
	 * @throws RemoteException
	 */
	private XuluServerStarter(String cmd, boolean showGUI)
			throws RemoteException {
		super();
		this.cmd = cmd;
		this.bind(bindingName, this, port);
		if (showGUI) {
			redIcon = Toolkit.getDefaultToolkit().getImage(
					"resource/xulu_icon_red.gif");
			greenIcon = Toolkit.getDefaultToolkit().getImage(
					"resource/xulu_icon_green.gif");
			yellowIcon = Toolkit.getDefaultToolkit().getImage(
					"resource/xulu_icon_yellow.gif");
			createGUI();
			createTray();
			setStatus(Status.serverOff);
		}
	}

	/**
	 * Should be called if the status has changed (changes tray icon)
	 * 
	 * @param newStatus
	 *            the new status
	 */
	private void setStatus(Status newStatus) {
		// if tray disabled: return
		if (trayIcon == null)
			return;
		switch (newStatus) {
		case serverOff:
			trayIcon.setImage(redIcon);
			trayIcon.setToolTip("XuluServer is disabled");
			break;
		case serverStarted:
			trayIcon.setImage(greenIcon);
			trayIcon.setToolTip("Xulu Server is ready and waiting for tasks");
			break;
		case serverActive:
			trayIcon.setImage(yellowIcon);
			trayIcon.setToolTip("A client is connected to the XuluServer");
			break;
		}
	}

	/**
	 * creates and inits the tray icon.
	 */
	private void createTray() {

		if (SystemTray.isSupported()) {

			SystemTray tray = SystemTray.getSystemTray();
			Image image = Toolkit.getDefaultToolkit().getImage(
					"temp/xulu_icon.gif");

			// create menu entries
			PopupMenu popup = new PopupMenu();
			menu_exit = new MenuItem("Exit");
			menu_start = new MenuItem("Start Server");
			menu_stop = new MenuItem("Stop Server");
			menu_showConsole = new MenuItem("Show contol window");
			menu_exit.addActionListener(this);
			menu_start.addActionListener(this);
			menu_stop.addActionListener(this);
			menu_showConsole.addActionListener(this);

			popup.add(menu_start);
			popup.add(menu_stop);
			popup.add(menu_showConsole);
			popup.add(menu_exit);
			trayIcon = new TrayIcon(image, "Xulu Server", popup);
			trayIcon.setImageAutoSize(true);
			trayIcon.addMouseListener(new MouseAdapter() {
				// on doubleclick popup the control-window
				public void mouseClicked(MouseEvent e) {
					if (e.getClickCount() > 1) {
						starterServerGUI.setVisible(!starterServerGUI
								.isVisible());
					}
				}
			});

			try {
				tray.add(trayIcon);
			} catch (AWTException e) {
				System.err.println("TrayIcon could not be added.");
				tray = null;
			}

		} else {
			// System Tray is not supported
			System.out.println("System Tray not supported"); 
			trayIcon = null;
		}

	}

	/**
	 * creates the GUI for the Starter
	 */
	private void createGUI() {
		// create new GUI
		starterServerGUI = new XuluStarterServerGUI();
		// register this instance as button listener
		starterServerGUI.button_exit.addActionListener(this);
		starterServerGUI.button_start.addActionListener(this);
		starterServerGUI.button_stop.addActionListener(this);

		// register TextArea as OutputStream
		textAreaOutput = new TextAreaPrintStream(starterServerGUI.consoleArea);
		System.setErr(textAreaOutput);
		System.setOut(textAreaOutput);

	}

	/**
	 * reads the configuration out of the {@link XuluConfig}
	 */
	private void loadConfig() {
		XuluConfig config = XuluConfig.getXuluConfig();
		if (config.getIntProperty("XuluServerStarter.port") != 0)
			port = config.getIntProperty("XuluServerStarter.port");
		if (config.getProperty("XuluServerStarter.javaprogram") != null)
			javaprogram = config.getProperty("XuluServerStarter.javaprogram");
		if (config.getProperty("XuluServerStarter.codebasedir") != null)
			codebase = config.getProperty("XuluServerStarter.codebasedir");
		if (config.getProperty("XuluServerStarter.securitypolicy") != null)
			securitypolicy = config
					.getProperty("XuluServerStarter.securitypolicy");
		if (config.getProperty("XuluServerStarter.furtherjavaarguments") != null)
			furtherjavaarguments = config
					.getProperty("XuluServerStarter.furtherjavaarguments");
		if (config.getProperty("XuluServerStarter.classpath") != null)
			classpath = config.getProperty("XuluServerStarter.classpath");
		if (config.getProperty("XuluServerStarter.memorymax") != null)
			xmx = config.getIntProperty("XuluServerStarter.memorymax");
		if (config.getProperty("XuluServerStarter.memorymin") != null)
			xms = config.getIntProperty("XuluServerStarter.memorymin");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.starter.Starter#restart()
	 */
	public void restart() throws RemoteException {
		this.stop();
		this.start();
	}

	/**
	 * Creates the Java commandline and starts the {@link XuluServer}
	 * 
	 * @see appl.parallel.starter.Starter#start()
	 */
	public void start() throws RemoteException {
		loadConfig();
		try {
			if (p != null) {
				System.out
						.println("Process is already running. Stop process first");
				return;
			}
			String commandline = "\"" + javaprogram + "\"" + " -cp \""
					+ codebase;
			if (!(classpath.equals("")))
				commandline += (";" + classpath);
			commandline += "\" ";
			if (xms != 0)
				commandline += (" -Xms" + xms + "M ");
			if (xmx != 0)
				commandline += (" -Xmx" + xmx + "M ");

			commandline += (" -Djava.rmi.server.codebase=file:///" + codebase
					+ " -Djava.security.policy=" + securitypolicy + " "
					+ furtherjavaarguments + " " + XuluServerString + " " + xuluserverarguments);
			// if there is a user commandline, use this instead
			if (cmd != null)
				commandline = cmd;
			System.out.println("Executing commandline: " + commandline);
			p = Runtime.getRuntime().exec(commandline);
			xuluServer = null;
			// forward the inputstream of the new process to the
			// console/textarea
			InputStream standardInput = p.getInputStream();
			InputStream errorInput = p.getErrorStream();
			thread = new SimpleConsoleOutputThread(standardInput, errorInput);
			thread.startThread();
			// wait until the server is bound to the registry
			Thread.sleep(5000);
			// lookup the new Server
			xuluServer = (ComputingResource) Naming
					.lookup("rmi://localhost/XuluServer");
			setStatus(Status.serverStarted);
			System.out
					.println("The Starter has successfully discovered the XuluServer instance");

		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (NotBoundException e) {
			System.out
					.println("Could not find the Xulu server instance(NotBoundException).\n Creation of the XuluServer failed..please correct the commandline...");

		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see appl.parallel.starter.Starter#stop()
	 */
	public void stop() throws RemoteException {
		try {
			if (twoStreams != null)
				twoStreams.close();
			if (thread != null)
				thread.stopThread();
			// if (input != null)
			// input.close();
			if (p != null)
				p.destroy();
			p = null;
			System.out.println("Process stopped");
			this.setStatus(Status.serverOff);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Starts the XuluServerStarter <br/> The parameter -start can be used to
	 * start the Server at launch of the Starter. <br>
	 * The parameter -cmd <command> will use the given command to try to start
	 * the server. Else the XuluConfig will be used. <br>
	 * The parameter -noGUI forces the starter to not use any GUI functionality.
	 */
	public static void main(String[] args) {
		try {
			System.out
					.println("Use the parameter -help to see all available commandline options");
			String cmd = null;
			boolean noGUI = false;
			boolean start = false;
			// search for command parameter
			for (int i = 0; i < args.length; i++) {
				if (args[i].toLowerCase().equals("-cmd") && i < args.length)
					cmd = args[i + 1];
				if (args[i].toLowerCase().equals("-nogui") && i < args.length)
					noGUI = true;
				if (args[i].toLowerCase().equals("-start"))
					start = true;
				if (args[i].toLowerCase().equals("-help"))
					printHelp();
			}
			XuluServerStarter starter = new XuluServerStarter(cmd, !noGUI);
			System.out.println("XuluServerStarter initialized .... ");
			if (start) {
				starter.start();
			}
			if (!starter.isRunning()) {
				System.out
						.println("Hint: use parameter '-start' to start the server directly");
			}
		} catch (RemoteException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * 
	 */
	private static void printHelp() {
		System.out
				.println("The following commands can be used: \n\n"
						+ "-start can be used to start the Server at launch of the Starter.\n"
						+ "-cmd <command> will use the given command to try to start the server."
						+ " Else the XuluConfig will be used.\n"
						+ "-noGUI doesn't use any GUI functionality.\n"
						+ "-help displays this help");

	}

	public boolean isRunning() throws RemoteException {
		return (p != null);

	}

	/**
	 * Copied from {@link Helper} and modified so that it does not depend on
	 * log4j. Binds the remote object's stub in the registry. Creates a registry
	 * if no running registry is found.
	 * 
	 * @param bindingName
	 *            the name to be used for binding
	 * @param bindingInstance
	 *            an instance of the type to bind to the registry
	 * @param defaultPort
	 *            the default registry port
	 * @throws RemoteException
	 *             if something goes wrong
	 */
	public void bind(String bindingName, Remote bindingInstance, int defaultPort)
			throws RemoteException {
		boolean failed = true;
		Registry registry;
		try {
			registry = LocateRegistry.getRegistry();
			registry.rebind(bindingName, bindingInstance);
			failed = false;
		} catch (AccessException e) {
			System.err.println("Not enough permissions to bind" + bindingName
					+ "to the registry! Adjust your security permission file!");
			e.printStackTrace();
		} catch (RemoteException e) {
			if (e instanceof java.rmi.ConnectException) {
				System.out
						.println("Could not connect to registry! Trying to create new one...");
				try {
					registry = LocateRegistry.createRegistry(defaultPort);
					registry.rebind(bindingName, bindingInstance);
					System.out.println("Registry successfully created at port "
							+ defaultPort);
					failed = false;
				} catch (RemoteException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			} else
				e.printStackTrace();
		}
		if (failed)
			throw new RemoteException("Could not find or create registry!");
	}

	/**
	 * The Thread is used to direct the output of the process to System.out.
	 * Notice that this stream may be redirected to a
	 * {@link TextAreaPrintStream} or any other printstream using
	 * {@link System#setOut(java.io.PrintStream)}
	 * 
	 * @author appl
	 */
	class SimpleConsoleOutputThread implements Runnable {
		private volatile Thread outputThread;

		private boolean exit;

		private final InputStream standardInput;

		private final InputStream errorInput;

		private final BufferedReader input;

		private final SequenceInputStream twoStreams;

		private boolean availableLastTime = false;

		/**
		 * Creates a new thread. The parmeters are the inputstreams which should
		 * be forwarded to System.out
		 * 
		 * @param standardInput
		 * @param errorInput
		 */
		public SimpleConsoleOutputThread(InputStream standardInput,
				InputStream errorInput) {
			this.standardInput = standardInput;
			this.errorInput = errorInput;
			// combine the streams
			twoStreams = new SequenceInputStream(standardInput, errorInput);
			input = new BufferedReader(new InputStreamReader(twoStreams));
		}

		public void startThread() {
			exit = false;
			if (outputThread == null)
				outputThread = new Thread(this);
			outputThread.start();
		}

		public void stopThread() {
			exit = true;
			outputThread.interrupt();
			try {
				input.close();
			} catch (IOException e) {
				System.out.println("Input closed");
				e.printStackTrace();
			}
		}

		public void run() {
			String line;
			while (exit == false) {
				try {
					while (input.ready()) {
						line = input.readLine();
						System.out.println(line);
					}
					Thread.sleep(1000);
					// check if server is active for tray icon change
					if (xuluServer != null)
						if (availableLastTime != xuluServer.isAvailable()) {
							if (availableLastTime)
								setStatus(Status.serverActive);
							else
								setStatus(Status.serverStarted);
							availableLastTime = (!availableLastTime);
						}

				} catch (IOException e) {
					System.out
							.println("Connection to server failed. Input closed");
					e.printStackTrace();
					exit = true;
				} catch (InterruptedException e) {
					if (exit == true)
						System.out.println("Stopped output thread...");
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		if (e.getSource() == starterServerGUI.button_stop
				|| e.getSource() == menu_stop) {
			try {
				this.stop();
			} catch (Exception e1) {
				System.out
						.println("Could not stop the Server. Reason follows: ");
				e1.printStackTrace();
			}
		} else if (e.getSource() == starterServerGUI.button_exit
				|| e.getSource() == menu_exit) {
			// if user does not confirm: return.
			if (JOptionPane
					.showConfirmDialog(
							starterServerGUI,
							"Are you sure that you want to exit the Starter-Application?",
							"Confirm exit", JOptionPane.OK_CANCEL_OPTION) != JOptionPane.OK_OPTION)
				return;
			try {
				this.stop();
			} catch (RemoteException e1) {
			}
			System.exit(1);

		} else if (e.getSource() == starterServerGUI.button_start
				|| e.getSource() == menu_start) {
			try {
				this.start();
			} catch (Exception e1) {
				System.out
						.println("Could not start the Server. Reason follows: ");
				e1.printStackTrace();
			}
		} else if (e.getSource() == menu_showConsole) {
			starterServerGUI.setVisible(true);
		}

	}
}
