package appl.parallel.util;

import java.rmi.AccessException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.util.Vector;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;


import appl.parallel.ComputingResource;
import appl.parallel.ComputingResourceContainer;
import appl.parallel.ComputingResourceProperties;
import appl.parallel.test.PingTestObject;

/**
 * See method description for details.
 *
 * @author Dominik Appl
 */
public class Helper {
	private static final Logger LOG = LogManager.getLogger("appl.parallel.util.Helper");

	/**
	 * 	Bind the remote object's stub in the registry. Creates a registry if no running registry
	 * is found.
	 *
	 * @param bindingName the name to be used for binding (with port specified, if necessary)
	 * @param bindingInstance an instance of the type to bind to the registry
	 * @param registryPort the default registry port
	 * @throws RemoteException if something goes wrong
	 */
	public static void bind(String bindingName, Remote bindingInstance, int registryPort) throws RemoteException{
		boolean failed=true;
		Registry registry;
		try {
			 registry = LocateRegistry.getRegistry(registryPort);
			 registry.rebind(bindingName, bindingInstance);
			 failed = false;
		} catch (AccessException e) {
			System.err.println("Not enough permissions to bind" + bindingName +
					"to the registry! Adjust your security permission file!");
			LOG.error("Not enough permissions to bind" + bindingName +
					"to the registry! Adjust your security permission file!",e);
			e.printStackTrace();
		} catch (RemoteException e) {
			if(e instanceof java.rmi.ConnectException){
				System.out.println("Could not connect to registry! Trying to create new one...");
				try {
					registry = LocateRegistry.createRegistry(registryPort);
					registry.rebind(bindingName, bindingInstance);
					System.out.println("Registry successfully created at port " + registryPort);
					failed = false;
				} catch (RemoteException e1) {
					// TODO Auto-generated catch block
					e1.printStackTrace();
				}
			}
			else
			  e.printStackTrace();
		}
		if(failed)
			throw new RemoteException("Could not find or create registry!");
	}

	/**
	 * Unbinds a a stub from the registry!
	 * @param bindingName
	 */
	public static void unbind(String bindingName){
	Registry registry;
	try {
		registry = LocateRegistry.getRegistry();
		registry.unbind(bindingName);
		LOG.debug("Sucessfully removed " + bindingName + " from reggie");
	} catch (RemoteException e) {
		LOG.debug("tried to unbind " + bindingName
				+ " from reggie, but an exception occured: "
				+ e.getMessage());
	} catch (NotBoundException e) {
		LOG.debug("tried to unbind " + bindingName
				+ " from reggie, but the name was not bound "
				+ e.getMessage());
	}
	}

	/**
	 * Gets the remote resources from all given {@link ComputingResource} objects.
	 * For this all resources are queried. If a ressouce does not respond the
	 * Resource is removed form the Vector(!!!!!) and a warning is given to the
	 * logger.
	 *
	 * @param res
	 */
	public static Vector<ComputingResourceContainer> getResourceContainersForVector(
			Vector<ComputingResource> res) {

	final Vector<ComputingResourceContainer> containers = new Vector<ComputingResourceContainer>();
    ExecutorService executor = Executors.newCachedThreadPool();

	//this is not a for each loop to avoid ConcurrentModificationException form threads

	Future[] results = new Future[res.size()];
	int i = res.size() - 1;
	while (i >= 0) {
			results[i] = executor.submit(new Helper().new SimpleConnectionThread(res,i));
		i--;
	}
	//wait for the threads to finish  /
	//add the responding containers to a new Vector and return it
	for (int j = 0; j < results.length; j++) {
		Object result;
		try {
			result = results[j].get();
			if(result!=null)
				containers.add((ComputingResourceContainer) result);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ExecutionException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	return containers;
}

	/**
	 * if a connection to the given resource is possible the ResourceProperties of this object
	 * are returned. Else NULL is returned.
	 *
	 * @author Dominik Appl
	 */
	public class SimpleConnectionThread implements Callable {

		private final Vector<ComputingResource> containers;

		private final int position;

		public SimpleConnectionThread(
				Vector<ComputingResource> containers, int position) {
			this.containers = containers;
			this.position = position;
		}

		/*
		 * (non-Javadoc)
		 *
		 * @see java.util.concurrent.Callable#call()
		 */
		public Object call() throws Exception {
//			will remain true, if there is a connection error
			boolean error = true;
			ComputingResource r = (ComputingResource) containers.get(position);
			try {
				// throw a exeception in case of a connection error:
				ComputingResourceProperties rp = r.getResourceInformation();
				error = false;
				if (LOG.isDebugEnabled())
					LOG.debug("Getting of resInfo of <"
							+ rp.getProperty("Name") + "> was successfull!");
				return new ComputingResourceContainer(r, rp);
			} catch (Exception e) {
				LOG.warn("A server was not responding .. removing ressouce!");
				if (LOG.isDebugEnabled())
					LOG.debug("Reason was: " + e.getMessage());
			}
			return null;
		}
	}


	/**
	 * Calculates weights out of ratings. A rating is a value >=1. If the rating
	 * is 0 an average rating is assumed. This method weights the ratings
	 * relative to each other so that a weight is a double value between 0 and
	 * 1 and the sum of all calculated weights is 1.
	 *
	 * @param ratings
	 */
	public static double[] calculateWeights(int ratings[]){
		// values with 0 are weightes average. Add all values and count the
		// values ==0
		double sum = 0;
		double nulls = 0;
		for (int i : ratings) {
			sum += i;
			if (i == 0)
				nulls++;
		}
		// for each null add the average
		double average = (sum / (ratings.length - nulls));
		// if there are ONLY nulls define the average seperately
		if (nulls == ratings.length)
			average = 1000.0 / ratings.length;
		sum += average * nulls;

		// calculate weights
		double[] weights = new double[ratings.length];
		for (int i = 0; i < weights.length; i++) {
			if (ratings[i] == 0)
				ratings[i] = (int) average;
			weights[i] = ratings[i] / sum;
		}

		if (true == true) {
			String weightString = "Calculated weights: ";
			for (double d : weights) {
				weightString += (d + " ");
			}
			LOG.debug("Calculated weights: + " + weightString);
			System.out.println(weightString);
		}
		return weights;



	}

}


