package appl.parallel.util;

import java.awt.Rectangle;
import java.awt.image.DataBuffer;

import schmitzm.data.WritableGrid;
import schmitzm.data.WritableGridArray;
import appl.parallel.data.WritableGridArrayPartition;
import appl.parallel.data.WritableGridArrayPartition.Double;
import appl.parallel.data.WritableGridArrayPartition.Float;
import appl.parallel.data.WritableGridArrayPartition.Integer;
import appl.parallel.spmd.split.SplittableResource;
import appl.parallel.spmd.split.WritableGridPartition;
import appl.util.RasterMetaData;

/**
 * See method description for details.
 * 
 * @author Dominik Appl
 */
public class PartitionUtil {

	/**
		 * Returns partition of a {@link WritableGrid} as a {@link WritableGridArray}.
		 * Notice, that the minX and minY values are ignored and will be set to 0
		 * (see {@link WritableGrid} for more information about minX and minY). If
		 * the given {@link WritableGrid} is not an instance of
		 * {@link WritableGridPartition} it is assumed that the topLeft corner of
		 * the source grid is (0,0) (like in standard image processing).<br>
		 * If the partition is a {@link WritableGridPartition} it is assumed that
		 * the the bounds of the two Partition refer to the same coordinate system.
		 * Real coordinates are supposed to reference the South(!)-West Corner.
		 * 
		 * @param sourceGrid
		 *            the grid to be partitioned
		 * @param partitionBounds
		 *            the rectangle describing the partition
		 * @param rootID
		 *            the id to identify the root partition
		 * @return the new partition including the given corners
		 * @see SplittableResource#getRootID()
		 */
		public static WritableGridPartition getPartitialGrid2D(WritableGrid sourceGrid,
				Rectangle partitionBounds, int rootID) {
			
	//				(int) partitionBounds.getX(),
	//				(int) partitionBounds.getY(),
	//				(int) (partitionBounds.getX() + partitionBounds.getWidth() - 1),
	//				(int) (partitionBounds.getY() + partitionBounds.getHeight() - 1));
			
	//		get Metadata of the big Grid
			RasterMetaData bigMeta = new RasterMetaData(sourceGrid);
	
			//create MetaData for the new partitial Grid
			//Height an width of the partitial Grid
			int pHeight = (int) partitionBounds.getHeight();
			int pWidth = (int)partitionBounds.getWidth();
	
			//real coordinates of the partitial Grid (which are referenced by the SW Corner)
			double pX = partitionBounds.getX() * bigMeta.getCellWidth() + bigMeta.getX();
			double pY = (partitionBounds.getY() + partitionBounds.getHeight() - 1) * bigMeta.getCellHeight() + bigMeta.getY();
			double pRealWidth = pWidth * bigMeta.getCellWidth();
			double pRealHeight = pHeight * bigMeta.getCellHeight();
	
			//the MetaData of the new Grid
			RasterMetaData pMetaData = new RasterMetaData(bigMeta.getDataType(),
					pWidth, pHeight, 0, 0, pX, pY, pRealWidth, pRealHeight, bigMeta.getCoordinateReferenceSystem());
	
			//create empty return Grid
			WritableGridPartition pGrid;
			switch(bigMeta.getDataType()){
			case DataBuffer.TYPE_INT:
				pGrid = new WritableGridArrayPartition.Integer(pMetaData,rootID,partitionBounds);	
				break;
			case DataBuffer.TYPE_DOUBLE:
				pGrid = new WritableGridArrayPartition.Double(pMetaData,rootID,partitionBounds);
				break;
			default:
				pGrid = new WritableGridArrayPartition.Float(pMetaData,rootID,partitionBounds);	
			}
			
			//if the sourceGrid is a partition itself then we do not want to to start reading from (0,0)
			//but shifted relative to the global coordinate system. So we calculate now the start values:
			
			//for standard WritableGrids these are simply the bounds of the new partition...
			int startX = (int) partitionBounds.getX();
			int startY = (int) partitionBounds.getY();
			
			// ...but for partitions these values are shifted:
			if(sourceGrid instanceof WritableGridPartition){
				Rectangle srcBounds = ((WritableGridPartition) sourceGrid).getPartitionBounds();
				startX -= srcBounds.getX();
				startY -= srcBounds.getY();
			}
			//read the data into the raster from left to right and downwards
			int x=0,y=0;
			try{
			for (y = 0; y < pHeight; y++)
				for (x = 0; x < pWidth; x++)
					pGrid.setRasterSample(
							sourceGrid.getRasterSample(startX  + x, startY + y),
							x, 
							y);
			}catch (Exception e) {
				System.out.println("stop");
			}
			return pGrid;
		}

	/**
	     * Overwrites the data at the location specified by the {@link Rectangle} with
	     * the given partition-data. If the given baseGrid is NOT an instance
		 * of {@link WritableGridPartition} it is assumed that its the topLeft corner (0,0). 
		 * If it IS a {@link WritableGridPartition} it is assumed that the the
		 * bounds of the two Partition refer to the same coordinate system and the partition
		 * is inserted at the correct absolut position.
	     * 
		 * @param baseGrid the grid in which the data is inserted
		 * @param gridPartition the grid to be inserted
		 * @param partitionBounds the excact location in coordinates of the baseGrid
		 */
		public static void setPartition(WritableGrid baseGrid, WritableGrid gridPartition, Rectangle partitionBounds) {
			//Check if partition and the rectangle fit
			if((gridPartition.getWidth()!=partitionBounds.getWidth()) ||
					(gridPartition.getHeight()!=partitionBounds.getHeight()))
			  throw new UnsupportedOperationException("The partition width/height does not match the rectangle width/height");
			
			//if the baseeGrid is a partition itself then we do not want to to start writing from (0,0)
			//but shifted relative to the global coordinate system. So we calculate now the start values:
			
			//for standard WritableGrids these are simply the bounds of the new partition...
			int startX = (int) partitionBounds.getX();
			int startY = (int) partitionBounds.getY();
			
			// ...but for partitions these values are shifted:
			if(baseGrid instanceof WritableGridPartition){
				Rectangle targetBounds = ((WritableGridPartition) baseGrid).getPartitionBounds();
				startX -= targetBounds.getX();
				startY -= targetBounds.getY();
			}
			//check if the partition fits in
	//		if(((partitionBounds.getX()+partitionBounds.getWidth()-1)>baseGrid.getWidth()  ||
	//				(partitionBounds.getY()+partitionBounds.getHeight()-1)>baseGrid.getHeight()))
	//				System.out.println("X");
				//throw new UnsupportedOperationException("The partition does not fit in the Grid!");
			//write the partition into the Grid
			
			for(int y = 0; y < partitionBounds.getHeight();y++)
				for(int x = 0; x < partitionBounds.getWidth(); x++)
					baseGrid.setRasterSample(
											gridPartition.getRasterSample(x,y),
											x + startX, 
											y + startY);
			
		}

}
