package appl.util.benchmark;

import java.awt.Rectangle;
import java.awt.image.DataBuffer;
import java.util.Random;

import appl.data.LoadingException;
import appl.ext.XuluConfig;
import appl.parallel.data.splittable.SplittableLLProxyGrid;
import appl.util.RasterMetaData;
import appl.util.RasterUtil;
import edu.bonn.xulu.plugin.io.grid.array.WritableGridArrayFactory;

/**
 * The algorithm simply iterates over a grid. The with and height can be
 * specified by the {@link XuluConfig} Property
 * <code>Benchmark.SimpleBench.gridwidth</code>. For each cell the average
 * over a neighborhood region is calculated. The neighborhoodRange can be
 * specified by <code>Benchmark.SimpleBench.neighborhood</code>. If the
 * algorithm reaches the end of the raster it starts again at (0,0). The rating
 * depends on the number of cells visited in the time (ms) specified by
 * <code>Benchmark.SimpleBench.time<code>. The number of cells is divided by
 * <code>Benchmark.SimpleBench.calibrator</code> and then given back as rating.
 * 
 * @author Dominik Appl
 */
public class SimpleBenchmark implements Benchmark {
	private SplittableLLProxyGrid inputGrid;

	int runningTime = 10000;

	private int gridwidth = 1000;

	private int neighborhoodRange;

	private double calibrator = 1;

	public int bench() {
		setUp();

		int cellCount = 0;
		long startTime = System.currentTimeMillis();
		Rectangle partition = new Rectangle(0, 0, inputGrid.getWidth(),
				inputGrid.getHeight());
		boolean exit = false;
		for (int step = 0; exit == false; step++) {
			// the overallsum is simply the sum over all elements
			float overallSum = 0f;
			for (int y = 0; y < inputGrid.getHeight(); y++)
				for (int x = 0; x < inputGrid.getWidth(); x++) {
					if (System.currentTimeMillis() - startTime > runningTime)
						exit = true;
					else
						cellCount++;
					float tmp = inputGrid.getRasterSampleAsFloat(x, y);
					if (!Float.isNaN(tmp))
						overallSum += tmp;
					// the local sum is simply the sum over all elements in the
					// neighborhood
					float localSum = 0;
					// number of cells over which the sum is calculated
					int noOfCells = 0;
					// for each cell: calculate the sum of all neighbors
					for (int y2 = y - neighborhoodRange; y2 <= y
							+ neighborhoodRange; y2++)
						for (int x2 = x - neighborhoodRange; x2 <= x
								+ neighborhoodRange; x2++)
							// check if the coordinates are valid (inside the
							// grid and not NaN)
							if (partition.contains(x2, y2))
								if (!Float.isNaN(inputGrid
										.getRasterSampleAsFloat(x2, y2))) {
									localSum += inputGrid
											.getRasterSampleAsFloat(x2, y2);
									noOfCells++;
								}

				}
		}
		return (int) (cellCount / calibrator);
	}

	/**
	 * The method is used to load grids and to initialize the benchmarks
	 * configuration.
	 */
	private void setUp() {
		// read config
		int width = XuluConfig.getXuluConfig().getIntProperty(
				"Benchmark.SimpleBench.gridwidth");
		if (width != 0)
			this.gridwidth = width;
		// init running time
		int time = XuluConfig.getXuluConfig().getIntProperty(
				"Benchmark.SimpleBench.runningtime");
		if (time != 0)
			this.runningTime = time;
		// init calibrator
		double calibrator = XuluConfig.getXuluConfig().getDoubleProperty(
				"Benchmark.SimpleBench.calibrator");
		if (calibrator != 0)
			this.calibrator = calibrator;
		// init neighborhoodRange
		neighborhoodRange = XuluConfig.getXuluConfig().getIntProperty(
				"Benchmark.SimpleBench.neighborhood");
		// create input Grid
		inputGrid = new SplittableLLProxyGrid(new WritableGridArrayFactory(),
				new RasterMetaData(DataBuffer.TYPE_DOUBLE, gridwidth,
						gridwidth, 0, 0, 0, 0, 50, null));
		// load grid (so late loading has no influence on the calculation)
		try {
			inputGrid.loadData();
		} catch (LoadingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		// fill Grid with random Data
		Random rand = new Random();
		for (int y = 0; y < inputGrid.getHeight(); y++)
			for (int x = 0; x < inputGrid.getWidth(); x++)
				inputGrid.setGridSample(rand.nextDouble(), x, y);
	}

	public static void main(String[] args) {
		System.out.println("Current rating: " + new SimpleBenchmark().bench());
	}
}
