########################################################################
########################################################################
# Skript to calculate logit regression for landcover analysis
# make random subset of input data to minimize computation time 
# input: datafile prepared with CLUS-S file conversion tool
# output: ascii file ready to use as alloc.1 in CLUE-S
# adapt the following parameters:
# - path to files and filenames
# - look if colnames are all acoorect!
# - Subset of data to use (exclude irrelevant predictors and cover classes)
# CHANGES:
#
# Version 1.3:
# - mehr automatismen, um daten anderer Konfiguration zu verarbeiten
#   z.B. Daten in 100m und fr Xulu -> LN muss nicht bei 0 anfangen
# Version 1.2:
# - balanced sampling also for small classes: 
#   equal samples of 0 and 1 (tip from Peter V.)
#
# Version 1.1:
# - extract certain methods to functions (write different values)
# 
# Version 1.0:
# - read data automaticaly from directory and assign colnames from filenames
# 

########################################################################
########################################################################
 

# load libraries
library(MASS)
library(ROCR)
library(sp)

#######################################################################
# read functions
#######################################################################
source(paste(locPath(dest="prog"),"CLUE-S/functions.R", sep=""))

# ---------------------------------------------------------------------
# define path to data
# ---------------------------------------------------------------------                             
setwd(locPath(dest="model"))

outputRegressionFile = "R_coefficients_lc.txt" 
outputRegXulu = "R_coefficients_xulu.txt"
outputRegEval = "R_regression_evaluation_all.txt"

# ---------------------------------------------------------------------
# import landcover data and driving forces into R
# import data are output from CLUE-S conversion tool
# ---------------------------------------------------------------------

# read directly from ASCII GRID
# take care the NO MINUS signs are pressent in file names!!
# replace with underscores
extension <- ".txt"
FileList = list.files(pattern = paste("\\",extension,"$",sep=""))
#FileList = list.files(pattern = "\\.txt$")
FileList = FileList[-which(FileList == "ac_orig_all.asc")]

# read ascii data into dataframe and test if all have same pixel amount
BeninData = read.asciigrid.df(FileList, 1434056 )  #for 250m resolution: 229863

# write colnames (filename) to dataframe
# delete file name extension
FileListN = gsub(extension,"",FileList)
# replace undescore in variable name by .
FileListN = gsub("_", ".",FileListN)
colnames(BeninData)=FileListN

# convert to factor levels
BeninData$soil = factor(BeninData$soil)
BeninData$tpi = factor(BeninData$tpi)

# resort df that lc data are first collumn
#BeninData = BeninData[c(8,1:7,9:length(BeninData))]
BeninData = data.frame(BeninData$lc2000, subset(BeninData, select= -lc2000))
#colnames(BeninData[1]) <- "lc.2000"        # funktiniert nicht!
## muss names(BeninData)[1]="lc.2000" sein! # colname ist Benindata.lc.2000
                                            # alle folgenden stellen ndern
                                            # in glm etc.
  
  

# ---------------------------------------------------------------------
# PRE-FORMATING OF DATA
# ---------------------------------------------------------------------
# create list of colnames
# ---------------------------------------------------------------------

# get number of fields in dataset
numLUClasses <- max(BeninData[[1]], na.rm = T)
# 
LuClasses <- seq(min(BeninData[[1]], na.rm = T) , max(BeninData[[1]], na.rm = T))

# get number of fields in dataset
numData = length(BeninData)

# create dataframe with driving force names and id numbers
# due to factor levels it is not possible to take the clomanes 
# of the dataframe.


dforce_name = vector()
for (i in 2:length(BeninData)) {
  if (is.factor(BeninData[[i]])){
    #print ("factor")
    factor.range <- range(as.numeric(levels(BeninData[[5]])))
    for (j in factor.range[1]+1 : factor.range[2]){ 
      dforce_name = append(dforce_name, paste(colnames(BeninData[i]),j, sep=""))
    }
  }
  else {
  dforce_name = append(dforce_name, colnames(BeninData[i]))
  }
}
dforce_number <- seq(0,length(dforce_name)-1)
dforce <- data.frame(dforce_name, dforce_number)




########################################################################
# Logistic regression with random subset of data
# SELECT BETWEEN THE FOLLLOWING METHODS TO SUBSET DATA11
########################################################################

# ---------------------------------------------------------------------
# fit binary logistic model to lc data
# loop that goes through all landcover types and calculate
# regression parameters with subset of data and stepwise
# regression
# 
# leave out first list in data_benin as this is AgricChange
# ---------------------------------------------------------------------


# start loop with all LC classes
i=1
for (i in LuClasses) {
  print("---------------------------") # leerzeile
  print(paste("calculation on landcover no:", i, sep=" "))

	# balanced sampling of data
	BeninDataS = balanced.sampling(BeninData, i, 25000)  
	
	# load data from previous sampling:
  load(list.files(pattern = paste("BeninDataSampleLc", i,sep="")))
  
  # save sample
	#save(BeninDataS, file=paste("BeninDataSampleLc", i, "_", Sys.Date(), ".Rdata", sep=""), compress=T)
	
  # --------------------------------------------------------------
	# do logitis regression fit
	# --------------------------------------------------------------
  #fit0 <- glm(lu ~ bev2000 + d.road.2000 +d.village  , family=binomial, 
  #  data=subset(BeninDataS, select=c(-BeninData.lc.2000.f,  -d.a.area.1991, -d.f.area.2000)))
  
  
  fit <- assign(paste("fit_lc", i, sep=""), glm(lu ~ . , family=binomial, 
    data=subset(BeninDataS, select=c(-BeninData.lc2000)) ))

  # --------------------------------------------------------------
  # do stepwise selection of coefficients
  # --------------------------------------------------------------   
	fitstep <- assign(paste("step_lc", i, sep=""), stepAIC(fit, direction="both"))
	
	
	# write results to file
  #WriteCoefClue(fit0, outputRegressionFile, i, "25000 samples, only lc0, selected df")
  WriteCoefXulu(fitstep, outputRegXulu, i,)
  #WriteAUC(fitstep, BeninDataS, i)
  
}

#######################################################################
############ plot fitted value as image
####################################################################### 

## plot.fitted(FileList, fitted object, data to predict, numrows(grid), numcols) 

plot.fitted(FileList, fitstep, BeninData)  


#######################################################################
# plot log curve data 
#######################################################################

# ---------------------------------------------------------------------
# real data:
# ---------------------------------------------------------------------
fit1 <- glm(lu ~ CostDist , family=binomial, data=BeninDataS[-1])
coef = fit1$coef

# ---------------------------------------------------------------------
# manual calculation of response values
# ---------------------------------------------------------------------
logistW = exp(coef[1:1]+(coef[2:2]*BeninDataS$DistRoad))/(1 + exp(coef[1:1]+(coef[2:2]*BeninDataS$DistRoad)))
plot(BeninDataS$DistRoad,logistW)
lines(BeninDataS$DistRoad, predict(fit1, type="response"))

# use function predict to calculate response values
p=predict(fit1, type="response")
d1=data.frame(BeninDataS$CostDist, p)
ord=order(d1$p, decreasing = TRUE)
d1=d1[ord,]
plot(d1), type="line")

# ---------------------------------------------------------------------   
#easyest possibility to plot fitted log curve:
# ---------------------------------------------------------------------
plot(BeninDataS$CostDist, fit1$fitted)

d1 = BeninDataS$CostDist
d2 = fit1$fitted
plot(d2)
d = data.frame(d1,d2)
plot(d)

# ---------------------------------------------------------------------
# test with artificial data
# ---------------------------------------------------------------------
fit1 <- glm(lu ~ DistRoad , family=binomial, data=BeninDataS[-1])
d1 = seq(0, 50000, 50)
logistW = exp(coef[1:1]+(coef[3:3]*d1))/(1 + exp(coef[1:1]+(coef[3:3]*d1)))
plot(d1,logistW)
lines(d1, predict(fit1, data.frame(DistRoad = d1, lc = rep(1, length(d1))), type="response"))


#######################################################################
# caculate some performance measures
#######################################################################

# ---------------------------------------------------------------------
# pseudo R McFadden
# ---------------------------------------------------------------------
r = 1-(step$deviance/step$null.deviance)


# ---------------------------------------------------------------------
# calculate ROC curve
# ---------------------------------------------------------------------

pred = prediction(fitted(fitstep),as.vector(fitstep$data$lu))
perf <- performance(pred, measure = "tpr", x.measure = "fpr")

# magnify graph
par.olf <- par(cex=1.5)
plot(perf, col="red", lwd=3)
par(par.old)

# ---------------------------------------------------------------------
# compute AUC
# ---------------------------------------------------------------------

auc = performance(pred, measure="auc")
# auc is object of class performance, data stored in slots
# to access slots (and datas) use @ (in Newsletter 4/1 is said don't use it
# but I don't know an other way
aucValue = auc@y.values[[1]]
print(aucValue)
write (paste("AUC for",colnames(BeninData[1]), ":", aucValue,sep=" "), outputRegEval)


