/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu;

import javax.swing.UIManager;
import java.util.Map;
import java.util.HashMap;
import java.util.Vector;
import java.util.TreeSet;
import java.util.SortedSet;
import java.util.Locale;
import java.io.FileInputStream;
import java.io.File;
import java.io.PrintStream;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.lang.SortableVector;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.SwingWorker;
import schmitzm.swing.ExceptionDialog;
import schmitzm.geotools.GTUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.lang.DefaultComparator;
import schmitzm.geotools.io.GeoImportUtil;

import edu.bonn.xulu.gui.XuluMainFrame;
import edu.bonn.xulu.gui.XuluStartingFrame;
import edu.bonn.xulu.appl.XuluComponent;
import edu.bonn.xulu.appl.XuluComponentUtil;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.XuluRegistry;
import edu.bonn.xulu.appl.XuluRegistryReader;
import edu.bonn.xulu.appl.VisualisationManager;
import edu.bonn.xulu.appl.ModelControlManager;
import edu.bonn.xulu.appl.EventManager;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.plugin.appl.DataScriptInterpreter_Basic;
import edu.bonn.xulu.plugin.appl.XuluRegistryReader_BasicAscii;

import appl.ext.XuluConfig;

/**
 * Diese Klasse implementiert das Hauptprogramm der Xulu-Modeling-Platform.
 * Beim Starten koennen folgende Kommandozeilenparameter angegeben werden:
 * <ul>
 * <li><code>? | /? | -? | -h | h | --help</code><br>
 *     Zeigt eine Hilfe zu den Kommandozeilenparametern an.</li>
 * <li><code>-kso</code><br>
 *     Haelt den Start-Dialog offen, bis "OK" gedrueckt wird.</li>
 * <li><code>-nogui</code><br>
 *     Xulu-GUI wird <b>nicht</b> gestartet! Dies macht nur Sinn, wenn autom.
 *     ein Skript gestartet wird oder die Xulu-Applikation per Konstruktor
 *     aus einer anderen Applikation instanziiert wird.</li>
 * <li><code>-noplugin</code><br>
 *     Ignoriert die in der Xulu-Registry angegebene Autostart-Funktion fuer
 *     Plugins.</li>
 * <li><code>-r <i>path</i></code><br>
 *     Setzt das Start-Verzeichnis (Root) fuer die Instanz, in dem die Registry- und
 *     XuluProperties-Dateien zu finden sind (Standard: Start-Verzeichnis).</li>
 * <li><code>-w <i>path</i></code><br>
 *     Setzt das Arbeitsverzeichnis fuer die Instanz.</li>
 * <li><code>-l <i>language</i></code><br>
 *     Setzt die Sprache fuer die GUI. Muss ein ISO-639 Code sein.</li>
 * <li><code>-rf <i>filepath</i></code><br>
 *     Setzt die zu Beginn einzulesende Registry-Datei (Default: <code>registry.xif</code>)</li>
 * <li><code>-rr <i>package.class</i></code><br>
 *     Setzt den RegistryReader, der zum Einlesen der Registry verwendet wird.
 *     Muss eine ueber einen Standard-Konstruktor instanziierbare Subklasse
 *     von {@link XuluRegistryReader} sein (Default: {@link XuluRegistryReader_BasicAscii}.</li>
 * <li><code>-d <i>path</i></code><br>
 *     Setzt den Suchpfad (Classpath) fuer die dynamisch zu aktualisierenden Klassen
 *     (zur Zeit nur Modell-Klassen, also das Paket <code>edu.bonn.xulu.plugin.model</code>).
 *     Standardmaessig wird das Verzeichnis <code>classes_dyn</code>
 *     unterhalb des Xulu-Programmverzeichnisses verwendet.<br>
 *     <b>Bemerke:</b><br>
 *     Dieses Verzeichnis darf <b>kein</b> Bestandteil des CLASSPATH der JVM
 *     sein, in der Xulu gestartet wird, da ansonsten das dynamische
 *     Aktualisieren veraenderter Klassen nicht funktioniert!</li>
 * <li><code>-s <i>filepath</i></code><br>
 *     Fuehrt nach dem Initialisieren automatisch die durch ';' getrennten Skripte aus</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XuluModellingPlatform implements XuluComponent {

  /** Liste aller in Xulu verwendeten {@link ResourceProvider} */
  public static final SortableVector<ResourceProvider> RESOURCE_BUNDLES = new SortableVector<ResourceProvider>();
  static {
    // Resource Bundles aus Standard-Paketen umlenken
    schmitzm.swing.SwingUtil.RESOURCE.resetResourceBundle(XuluConstants.EXTENTION_BUNDLE,"schmitzm.swing");
    schmitzm.geotools.gui.GeotoolsGUIUtil.RESOURCE.resetResourceBundle(XuluConstants.EXTENTION_BUNDLE,"schmitzm.gt");
    schmitzm.jfree.JFreeChartUtil.RESOURCE.resetResourceBundle(XuluConstants.EXTENTION_BUNDLE,"schmitzm.jfree");
    schmitzm.data.DataUtil.RESOURCE.resetResourceBundle(XuluConstants.EXTENTION_BUNDLE,"schmitzm.data");
    // Resource Bundles in Liste sammeln
    registerResourceProvider(schmitzm.swing.SwingUtil.RESOURCE);
    registerResourceProvider(schmitzm.geotools.gui.GeotoolsGUIUtil.RESOURCE);
    registerResourceProvider(schmitzm.jfree.JFreeChartUtil.RESOURCE);
    registerResourceProvider(schmitzm.data.DataUtil.RESOURCE);
  }

  private static boolean keepStartFrameOpen = false;
  private        boolean startGUI = true;
  private        boolean autostartPlugins = true;

  private boolean packFrame = false;

  // Art und Weise die Registry einzulesen
  // ACHTUNG: NACH AENDERUG AUCH DIE JAVADOC OBEN ANPASSEN
  private File               registryFile   = new File("registry.xif");
  private XuluRegistryReader registryReader = new XuluRegistryReader_BasicAscii();

  // Xulu-Komponenten
  private DataPool             dataPool       = null;
  private XuluRegistry         registry       = null;
  private VisualisationManager visManager     = null;
  private ModelControlManager  modelManager   = null;
  private EventManager         eventManager   = null;
  // Haupt-GUI
  private XuluMainFrame        mainFrame      = null;

  // Programm-Parameter
  private File                 startDir    = new File(".");
  private File                 pluginDir    = new File("plugin");
  private File                 workDir     = new File(".");
  private File                 dynClassDir = null;
  private File                 startupScriptFiles[] = null;

  /**
   * Erzeugt eine neue Instanz der Xulu-Modelling-Platform.
   */
  public XuluModellingPlatform() {
    this( new String[0] );
  }

  /**
   * Erzeugt eine neue Instanz der Xulu-Modelling-Platform.
   * @param args Kommandozeilenparameter
   */
  public XuluModellingPlatform(String[] args) {
    this(args,null);
  }

  private void giveStatus(String text, PrintStream stream) {
    if (stream == null)
      return;
    stream.print(text);
  }

  /**
   * Erzeugt eine neue Instanz der Xulu-Modelling-Platform.
   * @param args       Kommandozeilenparameter
   * @param infoStream Stream in den Status-Meldungen geschrieben werden, die
   *                   ueber den Fortschritt der Initialisierung Auskunft geben
   */
  public XuluModellingPlatform(String[] args, PrintStream infoStream) {
    //this.dynClassDir = new File("classes_dyn");
    try {
      this.dynClassDir = new File(ClassLoader.getSystemResource("").toURI());
    } catch(Exception err) {
      throw new RuntimeException(err);
    }
    
    giveStatus("Reading command line parameters...",infoStream);
    interpreteCommandLineParameters(args);
    giveStatus(" done.\n",infoStream);

    //Reading config file (automaticly loaded from File 'XuluProperties'
    appl.ext.XuluConfig.getXuluConfig( startDir );

    // Datenpool erzeugen und initialisieren
    giveStatus("Creating xulu data pool...",infoStream);
    this.dataPool = new DataPool();
    giveStatus(" done.\n",infoStream);

    // Registry erzeugen
    giveStatus("Creating xulu registry...",infoStream);
    this.registry = new XuluRegistry();
    giveStatus(" done.\n",infoStream);

    // Visualisierungsmanager erzeugen und initialisieren
    giveStatus("Creating xulu visualisation manager...",infoStream);
    this.visManager = new VisualisationManager();
    giveStatus(" done.\n",infoStream);

    // Manager fuer Modell-Kontrollkomponenten erzeugen und initialisieren
    giveStatus("Creating xulu model manager...",infoStream);
    this.modelManager = new ModelControlManager();
    giveStatus(" done.\n",infoStream);

    // Manager fuer Ereignisse erzeugen und initialisieren
    giveStatus("Creating xulu event manager...",infoStream);
    this.eventManager = new EventManager(this);
    giveStatus(" done.\n",infoStream);

    // MainFrame erzeugen und initialisieren
    if ( startGUI ) {
      giveStatus("Creating xulu main window...",infoStream);
      this.mainFrame = new XuluMainFrame(this);
      //Frames Ueberpruefen, die voreingestellte Groesse haben
      //Frames packen, die nutzbare bevorzugte Groesseninformationen enthalten, z.B. aus ihrem Layout
      if (packFrame)
        mainFrame.pack();
      else
        mainFrame.validate();
      SwingUtil.centerFrameOnScreen(mainFrame);
      mainFrame.setVisible(true);
      mainFrame.arrangeInnerFrames();
      giveStatus(" done.\n",infoStream);
    }

    // Registry einlesen
    giveStatus("Reading xulu registry file...",infoStream);
    try {
      FileInputStream fileIn = new FileInputStream(registryFile);
      registryReader.readRegistry(this,fileIn,registry);
      fileIn.close();
    } catch (Exception err) {
      System.err.println("Fehler beim Initialisieren der XuluRegistry: "+err);
    }
    giveStatus(" done.\n",infoStream);

    // RecentImports und RecentScripts aus Properties lesen
    registry.getRecentImports().loadFromXuluProperties(this);
    registry.getRecentScripts().loadFromXuluProperties(this);
    // Standard-CRS aus Properties lesen
    giveStatus("Read default CRS from XULU properties...",infoStream);
    String defaultCRS = XuluConfig.getXuluConfig().getProperty("General.DefaultCRS");
    CoordinateReferenceSystem crs = defaultCRS != null ? GTUtil.createCRS(defaultCRS) : null;
    if ( crs != null )  {
      GeoImportUtil.DEFAULT_CRS = crs;
      giveStatus(" done.\n",infoStream);
    } else {
      giveStatus(" failt!\n",infoStream);
    }


    // Autostart-Plugins starten
    // --> WICHTIG: sollte VOR dem Ausfuehren der Start-Skripte stattfinden,
    //              damit das LoggerPlugin bereits gestartet ist (fuer
    //              Meldungen waehrend der Start-Skripte!)
    if ( this.autostartPlugins ) {
      XuluPlugin[] autostartPlugin = registry.getAutostartPlugins();
      giveStatus("Executing Autostart-Plugins...",infoStream);
      for (int i=0; autostartPlugin != null && i<autostartPlugin.length; i++ )
        autostartPlugin[i].execute(this);
      giveStatus(" done.\n",infoStream);
    }

    // Start the scripts
    if (startupScriptFiles != null)
      for (File file : startupScriptFiles) {
        if (file != null) {
          giveStatus("Executing script " + file.getName() + "...", infoStream);
          try {
            FileInputStream fileIn = new FileInputStream(file);
            new DataScriptInterpreter_Basic().execute(fileIn, this);
            fileIn.close();
          }
          catch (Exception err) {
            System.err.println("Error while executing startup script: " + err);
          }
          giveStatus(" done.\n", infoStream);
        }
      }

  }

  /**
   * Zerstoert die Instanz der Xulu-Anwendung in dem ein
   * {@link System#exit(int) System.exit(0)} ausgefuehrt wird. Zuvor werden
   * folgende Komponenten zerstoert:
   * <ul>
   * <li>{@linkplain DataPool#dispose() Daten-Pool}</li>
   * <li>{@linkplain ModelControlManager#dispose() Model-Control-Manager}</li>
   * <li>{@linkplain VisualisationManager#dispose() Visualisation-Manager}</li>
   * <li>{@linkplain XuluMainFrame#dispose() GUI-Hauptfenster}</li>
   * </ul>
   */
  public void dispose() {
    XuluComponentUtil.checkDisposed(this);
    registry.getRecentImports().storeToXuluProperties(this);
    registry.getRecentScripts().storeToXuluProperties(this);
    XuluConfig.getXuluConfig().store();
    if ( modelManager != null )
      modelManager.removeAll();
    modelManager = null;
    if ( visManager != null )
      visManager.removeAll();
    visManager = null;
    if ( dataPool != null )
      dataPool.dispose();
    dataPool = null;
    if ( mainFrame != null )
      mainFrame.dispose();
    mainFrame = null;
    if ( startGUI )
      System.exit(0);
  }

  /**
   * Prueft, ob die Xulu-Anwendung zerstoert ist.
   * @see #dispose()
   */
  public boolean isDisposed() {
    return getRegistry() == null;
  }

  /**
   * Liefert das Startverzeichnis der Xulu-Applikation.
   */
  public File getStartingDirectory() {
    return startDir;
  }

  /**
   * Liefert das CLASSPATH-Verzeichnis unterhalb dessen die Klassen zu finden
   * sind, die dynamisch in Xulu geladen werden koennen.<br>
   * <b>Bemerkung:</b><br>
   * Dieses Verzeichnis darf <b>nicht</b> im CLASSPATH der JVM enthalten sein,
   * in der Xulu ausgefuehrt wird!!
   */
  public File getDynamicClassRootDirectory() {
    return dynClassDir;
  }

  /**
   * Setzt das CLASSPATH-Verzeichnis unterhalb dessen die Klassen zu finden
   * sind, die dynamisch in Xulu geladen werden koennen.<br>
   * <b>Bemerkung:</b><br>
   * Dieses Verzeichnis darf <b>nicht</b> im CLASSPATH der JVM enthalten sein,
   * in der Xulu ausgefuehrt wird!!
   */
  public void setDynamicClassRootDirectory(File searchPath) {
    dynClassDir = searchPath;
  }

  /**
   * Liefert das Verzeichnis ('{@code plugin}' unterhalb des Start-Verzeichnisses)
   * in dem von Plugins benoetigte Dateien abgelegt werden.
   */
  public File getPluginDirectory() {
    return pluginDir;
  }

  /**
   * Liefert das Arbeitsverzeichnis der Xulu-Applikation.
   */
  public File getWorkingDirectory() {
    return workDir;
  }

  /**
   * Setzt das Arbeitsverzeichnis der Xulu-Applikation.
   */
  public void setWorkingDirectory(File workDir) {
    if ( !workDir.isDirectory() || !workDir.exists() )
      throw new RuntimeException("Working directory must be a directory and must exist!");
    this.workDir = workDir;
  }

  /**
   * Liefert den Datenpool der Xulu-Instanz.
   */
  public DataPool getDataPool() {
    return dataPool;
  }

  /**
   * Liefert die Registry der Xulu-Instanz.
   */
  public XuluRegistry getRegistry() {
      return registry;
  }

  /**
   * Liefert den Verwaltungsmanager fuer die Visualisierungstool-Instanzen.
   */
  public VisualisationManager getVisualisationManager() {
      return visManager;
  }

  /**
   * Liefert den Verwaltungsmanager fuer die Modell-Kontroll-Komponenten.
   */
  public ModelControlManager getModelControlManager() {
      return modelManager;
  }

  /**
   * Liefert den Verwaltungsmanager fuer die Modell-Kontroll-Komponenten.
   */
  public EventManager getEventManager() {
      return eventManager;
  }

  /**
   * Liefert das Hauptfenster der Xulu-Instanz.
   */
  public XuluMainFrame getMainFrame() {
      return mainFrame;
  }

  /**
   * Interpretiert die Kommandozeilen-Parameter, die fuer die Xulu-Instanz
   * relevant sind.
   * Dies sind:
   * <ul>
   * <li><code>-r <i>path</i></code><br>
   *     Setzt das Start-Verzeichnis (Root) fuer die Instanz, in dem die Registry- und
   *     XuluProperties-Dateien zu finden sind.</li>
   * <li><code>-w <i>path</i></code><br>
   *     Setzt das Arbeitsverzeichnis fuer die Instanz.</li>
   * <li><code>-rf <i>path</i></code><br>
   *     Bestimmt die einzulesende Registry-Datei.</li>
   * <li><code>-rr <i>package.class</i></code><br>
   *     Bestimmt die Art, die Registry-Datei einzulesen. <i>package.class</i> muss eine
   *     Unterklasse von {@link XuluRegistryReader} sein und einen Standard-Konstruktor
   *     besitzen.</li>
   * <li><code>-d <i>path</i></code><br>
   *     Setzt den Suchpfad (Classpath) fuer die dynamisch zu ladenen Klassen (zur Zeit
   *     nur Modell-Klassen). Standardmaessig wird das Verzeichnis <code>classes_dyn</code>
   *     unterhalb des Xulu-Programmverzeichnisses verwendet.<br>
   *     <b>Bemerke:</b><br>
   *     Dieses Verzeichnis darf <b>kein</b> Bestandteil des CLASSPATH der JVM
   *     sein, in der Xulu gestartet wird, da ansonsten das dynamische
   *     Aktualisieren veraenderter Klassen nicht funktioniert!</li>
   * <li><code>-s <i>path<i></code><br>
   *     Fuehrt die angegebenen durch ';' getrennten Scripte beim Starten aus. </li>
   * <li><code>-nogui</code><br>
   *     Verhindert, dass die Xulu-GUI gestartet wird. Dies macht nur Sinn, wenn autom.
   *     ein Skript gestartet wird oder die Xulu-Applikation per Konstruktor
   *     aus einer anderen Applikation instanziiert wird.</li>
   * <li><code>-noplugin</code><br>
   *     Ignoriert die in der Xulu-Registry angegebene Autostart-Funktion fuer
   *     Plugins.</li>
   * </ul>
   */
  private void interpreteCommandLineParameters(String[] args) {
    for (int i=0; args!=null && i<args.length; i++) {
      // bereits verarbeitete Argumente ignorieren
      if ( args[i] == null )
        continue;

       // Angabe des Root-Verzeichnisses
       // -r <path>
       if ( args[i].equalsIgnoreCase("-r") )
         try {
           startDir = new File(args[i+1]);
           i++;
           continue;
         } catch (Exception err) {
           System.err.println("Invalid Argument: After -w a valid path is expected");
           continue;
         }
        // Angabe des Arbeitsverzeichnisses
        // -w <path>
        if ( args[i].equalsIgnoreCase("-w") )
          try {
            setWorkingDirectory( new File(args[i+1]) );
            i++;
            continue;
          } catch (Exception err) {
            System.err.println("Invalid Argument: After -w a valid path is expected");
            continue;
          }
       //Angabe von Startup-Skripten
       // -s <file>
       if ( args[i].equalsIgnoreCase("-s") )
         try {
             //splitte nach den einzelnen files
        	 String[] fileNames = args[i+1].split(";");
        	 startupScriptFiles = new File[fileNames.length];
        	 for (int j = 0; j < fileNames.length; j++) {
				startupScriptFiles [j] = new File(fileNames[j]);
			}
           i++;
           continue;
         } catch (Exception err) {
           System.err.println("Invalid Argument: After -s a valid paths sperated by ';' are expected");
           continue;
         }
        // Angabe der Registry-Datei
        // -rf <file>
        if ( args[i].equalsIgnoreCase("-rf") )
         try {
           registryFile = new File( args[i+1] );
           i++;
           continue;
         } catch (Exception err) {
           System.err.println("Invalid Argument: After -rf a valid path is expected");
           continue;
         }
         // Angabe des Registry-Readers
         // -r <package.class>
         if ( args[i].equalsIgnoreCase("-rr") )
          try {
            registryReader = (XuluRegistryReader)Class.forName(args[i+1]).newInstance();
            i++;
            continue;
          } catch (InstantiationException err) {
            System.err.println("Invalid Argument: Subclass of XuluRegistryReader needs a default constructor");
            continue;
          } catch (Exception err) {
            System.err.println("Invalid Argument: After -rr a subclass of XuluRegistryReader is expected");
            continue;
          }
          // Angabe des Such-Pfades fuer dynamische Klassen
          // -d <path>
          if ( args[i].equalsIgnoreCase("-d") )
           try {
             setDynamicClassRootDirectory( new File(args[i+1]) );
             i++;
             continue;
           } catch (Exception err) {
             System.err.println("Invalid Argument: After -d a valid path is expected");
             continue;
           }
         // Flag, dass keine GUI gestartet wird
         if ( args[i].equalsIgnoreCase("-nogui") ) {
           this.startGUI = false;
           continue;
         }
         // Flag, dass keine Plugins gestartet werden
         if ( args[i].equalsIgnoreCase("-noplugin") ) {
           this.autostartPlugins = false;
           continue;
         }
     System.err.println("Unknown Argument: "+args[i]);
   }
  }

  ///////////////////////////////////////////////////////////////////
  // Starter-Methoden
  ///////////////////////////////////////////////////////////////////
  /**
   * Interpretiert die Kommandozeilen-Parameter, die statisch relevant sind.
   * Dies sind:
   * <ul>
   * <li><code>? | /? | -? | -h | h | --help</code><br>
   *     Zeigt eine Hilfe zu den Kommandozeilenparametern an.</li>
   * <li><code>-kso</code><br>
   *     Haelt das Startup-Fenster nach dem Xulu-Start offen.</li>
   * <li><code>-l <i>language</i></code><br>
   *     Setzt die Sprache der GUI. <i>language</i> muss ein ISO-639 Code sein.</li>
   * </ul>
   */
  private static void interpreteStaticCommandLineParameters(String[] args) {
    for (int i=0; args!=null && i<args.length; i++) {
        // Parameter-Hilfe
        // ? -? /? h -h --help
        if ( args[i].startsWith("?")  || args[i].startsWith("-?") ||
             args[i].startsWith("/?") ||
             args[i].equalsIgnoreCase("h") || args[i].equalsIgnoreCase("-h") ||
             args[i].equalsIgnoreCase("--help") ) {
          showParameterHelp();
          args[i] = null; // Argument als verarbeitet kennzeichnen
          System.exit(0);
        }
        // Angabe der Sprache
        // -l <language>
        if ( args[i].equalsIgnoreCase("-l") )
          try {
            Locale.setDefault( new Locale(args[i+1]) );
            i++;
            // Argumente als verarbeitet kennzeichnen
            args[i-1] = null;
            args[i] = null;
            continue;
          } catch (Exception err) {
            System.err.println("Invalid Argument: After -l a valid ISO-639 code is expected");
            continue;
          }
        // Start-Up-Fenster offen halten
        if ( args[i].equalsIgnoreCase("-kso") ) {
          keepStartFrameOpen = true;
          args[i] = null; // Argument als verarbeitet kennzeichnen
        }
    }
  }

  private static void showParameterHelp() {
    System.out.println("\nUsage: java [jre commands] edu.bonn.xulu.XuluModellingPlatform [Xulu commands]");
    System.out.println("Xulu commands:  ? | /? | -?       Shows this help message");
    System.out.println("               -h |  h | --help   Shows this help message");
    System.out.println("               -kso       gs        Keeps the starting frame open.");
    System.out.println("               -r <path>          Sets the starting (root) directory to <path>");
    System.out.println("               -w <path>          Sets the working directory to <path>");
    System.out.println("               -l <language>      Sets the language for the application");
    System.out.println("                                  Note:");
    System.out.println("                                  <language> must be a lowercase two-letter");
    System.out.println("                                  ISO-639 code and the corresponding");
    System.out.println("                                  ResourceBundle classes must be implemented.");
    System.out.println("                                  Currently 'de' and 'en' are supported.");
    System.out.println("               -rf <filepath>     Sets the path to xulu registry file.");
    System.out.println("               -rr <class>        Sets the class to read the xulu registry");
    System.out.println("                                  Note:");
    System.out.println("                                  <class> must be a subclass of XuluRegistryReader");
    System.out.println("                                  and have a default constructor.");
    System.out.println("                                  A full class name (package.class) must be specified.");
    System.out.println("               -d <path>          Sets the directory where the dynamic (reloadable)");
    System.out.println("                                  classes are located.");
    System.out.println("                                  Note:");
    System.out.println("                                  <path> is not allowed to be a part of the CLASSPATH!");
    System.out.println("               -s <path>          Executes the given datapool scipt after xulu start.");
    System.out.println("               -nogui             Avoids starting the Xulu GUI. This makes only sence,");
    System.out.println("                                  if a script is executed automatically or the");
    System.out.println("                                  XuluModellingPlatform class is instanciated manually");
    System.out.println("                                  by another application.");
    System.out.println("               -noplugin          Ignores the autostart functionality specified for");
    System.out.println("                                  plugins in the Xulu registry.");
    System.out.println();
  }

  /**
   * Startet die Xulu-Anwendungung.
   * @param args Kommandozeilenparameter
   */
  public static void main(final String[] args) {
    try {
      if( XuluConfig.getXuluConfig().getBooleanProperty("General.useSystemLookAndFeel") )
        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());

      interpreteStaticCommandLineParameters(args);

      // Waehrend die Xulu-Applikation (im Hintergrund) gestartet wird, wird
      // ein Fenster angezeigt, in dem die Initialisierungsphasen angezeigt
      // werden
      SwingWorker.Work initWork = new SwingWorker.Work() {
        public Object execute() {
          new XuluModellingPlatform(args,((XuluStartingFrame)getSwingWorker().getDialog()).getPrintStream());
          return null;
        }
        public void performError(Throwable err) {
          err.printStackTrace();
          super.performError(err);
          System.exit(-1);
        }
      };
      SwingWorker startingWindow = new SwingWorker(initWork,new XuluStartingFrame()) {
        // wenn der Start-Dialog abgebrochen wird, wird die komplette
        // Applikation abgebrochen
        public void terminate() {
          System.exit(0);
        }
      };
      // Ueber den Kommandozeilenparameter -kso wird das Start-Fenster
      // offen gehalten und nicht automatisch geschlossen
      startingWindow.setKeepOpen( keepStartFrameOpen );
      startingWindow.start();

      // Begruessungsmeldung im Xulu-Fenster
      System.out.println("Start XULU by");
      System.out.println("   java [jre commands] edu.bonn.xulu.XuluModellingPlatform --help");
      System.out.println("to see available command line parameters.\n");
    } catch(Throwable e) {
      e.printStackTrace();
      ExceptionDialog.show(e);
      System.exit(-1);
    }
  }

  /**
   * Fuegt einen {@link ResourceProvider} in die Liste der in Xulu verwendeten
   * ResourceProvider ein.
   * @param rp ein ResourceProvider
   */
  public static void registerResourceProvider(ResourceProvider rp) {
    RESOURCE_BUNDLES.add(rp);
  }

  /**
   * Erzeugt einen {@link ResourceProvider} unterhalb von {@link XuluConstants#LOCALES_BASE}
   * ({@code locales}) und fuegt ihn in die Liste der in Xulu verwendeten
   * ResourceProvider ein.
   * @param bundleName    Name des Bundles
   * @param defaultLocale Standard-SPrache des Bundles
   * @see XuluConstants#LOCALES_BASE
   */
  public static ResourceProvider registerResourceProvider(String bundleName, Locale defaultLocale) {
    ResourceProvider rp = new ResourceProvider(LOCALES_BASE+"."+bundleName,defaultLocale);
    registerResourceProvider( rp );
    return rp;
  }

}
