/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.util.Vector;

import schmitzm.lang.DuplicateException;
import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.Invoker;

// Nur fuer Doku
import schmitzm.data.event.ObjectListener;

/**
 * Diese Klasse stellt einen allgemeinen Manager dar. Dieser dient dazu,
 * Komponenten, denen ein grafisches Fenster zugeordnet ist, in einer Liste zu
 * verwalten. z.B.
 * <ul>
 * <li>Visualisierungstools</li>
 * <li>Modell-Instanzen</li>
 * <li>...</li>
 * </ul>
 * Aus diesem Grund muessen die Objekte, bestimmte Anforderungen erfuellen, die
 * durch das Interface {@link AbstractManager.ContentType} spezifiziert werden.<br>
 * Der Manager fungiert als <code>AbstractObjectTraceable</code>, so dass
 * {@link ObjectListener} angeschlossen werden koennen, welche automatisch durch
 * ein {@link AbstractManager.ChangeEvent} informiert werden, wenn sich der
 * Inhalt des Managers aendert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractManager extends AbstractObjectTraceable implements XuluComponent {
  /** Speichert den Inhalt des Managers. */
  protected Vector content = new Vector<ContentType>();
  /** Speichert, ob in dem Manager mehrere gleiche Objekte erlaubt sind */
  protected boolean duplAllowed = true;

  /**
   * Erzeugt einen neuen (leeren) Manager.
   * @param duplAllowed steuert, ob Duplikate im Manager erlaubt sind
   */
  public AbstractManager(boolean duplAllowed) {
    this.duplAllowed = duplAllowed;
  }

  /**
   * Erzeugt einen neuen (leeren) Manager, der Duplikate erlaubt.
   */
  public AbstractManager() {
    this(true);
  }

  /**
   * Prueft, ob im Manager Objekte mit der gleichen Bezeichnung erlaubt sind.
   */
  public boolean getDuplicatedAllowed() {
    return this.duplAllowed;
  }

  /**
   * Zerstoert den Manager und alle darin enthaltenen Komponenten.
   */
  public void dispose() {
    this.removeAll();
    content = null;
  }

  /**
   * Prueft, ob der Manager zerstoert ist.
   * @see #dispose()
   */
  public boolean isDisposed() {
    return content == null;
  }

  /**
   * Liefert die Anzahl an aktuell registrierten Komponenten.
   */
  public int getCount() {
    XuluComponentUtil.checkDisposed(this);
    return content.size();
  }

  /**
   * Liefert eine registrierte Komponente.
   * @param index Listen-Index
   */
  public ContentType get(int index) {
    XuluComponentUtil.checkDisposed(this);
    return (ContentType)content.get(index);
  }

  /**
   * Liefert alle registrierte Komponenten als Array.
   * @return leerer Array falls keine Komponenten im Manager registriert sind.
   */
  public ContentType[] getAll() {
    XuluComponentUtil.checkDisposed(this);
    return (ContentType[])content.toArray(new ContentType[0]);
  }

  /**
   * Fuegt dem Manager eine Komponente hinzu. Das Fenster dieser
   * Komponente wird noch nicht angezeigt!
   * @param comp neue Komponente
   */
  public void add(ContentType comp) {
    add(comp,false);
  }

  /**
   * Fuegt dem Manager eine Komponente hinzu.
   * @param comp neue Komponente
   * @param visible bestimmt, ob die Komponente (bzw. deren Fenster) direkt
   *                angezeigt wird
   */
  public void add(ContentType comp, boolean visible) {
    XuluComponentUtil.checkDisposed(this);
    checkComponentAndError(comp);

    if ( !duplAllowed )
      checkDuplicateAndError(comp);

    if ( content.add( comp ) )
      fireEvent( new ChangeEvent(this,null,comp) );
    comp.setVisible(visible);
  }

  /**
   * Entfernt eine Komponente aus dem Manager. Zuvor wird diese
   * komplett geschlossen.
   * @param comp zu schliessende Komponente
   * @see ContentType#close()
   */
  public void remove(ContentType comp) {
    XuluComponentUtil.checkDisposed(this);
    if ( content.remove(comp) )
      fireEvent( new ChangeEvent(this,comp,null) );
    comp.setVisible(false);
    comp.close();
  }

  /**
   * Entfernt alle Komponenten aus dem Manager. Zuvor werden diese
   * komplett geschlossen.
   * @see ContentType#close()
   */
  public void removeAll() {
    while( getCount() > 0 )
      remove(0);
  }

  /**
   * Entfernt eine Komponente aus dem Manager. Zuvor wird diese
   * komplett geschlossen.
   * @param index Listen-Index
   * @see ContentType#close()
   */
  public void remove(int index) {
    XuluComponentUtil.checkDisposed(this);
    remove( (ContentType)content.elementAt(index) );
  }

  /**
   * Prueft, ob es bereits ein Objekt im Manager gibt, das dem uebergebenen
   * gleicht (diese Implementierung prueft auf die {@code equals(.)}.Methode).
   * Ist dies der Fall, wird eine {@link DuplicateException} geworfen.
   * @param cont Objekt auf das geprueft wird
   */
  protected void checkDuplicateAndError(ContentType cont) {
    ContentType[] content = getAll();
    for (ContentType c : content)
      if ( c == null && cont == null ||
           c != null && c.equals( cont ) )
        throw new DuplicateException("Duplicate Object in AbstractManager",c,cont);
  }

  /**
   * Diese Methode prueft, ob der Inhalt fuer die konkrete Implementierung
   * zulaessig ist.
   * @param cont ContentType
   */
  public abstract void checkComponentAndError(ContentType cont);

  /**
   * Dieses Interface definiert Mindestanforderungen, die die Komponenten
   * erfuellen muessen, die in dem Manager verwaltet werden sollen.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static interface ContentType {
    /**
     * Zeigt die Komponente (deren Fenster an) oder verbirgt sie.
     */
    public void setVisible(boolean visible);

    /**
     * Prueft, ob die Komponente (deren Fenster) aktuell angezeigt wird.
     */
    public boolean isVisible();

    /**
     * Liefert die Bezeichnung der Komponente. Hiermit wird die Komponente
     * in einer grafischen Darstellung (Tabelle) identifiziert.
     */
    public String getTitle();

    /**
     * Setzt die Bezeichnung der Komponente. Hiermit wird die Komponente
     * in einer grafischen Darstellung (Tabelle) identifiziert.
     */
    public void setTitle(String newTitle);

    /**
     * Schliesst die Komponente komplett. Alle internen Ressourcen sollten
     * freigegeben werden.
     */
    public void close();
  }

  /**
   * Diese Ereignis wird erzeugt (und an alle {@link ObjectListener}
   * gesendet), wenn sich der Inhalt des Managers aendert (eine Instanz hingefuegt
   * oder entfernt wurde).
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ChangeEvent extends ObjectChangeEvent {
    /**
     * Erzeugt ein neues Ereignis.
     * @param manager  Manager, der sich geaendert hat
     * @param oldComp  alte Komponente (<code>null</code> falls
     *                 eine Komponente hinzugefuegt wurde)
     * @param newComp  neue Komponente (<code>null</code> falls
     *                 eine Komponente entfernt wurde)
     */
    public ChangeEvent(AbstractManager manager, Object oldComp, Object newComp) {
      super( new Invoker(manager, new Invoker(oldComp)), oldComp, newComp );
    }
  }
}
