/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import schmitzm.lang.AbstractNamedObject;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.XuluPlugin;

/**
 * Diese Klasse stellt eine Basis-Implementierung von {@link XuluPlugin} dar.
 * Unterklassen muessen lediglich noch die {@link XuluPlugin#execute(XuluModellingPlatform)}-Methode
 * implementieren und darin das {@link #started}-Flag pruefen, damit das
 * Plugin beendet werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractXuluPlugin extends AbstractNamedObject implements XuluPlugin {
  /** Speichert, ob das Plugin gestartet ist oder nicht. */
  protected boolean started = false;
  /** Speichert, ob die Plugin-GUI angezeigt wird. */
  protected boolean visible = false;
  /** Speichert, ob die Plugin eine GUI besitzt. */
  protected boolean hasGUI = false;
  /** Speichert, ob die Xulu-Instanz, in der das Plugin ausgefuehrt wird. */
  protected XuluModellingPlatform appl = null;

  /**
   * Erzeugt eine neue Plugin-Instanz.
   * @param hasGUI gibt an, ob das Plugin eine GUI besitzt.
   */
  public AbstractXuluPlugin(boolean hasGUI) {
    super();
    this.hasGUI = hasGUI;
  }

  /**
   * Startet die Ausfuehrung des Plugins. Diese Basis-Implementierung speichert
   * lediglich die Xulu-Instanz in der globalen Variable {@link #appl}.
   * Unterklassen muessen diese Methode ueberschreiben <b>und die
   * {@code super.execute(.)}-Methode aufrufen</b>.
   * @param appl Instanz der Xulu-Applikation
   */
  public void execute(XuluModellingPlatform appl) {
    this.appl = appl;
  }

  /**
   * Stoppt die Ausfuehrung des Plugins. Diese Basis-Implementierung setzt
   * lediglich das {@link #started}-Flag auf {@code false}.
   */
  public void stop() {
    this.started = false;
  }

  /**
   * Prueft, ob das Plugin gestartet ist.
   * @return Wert des {@link #started}-Flag
   */
  public boolean isStarted() {
    return started;
  }

  /**
   * Prueft, ob die Plugin-GUI angezeigt wird.
   * @return Wert des {@link #visible}-Flag
   */
  public boolean isVisible() {
    return visible;
  }

  /**
   * Zeigt die Plugin-GUI an oder verbirgt sie. Diese Basis-Implementierung
   * setzt lediglich das Flag {@code #visible}. Macht nichts, wenn das Plugin keine
   * GUI besitzt.
   */
  public void setVisible(boolean visible) {
    this.visible = hasGUI && visible;
  }
}
