/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.util.Vector;
import java.util.Random;
import java.util.Hashtable;
import java.util.Enumeration;

import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.AbstractObjectEvent;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.NameChangeEvent;
import schmitzm.data.event.Invoker;

import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.data.DuplicateXuluObjectIDException;
import edu.bonn.xulu.data.DuplicateXuluObjectNameException;

/**
 * Diese Klasse stellt den Xulu-Datenpool dar, in dem saemtliche (globalen)
 * Daten fuer die Modellierung verwaltet werden.<br>
 * Die Objekte werden in einer {@link java.util.Hashtable} verwaltet und sind
 * durch eine eindeutige ID ansprechbar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DataPool extends AbstractObjectTraceable implements XuluComponent {
  /** Zufallsgenerator */
  private static final Random RANDOM = new Random();
  /** Enthaelt alle <code>DataPool</code>-Instanzen. */
  private static Vector pools = new Vector();

  /** Enthaelt den Inhalt des Datenpools */
  private Hashtable content = null;

  /** Enthaelt einen {@link ObjectListener} der auf interne Veraenderungen
   *  der Datenpool-Objekte horcht */
  private ObjectListener objectListener = null;

  /** Flag, ob sich mehrere Objekte mit demselben Namen im Datenpool
   *  befinden duerfen */
  private boolean duplNamesAllowed = false;
  /** Flag, ob beim Einfuegen eines neuen Objekte bestehende Objekte mit dem
   *  gleichen Namen automatisch aus dem Datenpool entfernt werden. Ist nur
   *  relevant, wenn Duplikate verboten sind.
   *  @see #duplNamesAllowed */
  private boolean autoOverwriteDuplNames = false;

  /**
   * Erzeugt eine neue Datenpool-Instanz.
   * @param cap initiale Kapazitaet des Pools
   */
  public DataPool(int cap) {
    pools.add(this);
    this.content = new Hashtable(cap);
    final DataPool dummyPool = this;
    this.objectListener = new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        if ( e instanceof NameChangeEvent )
          fireEvent( new DataPoolChangeEvent(dummyPool,e.getSource(),DataPoolChangeEvent.OBJECT_NAME_CHANGED) );
      }
    };
  }

  /**
   * Erzeugt einen neuen Datenpool. Dieser hat die initiale Groesse von
   * 50 Objekten.
   */
  public DataPool() {
    this(50);
  }

  /**
   * Zerstoert den Datenpool und alle darin enthaltenen Objekte.
   * @see XuluObject#dispose()
   */
  public void dispose() {
    XuluComponentUtil.checkDisposed(this);
    Enumeration e = asEnumeration();
    while( e.hasMoreElements() ) {
      XuluObject o = (XuluObject)e.nextElement();
      removeObject(o);
      o.dispose();
    }
  }

  /**
   * Prueft, ob der Datenpool zerstoert ist.
   * @see #dispose()
   */
  public boolean isDisposed() {
    return content == null;
  }

  /**
   * Liefert die Anzahl an Objekten im Datenpool.
   */
  public int getSize() {
    XuluComponentUtil.checkDisposed(this);
    return content.size();
  }

  /**
   * Liefert den Inhalt des Datenpools als Liste.
   */
  public Enumeration asEnumeration() {
    XuluComponentUtil.checkDisposed(this);
    return content.elements();
  }

  /**
   * Liefert ein Objekt aus dem Datenpool.
   * @param id ID des Objekts im Datenpool
   * @return <code>null</code> falls es im Pool kein Objekt mit dieser
   *         ID gibt
   */
  public XuluObject getObject(long id) {
    XuluComponentUtil.checkDisposed(this);
    return (XuluObject)content.get(id);
  }

  /**
   * Liefert Objekte aus dem Datenpool.
   * @param name Bezeichnung des Objekts im Datenpool
   * @return ein leerer Array falls es im Pool keine Objekte mit dieser
   *         Bezeichnung gibt
   */
  public XuluObject[] getObjects(String name) {
    Vector outObjects = new Vector();
    Enumeration e = asEnumeration();
    for (;e.hasMoreElements();) {
      XuluObject o = (XuluObject)e.nextElement();
      if ( o.getDescription() != null && o.getDescription().equals(name) )
        outObjects.add(o);
    }
    return (XuluObject[])outObjects.toArray(new XuluObject[0]);
  }

  /**
   * Liefert alle Objekte aus dem Datenpool, die Instanzen eines bestimmten
   * Typs sind.
   * @param type zu suchende Typ
   * @return leerer Array, falls es keine solchen Objekte gibt
   */
  public XuluObject[] getObjects(Class type) {
    Vector outObjects = new Vector();
    Enumeration e = asEnumeration();
    for (;e.hasMoreElements();) {
      XuluObject o = (XuluObject)e.nextElement();
      if ( o != null && type.isInstance(o) )
        outObjects.add(o);
    }
    return (XuluObject[])outObjects.toArray(new XuluObject[0]);
  }

  /**
   * Prueft, ob im Pool ein Objekt vorhanden ist.
   * @param id ID des Objekts
   */
  public boolean contains(long id) {
    return getObject(id) != null;
  }

  /**
   * Prueft, ob im Pool ein Objekt vorhanden ist.
   * @param name Bezeichnung des Objekts
   */
  public boolean contains(String name) {
    return getObjects(name).length > 0;
  }
  
  /**
   * Erstellt eine fuer den Datenpool eindeutige Bezeichnung.
   * Enthaelt der Datenpool bereits ein Objekt mit der angegeben
   * Bezeichnung, wird ein Index an die Bezeichnung angehaengt.
   * @param baseName Basis-Bezeichnung
   */
  public String createUniqueDescription(String baseName) {
    String newName = baseName;
    for (int i=2; contains(newName); i++)
      newName = baseName + " ("+i+")";
    return newName;
  }

  /**
   * Fuegt dem Datenpool ein Objekt hinzu. Erzeugt ein <code>DataPoolChangeEvent</code>
   * fuer das neue Objekt.
   * @see DataPoolChangeEvent#OBJECT_ADDED
   * @exception DuplicateXuluObjectIDException falls im Datenpool bereits ein Objekt
   *            mit der gleichen Objekt-ID existiert
   * @exception DuplicateXuluObjectNameException falls die Duplikat-Eigenschaft
   *            abgeschaltet ist und im Datenpool bereits ein Objekt mit der
   *            gleichen Objekt-Beschreibung existiert
   */
  public void addObject(XuluObject object) {
    XuluComponentUtil.checkDisposed(this);
    if ( contains(object.getID()) )
      throw new DuplicateXuluObjectIDException(object,"DataPool already contains an object with ID "+object.getID());

    // Sind doppelte Namen verboten, wird geprueft, ob es bereits Objekte
    // mit dem Namen im Datenpool gibt
    if ( !duplNamesAllowed ) {
      XuluObject[] obj = getObjects(object.getDescription());
      // Wenn kein AutoOverwrite, wird die Einfuege-Operation zurueckgewiesen
      if ( obj.length > 0 && !autoOverwriteDuplNames )
        throw new DuplicateXuluObjectNameException(object,"DataPool already contains object '"+object.getDescription()+"'");
      for (XuluObject o : obj)
        removeObject(o);
    }
    content.put(object.getID(),object);
    object.addObjectListener( objectListener );
    fireEvent( new DataPoolChangeEvent(this,object,DataPoolChangeEvent.OBJECT_ADDED) );
  }

  /**
   * Entfernt ein Objekt aus dem Datenpool. Erzeugt ein <code>DataPoolChangeEvent</code>
   * fuer das entfernte Objekt. Existiert das Objekt nicht im DatenPool, geschieht
   * nichts (es wird auch kein Event erzeugt).
   * @param object zu entfernendes Xulu-Objekt
   * @see DataPoolChangeEvent#OBJECT_REMOVED
   */
  public void removeObject(XuluObject object) {
    removeObject( object.getID() );
  }

  /**
   * Entfernt ein Objekt aus dem Datenpool. Erzeugt ein <code>DataPoolChangeEvent</code>
   * fuer das entfernte Objekt. Existiert im DatenPool kein Objekt mit der ID,
   * geschieht nichts (es wird auch kein Event erzeugt).<br>
   * <b>Bemerke:</b><br>
   * Durch das Loeschen des Objekts wird dieses noch nicht zerstoert. Hierdurch
   * kann es einfach in eine andere Datenpool-Instanz eingefuegt werden!
   * @param id ID des Xulu-Objekts im Datenpool
   * @see DataPoolChangeEvent#OBJECT_REMOVED
   */
  public void removeObject(long id) {
    XuluComponentUtil.checkDisposed(this);
    XuluObject object = (XuluObject)content.remove(id);
    object.removeObjectListener( objectListener );
    if ( object != null )
      fireEvent( new DataPoolChangeEvent(this,object,DataPoolChangeEvent.OBJECT_REMOVED) );
  }

  /**
   * Prueft, ob sich mehrere Objekte mit demselben Namen im Datenpool
   * befinden duerfen.
   * @see #setAutoOverwriteDuplicateNames(boolean)
   */
  public boolean getDuplicateNamesAllowed() {
    return this.duplNamesAllowed;
  }

  /**
   * Bestimmt, ob sich mehrere Objekte mit demselben Namen im Datenpool
   * befinden duerfen.<br>
   * Wird diese Eigenschaft auf {@code false} gesetzt haengt es von der
   * {@link #setAutoOverwriteDuplicateNames(boolean) autoOverwriteDuplNames}-Eigenschaft
   * ab, ob eine Einfuege-Operation zurueckgewiesen wird, oder das bestehende Objekt
   * automatisch entfernt wird.<br>
   * Standardwert: {@code false}
   * @param duplAllowed wenn {@code true}, koennen mehrere Objekte mit demselben
   *                    Namen im Datenpool vorkommen
   */
  public void setDuplicateNamesAllowed(boolean duplAllowed) {
    this.duplNamesAllowed = duplAllowed;
  }

  /**
   * Prueft, ob beim Einfuegen eines neuen Objekte bestehende Objekte mit dem
   * gleichen Namen automatisch aus dem Datenpool entfernt werden ({@code true}),
   * oder die Einfuege-Operation zurueckgewiesen wird ({@code false}).
   * Diese Eigenschaft ist nur relevant, wenn Duplikate verboten sind. <br>
   * Standardwert: {@code false}
   * @see #getDuplicateNamesAllowed() */
  public boolean getAutoOverwrideDuplicateNames() {
    return this.autoOverwriteDuplNames;
  }

  /**
   * Bestimmt, ob beim Einfuegen eines neuen Objekte bestehende Objekte mit dem
   * gleichen Namen automatisch aus dem Datenpool entfernt werden ({@code true}),
   * oder die Einfuege-Operation zurueckgewiesen wird ({@code false}).
   * Diese Eigenschaft ist nur relevant, wenn Duplikate verboten sind.<br>
   * Standardwert: {@code false}
   * @see #setDuplicateNamesAllowed(boolean) */
  public void setAutoOverwriteDuplicateNames(boolean autoOverwrite) {
    this.autoOverwriteDuplNames = autoOverwrite;
  }

  ///////////////////////////////////////////////////////////////////
  // Statische Methoden
  ///////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob die angegebene ID bereits in irgendeinem Datenpool fuer
   * ein Objekt verwendet wird.
   */
  private static boolean checkUnused(long id) {
    for (int i=0; i<pools.size(); i++)
      if ( ((DataPool)pools.elementAt(i)).contains(id) )
        return false;
    return true;
  }

  /**
   * Generiert eine Objekt-ID, die (ueber alle <code>DataPool</code>-Instanzen!)
   * eindeutig ist.
   */
  public static long generateUniqueID() {
    long newID = 0;
    while( !checkUnused( newID = System.currentTimeMillis() % RANDOM.nextInt() ) );
    return newID;
  }

  /**
   * Diese Ereignis wird erzeugt (und an alle {@link schmitzm.data.event.ObjectListener}
   * gesendet), wenn sich der Inhalt des Datenpools aendert (ein Objekt hingefuegt
   * oder entfernt wurde).
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class DataPoolChangeEvent extends AbstractObjectEvent {
    /**
     * Ereignis-Typ, der anzeigt, dass ein Objekt hinzugefuegt wurde.
     */
    public static final int OBJECT_ADDED = 30;

    /**
     * Ereignis-Typ, der anzeigt, dass ein Objekt entfernt wurde.
     */
    public static final int OBJECT_REMOVED = 40;

    /**
     * Ereignis-Typ, der anzeigt, dass ein Objekt seinen Namen geaendert hat.
     */
    public static final int OBJECT_NAME_CHANGED = 50;

    /**
     * Erzeugt ein neues Ereignis.
     * @param pool DataPool der sich geaendert hat
     * @param obj  Xulu-Objekt, das betroffen ist
     * @param type Zeigt an, was mit dem Objekt geschehen ist.
     */
    public DataPoolChangeEvent(DataPool pool, XuluObject obj, int type) {
      this(pool,new Invoker(obj), type );
    }

    /**
     * Erzeugt ein neues Ereignis.
     * @param pool   DataPool der sich geaendert hat
     * @param source betroffene Objekt-Kette
     * @param type   Zeigt an, was mit dem Objekt geschehen ist.
     */
    public DataPoolChangeEvent(DataPool pool, Invoker source, int type) {
      super( new Invoker( pool, source ), type );
    }
  }

}

