/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.util.Vector;
import java.util.Enumeration;

import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectTraceable;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.GeneralObjectChangeEvent;
import schmitzm.data.event.NameChangeEvent;
import schmitzm.data.event.ObjectCloseEvent;

import schmitzm.lang.HashtableWithCollisionList;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.appl.EventHandler;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.ModelEvent;
import edu.bonn.xulu.gui.ModelControlFrame;

import edu.bonn.xulu.plugin.appl.VisualisationHandlerFactory;
import edu.bonn.xulu.plugin.appl.FileExportHandlerFactory;

// nur fuer Doku
import edu.bonn.xulu.plugin.appl.VisualisationHandler;
import edu.bonn.xulu.appl.ModelControlManager;
import edu.bonn.xulu.appl.Handler;

/**
 * Ueber den Xulu-Event-Manager werden verschiedene Reaktionen auf Ereignisse
 * bestimmter Komponenten innerhalb Xulu-Applikation gesteuert.<br>
 * Hierzu koppelt sich der Event-Manager als Listener an den {@linkplain DataPool Xulu-Datenpool}
 * und den {@linkplain ModelControlManager Xulu-Modell-Manager}. Sobald von diesen
 * ein neues Objekt/Modell gemeldet wird, koppelt sich der Modell-Manager
 * auch an dieses neue Objekt/Modell und kann somit im Folgenden auf saemtliche
 * Ereignisse dieses Objekts reagieren (mit einem Handler).<br>
 * Der Event-Manager unterscheidet zwei Arten von {@linkplain EventHandler Event-Handlern}.
 * <ol>
 *   <li><b>Objekt-Event-Handler</b>: Der zugehoerige {@link Handler} wird
 *       ausgefuerhrt, wenn ein bestimmtes Objekt/Modell ein bestimmtes Ereignis
 *       ausloest.</li>
 *   <li><b>globale Event-Handler</b>: Nicht an ein spezielles Objekt gebunden. Der
 *       zugehoerige {@link Handler} wird ausgefuerhrt, unabhaengig davon,
 *       welches Objekt/Modell das entsprechende Ereignis initiiert</li>
 * </ol>
 * Neben fest definierten Event-Handlern besteht fuer den Anwender die Moeglichkeit,
 * ueber die GUI des Event-Managers eigene Event-Handler zu spezifizieren.
 * Zur Zeit ist ein Handler implementiert, mit dem ein Visualisierungsupdate
 * fuer ein bestimmtes Objekt vorgenommen werden kann (siehe {@link VisualisationHandler}).
 * @todo Handler zum automatisierten Speichern eines Objekts
 * @todo HandlerFactory-Klassen ueber Xulu-Registry anstelle des statischen Arrays
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class EventManager extends AbstractManager implements ObjectListener, ModelListener {
  private XuluModellingPlatform appl = null;

  /** Speichert die globalen Event-Handler, die fuer alle Objekte gueltig sind. */
  protected Vector globalEventHandler = null;
//  protected HashtableWithCollisionList globalEventHandler = null;

  /** Speichert die objekt-spezifischen Event-Handler. */
  protected HashtableWithCollisionList objectEventHandler = null;

  /** Speichert die zur Verfuegung stehenden Handler. */
  protected HandlerFactory[] handler = new HandlerFactory[] {
      new VisualisationHandlerFactory( XuluConstants.XULUGUI_RES.getString("VisualisationUpdate") ),
      new FileExportHandlerFactory( XuluConstants.XULUGUI_RES.getString("FileExport") )
  };

  /** Speichert die zur Verfuegung stehenden Objekt-Ereignisse. */
  private static final Class<ObjectEvent>[] objectEvents = new Class[] {
      GeneralObjectChangeEvent.class,
      ObjectChangeEvent.class,
      ObjectCloseEvent.class,
      NameChangeEvent.class
  };

  /**
   * Erzeugt den Event-Manager.
   * @param appl Xulu-Applikation in der die Event-Handler ausgefuehrt werden
   */
  public EventManager(XuluModellingPlatform appl) {
    super();

    globalEventHandler = new Vector(); //new HashtableWithCollisionList<Class<ObjectEvent>,EventHandler>();
    objectEventHandler = new HashtableWithCollisionList<Object,EventHandler>();
    initGlobalEventHandler();

    this.appl = appl;
    appl.getDataPool().addObjectListener(this);
    appl.getModelControlManager().addObjectListener(this);
  }

  /**
   * Zerstoert den Event-Manager. Durch den Aufruf von
   * <code>super.dispose()</code> werden zunaechst alle Objekte (Event-Handler)
   * aus dem Event-Manager entfernt ({@link #remove(AbstractManager.ContentType)}).
   * Anschliessend werden die Kopplungen zu allen Xulu-Objekten und -Modellen
   * entfernt, sowie die Kopplung zum {@linkplain DataPool Daten-Pool}
   * und {@linkplain ModelControlManager Modell-Manager}.
   */
  public void dispose() {
    super.dispose();

    // Listener von allen Datenpool-Objekten entfernen
    Enumeration list = appl.getDataPool().asEnumeration();
    while( list.hasMoreElements() )
      ((ObjectTraceable)list.nextElement()).removeObjectListener(this);

    // Listener von allen Modellen entfernen
    ModelControlFrame[] modelFrame = appl.getModelControlManager().getAll();
    for (int i=0; i<modelFrame.length; i++)
      modelFrame[i].getModel().removeModelListener(this);

    // Listener von Datenpool und Modell-Manager entfernen
    appl.getDataPool().removeObjectListener(this);
    appl.getModelControlManager().removeObjectListener(this);
  }

  /**
   * Prueft (fuer add/remove), ob es sich einem Objekt um einen {@link EventHandler} handelt.
   */
  public void checkComponentAndError(AbstractManager.ContentType comp) {
    if ( !(comp instanceof EventHandler) )
      throw new UnsupportedOperationException(getClass().getSimpleName().concat(" can not manage instances of ").concat(comp.getClass().getName()));
  }

  /**
   * Fuegt dem Event-Manager einen Event-Handler hinzu.
   * @param comp {@linkplain EventHandler Event-Handler}
   */
  public void add(AbstractManager.ContentType comp) {
    super.add(comp);
    // Auch in die Hash-Tabelle einfuegen
    EventHandler eh = (EventHandler)comp;
    if ( eh.isGlobal() )
      globalEventHandler.add(eh);
    else
      objectEventHandler.add(eh.getObject(),eh);
  }

  /**
   * Entfernt einen Event-Handler aus dem Event-Manager.
   * @param comp {@linkplain EventHandler Event-Handler}
   */
  public void remove(AbstractManager.ContentType comp) {
    super.remove(comp);
    // Auch aus den Hash-Tabelle einfuegen
    EventHandler eh = (EventHandler)comp;
    if ( eh.isGlobal() )
      globalEventHandler.remove(eh);
    else
      objectEventHandler.remove(eh.getObject(),eh);
  }

  /**
   * Liefert alle Factorys, die fuer die Erzeugung von Event-Handlern zur
   * Verfuegung stehen.
   */
  public HandlerFactory[] getHandlerFactories() {
    return handler.clone();
  }

  /**
   * Liefert alle Ereignis-Typen, die fuer ein bestimmtes Objekt zur Verfuegung stehen.
   */
  public Class[] getEventTypes(Object obj) {
    if ( obj instanceof XuluModel )
      return ((XuluModel) obj).getEvents().clone();
    if ( obj instanceof ObjectTraceable )
      return this.objectEvents.clone();
    return new Class[0];
  }


  /**
   * Verarbeitet saemtliche Ereignisse.
   * @param e ein Ereignis.
   */
  public void performObjectEvent(ObjectEvent e) {
    ///////////////////  DATENPOOL //////////////////////
    // Objekt dem Datenpool hinzugefuegt
    if ( e instanceof DataPool.DataPoolChangeEvent && e.getType() ==  DataPool.DataPoolChangeEvent.OBJECT_ADDED ) {
      performDatapoolObjectAdded(e.getSource().getRoot());
      return;
    }
    // Objekt dem Datenpool entfernt
    if ( e instanceof DataPool.DataPoolChangeEvent && e.getType() ==  DataPool.DataPoolChangeEvent.OBJECT_REMOVED ) {
      performDatapoolObjectRemoved(e.getSource().getRoot());
      return;
    }

    ///////////////////  MODELL-MANAGER //////////////////////
    if ( e instanceof ModelControlManager.ChangeEvent ) {
      // Modell dem Modell-Manager hinzugefuegt
      if ( ((ModelControlManager.ChangeEvent)e).getOldValue() == null ) {
        performModelAdded( ((ModelControlFrame)((ModelControlManager.ChangeEvent)e).getNewValue()).getModel() );
        return;
      }
      // Modell aus dem Modell-Manager entfernt
      if ( ((ModelControlManager.ChangeEvent)e).getNewValue() == null ) {
        performModelRemoved( ((ModelControlFrame)((ModelControlManager.ChangeEvent)e).getOldValue()).getModel() );
        return;
      }

    }

    /////////////////// OBJEKT  //////////////////////
    // globale Handler ausfuehren
    Vector list = globalEventHandler; //(Vector)globalEventHandler.get(e.getClass());
    for (int i=0; list!=null && i<list.size(); i++) {
      EventHandler eh = (EventHandler)list.elementAt(i);
      if ( eh.getEventType().isAssignableFrom(e.getClass()) )
           eh.getHandler().execute(e);
    }
    // objektspezifische Handler ausfuehren
    list = (Vector)objectEventHandler.get(e.getSource().getRoot());
    for (int i=0; list!=null && i<list.size(); i++) {
      EventHandler eh = (EventHandler)list.elementAt(i);
      if ( eh.getObject() == e.getSource().getRoot() && eh.getEventType().isAssignableFrom(e.getClass()) )
        eh.getHandler().execute(e);
    }
  }

  /**
   * Ruft lediglich {@link #performObjectEvent(ObjectEvent)} auf.
   * @param e ein Ereignis.
   */
  public void performModelEvent(ModelEvent e) {
    this.performObjectEvent(e);
  }

  ///////////////////////////////////////////////////////////////////
  /////  Methoden fuer die Ereignisbehandlung im Event-Manager  /////
  ///////////////////////////////////////////////////////////////////
  /**
   * Erzeugt die globalen EventHandler, die automatisch fuer alle
   * Modelle/Objekte gueltig sind.<br>
   * Zur Zeit macht diese Methode nichts!
   */
  private void initGlobalEventHandler() {
//    add( new EventHandler(
//      schmitzm.data.event.ObjectChangeEvent.class,
//      new Handler() {
//        public HandlerFactory getFactory() {
//          return null;
//        }
//
//        public void execute(ObjectEvent e) {
//          System.out.println("Test: "+e.getSource().getRoot()+" > "+e.getClass().getSimpleName());
//        }
//
//        public void close() {
//
//        }
//      },
//      appl
//    ));
  }

  /**
   * Wird dem Datenpool ein Objekt hinzugefuegt, koppelt sich der Event-Manager
   * automatisch als Listener an dieses Objekt.
   * @param obj neues Objekt im Datenpool
   */
  private void performDatapoolObjectAdded(Object obj) {
    // Event-Manager ankoppeln
    if ( obj instanceof ObjectTraceable )
      ((ObjectTraceable)obj).addObjectListener(this);
  }

  /**
   * Wird ein Objekt aus dem Datenpool geloescht, entkoppelt sich der Event-Manager
   * von diesem Objekt und loescht alle mit dem Objekt verknuepften Event-Handler.
   * @param obj aus dem Datenpool geloschtes Objekt
   */
  private void performDatapoolObjectRemoved(Object obj) {
    // alle EventHandler entfernen!
    Vector list = (Vector)objectEventHandler.get(obj);
    if ( list != null )
      list = (Vector)list.clone();

    for (int i=0; list!=null && i<list.size(); i++) {
      EventHandler eh = (EventHandler)list.elementAt(i);
      if ( eh.getObject().equals(obj) )
        remove(eh);
    }

    // Event-Manager abkoppeln
    if ( obj instanceof ObjectTraceable )
      ((ObjectTraceable)obj).removeObjectListener(this);
  }

  /**
   * Wird dem Modell-Manager ein Modell hinzugefuegt, koppelt sich der Event-Manager
   * automatisch als Listener an dieses Modell.
   * @param model neues Modell im Modell-Manager
   */
  private void performModelAdded(XuluModel model) {
    // Event-Manager ankoppeln
    model.addModelListener(this);
  }

  /**
   * Wird ein Modell aus dem Modell-Manager entfernt, entkoppelt sich der Event-Manager
   * von diesem Modell und loescht alle mit dem Modell verknuepften Event-Handler.
   * @param model aus dem Modell-Manager entferntes Modell.
   */
  private void performModelRemoved(XuluModel model) {
    // alle EventHandler entfernen!
    Vector list = (Vector)objectEventHandler.get(model);
    if ( list!=null )
      list = (Vector)list.clone();
    for (int i=0; list!=null && i<list.size(); i++) {
      EventHandler eh = (EventHandler)list.elementAt(i);
      if ( eh.getObject().equals(model) )
        remove(eh);
    }

    // Event-Manager abkoppeln
    model.removeModelListener(this);
  }

}
