/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.io.InputStream;
import java.io.File;
import java.io.FileInputStream;

import schmitzm.lang.NamedObject;
import schmitzm.lang.AbstractNamedObject;
import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.io.IODefinition;

/**
 * Ein <code>ScriptInterpreter</code> fuehrt eine Reihe von Befehlen
 * in einer Xulu-Umgebung aus. Die Befehle erhaelt er ueber einen Eingabe-Stream.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface ScriptInterpreter extends NamedObject {
  /**
   * Fuehrt ein Script aus.
   * @param in   Eingabe-Stream in dem das Skript zu finden ist.
   * @param appl Applikation, in der das Skript auszufuehren ist
   * @exception java.lang.Exception wenn ein Fehler bei der Ausfuehrung aufritt
   */
  public void execute(InputStream in, XuluModellingPlatform appl) throws Exception;

  ///////////////////////////////////////////////////////////////////
  ////////////////   Script-Ausfuehungs-Definition   ////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt die Definition einer Script-Ausfuehrung dar.
   * Die besteht aus einem {@link ScriptInterpreter} und einem Quell-Objekt
   * aus dem der {@link ScriptInterpreter} das Script einliest.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static abstract class ScriptExecutionDefinition extends AbstractNamedObject {
    /** Speichert das Quell-Objekt, aus dem das Script gelesen wird. */
    protected Object source = null;
    /** Speichert den {@link ScriptInterpreter}, der das Objekt einliest. */
    protected ScriptInterpreter interpreter = null;

    /**
     * Erzeugt eine neue Import-Definition.
     * @param source      Objekt aus dem das Script eingelesen wird; z.B. eine Datei
     * @param interpreter {@link ScriptInterpreter} der das Script einliest
     */
    public ScriptExecutionDefinition(Object source, ScriptInterpreter interpreter) {
      if ( source == null )
        throw new IllegalArgumentException("null-source not allowed!");
      if ( interpreter == null )
        throw new IllegalArgumentException("null-interpreter not allowed!");
      this.source = source;
      this.interpreter = interpreter;
      setName( toString() );
    }

    /**
     * Liefert den Script-Interpreter, die das Script einliest.
     */
    public ScriptInterpreter getScriptInterpreter() {
      return interpreter;
    }

    /**
     * Liefert den Quell-Objekt, aus dem das Script eingelesen wird.
     */
    public Object getScriptSource() {
      return source;
    }

    /**
     * Liefert einen Stream zum Quell-Objekt.
     */
    public abstract InputStream getSourceAsStream() throws Exception;


    /**
     * Fuehrt das durch die {@code ScriptDefinition} definitierten Script
     * aus.
     * @param appl Xulu-Instanz unter der das Script ausgefuehrt wird.
     * @throws Exception
     */
    public void execute(XuluModellingPlatform appl) throws Exception {
      getScriptInterpreter().execute( getSourceAsStream(),appl );
    }

    /**
     * Vergleicht die {@code ScriptExecutionDefinition} mit einer anderen.
     * Beide sind gleich, gdw. beide den gleichen {@link ScriptInterpreter}
     * und die gleiche Script-Quelle enthalten.
     * @param o andere {@code IODefinition}
     */
    public boolean equals(Object o) {
      return o != null &&
             o instanceof ScriptExecutionDefinition &&
             ((ScriptExecutionDefinition)o).getScriptInterpreter().equals( getScriptInterpreter() ) &&
             ((ScriptExecutionDefinition)o).getScriptSource().equals( getScriptSource() );
    }
  }

  /**
   * Diese Klasse stellt die Definition einer Script-Ausfuehrung dar.
   * Die besteht aus einem {@link ScriptInterpreter} und einer Quell-Datei
   * aus der der {@link ScriptInterpreter} das Script einliest.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ScriptFileExecutionDefinition extends ScriptExecutionDefinition {
    /**
     * Erzeugt eine neue Import-Definition.
     * @param sourceFile  Datei aus der das Script eingelesen wird
     * @param interpreter {@link ScriptInterpreter} der das Script einliest
     */
    public ScriptFileExecutionDefinition(File sourceFile, ScriptInterpreter interpreter) {
      super(sourceFile,interpreter);
      this.setName(sourceFile.getAbsolutePath());
    }

    /**
     * Liefert die Quell-Datei, aus der das Script eingelesen wird.
     */
    public File getScriptSource() {
      return (File)source;
    }

    /**
     * Setzt das Xulu-Arbeitsverzeichnis auf das der Script-Datei und fuehrt
     * das Skript aus.
     * @param appl Xulu-Instanz unter der das Script ausgefuehrt wird.
     */
    public void execute(XuluModellingPlatform appl) throws Exception {
      appl.setWorkingDirectory( getScriptSource().getParentFile() );
      super.execute(appl);
    }

    /**
     * Liefert einen {@link FileInputStream} zur Quell-Datei.
     */
    public FileInputStream getSourceAsStream() throws Exception {
      return new FileInputStream( getScriptSource() );
    }
  }

}
