/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.ObjectCloseEvent;

import edu.bonn.xulu.appl.VisualisationTool;

/**
 * Zwischen Xulu-Datenobjekt ("source") und visualisiertem Objekt ("visObject")
 * besteht in der Regel ein Unterschied, da das Datenobjekt meist nicht direkt
 * dargestellt werden kann. Es wird ein eigenes Objekt für die Visualisierung
 * erstellt. Diese Klasse stellt hierfuer die Schnittstelle zwischen dem
 * allgemeinem Objekt und dessen Visualisierungsobjekt dar.<br>
 * Während ein Visualisierungstool fuer die Darstellung von vielen Objekten
 * verantwortlich ist, besteht eine 1:1-Beziehung zwischen einem Objekt und
 * einem <code>VisualisationUpdateListener</code>. M.a.W.: Ein Visualisierungstool
 * (PlugIn) muss für jedes dargestellte Objekt einen <code>VisualisationUpdateListener</code>
 * verwalten. U.U. können einem Objekt aber auch mehrere VisualisationUpdateListener
 * zugeordnet werden.<br>
 * Der <code>VisualisationUpdateListener</code> ist stark Tool-abhaengig und
 * sollte als innere Klasse des jeweiligen Visualisierungstools implementiert werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class VisualisationUpdateListener {//implements ObjectListener {
  /** Speichert das Visualisierungstool, in dem das Objekt dargestellt wird. */
  protected VisualisationTool visTool = null;
  /** Speichert das darzustellende Objekt */
  protected Object visObject = null;

//  /** Speichert, ob die automatische Aktualisierung (wenn ein Objekt sich
//   *  aendert) aktiviert ist. Default: <code>true</code> **/
//  private   boolean autoRefresh = false;

  /**
   * Erzeugt einen neuen Listener.
   * @param tool Visualisierungstool, an den der Listener angeschlossen ist
   * @param obj  Objekt, welches durch diesen Listener dargestellt wird
   */
  public VisualisationUpdateListener(VisualisationTool tool, Object obj) {
    this.visTool   = tool;
    this.visObject = obj;
  }

//  /**
//   * Verarbeitet ein <code>ObjectEvent</code>. Handelt es sich um ein
//   * {@link schmitzm.data.event.ObjectChangeEvent}, wird bei aktiviertem
//   * AutoRefresh ein {@link #refresh()} aufgerufen.<br>
//   * Bei einem {@link schmitzm.data.event.ObjectCloseEvent}
//   * wird immer {@link #close()} initiiert.
//   */
//  public void performObjectEvent(ObjectEvent e) {
//    if ( autoRefreshEnabled() && e instanceof ObjectChangeEvent && e.getSource().contains( getSourceObject() ) )
//      refresh();
//    if ( e instanceof ObjectCloseEvent )
//      close();
//  }

  /**
   * Liefert das Objekt, das durch diesen <code>VisualisationUpdateListener</code>
   * visualisiert wird.
   */
  public Object getSourceObject() {
    return this.visObject;
  }

//  /**
//   * Setzt den Modus fuer die automatische Aktualisierung der Visualisierung.
//   * Ist diese aktiviert erfolgt bei einem {@link schmitzm.data.event.ObjectChangeEvent}
//   * automatisch ein {@link #refresh()}. Standardmaessig ist sie deaktiviert!
//   */
//  public void setAutoRefresh(boolean autoRefresh) {
//    this.autoRefresh = autoRefresh;
//  }
//
//  /**
//   * Prueft, ob die automatische Aktualisierung der Visualisierung aktiviert ist.
//   * @see #setAutoRefresh(boolean)
//   */
//  public boolean autoRefreshEnabled() {
//    return autoRefresh;
//  }

  /**
   * Aktualisiert die Darstellung des Objekts.
   */
  public abstract void refresh();

  /**
   * Schliesst den VisualisationUpdateListener und entfernt die Darstellung
   * aus dem Visualisierungstool.
   */
  public abstract void close();

}
