/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.appl;

import java.util.Vector;
import java.util.Hashtable;

import schmitzm.lang.NamedObject;
import schmitzm.lang.LimitedVector;
import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.Invoker;
import edu.bonn.xulu.io.IODefinition;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.TypeMapping;
import edu.bonn.xulu.appl.ScriptInterpreter;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.model.XuluModel;


/**
 * In der Xulu-Registry werden alle dynamischen Komponenten registriert,
 * die in der Xulu-Applikation verwendet werden sollen. Hierzu zaehlen
 * <ul>
 * <li><b><u>Datentypen</u></b><br>
 *     alle Datentyp die im Xulu-Datenpool verwaltet werden koennen
 *     (Subklassen von {@link XuluObject})</li>
 * <li><b><u>DefaultFactorys</u></b><br>
 *     alle Factorys, die Standard-Instanzen der Datentypen erzeugen</li>
 * <li><b><u>ImportFactorys</u></b><br>
 *     alle Factorys, die Instanzen der Datentypen importieren</li>
 * <li><b><u>ExportFactorys</u></b><br>
 *     alle Factorys, die Instanzen der Datentypen exportieren</li>
 * <li><b><u>TypeMappings</u></b><br>
 *     Zuordnungen, welche welche Factorys fuer einen Datentyp zustaendig sind</li>
 * <li><b><u>Visualisierungstools</u></b><br>
 *     alle Tools die Xulu-Daten visualisieren koennen
 *     (Subklassen von {@link VisualisationTool})</li>
 * <li><b><u>Modell-Klassen</u></b><br>
 *     alle Modelle die im Xulu-Daten visualisieren koennen
 *     (Subklassen von {@link XuluModel})</li>
 * <li><b><u>Skript-Interpreter</u></b><br>
 *     alle Klassen die in der Xulu-Anwendung Skripte interpretieren koennen
 *     (Subklassen von {@link ScriptInterpreter})</li>
 * <li><b><u>Plugins</u></b><br>
 *     alle Klassen die in der Xulu-Anwendung als Plugin ausgefuehrt werden
 *     (Subklassen von {@link XuluPlugin})</li>
 * </ul>
 * Daneben bietet die Registry {@linkplain LimitedVector Listen}, in denen
 * die letzten Daten-Importe, Daten-Exporte und Script-Ausfuehrungen
 * hinterlegt sind.<br>
 * <br>
 * Die Registry fungiert als <code>AbstractObjectTraceable</code>, so dass
 * {@link ObjectListener} angeschlossen werden koennen, welche automatisch
 * durch ein {@link XuluRegistry.ChangeEvent} informiert werden, wenn sich der
 * Inhalt der Registry aendert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XuluRegistry extends AbstractObjectTraceable implements XuluComponent {
  private final XuluRegistry  THIS = this;

  private Vector    dataTypes         = new Vector<Class>();
  private Vector    defaultFac        = new Vector<InstantiationFactory>();
  private Vector    importFac         = new Vector<ImportFactory>();
  private Vector    exportFac         = new Vector<ExportFactory>();
  private Hashtable typeMappings      = new Hashtable<Class,TypeMapping>();
  private Vector    visTools          = new Vector<Class>();
  private Vector    modelClasses      = new Vector<Class>();
  private Vector    scriptInterpreter = new Vector<ScriptInterpreter>();
  private Vector    xuluPlugins       = new Vector<XuluPlugin>();
  private Vector    autostartPlugins  = new Vector<XuluPlugin>();
  private boolean   disposed          = false;

  // History der Importe, Skripte, ...
  private RecentImports recentImports = new RecentImports(10);
  private RecentScripts recentScripts = new RecentScripts(10);

  // Proxy-Listener, der die Ereignisse der recentImports, recentExports, ...
  // an die der Registry angeschlossenen Listener weiterleitet
  private ObjectListener listenerProxy = new ObjectListener() {
    public void performObjectEvent(ObjectEvent e) {
      e.expandSource(THIS);
      fireEvent( e );
    }
  };

  /**
   * Erzeugt eine neue (leere) Registry.
   */
  public XuluRegistry() {
    recentImports.addObjectListener( listenerProxy );
    recentImports.setDuplicatedAllowed(false);
    recentScripts.addObjectListener( listenerProxy );
    recentScripts.setDuplicatedAllowed(false);
  }

  /**
   * Zerstoert den Manager und alle darin enthaltenen Komponenten.
   */
  public void dispose() {
    XuluComponentUtil.checkDisposed(this);
    dataTypes.removeAllElements();
    defaultFac.removeAllElements();
    importFac.removeAllElements();
    exportFac.removeAllElements();
    typeMappings.clear();
    visTools.removeAllElements();
    modelClasses.removeAllElements();
    scriptInterpreter.removeAllElements();
    xuluPlugins.removeAllElements();
    disposed = true;
  }

  /**
   * Prueft, ob der Manager zerstoert ist.
   * @see #dispose()
   */
  public boolean isDisposed() {
    return disposed;
  }

  ///////////////////////////////////////////////////////////////////
  // Allgemeine Methoden
  ///////////////////////////////////////////////////////////////////
  /**
   * Ueberschreibt {@link AbstractObjectTraceable#createGeneralEvent()}, damit
   * ein {@link XuluRegistry.ChangeEvent XuluRegistry.ChangeEvent} zurueckgegeben wird.
   */
  protected ChangeEvent createGeneralEvent() {
    return new ChangeEvent(this,null,null);
  }

  /**
   * Fuegt ein Object in einen Vector ein.
   */
  private boolean register(Object o, Vector vector) {
    XuluComponentUtil.checkDisposed(this);
    if (vector.contains(o))
      return false;
    boolean ret = vector.add(o);
    this.fireEvent( new ChangeEvent(this,null,o) );
    return ret;
  }

  /**
   * Entfernt ein Object aus einem Vector.
   * Ist das Objekt nicht im Vector vorhanden wird nichts gemacht und
   * <code>false</code> zurueckgegeben.
   */
  private boolean unregister(Object o, Vector vector) {
    XuluComponentUtil.checkDisposed(this);
    if (!vector.contains(o))
      return false;
    boolean ret = vector.remove(o);
    this.fireEvent( new ChangeEvent(this,o,null) );
    return ret;
  }

  /**
   * Ersetzt ein Object in einem Vector durch ein anderes.
   * Ist das <code>o1</code> nicht im Vector vorhanden wird nichts gemacht und
   * <code>false</code> zurueckgegeben.
   */
  private boolean replace(Object o1, Object o2, Vector vector) {
    XuluComponentUtil.checkDisposed(this);
    int idx = vector.indexOf(o1);
    if ( idx < 0 )
      return false;
    vector.set(idx,o2);
    this.fireEvent( new ChangeEvent(this,o1,o2) );
    return true;
  }

  /**
   * Prueft, ob ein Object in einem Vector enthalten ist.
   */
  private boolean contains(Object o, Vector vector) {
    XuluComponentUtil.checkDisposed(this);
    return vector.contains(o);
  }

  /**
   * Prueft, ob diese angegebene Klasse zulaessig ist.
   */
  private void checkClass(Class c) {
    if ( c == null )
      throw new UnsupportedOperationException("null-Classes are not allowed!");
  }

  /**
   * Registriert ein Objekt. Dieses wird (wenn moeglich) fuer alle
   * folgenden Punkte registriert:
   * <ul>
   * <li><code>DefaultFactory</code></li>
   * <li><code>ImportFactory</code></li>
   * <li><code>ExportFactory</code></li>
   * <li><code>TypeMapping</code></li>
   * <li><code>VisualisationTool</code></li>
   * <li><code>Skript-Interpreter</code></li>
   * </ul>
   * @param obj ein beliebiges Objekt
   * @return <code>true</code> wenn mindestens eine neue Registrierung
   *         erfolgt ist
   */
  public boolean register(Object obj) {
    XuluComponentUtil.checkDisposed(this);

    boolean ret = false;
    if ( obj instanceof Class && XuluObject.class.isAssignableFrom((Class)obj) )
      ret |= registerDataType((Class)obj);
    if ( obj instanceof InstantiationFactory )
      ret |= registerDefaultFactory((InstantiationFactory)obj);
    if ( obj instanceof ImportFactory )
      ret |= registerImportFactory((ImportFactory)obj);
    if ( obj instanceof ExportFactory )
      ret |= registerExportFactory((ExportFactory)obj);
    if ( obj instanceof TypeMapping )
      ret |= registerTypeMapping((TypeMapping)obj);
    if ( obj instanceof Class && VisualisationTool.class.isAssignableFrom((Class)obj) )
      ret |= registerVisualisationTool((Class)obj);
    if ( obj instanceof ScriptInterpreter )
      ret |= registerScriptInterpreter((ScriptInterpreter)obj);

    return ret;
  }

  /**
   * Liefert alle Objekte eines Arrays, die eine bestimmte Bezeichnung (oder
   * Klassennamen) haben.
   * @param source Eingabe-Array
   * @param name Bezeichnung
   * @param outArray Ausgabe-Array (kann ein leerer Array sein!)
   */
  private static NamedObject[] findObjectsByName(NamedObject[] source, String name, Object[] outArray) {
    Vector out = new Vector();
    for (int i=0; i<source.length; i++)
      if ( source[i].getName().equals(name) ||
           source[i].getClass().getName().equals(name) ||
           source[i].getClass().getSimpleName().equals(name) )
        out.add(source[i]);
    return (NamedObject[])out.toArray(outArray);
  }

  ///////////////////////////////////////////////////////////////////
  // Datentypen
  ///////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob diese angegebene Klasse als Datentyp zulaessig ist.
   */
  private void checkTypeClass(Class c) {
    checkClass(c);
    if ( !XuluObject.class.isAssignableFrom( c ) )
      throw new UnsupportedOperationException( "Only subclasses of ".concat(XuluObject.class.getName()).concat(" can be registered as data types!") );
  }

  /**
   * Registriert einen Datentyp. Hierfuer koennen nur (Sub)Klassen von
   * {@link edu.bonn.xulu.data.XuluObject} angegeben werden.
   * @param type Datentyp
   * @return <code>false</code> falls der Datentyp bereits registriert ist
   * @exception java.lang.UnsupportedOperationException falls der angegebene
   *            Typ nicht einem {@link edu.bonn.xulu.data.XuluObject}
   *            entspricht
   */
  public boolean registerDataType(Class type) {
    checkTypeClass(type);
    return register(type,dataTypes);
  }

  /**
   * Entfernt die Registrierung eines Datentyps.
   * @param type Datentyp
   * @return <code>false</code> falls der Datentyp nicht registriert war
   */
  public boolean unregisterDataType(Class type) {
    return unregister(type,dataTypes);
  }

  /**
   * Ersetzt die Registrierung eines Datentyps durch einen anderen.
   * @param typeOld zu ersetzender Datentyp
   * @param typeNew neuer Datentyp
   * @return <code>false</code> falls <code>typeOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   * @exception java.lang.UnsupportedOperationException falls der neue
   *            Typ nicht einem {@link edu.bonn.xulu.data.XuluObject}
   *            entspricht
   */
  public boolean replaceDataType(Class typeOld, Class typeNew) {
    checkTypeClass(typeNew);
    return replace(typeOld,typeNew,dataTypes);
  }

  /**
   * Prueft, ob ein Datentyp registriert ist.
   */
  public boolean isDataTypeRegistered(Class type) {
    return contains(type,dataTypes);
  }

  /**
   * Liefert alle registrierten Datentypen.
   */
  public Class[] getDataTypes() {
    XuluComponentUtil.checkDisposed(this);
    return (Class[])dataTypes.toArray(new Class[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Standard-Factories
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert eine Standard-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory bereits registriert ist
   */
  public boolean registerDefaultFactory(InstantiationFactory fac) {
    return register(fac,defaultFac);
  }

  /**
   * Entfernt die Registrierung einer Standard-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory nicht registriert war
   */
  public boolean unregisterDefaultFactory(InstantiationFactory fac) {
    return unregister(fac,defaultFac);
  }

  /**
   * Ersetzt die Registrierung einer Standard-Factory durch eine andere.
   * @param facOld zu ersetzende Factory
   * @param facNew neue Factory
   * @return <code>false</code> falls <code>facOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceDefaultFactory(InstantiationFactory facOld, InstantiationFactory facNew) {
    return replace(facOld,facNew,defaultFac);
  }

  /**
   * Prueft, ob eine Standard-Factory registriert ist.
   */
  public boolean isDefaultFactoryRegistered(InstantiationFactory fac) {
    return contains(fac,defaultFac);
  }

  /**
   * Liefert alle registrierten Standard-Factorys.
   */
  public InstantiationFactory[] getDefaultFactories() {
    XuluComponentUtil.checkDisposed(this);
    return (InstantiationFactory[])defaultFac.toArray(new InstantiationFactory[0]);
  }

  /**
   * Liefert alle Default-Factorys, die einen bestimmten Namen besitzen.
   * @param name Bezeichnung der Factory
   * @return ein leerer <code>DeafultFactory</code>-Array, falls zu der Bezeichnung
   *         keine Factory registriert ist
   */
  public InstantiationFactory[] getDefaultFactories(String name) {
    return (InstantiationFactory[])findObjectsByName(getDefaultFactories(),name,new InstantiationFactory[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Import-Factories
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert eine Import-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory bereits registriert ist
   */
  public boolean registerImportFactory(ImportFactory fac) {
    return register(fac,importFac);
  }

  /**
   * Entfernt die Registrierung einer Import-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory nicht registriert war
   */
  public boolean unregisterImportFactory(ImportFactory fac) {
    return unregister(fac,importFac);
  }

  /**
   * Ersetzt die Registrierung einer Import-Factory durch eine andere.
   * @param facOld zu ersetzende Factory
   * @param facNew neue Factory
   * @return <code>false</code> falls <code>facOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceImportFactory(ImportFactory facOld, ImportFactory facNew) {
    return replace(facOld,facNew,importFac);
  }

  /**
   * Prueft, ob eine Import-Factory registriert ist.
   */
  public boolean isImportFactoryRegistered(ImportFactory fac) {
    return contains(fac,importFac);
  }

  /**
   * Liefert alle registrierten Import-Factorys.
   */
  public ImportFactory[] getImportFactories() {
    XuluComponentUtil.checkDisposed(this);
    return (ImportFactory[])importFac.toArray(new ImportFactory[0]);
  }

  /**
   * Liefert alle Import-Factorys, die einen bestimmten Namen besitzen.
   * @param name Bezeichnung der Factory
   * @return ein leerer <code>ImportFactory</code>-Array, falls zu der Bezeichnung
   *         keine Factory registriert ist
   */
  public ImportFactory[] getImportFactories(String name) {
    return (ImportFactory[])findObjectsByName(getImportFactories(),name,new ImportFactory[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Export-Factories
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert eine Export-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory bereits registriert ist
   */
  public boolean registerExportFactory(ExportFactory fac) {
    return register(fac,exportFac);
  }

  /**
   * Entfernt die Registrierung einer Export-Factory.
   * @param fac Factory
   * @return <code>false</code> falls die Factory nicht registriert war
   */
  public boolean unregisterExportFactory(ExportFactory fac) {
    return unregister(fac,exportFac);
  }

  /**
   * Ersetzt die Registrierung einer Export-Factory durch eine andere.
   * @param facOld zu ersetzende Factory
   * @param facNew neue Factory
   * @return <code>false</code> falls <code>facOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceExportFactory(ExportFactory facOld, ExportFactory facNew) {
    return replace(facOld,facNew,exportFac);
  }

  /**
   * Prueft, ob eine Export-Factory registriert ist.
   */
  public boolean isExportFactoryRegistered(ExportFactory fac) {
    return contains(fac,exportFac);
  }

  /**
   * Liefert alle registrierten Export-Factorys.
   */
  public ExportFactory[] getExportFactories() {
    XuluComponentUtil.checkDisposed(this);
    return (ExportFactory[])exportFac.toArray(new ExportFactory[0]);
  }

  /**
   * Liefert alle Export-Factorys, die einen bestimmten Namen besitzen.
   * @param name Bezeichnung der Factory
   * @return ein leerer <code>ExportFactory</code>-Array, falls zu der Bezeichnung
   *         keine Factory registriert ist
   */
  public ExportFactory[] getExportFactories(String name) {
    return (ExportFactory[])findObjectsByName(getExportFactories(),name,new ExportFactory[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Type-Mappings
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert ein TypeMapping. War bereits ein TypeMapping fuer den
   * entsprechenden Datentyp registriert, wird das bestehende TypeMapping
   * ersetzt.<br>
   * Sind die darin enthaltenen Komponenten (Datentyp, Factorys, ...) noch
   * nicht registriert, werden diese ebenfalls in die XuluRegistry aufgenommen.
   * @param mapping TypeMapping
   * @return <code>true</code>
   */
  public boolean registerTypeMapping(TypeMapping mapping) {
    // Alle Komponenten des Mappings registrieren
    registerDataType(mapping.getType());
    registerDefaultFactory(mapping.getDefaultFactory());
    ImportFactory[] iFac = mapping.getImportFactories();
    for (int i=0; i<iFac.length; i++)
      registerImportFactory(iFac[i]);
    ExportFactory[] eFac = mapping.getExportFactories();
    for (int i=0; i<eFac.length; i++)
      registerExportFactory(eFac[i]);
    // Wenn alle Komponenten erfolgreich registriert wurden, wird
    // das TypeMapping registriert
    typeMappings.put(mapping.getType(),mapping);
    this.fireEvent( new ChangeEvent(this,null,mapping) );
    return true;
  }

  /**
   * Entfernt die Registrierung eines TypeMappings.
   * @param mapping TypeMapping
   * @return <code>false</code> falls das TypeMapping nicht registriert war
   */
  public boolean unregisterTypeMapping(TypeMapping mapping) {
    XuluComponentUtil.checkDisposed(this);
    if ( typeMappings.remove(mapping.getType()) != null ) {
      fireEvent( new ChangeEvent(this,mapping,null) );
      return true;
    }
    return false;
  }

  /**
   * Ersetzt die Registrierung eines TypeMappings durch ein anderes.
   * @param mapOld zu ersetzendes TypeMapping
   * @param mapNew neues TypeMapping
   * @return <code>false</code> falls <code>mapOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceTypeMapping(TypeMapping mapOld, TypeMapping mapNew) {
    // Da die Hashtable keine feste Reihenfolge impliziert, entspricht
    // replace(TypeMapping,TypeMapping) einfach dem Entfernen und Hinzufuegen
    // Die Methode wird nur der Vollstaendigkeit halber implementiert.
    if ( !unregisterTypeMapping(mapOld) )
      return false;
    return registerTypeMapping(mapNew);
  }

  /**
   * Prueft, ob ein TypeMapping (fuer den Datentyp) registriert ist.
   */
  public boolean isTypeMappingRegistered(TypeMapping mapping) {
    XuluComponentUtil.checkDisposed(this);
    return typeMappings.containsKey(mapping.getType());
  }

  /**
   * Liefert alle registrierten TypeMappings.
   */
  public TypeMapping[] getTypeMappings() {
    XuluComponentUtil.checkDisposed(this);
    return (TypeMapping[])typeMappings.values().toArray(new TypeMapping[0]);
  }

  /**
   * Liefert ein TypeMapping aus der Xulu-Registry.
   * @param type Datentyp
   * @return <code>null</code> falls zu dem Datentyp kein TypeMapping registriert
   *         ist
   */
  public TypeMapping getTypeMapping(Class type) {
    XuluComponentUtil.checkDisposed(this);
    return (TypeMapping)typeMappings.get(type);
  }

  /**
   * Liefert eine Datentyp-Factory aus der Xulu-Registry.
   * @param type Datentyp
   * @return <code>null</code> falls zu dem Datentyp kein TypeMapping registriert
   *         ist
   */
  public InstantiationFactory getDefaultFactory(Class type) {
    TypeMapping tm = getTypeMapping(type);
    return tm==null ? null : tm.getDefaultFactory();
  }

  /**
   * Liefert alle Import-Factorys zu einem Datentyp.
   * @param type Datentyp
   * @return ein leerer <code>ImportFactory</code>-Array, falls zu dem Datentyp
   *         kein TypeMapping registriert ist
   */
  public ImportFactory[] getImportFactories(Class type) {
    TypeMapping tm = getTypeMapping(type);
    return tm==null ? new ImportFactory[0] : tm.getImportFactories();
  }

  /**
   * Liefert eine Export-Factory aus der Xulu-Registry.
   * @param type Datentyp
   * @return ein leerer <code>ExportFactory</code>-Array, falls zu dem Datentyp
   *         kein TypeMapping registriert ist
   */
  public ExportFactory[] getExportFactories(Class type) {
    TypeMapping tm = getTypeMapping(type);
    return tm==null ? new ExportFactory[0] : tm.getExportFactories();
  }

  ///////////////////////////////////////////////////////////////////
  // Visualisierungs-Tools
  ///////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob diese angegebene Klasse als Visualisierungstool zulaessig ist.
   */
  private void checkVisualisationToolClass(Class c) {
    checkClass(c);
    if ( !VisualisationTool.class.isAssignableFrom( c ) )
      throw new UnsupportedOperationException( "Only subclasses of ".concat(VisualisationTool.class.getName()).concat(" can be registered as visualisation tool!") );
    // Pruefen, ob Klasse einen Standard-Konstruktor hat
    try {
      c.getConstructor();
    } catch ( Exception err ) {
      throw new UnsupportedOperationException( "Visualisation tool must have a default constructor!" );
    }
  }

  /**
   * Registriert ein Visualisierungstool.
   * @param toolClass Visualisierungstool
   * @return <code>false</code> falls das Visualisierungstool bereits registriert ist
   */
  public boolean registerVisualisationTool(Class toolClass) {
    checkVisualisationToolClass(toolClass);
    return register(toolClass,visTools);
  }

  /**
   * Entfernt die Registrierung eines Visualisierungstool.
   * @param toolClass Visualisierungstool
   * @return <code>false</code> falls das Visualisierungstool nicht registriert war
   */
  public boolean unregisterVisualisationTool(Class toolClass) {
    return unregister(toolClass,visTools);
  }

  /**
   * Ersetzt die Registrierung eines Visualisierungstools durch ein andere.
   * @param toolOld zu ersetzendes Visualisierungstool
   * @param toolNew neues Visualisierungstool
   * @return <code>false</code> falls <code>toolOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceVisualisationTool(Class toolOld, Class toolNew) {
    checkVisualisationToolClass(toolNew);
    return replace(toolOld,toolNew,visTools);
  }

  /**
   * Prueft, ob eine Visualisierungstool registriert ist.
   */
  public boolean isVisualisationToolRegistered(Class toolClass) {
    return contains(toolClass,visTools);
  }

  /**
   * Liefert alle registrierten Visualisierungstools.
   */
  public Class[] getVisualisationTools() {
    XuluComponentUtil.checkDisposed(this);
    return (Class[])visTools.toArray(new Class[0]);
  }

  /**
   * Diese Ereignis wird erzeugt (und an alle {@link schmitzm.data.event.ObjectListener}
   * gesendet), wenn sich der Inhalt der Xulu-Registrierung aendert (ein Objekt hingefuegt,
   * entfernt oder ersetzt wurde).
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ChangeEvent extends ObjectChangeEvent {
    /**
     * Erzeugt ein neues Ereignis.
     * @param registry XuluRegistry, die sich geaendert wird
     * @param oldObj   altes Objekt (<code>null</code> falls ein Registry-Objekt
     *                 hinzugefuegt wurde)
     * @param newObj   neues Objekt (<code>null</code> falls ein Registry-Objekt
     *                 geloescht wurde)
     */
    public ChangeEvent(XuluRegistry registry, Object oldObj, Object newObj) {
      super( new Invoker(registry, new Invoker(oldObj)), oldObj, newObj );
    }
  }

  ///////////////////////////////////////////////////////////////////
  // Modell-Klassen
  ///////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob diese angegebene Klasse als Modell zulaessig ist.
   */
  private void checkModelClass(Class c) {
    checkClass(c);
    if ( !XuluModel.class.isAssignableFrom( c ) )
      throw new UnsupportedOperationException( "Only subclasses of ".concat(XuluModel.class.getName()).concat(" can be registered as model!") );
    // Pruefen, ob Klasse einen Standard-Konstruktor hat
    try {
      c.getConstructor();
    } catch ( Exception err ) {
      throw new UnsupportedOperationException( "Model class must have a default constructor!" );
    }
  }

  /**
   * Registriert eine Modellklasse.
   * @param modelClass Modellklasse
   * @return <code>false</code> falls die Modellklasse bereits registriert ist
   */
  public boolean registerModelClass(Class modelClass) {
    checkModelClass(modelClass);
    return register(modelClass,modelClasses);
  }

  /**
   * Entfernt die Registrierung einer Modellklasse.
   * @param modelClass Modellklasse
   * @return <code>false</code> falls die Modellklasse nicht registriert war
   */
  public boolean unregisterModelClass(Class modelClass) {
    return unregister(modelClass,modelClasses);
  }

  /**
   * Ersetzt die Registrierung einer Modellklasse durch eine andere.
   * @param modelOld zu ersetzende Modellklasse
   * @param modelNew neue Modellklasse
   * @return <code>false</code> falls <code>modelOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceModelClass(Class modelOld, Class modelNew) {
    checkModelClass(modelNew);
    return replace(modelOld,modelNew,modelClasses);
  }

  /**
   * Prueft, ob eine Modellklasse registriert ist.
   */
  public boolean isModelClassRegistered(Class modelClass) {
    return contains(modelClass,modelClasses);
  }

  /**
   * Liefert alle registrierten Modellkassen.
   */
  public Class[] getModelClasses() {
    XuluComponentUtil.checkDisposed(this);
    return (Class[])modelClasses.toArray(new Class[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Skript-Interpreter
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert einen Skript-Interpreters.
   * @param interpreter {@link ScriptInterpreter}-Instanz
   * @return <code>false</code> falls die Klasse bereits registriert ist
   */
  public boolean registerScriptInterpreter(ScriptInterpreter interpreter) {
    return register(interpreter,scriptInterpreter);
  }

  /**
   * Entfernt die Registrierung eines Skript-Interpreters.
   * @param interpreter {@link ScriptInterpreter}-Instanz
   * @return <code>false</code> falls die Klasse nicht registriert war
   */
  public boolean unregisterScriptInterpreter(ScriptInterpreter interpreter) {
    return unregister(interpreter,scriptInterpreter);
  }

  /**
   * Ersetzt die Registrierung einer Skript-Interpreters durch einen andere.
   * @param interpreterOld zu ersetzender Skript-Interpreter
   * @param interpreterNew neuer Skript-Interpreter
   * @return <code>false</code> falls <code>interpreterOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceScriptInterpreter(ScriptInterpreter interpreterOld, ScriptInterpreter interpreterNew) {
    return replace(interpreterOld,interpreterNew,scriptInterpreter);
  }

  /**
   * Prueft, ob ein Skript-Interpreter registriert ist.
   */
  public boolean isScriptInterpreterRegistered(ScriptInterpreter interpreter) {
    return contains(interpreter,scriptInterpreter);
  }

  /**
   * Liefert alle registrierten Skript-Interpreter.
   */
  public ScriptInterpreter[] getScriptInterpreter() {
    XuluComponentUtil.checkDisposed(this);
    return (ScriptInterpreter[])scriptInterpreter.toArray(new ScriptInterpreter[0]);
  }

  /**
   * Liefert alle Skript-Interpreter, die einen bestimmten Namen besitzen.
   * @param name Bezeichnung des Skript-Interpreter
   * @return ein leerer <code>Skript-Interpreter</code>-Array, falls zu der Bezeichnung
   *         kein Skript-Interpreter registriert ist
   */
  public ScriptInterpreter[] getScriptInterpreter(String name) {
    return (ScriptInterpreter[])findObjectsByName(getScriptInterpreter(),name,new ScriptInterpreter[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Xulu-Plugins
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert ein Xulu-Plugin.
   * @param plugin {@link XuluPlugin}-Instanz
   * @return <code>false</code> falls die Klasse bereits registriert ist
   */
  public boolean registerPlugin(XuluPlugin plugin) {
    return register(plugin, xuluPlugins);
  }

  /**
   * Entfernt die Registrierung eines Xulu-Plugins. Wenn das
   * Plugin auch als Autostart-Plugin registriert war, wird dieses ebenfalls
   * entfernt.
   * @param plugin {@link XuluPlugin}-Instanz
   * @return <code>false</code> falls die Klasse nicht registriert war
   */
  public boolean unregisterPlugin(XuluPlugin plugin) {
    unregisterAutostartPlugin(plugin);
    return unregister(plugin,xuluPlugins);
  }

  /**
   * Ersetzt die Registrierung eines Xulu-Plugins durch eine andere. Wenn das
   * Plugin auch als Autostart-Plugin registriert war, wird dieses ebenfalls
   * ersetzt.
   * @param pluginOld zu ersetzendes Plugin
   * @param pluginNew neues Plugin
   * @return <code>false</code> falls <code>pluginOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replacePlugin(XuluPlugin pluginOld, XuluPlugin pluginNew) {
    replaceAutostartPlugin(pluginOld, pluginNew);
    return replace(pluginOld, pluginNew, xuluPlugins);
  }

  /**
   * Prueft, ob ein Plugin registriert ist.
   */
  public boolean isPluginRegistered(XuluPlugin plugin) {
    return contains(plugin, xuluPlugins);
  }

  /**
   * Liefert alle registrierten Xulu-Plugins.
   */
  public XuluPlugin[] getPlugins() {
    XuluComponentUtil.checkDisposed(this);
    return (XuluPlugin[]) xuluPlugins.toArray(new XuluPlugin[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // Autostart-Plugins
  ///////////////////////////////////////////////////////////////////
  /**
   * Registriert ein Xulu-Plugin als Autostart-Plugin. Wenn noch nicht
   * registriert, wird es (automatisch) auch in die Liste der Xulu-Plugins
   * aufgenommen.
   * @param plugin {@link XuluPlugin}-Instanz
   * @return <code>false</code> falls das Plugin bereits registriert ist
   */
  public boolean registerAutostartPlugin(XuluPlugin plugin) {
    registerPlugin(plugin);
    return register(plugin, autostartPlugins);
  }

  /**
   * Entfernt die Registrierung eines Xulu-Plugins als Autostart-Plugin.
   * Als "normalen" Plugin bleibt es jedoch registriert.
   * @param plugin {@link XuluPlugin}-Instanz
   * @return <code>false</code> falls die Klasse nicht registriert war
   */
  public boolean unregisterAutostartPlugin(XuluPlugin plugin) {
    return unregister(plugin,autostartPlugins);
  }

  /**
   * Ersetzt die Registrierung eines Xulu-Plugins als Autostart-Plugin
   * durch eine andere.
   * @param pluginOld zu ersetzendes Plugin
   * @param pluginNew neues Plugin
   * @return <code>false</code> falls <code>pluginOld</code> nicht registriert war
   *         und deshalb keine Ersetzung moeglich ist
   */
  public boolean replaceAutostartPlugin(XuluPlugin pluginOld, XuluPlugin pluginNew) {
    boolean success = replace(pluginOld, pluginNew, autostartPlugins);
    // Wenn die Ersetzung erfolgreich war, das neue Plugin auch in die
    // Liste der normalen Plugins aufnehmen
    if ( success )
      registerPlugin(pluginNew);
    return success;
  }

  /**
   * Prueft, ob ein Plugin als Autostart-Plugin registriert ist.
   */
  public boolean isAutostartPluginRegistered(XuluPlugin plugin) {
    return contains(plugin, autostartPlugins);
  }

  /**
   * Liefert alle registrierten Autostart-Plugins.
   */
  public XuluPlugin[] getAutostartPlugins() {
    XuluComponentUtil.checkDisposed(this);
    return (XuluPlugin[]) autostartPlugins.toArray(new XuluPlugin[0]);
  }

  ///////////////////////////////////////////////////////////////////
  ///////////////   Letzte Importe, Exporte, ...   //////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert eine Liste, in der die letzten Daten-Importe hinterlegt werden.
   */
  public RecentImports getRecentImports() {
    return recentImports;
  }

  /**
   * Liefert eine Liste, in der die letzten Script-Aufrufe hinterlegt werden.
   */
  public RecentScripts getRecentScripts() {
    return recentScripts;
  }

}
