/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.data;

import java.util.Enumeration;

import edu.bonn.xulu.appl.DataPool;
import schmitzm.data.property.PropertySet;
import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import schmitzm.data.property.PropertyType;
import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.ObjectCloseEvent;
import schmitzm.data.event.NameChangeEvent;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.Invoker;
// nur fuer Doku
import schmitzm.data.ObjectStructure;
import schmitzm.data.property.ValuePropertyType;


/**
 * Diese Klasse bildet die Basisimplementierung eines Xulu-Datenobjekts. Diese
 * stellen die Objekte dar, die im {@link DataPool Xulu-Datenpool} verwaltet
 * werden koennen.
 * <code>AbstractXuluObject</code> ist <b>fest</b> aus einer Menge an Properties
 * zusammengesetzt, welche in einem {@link schmitzm.data.property.PropertySet}
 * verwaltet werden. Diese sind von der jeweiligen Implementierung (im
 * Konstruktor) festzulegen. Moeglichkeiten, dynamisch Eigenschaften hinzufuegen
 * oder zu entfernen besteht nicht. Hierzu kann {@link DynamicXuluObject} genutzt
 * werden.<br>
 * Jedes Xulu-Datenobjekt erhaelt (automatisch) bei der Erzeugung eine ID, mit
 * der es im Datenpool (sogar ueber alle Datenpool-Instanzen!) eindeutig
 * identifiziert werden kann ({@link DataPool#generateUniqueID()}).
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractXuluObject extends AbstractObjectTraceable implements XuluObject {
  /** Zaehlt die Instanzen fuer die Default-Beschreibung */
  private static int instanceNo = 0;

  /** Speichert die Eigenschaften des Xulu-Objekts. Darf nicht mit einem neuen
   *  <code>PropertySet</code> belegt werden (<code>final</code>),
   *  da es mit einem {@link ObjectListener} belegt wurde, der auf Aenderungen
   *  an dem <code>PropertySet</code> reagiert (jeder hinzugefuegten Eigenschaft
   *  wird ein {@link ObjectListener} hinzugefuegt, damit das <code>XuluObject</code>
   *  Aenderungen an seinen Properties mitbekommt und ggf. weiterleiten kann).*/
  protected final PropertySet propSet = new PropertySet();

  /** Flag das anzeigt, dass das Objekt zerstoert ist. */
  private boolean disposed = false;

  /** Beschreibung fuer den Xulu-Datenpool */
  private String desc = "";

  /** eindeutige ID mit der das Datenobjekt im Datenpool identifizierbar ist */
  private long id = 0;

  /**
   * Lauscht auf Aenderungen an den Properties und propagiert diese an
   * alle Listener auf dem <code>XuluObject</code> weiter */
  protected ObjectListener propertyChangeListener = null;

  ///////////////////////////////////////////////////////////////////
  // Konstruktoren
  ///////////////////////////////////////////////////////////////////
  /**
   * Erzeugt eine neue <code>XuluObject</code>-Instanz.
   * @param desc Beschreibung, welche im Datenpool angezeigt wird (wird
   *         <code>null</code> uebergeben, wird eine Standardbeschreibung
   *         erzeugt)
   */
  public AbstractXuluObject(String desc) {
    instanceNo++; // Anzahl an Instanzen zaehlen (nur fuer Default-Beschreibung!)
    // Standard-Beschreibung generieren
    if ( desc == null )
      desc = this.getClass().getSimpleName().concat(String.valueOf(instanceNo));
    setDescription(desc);

    this.id = DataPool.generateUniqueID();

    // Listener achtet auf Aenderungen (an Eigenschaften des XuluObject)
    // und propagiert diese an die Listener des XuluObject weiter
    this.propertyChangeListener = new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        propagateDeeperEvent(e);
      }
    };

    // Listener achtet darauf, ob dem XuluObject eine Property hingefuegt
    // oder entfernt wird
    this.propSet.addObjectListener( new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        if ( e instanceof ObjectChangeEvent ) {
          ObjectChangeEvent oce = (ObjectChangeEvent)e;
          // Property wurde entfernt > Listener entfernen
          if ( oce.getNewValue() == null )
            ((Property)oce.getOldValue()).removeObjectListener(propertyChangeListener);
          // Property wurde hinzugefuegt > Listener hinzufuegen
          if ( oce.getOldValue() == null )
            ((Property)oce.getNewValue()).addObjectListener(propertyChangeListener);
        }
      }
    });

  }

  /**
   * Erweitert die Ereigniskette eines Events, das von einem Unterobjekt
   * erhalten wurde, und gibt es an alle Listener weiter.
   */
  private void propagateDeeperEvent(ObjectEvent e) {
    e.expandSource(this);
    fireEvent(e);
  }

  /**
   * Erzeugt eine neue <code>XuluObject</code>-Instanz. Diese erhaelt eine Default-Beschreibung,
   * bestehend aus dem Klassennamen und einer laufenden Nummer.
   */
  public AbstractXuluObject() {
    this(null);
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von XuluObject
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert die Beschreibung des Xulu-Objekts fuer den Datenpool.
   */
  public String getDescription() {
    return this.desc;
  }

  /**
   * Setzt die Beschreibung des Xulu-Objekts fuer den Datenpool. Aendert sich
   * der Name, wird ein {@link NameChangeEvent} an alle Listener weitergeleitet.
   * @param desc Beschreibung fuer das Objekt
   */
  public void setDescription(String desc) {
    if ( desc == null || desc.trim().equals("") )
      throw new XuluDataException("Object description can not be empty!");
    String oldDesc = getDescription();
    this.desc = desc;
    if ( !desc.equals(oldDesc) )
      this.fireEvent( new NameChangeEvent( new Invoker(this),oldDesc,desc ) );
  }

  /**
   * Liefert die Beschreibung des Xulu-Objekts fuer den Datenpool. Synonym fuer
   * {@link #getDescription()}.
   */
  public String getName() {
    return getDescription();
  }

  /**
   * Setzt die Beschreibung des Xulu-Objekts fuer den Datenpool. Synonym fuer
   * {@link #setDescription(String)}.
   * @param desc Beschreibung fuer das Objekt
   */
  public void setName(String desc) {
    setDescription(desc);
  }


  /**
   * Liefert die ID, mit der das Objekt in allen Datenpools eindeutig
   * angesprochen werden kann.
   */
  public long getID() {
    return this.id;
  }

  /**
   * Zerstoert das Xulu-Objekt, indem alle seine Properties zerstoert
   * werden. Unterklassen sollten diese Methode ueberschreiben, wenn
   * weitere Ressourcen bei der Zerstoerung des Objekts oder seiner
   * Propertys notwendig sind!!
   * @see Property#dispose()
   */
  public void dispose() {
    this.fireEvent( new ObjectCloseEvent(new Invoker(this)) );
    Property[] prop = getProperties();
    for (int i=0; i<prop.length; i++)
      prop[i].dispose();
    disposed = true;
  }

  /**
   * Prueft, ob das Objekt zerstoert ist.
   */
  public boolean isDisposed() {
    return disposed;
  }

  /**
   * Prueft, ob aktuell auf eine Property des Xulu-Objekts zugegriffen
   * wird.
   */
  public boolean isAccessed() {
    Property[] prop = getProperties();
    for (int i=0; i<prop.length; i++)
      if (prop[i].hasAccess())
        return true;
    return false;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von Properties
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert die Anzahl an Eigenschaften des Xulu-Objekts.
   */
  public int getPropertyCount() {
    return propSet.getPropertyCount();
  }

  /**
   * Liefert die Eigenschaften des Xulu-Objekts als Array.
   */
  public Property[] getProperties() {
    return propSet.getProperties();
  }

  /**
   * Liefert alle Eigenschaften des Xulu-Objekts, die einem bestimmten
   * Typ angehoeren. Als Typ kann sowohl ein bestimmter <code>Property</code>-Untertyp
   * angegeben werden (z.B. {@link schmitzm.data.property.ListProperty}),
   * als auch ein Objekt-Typ (Typ den eine Property aufnehmen kann; z.B.
   * <code>int.class</code>).<br>
   * Da die Methode wiederum eine <code>Properties</code>-Instanz liefert,
   * lassen sich recht einfach verschiedene Filter hintereinander anwenden:<br>
   * <br>
   * <center><code>
   * Properties.getProperties(ListProperty.class).getProperties(int.class).getPropertyNames()
   * </code></center>
   * <br>
   * liefert z.B. die Namen aller Listen-Eigenschaften, die <code>int</code>-Werte
   * aufnehmen koennen.
   */
  public Properties getProperties(Class type) {
    return propSet.getProperties(type);
  }

  /**
   * Liefert die Namen aller Eigenschaften des Xulu-Objekts.
   */
  public String[] getPropertyNames() {
    return propSet.getPropertyNames();
  }

  /**
   * Liefert die Typen aller Eigenschaften des Xulu-Objekts.
   */
  public PropertyType[] getPropertyTypes() {
    return propSet.getPropertyTypes();
  }

  /**
   * Liefert eine bestimmte Eigenschaft des Xulu-Objekts.
   * @param name Name der Eigenschaft
   * @return <code>null</code> falls das Xulu-Objekt keine Eigenschaft dieses
   *         Namens enthaelt
   */
  public Property getProperty(String name) {
    return propSet.getProperty(name);
  }

  /**
   * Prueft, ob das Xulu-Objekt eine bestimmte Eigenschaft enthaelt.
   * @param name Name der Eigenschaft
   */
  public boolean containsProperty(String name) {
    return propSet.containsProperty(name);
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>true</code>.
   */
  public boolean isStructureNamed() {
    return true;
  }

  /**
   * Liefert die Bezeichnung des Xulu-Objekts.
   * @see #getDescription()
   */
  public String getStructureName() {
    return getDescription();
  }

  /**
   * Liefert die Anzah1 an Properties des Xulu-Objekt.
   * @see #getPropertyCount()
   */
  public int getAttrCount() {
    return propSet.getAttrCount();
  }

  /**
   * Liefert eine Liste, in der genau so viele Elemente enthaelt, wie das
   * Xulu-Objekt Eigenschaften hat. Alle Listen-Elemente sind vom Typ
   * {@link ObjectStructure}.
   */
  public Enumeration<ObjectStructure> getAttrTypes() {
    return propSet.getAttrTypes();
  }

  /**
   * Vergleicht die Struktur der Property mit einer anderen auf Gleichheit.
   * @param object eine anderes {@link ObjectStructure}-Objekt
   * @see ValuePropertyType#equalsInStructure(Object)
   */
  public boolean equalsInStructure(Object object) {
    return propSet.equalsInStructure(object);
  }

  /**
   * Liefert <code>false</code>, da ein Xulu-Objekt von jeder Eigenschaft
   * immer nur eine Instanz enthaelt.
   */
  public boolean containsMultipleValues() {
    return false;
  }


}
