/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.gui;

import javax.swing.tree.TreeModel;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;
import java.util.Enumeration;

import schmitzm.data.property.Property;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.Properties;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ListPropertyReadAccess;

import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.swing.tree.EmptyNode;
import schmitzm.swing.tree.ContentNode;

import edu.bonn.xulu.appl.DataPool;
import edu.bonn.xulu.data.XuluObject;

/**
 * Diese Klasse stellt ein TreeModel fuer den {@link edu.bonn.xulu.appl.DataPool} dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DataPoolTreeModel extends DefaultTreeModel {
  private DataPool dataPool               = null;
  private boolean  editable               = false;
  private ObjectListener dataPoolListener = null;

  /**
   * Erzeugt ein neues Daten-Modell fuer einen Datenpool.
   * @param dp Instanz eines Xulu-Datenpools
   * @param editable bestimmt, ob die Baum-Eintraege editierbar sein duerfen
   * @param rootName Bezeichnung fuer den Wurzelknoten
   */
  public DataPoolTreeModel(DataPool dp, boolean editable, String rootName) {
    super( new EmptyNode(rootName) );
    this.dataPool = dp;
    this.editable = editable;
    reorganize();
    this.dataPoolListener = new ObjectListener() {
      public void performObjectEvent(ObjectEvent e) {
        DataPool.DataPoolChangeEvent dpEvent = (DataPool.DataPoolChangeEvent)e;
        switch (dpEvent.getType()) {
          // Objekt wurde hinzugefuegt
          case DataPool.DataPoolChangeEvent.OBJECT_ADDED:
            add( (XuluObject)dpEvent.getSource().getRoot() );
            break;
          // Objekt wurde entfernt
          case DataPool.DataPoolChangeEvent.OBJECT_REMOVED:
            remove( (XuluObject)dpEvent.getSource().getRoot() );
            break;
        }
        reload();
      }
    };
    dataPool.addObjectListener( dataPoolListener );
  }

  /**
   * Zerstoert das Tree-Model.
   */
  public void finalize() throws Throwable {
    dataPool.removeObjectListener( dataPoolListener );
    super.finalize();
  }

  /**
   * Baut den Baum komplett neu auf.
   */
  public void reorganize() {
    ((EmptyNode)getRoot()).removeAllChildren();
    // Fuer alle Xulu-Objekte einen Knoten einfuegen
    for ( Enumeration e = dataPool.asEnumeration(); e.hasMoreElements(); )
      add( (XuluObject)e.nextElement() );
    reload();
  }

  /**
   * Fuegt dem Baum auf der Root-Ebene ein Xulu-Objekt hinzu.
   * @param obj hinzuzufuegendes Xulu-Objekt
   */
  public void add(XuluObject obj) {
    ((EmptyNode)getRoot()).add( new XuluObjectNode(obj,editable) );
  }

  /**
   * Entfernt ein Xulu-Objekt aus der Root-Ebene des Baums.
   * @param obj zu entfernendes Xulu-Objekt
   */
  public void remove(XuluObject obj) {
    // Knoten des Objekts suchen und entfernen
    for (int i=0; i<((TreeNode)getRoot()).getChildCount();i++) {
      ContentNode child = (ContentNode) ((EmptyNode)getRoot()).getChildAt(i);
      if ( child.getUserObject() == obj )
        ((EmptyNode)getRoot()).remove( i-- );
    }
  }

  /**
   * Diese Klasse stellt einen Baum-Knoten fuer ein {@link Properties}-Objekt dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class PropertiesNode extends ContentNode {
    public PropertiesNode(Properties obj, boolean editable) {
      super(obj,editable);
      Property[] prop = obj.getProperties();
      for (int i=0; i<prop.length; i++)
        this.add( new PropertyNode(prop[i]) );
    }
  }

  /**
   * Diese Klasse stellt einen Baum-Knoten fuer ein {@link XuluObject}-Objekt dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class XuluObjectNode extends PropertiesNode {
    public XuluObjectNode(XuluObject obj, boolean editable) {
      super(obj,editable);
    }

    // bei Aenderung aendert sich der Name des Objekts
    public void setUserObject(Object newValue) {
      if ( newValue instanceof String && getUserObject() instanceof XuluObject )
        ((XuluObject)getUserObject()).setDescription( (String)newValue );
    }

    // Xulu-Objekte haben eine spezielle Beschreibung
    public String toString() {
      return ((XuluObject)getUserObject()).getDescription();
    }
  }


  /**
   * Diese Klasse stellt einen Baum-Knoten fuer ein {@link Property}-Objekt dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class PropertyNode extends ContentNode {
    public PropertyNode(Property prop) {
      super(prop,prop.getName(),false);
      if ( prop instanceof ScalarProperty )
        expandNode(this,((ScalarProperty)prop).getOneTimeReadAccess().getValue(),"");
      if ( prop instanceof ListProperty ) {
        ListPropertyReadAccess lpa = ((ListProperty)prop).getReadAccess(this);
        EmptyNode node = new EmptyNode("List["+prop.getPropertyType().getType().getSimpleName()+"] >> "+lpa.getCount());
        for (int i=0; i<lpa.getCount(); i++)
          expandNode(node,lpa.getValue(i),(i+1)+". ");
        lpa.release();
        add( node );
      }
      if ( prop instanceof MatrixProperty ) {
        MatrixProperty m = (MatrixProperty)prop;
        EmptyNode node = new EmptyNode(m.getSizeText()+"-Matrix["+prop.getPropertyType().getType().getSimpleName()+"]");
        add( node );
      }
    }

    private void expandNode(DefaultMutableTreeNode node, Object o, String namePrefix) {
      if (o instanceof Property)
        node.add(new PropertyNode( (Property) o));
      else if (o instanceof Properties)
        node.add(new PropertiesNode( (Properties) o, false));
      else if (o != null)
        node.add(new ContentNode(o,namePrefix+o.getClass().getSimpleName(),false));
      else
        node.add(new EmptyNode());
    }


  }

}
