/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.io;

import java.io.File;
import schmitzm.lang.AbstractNamedObject;
import edu.bonn.xulu.gui.DataPoolMenu;
import edu.bonn.xulu.appl.XuluRegistry;

/**
 * Diese Klasse stellt eine Oberklasse fuer die Informationen dar, die zum
 * Import oder Export von Xulu-Objekten notwendig sind.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class IODefinition /*extends AbstractNamedObject*/ {
  /** Speichert das Object aus dem importiert (in das exportiert) wird. */
  protected Object       object   = null;
  /** Speichert die Factory, die den Import (Export) vornimmt. */
  protected Factory      factory  = null;
  /** Speichert die Xulu-Registry, in der die Factories registriert sind. */
  protected XuluRegistry registry = null;

  /**
   * Erzeugt eine neue Import/Export-Definition.
   * @param object   Objekt aus dem importiert (bzw. in das exportiert) wird; z.B. eine Datei
   * @param factory  Factory, die den Import/Export vornimmt
   * @param registry Xulu-Registry in der die Factories registriert sind
   */
  public IODefinition(Object object, Factory factory, XuluRegistry registry) {
    if ( object == null )
      throw new IllegalArgumentException("null-object not allowed!");
    if ( factory == null )
      throw new IllegalArgumentException("null-factory not allowed!");
    this.object     = object;
    this.factory    = factory;
    this.registry   = registry;
  }

  /**
   * Liefert die Xulu-Registry, in der die Factories registriert sind.
   */
  public XuluRegistry getRegistry() {
    return this.registry;
  }

  /**
   * Liefert die Factory, die den Import (Export) vornimmt.
   */
  public Factory getFactory() {
    return factory;
  }

  /**
   * Liefert das Objekt aus dem importiert wird (in das exportiert
   * wird).
   */
  public Object getIOObject() {
    return object;
  }

  /**
   * Vergleicht die {@code IODefinition} mit einer anderen. Beide sind
   * gleich, gdw. beide die gleiche Factory und das gleiche I/O-Objekt
   * enthalten.
   * @param o andere {@code IODefinition}
   * @see Factory#equals(Factory)
   * @see Object#equals(Object)
   */
  public boolean equals(Object o) {
    return o != null &&
           o instanceof IODefinition &&
           ((IODefinition)o).getFactory().equals( getFactory() ) &&
           ((IODefinition)o).getIOObject().equals( getIOObject() );
  }

  ///////////////////////////////////////////////////////////////////
  //////////////////////   Import-Definition   //////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt eine Import-Definition dar. Sie besteht aus
   * einer {@link ImportFactory} und einem Quell-Objekt aus dem
   * die Factory ein Objekt generiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ImportDefinition extends IODefinition {
    /**
     * Erzeugt eine neue Import-Definition.
     * @param source   Objekt aus dem importiert wird; z.B. eine Datei
     * @param fac      Factory, die den Import vornimmt
     * @param registry Xulu-Registry in der die Factories registriert sind
     */
    public ImportDefinition(Object source, ImportFactory fac, XuluRegistry registry) {
      super(source,fac,registry);
    }

    /**
     * Liefert die Factory, die den Import vornimmt.
     */
    public ImportFactory getFactory() {
      return (ImportFactory)super.getFactory();
    }
  }

  ///////////////////////////////////////////////////////////////////
  //////////////////////   Export-Definition   //////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt eine Export-Definition dar. Sie besteht aus
   * einer {@link ExportFactory} und einem Ziel-Objekt in das
   * die Factory ein Objekt schreibt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ExportDefinition extends IODefinition {
    /**
     * Erzeugt eine neue Export-Definition.
     * @param dest     Objekt in das exportiert wird; z.B. eine Datei
     * @param fac      Factory, die den Export vornimmt
     * @param registry Xulu-Registry in der die Factories registriert sind
     */
    public ExportDefinition(Object dest, ExportFactory fac, XuluRegistry registry) {
      super(dest,fac,registry);
    }

    /**
     * Liefert die Factory, die den Export vornimmt.
     */
    public ExportFactory getFactory() {
      return (ExportFactory)super.getFactory();
    }
  }
}

