/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.io;

import java.util.Vector;

/**
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class TypeMapping {
  private Class                type       = null;
  private InstantiationFactory defaultFac = null;
  private Vector               importFac  = new Vector<ImportFactory>();
  private Vector               exportFac  = new Vector<ExportFactory>();

  /**
   * Erzeugt ein neues TypeMapping.
   * @param type Datentyp
   * @param fac  Factory, die eine Standard-Instanz des Datentyps erzeugt
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            Default-Factory den Datentyp nicht erzeugt oder die Factory
   *            <code>null</code> ist
   */
  public TypeMapping(Class type, InstantiationFactory fac) {
    this.type = type;
    setDefaultFactory(fac);
  }

  ///////////////////////////////////////////////////////////////////
  // Allgemeine Methoden
  ///////////////////////////////////////////////////////////////////
  /**
   * Ersetzt ein Object in einem Vector durch ein anderes.
   * Ist das <code>o1</code> nicht im Vector vorhanden wird nichts gemacht und
   * <code>false</code> zurueckgegeben.
   */
  private boolean replace(Object o1, Object o2, Vector vector) {
    int idx = vector.indexOf(o1);
    if ( idx < 0 )
      return false;
    vector.set(idx,o2);
    return true;
  }

  /**
   * Prueft, ob diese angegebene allgemeine Factory zulaessig ist.
   */
  private void checkFactory(Factory fac) {
    if ( fac == null )
      throw new TypeMappingException("null-Factories are not allowed!");
  }

  /**
   * Liefert den Datentyp des Mappings.
   */
  public Class getType() {
    return this.type;
  }

  ///////////////////////////////////////////////////////////////////
  // Default-Factory
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert die Factory, die den Datentyp standardmaessig instanziiert.
   */
  public InstantiationFactory getDefaultFactory() {
    return this.defaultFac;
  }

  /**
   * Prueft, ob diese angegebene allgemeine Factory als DefaultFactory zulaessig ist.
   */
  private void checkDefaultFactory(Factory fac) {
    checkFactory(fac);
    if ( !(fac instanceof InstantiationFactory) )
      throw new TypeMappingException("DefaultFactory must be an instance of schmitzm.data.io.InstantiationFactory!");
    if ( !getType().isAssignableFrom(((InstantiationFactory)fac).getInstanceType()) )
      throw new TypeMappingException(getType(),fac,"Factory does not instantiate this type!");

  }

  /**
   * Setzt die Factory, die den Datentyp standardmaessig instanziiert.
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            Default-Factory den Datentyp nicht erzeugt oder die Factory
   *            <code>null</code> ist
   */
  public void setDefaultFactory(InstantiationFactory fac) {
    checkDefaultFactory(fac);
    this.defaultFac = fac;
  }

  ///////////////////////////////////////////////////////////////////
  // Import-Factory
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert alle Factorys, die den Datentyp importieren.
   */
  public ImportFactory[] getImportFactories() {
    return (ImportFactory[])importFac.toArray(new ImportFactory[0]);
  }

  /**
   * Prueft, ob diese angegebene allgemeine Factory als ImportFactory zulaessig ist.
   */
  private void checkImportFactory(Factory fac) {
    checkFactory(fac);
    if ( !(fac instanceof ImportFactory) )
      throw new TypeMappingException("ImportFactory must be an instance of schmitzm.data.io.ImportFactory!");
    if ( !getType().isAssignableFrom(((ImportFactory)fac).getImportType()) )
      throw new TypeMappingException(getType(),fac,"Factory does not instantiate this type!");
  }

  /**
   * Fuegt eine Import-Factory fuer den Datentyp hinzu.
   * @return <code>false</code> falls diese Factory bereits fuer den Datentyp
   *         registriert ist
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            Import-Factory den Datentyp des TypeMappings nicht erzeugt
   *            oder die Factory <code>null</code> ist
   */
  public boolean addImportFactory(ImportFactory fac) {
    checkImportFactory(fac);
    if ( importFac.contains(fac) )
      return false;
    return importFac.add(fac);
  }

  /**
   * Entfernt eine Import-Factory aus dem TypeMapping.
   * @return <code>false</code> falls diese Factory nicht fuer den Datentyp
   *         registriert ist
   */
  public boolean removeImportFactory(ImportFactory fac) {
    return importFac.remove(fac);
  }

  /**
   * Ersetzt eine Import-Factory des TypeMappings durch eine andere.
   * @param facOld zu ersetzende Factory
   * @param facNew neue Factory
   * @return <code>false</code> falls <code>facOld</code> nicht fuer den Datentyp
   *         registriert war und deshalb keine Ersetzung moeglich ist
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            neue Import-Factory den Datentyp des TypeMappings nicht erzeugt
   *            oder die Factory <code>null</code> ist
   */
  public boolean replaceImportFactory(ImportFactory facOld, ImportFactory facNew) {
    checkImportFactory(facNew);
    return replace(facOld,facNew,importFac);
  }

  ///////////////////////////////////////////////////////////////////
  // Export-Factory
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert alle Factorys, die den Datentyp exportieren.
   */
  public ExportFactory[] getExportFactories() {
    return (ExportFactory[])exportFac.toArray(new ExportFactory[0]);
  }

  /**
   * Prueft, ob diese angegebene allgemeine Factory als ExportFactory zulaessig ist.
   */
  private void checkExportFactory(Factory fac) {
    checkFactory(fac);
    if ( !(fac instanceof ExportFactory) )
      throw new TypeMappingException("ExportFactory must be an instance of schmitzm.data.io.ExportFactory!");
    if ( !((ExportFactory)fac).isExportable(getType()) )
      throw new TypeMappingException(getType(),fac,"Factory can not export this type!");
  }

  /**
   * Fuegt eine Export-Factory fuer den Datentyp hinzu.
   * @return <code>false</code> falls diese Factory bereits fuer den Datentyp
   *         registriert ist
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            Export-Factory den Datentyp des TypeMappings nicht exportieren
   *            kann oder die Factory <code>null</code> ist
   */
  public boolean addExportFactory(ExportFactory fac) {
    checkExportFactory(fac);
    if ( exportFac.contains(fac) )
      return false;
    return exportFac.add(fac);
  }

  /**
   * Entfernt eine Export-Factory aus dem TypeMapping.
   * @return <code>false</code> falls diese Factory nicht fuer den Datentyp
   *         registriert ist
   */
  public boolean removeExportFactory(ImportFactory fac) {
    return exportFac.remove(fac);
  }

  /**
   * Ersetzt eine Export-Factory des TypeMappings durch eine andere.
   * @param facOld zu ersetzende Factory
   * @param facNew neue Factory
   * @return <code>false</code> falls <code>facOld</code> nicht fuer den Datentyp
   *         registriert war und deshalb keine Ersetzung moeglich ist
   * @exception schmitzm.data.io.TypeMappingException falls die
   *            neue Export-Factory den Datentyp des TypeMappings nicht exportieren
   *            kann oder die Factory <code>null</code> ist
   */
  public boolean replaceExportFactory(ExportFactory facOld, ExportFactory facNew) {
    checkExportFactory(facNew);
    return replace(facOld,facNew,exportFac);
  }
}
