/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import java.util.Map;

import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.data.XuluDataException;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.CaptionsChangeable;

import org.apache.log4j.Logger;

/**
 * Diese Klasse stellt eine abstrakte Basis-Implementierung fuer den
 * {@link ModelContentManager} dar. Das einizge, das fuer ein Modell
 * noch implementieren bleibt, ist die modellspezifische Befuellung
 * des {@link #resource}-Array (im Konstruktor).<br>
 * Am besten wird wird die modellspezifische Implementierung als innere
 * Klasse in die Modell-Klasse integriert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractModelContentManager implements ModelContentManager {
  private ResourceProvider RESOURCE = XuluConstants.XULUGUI_RES;

  /** {@link Logger} fuer Debug- oder Warn-Meldungen. */
  protected Logger LOGGER = LangUtil.createLogger(this);

  /** Speichert die Ressourcen, die ein Modell benoetigt. */
  protected ModelResource[] resource = null;

  /**
   * Erzeugt einen neuen ContentManager. Der {@link #resource}-Array wird auf
   * die benoetigte Ressourcen-Anzahl initialisiert.
   * @param count Anzahl an Ressourcen, die das Model benoetigt.
   */
  public AbstractModelContentManager(int count) {
    this.resource = new ModelResource[count];
  }

  /**
   * Erzeugt einen neuen ContentManager. Der {@link #resource}-Array wird auf
   * eine groesse von 0 initialisiert.
   */
  public AbstractModelContentManager() {
    this(0);
  }

  /**
   * Liefert eine Ressource, die ein Modell benoetigt.
   * @param idx Nummer der Ressource
   */
  public ModelResource getResource(int idx) {
    return resource[idx];
  }

  /**
   * Liefert das Datenobjekt einer Ressourcen.
   * @param idx Nummer der Ressource
   */
  public Object getResourceData(int idx) {
    return getResource(idx).getData();
  }

  /**
   * Liefert die Anzahl der Ressourcen, die das Modell benoetigt.
   */
  public int getResourceCount() {
    return resource.length;
  }

  /**
   * Ueberprueft alle Ressourcen auf Korrektheit.
   * @see ModelResource#check()
   * @return <code>false</code> falls eine Ressource nicht korret gefuellt ist
   */
  public boolean check() {
    try {
      checkAndError();
    } catch (Exception err) {
      return false;
    }
    return true;
  }

  /**
   * Ueberprueft alle Ressourcen auf Korrektheit. Unterklassen muessen diese
   * Methode ueberschreiben, falls eine Uberpruefung der Konsistenz
   * <b>zwischen</b> den einzelnen Ressourcen notwendig ist.
   * @exception XuluDataException falls eine Ressource nicht korret gefuellt ist
   * @see ModelResource#checkAndError()
   */
  public void checkAndError() throws XuluDataException {
    for (int i=0; resource != null & i<resource.length; i++)
      resource[i].checkAndError();
  }

  /**
   * Belegt die sprachspezifischen Elemente des Modells (Ressourcen-Beschreibungen)
   * neu mit Elementen aus der Map. Diese Standard-Implementierung macht nichts.
   * @param captionMap Label-Beschreibungen
   */
  public void resetCaptions(Map<String, Object> captionMap) { 
  }

  /**
   * Belegt die Beschreibung einer Ressource neu. Zunaechst wird die {@code captionMap}
   * herangezogen. Wenn dort zu dem {@code captionKey} kein Wert hinterlegt ist,
   * wird auf dan angegebenen ResourceProvider zurueckgegriffen.
   * @param resourceNo    Index einer Modell-Ressource
   * @param captionMap    Map mit neuen Bezeichnungen
   * @param rp            ResourceProvider
   * @param captionKey    Key unter dem in der Map und im ResourceProvider gesucht wird 
   * @param captionParams Parameter, die durch den ResourceProvider in die Caption "eingebaut" werden
   * @see ResourceProvider#getString(String, Object...)
   */
  public void resetResourceDescription(int resourceNo, Map<String,Object> captionMap, ResourceProvider rp, String captionKey, Object... captionParams) {
    String newCaption = AbstractXuluModel.determineCaption(captionMap, rp, captionKey, captionParams);
    if ( newCaption != null )
      resource[resourceNo].setDesc( newCaption );
  }
}
