/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import schmitzm.lang.ResourceProvider;

import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.appl.XuluConstants;

/**
 * Diese Klasse bildet eine abstrakte Basis-Implementierung fuer eine Modell-Ressource.
 * Die einzigen Methoden, die eine Unterklasse noch implementieren muss, sind
 * {@link #checkAndError(Object)} und {@link #checkTypeAndError(Object)},
 * ueber die geprueft wird, ob ein Objekt fuer die Ressource zulaessig ist.<br>
 * Details zur Funktionalitaet und Bestimmung dieser Klasse sind in
 * {@link ModelResource} zu finden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractModelResource implements ModelResource {
  private ResourceProvider RESOURCE = XuluConstants.XULUGUI_RES;

    /** Speichert den Datentyp der Ressource. */
    protected Class type   = null;
    /** Speichert das Daten-Objekt der Ressource. */
    protected Object data = null;
    /** Speichert die Beschreibung der Ressource. */
    protected String desc   = "";
    /** Speichert, ob die Ressource unbelegt bleiben darf */
    protected boolean nullable = false;
    /** Speichert die Kategorie der Ressource (Default: {@link #CATEGORY_OTHER}) */
    protected int category = CATEGORY_OTHER;

    /**
     * Erzeugt eine leere Ressource.
     * @param category Ordnet die Ressource in eine Kategorie ein (siehe
     *                 {@link #setCategory(int)})
     * @param desc Beschreibung der Ressource
     * @param type Datentyp der Ressource
     * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
     * @exception XuluDataException falls der Datentyp <code>null</code> ist.
     */
    public AbstractModelResource(int category, String desc, Class type, boolean nullable) {
      if ( type == null )
        throw new XuluDataException( RESOURCE.getString("ModelResource.Type.Null") );
      checkCategory(category);
      this.category = category;
      this.type     = type;
      this.desc     = desc;
      this.nullable = nullable;
    }

    /**
     * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
     * wird implizit aus dem uebergebenen Objekt ermittelt.
     * @param category Ordnet die Ressource in eine Kategorie ein (siehe
     *                 {@link #setCategory(int)})
     * @param desc Beschreibung der Ressource
     * @param data Daten-Objekt fuer die Ressource
     * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
     * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
     *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
     */
    public AbstractModelResource(int category, String desc, Object data, boolean nullable) {
      if ( data == null )
        throw new XuluDataException( RESOURCE.getString("ModelResource.Data.Null") );
      checkCategory(category);
      this.category = category;
      this.type = data.getClass();
      this.desc = desc;
      this.nullable = nullable;
      setData(data);
    }

    /**
     * Erzeugt eine leere Ressource.
     * @param desc Beschreibung der Ressource
     * @param type Datentyp der Ressource
     * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
     * @exception XuluDataException falls der Datentyp <code>null</code> ist.
     * @exception IllegalArgumentException falls eine unbekannte Kategorie
     *            angegeben wird
     */
    public AbstractModelResource(String desc, Class type, boolean nullable) {
      this(CATEGORY_OTHER,desc,type,nullable);
    }

    /**
     * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
     * wird implizit aus dem uebergebenen Objekt ermittelt.
     * @param desc Beschreibung der Ressource
     * @param data Daten-Objekt fuer die Ressource
     * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
     * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
     *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
     * @exception IllegalArgumentException falls eine unbekannte Kategorie
     *            angegeben wird
     */
    public AbstractModelResource(String desc, Object data, boolean nullable) {
      this(CATEGORY_OTHER,desc,data,nullable);
    }

    /**
     * Prueft, ob die angegebene Kategorie korrekt ist. Unterklassen, die neue
     * Kategorien definieren, muessen diese Methode ueberschreiben.
     * @exception IllegalArgumentException falls eine unbekannte Kategorie
     *            angegeben wird
     */
    protected void checkCategory(int category) {
      if ( category < 0 ||
           category > (CATEGORY_INPUT | CATEGORY_OUTPUT | CATEGORY_TEMP) )
        throw new XuluDataException( RESOURCE.getString("ModelResource.Cat.Unknown", category) );
    }

    /**
     * Gibt an, ob die Ressource opional ist (also unbelegt bleiben darf).
     */
    public boolean isNullable() {
      return nullable;
    }

    /**
     * Bestimmt, ob die Ressource opional ist (also unbelegt bleiben darf).
     */
    public void setNullable(boolean nullable) {
      this.nullable = nullable;
    }

    /**
     * Liefert die Kategorie der Ressource. Dies kann die Konstante
     * {@link #CATEGORY_OTHER} sein, oder eine logische
     * Oder-Verknuepfung der Konstanten {@link #CATEGORY_INPUT},
     * {@link #CATEGORY_OUTPUT} und {@link #CATEGORY_TEMP}.
     */
    public int getCategory() {
      return this.category;
    }

    /**
     * Setzt die Kategorie der Ressource. Dies kann die Konstante
     * {@link #CATEGORY_OTHER} sein, oder eine logische
     * Oder-Verknuepfung der Konstanten {@link #CATEGORY_INPUT},
     * {@link #CATEGORY_OUTPUT} und {@link #CATEGORY_TEMP}.
     * @exception IllegalArgumentException falls eine unbekannte Kategorie
     *            angegeben wird
     */
    public void setCategory(int category) {
      checkCategory(category);
      this.category = category;
    }

    /**
     * Prueft, ob die Ressource einer bestimmten Kategorie
     * zugeordnet ist. Liefert <code>true</code> wenn die Ressource <b>einer</b>
     * der angegebenen Kategorien zugeordnet ist (ODER).
     * @param category Oder-Verknuepfung von Kategorien
     * @return <code>this.getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
     *         (this.getCategory() & category) > 0</code>
     */
    public boolean isCategory(int category) {
      return getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
            (getCategory() & category) > 0;
    }

    /**
     * Prueft, ob die Ressource einer bestimmten Kategorie-Kombination
     * zugeordnet ist. Liefert <code>true</code> wenn die Ressource <b>allen</b>
     * angegebenen Kategorien zugeordnet ist (UND).
     * @return <code>this.getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
     *         (this.getCategory() & category) == category</code>
     */
    public boolean isCategoryAll(int category) {
      return getCategory() == CATEGORY_OTHER && category == CATEGORY_OTHER ||
            (getCategory() & category) == category;
    }

    /**
     * Prueft ein beliebiges Objekt nur auf Typ-Vertraeglichkeit mit der
     * Ressource.
     */
    public boolean checkType(Object data) {
      try {
        checkTypeAndError(data);
      } catch (Exception err) {
        return false;
      }
      return true;
    }

    /**
     * Prueft ein beliebiges Objekt nur auf Typ-Vertraeglichkeit mit der
     * Ressource.
     * @exception XuluDataException falls der Objekt-Typ nicht mit der Ressourcen
     *            vertraeglich ist
     */
    public abstract void checkTypeAndError(Object data) throws XuluDataException;

    /**
     * Prueft ein beliebiges Objekt auf Typ- und Wert-Konsistenz mit der
     * Ressource.
     * @see #checkAndError(Object)
     */
    public boolean check(Object data) {
      try {
        checkAndError(data);
      } catch (Exception err) {
        return false;
      }
      return true;
    }

    /**
     * Prueft ein beliebiges Objekt auf Typ- und Wert-Konsistenz mit der
     * Ressource. Kann manuell aufgerufen werden und wird automatisch von
     * {@link #setData(Object)} aufgerufen.
     * @exception XuluDataException falls der Typ oder der Wert des zugeordneten
     *            Objekts nicht mit der Ressource vertraeglich ist
     */
    public abstract void checkAndError(Object data) throws XuluDataException;

    /**
     * Prueft das der Ressource zugeordnete Objekt auf Typ- und Wert-Konsistenz
     * mit der Ressource.
     * @see #checkAndError()
     */
    public boolean check() {
      return check( getData() );
    }

    /**
     * Prueft das der Ressource zugeordnete Objekt auf Typ- und Wert-Konsistenz
     * mit der Ressource.
     * @exception XuluDataException falls der Typ oder der Wert des zugeordneten
     *            Objekts nicht mit der Ressource vertraeglich ist
     * @see #checkAndError(Object)
     */
    public void checkAndError() throws XuluDataException {
      checkAndError( getData() );
    }

    /**
     * Liefert den Datentyp der Ressource.
     */
    public Class getType() {
      return this.type;
    }

    /**
     * Liefert das Daten-Objekt der Ressource.
     */
    public Object getData() {
      return this.data;
    }

    /**
     * Liefert eine Beschreibung, worum es sich bei der Ressource handelt.
     */
    public String getDesc() {
      return this.desc;
    }

    /**
     * Setzt eine Beschreibung, worum es sich bei der Ressource handelt.
     */
    public void setDesc(String desc) {
      this.desc = desc;
    }

    /**
     * Ordnet der Ressource ein bestimmtes Datenobjekt zu.
     * @param data Daten-Objekt
     * @param check bestimmt, ob das Datenobjekt sofort auf Typ-Vertraeglichkeit
     *   ueberprueft werden soll
     * @see #checkAndError(Object)
     */
    public void setData(Object data, boolean check) {
      if ( check )
        checkAndError( data );
      this.data = data;
    }

    /**
     * Ordnet der Ressource ein bestimmtes Datenobjekt zu.
     * @param data Daten-Objekt
     * @throws XuluDataException falls das nicht mit dem Ressourcen-Typ
     *         vertraeglich ist
     * @see #checkAndError(Object)
     */
    public void setData(Object data) throws XuluDataException {
      setData(data,true);
    }

    /**
     * Erzeugt und wirft eine {@link XuluDataException} mit der Beschreibung
     * <center>
     * "Resource '<i>resource_name</i>': <i>expected</i> expected"
     * </center>
     * Der Ressourcenname wird automatisch ueber {@link #getDesc() this.getDesc()}
     * ermittelt. Der erwartete Wert/Typ wird als Parameter uebergeben.
     * @param expected Bezeichnung des erwarteten Typ/Wert
     */
    protected void throwDataException(String expected) {
      throw new XuluDataException( RESOURCE.getString("ModelResource.Data.Error",desc,expected));
    }

}
