/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import schmitzm.lang.ResourceProvider;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.model.AbstractXuluModel;
import edu.bonn.xulu.model.StepModel;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.ModelStepStartedEvent;
import edu.bonn.xulu.model.event.ModelStepFinishedEvent;

/**
 * Diese Klasse implementiert ein Modell, welches schrittweise ausgefuehrt
 * wird. Das Modell arbeitet eine feste Anzahl von Arbeitsschritten ab, bevor
 * es automatisch auslaeuft. Die Anzahl wird in der Variablen {@link #stepCount}
 * hinterlegt und <b>muss</b> von einer Implementierung befuellt werden!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractStepModel extends AbstractXuluModel implements StepModel {
  /** ResourceProvider fuer allgemeine Uebersetzungen. */
  private ResourceProvider RESOURCE = XuluConstants.XULUGUI_RES;

  /** Speichert die Anzahl der zu modellierenden Schritte.
   *  muss vor dem Modellstart von der Implementierung gefuellt werden! */
  protected int stepCount = 0;

  /**
   * Erzeugt ein neues Model.
   * @param contentManager verwaltet die Ressourcen fuer das Model
   */
  public AbstractStepModel(ModelContentManager contentManager) {
    super(contentManager);
    events.add( ModelStepStartedEvent.class );
    events.add( ModelStepFinishedEvent.class );
  }

  /**
   * Implementiert den Rahmen des Modellablaufs. Ruft sukzessive
   * {@link #performModelStep(int)} in einer Schleife auf, bis die
   * vorgegebene Schrittzahl erreicht ist. Wird das Modell zuvor unterbrochen
   * ({@link #stopModelling()} wird auch die Schleife unterbrochen. Loest
   * {@link #fireModelStepStarted(int)} und {@link #fireModelStepFinished(int)}
   * aus.
   */
  public void performModelStart() {
    int stepCount = getStepCount();
    for (int step=1; !isStopped() && step<=stepCount; step++) {
      fireModelStepStarted(step);
      performModelStep(step);
      fireModelStepFinished(step);
    }
  }

  /**
   * Implementiert einen Modellschritt.
   * @param stepNo Nummer des Modellschritts (beginnend bei 1)
   */
  public abstract void performModelStep(int stepNo);

  /**
   * Liefert die Anzahl der zu modellierenden Schritte.
   * @see #stepCount
   */
  public int getStepCount() {
    return stepCount;
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell einen
   * neuen Modellierungsschritt beginnt.
   */
  protected void fireModelStepStarted(int step) {
    getStatusOut().println( RESOURCE.getString("Model.Step.Started",step) );
    for (int i=0; i<listeners.size(); i++)
      ((ModelListener)listeners.get(i)).performModelEvent(new ModelStepStartedEvent(this,step));
  }

  /**
   * Informiert alle {@link ModelListener} darueber, dass das Modell einen
   * neuen Modellierungsschritt beendet hat.
   */
  protected void fireModelStepFinished(int step) {
    getStatusOut().println( RESOURCE.getString("Model.Step.Finished",step) );
    for (int i=0; i<listeners.size(); i++)
      ((ModelListener)listeners.get(i)).performModelEvent(new ModelStepFinishedEvent(this,step));
  }
}
