/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import schmitzm.data.property.Properties;
import edu.bonn.xulu.model.AbstractModelResource;
import edu.bonn.xulu.data.XuluDataException;

// nur fuer Doku
import edu.bonn.xulu.model.ModelResource;

/**
 * Diese Implementierung von <code>ModelResource</code> laesst nur
 * {@linkplain Properties Properties-Objekte} als Datenobjekt zu.<br>
 * Details zur Funktionalitaet und Bestimmung dieser Klasse sind in
 * {@link ModelResource} zu finden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PropertiesResource extends AbstractModelResource {
  /**
   * Erzeugt eine leere Ressource.
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param desc   Beschreibung der Ressource
   * @param type   Datentyp der Ressource; muss eine (Sub)Klasse von
   *               {@link Properties} sein
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls der Datentyp <code>null</code> oder
   *            keine (Sub)Klasse von {@link Properties} ist.
   */
  public PropertiesResource(int category, String desc, Class type, boolean nullable) {
    super(category, desc, type, nullable);
    if (!Properties.class.isAssignableFrom(type))
      throw new XuluDataException("Resource '"+desc+"': Only (sub)classes of ".concat(Properties.class.
          getName()).concat(" are allowed for ModelResource!"));
  }

  /**
   * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
   * wird implizit aus dem uebergebenen Objekt ermittelt.
   * @param category Ordnet die Ressource in eine Kategorie ein (siehe
   *                 {@link #setCategory(int)})
   * @param data Daten-Objekt fuer die Ressource
   * @param desc Beschreibung der Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
   *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
   */
  public PropertiesResource(int category, String desc, Properties data, boolean nullable) {
    super(category, desc, data, nullable);
  }

  /**
   * Erzeugt eine leere Ressource.
   * @param desc   Beschreibung der Ressource
   * @param type   Datentyp der Ressource; muss eine (Sub)Klasse von
   *               {@link Properties} sein
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls der Datentyp <code>null</code> oder
   *            keine (Sub)Klasse von {@link Properties} ist.
   */
  public PropertiesResource(String desc, Class type, boolean nullable) {
    this(CATEGORY_OTHER,desc,type,nullable);
  }

  /**
   * Erzeugt Ressource und ordnet ihr ein konretes Datenobjekt zu. Der Datentyp
   * wird implizit aus dem uebergebenen Objekt ermittelt.
   * @param data Daten-Objekt fuer die Ressource
   * @param desc Beschreibung der Ressource
   * @param nullable gibt an, ob die Ressource unbelegt bleiben darf
   * @exception XuluDataException falls das Datenobjekt <code>null</code> ist,
   *           denn aus <code>null</code> kann kein Datentyp ermittelt werden
   */
  public PropertiesResource(String desc, Properties data, boolean nullable) {
    this(CATEGORY_OTHER,desc,data,nullable);
  }

  /**
   * Liefert das Daten-Objekt der Ressource.
   */
  public Properties getData() {
    return (Properties)super.getData();
  }

  /**
   * Prueft ein beliebiges Objekt auf Typ-Konsistenz mit der Ressource und
   * ggf. auf <code>not null</code>.
   * Kann manuell aufgerufen werden und wird automatisch von
   * {@link #checkAndError(Object)} aufgerufen.
   * @exception XuluDataException falls das Objekt unerlaubterweise <code>null</code>
   *         ist, oder dessen Klasse nicht mit dem Ressourcen-Typ vertraeglich ist
   */
  public void checkTypeAndError(Object data) throws XuluDataException {
      // Datenobjekt auf null checken
      if ( data == null ) {
        if ( !isNullable() )
          throw new XuluDataException("Resource '"+desc+": Resource-Data '".concat(this.getDesc()).concat("' is not allowed to be null!"));
        else
          return;
      }
    if ( !getType().isAssignableFrom( data.getClass() ) )
      throw new XuluDataException("Resource '"+desc+"': Resource-Data must be an instance of ".concat(getType().getName()).concat("!"));
  }

  /**
   * Prueft ein beliebiges Objekt auf Typ-Konsistenz mit der Ressource und auf
   * die Erfuelltheit der Constraints.<br>
   * Kann manuell aufgerufen werden und wird automatisch von
   * {@link #setData(Object)} aufgerufen.
   * @exception XuluDataException falls das Objekt unerlaubterweise <code>null</code>
   *         ist, oder dessen Klasse nicht mit dem Ressourcen-Typ vertraeglich ist
   */
  public void checkAndError(Object data) throws XuluDataException {
    checkTypeAndError(data);
  }
}
