/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.XuluModelThread;
import edu.bonn.xulu.model.StepModel;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.StepModelAdapter;

/**
 * Diese Klasse implementiert einen Thread fuer ein Modell, welches schrittweise
 * ausgefuehrt wird. Diese Modell muss das Interface {@link StepModel}
 * implementieren, z.B. {@link AbstractStepModel}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class StepModelThread extends XuluModelThread {
  /** Modus "Thread fuehrt aktuellen Schritt zu Ende und wechselt dann in
   *  'Pause'". */
  public static final int MODE_STEP = 5;

  private ModelListener modelListener = null;

  /**
   * Erzeugt einen neuen Modell-Prozess.
   * @param model zu startendes Modell
   */
  public StepModelThread(StepModel model) {
    super(model);
    modelListener = new StepModelAdapter() {
      public void modelStepFinished(XuluModel model, int stepNo) {
        if ( mode == MODE_STEP && model instanceof StepModel && stepNo < ((StepModel)model).getStepCount() )
          pause();
        checkBreakingCommands();
      }
    };
    model.addModelListener( modelListener );
  }

  /**
   * Fuehrt einen einzelnen Arbeitsschritt aus. Danach wird der Thread
   * wieder schlafen gelegt.
   */
  public void step() {
    boolean wasSleeping = isSleeping();

    this.mode = MODE_STEP;
    // beim allerersten Mal muss der Thread gestartet werden
    if ( !isAlive() ) {
      super.start();
      this.mode = MODE_STEP;
    } else
      if ( wasSleeping )
        wakeUp();
  }

  /**
   * Entfernt die gesetzten Listener wieder vom Modell, damit der Thread nicht
   * weiterhin ueber Model-Aktivitaeten informiert wird.
   */
  public void performDispose() {
    // WICHTIG!! ModelListener wieder entfernen!! Ansonsten wird der Thread
    // immer wieder informiert, was zu einem Abbruch ALLER anderen Threads
    // fuehrt, da checkBreakingCommands() fuer diesen Thread eine
    // ThreadDeath-Exception wirft!
    model.removeModelListener( modelListener );
    super.performDispose();
  }
}
