/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.model;

import java.io.PrintStream;
import java.awt.Component;
import schmitzm.lang.NamedObject;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.CaptionsChangeable;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.ModelEvent;

/**
 * Dieses Interface definiert, wie eine Modell-Klasse aufgebaut sein muss,
 * damit das Modell in die Xulu-Modelling-Platform integiert werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface XuluModel extends NamedObject, CaptionsChangeable {
  /**
   * Liefert den ContentManager des Modells.
   */
  public ModelContentManager getContentManager();

  /**
   * Diese Methode setzt alle <b>internen</b> Modell-Ressourcen auf einen
   * Ausgangszustand zurueck, so dass das Modell komplett neu gestartet werden
   * kann.<br>
   * <b>Bemerke:</b> Auf die Ruecksetzung der externen Ressourcen, die von der
   * Modellierungsplatform bereitgestellt werden hat das Modell i.d.R. keinen
   * Einfluss. Hierfuer ist der Anwender verantwortlich.
   */
  public void initModel();

  /**
   * Prueft, ob das Modell initialisiert ist - sich also im Ausgangszustand
   * befindet - und bereit ist gestartet zu werden.<br>
   * Nach dem Modell-Start sollte diese Methode <code>false</code> liefern, da
   * sich die internen Ressourcen dann nicht mehr im Ausgangszustand befinden!
   */
  public boolean isInitialised();

  /**
   * Diese Methode loescht alle <b>internen</b> Modell-Ressourcen aus dem Speicher
   * und gibt die externen wieder frei.
   */
  public void disposeModel();

  /**
   * Prueft, ob das Modell (dessen interen Ressourcen) "zerstoert" sind.
   */
  public boolean isDisposed();

  /**
   * Diese Methode startet den Modell-Ablauf.
   * @return {@code false} wenn das Modell auf irgendeine Art (kontrolliert oder
   *         Fehler) abgebrochen wurde; {@code true} nur wenn das Modell komplett
   *         durchgelaufen ist
   */
  public boolean startModelling();


  /**
   * Prueft, ob die Modellierung gerade am Laufen ist.
   */
  public boolean isRunning();

  /**
   * Diese Methode unterbricht den Modellablauf komplett. Ein Fortsetzen ist
   * nicht moeglich.
   * Alle Modell-Ressourcen sollten beim Aufrufen dieser Methode wieder
   * freigegeben werden.
   */
  public void stopModelling();

  /**
   * Prueft, ob die Modellierung beendet ist.
   */
  public boolean isStopped();

  /**
   * Diese Methode wird aufgrufen, wenn ein unerwarteter Fehler bei der
   * Modellierung auftritt, welcher das Beenden der Modellierung zu Folge
   * hat. In dieser Methode muss das Modell dafuer sorgen, dass alle
   * Ressourcen - die internen und insbesondere auch die externen -
   * wieder freigegenen und ggf. aus dem Speicher entfernt werden.
   */
  public void performError();

  /**
   * Fuegt dem Model einen Listener hinzu.
   * @param listener neuer ModelListener
   */
  public boolean addModelListener(ModelListener listener);

  /**
   * Entfernt einen Listener von dem Modell.
   * @param listener zu entfernender ModelListener
   */
  public boolean removeModelListener(ModelListener listener);

  /**
   * Liefert den Stream fuer die Status-Ausgaben des Modells.
   */
  public PrintStream getStatusOut();

  /**
   * Setzt den Stream fuer die Status-Ausgaben des Modells.
   */
  public void setStatusOut(PrintStream out);


  /**
   * Liefert (optional) eine modellspezifische GUI-Komponente.
   * @return <code>null</code> wenn das Modell keine zusaetzliche GUI
   *         zur Verfuegung stellt
   */
  public Component getModelSpecificGUI();

  /**
   * Liefert alle Ereignis-Typen, die waehrend  des Modellablaufs initiiert werden.
   * @return leerer Array, falls keine Ereignisse initiiert werden
   */
  public Class<ModelEvent>[] getEvents();
  
  /**
   * Liefert den {@link ResourceProvider}, der die Uebersetzungen
   * fuer Modell-Ausgaben und modellspezifische GUI verwaltet.
   * @return {@code null}, wenn das Modell nicht multi-lingual ist
   */
  public ResourceProvider getResourceProvider();

}

