/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.InputOption;
import schmitzm.swing.SelectionInputOption;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.Handler;
import edu.bonn.xulu.appl.HandlerFactory;
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.io.Factory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.gui.DataPoolInputOption;
import edu.bonn.xulu.io.ExportFactory;

/**
 * Diese Klasse stellt eine abstrakte Oberklasse fuer alle Factories dar,
 * mit denen Instanzen des {@link DataExportHandler} erstellt und modifiziert
 * werden. Unterklassen muessen nur noch die Methoden implementieren, ueber
 * die benoetigten Informationen interaktiv (ueber Dialoge) vom Anwender abgefragt
 * werden.
 * @author <a href="mailto:schmitzm@bonn.edu">Martin Schmitz</a> (University of  Bonn/Germany)
 * @version 1.0
 */
public abstract class DataExportHandlerFactory extends AbstractFactory implements HandlerFactory {
  /**
   * Erzeugt eine neue Handler-Factory.
   */
  public DataExportHandlerFactory() {
    this("");
  }

  /**
   * Erzeugt eine neue Handler-Factory.
   * @param name Beschreibung der Factory
   */
  public DataExportHandlerFactory(String name) {
    super(name);
  }

  /**
   * Erzeugt einen neuen {@link DataExportHandler}.
   * @param appl Instanz der Xulu-Applikation
   */
  public DataExportHandler createHandler(XuluModellingPlatform appl) {
    Object[] input = showDialog(appl,null,null,null);
    if ( input == null || input.length < 3 )
      return null;
    return new DataExportHandler(input[0],(ExportFactory)input[1],input[2],this,appl);
  }

  /**
   * Veraendert einen bestehenden {@link FileExportHandler}.
   * @param appl Instanz der Xulu-Applikation
   * @param handler Handler, der veraendert werden soll
   * @exception UnsupportedOperationException falls der uebergebene Handler
   *            kein {@link FileExportHandler} ist
   */
  public void modifyHandler(XuluModellingPlatform appl, Handler handler) {
    if ( !(handler instanceof DataExportHandler) )
      throw new UnsupportedOperationException("FileExportHandlerFactory can only modify instances of FileExportHandler.");
    DataExportHandler exportHandler = (DataExportHandler)handler;
    Object[] input = showDialog(appl,exportHandler.getObject(), exportHandler.getExportFactory(), exportHandler.getExportDestination());
    if ( input != null && input.length >= 3 ) {
      exportHandler.setObject(input[0]);
      exportHandler.setExportFactory( (ExportFactory)input[1] );
      exportHandler.setExportDestination( input[2] );
    }
  }

  /**
   * Ermittelt die fuer den {@link DataExportHandler} benoetigten Informationen
   * ueber einen Anwender-Dialog.
   * @param appl Instanz der XuluModellingPlatform
   * @param object vorgeblendetes Objekt, welches exportiert wird (kann {@code null} sein)
   * @param expFac vorgeblendete Export-Factory (kann {@code null} sein)
   * @param destObj vorgeblendetes Objekt, in welches der Export vorgenommen wird (kann {@code null} sein)
   * @return Array in dem in Element 0 das Objekt, in Element 1 die Export-Factory und in
   *         Element 2 das Ziel-Objekt enthalten ist
   */
  protected abstract Object[] showDialog(XuluModellingPlatform appl, Object object, ExportFactory expFac, Object destObj);
}
