/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.io.File;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.StringTokenizer;
import java.awt.Color;
import java.util.Hashtable;
import java.util.Enumeration;

import org.geotools.styling.ColorMap;
import org.geotools.styling.ColorMapImpl;

import schmitzm.io.IOUtil;
import schmitzm.lang.AbstractNamedObject;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;
import schmitzm.swing.SwingUtil;

import schmitzm.geotools.gui.LayeredMapFrame;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.geotools.styling.ColorMapManager;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.appl.VisualisationManager;
// nur fuer Doku
import edu.bonn.xulu.appl.VisualisationTool;
import edu.bonn.xulu.plugin.vis.GTVisualisationTool;

/**
 * Dieses Plugin liest Farbpaletten aus der Datei <code>gtcolormaps.xif</code>
 * und fuegt sie automatisch den {@link GTVisualisationTool}-Instanzen hinzu,
 * sobald diese in Xulu geoeffnet werden. Die Datei muss im Verzeichnis
 * "GTVisualisationColorMapPlugin" unterhalb des
 * {@linkplain XuluModellingPlatform#getPluginDirectory() Xulu-Plugin-Verzeichnisses}
 * abgelegt sein.<br><br>
 * Die Datei kann mehrere Farbpaletten enthalten. Jede Farbpalette wird
 * von einem <code>[..]</code>-Tag eingeleitet, wobei die Bezeichnung zwischen den
 * Klammern als Name fuer die Farbpalette verwendet wird.
 * Nach dem <code>[..]</code>-Tag folgen die Farbpaletten-Eintraege der Form
 * <center><code><i>rasterwert</i> <i>color</i> [<i>label</i>]</code></center>
 * Die Farbe kann auf mehrere Arten spezifiziert werden:
 * <ol>
 *   <li>Als RGB-Werte der Form <code>"RGB(<i>red</i>,<i>green</i>,<i>blue</i>)"</code>.<br>
 *       Die RGB-Werte muessen durch Komma und OHNE Leerzeichen voneinander
 *       getrennt sein.</li>
 *   <li>Als dezimaler (1234) oder hexadezimaler Integer-Wert (0x123456),
 *       aus dem die 3 RGB-Komponenten extrahiert werden.</li>
 *   <li>Als String, der die Farbe identifiziert. Hierfuer sind alle
 *       Farb-Bezeichnungen zulaessig, die in der Klasse {@link Color java.awt.Color}
 *       als (statische) Felder deklariert sind.<br>
 *       z.B. Black, Blue, Cyan, Gray, Green, Magenta, Orange, Pink,  Red, White,
 *            Yellow</li>
 * </ol>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GTVisualisationColorMapPlugin extends AbstractNamedObject implements XuluPlugin, ObjectListener {
  /** Die Datei, aus der die {@linkplain ColorMap Farbpaletten} eingelesen werden
   * (<code>gtcolormaps.xif</code>). */
  public static final String INFILE = "GTVisualisationColorMapPlugin"+File.separatorChar+"gtcolormaps.xif";

  /** Flag, ob das Plugin gestartet ist, oder nicht */
  protected boolean started = false;

  /** Speichert die aus der Datei eingelesenen Farbpaletten */
  private Hashtable<String,ColorMap> colorMaps = new Hashtable<String,ColorMap>();
  private String                     defaultColorMap = null;

  /**
   * Erzeugt eine neue Instanz des Plugins.
   */
  public GTVisualisationColorMapPlugin() {
  }

  /**
   * Startet die Ausfuehrung des Plugins. Die {@linkplain ColorMap Farbpaletten}
   * werden aus der Datei eingelesen
   * @param appl Instanz der Xulu-Applikation
   */
  public void execute(XuluModellingPlatform appl) {
    started = true;
    try {
      readColorMapsFromFile(new File(appl.getPluginDirectory(),INFILE));
    } catch (Exception err) {
      XuluGUIMessages.showError(appl.getMainFrame(), err, "Plugin execution of "+getClass().getSimpleName()+" aborted!");
      stop();
    }
    if ( !appl.getVisualisationManager().containsObjectListener(this) )
      appl.getVisualisationManager().addObjectListener( this );
  }

  /**
   * Wird aufgerufen, wenn sich der Inhalt des {@link VisualisationManager Xulu-Visualisierungsmanagers}
   * aendert. Wurde ein neues {@link GTVisualisationTool} hinzugefuegt (oder ein
   * anderers von {@link LayeredMapFrame} abgeleitetes Tool), werden dessen
   * Farbpaletten alle eingelesenen Paletten hinzugefuegt.
   * @param e Ereignis
   */
  public void performObjectEvent(ObjectEvent e) {
    if ( !isStarted() || !(e instanceof VisualisationManager.ChangeEvent) )
      return;
    VisualisationManager.ChangeEvent event = (VisualisationManager.ChangeEvent)e;
    // ColorMaps nur hinzufuegen, wenn VisualisationTool neu hinzugefuegt
    // wurde und wenn es sich um ein GTVisualisationTool handelt
    if ( event.getOldValue()==null && event.getNewValue()!=null && event.getNewValue() instanceof GTVisualisationTool ) {
      ColorMapManager colorMapManager = ((GTVisualisationTool)event.getNewValue()).getLayeredMapFrame().getLayeredMapPane().getColorMapManager();
      Enumeration<String> colorMapNames = colorMaps.keys();
      for (; colorMapNames.hasMoreElements();) {
        String name = colorMapNames.nextElement();
        boolean isDefault = name.equals(defaultColorMap);
        colorMapManager.put(name,colorMaps.get(name),isDefault);
      }
    }
  }

  /**
   * Prueft, ob das Plugin gestartet ist.
   */
  public boolean isStarted() {
    return started;
  }

  /**
   * Prueft, ob die GUI zur Plugin-Konfiguration angezeigt wird.
   * @return immer <code>false</code>, da das Plugin keine eigene GUI besitzt
   */
  public boolean isVisible() {
    return false;
  }

  /**
   * Macht nichts, da das Plugin keine GUI besitzt.
   */
  public void setVisible(boolean visible) {
  }

  /**
   * Stoppt die Ausfuehrung des Plugins. Neuen Visualisierungs-Fenstern werden
   * keine Farbpaletten mehr hinzugefuegt.
   */
  public void stop() {
    started = false;
  }

  /**
   * Liest die Farbpaletten aus einer Datei.
   * @param inFile Eingabe-Datei
   */
  private void readColorMapsFromFile(File inFile) throws Exception {
    colorMaps.clear();
    BufferedReader input  = new BufferedReader( new FileReader(inFile) );
    String   line         = null;
    int      lineNo       = 0;
    ColorMap colorMap     = null;
    String   colorMapName = null;
    for(lineNo=1; input.ready() && (line=input.readLine().trim())!=null; lineNo++ ) {
      // Kommentarzeilen und Leerzeilen uebergehen
      if ( line.equals("") || IOUtil.isCommentLine(line) )
        continue;

      // Bei einem [<name>]-Tag beginnt eine neue ColorMap mit dem Namen <name>
      if ( line.startsWith("[") && line.endsWith("]") ) {
        colorMapName = line.substring(1,line.length()-1);
        colorMap     = new ColorMapImpl();
        colorMap.setType(ColorMap.TYPE_VALUES);
        colorMaps.put(colorMapName,colorMap);
        continue;
      }

      // Beim Tag DEFAULT wird die aktuelle ColorMap als Standard festgelegt
      if ( line.equals("DEFAULT") ) {
        defaultColorMap = colorMapName;
        continue;
      }

      // Ansonsten Zeile als ColorMapEntry interpretieren:
      //     <Quantity> <Color> [<Label>]
      if ( colorMap == null )
        throw new IOException("Line "+lineNo+": ColorMapEnty outside of [..]-section");
      StringTokenizer tok      = new StringTokenizer(line);
      Double          quantity = null;
      Color           color    = null;
      String          label    = "";
      // 1) Wert
      try {
        quantity = Double.parseDouble(tok.nextToken());
      } catch (Exception err) {
        String mess = "Line "+lineNo+": Quantity expected (double value) >> ";
        if ( err.getMessage() == null || err.getMessage().trim().equals("") )
          mess = mess + err.getClass().getSimpleName();
        else
          mess = mess + err.getMessage();
        throw new Exception(mess);
      }
      // 2) Farbe
      try {
        color = SwingUtil.parseColor(tok.nextToken());
      } catch (Exception err) {
        String mess = "Line "+lineNo+": Color specification expected (rgb or color name) >> ";
        if ( err.getMessage() == null || err.getMessage().trim().equals("") )
          mess = mess + err.getClass().getSimpleName();
        else
          mess = mess + err.getMessage();
        throw new Exception(mess);
      }
      // 3) Label (optional)
      try {
        String labelStr = tok.nextToken("#\n").trim();
        if ( !IOUtil.isCommentLine(labelStr) )
          label = labelStr;
      } catch (Exception err) {
        // Label ignorieren (wenn keine Token mehr uebrig sind)
        label = "";
      }

      // neuen Eintrag in der Farbpalette aufnehmen
      colorMap.addColorMapEntry( StylingUtil.createColorMapEntry(label,quantity,color,1.0) );
    }
// TEST
//    colorMaps.put("B/W", StylingUtil.builder.createColorMap(
//      new String[] { "","" },
//      new double[] { 0f,  999f },
//      new Color[]  { Color.BLACK, Color.WHITE },
//      ColorMap.TYPE_VALUES
//    ));
//    colorMaps.put("Test", StylingUtil.builder.createColorMap(
//      new String[] { "","","","","","" },
//      new double[] { 0f,  1f,  2f,  3f,  4f, 5f },
//      new Color[]  { Color.BLUE, Color.GRAY,Color.YELLOW,Color.RED,Color.CYAN, Color.MAGENTA },
//      ColorMap.TYPE_VALUES
//    ));
  }
}
