/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.BufferedWriter;
import java.io.FileWriter;

import java.util.Vector;
import java.util.StringTokenizer;
import java.util.Arrays;
import java.util.Comparator;

import schmitzm.lang.LangUtil;
import schmitzm.io.dyntxt.DynamicTextGenerator;
import schmitzm.io.dyntxt.DynamicInputProvider;
import schmitzm.io.dyntxt.DynamicField;
import schmitzm.io.dyntxt.DynamicLoop;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.temp.BaseTypeUtil;

import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.AbstractXuluModel;
import edu.bonn.xulu.model.StepModel;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.PropertiesResource;
// nur fuer Doku
import edu.bonn.xulu.model.ModelContentManager;

/**
 * Diese Klasse stellt einen Sourcecode-Generator dar, mit dem ein Grundgeruest
 * fuer {@link XuluModel}- und {@link StepModel}-Klassen (mit entsprechendem
 * {@link ModelContentManager}) erzeugt werden koennen.<br>
 * Der Generator basiert auf folgenden Vorlagen-Dateien:
 * <ul>
 *   <li>{@link #XULUMODEL_TEMPLATE_FILENAME XuluModel.tpl}</li>
 *   <li>{@link #STEPMODEL_TEMPLATE_FILENAME StepModel.tpl}</li>
 *   <li>{@link #CM_TEMPLATE_FILENAME        ModelContentManager.tpl}</li>
 * </ul>
 * Das Verzeichnis, aus dem die Dateien gezogen werden, wird im Konstruktor
 * angegeben.<br>
 * Auf Basis von
 * <ul>
 *   <li>Modell-Name</li>
 *   <li>Modell-Art (Oberklasse vom Typ {@link XuluModel})</li>
 *   <li>Package-Name</li>
 *   <li>Klassen-Verzeichnis</li>
 *   <li>Einer Liste von Ressourcen und Variablen-Namen</li>
 * </ul>
 * wird aus den Vorlagen-Dateien eine Klasse für den {@linkplain ModelContentManager ContentManager}
 * und eine Klasse fuer das {@linkplain XuluModel Xulu-Modell} erzeugt, in denen
 * die Ressourcen-Spezifikation, das Ressourcen-Handling, sowie deren Referenzierung
 * in lokalen Variablen bereits implementiert ist.<br>
 * Der Modell-Entwickler (unmittelbar) mit der Programmierung des
 * Modell-Algorithmus (z.B. in {@link StepModel#performModelStep(int)}) beginnen.
 * @see schmitzm.io.dyntxt.DynamicTextGenerator
 * @see schmitzm.io.dyntxt.DynamicInputProvider
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GeoModelCodeGenerator {
  /** Name der Vorlagen-Datei fuer {@link XuluModel}. */
  public static final String XULUMODEL_TEMPLATE_FILENAME = "XuluModel.tpl";
  /** Name der Vorlagen-Datei fuer {@link StepModel}. */
  public static final String STEPMODEL_TEMPLATE_FILENAME = "StepModel.tpl";
  /** Vorlagen-Datei fuer {@link ModelContentManager}. */
  public static final String CM_TEMPLATE_FILENAME        = "ModelContentManager.tpl";

  private File                     tplDir = null;
  private File                     modelFile = null;
  private File                     cmFile = null;
  private String                   modelName = null;
  private String                   packageName = null;
  private Class<XuluModel>         modelClass = null;
  private ModelResource[]          resource = null;
  private String[]                 resourceVarName = null;
  private ValuePropertyResource[]  propResource = null;
  private Integer[]                propResourceRefIdx = null;
  private Class[]                  neededImports = null;
  private DynamicInputProvider     inputProvider = null;


  /**
   * Erzeugt einen neuen ModellCode-Generator.
   * @param modelName       Name des Modells (darf keine Leer/Sonderzeichen enthalten)
   * @param packageName     Name des Pakets in dem Modell und ContentManager
   *                        implementiert werden
   * @param classDir File   Pfad unterhalb dem die Klassen abgelegt werden
   * @param modelClass      Typ der Modell-Klasse (nur {@link StepModel} oder {@link XuluModel} erlaubt)
   * @param resource        fuer das Modell benoetigte Ressourcen
   * @param resourceVarName Variablen-Namen fuer die modell-Internen Referenzen
   *                        auf die Ressourcen
   * @param tplDir          Verzeichnis, in dem die Vorlagen-Dateien zur Code-Generierung
   *                        zu finden sind
   */
  public GeoModelCodeGenerator(String modelName, String packageName, File classDir, Class<XuluModel> modelClass, ModelResource[] resource, String[] resourceVarName, File tplDir) {
    checkInputs(modelName,packageName,classDir,modelClass,resource,resourceVarName,tplDir);
    this.tplDir          = tplDir;
    this.modelFile       = new File(classDir, packageName.replace('.',File.separatorChar) + File.separator + modelName + ".java");
    this.cmFile          = new File(classDir, packageName.replace('.',File.separatorChar) + File.separator + modelName + "ContentManager.java");
    this.modelName       = modelName;
    this.packageName     = packageName;
    this.modelClass      = modelClass;
    this.resource        = resource;
    this.resourceVarName = resourceVarName;
    // Property-Ressourcen werden an einigen Stellen anders behandelt
    // -> Ressoucen und Referenz-Index auf normale Ressourcen-Arrays isolieren
    Vector<ValuePropertyResource> propRes       = new Vector<ValuePropertyResource>();
    Vector<Integer>               propResRefIdx = new Vector<Integer>();
    for (int i=0; i<resource.length; i++)
      if ( resource[i] instanceof ValuePropertyResource ) {
        propRes.add( (ValuePropertyResource)resource[i] );
        propResRefIdx.add( i );
      }
    this.propResource        = propRes.toArray( new ValuePropertyResource[0] );
    this.propResourceRefIdx  = propResRefIdx.toArray( new Integer[0] );

    // InputProvider fuer die Template-Dateien
    inputProvider = new GeoModelInputProvider();
  }

  /**
   * Prueft die Parameter des Konstruktors auf Korrektheit
   * @param modelName       Name des Modells (darf keine Leer/Sonderzeichen enthalten)
   * @param packageName     Name des Pakets in dem Modell und ContentManager
   *                        implementiert werden
   * @param classDir File   Pfad unterhalb dem die Klassen abgelegt werden
   * @param modelClass      Typ der Modell-Klasse (nur {@link StepModel} oder {@link XuluModel} erlaubt)
   * @param resource        fuer das Modell benoetigte Ressourcen
   * @param resourceVarName Variablen-Namen fuer die modell-Internen Referenzen
   *                        auf die Ressourcen
   * @exception IllegalArgumentException falls ein Parameter nicht korrekt ist
   */
  protected void checkInputs(String modelName, String packageName, File classDir, Class<XuluModel> modelClass, ModelResource[] resource, String[] resourceVarName, File tplDir) {
    // Verzeichnis fuer Vorlagen-Dateien muss existieren
    if ( tplDir == null || !tplDir.isDirectory() || !tplDir.exists() )
      throw new IllegalArgumentException("Directory for template files does not exist: "+(tplDir == null ? "null" : tplDir.getAbsolutePath()));
    // Modell-Namen pruefen
    try {
      LangUtil.checkIdentifierAndError(modelName);
    } catch (Exception err) {
      throw new IllegalArgumentException("Illegal model name: "+err.getMessage());
    }

    // Paket-Namen pruefen
    try {
      if ( packageName.startsWith(".") || packageName.endsWith(".") )
        throw new IllegalArgumentException("'.' not allowed as first or last charakter!");
      StringTokenizer tok = new StringTokenizer(packageName);
      for (;tok.hasMoreTokens();)
        LangUtil.checkIdentifierAndError(tok.nextToken(".\n"));
    } catch (Exception err) {
      throw new IllegalArgumentException("Illegal package name: "+err.getMessage());
    }

    // nur XuluModel oder StepModel zugelassen als Modell-Typ
    if ( modelClass == null )
      throw new IllegalArgumentException("NULL not allowed for model type!");
    if ( !modelClass.equals( XuluModel.class ) &&
         !modelClass.equals( StepModel.class ) )
      throw new IllegalArgumentException("Only XuluModel or StepModel allowed for model type!");

    // Ressourcen-Array und Variablennamen-Array muessen dieselbe Laenge haben
    if ( resource == null )
      throw new IllegalArgumentException("Resource array required (NULL not allowed)");
    if ( resourceVarName == null )
      throw new IllegalArgumentException("Variable name array required (NULL not allowed)");
    if ( resource.length != resourceVarName.length )
      throw new IllegalArgumentException("Arrays for resources and variable names must have the same length");

    // Variablen-Namen muessen Identifier sein
    int i = -1;
    try {
      for (i=0; i<resourceVarName.length; i++)
        LangUtil.checkIdentifierAndError(resourceVarName[i]);
    } catch (Exception err) {
      throw new IllegalArgumentException("Illegal variable name '"+resourceVarName[i]+"': "+err.getMessage());
    }
  }

  /**
   * Fuehrt die Generierung der Quell-Dateien fuer ContentManager und Modell aus.
   * @exception Exception wenn irgendein Fehler auftritt
   */
  public void generateSource() throws Exception {
    modelFile.getParentFile().mkdirs(); // Verzeichnisstruktur erzeugen
    generateModelContentManagerClass();
    generateModelClass();
  }

  /**
   * Generierung der Content-Manager-Klasse.
   * @exception Exception wenn irgendein Fehler auftritt
   */
  private void generateModelContentManagerClass() throws Exception {
    // Fuer ContentManager benoetigte Import-Klassen bestimmen
    Vector<Class> neededImp = new Vector<Class>();
    neededImp.add( AbstractModelContentManager.class );
    neededImp.add( ModelResource.class );

    for (int i=0; i<resource.length; i++) {
      // Basis-Klassen fuer Ressourcen
      if ( resource[i] instanceof PropertiesResource && !neededImp.contains( PropertiesResource.class ) )
        neededImp.add( PropertiesResource.class );
      if ( resource[i] instanceof ValuePropertyResource && !neededImp.contains( ValuePropertyResource.class ) )
        neededImp.add( ValuePropertyResource.class );
      // Ressourcen-Typ
      if ( !neededImp.contains( resource[i].getType() ) )
        neededImp.add( resource[i].getType() );
      // Datentyp eine ValuePropertyResource
      if ( resource[i] instanceof ValuePropertyResource &&
          !BaseTypeUtil.isBaseType(((ValuePropertyResource)resource[i]).getPropertyType().getType()) &&
          !neededImp.contains( ((ValuePropertyResource)resource[i]).getPropertyType().getType() ) )
        neededImp.add( ((ValuePropertyResource)resource[i]).getPropertyType().getType() );
    }
    neededImports = neededImp.toArray(new Class[0]);
    sortClassesByName( neededImports );

    // Ausgabe-Datei fuer den Content-Manager erzeugen
    FileInputStream  template = new FileInputStream( new File(tplDir,CM_TEMPLATE_FILENAME) );
    FileOutputStream output   = new FileOutputStream(cmFile);
    new DynamicTextGenerator(template).generateDynamicText(output,inputProvider);
    output.flush();
    output.close();

    // Klasse kompilieren
//    System.out.println( packageName+"."+modelName+".java --> "+ Compiler.compileClasses(packageName+"."+modelName+".java") );
  }

  /**
   * Generierung der Modell-Klasse.
   * @exception Exception wenn irgendein Fehler auftritt
   */
  private void generateModelClass() throws Exception  {
      // Fuer Modell benoetigte Import-Klassen bestimmen
      Vector<Class> neededImp = new Vector<Class>();
      neededImp.add( modelClass.equals(StepModel.class) ? AbstractStepModel.class : AbstractXuluModel.class );
      for (int i=0; i<resource.length; i++) {
        // Basis-Klassen fuer Ressourcen
        if ( resource[i] instanceof ValuePropertyResource ) {
          if ( ScalarProperty.class.isAssignableFrom(resource[i].getType()) ||
               MatrixProperty.class.isAssignableFrom(resource[i].getType()) ) {
            if ( !neededImp.contains( PropertyReadAccess.class ) )
              neededImp.add( PropertyReadAccess.class );
            if ( !neededImp.contains( PropertyWriteAccess.class ) )
              neededImp.add( PropertyWriteAccess.class );
          }
          if ( ListProperty.class.isAssignableFrom(resource[i].getType())  ) {
            if ( !neededImp.contains( ListPropertyReadAccess.class ) )
              neededImp.add( ListPropertyReadAccess.class );
            if ( !neededImp.contains( ListPropertyWriteAccess.class ) )
              neededImp.add( ListPropertyWriteAccess.class );
          }
        }
        // Ressourcen-Typ
        if ( !neededImp.contains( resource[i].getType() ) )
          neededImp.add( resource[i].getType() );
        // Datentyp eine ValuePropertyResource
        if ( resource[i] instanceof ValuePropertyResource &&
            !BaseTypeUtil.isBaseType(((ValuePropertyResource)resource[i]).getPropertyType().getType()) &&
            !neededImp.contains( ((ValuePropertyResource)resource[i]).getPropertyType().getType() ) )
          neededImp.add( ((ValuePropertyResource)resource[i]).getPropertyType().getType() );
      }
      neededImports = neededImp.toArray(new Class[0]);
      sortClassesByName( neededImports );

      // Ausgabe-Datei fuer die Modell-Klasse erzeugen
      String modelTemplateFileName = modelClass.equals(StepModel.class) ? STEPMODEL_TEMPLATE_FILENAME : XULUMODEL_TEMPLATE_FILENAME;
      FileInputStream  template = new FileInputStream( new File(tplDir,modelTemplateFileName) );
      FileOutputStream output   = new FileOutputStream( modelFile );
      new DynamicTextGenerator(template).generateDynamicText(output,inputProvider);
      output.flush();
      output.close();
  }

  private class GeoModelInputProvider implements DynamicInputProvider {
    /**
     * Definiert die Ersetzung der {@link DynamicTextGenerator}-Felder.
     * @param field zu verarbeitendes Feld
     * @exception UnsupportedOperationException falls das uebergebene Feld nicht
     *            verarbeitet werden kann
     */
    public String performField(DynamicField field) {
      // Java-Paket, in dem Modell und Content-Manager implementiert sind
      if (field.getID().equalsIgnoreCase("Package"))
        return packageName;
      // Ein benoetigter Klassen-Import (innerhalb der ClassImport-Schleife)
      if (field.getID().equalsIgnoreCase("ClassImport"))
        return neededImports[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).
            getActualLoop()].getName();
      // Modell-Name (= Klassenname+Konstruktor des Modells)
      if (field.getID().equalsIgnoreCase("ModelName"))
        return modelName;
      // Modell-OberKlasse
      if (field.getID().equalsIgnoreCase("ModelSuperClass"))
        return modelClass.equals(StepModel.class) ? "AbstractStepModel" : "AbstractXuluModel";
      // ContentManager-Name (= Klassenname+Konstruktor des Content-Manager)
      if (field.getID().equalsIgnoreCase("ModelCMName"))
        return modelName + "ContentManager";
      // ContentManager-OberKlasse
      if (field.getID().equalsIgnoreCase("ModelCMSuperClass"))
        return "AbstractModelContentManager";
      // Anzahl an Ressourcen
      if (field.getID().equalsIgnoreCase("ResourceCount"))
        return String.valueOf(resource.length);
      // Nummer einer Ressource (innerhalb einer Schleife ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceNo"))
        return String.valueOf( ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop());
      // Beschreibung einer Ressource (innerhalb einer Schleife ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceDesc"))
        return resource[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()].getDesc();
      // Variable einer Ressource (innerhalb einer Schleife ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceVar"))
        return resourceVarName[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()];
      // Typ einer Ressource (innerhalb einer Schleife ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceType"))
        return resource[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()].getType().getSimpleName();
      // Instanziierung einer Ressource (innerhalb einer Schleife ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceDefinition")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        StringBuffer buffer = new StringBuffer();
        if (resource[idx] instanceof ValuePropertyResource) {
          buffer.append("new ValuePropertyResource(")
                .append(getResourceCategoryString(resource[idx])).append(",")
                .append("\"").append(resource[idx].getDesc()).append("\",")
                .append(resource[idx].getType().getSimpleName()).append(".class,")
                .append( ( (ValuePropertyResource) resource[idx]).getPropertyType().getType().getSimpleName()).append(".class,")
                .append(resource[idx].isNullable())
                .append(")");
          return buffer.toString();
        }
        if (resource[idx] instanceof PropertiesResource) {
          buffer.append("new PropertiesResource(")
                .append(getResourceCategoryString(resource[idx])).append(",")
                .append("\"").append(resource[idx].getDesc()).append("\",")
                .append(resource[idx].getType().getSimpleName()).append(".class,")
                .append(resource[idx].isNullable())
                .append(")");
          return buffer.toString();
        }
      }
      // Deklaration der Access-Variable zu einer Ressource (innerhalb einer Schleife
      // ueber die Property-Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceAccessType")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        if (categoryHandledByWriteAccess(propResource[idx].getCategory()))
          return ListProperty.class.isAssignableFrom(propResource[idx].getType()) ? "ListPropertyWriteAccess" : "PropertyWriteAccess";
        else
          return ListProperty.class.isAssignableFrom(propResource[idx].getType()) ? "ListPropertyReadAccess" : "PropertyReadAccess";
      }
      if (field.getID().equalsIgnoreCase("ResourceAccessVar")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        if (categoryHandledByWriteAccess(propResource[idx].getCategory()))
          return "WA_" + resourceVarName[propResourceRefIdx[idx]];
        else
          return "RA_" + resourceVarName[propResourceRefIdx[idx]];
      }
      if (field.getID().equalsIgnoreCase("PropertyResourceDesc"))
        return propResource[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()].getDesc();
      if (field.getID().equalsIgnoreCase("PropertyResourceNo"))
        return String.valueOf(propResourceRefIdx[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()]);
      if (field.getID().equalsIgnoreCase("PropertyResourceType"))
        return propResource[ ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop()].getType().getSimpleName();
      // Deklaration der Zugriffs-Variable zu einer Ressource (innerhalb einer Schleife
      // ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceVarType")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        // Nicht-Property-Ressource werden direkt ueber das Objekt angesprochen
        if (! (resource[idx] instanceof ValuePropertyResource))
          return resource[idx].getType().getSimpleName();
        ValuePropertyResource propRes = (ValuePropertyResource) resource[idx];
        // Aus skalaren Propertys, die keinen Basis-Typ beinhalten, wird das
        // Objekt herausgeholt
        if (ScalarProperty.class.isAssignableFrom(propRes.getType()) &&
            !BaseTypeUtil.isBaseType(propRes.getPropertyType().getType()))
          return propRes.getPropertyType().getType().getSimpleName();
        // Andernfalls wird ueber das Lese/Schreibrecht auf die Property zugegriffen
        if (ListProperty.class.isAssignableFrom(propRes.getType()))
          return categoryHandledByWriteAccess(propRes.getCategory()) ? "ListPropertyWriteAccess" : "ListPropertyReadAccess";
        else
          return categoryHandledByWriteAccess(propRes.getCategory()) ? "PropertyWriteAccess" : "PropertyReadAccess";
      }
      // Wertzuweisung an die Access-Variable einer Ressource (innerhalb einer Schleife
      // ueber die Property-Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceAccessAlloc")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        if (categoryHandledByWriteAccess(propResource[idx].getCategory()))
          return ".getWriteAccess(this)";
        else
          return ".getReadAccess(this)";
      }
      // Wertzuweisung an die Zugriffs-Variable einer Ressource (innerhalb einer Schleife
      // ueber die Resourcen)
      if (field.getID().equalsIgnoreCase("ResourceVarAlloc")) {
        int idx = ( (DynamicLoop) field.getFather(DynamicLoop.class)).getActualLoop();
        // Nicht-Property-Ressource werden direkt ueber das Objekt angesprochen
        if (! (resource[idx] instanceof ValuePropertyResource))
          return "(" + resource[idx].getType().getSimpleName() + ")contManager.getResource(" + idx + ").getData()";
        ValuePropertyResource propRes = (ValuePropertyResource) resource[idx];
        // Aus skalaren Propertys, die keinen Basis-Typ beinhalten, wird das
        // Objekt herausgeholt
        if (ScalarProperty.class.isAssignableFrom(propRes.getType()) &&
            !BaseTypeUtil.isBaseType(propRes.getPropertyType().getType())) {
          StringBuffer buffer = new StringBuffer();
          buffer.append("(")
                .append(propRes.getPropertyType().getType().getSimpleName())
                .append(")")
                .append(categoryHandledByWriteAccess(propRes.getCategory()) ? "WA_" : "RA_")
                .append(resourceVarName[idx])
                .append(".getValue()");
          return buffer.toString();
        }
        // Andernfalls wird ueber das Lese/Schreibrecht auf die Property zugegriffen
        StringBuffer buffer = new StringBuffer();
        buffer.append(categoryHandledByWriteAccess(propRes.getCategory()) ? "WA_" : "RA_")
              .append(resourceVarName[idx]);
        return buffer.toString();
      }

      // Unbekanntes Feld
      throw new UnsupportedOperationException("Unknown field-ID: " + field.getID());
    }

    /**
     * Definiert das Abbruch-Kriterium der {@link DynamicTextGenerator}-Schleifen.
     * @param field zu verarbeitende Schleife
     * @exception UnsupportedOperationException falls die uebergebene Schleife nicht
     *            verarbeitet werden kann
     */
    public boolean performLoop(DynamicLoop loop) {
      // Schleife fuer alle zu importierenden Klassen
      if (loop.getID().equalsIgnoreCase("ClassImport"))
        return loop.getActualLoop() < neededImports.length;
      // Schleife ueber alle Ressourcen
      if (loop.getID().equalsIgnoreCase("Resources"))
        return loop.getActualLoop() < resource.length;
      // Schleife ueber alle Property-Ressourcen
      if (loop.getID().equalsIgnoreCase("PropertyResources"))
        return loop.getActualLoop() < propResource.length;

      throw new UnsupportedOperationException("Unknown loop-ID: " + loop.getID());
    }
  }

  /////////////////////////////////////////////////////////////////////////
  ////////////////////   statische Hilfsmethoden   ////////////////////////
  /////////////////////////////////////////////////////////////////////////
  /**
   * Prueft anhand der Ressourcen-Kategorie, ob eine Ressource ueber ein
   * Schreibrecht angesprochen wird
   * @param category Ressourcen-Kategorie
   */
  private static boolean categoryHandledByWriteAccess(int category) {
    return (category & (ModelResource.CATEGORY_OUTPUT | ModelResource.CATEGORY_TEMP)) > 0 ||
            category == ModelResource.CATEGORY_OTHER;
  }

  /**
   * Liefert die Bezeichnung einer Kategorie-Konstante fuer eine Modell-Ressource.
   * @param resource Model-Ressource
   */
  public static String getResourceCategoryString(ModelResource resource) {
    StringBuffer catString = new StringBuffer();
    if ( resource.isCategory( ModelResource.CATEGORY_INPUT ) )
      catString.append( "ModelResource.CATEGORY_INPUT" );
    if ( resource.isCategory( ModelResource.CATEGORY_OUTPUT ) )
      catString.append( catString.length() > 0 ? " | " : "").append( "ModelResource.CATEGORY_OUTPUT" );
    if ( resource.isCategory( ModelResource.CATEGORY_TEMP ) )
      catString.append( catString.length() > 0 ? " | " : "").append( "ModelResource.CATEGORY_TEMP" );
    if ( resource.isCategory( ModelResource.CATEGORY_OTHER ) )
      catString.append( catString.length() > 0 ? " | " : "").append( "ModelResource.CATEGORY_OTHER" );
    return catString.toString();
  }

  /**
   * Sortiert einen Array von Klassen nach der natuerlichen Sortierung ihrer
   * Namen (inklusive Package).
   * @param clazz Klassen
   * @see Class#getName()
   */
  public static void sortClassesByName(Class[] clazz) {
    Arrays.sort( clazz, new Comparator<Class>() {
      public int compare(Class c1, Class c2) {
        if ( c1 == null )
          return -1;
        return c1.getName().compareTo(c2.getName());
      }
    });
  }
}
