/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.io.File;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.KeyAdapter;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JCheckBox;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.ListSelectionModel;
import javax.swing.DefaultCellEditor;
import java.util.Vector;

import org.geotools.feature.FeatureCollection;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.lang.LangUtil;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.SwingWorker;
import schmitzm.swing.InputOption;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.ManualInputOption;
import schmitzm.swing.FileInputOption;
import schmitzm.swing.BooleanInputOption;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.event.InputOptionAdapter;
import schmitzm.swing.table.AbstractMutableTableModel;
import schmitzm.swing.table.MutableTable;
import schmitzm.swing.table.MutableTableModel;
import schmitzm.swing.table.ComponentRenderer;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.gui.XuluInternalFrame;
import edu.bonn.xulu.gui.XuluGUIMessages;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.PropertiesResource;
import edu.bonn.xulu.model.StepModel;
import edu.bonn.xulu.model.XuluModel;

// nur fuer Doku
import javax.swing.table.TableModel;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.plugin.appl.GeoModelCodeGeneratorPlugin;

/**
 * Diese Klasse stellt eine GUI zur Vefuegung, mit der der {@link GeoModelCodeGenerator}
 * gesteuert werden kann. Sie ist als {@linkplain XuluInternalFrame internes}
 * Fenster der {@linkplain XuluModellingPlatform Xulu-Modelling-Platform}
 * realisiert. Ueber das Xulu-Plugin {@link GeoModelCodeGeneratorPlugin} wird
 * sie automatisch in das Xulu-Menue eingebunden.<br>
 * Ueber die GUI spezifiert der Modell-Entwickler
 * <ul>
 *   <li>den Modell-Typ</li>
 *   <li>den Modell-Namen</li>
 *   <li>die zur Modelllierung benoetigten Ressourcen</li>
 *   <li>Variablennamen, ueber die die Ressourcen Modell-<b>intern</b>
 *       angesprochen werden sollen</li>
 * </ul>
 * Daraus erzeugt der {@link GeoModelCodeGenerator} je eine Java-Klasse (als
 * Source-Code) fuer {@linkplain ModelContentManager Content-Manager} und
 * {@linkplain XuluModel Xulu-Modell}, in denen die Ressourcen-Spezifikation,
 * das Ressourcen-Handling, sowie deren Referenzierung in lokalen Variablen
 * bereits implementiert ist.<br>
 * Der Modell-Entwickler (unmittelbar) mit der Programmierung des
 * Modell-Algorithmus (z.B. in {@link StepModel#performModelStep(int)}) beginnen.<br>
 * Der {@link GeoModelCodeGenerator} benoetigt einige Vorlagen-Dateien, aus denen
 * der Sourcecode generiert wird. Diese muessen im Verzeichnis "GeoModelCodeGenerator"
 * unterhalb des {@linkplain XuluModellingPlatform#getPluginDirectory() Xulu-Plugin-Verzeichnisses}
 * abgelegt sein.
 * @see GeoModelCodeGenerator
 * @see GeoModelCodeGeneratorPlugin
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GeoModelCodeGeneratorGUI extends XuluInternalFrame {
  /** Verzeichnis-Name (unterhalb des {@linkplain XuluModellingPlatform#getPluginDirectory() Xulu-Plugin-Verzeichnisses}
   *  in dem die Vorlagen-Dateien fuer Code-Generierung hinterlegt sind. */
  public static final String tplDirName = "GeoModelCodeGenerator";

  /**
   * Vorgabe, aus welchen Modell-Typen der Anwender auswaehlen kann.
   */
  public static final Class[] MODELTYPE = new Class[] {
      StepModel.class,
      XuluModel.class
  };

  /**
   * Vorgabe, aus welchen Ressourcen-Categorien der Anwender auswaehlen kann.
   */
  public static final Integer[] CATEGORY = new Integer[] {
      ModelResource.CATEGORY_OTHER,
      ModelResource.CATEGORY_INPUT,
      ModelResource.CATEGORY_OUTPUT,
      ModelResource.CATEGORY_INPUT | ModelResource.CATEGORY_OUTPUT,
      ModelResource.CATEGORY_TEMP,
  };

  /**
   * Beschreibungen fuer die Ressourcen-Categorien, aus welchen der Anwender auswaehlen kann.
   */
  public static final String[] CATEGORY_DESC = new String[] {
      "",
      "Input",
      "Output",
      "Input/Output",
      "Temporal",
  };

  /**
   * Vorgabe, aus welchen Ressourcen-Typen der Anwender auswaehlen kann (
   * {@link ScalarProperty} , {@link ListProperty}, {@link MatrixProperty}).
   */
  public static final Class[] RESOURCETYPE = new Class[] {
      ScalarProperty.class,
      ListProperty.class,
      MatrixProperty.class
  };

  /**
   * Vorgabe, aus welchen Property-Datentypen der Anwender auswaehlen kann (
   * {@link Integer} , {@link Float}, {@link Long}, {@link Double}, {@link Byte},
   * {@link Short}, {@link Boolean}, {@link WritableGrid} , {@link FeatureCollection}).
   */
  public static final Class[] DATATYPE = new Class[] {
      Integer.class,
      Float.class,
      Long.class,
      Double.class,
      Byte.class,
      Short.class,
      Boolean.class,
      WritableGrid.class,
      FeatureCollection.class,
  };

  /** Speichert alle zur Auswahl stehende Ressourcen-Typen ({@link #RESOURCETYPE} + alle
   *  in Xulu registrierten {@link XuluObject}-Typen). */
  protected Class[] resourceType = null;
  /** Speichert eine Referenz auf die Xulu-Applikation, in der der Code-Generator
   *  ausgefuehrt wird. */
  protected XuluModellingPlatform appl = null;
  /** Auswahlfeld fuer den Modell-Typ. */
  protected SelectionInputOption.Combo modelType = null;
  /** Eingabefeld fuer den Modell-Namen. */
  protected ManualInputOption.Text modelName = null;
  /** Tabelle in der die Ressourcen spezifiziert werden. */
  protected MutableTable resources = null;
  /** Butten zum Starten der Generierung. */
  protected JButton startButton = null;
  /** Referenz auf den Container des Fensters. */
  protected Container contentPane = null;

  /**
   * Erzeugt eine neue GUI fuer den {@link GeoModelCodeGenerator}.
   * @param appl Instanz der Xulu-Applikation, in der der Code-Generator
   *             ausgefuehrt wird
   */
  public GeoModelCodeGeneratorGUI(XuluModellingPlatform appl) {
    super("Source code generator for geo models");
    this.appl = appl;
    resourceType = this.appendXuluTypesToResourceTypes(appl);

    this.setVisible(false);
    this.setSize( new Dimension(500,500) );
    this.setClosable(true);
    this.setDefaultCloseOperation( HIDE_ON_CLOSE );
    this.getContentPane().setLayout( new GridBagLayout() );
    contentPane  = this.getContentPane();
    initGUI();
  }

  /**
   * Initalisiert die GUI des Fensters.
   */
  protected void initGUI() {
    modelType = new SelectionInputOption.Combo(null,true,MODELTYPE,LangUtil.getClassNames(MODELTYPE,true));
    modelName = new ManualInputOption.Text(null,true) {
      public boolean performIsInputValid() {
        // Modell-Name darf keine Leerzeichen, Punkte, Slashes und Backslashes enthalten
        return super.performIsInputValid() &&
              ((String)performGetValue()).indexOf(" ") < 0 &&
              ((String)performGetValue()).indexOf(".") < 0 &&
              ((String)performGetValue()).indexOf("/") < 0 &&
              ((String)performGetValue()).indexOf("\\") < 0;
      }
    };
    // Modellname soll immer mit einem Grossbuchstaben beginnen,
    // wegen Java-Konvention fuer Klassen-Namen
    modelName.addInputOptionListener(new InputOptionAdapter() {
      public void optionChanged(InputOption option, Object oldValue, Object newValue) {
        String newName = (String)newValue;
        if ( newName == null )
          return;
        newName = newName.trim();
        if ( newName.length() > 1 )
          newName = newName.substring(0,1).toUpperCase() + newName.substring(1);
        modelName.setValue( newName );
      }
    });

    resources = new MutableTable( new ResourceTableModel(), MutableTable.ITEM_ADD | MutableTable.ITEM_REMOVE );
    resources.setColumnSelectionAllowed(false);
    resources.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    resources.setRowHeight( new JComboBox(new String[] {"Dummy"}).getPreferredSize().height );
    // Renderer fuer die Tabelle setzen
    resources.getColumnModel().getColumn(0).setCellRenderer( new ComponentRenderer.JComboBox() );
    resources.getColumnModel().getColumn(1).setCellRenderer( new ComponentRenderer.JTextField() );
    resources.getColumnModel().getColumn(2).setCellRenderer( new ComponentRenderer.JTextField() );
    resources.getColumnModel().getColumn(3).setCellRenderer( new ComponentRenderer.JComboBox() );
    resources.getColumnModel().getColumn(4).setCellRenderer( new ComponentRenderer.JComboBox() );
    resources.getColumnModel().getColumn(5).setCellRenderer( new ComponentRenderer.JCheckBox() );
    // Editoren fuer die Tabelle setzen
    resources.getColumnModel().getColumn(0).setCellEditor( new DefaultCellEditor( new JComboBox( CATEGORY_DESC ) ) );
    resources.getColumnModel().getColumn(1).setCellEditor( new DefaultCellEditor( new JTextField() ) );
    resources.getColumnModel().getColumn(2).setCellEditor( new DefaultCellEditor( new JTextField()) );
    resources.getColumnModel().getColumn(3).setCellEditor( new DefaultCellEditor( new JComboBox( LangUtil.getClassNames(resourceType,true) ) ) );
    resources.getColumnModel().getColumn(4).setCellEditor( new DefaultCellEditor( new JComboBox( LangUtil.getClassNames(DATATYPE,true) ) ) );
    resources.getColumnModel().getColumn(5).setCellEditor( new DefaultCellEditor( new JCheckBox() {
      public int getHorizontalAlignment() {
        return this.CENTER;
      }
      public int getVerticalAlignment() {
        return this.CENTER;
      }
    }));

    // Button zum Starten
    startButton = new JButton("Generate...");
    startButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        startGeneration();
      }
    });

    // Komponenten in Fenster anordnen
    contentPane.add( new JLabel("Modell-Typ:"),   new GridBagConstraints(0,0,1,1,  0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,10,5,10),0,5) );
    contentPane.add( modelType,                   new GridBagConstraints(1,0,1,1,1.0,0,GridBagConstraints.WEST, GridBagConstraints.BOTH, new Insets(5,10,5,10),0,5) );
    contentPane.add( new JLabel("Modell-Name:"),  new GridBagConstraints(0,1,1,1,  0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,10,5,10),0,0) );
    contentPane.add( modelName,                   new GridBagConstraints(1,1,1,1,1.0,0,GridBagConstraints.WEST, GridBagConstraints.BOTH, new Insets(5,10,5,10),0,0) );
    contentPane.add( new JLabel("Ressourcen:"),   new GridBagConstraints(0,2,1,1,  0,0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,10,0,10),0,0) );
    contentPane.add( resources.createScrollPane(),new GridBagConstraints(0,3,2,1,1.0,1.0,GridBagConstraints.WEST, GridBagConstraints.BOTH, new Insets(5,10,5,10),0,0) );
    contentPane.add( startButton,                 new GridBagConstraints(0,4,1,1,0.0,0.0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,10,5,10),0,0) );
  }

  /**
   * Aktualisiert die Anzeige des Xulu-internen Fensters, in dem die
   * Ressourcen-Tabelle aktualisiert wird.
   */
  public void refresh() {
    ((MutableTableModel)resources.getModel()).fireTableDataChanged();
  }

  /**
   * Prueft die in der Maske spezifizierten Angaben auf Korrektheit. Ist eine
   * Angabe nicht korrekt, wird ein entsprechender Fehler-Dialog angezeigt.
   */
  private boolean checkInputs() {
    int i = -1;
    try {
      // Modell-Angaben checken
      if ( !modelType.isInputValid() )
        throw new IllegalArgumentException("Der angegebene Modelltyp ist ungueltig!");
      if ( !modelName.isInputValid() || !LangUtil.checkIdentifier(modelName.getValue()) )
        throw new IllegalArgumentException("Der angegebene Modellname ist ungueltig (keine Leer- oder Sonderzeichen erlaubt)!");

      // Ressourcen checken
      ResourceTableModel.ResourceDefinition[] resourceDef = ((ResourceTableModel)resources.getModel()).getResources();
      for (i=0; i<resourceDef.length; i++) {
        if ( !resourceDef[i].category.isInputValid() )
          throw new IllegalArgumentException("Kategorie ungueltig!");
        if ( !resourceDef[i].name.isInputValid() || !LangUtil.checkIdentifier(resourceDef[i].name.getValue()) )
          throw new IllegalArgumentException("Variablen-Name ungueltig (keine Leer- oder Sonderzeichen erlaubt)!");
        if ( !resourceDef[i].desc.isInputValid() )
          throw new IllegalArgumentException("Beschreibung ungueltig!");
        if ( !resourceDef[i].type.isInputValid() )
          throw new IllegalArgumentException("Ressourcen-Typ ungueltig!");
        if ( !resourceDef[i].datatype.isInputValid() )
          throw new IllegalArgumentException("Datentyp ungueltig!");
        if ( ValueProperty.class.isAssignableFrom( (Class)resourceDef[i].type.getValue() ) &&
             resourceDef[i].datatype == null )
          throw new IllegalArgumentException("Es muss ein Property-Typ angegeben werden!");
      }
    } catch (Exception err) {
      if ( i >= 0 )
        XuluGUIMessages.showError(this,err,"Fehler in Ressource "+(i+1)+": "+err.getMessage());
      else
        XuluGUIMessages.showError(this,err);
      return false;
    }
    return true;
  }

  /**
   * Erzeugt den kompletten Source-Dateinamen aus Klassenname, Paketname und
   * Source-Basisverzeichnis
   * @param className          Eingabenfeld fuer Klassenname
   * @param packageName        Eingabefeld fuer Package-Name
   * @param classDir           Eingabefeld fuer Source-Verzeichniss
   * @param contentManagerFile wenn {@code true} wird dem Klassenname das Suffix
   *                           "ContentManager" angehaengt
   */
  private static String generateOutFileName(ManualInputOption.Text className, ManualInputOption.Text packageName, FileInputOption classDir, boolean contentManagerFile) {
    StringBuffer buffer = new StringBuffer();
    if ( classDir.getValue() != null )
//    FileInputOption.getValue() liefert waehrend der Text-Eingabe noch nicht
//    das neue File-Objekt. Dies wird erst erzeugt, wenn das Text-Feld verlassen wird!
//    -> Dynamische Anzeige der Dateien waehrend der Text-Eingabe klappt nur
//       wenn direkt der Wert des Text-Feldes abgegriffen wird!
//    buffer.append( classDir.getValue().getAbsolutePath() );
      buffer.append( ((JTextField)classDir.getInputComponent()).getText() );
    // Seperator nur anhaengen, wenn Klassen-Verzeichnis nicht bereits mit einem
    // solchen  endet
    if ( buffer.length() == 0 || buffer.charAt( buffer.length()-1 ) != File.separatorChar )
      buffer.append( File.separatorChar );
    // Package als Verzeichnis + Seperator + Klassenname
    buffer.append( packageName.getValue().replace('.',File.separatorChar) )
          .append( File.separator )
          .append( className.getValue() );
    // Fuer die CM-Klasse noch das entsprechende Suffix anhaengen
    if ( contentManagerFile )
      buffer.append( "ContentManager" );
    // Endung ".java"
    buffer.append( ".java" );
    return buffer.toString();
  }

  /**
   * Startet die Sourcecode-Generierung, wenn der Start-Button aktiviert wurde.
   * Zuvor werden noch technische Information (wie Package-Name und Klassen-Verzeichnis)
   * durch einen Anwender-Dialog abgefragt.
   */
  private void startGeneration() {
    // Eingaben pruefen
    if ( !checkInputs() )
      return;

    // Ressourcen erzeugen
    ResourceTableModel.ResourceDefinition[] resourceDef = ((ResourceTableModel)resources.getModel()).getResources();
    ModelResource[]                         resource = new ModelResource[resourceDef.length];
    String[]                                varName  = new String[resourceDef.length];
    for (int i=0; i<resource.length; i++) {
      // Variablenname fuer Ressource
      varName[i] = (String)resourceDef[i].name.getValue();
      // Wenn "Datentyp" befuellt ist, handelt es sich um eine Property-Ressource,
      // ansonsten wird eine PropertiesRessource (fuer ein Xulu-Objekt) erzeugt
      if ( ValueProperty.class.isAssignableFrom( (Class)resourceDef[i].type.getValue() ) )
        resource[i] = new ValuePropertyResource(
          (Integer)resourceDef[i].category.getValue(),
          (String)resourceDef[i].desc.getValue(),
          (Class)resourceDef[i].type.getValue(),
          (Class)resourceDef[i].datatype.getValue(),
          (Boolean)resourceDef[i].optional.getValue()
        );
      else
        resource[i] = new PropertiesResource(
          (Integer)resourceDef[i].category.getValue(),
          (String)resourceDef[i].desc.getValue(),
          (Class)resourceDef[i].type.getValue(),
          (Boolean)resourceDef[i].optional.getValue()
        );
    }

    // Generierungs-Optionen abfragen
    final ManualInputOption.Text packageName    = new ManualInputOption.Text("Java-Package fuer Modell:",true,"edu.bonn.xulu.plugin.model");
    final FileInputOption        classDirName   = new FileInputOption("Compile-Verzeichnis:",true,/*"temp"*/appl.getDynamicClassRootDirectory());
    final ManualInputOption.Text modelFileName  = new ManualInputOption.Text("Ausgabe-Dateien:",true,generateOutFileName(modelName,packageName,classDirName,false));
    final ManualInputOption.Text cmFileName     = new ManualInputOption.Text(null,true,generateOutFileName(modelName,packageName,classDirName,true));
    classDirName.getFileChooser().setFileSelectionMode( JFileChooser.DIRECTORIES_ONLY );
    classDirName.getFileChooser().setDialogType( JFileChooser.OPEN_DIALOG );
    ((JTextField)modelFileName.getInputComponent()).setEditable(false);
    ((JTextField)cmFileName.getInputComponent()).setEditable(false);
    // Angabe der Ausgabe-Datei soll sich dynamisch aendern, sobald sich
    // Paket oder Klassen-Pfad aendern
    KeyListener keyListener = new KeyAdapter() {
      public void keyReleased(KeyEvent e) {
        modelFileName.setValue( generateOutFileName(modelName,packageName,classDirName,false) );
        cmFileName.setValue( generateOutFileName(modelName,packageName,classDirName,true) );
      }
    };
    InputOptionAdapter inputListener = new InputOptionAdapter() {
      public void optionChanged(InputOption option, Object oldValue, Object newValue) {
        modelFileName.setValue( generateOutFileName(modelName,packageName,classDirName,false) );
        cmFileName.setValue( generateOutFileName(modelName,packageName,classDirName,true) );
      }
    };

    packageName.getInputComponent().addKeyListener( keyListener );
    classDirName.getInputComponent().addKeyListener( keyListener );
    classDirName.addInputOptionListener( inputListener );

    SwingWorker worker = null;
    for (;;) {
      Object[] output = MultipleOptionPane.showMultipleInputDialog(this,"Output-Configuration", new InputOption[] {
          packageName,
          classDirName,
          modelFileName,
          cmFileName
      });
      // Dialog abgebrochen
      if ( output == null )
        return;
      // Dialog abgeschlossen
      File modelFile = new File((String)output[2]);
      File cmFile    = new File((String)output[3]);
      // Sicherheitsabfrage, falls Dateien bereits existieren
      if ( modelFile.exists() && MultipleOptionPane.showConfirmDialog(this,"Modell-Datei existiert bereits! Überschreiben?","Warnung...",MultipleOptionPane.OK_CANCEL_OPTION,MultipleOptionPane.WARNING_MESSAGE) == MultipleOptionPane.CANCEL_OPTION )
        continue; // zum Dialog zurueckkehren
      if ( cmFile.exists() && MultipleOptionPane.showConfirmDialog(this,"ContentManager-Datei existiert bereits! Überschreiben?","Warnung...",MultipleOptionPane.OK_CANCEL_OPTION,MultipleOptionPane.WARNING_MESSAGE) == MultipleOptionPane.CANCEL_OPTION )
        continue; // zum Dialog zurueckkehren

      // Generator erzeugen (dabei Check der zusaetzlichen Angaben)
      try {
        final GeoModelCodeGenerator generator = new GeoModelCodeGenerator(
            (String) modelName.getValue(),
            (String) packageName.getValue(),
            classDirName.getValue(),
            (Class) modelType.getValue(),
            resource,
            varName,
            new File(appl.getPluginDirectory(),tplDirName)
        );
        // SwingWorker erzeugen, der die Generierung durchfuehrt
        worker = new SwingWorker(
          new SwingWorker.Work() {
            public Object execute() throws Exception {
              generator.generateSource();
              return null;
            }
            public void performError(Exception err) {
              // Fehler bei der Code-Generierung verarbeiten
              XuluGUIMessages.showError(appl.getMainFrame(),err);
            }
          },
          appl.getMainFrame(),
          XuluConstants.SWING_RES.getString("WaitMess")
        );
        // keine Fehler -> Schleife verlassen
        break;
      } catch (Exception err) {
        // Fehler bei der CodeGenerator-Erzeugung verarbeiten
        XuluGUIMessages.showError(appl.getMainFrame(),err);
        continue; // zum Dialog zurueckkehren
      }
    } // End Loop

    // Generierung starten
    worker.start();
    if ( !worker.isCanceled() )
      XuluGUIMessages.showInfoMessage(appl.getMainFrame(),"Modell-Dateien wurden erfolgreich erzeugt...");
  }

  /**
   * Liefert die Eintraege von {@link #RESOURCETYPE} plus alle aktuell in der
   * Xulu-Applikation geladenen {@link XuluObject}-Typen in einem gemeinsamen
   * Array.
   */
  private static Class[] appendXuluTypesToResourceTypes(XuluModellingPlatform appl) {
    Class[] xuluType    = appl.getRegistry().getDataTypes();
    Class[] allResType  = new Class[ RESOURCETYPE.length + xuluType.length ];
    // die registrierten Xulu-Datentypen an den Array der
    // Standard-Typen (Propertys) anhaengen
    for (int i=0; i<allResType.length; i++)
      allResType[i] = ( i < RESOURCETYPE.length ) ? RESOURCETYPE[i] : xuluType[i-RESOURCETYPE.length];
    return allResType;
  }

  /////////////////////////////////////////////////////////////////////////
  ////////////   Tabellen-Modell fuer die Ressourcen-Tabelle   ////////////
  /////////////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt ein {@link TableModel} fuer die Ressourcen-Tabelle
   * der {@link GeoModelCodeGeneratorGUI} dar. Diese enthaelt 6 Spalten:
   * <ul>
   *   <li>Ressourcen-Kategorie (Input/Output/Temporaer/.., siehe {@link ModelResource#getCategory()})</li>
   *   <li>Variablenname, mit dem die Ressource im Source-Code angesprochen werden soll</li>
   *   <li>Ressourcen-Beschreibung</li>
   *   <li>Typ des in der Ressourc gespeicherten Objekts</li>
   *   <li>Property-Typ (falls es sich bei der Ressource um eine Property-Ressource handelt)</li>
   *   <li>Flag, ob es sich um eine optionale Ressource handelt</li>
   * </ul>
   * Die Tabelle ist inital leer und wird sukzessive durch den Anwender befuellt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class ResourceTableModel extends AbstractMutableTableModel {
    /** Speichert die Eintraege der Tabelle */
    protected Vector<ResourceDefinition> resourceDef = new Vector<ResourceDefinition>();

    /**
     * Erzeugt ein neues Tabellenmodell.
     */
    public ResourceTableModel() {
      super();
//      resourceDef.add( new ResourceDefinition( ModelResource.CATEGORY_OTHER,
//                                               "intRes",
//                                               "Dummy-Ressource 1",
//                                               ScalarProperty.class,
//                                               Integer.class,
//                                               true) );
//      resourceDef.add( new ResourceDefinition( ModelResource.CATEGORY_INPUT,
//                                               "gridListRes1",
//                                               "Dummy-Ressource 2",
//                                               edu.bonn.xulu.plugin.data.grid.GridList.class,
//                                               null,
//                                               true) );
//      resourceDef.add( new ResourceDefinition( ModelResource.CATEGORY_OUTPUT,
//                                               "gridListRes2",
//                                               "Dummy-Ressource 3",
//                                               ListProperty.class,
//                                               schmitzm.data.WritableGrid.class,
//                                               false) );
//      resourceDef.add( new ResourceDefinition( ModelResource.CATEGORY_OUTPUT,
//                                               "gridRes",
//                                               "Dummy-Ressource 4",
//                                               ScalarProperty.class,
//                                               schmitzm.data.WritableGrid.class,
//                                               false) );
    }

    /**
     * Liefert die Spaltennamen der Tabelle.
     */
    public String[] createColumnNames() {
      return new String[] {
          "Kategorie", "Variablen-Name", "Beschreibung", "Ressourcen-Typ", "Datentyp", "Optional"
      };
    }
    
    /**
     * Liefert die aktruellen Eintraege der Tabelle.
     */
    public ResourceDefinition[] getResources() {
      return resourceDef.toArray( new ResourceDefinition[0] );
    }

    /**
     * Fuegt eine neue (zunaechst leere) Ressouce hinzu.
     */
    public void performAddRow() {
      resourceDef.add( new ResourceDefinition() );
    }

    /**
     * Loescht eine Ressouce aus der Liste.
     */
    public void performRemoveRow(int row) {
      resourceDef.remove(row);
    }

    /**
     * Machts nichts, da die Werte direkt geaendert werden und nicht auf einen
     * Doppelklick reagiert wird.
     * @param row Zeilenindex
     * @param col Spaltenindex
     */
    public void performChangeData(int row, int col) {
    }

    /**
     * Liefert die Zeilenanzahl der Tabelle.
     */
    public int getRowCount() {
      return resourceDef.size();
    }

    /**
     * Liefert einen Wert der Tabelle.
     * @param row Zeilenindex
     * @param col Spaltenindex
     */
    public Object getValueAt(int row, int col) {
      switch (col) {
        case 0: return resourceDef.elementAt(row).category.getSelectedDisplayItem();
        case 1: return resourceDef.elementAt(row).name.getValue();
        case 2: return resourceDef.elementAt(row).desc.getValue();
        case 3: return resourceDef.elementAt(row).type.getSelectedDisplayItem();
        case 4: return resourceDef.elementAt(row).datatype.getSelectedDisplayItem();
        case 5: return resourceDef.elementAt(row).optional.getValue();
      }

      return null;
    }

    /**
     * Liefert nur {@code false}, wenn es sich um Spalte 4 handelt
     * ("Datentyp") und die betreffende Ressource <b>keine</b> Property-Ressource
     * ist.
     * @param row Zeilenindex
     * @param col Spaltenindex
     */
    public boolean isCellEditable(int row, int col) {
      switch( col ) {
        case 4: // Datentyp soll nur aktiviert sein, wenn als Resource-Typ eine
                // ValueProperty angewaehlt ist
                Class type = (Class)resourceDef.elementAt(row).type.getValue();
                return type != null && ValueProperty.class.isAssignableFrom(type);
      }
      return true;
    }

    /**
     * Veraendert einen Wert einer Tabellenzelle.
     * @param value neuer Zellen-Wert
     * @param row Zeilenindex
     * @param col Spaltenindex
     */
    public void setValueAt(Object value, int row, int col) {
      if ( row >= getRowCount() )
        return;
      switch (col) {
        case 0: resourceDef.elementAt(row).category.setSelectedDisplayItem(value); break;
        case 1: resourceDef.elementAt(row).name.setValue(value); break;
        case 2: resourceDef.elementAt(row).desc.setValue(value); break;
        case 3: resourceDef.elementAt(row).type.setSelectedDisplayItem(value); break;
        case 4: resourceDef.elementAt(row).datatype.setSelectedDisplayItem(value); break;
        case 5: resourceDef.elementAt(row).optional.setValue( value/*!(Boolean)resourceDef.elementAt(row).optional.getValue()*/ ); break;
      }
    }

    /**
     * Diese Klasse repraesentiert eine Zeile der Ressourcen-Tabelle. Sie besteht
     * aus 6 Eingabefeldern:
     * <ul>
     *   <li>Auswahlfeld fuer Ressourcen-Kategorie</li>
     *   <li>Text-Eingabe fuer Variablen-Name</li>
     *   <li>Text-Eingabe fuer Ressourcen-Beschreibung</li>
     *   <li>Auswahlfeld fuer Ressourcen-Typ</li>
     *   <li>Auswahlfeld fuer Datentyp einer Property-Ressource</li>
     *   <li>Checkbox fuer "optional"-Flag</li>
     * </ul>
     * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
     * @version 1.0
     */
    protected class ResourceDefinition {
      /** Auswahlfeld fuer die Resourcen-Kategorie. */
      protected SelectionInputOption.Combo category = null;
      /** Textfeld fuer den Resourcen-Namen (mit dem die Ressource im Source-
       *  Code angesprochen wird. */
      protected ManualInputOption.Text name = null;
      /** Textfeld fuer die Resourcen-Beschreibung. */
      protected ManualInputOption.Text desc = null;
      /** Auswahlfeld fuer den Resourcen-Typ. */
      protected SelectionInputOption.Combo type = null;
      /** Auswahlfeld fuer den Datentyp einer Property-Resource. */
      protected SelectionInputOption.Combo datatype = null;
      /** Checkbox fuer die Optional-Angabe. */
      protected BooleanInputOption optional = null;

      /**
       * Erzeugt eine neue Zeile. Die Eingabefelder sind alle leer.
       */
      public ResourceDefinition() {
        super();
        category = new SelectionInputOption.Combo(null,true,CATEGORY,CATEGORY_DESC);
        name     = new ManualInputOption.Text(null,true);
        SwingUtil.setPreferredWidth(name,150);
        desc     = new ManualInputOption.Text(null,true);
        SwingUtil.setPreferredWidth(desc,150);
        type     = new SelectionInputOption.Combo(null,true,resourceType,LangUtil.getClassNames(resourceType,true));
        datatype = new SelectionInputOption.Combo(null,true,DATATYPE,LangUtil.getClassNames(DATATYPE,true));
        optional = new BooleanInputOption(null,false);
        // Wenn kein ValueProperty-Typ fuer die Ressource ausgewaehlt ist,
        // wird der Datentyp autom. auf NICHTS gesetzt
        type.addInputOptionListener(new InputOptionAdapter() {
          public void optionChanged(InputOption option, Object oldValue, Object newValue) {
            if (!ValueProperty.class.isAssignableFrom( (Class) newValue) ) {
              datatype.setValue(null);
              datatype.setInputNeeded(false);
              fireTableRowsUpdated(0,0); // klappt merkwuerdigerweise auch fuer andere Zeilen!
            } else {
              datatype.setInputNeeded(true);
            }
          }
        });
      }

      /**
       * Erzeugt eine neue vorbelegte Zeile.
       * @param category Ressourcen-Kategorie (siehe {@link ModelResource#getCategory()})
       * @param name     Name der Ressourcen-Variable
       * @param desc     Beschreibung der Ressource
       * @param type     Typ der Ressource
       * @param datatype Datentyp (falls es sich beim Typ eine {@link ValueProperty} handelt)
       */
      public ResourceDefinition(int category, String name, String desc, Class type, Class datatype, boolean optional) {
        this();
        this.category.setSelectedItem( category );
        this.name.setValue(name);
        this.desc.setValue(desc);
        this.type.setSelectedItem(type);
        this.datatype.setSelectedItem(datatype);
        this.optional.setValue(optional);
      }

    }

  }

}
