/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.appl;

import java.util.Vector;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.appl.XuluPlugin;
import edu.bonn.xulu.appl.ModelControlManager;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.event.ModelListener;
import edu.bonn.xulu.model.event.ModelEvent;
import edu.bonn.xulu.model.event.ModelStartedEvent;
import edu.bonn.xulu.model.event.ModelStoppedEvent;
import edu.bonn.xulu.model.event.ModelStepStartedEvent;
import edu.bonn.xulu.model.event.ModelStepFinishedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;
import edu.bonn.xulu.gui.ModelControlFrame;

import schmitzm.lang.AbstractNamedObject;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectEvent;

/**
 * Dieses Plugin koppelt sich an alle in Xulu geladenen Modelle und
 * gibt deren absolute Laufzeit (in Sekunden aus). Bei Step-Modellen wird
 * auch die Laufzeit pro Modell-Schitt ausgegeben.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ModelTimeSnifferPlugin extends AbstractNamedObject implements XuluPlugin, ObjectListener {
  /** Flag, ob das Plugin gestartet ist, oder nicht */
  protected boolean started = false;
  /** Speichert die Xulu-Instanz in der das Plugin ausgefuehrt wird. */
  protected XuluModellingPlatform appl = null;
  /** Speichert die Listener, die an die Modelle angeschlossen wurden. */
  protected Vector<TimeSniffer> sniffers = new Vector<TimeSniffer>();


  /**
   * Erzeugt eine neue Instanz des Plugins.
   */
  public ModelTimeSnifferPlugin() {
  }


  /**
   * Startet die Ausfuehrung des Plugins.
   * @param appl Instanz der Xulu-Applikation
   */
  public void execute(XuluModellingPlatform appl) {
    this.appl = appl;
    started = true;
    if ( !appl.getModelControlManager().containsObjectListener(this) )
      appl.getModelControlManager().addObjectListener( this );
    //add Models which are currently loaded
    for(ModelControlFrame frame : appl.getModelControlManager().getAll())
    {
    	 XuluModel model = frame.getModel();
         TimeSniffer sniffer = new TimeSniffer(model);
         sniffers.add( sniffer );
         model.addModelListener( sniffer );
    }
  }


  /**
   * Wird aufgerufen, wenn sich der Inhalt des {@link ModelControlManager Xulu-Modellmanagers}
   * aendert. Wurde ein neues {@link XuluModel} hinzugefuegt, wird diesem
   * ein Listener hinzugefuegt, der die Modellierungszeit misst.
   * @param e Ereignis
   */
  public void performObjectEvent(ObjectEvent e) {
    if ( !isStarted() || !(e instanceof ModelControlManager.ChangeEvent) )
      return;
    ModelControlManager.ChangeEvent event = (ModelControlManager.ChangeEvent)e;
    // Listener nur hinzufuegen, wenn Modell neu hinzugefuegt
    // wurde und wenn es sich um ein XuluModell handelt
    if ( event.getOldValue()==null && event.getNewValue()!=null && event.getNewValue() instanceof ModelControlFrame ) {
      XuluModel model = ((ModelControlFrame)event.getNewValue()).getModel();
      TimeSniffer sniffer = new TimeSniffer(model);
      sniffers.add( sniffer );
      model.addModelListener( sniffer );
    }
  }

  /**
   * Stoppt die Ausfuehrung des Plugins.
   */
  public void stop() {
    started = false;
    appl.getModelControlManager().removeObjectListener(this);
    for (int i=0; i<sniffers.size(); i++)
      sniffers.elementAt(i).removeFromModel();
    sniffers.clear();
  }

  /**
   * Prueft, ob das Plugin gestartet ist.
   */
  public boolean isStarted() {
    return started;
  }

  /**
   * Prueft, ob die GUI zur Plugin-Konfiguration angezeigt wird.
   * @return immer {@code false}, da das Plugin keine GUI besitzt
   */
  public boolean isVisible() {
    return false;
  }

  /**
   * Macht nichts, da das Plugin keine GUI besitzt.
   */
  public void setVisible(boolean visible) {
  }


  private static class TimeSniffer implements ModelListener {
    private long      modelStartMillis     = -1; // Startzeit des Modells
    private long      modelStepStartMillis = -1; // Startzeit eines Modellschritts
    private long      modelIterStartMillis = -1; // Startzeit der Modell-Iteration
    private long      modelIterCount       = -1;
    private XuluModel model             = null;

    public TimeSniffer(XuluModel model) {
      this.model = model;
    }

    public void performModelEvent(ModelEvent e) {
      long eventMillis = System.currentTimeMillis();
      // Beginn des Modells
      if ( e instanceof ModelStartedEvent ) {
        modelStartMillis     = eventMillis;
        modelStepStartMillis = -1;
        modelIterStartMillis = -1;
      }
      // Beginn eines Zeitschritts
      if ( e instanceof ModelStepStartedEvent ) {
        modelStepStartMillis = eventMillis;
        modelIterStartMillis = -1;
      }
      // Beginn der eigentlichen Iterationen des Zeitschritts
      if ( e instanceof ModelIterationStepStartedEvent && ((ModelIterationStepStartedEvent)e).getIterNo() == 1 )
        modelIterStartMillis = eventMillis;
      if ( e instanceof ModelIterationStepFinishedEvent )
        modelIterCount = ((ModelIterationStepFinishedEvent)e).getIterNo();
      // Ende eines Zeitschritts
      if ( e instanceof ModelStepFinishedEvent ) {
        System.out.print(model.getName()+": step "+((ModelStepFinishedEvent)e).getStepNo()+" time = "+(eventMillis-modelStepStartMillis)/1000.0+"s");
        if ( modelIterStartMillis >= 0) {
          double secs = (eventMillis-modelIterStartMillis)/1000.0;
          System.out.println("   (Iteration cnt "+modelIterCount+", time = "+secs+"s avg = "+secs/modelIterCount+")");
        } else
          System.out.println();
      }
      // Ende des Modells
      if ( e instanceof ModelStoppedEvent )
        System.out.println(model.getName()+": model time = "+(eventMillis-modelStartMillis)/1000.0+"s");
    }

    public void performObjectEvent(ObjectEvent e) {
      if ( e instanceof ModelEvent )
        performModelEvent((ModelEvent)e);
    }

    public void removeFromModel() {
      model.removeModelListener(this);
    }

  }

}
