/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.data.misc;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.util.Vector;

/**
 * Dieser Datentyp erweitert die allgemeine {@linkplain CAArea CA-Region} um
 * folgende Eigenschaften:
 * <ul>
 *   <li>Einwohnerzahl (zum Zeitpunkt 0)</li>
 *   <li>Wachstumsrate der Einwohnerzahl pro Zeitschritt</li>
 *   <li>Bedarf pro Einwohner (in qm)</li>
 * </ul>
 * @see edu.bonn.xulu.plugin.model.ca.MultiCellularAutomaton
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CASettlementArea extends CAArea {
  /** Speichert die Einwohnerzahlen zum Start-Zeitpunkt. */
  protected double          inhabitants  = 0.0;
  /** Speichert die Wachstumsrate fuer die Einwohnerzahl. */
  protected double          growthRate   = 0.0;
  /** Speichert den Flaechenbedarf in qm pro Einwohner. */
  protected double          demandPerInh = 0.0;

  /** Umrechnungsfaktor, falls der Bedarf (pro Einwohner) nicht in qm angegeben
   *  ist (Standard: 1.0). */
  protected double          demandConvFact = 1.0;

  /**
   * Erzeugt eine neue Area.
   */
  public CASettlementArea() {
    this("",0,0,0.0,0.0,1);
  }

  /**
   * Erzeugt eine neue Area.
   * @param name Name der Area
   * @param x    X-Koodinate (Longitude)
   * @param y    Y-Koodinate (Latitude)
   * @param inhabitants Einwohner-Zahl zum Start-Zeitpunkt
   * @param growthRate  Wachstumsrate der Einwohner pro Zeitschritt
   * @param dpi  qm-Bedarf pro Einwohner
   */
  public CASettlementArea(String name, double x, double y, double inhabitants, double growthRate, double dpi) {
    super(name,x,y);
    this.inhabitants    = inhabitants;
    this.growthRate     = growthRate;
    this.demandPerInh   = dpi;
    this.demandConvFact = 1.0;
  }

  /**
   * Liefert den Bedarf pro Einwohner in qm.
   */
  public double getDemandPerInhabitant() {
    return this.demandPerInh;
  }

  /**
   * Setzt den Bedarf pro Einwohner.
   * @param dpi Bedarf in qm
   */
  public void setDemandPerInhabitant(double dpi) {
    this.demandPerInh = dpi;
  }

  /**
   * Liefert den Umrechnungsfaktor, um den Bedarf pro Einwohner in qm
   * zu transformieren.
   */
  public double getDemandToQMFactor() {
    return this.demandConvFact;
  }

  /**
   * Setzt einen Umrechnungsfaktor, falls der Bedarf pro Einwohner nicht in
   * qm angegeben ist. Mit diesem Wert wird der {@link #demandPerInh} multipliziert
   * um festzustellen, ob der Bedarf gedeckt ist.
   * @param dcf Faktor
   * @see #isDemandSatisfied(int)
   */
  public void setDemandToQMFactor(double dcf) {
    this.demandConvFact = dcf;
  }

  /**
   * Liefert die Wachstumsrate der Bevoelkerung (z.B. <code>0.3</code> bei
   * 30%igem Wachstum).
   */
  public double getGrowthRate() {
    return this.growthRate;
  }

  /**
   * Setzt die Wachstumsrate der Bevoelkerung (z.B. <code>0.3</code> bei
   * 30%igem Wachstum).
   * @param rate Wachstumsrate
   */
  public void setGrowthRate(double rate) {
    this.growthRate = rate;
  }

  /**
   * Liefert die Einwohnerzahl fuer einen bestimmten Zeitschritt. Diese
   * berechnet sich ueber folgende Formel:<br>
   * <code>inhabitants(i) = inhabitants(0) * growthRate^step</code>
   * @param step Schrittnummer (beginnend bei 0)
   * @return 0, falls <code>step < 0</code>
   */
  public double getInhabitants(int step) {
    if ( step < 0 )
      return 0;
    return this.inhabitants * Math.pow(1+growthRate,step);
  }

  /**
   * Liefert die Einwohnerzahl zum Start-Zeitpunkt (0).
   */
  public double getInhabitants() {
    return this.inhabitants;
  }

  /**
   * Setzt die Einwohnerzahl zum Start-Zeitpunkt (0).
   */
  public void setInhabitants(double inhabitants) {
    this.inhabitants = inhabitants;
  }

  /**
   * Prueft, ob die aktuell belegte Flaeche den Bedarf deckt.
   * @param step Schrittnummer (beginnend bei 0!)
   */
  public boolean isDemandSatisfied(int step) {
    return getCurrentCover() >= getInhabitants(step) * getDemandPerInhabitant() * getDemandToQMFactor();
  }
}
