/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.gui;

import java.util.Vector;
import java.awt.Component;
import javax.swing.JTable;
import javax.swing.JLabel;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableCellRenderer;

import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.AccessViolationException;

import edu.bonn.xulu.gui.XuluGUIMessages;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt eine GUI-Komponente dar, in der eine
 * 2-dimensionale (m,n)-{@link MatrixProperty} angezeigt werden kann.
 * Die einzelnen Matrix-Werte werden tabellarisch in m Spalten und n Zeilen
 * aufgelistet.<br>
 * Darstellbar (und editierbar) sind Elemente, die einen Basis-Datentyps
 * (z.B. Integer) repraesentieren.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DisplayContainer_MatrixProperty extends DisplayContainer_Table {
  /**
   * Erzeugt eine neue Darstellungskomponente.
   */
  public DisplayContainer_MatrixProperty() {
    this(null);
  }

  /**
   * Erzeugt eine neue Darstellungskomponente und zeigt sofort ein
   * Objekt an.
   * @param data anzuzeigendes Objekt
   */
  public DisplayContainer_MatrixProperty(Object data) {
    super(data);
    this.table.getTableHeader().setReorderingAllowed(false);
    this.table.setCellSelectionEnabled(true);
    this.table.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    // Die erste Spalte, in der die Zeilennummern ausgegeben werden, soll
    // wie der Header aussehen
    this.table.setDefaultRenderer(Object.class, new DefaultTableCellRenderer() {
      public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
        if ( column == 0 )
          return table.getTableHeader().getDefaultRenderer().getTableCellRendererComponent(table,value,false,false,row,column);
        JLabel comp = (JLabel)super.getTableCellRendererComponent(table,value,isSelected,hasFocus,row,column);
        comp.setHorizontalAlignment(JLabel.CENTER);
        return comp;
      }
    });
  }

  /**
   * Erzeugt das TableModel fuer die Tabelle
   * @param o darzustellendes <code>MatrixProperty</code>-Objekt (kann <code>null</code> sein!)
   */
  protected MatrixPropertyTableModel createTableModel(Object o) {
    return new MatrixPropertyTableModel((MatrixProperty)o);
  }

  /**
   * Liefert die aktuell dargestellte {@link MatrixProperty}-Instanz.
   * @return <code>null</code> falls aktuell kein Objekt dargestellt wird
   */
  public MatrixProperty getObject() {
    return (MatrixProperty)super.getObject();
  }

  /**
   * Prueft, ob ein Objekt darstellbar ist.
   * @return <code>true</code> gdw. das angegebene Objekt eine Instanz von
   *         {@link MatrixProperty} ist, dessen Dimension 2 ist.
   */
  public boolean canDisplay(Object data) {
    return data instanceof MatrixProperty &&
           ((MatrixProperty)data).getDimension() == 2;
  }

  /**
   * Liefert alle in der Tabelle selektierten Matrixelemente.
   */
  public Object[] getSelectedElements() {
    Vector selElem = new Vector();
    int[] selRowIdx = table.getSelectedRows();
    int[] selColIdx = table.getSelectedRows();
    for (int c=0; c<selColIdx.length; c++)
      if (selColIdx[c] > 0 ) // erste Spalte sind die Zeilennummern! Nicht selektierbar!
        for (int r=0; r<selRowIdx.length; r++)
          if ( table.isCellSelected(r,c) )
            selElem.add( ((MatrixPropertyTableModel)tableModel).getMatrixElement(selRowIdx[r],selColIdx[c]) );
    return selElem.toArray(new Object[0]);
  }

  ///////////////////////////////////////////////////////////////////
  // TableModel fuer die MatrixProperty-Tabelle
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt ein TableModel fuer MatrixProperty-Instanzen dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class MatrixPropertyTableModel extends DisplayContainerTableModel {
    private MatrixProperty matrixProp = null;

    /**
     * Erzeugt ein neues Daten-Modell fuer ein <code>MatrixProperty</code>-Objekt.
     * @param data matrixProp Matrix-Eigenschaft
     * @see schmitzm.data.property.MatrixProperty
     */
    public MatrixPropertyTableModel(MatrixProperty matrixProp) {
      setData(matrixProp);
    }

    /**
     * Erzeugt ein leeres Daten-Modell fuer ein <code>MatrixProperty</code>-Objekt.
     * @see schmitzm.data.property.MatrixProperty
     */
    public MatrixPropertyTableModel() {
      this(null);
    }

    /**
     * Setzt die Matrix, deren Elemente dargestellt werden sollen.
     * @param matrixProp Matrix-Eigenschaft (kann <code>null</code> sein!)
     */
    public void setData(Object matrixProp) {
      this.matrixProp = (MatrixProperty)matrixProp;
      fireTableDataChanged();
    }

    /**
     * Liefert ein Matrix-Element.
     * @param row Zeile des Datenmodells
     * @param col Spalte des Datenmodells
     */
    public Object getMatrixElement(int row, int col) {
      if ( col == 0 )
        return null;
      return this.matrixProp.getOneTimeReadAccess().getValue(col-1,row);
    }

    /**
     * Liefert einen Eintrag der Tabelle. Es wird das entsprechende Matrix-Element
     * geliefert
     * @param row Tabellenzeile
     * @param col Tabellenspalte
     */
    public Object getValueAt(int row, int col) {
      try {
        // in der ersten Spalte stehen die Zeilennummern
        if ( col == 0 )
          return row;
        // in den weiteren Spalten die Matrix-Spalten
        return getMatrixElement(row,col);
      } catch( AccessViolationException err ) {
        return "<no access!!>";
      }
    }

    /**
     * Liefert die Anzahl an Tabellen-Zeilen. Diese entspricht der
     * Matrix-Groesse in Dimension 1.
     */
    public int getRowCount() {
      return (matrixProp!=null) ? matrixProp.getSize(1) :0;
    }

    /**
     * Liefert die Anzahl an Tabellen-Spalten. Diese entspricht der
     * Matrix-Groesse in Dimension 0 plus 1 fuer die erste Spalte, die die
     * Zeilennummern enthaelt.
     */
    public int getColumnCount() {
      // Matrix-Spalten plus eine Spalte fuer die Zeilennummern
      return (matrixProp!=null) ? matrixProp.getSize(0)+1 :0;
    }

    /**
     * Liefert die Spaltennamen. Dieser entspricht der Spaltennummer minus 1,
     * da in der allerersten Spalte die Zeilennummern stehen.
     */
    public String getColumnName(int col) {
      // in der ersten Spalte stehen die Zeilennummern
      if ( col == 0 )
        return matrixProp.getSize(0)+"x"+matrixProp.getSize(1);
      // in den restlichen Spalten die Matrix-Spalten
      return String.valueOf(col-1);
    }

    /**
     * Prueft, ob eine Tabellen-Zelle editierbar ist. Dies ist nur fuer Zellen
     * der Fall bei denen es sich um einen Build-In-Type beinhaltet.
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see schmitzm.temp.BaseTypeUtil#isBaseType(Class)
     */
    public boolean isCellEditable(int row, int col) {
      // in der ersten Spalte stehen die Zeilennummern
      return col > 0 && BaseTypeUtil.isBaseType( getMatrixElement(row,col) );
    }

    /**
     * Setzt den jeweiligen Property-Wert neu.
     * @param obj neue Objektbeschreibung (<code>String</code>)
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see #isCellEditable(int,int)
     * @see schmitzm.temp.BaseTypeUtil#convertFromString(String,Class)
     */
    public void setValueAt(Object obj, int row, int col) {
      if ( isCellEditable(row,col) ) {
        PropertyWriteAccess w = null;
        try {
          w = matrixProp.getWriteAccess(this);
          w.setValue(BaseTypeUtil.convertFromString((String)obj, matrixProp.getType()),col-1,row);
        } catch (AccessViolationException err) {
          XuluGUIMessages.showAccessError(GUI_COMPONENT,err);
        } catch (NumberFormatException err) {
          XuluGUIMessages.showBaseTypeError(GUI_COMPONENT,err);
        } catch (Exception err) {
          XuluGUIMessages.showError(GUI_COMPONENT,err);
        } finally {
          // Schreibrecht muss unbedingt wieder freigegeben werden!!
          if ( w!=null ) w.release();
        }

      }
    }
  }

}
