/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.gui;

import java.awt.BorderLayout;
import javax.swing.table.AbstractTableModel;
import javax.swing.JScrollPane;
import javax.swing.JTable;

import schmitzm.swing.ObjectDisplayContainer;
import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import schmitzm.data.property.PropertySet;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.AccessViolationException;

import edu.bonn.xulu.gui.XuluGUIMessages;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt eine GUI-Komponente dar, in der eine
 * {@link Properties}-Instanz angezeigt werden kann. Die einzelnen Eigenschaften
 * werden tabellarisch in 3 Spalten aufgelistet:
 * <ol>
 * <li>Name der Property</li>
 * <li>Typ der Property</li>
 * <li>Wert der Property (sofern darstellbar)</li>
 * </ol>
 * Darstellbar (und aenderbar) sind skalare Propertys, die einen Basis-Datentyps
 * (z.B. Integer) repraesentieren.<br>
 * Dieser Container kann auch fuer die darstellung einer einzelnen {@link Property}
 * genutzt werden. Die Methode {@link #setObject(Object)} legt hierzu ein
 * {@link PropertySet} an, in dem nur die eine Property enthalten ist.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DisplayContainer_Properties extends DisplayContainer_Table {
  /**
   * Erzeugt eine neue Darstellungskomponente.
   */
  public DisplayContainer_Properties() {
    this(null);
  }

  /**
   * Erzeugt eine neue Darstellungskomponente und zeigt sofort ein
   * Objekt an.
   * @param data anzuzeigendes Objekt
   */
  public DisplayContainer_Properties(Object data) {
    super(data);
  }

  /**
   * Erzeugt das TableModel fuer die Tabelle
   * @param o darzustellendes <code>Properties</code>-Objekt (kann <code>null</code> sein!)
   */
  protected PropertiesTableModel createTableModel(Object o) {
    return new PropertiesTableModel((Properties)o);
  }

  /**
   * Liefert die aktuell dargestellte {@link Properties}-Instanz.
   * @return <code>null</code> falls aktuell kein Objekt dargestellt wird
   */
  public Properties getObject() {
    return (Properties)super.getObject();
  }

  /**
   * Setzt das darzustellende Objekt. Wird eine einzelne {@link Property}
   * uebergeben, wird zur Darstellung ein {@link PropertySet} angelegt, in dem
   * sich nur die eine Property befindet.
   * @exception UnsupportedOperationException falls das angegebene Objekt
   *            nicht dargestellt werden kann.
   * @see #canDisplay(Object)
   */
  public void setObject(Object data) {
    if ( data instanceof Property )
      data = createPropertySet((Property)data);
    super.setObject(data);
  }

  /**
   * Erzeugt ein PropertySet fuer eine einzelne Property.
   */
  private static PropertySet createPropertySet(Property p) {
    PropertySet ps = new PropertySet();
    ps.addProperty(p);
    return ps;
  }

  /**
   * Prueft, ob ein Objekt darstellbar ist.
   * @return <code>true</code> gdw. das angegebene Objekt eine Instanz von
   *         {@link Properties} oder {@link Property} ist.
   */
  public boolean canDisplay(Object data) {
    return data instanceof Properties || data instanceof Property;
  }

  /**
   * Liefert alle in der Tabelle selektierten Propertys.
   */
  public Property[] getSelectedElements() {
    int[] selIdx = table.getSelectedRows();
    Property[] selProp = new Property[selIdx.length];
    for (int i=0; i<selProp.length; i++)
      selProp[i] = ((PropertiesTableModel)tableModel).getProperty( selIdx[i] );
    return selProp;
  }

  ///////////////////////////////////////////////////////////////////
  // TableModel fuer die Properties-Tabelle
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Klasse stellt ein TableModel fuer Properties (z.B. eines
   * {@link schmitzm.data.XuluObject}) dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  private class PropertiesTableModel extends DisplayContainerTableModel {
    protected final String[] COLUMN_NAME = {"Property","Type","Value"};
    private Properties propObj = null;
    private Property[] prop    = null;

    /**
     * Erzeugt ein neues Daten-Modell fuer ein <code>Properties</code>-Objekt.
     * @param data propObj Objekt das aus Eigenschaften zusammengesetzt ist
     * @see schmitzm.data.property.Properties
     */
    public PropertiesTableModel(Properties propObj) {
      setData(propObj);
    }

    /**
     * Erzeugt ein leeres Daten-Modell fuer ein <code>Properties</code>-Objekt.
     * @see schmitzm.data.property.Properties
     */
    public PropertiesTableModel() {
      this((Properties)null);
    }

    /**
     * Setzt das Objekt, dessen Properties dargestellt werden sollen.
     * @param propObj Objekt das aus Eigenschaften zusammengesetzt ist (kann
     *                <code>null</code> sein!)
     */
    public void setData(Object propObj) {
      this.propObj = (Properties)propObj;
      this.prop    = (propObj != null) ? this.propObj.getProperties() : new Property[0];
      fireTableDataChanged();
    }

    /**
     * Liefert eine angezeigte Property.
     * @param row Zeile des Datenmodells
     */
    public Property getProperty(int row) {
      return this.prop[row];
    }

    /**
     * Liefert einen Eintrag der Tabelle. Fuer Spalte 0 wird die Property-Bezeichnung
     * zurueckgegeben, fuer Spalte 1 der Property-Datentyp und fuer Spalte 2
     * der aktuelle Wert der Property. Letzter wird nur angezeigt, wenn es
     * sich um eine {@link schmitzm.data.property.ScalarProperty} handelt.
     * @param row Tabellenzeile
     * @param col Tabellenspalte
     */
    public Object getValueAt(int row, int col) {
      switch( col ) {
        case 0: // Name der Property
                return prop[row].getName();
        case 1: // Typ der Property
                //  fuer Skalare: "Typ-Name"
                //  fuer Listen/Matrizen: "Klassenname [Typ-Name]"
                String typeStr = prop[row].getType().getSimpleName();
                if ( !(prop[row] instanceof ScalarProperty) )
                  typeStr = prop[row].getClass().getSimpleName().concat(" [").concat(typeStr).concat("]");
                return typeStr;
        case 2: // aktueller Wert der Property
                if ( prop[row] instanceof ScalarProperty &&
                     BaseTypeUtil.isBaseType(prop[row].getType())) {
                  try {
                    return ((ScalarProperty)prop[row]).getOneTimeReadAccess().getValue();
                  } catch( AccessViolationException err ) {
                    return "<no access!!>";
                  }
                }
                return "<can not be displayed>";
      }
      return null;
    }

    /**
     * Liefert die Anzahl an Tabellen-Zeilen. Diese entspricht der Anzahl an
     * Eigenschaften (Properties) des Objekts.
     */
    public int getRowCount() {
      return prop.length;
    }

    /**
     * Liefert die Anzahl an Tabellen-Spalten. Diese ist konstant 3 (je eine
     * Spalte fuer Property-Bezeichnung Property-Typ und Property-Wert).
     */
    public int getColumnCount() {
      return COLUMN_NAME.length;
    }

    /**
     * Liefert die Spaltennamen.
     * @see #COLUMN_NAME
     */
    public String getColumnName(int col) {
      return COLUMN_NAME[col];
    }

    /**
     * Prueft, ob eine Tabellen-Zelle editierbar ist. Dies ist nur fuer die
     * Spalte mit dem Property-Wert der Fall und nur wenn es sich um eine
     * {@link schmitzm.data.property.ScalarProperty} handelt, welche
     * einen Build-In-Type beinhaltet.
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see schmitzm.temp.BaseTypeUtil#isBaseType(Class)
     */
    public boolean isCellEditable(int row, int col) {
      return col == 2 &&
             prop[row] instanceof ScalarProperty &&
             BaseTypeUtil.isBaseType( prop[row].getType() );
    }

    /**
     * Setzt den jeweiligen Property-Wert neu.
     * @param obj neue Objektbeschreibung (<code>String</code>)
     * @param row Tabellen-Zeile
     * @param col Tabellen-Spalte
     * @see #isCellEditable(int,int)
     * @see schmitzm.temp.BaseTypeUtil#convertFromString(String,Class)
     */
    public void setValueAt(Object obj, int row, int col) {
      if ( isCellEditable(row,col) ) {
        PropertyWriteAccess w = null;
        try {
          w = ((ScalarProperty)prop[row]).getWriteAccess(this);
          w.setValue(BaseTypeUtil.convertFromString( (String) obj, prop[row].getType()));
        } catch (AccessViolationException err) {
          XuluGUIMessages.showAccessError(GUI_COMPONENT,err);
        } catch (NumberFormatException err) {
          XuluGUIMessages.showBaseTypeError(GUI_COMPONENT,err);
        } catch (Exception err) {
          XuluGUIMessages.showError(GUI_COMPONENT,err);
        } finally {
          // Schreibrecht muss unbedingt wieder freigegeben werden!!
          if ( w!=null ) w.release();
        }
      }
    }
  }
}
