/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io;

import java.io.File;

import java.net.URL;

import schmitzm.data.property.Property;
import schmitzm.data.property.ScalarProperty;

import edu.bonn.xulu.data.XuluObject;

import skrueger.geotools.StyledLayerStyle;
import skrueger.geotools.StyledLayerUtil;

// fuer Doku
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;

/**
 * Diese Klasse enthaelt statische Hilfs-Methoden fuer das Importieren und
 * Exportieren von {@linkplain XuluObject Xulu-Objekten}.
 * @see ImportFactory
 * @see ExportFactory
 * @author <a href="mailto:schmitzm@bonn.edu">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class IOUtil {

  /**
   * Hinterlegt einen Style fuer ein {@link XuluObject} in der Property
   * {@linkplain XuluObject#PROP_STYLE "Style"}.
   * Macht nichts, wenn das {@link XuluObject} keine entsprechende Property besitzt.
   * @param object ein {@link XuluObject}
   * @param style  Style
   */
  public static void setStyleToXuluObject(XuluObject object, Object style) {
    // Style-Property ermitteln
    Property styleProp = object.getProperty(object.PROP_STYLE);
    if ( styleProp != null && styleProp instanceof ScalarProperty )
      ((ScalarProperty)styleProp).getOneTimeWriteAccess().setValue(style);
  }

  /**
   * Liefert den eines {@link XuluObject} aus der Property
   * {@linkplain XuluObject#PROP_STYLE "Style"}.
   * @param object ein {@link XuluObject}
   * @return {@code null}, wenn das {@link XuluObject} keine entsprechende Property besitzt.
   */
  public static Object getStyleFromXuluObject(XuluObject object) {
    if ( object == null )
      return null;
    // Style-Property ermitteln
    Property styleProp = object.getProperty(XuluObject.PROP_STYLE);
    if ( styleProp != null && styleProp instanceof ScalarProperty )
      return ((ScalarProperty)styleProp).getOneTimeReadAccess().getValue();

    return null;
  }

  /**
   * Importiert einen {@linkplain StyledLayerStyle Raster-Style} fuer ein
   * {@link XuluObject} und hinterlegt diesen in der Property
   * {@linkplain XuluObject#PROP_STYLE "Style"}.
   * Macht nichts, wenn das {@link XuluObject} keine entsprechende Property besitzt.
   * @param object ein {@link XuluObject}
   * @param objectURL URL unter der das XuluObject hinterlegt ist (bestimmt
   *                  Pfad und Dateiname fuer die Style-Dateien)
   * @see StyledLayerUtil#loadStyledRasterStyle(URL)
   */
  public static void importRasterStyle(XuluObject object, URL objectURL) {
    Object style = StyledLayerUtil.loadStyledRasterStyle( objectURL );
    setStyleToXuluObject(object, style);
  }

  /**
   * Importiert einen {@linkplain StyledLayerStyle Feature-Style} fuer ein
   * {@link XuluObject} und hinterlegt diesen in der Property
   * {@linkplain XuluObject#PROP_STYLE "Style"}.
   * Macht nichts, wenn das {@link XuluObject} keine entsprechende Property besitzt.
   * @param object ein {@link XuluObject}
   * @param objectURL URL unter der das XuluObject hinterlegt ist (bestimmt
   *                  Pfad und Dateiname fuer die Style-Dateien)
   * @see StyledLayerUtil#loadStyledFeatureStyle(URL)
   */
  public static void importFeatureStyle(XuluObject object, URL objectURL) {
    Object style = StyledLayerUtil.loadStyledFeatureStyle( objectURL );
    setStyleToXuluObject(object, style);
  }

  /**
   * Exportiert den Style eines {@link XuluObject}, der in der Property
   * {@linkplain XuluObject#PROP_STYLE "Style"} hinterlegt ist. Macht nichts,
   * wenn das {@link XuluObject} keine entsprechende Property besitzt, oder
   * in der Property nichts hinterlegt ist.
   * @param object ein {@link XuluObject}
   * @param objectURL URL unter der das XuluObject hinterlegt ist (bestimmt
   *                  Pfad und Dateiname fuer die Style-Dateien)
   * @throws Exception wenn ein Fehler auftritt
   */
  public static void exportStyle(XuluObject object, URL objectURL) throws Exception {
    // Style-Property ermitteln
    Property styleProp = object.getProperty(object.PROP_STYLE);
    if ( styleProp != null && styleProp instanceof ScalarProperty ) {
      Object style = ((ScalarProperty)styleProp).getOneTimeWriteAccess().getValue();
      if ( style != null ) {
        if ( style instanceof StyledLayerStyle )
          StyledLayerUtil.saveStyledLayerStyle( (StyledLayerStyle)style, objectURL );
        else
          throw new UnsupportedOperationException("Export of style not supported: "+style.getClass().getSimpleName());
      }
    }

  }
}
