/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.array;

import java.io.File;

import schmitzm.io.IOUtil;
import schmitzm.data.WritableGridArray;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyReadAccess;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.appl.XuluRegistry;

import edu.bonn.xulu.plugin.data.grid.GridList;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link GridList} aus/in Dateien im ArcInfo-ASCII-Grid-Format.
 * Beim Import liefert die Factory eine Standard-Arrays basierendes Raster-Liste
 * ({@link WritableGridArray}).
 * Entsprechend koennen beim Export nur {@link GridList GridLists}
 * verarbeitet werden, die {@link WritableGridArray}-Instanzen in ihrer
 * Raster-Property besitzen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridListFactory_ArcInfoAsciiGrid extends AbstractFactory implements ImportFactory, ExportFactory {
  private static final WritableGridArrayFactory_ArcInfoAsciiGrid GRID_FAC = new WritableGridArrayFactory_ArcInfoAsciiGrid();

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link GridList GridList.class}
   */
  public Class getImportType() {
    return GridList.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File[].class}
   */
  public Class getImportSourceType() {
    return File[].class;
  }

  /**
   * Importiert Raster aus Dateien im ArcInfoAsciiGrid-Format
   * und fuegt diese in einer {@link GridList} zusammen.
   * @param input Eingabe-Quelle (muss ein {@link File File[]} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link GridList} erzeugt, in
   *              die das Objekt importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Dateien angegeben werden
   */
  public GridList importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    GridList gl = (GridList)instFac.newInstance(false);
    // Raster importieren und in GridList einfuegen
    for (int i=0; i<((Object[])input).length; i++) {
      gl.addGrid(
        GRID_FAC.importObject( ((File[])input)[i], reg )
      );
      // Style vom ersten Raster uebernehmen
      if ( edu.bonn.xulu.plugin.io.IOUtil.getStyleFromXuluObject( gl ) == null )
        edu.bonn.xulu.plugin.io.IOUtil.importRasterStyle( gl, ((File[])input)[i].toURI().toURL() );
    }

//    // Dateinamen in die Bezeichnung aufnehmen
//    sg.setDescription(sg.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return gl;
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um eine {@link GridList} handelt, die nur Instanzen
   * von {@link WritableGridArray} als Elemente enthaelt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    if ( obj==null || !(obj instanceof GridList) )
      return false;
    GridList list = (GridList)obj;
    for (int i=0; i<list.getGridCount(); i++)
      if ( !(list.getGrid(i) instanceof WritableGridArray) )
        return false;
    return true;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link GridList} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return GridList.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt. Dabei handelt es sich um eine <b>einzelne Datei</b>, da
   * die die letztendlichen Export-Dateien automatisch durchnummeriert werden!!
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link GridList} in Dateien des
   * ArcInfoAsciiGrid-Formats.<br>
   * <b>Beachte:</b><br>
   * Als Ausgabe-Objekt ist eine <b>einzelne</b> Datei anzugeben, kein
   * Array! Die Dateien werden ausgehend von dem angegebenen Dateinamen
   * durchnummeriert!!
   * @param object zu exportiertende Raster (muss eine {@link GridList} sein
   * @param output Export-Ziel (muss ein einzelner {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            einzelne Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Objekt
   *            nicht um eine {@link GridList} handelt
   */
  public void exportObject(Object object, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(object) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(object.getClass().getName()));

    // Dateinamen-Teile ermitteln
    String fileName = IOUtil.getBaseFilePath((File)output);
    String fileExt  = IOUtil.getFileExt((File)output);

    // Grids in der Liste einzeln exportieren
    GridList               gl         = (GridList)object;
    ListPropertyReadAccess listAccess = ((ListProperty) (gl.getProperty(GridList.PROP_GRIDS))).getReadAccess(this);
    for (int i=0; i<listAccess.getCount(); i++) {
      File outFile = new File( fileName.concat("_"+i).concat(fileExt) );
      GRID_FAC.exportObject( listAccess.getValue(i), outFile );
      // Style exportieren
      edu.bonn.xulu.plugin.io.IOUtil.exportStyle(gl, outFile.toURI().toURL() );
    }
    listAccess.release();

  }

}
