/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.array;

import java.io.File;

import schmitzm.data.WritableGridArray;

import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.appl.XuluRegistry;

import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import edu.bonn.xulu.plugin.io.IOUtil;

// nur fuer Doku

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link SingleGrid} aus/in das GeoTiff-Format.
 * Beim Import liefert die Factory ein auf Standard-Arrays basierendes Raster
 * ({@link WritableGridArray}).
 * Entsprechend koennen beim Export nur {@link SingleGrid SingleGrids}
 * verarbeitet werden, die eine {@link WritableGridArray}-Instanz in ihrere
 * Raster-Property besitzen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SingleGridFactory_GeoTiff extends AbstractFactory implements ImportFactory, ExportFactory {
  private static final WritableGridArrayFactory_GeoTiff GRID_FAC = new WritableGridArrayFactory_GeoTiff();

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link SingleGrid SingleGrid.class}
   */
  public Class getImportType() {
    return SingleGrid.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link SingleGrid} aus einer Datei im
   * GeoTiff-Format.
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @param reg   Instanz der Xulu-Registry, ueber die eine Factory ermittelt
   *              wird, die eine Standard-Instanz von {@link SingleGrid} erzeugt, in
   *              die das Raster importiert wird
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   */
  public SingleGrid importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    // Aus Registry die passende Factory fuer SingleGrid suchen
    InstantiationFactory instFac = AbstractFactory.getInstantiationFactoryFromRegistry(reg,getImportType(),true);
    // Raster importieren und in ein SingleGrid einfuegen
    SingleGrid sg = (SingleGrid)instFac.newInstance(false);
    sg.setGrid( GRID_FAC.importObject( (File)input,reg) );
    // Style importieren und falls vorhanden in Property einfuegen
    IOUtil.importRasterStyle( sg, ((File)input).toURI().toURL() );
    // Dateinamen in die Bezeichnung aufnehmen
    sg.setDescription(sg.getDescription().concat(" [").concat(((File)input).getName()).concat("]"));
    return sg;
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link SingleGrid} handelt, das ein
   * {@link WritableGridArray} in seiner Raster-Eigenschaft besitzt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null &&
           obj instanceof SingleGrid &&
           ((SingleGrid)obj).getGrid() instanceof WritableGridArray;
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link SingleGrid} (oder eine Unterklasse)
   * handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return SingleGrid.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link SingleGrid} in eine GeoTiff-Datei.
   * @param grid zu exportiertendes Raster (muss ein {@link SingleGrid} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Raster
   *            nicht um ein {@link SingleGrid} handelt
   */
  public void exportObject(Object grid, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(grid) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(grid.getClass().getName()));
    SingleGrid sg = (SingleGrid)grid;
    GRID_FAC.exportObject( sg.getGrid(),(File)output );
    IOUtil.exportStyle( sg, ((File)output).toURI().toURL() );

  }

}
