/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.array;

import java.io.File;

import org.geotools.coverage.grid.GridCoverageFactory;

import schmitzm.data.WritableGridArray;
import schmitzm.data.WritableGridArray.Float;
import schmitzm.data.WritableGridArray.Double;
import schmitzm.data.WritableGridArray.Integer;
import schmitzm.geotools.GTUtil;
import schmitzm.geotools.io.GeoImportUtil;
import schmitzm.geotools.io.GeoExportUtil;
import schmitzm.geotools.grid.WritableGridCoverage;

import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.ImportFactory;
import edu.bonn.xulu.io.ExportFactory;
import edu.bonn.xulu.appl.XuluRegistry;

// nur fuer Doku
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import org.geotools.coverage.grid.GridCoverage2D;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

/**
 * Diese Factory importiert und exportiert Instanzen des Datentyps
 * {@link WritableGridArray} aus/in das GeoTiff-Format.<br>
 * <b>Bemerke:</b><br>
 * Die erzeugten Instanzen stellen <b>keine</b> {@linkplain XuluObject Xulu-Objekte} dar,
 * koenn also nicht direkt im Xulu-Datenpool gespeichert werden. Hierzu ist der Datentyp
 * {@link SingleGrid} (bzw. eine entsprechende Factory) zu verwenden!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class WritableGridArrayFactory_GeoTiff extends AbstractFactory implements ImportFactory, ExportFactory {
  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ImportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert den Datentyp, den die Factory erzeugt.
   * @return immer {@link WritableGridArray WritableGridArray.class}
   */
  public Class getImportType() {
    return WritableGridArray.class;
  }

  /**
   * Liefert den Datentyp, den die Factory als Import-Quelle benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getImportSourceType() {
    return File.class;
  }

  /**
   * Importiert eine Instanz von {@link WritableGridArray} aus einer
   * Datei.
   * @param input Eingabe-Quelle (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Eingabe-Quelle keine
   *            Datei angegeben wird
   * @param reg wird nicht beruecksichtigt, da immer Instanzen von
   *            {@link WritableGridArray} verwendet werden!
   * @see GeoImportUtil#readGridFromGeoTiff(File,CoordinateReferenceSystem)
   */
  public WritableGridArray importObject(Object input, XuluRegistry reg) throws Exception {
    AbstractFactory.checkImportSourceObject(this,input,getImportSourceType(),true);
    GridCoverage2D grid = GeoImportUtil.readGridFromGeoTiff((File)input);
    return WritableGridArray.create(grid.getRenderedImage(),grid.getEnvelope2D(),grid.getCoordinateReferenceSystem());
  }

  ////////////////////////////////////////////////////////////////////////
  ///////////////     Implementierung von ExportFactory     //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob ein Objekt exportiert werden kann. Dies ist der
   * Fall, wenn es sich um ein {@link WritableGridArray.Integer},
   * {@link WritableGridArray.Float} oder {@link WritableGridArray.Double}
   * handelt.
   * @param obj zu pruefendes Objekt
   */
  public boolean isExportable(Object obj) {
    return obj!=null &&
         ( obj instanceof WritableGridArray.Integer ||
           obj instanceof WritableGridArray.Float ||
           obj instanceof WritableGridArray.Double );
  }

  /**
   * Prueft, ob ein Objekttyp exportiert werden kann. Dies ist der
   * Fall, wenn es sich um {@link WritableGridArray.Integer},
   * {@link WritableGridArray.Float}, {@link WritableGridArray.Double} oder
   * eine Unterklasse handelt.
   * @param c zu pruefender Objekttyp
   */
  public boolean isExportable(Class c) {
    return WritableGridArray.Integer.class.isAssignableFrom(c) ||
           WritableGridArray.Float.class.isAssignableFrom(c) ||
           WritableGridArray.Double.class.isAssignableFrom(c);
  }

  /**
   * Liefert den Objekt-Typ, den die Factory als Ziel zum Exportieren
   * benoetigt.
   * @return immer {@link File File.class}
   */
  public Class getExportDestinationType() {
    return File.class;
  }

  /**
   * Exportiert eine Instanz von {@link WritableGridArray} in eine Datei.
   * @param grid zu exportiertendes Raster (muss ein {@link WritableGridArray.Float} sein
   * @param output Export-Ziel (muss ein {@link File} sein!)
   * @exception UnsupportedOperationException falls als Export-Ziel keine
   *            Datei angegeben wird
   * @exception IllegalArgumentException falls es sich bei dem angegeben Raster
   *            nicht um ein {@link WritableGridArray} handelt
   * @see GeoExportUtil#writeGridToGeoTiff(GridCoverage2D,File)
   */
  public void exportObject(Object grid, Object output) throws Exception {
    AbstractFactory.checkExportDestinationObject(this,output,getExportDestinationType(),true);
    if ( !isExportable(grid) )
      throw new IllegalArgumentException(getClass().getName().concat(" can not export instances of ").concat(grid.getClass().getName()));
    WritableGridArray wga = (WritableGridArray)grid;
    float[][] floatData = new float[wga.getHeight()][wga.getWidth()];
    if ( wga instanceof WritableGridArray.Float )
      ((WritableGridArray.Float)wga).getData(floatData);
    else if ( wga instanceof WritableGridArray.Integer )
      ((WritableGridArray.Integer)wga).getData(floatData);
    else  if ( wga instanceof WritableGridArray.Double )
      ((WritableGridArray.Double)wga).getData(floatData);
    GeoExportUtil.writeGridToGeoTiff(
      new GridCoverageFactory().create("",floatData,GTUtil.createEnvelope2D(wga.getEnvelope(),wga.getCoordinateReferenceSystem())),
      (File)output
    );
  }
}
