/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.io.grid.awt;

import schmitzm.data.WritableGrid;
import edu.bonn.xulu.io.AbstractFactory;
import edu.bonn.xulu.io.InstantiationFactory;
import edu.bonn.xulu.io.FactoryCanceledException;
import edu.bonn.xulu.data.XuluObject;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import edu.bonn.xulu.plugin.data.grid.SingleGrid;
import edu.bonn.xulu.plugin.io.grid.WritableGridFactory;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.InputOption;
import schmitzm.swing.ManualInputOption;

/**
 * Diese Factory erzeugt Standard-Instanzen von {@link MultiGrid} mit
 * durch die Factory {@link WritableGridRasterFactory} erzeugten (auf Standard-AWT
 * basierenden) {@link WritableGrid}-Instanzen als Inhalt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MultiGridFactory extends AbstractFactory implements InstantiationFactory {
  /**
   * Liefert den Datentyp, der von der Factory erzeugt wird.
   * @return immer {@link MultiGrid MultiGrid.class}
   */
  public Class getInstanceType() {
    return MultiGrid.class;
  }

  /**
   * Erzeugt eine neue Instanz von {@link MultiGrid}. Hierzu wird ueber einen
   * Anwenderdialog abgefragt, wie viele Raster die Liste enthalten soll und
   * die Factory {@link WritableGridRasterFactory} herangezogen, um die Raster zu
   * erzeugen.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird die Vorlage ohne Dialog kopiert)
   * @param sample Vorlage
   * @exception IllegalArgumentException falls es sich bei der Vorlage nicht
   *            um eine {@link MultiGrid}, {@link SingleGrid} oder {@link WritableGrid}
   *            handelt
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
   *            "Abbrechen" verlassen wird
   */
  public MultiGrid newInstance(boolean dialog, Object sample) throws Exception {
//    if ( sample == null )
//      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from null") );
// wird nun in createFromSample(..) geprueft
//    if ( !getInstanceType().isInstance(sample) )
//      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getName()) );

    return createFromSample(dialog,sample);
  }

  /**
   * Erzeugt eine neue Instanz von {@link MultiGrid}. Hierzu wird ueber einen
   * Anwenderdialog abgefragt wie viele Raster die Liste enthalten soll und
   * die Factory {@link WritableGridRasterFactory} herangezogen, um die Raster zu
   * erzeugen.
   * @param dialog gibt an, ob die Factory einen Dialog mit dem Anwender
   *               fuehren darf, um die Instanz zu initialisieren (bei
   *               <code>false</code> wird eine Standard-Instanz erzeugt)
   * @exception FactoryCanceledException falls der Anwender-Dialog ueber
   *            "Abbrechen" verlassen wird
   */
  public MultiGrid newInstance(boolean dialog) throws Exception {
    return createFromSample(dialog,null);
  }

  /**
   * Erzeugt ein neues {@link MultiGrid} aus einer Vorlage. Trotzdem wird ein Dialog
   * angezeigt, in dem die Vorlagen-Daten nacheditiert werden koennen.
   * Falls die Vorlage <code>null</code> ist, werden Default-Angaben vorgeblendet.
   * @param sample Vorlage ({@link WritableGrid}, {@link SingleGrid} oder {@link MultiGrid})
   * @param dialog bestimmt, ob die Vorlagen-Daten sofort uebernommen werden, oder
   *               ueber einen Anwender-Dialog abgeaendert werden koennen
   */
  private MultiGrid createFromSample(boolean dialog, Object sample) throws Exception {
    if ( sample != null && !(sample instanceof WritableGrid) &&
                           !(sample instanceof SingleGrid) &&
                           !(sample instanceof MultiGrid) )
      throw new IllegalArgumentException(this.getClass().getSimpleName().concat(" can not create an instance of ").concat(getInstanceType().getSimpleName()).concat(" from ").concat(sample.getClass().getSimpleName()) );

    // Anzahl an Grids aus dem Sample uebernehmen
    int count = 0;
    if ( sample != null && dialog && sample instanceof MultiGrid )
      count = ((MultiGrid)sample).getGridCount();

    // Anzahl zu erzeugender Grids abfragen
    if (dialog) {
      Object[] inpValue = MultipleOptionPane.showMultipleInputDialog(
                               null,
                               this.getInstanceType().getSimpleName().concat(" erzeugen"),
                               new InputOption[] { new ManualInputOption.Integer("Anzahl an Grids in der Liste",true,count) }
      );
      if ( inpValue == null )
        throw new FactoryCanceledException(this);
      count = (Integer)inpValue[0];
    }

    // nacheinander die Grids erzeugen
    WritableGridFactory fac = getWritableGridFactory();
    if ( !WritableGrid.class.isAssignableFrom(fac.getInstanceType()) )
      throw new UnsupportedOperationException(getClass().getSimpleName().concat(".getWritableGridFactory() must provide a factory that creates a WritableGrid!"));

    // Vorlage erstellen
    WritableGrid sampleGrid = null;
    if ( sample instanceof MultiGrid )
      sampleGrid = ((MultiGrid)sample).getTemplateGrid();
    else if ( sample instanceof SingleGrid )
      sampleGrid = ((SingleGrid)sample).getGrid();
    else if ( sample instanceof WritableGrid )
      sampleGrid = (WritableGrid)sample;
    else if ( dialog )
      sampleGrid = fac.newInstance(dialog);

    // MultiGrid erstellen
    MultiGrid multiGrid = new MultiGrid(sampleGrid,fac);
    for (int i=0; i<count; i++)
      multiGrid.addGrid();
    // Style von der Vorlage uebernehmen
    if ( sample instanceof XuluObject )
      edu.bonn.xulu.plugin.io.IOUtil.setStyleToXuluObject(
        multiGrid,
        edu.bonn.xulu.plugin.io.IOUtil.getStyleFromXuluObject( (XuluObject)sample )
      );

    return multiGrid;
  }

  /**
   * Liefert eine {@link InstantiationFactory}, die Instanzen von {@link WritableGrid}
   * erzeugt. Mit diesen Instanzen wird die Liste gefuellt.<br>
   * Standardmaessig wird eine Instanz von {@link WritableGridRasterFactory}
   * zurueckgegeben. Sub-Klassen koennen diese Methode ueberschreiben, um
   * andere Arten von WritableGrid zu erzeugen.
   */
  protected WritableGridFactory getWritableGridFactory() {
    return new WritableGridRasterFactory();
  }

}
