/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca;

import java.util.Vector;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.TreeSet;
import java.awt.Point;

import schmitzm.lang.ComparableObject;

import schmitzm.data.WritableGrid;
import schmitzm.data.AbstractWritableGrid;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;

import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.plugin.data.misc.CAArea;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;

// nur fuer Doku
import edu.bonn.xulu.model.ModelContentManager;

/**
 * Dieses Modell stellt einen abstrakten zellularen Automaten (CA) dar,
 * der auf Basis einer {@linkplain #getPreferenceValue(Point,int,CAArea,int) Bewertungsfunktion}
 * die Expansion <b>eines</b> LUC-Typs simuliert. Es basiert auf folgenden
 * Ressourcen:
 * <ul>
 *   <li>eine Liste mit Start-Regionen ({@link CAArea}), die die Startkonfiguration
 *       des CA definiert (Siedlungen mit Einwohnerzahlen)</li>
 *   <li>ein LUC-Typ U (<code>int</code>), dessen Ausbreitung simuliert wird</li>
 *   <li>Anzahl T (<code>int</code>) der zu modellierender Schritte</li>
 *   <li>ein Raster ({@link WritableGrid}) mit der initialen LUC</li>
 *   <li>ein temp. Raster ({@link WritableGrid}), in dem vom Modell die Zuordnung
 *       der simulierten U-Zellen zu den Regionen hinterlegt werden kann</li>
 *   <li>ein optionales Ausgabe-Raster ({@link WritableGrid})</li>
 *   <li>ein Schwellwert MIN_U (<code>int</code>), ab welche Eignungswert eine Zelle
 *       fuer den LUC-Typ U geeignet ist</li>
 * </ul>
 * In jedem Zeitschritt wird iteriert bis Bedarf fuer alle Regionen gedeckt ist.
 * In jeder Iteration wird jede Zelle des Regionen-Rand (initial die
 * 8 Zellen um die jeweilige Start-Zelle) betrachtet und bewertet (durch die
 * Methode {@link #getPreferenceValue(Point,int,CAArea,int)}). Alle fuer U in Frage kommenden
 * Zellen werden (samt dem Bewertungswert) in einer (sortierten) Liste gespeichert.
 * Nachdem alle Zellen im Rand betrachtet wurden, werden (reihum) solange die
 * geeignetsten umgewandelt, bis der Bedarf gedeckt ist.<br>
 * Fuer jede umgewandelte Zelle werden die 8 umgebenden Zellen in den
 * Rand aufgenommen, sofern sie noch nicht durch das Modell umgewandelt
 * wurden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class MultiCellularAutomaton extends AbstractStepModel {
  /** Anzahl an Ressourcen, die durch diese Modell-Oberklasse definiert sind. */
  public static final int BASE_RESCOUNT = CAContentManager.BASE_RESCOUNT;

  // Zugriffsrechte auf Ressourcen
  private PropertyWriteAccess baseLUC_WA     = null;
  private PropertyWriteAccess outLUC_WA      = null;
  private PropertyWriteAccess areaMapping_WA = null;
  private PropertyWriteAccess areaConf_WA    = null;

  /** Ausgangs-Szenario.*/
  protected WritableGrid baseLUC = null;
  /** Ausgabe-Raster. Wenn die entsprechende optionale Ressource unbelegt bleibt,
   *  ist <code>outLUC = baseLUC</code>. */
  protected WritableGrid outLUC = null;
  /** Anzahl der modellierten Gebiete/Siedlungen. */
  protected int areaCount = 0;
  /** Die modellierten Gebiete/Siedlungen (@linkplain CAArea Areas). */
  protected ListPropertyWriteAccess areas = null;
  /** Speichert die Zuordnung, welche Zelle welcher Area zugeordnet wurde. */
  protected WritableGrid areaMapping = null;
  /** LUC-Typ fuer den die Wechsel modelliert werden.<br>
   *  <b>Bemerke:</b><br>
   *  Wird in einer zukuenftigen Version ersetzt durch ein Feld in {@link CAArea},
   *  so dass pro Gebiet/Siedlung unterschiedliche LUC-Typen expandiert werden
   *  koennen. */
  protected int luccType = 0;
  /** Schwellwert der {@link #getPreferenceValue(Point,int,CAArea,int) Praeferenzfunktion}
   *  ab dem eine Zelle fuer einen potentiellen LUC-Wechsel in Betracht gezogen wird.<br>
   *  <b>Bemerke:</b><br>
   *  Wird in einer zukuenftigen Version ersetzt durch ein Feld in {@link CAArea},
   *  so dass pro Gebiet/Siedlung unterschiedliche Schwellwerte verwendet werden
   *  koennen. */
  protected double luccThreshold = 0.0;

  /** Speichert die Groesse (in qm) eine Zelle des {@linkplain #baseLUC Untersuchungsgebiets}. */
  protected double cellSize      = 0.0;

  /** Definiert die Nachbarschaftsumgebung, ueber die der zellulare Automat
   *  expandiert. Standardmaessig sind dies die 8 umgebenende Zellen. Kann
   *  von Sub-Klassen neu definiert werden:
   *  <ul>
   *    <li><code>NEIGHBORHOOD_AREA[i,<b>0</b>]</code> definiert die relative <b>horizontale</b>
   *      Position der i-ten Nachbarschaftszelle (z.B. 1 = eine Zelle weiter rechts,
   *      -2 = zwei zellen weiter links als die betrachtete Zelle).</li>
   *    <li><code>NEIGHBORHOOD_AREA[i,<b>1</b>]</code> definiert die relative <b>vertikale</b>
   *      Position der i-ten Nachbarschaftszelle (z.B. 0 = gleiche Hoehe wie die betrachtete
   *      Zelle, 1 = eine Zelle weiter unten).</li>
   *  </ul>
   *  <b>Bemerke:</b><br>
   *  <code>NEIGHBORHOOD_AREA[i] = (0,0)</code> wuerde die betrachtete Zelle,
   *  als einen ihrer eigenen Nachbarn identifizieren! Dies ist nicht sinnvoll
   *  und wird vom zellularen Automaten abgefangen!
   */
  protected int[][] NEIGHBORHOOD_AREA = {
//     oben    links   rechts   unten
      { 0, 1},{-1, 0},{ 1, 0}, { 0,-1},
//     li-ob   re-ob   li-un    re-un
      {-1, 1},{ 1, 1},{-1,-1}, { 1,-1}
  };

  /** Speichert (aufsteigend soriert) die bewerteten Zellen fuer jede Area. */
  private TreeSet[] bestCells = null;

  private Vector<Point>[]  startCells = null;


  /**
   * Erzeugt einen neuen zellularen Automaten (CA).
   * @param cm ContentManager fuer den CA.
   */
  public MultiCellularAutomaton(CAContentManager cm) {
    super( cm );
    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
  }

  /**
   * Initialisiert das Modell.
   */
  public void performModelInit() {
    //########## Lese/Schreib-Rechte auf Ressourcen ##########
    // Schreibrecht auf Ausgangs-Szenario
    baseLUC_WA = ((ScalarProperty)contManager.getResource(CAContentManager.BASELUC).getData()).getWriteAccess(this);
    // Wenn kein Ausgabe-Raster angegeben wird, wird die Ausgabe in das
    // Ausgangs-Szenatio geschrieben
    if ( contManager.getResource(CAContentManager.OUTLUC).getData() == null )
      outLUC_WA = null;
    else
      outLUC_WA = ((ScalarProperty)contManager.getResource(CAContentManager.OUTLUC).getData()).getWriteAccess(this);
    // Schreibrecht auf Temp-Raster fuer Regionen-Zuordnung
    areaMapping_WA = ((ScalarProperty)contManager.getResource(CAContentManager.AREAMAPPING).getData()).getWriteAccess(this);
    // Leserecht auf Start-Konfiguration
    areaConf_WA = ((ListProperty)contManager.getResource(CAContentManager.AREACONF).getData()).getWriteAccess(this);


    //########## Ressourcen in lokalen Variablen verlinken ##########
    stepCount     = ((ScalarProperty)contManager.getResource(CAContentManager.STEPCOUNT).getData()).getOneTimeReadAccess().getValueAsInt();
    areas         = (ListPropertyWriteAccess)areaConf_WA;
    areaCount     = areas.getCount();
    luccType      = ((ScalarProperty)contManager.getResource(CAContentManager.LUCCTYPE).getData()).getOneTimeReadAccess().getValueAsInt();
    luccThreshold = ((ScalarProperty)contManager.getResource(CAContentManager.THRESHOLD).getData()).getOneTimeReadAccess().getValueAsDouble();
    baseLUC       = (WritableGrid)baseLUC_WA.getValue();
    outLUC        = (outLUC_WA == null ) ? baseLUC : (WritableGrid)outLUC_WA.getValue();
    areaMapping   = (WritableGrid)areaMapping_WA.getValue();

    // Zellengroesse in qm
    cellSize      = baseLUC.getCellWidth() * baseLUC.getCellWidth();


    // modellierte Gebiete initialisieren (Border, Cover)
    bestCells = new TreeSet[areaCount];
    for (int i=0; i<areaCount; i++) {
      CAArea area = (CAArea)areas.getValue(i);
      area.init();
      // Raster-Zelle des Start-Punkts in Rand aufnehmen
      area.getBorder().add( new Point(
        baseLUC.convertRealToRaster( area.getStartLocation().getX(), 0 ),
        baseLUC.convertRealToRaster( area.getStartLocation().getY(), 1 )
      ));

      // Fuer jedes Gebiet werden erst alle Zellen des Rands bewertet und
      // anschliessend der Reihe nach die besten umgewandelt.
      bestCells[i] = new TreeSet<ComparableObject<Double, Point>> ();
    }
    // Zuordnungsraster (und ggf. das Ausgaberaster) initialisieren.
    // Das Zuordnungsraster speichert, welche Zelle zu welcher Zone (Area) gehoert
    int minX = areaMapping.getMinX();
    int minY = areaMapping.getMinY();
    for (int y=0; y<areaMapping.getHeight(); y++)
      for (int x=0; x<areaMapping.getWidth(); x++) {
        // am Anfang gehoert noch keine Zelle zu irgendeinem Gebiet
        areaMapping.setRasterSample(-1,x+minX,y+minY);
        // wenn ein spezielles Ausgabe-Raster verwendet wird, wird das
        // Ursprungsszenarion dort rein kopiert
        if ( outLUC != baseLUC )
          outLUC.setRasterSample( baseLUC.getRasterSample(x+minX,y+minY), x+minX, y+minY);
      }
  }

  /**
   * Zerstoert das Modell. Alle gehaltenen Zugriffsrechte auf Ressourcen
   * werden wieder freigegeben.
   */
  public void performModelDispose() {
    releaseAccess(baseLUC_WA);
    releaseAccess(outLUC_WA);
    releaseAccess(areaMapping_WA);
    releaseAccess(areaConf_WA);
  }

  /**
   * Modelliert genau einen Zeitschritt.
   * @param stepNo Nummer des Zeitschritts
   */
  public void performModelStep(int stepNo) {
    if ( stepNo == 1 )
      checkStartPoints();

    // Iterieren, bis alle Gebiete die Anforderung erfuellen
    boolean allAreasSatisfied = false;
    for (int iterNo=1; !allAreasSatisfied; iterNo++) {
      allAreasSatisfied = true;
      // Iterations-Ereignis initiieren
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,stepNo,iterNo) );

      // Fuer alle Gebiete den Rand bewerten
      for (int areaNo=0; areaNo<areaCount; areaNo++) {
        bestCells[areaNo].clear();
        CAArea area = (CAArea)areas.getValue(areaNo);
        // Gebiet nur bewerten, wenn Bedarf noch nicht gedeckt
        // Bemerke: -1, da stepNo von 1 hochgezaehlt wird!
        if ( area.isDemandSatisfied(stepNo-1) || area.isDeadlock() )
          continue;
        // ein Gebiet erfuellt den Bedarf noch nicht
        allAreasSatisfied = false;

        // Rand des Gebiets bewerten
        for (int i = 0; i < area.getBorder().size(); i++) {
          checkBreakingCommands();
          Point p = area.getBorder().elementAt(i);
          // Zelle kommt nur in Frage, wenn sie noch nicht zu der Area gehoert
          // Diese Abfrage stellt eine Sicherheitsabfrage dar. Fall kann nur
          // vorkommen, wenn die Startbereiche nicht aus dem Rand geloescht werden
          // um den Bezug zum Start-Gebiet zu erhalten.
          if ( areaMapping.getRasterSampleAsByte(p.x,p.y) == areaNo )
            continue;

          // Bewerten
          double value = getPreferenceValue(p, luccType, area, areaNo);
          if ( value >= luccThreshold )
            bestCells[areaNo].add( new ComparableObject<Double, Point>(value, p) );
        }

        // Wenn keine geeigneten Zellen gefunden wurden, befindet sich der
        // CA fuer dieses Gebiet in einem Deadlock!
        if ( bestCells[areaNo].isEmpty() ) {
          area.setDeadlock(true);
          statusOut.println("Warning for area "+areaNo+": No suitable cells found in border! Automaton reached deadlock!");
        }
      }

      // Reihum alle Gebiete expandieren
      boolean allAreasExpanded = false;
      for (int expRound=1; !allAreasExpanded; expRound++) {
        allAreasExpanded = true;
        for (int areaNo=0; areaNo<areaCount; areaNo++) {
          checkBreakingCommands();
          CAArea  area = (CAArea)areas.getValue(areaNo);
          TreeSet best = bestCells[areaNo];
          // Gebiet nur expandieren, wenn noch Bedarf noch nicht gedeckt und
          // noch Randzellen uebrig
          // Bemerke: -1, da stepNo von 1 hochgezaehlt wird!
          if ( area.isDemandSatisfied(stepNo-1) || area.isDeadlock() || best.isEmpty() )
            continue;
          allAreasExpanded = false;

          // Beste Zelle ermitteln
          ComparableObject<Double,Point> o = (ComparableObject)best.last();
          Point p = o.getObject();
          // Zelle umwandeln
          int prevLUC = outLUC.getRasterSampleAsInt(p.x,p.y);
          outLUC.setRasterSample(luccType,p.x,p.y);
          areaMapping.setRasterSample(areaNo,p.x,p.y);
          area.incCurrentCover(cellSize);
          // Umgewandelte Zelle aus Rand und besten Zellen (aller Areas!)
          // entfernen
          for (int i=0; i<areaCount; i++) {
            bestCells[i].remove(o);
            // Start-Punkte bleiben immer im Rand (damit Bezug zum Start - z.B.
            // beim Brach-Fallen) nie verloren geht)
            if ( !startCells[i].contains(p) )
              ((CAArea)areas.getValue(i)).getBorder().remove(p);
          }
          // Umgebung der Zelle in Rand aufnehmen
          addNeighborhoodToBorder(area,p);
          // Sub-Klassen ermoeglichen, auf Aenderung zu reagieren
          cellChanged(p,area,luccType,prevLUC);
        }
      }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,stepNo,iterNo) );
//      showDifference(stepNo,iterNo);
    }
  }

  /**
   * Macht nichts. Wird aufgerufen, nachdem eine Zelle umgewandelt wurde.
   * Sub-Klassen koennen diese Methode ueberschreiben, um auf die Umwandlung
   * zu reagieren.
   * @param p       Raster-Koordinaten der umgewandelten Zelle
   * @param area    Gebiet dem die Zelle zugeordnet wurde
   * @param newLUC  LUC-Typ, der der Zelle zugeordnet wurde
   * @param prevLUC LUC-Typ, den die Zelle zuvor hatte
   */
  protected void cellChanged(Point p, CAArea area, int newLUC, int prevLUC) {

  }

  private void checkStartPoints() {
      // Standardmaessig beginnt der CA an dem uebergebenen Startpunkt (Lat/Lon).
      // Falls die Bewertung dieser Zelle nicht dem Schwellwert entspricht,
      // iteriert sich der CA jedoch "zu Tode".
      // Radius um die Startzelle vergroessern, bis eine akzeptable Startzelle
      // gefunden wird
      for (int a=0; a<areaCount; a++) {
        CAArea area = (CAArea)areas.getValue(a);
        Point p = area.getBorder().firstElement();
        double val = this.getPreferenceValue(p,luccType,area,a);
        if ( val < luccThreshold ) {
          area.getBorder().clear();
          for (int r=1; area.getBorder().isEmpty() && r<=10; r++) {
            for (int y=p.y-r; y<=p.y+r; y++)
              for (int x=p.x-r; x<=p.x+r; x++)
                if ( x >= baseLUC.getMinX() && x < baseLUC.getMinX()+baseLUC.getWidth() &&
                     y >= baseLUC.getMinY() && y < baseLUC.getMinY()+baseLUC.getHeight() ) {
                  Point testPoint = new Point(x,y);
                  val = this.getPreferenceValue(testPoint,luccType,area,a);
                  if ( val >= luccThreshold )
                    area.getBorder().add( testPoint );
                }
          }
          // Warnmeldung geben, falls keine geeigneten Start-Zelle(n)
          // gefunden wurden
          if ( area.getBorder().isEmpty() ) {
            area.setDeadlock(true);
            statusOut.println("Warning for area "+a+": No suitable starting cell(s) found!");
          }
        }
      }


      // Start-Zellen merken
      startCells = new Vector[areaCount];
      for (int i=0; i<startCells.length; i++)
        startCells[i] = (Vector) ((CAArea)areas.getValue(i)).getBorder().clone();
  }


  private void addNeighborhoodToBorder(CAArea area, Point p) {
    // Umgebung der Zelle in Rand aufnehmen
    for (int b = 0; b < NEIGHBORHOOD_AREA.length; b++) {
      Point bp = new Point(p.x + NEIGHBORHOOD_AREA[b][0], p.y + NEIGHBORHOOD_AREA[b][1]);

      if ( !bp.equals(p) && // Zelle entspricht nicht der betrachtaten Zelle
           areaMapping.getRasterSampleAsByte(bp.x, bp.y) == -1 && // Zelle noch unbelegt
           !area.getBorder().contains(bp) ) // Zelle noch nicht im Rand

        area.getBorder().add(bp);
    }
  }


//  private void showDifference(int step, int iter) {
//    statusOut.print("After Iteration "+step+"/"+iter+":  ");
//    for (int a=0; a<areaCount; a++) {
//      CAArea area = (CAArea)areas.getValue(a);
//      double diff = area.getInhabitants(step-1)*area.getDemandPerInhabitant() - area.getCurrentCover();
//      statusOut.print("("+a+") "+diff+(area.isDeadlock() ? "*":"")+"\t");
//    }
//    statusOut.println();
//  }

  /**
   * Bewertet eine Zelle des Untersuchungsgebiets hinsichtlich eines bestimmten
   * LUC-Typs. Je groesser der Wert, desto geeigneter die Zelle fuer den
   * LUC-Typ <code>luc</code>.
   * @param p      Zelle des Untersuchungsgebiets
   * @param luc    LUC-Typ hinsichtlich dessen die Bewertung durchgefuehrt wird
   * @param area   CA-Gebiet fuer das die Bewertung durchgefuehrt wird
   * @param areaNo Nummer des Gebiets
   */
  public abstract double getPreferenceValue(Point p, int luc, CAArea area, int areaNo);

  /**
   * Diese Klasse definiert einen {@link ModelContentManager} fuer einen
   * {@linkplain MultiCellularAutomaton Multi-Zellularen-Automat} mit folgenden
   * Ressourcen:
   * <ol>
   *   <li><b>StepCount (<code>int</code>)</b><br>
   *       Spezifiziert die Anzahl zu modellierender Schritte.</li>
   *   <li><b>LUC-Typ (<code>int</code>)</b><br>
   *       Spezifiziert den LUC-Typ, dessen Wechsel durch den CA modelliert wird.<br>
   *       <b>Bemerke:</b><br>
   *       Wird in einer zukuenftigen Version ersetzt durch ein Feld in {@link CAArea},
   *       so dass pro Gebiet/Siedlung unterschiedliche LUC-Typen expandiert werden
   *       koennen.</li>
   *   <li><b>Schwellwert (<code>double</code>)</b><br>
   *       Spezifiziert den der {@link #getPreferenceValue(Point,int,CAArea,int) Praeferenzfunktion},
   *       ab dem eine Zelle fuer einen potentiellen LUC-Wechsel in Betracht gezogen wird.<br>
   *       <b>Bemerke:</b><br>
   *       Wird in einer zukuenftigen Version ersetzt durch ein Feld in {@link CAArea},
   *       so dass pro Gebiet/Siedlung unterschiedliche Schwellwerte verwendet werden
   *       koennen.</li>
   *  <li><b>Base-Scenario ({@link WritableGrid})</b><br>
   *      Ausgangs-Szenarion (LUC) des Untersuchungsgebiets.</li>
   *  <li><b>Startpunkte (<code>List[{@link CAArea}]</code>)</b><br>
   *      Startpunkte (Gebiete/Siedlungen) des CA definiert.</li>
   *  <li><b>Ausgabe-Raster ({@link WritableGrid})</b><br>
   *      Nimmt die modellierte LUC auf (optional).</li>
   *  <li><b>Area-Mapping ({@link WritableGrid})</b><br>
   *      Wird dazu genutzt, die Zuordnung der modellierten Zellen zu den
   *      Startpunkten (Gebiete/Siedlungen) zu speichern.</li>
   * </ol>
   * Fuer Sub-Klassen von {@link MultiCellularAutomaton} kann dieser Content-Manager
   * um weitere Ressourcen erweitert werden.
   */
  public abstract static class CAContentManager extends AbstractModelContentManager {
    /** Definiert die Anzahl der durch diesen Content-Manager definierten
     *  Ressourcen */
    public static final int    BASE_RESCOUNT         = 7;

    /** ID der <code>int</code>-Ressource fuer die Anzahl zu modellierender Modell-Schritte. */
    public static final int    STEPCOUNT             = 0;
    /** Beschreibung fuer die {@link #STEPCOUNT}-Ressource. */
    public static final String STEPCOUNT_DESC        = "Time steps (int)";

    /** ID der <code>int</code>-Ressource fuer den LUC-Typ, dessen Wechsel durch den CA
     *  simuliert wird. */
    public static final int    LUCCTYPE              = 1;
    /** Beschreibung fuer die {@link #LUCCTYPE}-Ressource. */
    public static final String LUCCTYPE_DESC         = "Modelled LUC-Type (int)";

    /** ID der <code>double</code>-Ressource fuer den Schwellwert der
     *  {@link MultiCellularAutomaton#getPreferenceValue(Point,int,CAArea,int) Praeferenzfunktion},
     *  ab dem eine Zelle fuer einen potentiellen LUC-Wechsel in Betracht gezogen wird. */
    public static final int    THRESHOLD             = 2;
    /** Beschreibung fuer die {@link #THRESHOLD}-Ressource. */
    public static final String THRESHOLD_DESC        = "Preference threshold for LUCC (double)";

    /** ID der {@link WritableGrid}-Ressource fuer das Ausgangsszenario des Untersuchungsgebiets. */
    public static final int    BASELUC               = 3;
    /** Beschreibung fuer die {@link #BASELUC}-Ressource. */
    public static final String BASELUC_DESC          = "Base Scenario (Raster)";

    /** ID der <code>List[{@link CAArea}]</code>-Ressource, die die Startpunkte
     *  (Gebiete/Siedlungen) des CA definiert. */
    public static final int    AREACONF              = 4;
    /** Beschreibung fuer die {@link #AREACONF}-Ressource. */
    public static final String AREACONF_DESC         = "Area configurations (CAArea-List)";

    /** ID der (optionalen) {@link WritableGrid}-Ressource fuer das Raster, in dem
     *  die Modell-Ausgabe hinterlegt wird. */
    public static final int    OUTLUC                = 5;
    /** Beschreibung fuer die {@link #OUTLUC}-Ressource. */
    public static final String OUTLUC_DESC           = "Output Raster (optional)";

    /** ID der (temporaeren) {@link WritableGrid}-Ressource fuer das Raster, in dem
     *  die Zuordnung der modellierten LUC zu den Start-Punkten (Gebiete/Siedlungen)
     *  hinterlegt wird. */
    public static final int    AREAMAPPING           = 6;
    /** Beschreibung fuer die {@link #AREAMAPPING}-Ressource. */
    public static final String AREAMAPPING_DESC      = "Temp-Raster for area mapping";


    /**
     * Erzeugt einen neuen Content-Manager.
     * @param addResourceCount Anzahl der zuseatzlichen durch eine Sub-Klasse
     *                         definierten Ressourcen
     */
    public CAContentManager(int addResourceCount) {
      super(BASE_RESCOUNT+addResourceCount);
      if ( addResourceCount < 0 )
        throw new IllegalArgumentException("Count of additional resources must be greater zero!");

      // Anzahl Zeitschritte
      resource[STEPCOUNT] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,STEPCOUNT_DESC,ScalarProperty.class,Integer.class,false);
      // Modellierter LUC-Typ
      resource[LUCCTYPE] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,LUCCTYPE_DESC,ScalarProperty.class,Integer.class,false);
      // Schwellwert fuer LUCC-Bewertung
      resource[THRESHOLD] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT,THRESHOLD_DESC,ScalarProperty.class,Double.class,false);
      // Raster fuer Base Szenario
      resource[BASELUC] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT | ValuePropertyResource.CATEGORY_OUTPUT,BASELUC_DESC,ScalarProperty.class,WritableGrid.class,false);
      // Area-Konfiguration
      resource[AREACONF] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_INPUT | ValuePropertyResource.CATEGORY_OUTPUT,AREACONF_DESC,ListProperty.class,CAArea.class,false);
      // Raster fuer Ausgabe
      resource[OUTLUC] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_OUTPUT,OUTLUC_DESC,ScalarProperty.class,WritableGrid.class,true);
      // Temp-Raster um Area-Zuordnung zu speichern
      resource[AREAMAPPING] = new ValuePropertyResource(ValuePropertyResource.CATEGORY_TEMP,AREAMAPPING_DESC,ScalarProperty.class,WritableGrid.class,false);
    }

    /**
     * Prueft die Ressourcen-Zuordnung auf Korrektheit.
     * <ol>
     *   <li>Die Anzahl der simulierten Modell-Schritte muss > 0 sein.</li>
     *   <li>Das Ausgabe-Raster muss das gleiche Format haben, wie das Ausgangsszenario.</li>
     *   <li>Das temp. Raster fuer die Area-Zuordnung muss das gleiche Format haben, wie das Ausgangsszenario.</li>
     *   <li>Alle Areas muessen ueber ausreichend Bedarfswerte verfuegen.</li>
     * </ol>
     */
    public void checkAndError() {
      super.checkAndError();

      // Modellschritte muss > 0 sein
      int stepCount = ((ScalarProperty)resource[STEPCOUNT].getData()).getOneTimeReadAccess().getValueAsInt();
      if ( stepCount <= 0 )
        throw new XuluDataException("Step count must be greater zero!");

      // Wenn ein Ausgabe-Raster angegeben wird, muss dieses das gleiche Format haben,
      // wie das Ausgangs-Szenario
      if ( resource[OUTLUC].getData() != null )
        AbstractWritableGrid.compareStructure(
                        (WritableGrid)((ScalarProperty)resource[OUTLUC].getData()).getOneTimeReadAccess().getValue(),
                        OUTLUC_DESC,
                        (WritableGrid)((ScalarProperty)resource[BASELUC].getData()).getOneTimeReadAccess().getValue(),
                        BASELUC_DESC,
                        false
        );

      // Temp-Raster fuer die Region-Zuordnung muss das gleiche Format haben, wie
      // das Ausgangs-Szenarion
      AbstractWritableGrid.compareStructure(
                      (WritableGrid)((ScalarProperty)resource[AREAMAPPING].getData()).getOneTimeReadAccess().getValue(),
                      AREAMAPPING_DESC,
                      (WritableGrid)((ScalarProperty)resource[BASELUC].getData()).getOneTimeReadAccess().getValue(),
                      BASELUC_DESC,
                      false
      );

    }
  }
}
