/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca.reservoir;

import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.GregorianCalendar;
import java.util.Set;
import java.util.HashSet;
import java.util.Vector;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.text.DecimalFormat;
import java.awt.Component;
import java.awt.Point;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import java.awt.Insets;
import javax.swing.JLabel;
import javax.swing.JCheckBox;

import org.geotools.feature.FeatureCollection;
import org.geotools.coverage.grid.GridCoverage2D;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.JPanel;
import schmitzm.swing.SliderSpinnerPanel;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;

import edu.bonn.xulu.XuluModellingPlatform;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;
import edu.bonn.xulu.plugin.model.ca.fire.ImpetusFireCA;
import static edu.bonn.xulu.plugin.model.ca.reservoir.SmallReservoirModelContentManager.*;

import org.apache.log4j.Logger;

/**
 * Diese Klasse stellt eine Implementierung des Modells {@linkplain ImpetusFireCA}
 * dar. Dieses benoetigt 7 Ressourcen, welche durch den
 * {@linkplain ImpetusFireCAContentManager} spezifiziert werden:
 * <ol>
 * <li><b>DHM ({@code dhm}):</b>
 *     Raster mit einem Hoehenmodell (Ausgangsszenario).
 *     </li>
 * <li><b>Staumauer ({@code damWall}):</b>
 *     FeatureCollection (z.B. aus Shape-File) mit einem Line-String, die die
 *     Lage der Staumauer definiert.
 *     </li>
 * <li><b>Hoehe der Staumauer ({@code damWallHeight}):</b>
 *     Hoehe der Staumauer ueber NN.
 *     </li>
 * <li><b>Versickerung pro Tag ({@code infiltration}):</b>
 *     Versickerung in mm pro Tag.
 *     </li>
 * <li><b>Verdunstung pro Tag ({@code evaporation}):</b>
 *     Verdunstung in mm pro Tag.
 *     </li>
 * <li><b>Modell-Schrittweite ({@code stepSize}):</b>
 *     Schrittweite in Tagen, mit welcher das "Entleeren" des Stausees
 *     (durch Verdunstung und Versickerung) in Rastern und Zwischenwerten
 *     in den Ausgabe-Ressourcen protokolliert wird.
 *     </li>
 * <li><b>Ausgabe: Groesse der ueberfluteten Flaeche in qm ({@code outFloodedAreaSize}):</b>
 *     Flaeche in Hektar, die ueberflutet wird, sowie die Flaeche zu den
 *     spaeteren Zeitpunkten, wenn Versickerung und Verdunstung (pro Tag)
 *     beruecksichtigt wird.
 *     </li>
 * <li><b>Ausgabe: Groesse der ueberfluteten Flaeche in qqm ({@code outFloodedAreaVol}):</b>
 *     Volumen in Qubikmeter, die ueberflutet wird, das Volumen zu den
 *     spaeteren Zeitpunkten, wenn Versickerung und Verdunstung (pro Tag)
 *     beruecksichtigt wird.
 *     </li>
 * <li><b>Ausgabe: Ueberfluteten Flaeche als Raster ({@code outFloodedArea}):</b>
 *     Enthaelt die ueberfluteten Zellen als Raster zum Zeitpunkt 0 (wenn der
 *     Stausee voll ist) und zu den spaeteren Zeitpunkten, wenn Versickerung
 *     und Verdunstung (pro Tag) beruecksichtigt wird.
 *     </li>
 * <li><b>Temporaere: Staumauer als DHM ({@code damDHM}):</b>
 *     Enthaelt ein Hoehenmodell, in dem der Staumauer-Vektor dargestellt wird.
 *     Dieses wird automatisch berechnen.
 *     </li>
 * </ol>
 * @see SmallReservoirModelContentManager
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SmallReservoirModel extends AbstractStepModel {
  /** {@link ResourceProvider}, der die Uebersetzungen fuer die GUI
   *  bereitstellt: {@code locales.XuluModel_SmallReservoirModel} */
  protected static ResourceProvider RESOURCE = XuluModellingPlatform.registerResourceProvider("XuluModel_SmallReservoirModel",Locale.ENGLISH);

	  /** Logger fuer Debug-Ausgaben und Warnings. */
  public static final Logger LOGGER = LangUtil.createLogger(SmallReservoirModel.class);

  /** Speichert den ContentManager fuer das Modell.
   *  @see SmallReservoirModelContentManager */
  protected SmallReservoirModelContentManager contManager;

  //********** Lese/Schreibrechte, die fuer den gesamten ***********
  //********** Modellanlauf gehalten werden muessen      ***********
  private PropertyReadAccess       RA_dhm = null;  // Hoehenmodell
  private PropertyReadAccess       RA_damWall = null;  // Lage der Staumauer
  private PropertyReadAccess       RA_damWallHeight = null;  // Damm-Hoehe
  private PropertyReadAccess       RA_infiltration = null;  // Versickerung
  private PropertyReadAccess       RA_evaporation = null;  // Verdunstung
  private PropertyReadAccess       RA_stepSize = null;  // Schrittgroesse des Modells
  private ListPropertyWriteAccess  WA_outFloodedAreaSize= null;  // Ausgabe: Ueberflutete Flaeche in qm
  private ListPropertyWriteAccess  WA_outFloodedAreaVol= null;  // Ausgabe: Ueberflutete Flaeche in m³
  private PropertyWriteAccess      WA_outFloodedArea = null;  // Ausgabe-Raster: Ueberflutete Flaeche als Raster
  private PropertyWriteAccess      WA_damWallDHM = null;  // Temp-Raster: Staumauer als DHM

  //**************** Variablen mit denen gearbeitet wird *******************
  protected  WritableGrid       dhm = null; // Hoehenmodell
  protected  FeatureCollection  damWall = null; // Lage der Staumauer
  protected  double[]           damWallHeight = null; // Hoehe der Staumauer ueber NN
  protected  double[]           damWallHeightMin = null; // min. Hoehe der Staumauer ueber NN
  protected  double[]           damWallHeightMax = null; // max. Hoehe der Staumauer ueber NN
  protected  double             infiltration = 0; // Versickerung in mm
  protected  double             evaporation = 0; // Verdunstung in mm
  protected  MultiGrid          outFloodedArea = null; // Ueberflutete Flaechen als Raster
  protected  double             outFloodedAreaSize = 0; // akt. ueberflutete Flaeche in qm
  protected  double             outFloodedAreaVol = 0; // akt. ueberflutetes Volumen in qqm
  protected  WritableGrid       outFloodedAreaGrid = null; // akt. ueberflutete Flaeche als Raster
  protected  WritableGrid       damWallDHM = null; // Staumauer als DHM
  protected  int                stepSize = 0; // Groesse in Tagen eines Modellschritts
  protected  boolean            storeStepResults = true;
  protected  int                areaMinX = 0;
  protected  int                areaMaxX = 0;
  protected  int                areaMinY = 0;
  protected  int                areaMaxY = 0;
  protected  double             areaCellSize = 0;
  protected  int                iterCount = 0;
  protected  Set<Point>[]       damCells = null;
  protected  Set<Point>         floodedCells = new HashSet<Point>();

// Nachbar-Zellen relativ zur Zentrum-Zelle
//  private  static final int[][] NEIGHBOR = new int[][] {
//      {-1,-1}, {0,-1}, {1,-1},
//      {-1, 0},         {1, 0},
//      {-1, 1}, {0, 1}, {1, 1}
//  };
  private  static final int[][] NEIGHBOR = new int[][] {
               {0,-1},
      {-1, 0},         {1, 0},
               {0, 1}
  };

  // Modell-Spezifische GUI
  protected  GUI modelGUI = null;

  /**
   * Erzeugt eine neue Instanz des Modells.
   */
  public SmallReservoirModel() {
    super( new SmallReservoirModelContentManager() );
    contManager = (SmallReservoirModelContentManager)super.contManager;
    modelResourceProvider = RESOURCE;
    modelGUI    = new GUI();
    modelGUI.setEnabled(false);
    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
  }

  /**
   * Liefert die modell-spezifische GUI des Kleinstausee-Modells.
   */
  public Component getModelSpecificGUI() {
    return modelGUI;
  }


  ////////////////////////////////////////////////////////////////////////
  ///////////////   Implementierung von AbstractXuluModel   //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
   * werden Referenzen auf die Ressourcen in lokalen Variablen
   * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
   * zu koennen.
   */
  public void performModelInit() {
    if ( contManager == null )
      return;
    // Zugriffsrechte aus Ressourcen/Propertys holen
    RA_dhm                = ((ScalarProperty)contManager.getResource(DHM_NO).getData()).getReadAccess(this);
    RA_damWall            = ((ScalarProperty)((SingleFeatureCollection)contManager.getResource(DAM_LOC_NO).getData()).getProperty("FEATURES")).getReadAccess(this);
    RA_damWallHeight      = ((ScalarProperty)contManager.getResource(DAM_HEIGHT_NO).getData()).getReadAccess(this);
    RA_evaporation        = ((ScalarProperty)contManager.getResource(EVAPO_NO).getData()).getReadAccess(this);
    RA_infiltration       = ((ScalarProperty)contManager.getResource(INFILT_NO).getData()).getReadAccess(this);
    RA_stepSize           = ((ScalarProperty)contManager.getResource(STEP_SIZE_NO).getData()).getReadAccess(this);
    WA_outFloodedAreaSize = ((ListProperty)contManager.getResource(OUT_AREA_SIZE_NO).getData()).getWriteAccess(this);
    WA_outFloodedAreaVol  = ((ListProperty)contManager.getResource(OUT_AREA_VOL_NO).getData()).getWriteAccess(this);
    WA_damWallDHM         = ((ScalarProperty)contManager.getResource(TMP_DHM_NO).getData()).getWriteAccess(this);

    // Variablen belegen mit denen gearbeitet wird
    dhm                = (WritableGrid)RA_dhm.getValue();
    damWall            = (FeatureCollection)RA_damWall.getValue();
    damWallDHM         = (WritableGrid)WA_damWallDHM.getValue();
    outFloodedAreaSize = 0;
    outFloodedAreaVol  = 0;
    outFloodedArea     = (MultiGrid)contManager.getResource(OUT_AREA_NO).getData();
    evaporation        = RA_evaporation.getValueAsDouble();
    infiltration       = RA_infiltration.getValueAsDouble();
    stepSize           = RA_stepSize.getValueAsInt();
    // Ausmasse des Rasters
    areaMinX       = dhm.getMinX();
    areaMaxX       = dhm.getMinX() + dhm.getWidth() - 1;
    areaMinY       = dhm.getMinY();
    areaMaxY       = dhm.getMinY() + dhm.getHeight() - 1;
    areaCellSize   = dhm.getCellWidth() * dhm.getCellHeight();

    // Ausgabe und Temp-Raster initialisieren
    for (;outFloodedArea.getGridCount() > 0;)
      outFloodedArea.removeGrid(0);
    outFloodedArea.setTemplateDimension(dhm);
    //    LangUtil.gc();
    WA_outFloodedAreaSize.removeAll();
    WA_outFloodedAreaVol.removeAll();
    for (int y=areaMinY; y<=areaMaxY; y++)
      for (int x=areaMinX; x<=areaMaxX; x++) {
        damWallDHM.setRasterSample(0, x, y);
      }

    // Schrittanzahl unbestimmt
    this.stepCount = Integer.MAX_VALUE;

    // Rasterzellen und min/max. Hoehe der Staumauer(n) ermitteln
    determineDamWallCells();

    if ( modelGUI != null ) {
      // Ausmasse des Sliders neu einstellen
      modelGUI.damHeightSlider.setRestrictions(
        RA_damWallHeight.getValueAsDouble(),
        damWallHeightMin[0],
        damWallHeightMax[0],
        1
      );
      // Schnittweite, Infiltration und Evaporation aus
      // Modell-Parametern anzeigen
      modelGUI.setModelStepSize( stepSize );
      modelGUI.setEvaporation( evaporation );
      modelGUI.setInfiltration( infiltration );

      // GUI aktivieren
      modelGUI.setEnabled(true);
    }

  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    // Ressourcen wieder freigeben
    releaseAccess(RA_dhm);
    releaseAccess(RA_damWall);
    releaseAccess(RA_damWallHeight);
    releaseAccess(RA_evaporation);
    releaseAccess(RA_infiltration);
    releaseAccess(RA_stepSize);
    releaseAccess(WA_outFloodedAreaSize);
    releaseAccess(WA_outFloodedAreaVol);
    releaseAccess(WA_outFloodedArea);
    releaseAccess(WA_damWallDHM);
  }

  /**
   * Fuehrt die Modell-Berechnung durch.
   */
  public void performModelStep(int stepNo) {
//    Random            RANDOM   = new Random();
//    GregorianCalendar calendar = new GregorianCalendar();
//    DateFormat        format   = new SimpleDateFormat("HH:mm:ss.SSS");

    if ( stepNo == 1 ) {
      // GUI sperren
    	modelGUI.setEnabled( false );
    	// Staudam-Hoehe aus GUI verwenden
    	damWallHeight[0] = modelGUI.getDamnHeight();
    	// Schrittweite in welcher die Stausee-Daten
    	// protokolliert werden
    	stepSize = modelGUI.getModelStepSize();
    	storeStepResults = modelGUI.checkStoreStepGrids();
    	
    	// Ausgabe-Raster fuer initialen Zustand erzeugen und
    	// Variablen fuer initialien Schritt initialisieren
    	outFloodedArea.addGrid(Float.NaN);
    	outFloodedAreaSize = 0;
    	outFloodedAreaVol  = 0;
    	outFloodedAreaGrid = outFloodedArea.getGrid(0);
    	// Startzellen initialisieren
    	determineDamWallDHMAndInitStartCells();
    	// Ausmasse des gefuellten Stausees berechnen
    	fillReservoir();
    	// Ergebnis in Ausgabe schreiben
    	WA_outFloodedAreaSize.addValue( (float)outFloodedAreaSize );
    	WA_outFloodedAreaVol.addValue( (float)outFloodedAreaVol );
    }

    // Ausgabe-Raster fuer Entleerungs-Schritt erzeugen und
    // Variablen fuer Entleerungs-Schritt initialisieren
    if ( storeStepResults ) {
      outFloodedArea.addGrid(Float.NaN);
      outFloodedAreaGrid = outFloodedArea.getGrid( outFloodedArea.getGridCount()-1 );
    } else
      outFloodedAreaGrid = null;
    outFloodedAreaSize = 0;
    outFloodedAreaVol  = 0;
    // Stausee fuer einen Zeit-Abschnitt leeren
    emptyReservoir(stepNo, stepSize);
    // Ergebnis in Ausgabe schreiben
    WA_outFloodedAreaSize.addValue( (float)outFloodedAreaSize );
    WA_outFloodedAreaVol.addValue( (float)outFloodedAreaVol );

    // Wenn keine ueberfluteten Zellen mehr vorhanden sind, Modell abbrechen
    if ( floodedCells.isEmpty() ) {
      // letztes Raster aus der Liste entfernen, da dieses leer ist
      outFloodedArea.removeGrid(outFloodedArea.getGridCount()-1);
      this.stopModelling();
    }
  }

  /**
   * Entleert den Stausee und berechnet pro Tag (oder je Modellschritt-Groesse)
   * verbleibende die Ausmasse des Stausees.
   */
  protected void emptyReservoir(int stepNo, int dayCount) {
    statusOut.print( RESOURCE.getString("Model.Calc.Empty",stepNo*dayCount) );

    //******* vorerst fuer alles die Hoehe des ersten Damms verwenden *********
    // @todo: individuelle Damm-Hoehe verwenden
    final double damHeight = damWallHeight[0];

    // Hoehe, um die der Wasserstand reduziert wird
    final double reducing = dayCount * (infiltration + evaporation) * 0.001;

    double newHeightAbs     = damHeight - stepNo*reducing;
    double cellGroundHeight = 0;
    double newHeightRel     = 0;
    Set<Point> lastStepFloodedCells = new HashSet<Point>(floodedCells);
    for (Point p : lastStepFloodedCells) {
      checkBreakingCommands();

      // Hoehe der (unueberfluteten) Zelle ueber NN
      cellGroundHeight = dhm.getRasterSampleAsFloat(p.x,p.y);
      // Wenn neue Stau-Hoehe noch groesser ist, als die
      // Hoehe der Zellen, bleibt die Zelle ueberflutet
      if ( cellGroundHeight < newHeightAbs ) {
        // Relative Stau-Hoehe (ueber Grund)
        newHeightRel = newHeightAbs - cellGroundHeight;
        // Zelle bleibt ueberflutet
        if ( outFloodedAreaGrid != null )
          outFloodedAreaGrid.setRasterSample(newHeightRel,p.x,p.y);
        // Ueberflutete Flaeche in qm summieren
        outFloodedAreaSize += areaCellSize;
        // Ueberflutetes Volumen in m³ erhoehen
        outFloodedAreaVol  += areaCellSize * newHeightRel;
      } else {
        // Zelle aus den ueberfluteten entfernen
        floodedCells.remove(p);
      }
    }
    // berechnete Werte in der GUI anzeigen
    if ( modelGUI != null ) {
      modelGUI.setCalculatedArea( outFloodedAreaSize );
      modelGUI.setCalculatedVolume( outFloodedAreaVol );
    }
    statusOut.println( RESOURCE.getString("Model.Calc.Done") );
  }

  /**
   * Berechnet die Ausmasse des Stausees (fuellt den Stausee).
   */
  protected void fillReservoir() {
    statusOut.print( RESOURCE.getString("Model.Calc.Fill") );

    //******* vorerst fuer alles die Hoehe des ersten Damms verwenden *********
    // @todo: individuelle Damm-Hoehe verwenden
    final double damHeight = damWallHeight[0];

    double nbhHeight        = 0;
    int    nbhX             = 0;
    int    nbhY             = 0;
    double relHeight        = 0;
    // "floodedCells" soll nach der Befuellung alle ueberfluteten Zellen
    // enthalten. Fuer das Ueberfluten wird aber immer nur der "Rand"
    // betrachtet, aus dem immer wieder Zellen entfernt werden (bis der Rand
    // leer ist) -> neben "floodedCells" eine temp. Menge fuer Rand verwalten
    Set<Point> floodedCellsBorder = new HashSet<Point>(floodedCells);
    for (int iterNo=1; !floodedCellsBorder.isEmpty(); iterNo++) {
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,0,iterNo) );
      // Die Nachbarn aller im letzten Schritt zur Ueberflutung markierte
      // Zellen ueberfluten und die geeigneten Nachbarzellen markieren
      Set<Point> lastStepFloodedCells = new HashSet<Point>(floodedCellsBorder);
      for (Point cell : lastStepFloodedCells) {
        // Zelle aus Rand entfernen
        floodedCellsBorder.remove(cell);
        // Stausee-Hoehe ueber Grund
        relHeight = damHeight - dhm.getRasterSampleAsFloat(cell.x,cell.y);
        // Zelle ueberfluten (im Ausabe-Raster markieren)
        outFloodedAreaGrid.setRasterSample(relHeight,cell.x,cell.y);
        floodedCells.add( cell );
        // Ueberflutete Flaeche in qm summieren
        outFloodedAreaSize += areaCellSize;
        // Ueberflutetes Volumen in m³ erhoehen
        outFloodedAreaVol  += areaCellSize * relHeight;

        // berechnete Werte in der GUI anzeigen
        if ( modelGUI != null ) {
          modelGUI.setCalculatedArea( outFloodedAreaSize );
          modelGUI.setCalculatedVolume( outFloodedAreaVol );
        }

        // Noch nicht ueberflutete Nachbarzellen ueberpruefen und
        // ggf. markieren (in Rand aufnehmen)
        for (int[] nbhOffset : NEIGHBOR) {
          checkBreakingCommands();

          nbhX = cell.x + nbhOffset[0];
          nbhY = cell.y + nbhOffset[1];
          if ( nbhX >= areaMinX && nbhX <= areaMaxX &&
               nbhY >= areaMinY && nbhY <= areaMaxY &&
               Float.isNaN( outFloodedAreaGrid.getRasterSampleAsFloat(nbhX,nbhY) ) ) {
            // Hoehe der Zelle bestimmen
            nbhHeight = dhm.getRasterSampleAsFloat(nbhX,nbhY);
            // Staumauer beruecksichtigen
            nbhHeight = Math.max(nbhHeight, damWallDHM.getRasterSampleAsFloat(nbhX,nbhY));
            // Wenn Hoehe unterhalb der Staumauer-Hoehe liegt, wird
            // Zelle ueberflutet
            if ( nbhHeight < damHeight )
              floodedCellsBorder.add( new Point(nbhX,nbhY) );
          }
        }
      }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,0,iterNo) );
    }

    statusOut.println( RESOURCE.getString("Model.Calc.Done") );
  }

  /**
   * Ermittelt die von dem Staumauer-Shape ueberlagerten Rasterzellen.
   * Zudem wird die minimale/maximale Staumauer-Hoehe ermittelt.
   */
  protected void determineDamWallCells() {
    // Erstes Feature aus der Collection holen
    Geometry           g   = damWall.features().next().getDefaultGeometry();
    // LineString(s) in Vector einfuegen
    Vector<LineString> lineStrings  = new Vector<LineString>();
    if ( g instanceof LineString )
      lineStrings.add( (LineString)g );
    if ( g instanceof MultiLineString )
      for (int i=0; i<((MultiLineString)g).getNumGeometries(); i++)
        lineStrings.add( (LineString)g.getGeometryN(i) );
    if ( lineStrings.isEmpty() )
      throw new XuluDataException( RESOURCE.getString("Error.In.Dam.Loc.Type") );

    GridCoverage2D gc = GridUtil.convertToGridCoverage2D( dhm );
    damCells         = new HashSet[lineStrings.size()];
    damWallHeight    = new double[damCells.length];
    damWallHeightMin = new double[damCells.length];
    damWallHeightMax = new double[damCells.length];
    for (int i=0; i<damCells.length; i++) {
      LineString ls = lineStrings.elementAt(i);
      damCells[i] = GridUtil.getOverlappingCells( gc, ls, (Set)null );
      // Minimale Staumauer-Hoehe = Hoehe der Tiefenlinie
      damWallHeightMin[i] = Double.MAX_VALUE;
      for (Point p : damCells[i])
        damWallHeightMin[i] = Math.min(
            damWallHeightMin[i],
            dhm.getRasterSampleAsDouble(p.x,p.y)
        );
        // Maximale Staumauer-Hoehe = Niedrigstes Ende der Staumauer
        damWallHeightMax[i] = Math.min(
            dhm.getGridSampleAsDouble( ls.getStartPoint().getX(), ls.getStartPoint().getY() ),
            dhm.getGridSampleAsDouble( ls.getEndPoint().getX(), ls.getEndPoint().getY() )
        );
        // Default-Hoehe = Minimale Hoehe
        damWallHeight[i] = damWallHeightMax[i];
    }
  }

  /**
   * Berechnet ein DHM, welches an den Stellen der Staumauer die Hoehe der
   * Staumauer ueber NN enthaelt und ansonsten 0. Zudem werden die Start-Zellen
   * fuer den Wasserzufluss, sowie die minimale/maximale Staumauer-Hoehe ermittelt.
   */
  protected void determineDamWallDHMAndInitStartCells() {
    statusOut.print( RESOURCE.getString("Model.Calc.Dam.DHM") );

    
    for (int i=0; i<damCells.length; i++)
      for (Point p : damCells[i])
        this.damWallDHM.setRasterSample(damWallHeight[i], p.x, p.y );

    // Start-Zelle(n) fuer die Ueberflutung berechnen
    // --> Fuer alle Damm-Zellen jeweils die 4 Umgebungen NO, NW, SW, SO
    //     betrachten und die Umgebung mit der GERINGSTE Durchschnittshoehe
    //     suchen
    // --> die Umgebung liegt IN Fluss-Richtung vom Staudamm
    // --> fuer die gefundene Damm-Zelle diejenige Umgebung als Start-Zellen
    //     waehlen, welche die GROESSTE Durchschnittshoehe hat
    // --> diese Umgebung liegt vom Staudamm aus GEGEN die Fluss-Richtung
    final int[][][] NBH_AREA = initNbhOffset(5);
    Point        minCell                = null;
    float        minCellMinNbhHeight    = 0;
    Set<Point>[] minCellNbhCells        = new HashSet[4];
    float[]      minCellNbhHeights      = new float[4];
    Point        nbhCell                = new Point();
    float        nbhCellHeight          = 0;

    // Alle Daemme betrachten
    floodedCells.clear();
    for (int damNo=0; damNo<damCells.length; damNo++) {
      // Alle Damm-Zellen betrachten
      for (Point p : damCells[damNo]) {
        Set<Point>[] nbhCells     = new HashSet[4];
        float[]      nbhHeights   = new float[4];
        // Alle 4 Richtungen um die Zelle betrachten
        for (int d=0; d<NBH_AREA.length; d++) {
          float nbhHeightSum = 0;
          nbhCells[d] = new HashSet<Point>();
          // Alle Nachbarzellen betrachten und Hoehen-Summe berechnen
          for (int i=0; i<NBH_AREA[d].length; i++) {
            nbhCell.x = p.x + NBH_AREA[d][i][0];
            nbhCell.y = p.y + NBH_AREA[d][i][1];
            if ( nbhCell.x >= areaMinX && nbhCell.x <= areaMaxX &&
                 nbhCell.y >= areaMinY && nbhCell.y <= areaMaxY  &&
                 !damCells[damNo].contains(nbhCell) ) {
              nbhCellHeight = dhm.getRasterSampleAsFloat(nbhCell.x,nbhCell.y);
              if ( nbhCellHeight < damWallHeight[damNo] ) {
                nbhCells[d].add( (Point)nbhCell.clone() );
                nbhHeightSum += dhm.getRasterSampleAsFloat(nbhCell.x,nbhCell.y);
              }
            }
          }
          // Durchschnittshoehe berechnen und ggf. das Minimum neu setzen
          if ( !nbhCells[d].isEmpty() ) {
            nbhHeights[d] = nbhHeightSum / nbhCells[d].size();
            if ( nbhHeights[d] < minCellMinNbhHeight || minCell == null ) {
              minCell             = p;
              minCellMinNbhHeight = nbhHeights[d];
              minCellNbhHeights   = nbhHeights;
              minCellNbhCells     = nbhCells;
            }
          }
        }
      }

      // Die Richtung der minimalen Umgebung suchen, die die groesste
      // Durchschnittshoehe besitzt (diese Umgebung liegt von der Staumauer
      // aus GEGEN die Flussrichtung)
      int   minCellMaxDir       = 0;
      float minCellMaxNbhHeight = 0;
      for (int d=0; d<minCellNbhHeights.length; d++) {
        if ( minCellNbhHeights[d] > minCellMaxNbhHeight ) {
          minCellMaxNbhHeight = minCellNbhHeights[d];
          minCellMaxDir       = d;
        }
      }
      // Die Umgebungszellen der Maximum-Richtung als Start-Zellen setzen
      for (Point cell : minCellNbhCells[minCellMaxDir])
        this.floodedCells.add(cell);
    }
    statusOut.println( RESOURCE.getString("Model.Calc.Done") );
  }

  /**
   * Erstellt eine Array von 2-dimensionalen Offset-Koordinaten.
   * Dieser ist in die 4 Quadranten eingeteilt.<br>
   * Dimension 1: Quadrant (0=NO, 1=NW, 2=SW, 3=SO)<br>
   * Dimension 2: Koordinaten-Index (0 bis {@code (size+1)^2)})<br>
   * Dimension 3: X oder Y (0=X, 1=Y)<br>
   * Beispiel: area[2][3][0] = 3. X-Koordinate in 2. Quadranten
   * @param size Radius der Umgebung (in Zellen)
   */
  private int[][][] initNbhOffset(int size) {
    int arrSize = size+1;

    int[][][] offset = new int[4][arrSize*arrSize][2];
    for (int d=0; d<4; d++) {
      int fx = (d==0 | d==3) ? 1 : -1;
      int fy = (d==2 | d==3) ? 1 : -1;
      for (int y=0; y<arrSize; y++)
        for (int x=0; x<arrSize; x++) {
          offset[d][y*arrSize+x][0] = x*fx;
          offset[d][y*arrSize+x][1] = y*fy;
        }
    }
    return offset;
  }

  /////////////////////////////////////////////////////////////////////////
  /////////////////////   Modellspezifische GUI   /////////////////////////
  /////////////////////////////////////////////////////////////////////////
  /**
   * Modell-spezifische GUI des {@linkplain SmallReservoirModel Kleinstausee-Modells}.
   * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
   */
  protected class GUI extends JPanel {
    /** Zahlenformat, in dem die berechneten Werte angezeigt werden. */
    protected final DecimalFormat DECIMAL_FORMAT = new DecimalFormat("0.00");


    /** Slider, mit dem die Hoehe der Staumauer ueber NN eingestellt wird. */
    protected SliderSpinnerPanel damHeightSlider = null;
    /** Slider, mit dem die Schrittweite (in Tagen) eingestellt wird, mit
     *  der sie Stausee-Daten protokolliert werden. */
    protected SliderSpinnerPanel modelStepSizeSlider = null;
    /** Slider, mit dem die Evaporation (in mm) eingestellt wird. */
    protected SliderSpinnerPanel evaporationSlider = null;
    /** Slider, mit dem die Infiltration (in mm) eingestellt wird. */
    protected SliderSpinnerPanel infiltrationSlider = null;
    /** CheckBox, ob Zwischenschritte als Raster gespeichert werden sollen. */
    protected JCheckBox storeStepGridsCheckBox = null;
    /** Label in dem die Ueberschrift der berechneten Stau-Flaeche angezeigt wird. */
    protected JLabel dammedAreaTitle = null;
    /** Label in dem die berechnete Stau-Flaeche angezeigt wird. */
    protected JLabel dammedAreaValue = null;
    /** Label in dem die Ueberschrift des berechneten Stau-Volumens angezeigt wird */
    protected JLabel dammedVolumeTitle = null;
    /** Label in dem das berechnete Stau-Volumen angezeigt wird. */
    protected JLabel dammedVolumeValue = null;

    /**
     * Erzeugt eine neue GUI. Ruft {@link #initGUI()} auf.
     */
    public GUI() {
      this.setLayout( new GridBagLayout() );

      initGUI();

      this.add(damHeightSlider, new GridBagConstraints(
          0, 0, 2, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5,5,5,5), 0, 0) );
      this.add(modelStepSizeSlider, new GridBagConstraints(
          0, 1, 2, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5,5,5,5), 0, 0) );
      this.add(infiltrationSlider, new GridBagConstraints(
          0, 2, 2, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5,5,5,5), 0, 0) );
      this.add(evaporationSlider, new GridBagConstraints(
          0, 3, 2, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5,5,5,5), 0, 0) );
      this.add(storeStepGridsCheckBox, new GridBagConstraints(
          0, 4, 2, 1, 1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.HORIZONTAL, new Insets(5,5,5,5), 0, 0) );
      this.add(dammedAreaTitle, new GridBagConstraints(
          0, 5, 1, 1, 1.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,5,5,5), 0, 0) );
      this.add(dammedAreaValue, new GridBagConstraints(
          1, 5, 1, 1, 1.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,5,5,5), 0, 0) );
      this.add(dammedVolumeTitle, new GridBagConstraints(
          0, 6, 1, 1, 1.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,5,5,5), 0, 0) );
      this.add(dammedVolumeValue, new GridBagConstraints(
          1, 6, 1, 1, 1.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(5,5,5,5), 0, 0) );
    }

    /**
     * Erzeugt die Komponenten der modellspezifischen GUI.
     */
    protected void initGUI() {
      this.damHeightSlider = new SliderSpinnerPanel(
          SliderSpinnerPanel.HORIZONTAL, "");
      this.modelStepSizeSlider = new SliderSpinnerPanel(
          SliderSpinnerPanel.HORIZONTAL, 1, 500, 1, 1, "" );
      this.infiltrationSlider = new SliderSpinnerPanel(
          SliderSpinnerPanel.HORIZONTAL, 0.1, 100, 0.1, 1, "");
      this.evaporationSlider = new SliderSpinnerPanel(
          SliderSpinnerPanel.HORIZONTAL, 0.1, 100, 0.1, 1, "");
      this.storeStepGridsCheckBox = new JCheckBox("",true);
      this.dammedAreaTitle        = new JLabel("");
      this.dammedAreaValue        = new JLabel("");
      this.dammedVolumeTitle      = new JLabel("");
      this.dammedVolumeValue      = new JLabel("");
      // Labels lokalisieren
      resetCaptions( null );
    }

    /**
     * Belegt die Labels in der GUI neu. Sind in der uebergebenen Map fuer die Keys
     * keine Werte hinterlegt (oder wenn {@code captionMap = null}, wird auf den
     * {@linkplain ImpetusFireCA#RESOURCE Resource-Provider} des Modells zurueckgegriffen.
     * @param captionMap Map in der nach den Label-Bezeichnungen gesucht wird (kann
     *                   {@code null} sein)
     */
    public void resetCaptions(Map<String,Object> captionMap) {
      this.damHeightSlider.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.DamHeight") );
      this.modelStepSizeSlider.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.ModelSteps") );
      this.infiltrationSlider.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.Infilt") );
      this.evaporationSlider.getHeaderLabel().setText( determineCaption(captionMap, RESOURCE, "Model.GUI.Evapo") );
      this.storeStepGridsCheckBox.setText( determineCaption(captionMap, RESOURCE, "Model.GUI.Store.StepResults") );
      this.dammedAreaTitle.setText( determineCaption(captionMap, RESOURCE, "Model.GUI.Calc.Area") );
      this.dammedVolumeTitle.setText( determineCaption(captionMap, RESOURCE, "Model.GUI.Calc.Vol") );
    } 

    /**
     * Deaktiviert die Eingabe-Komponenten der GUI (nicht wie die
     * {@code super}-Methode alle Komponenten)
     */
    public void setEnabled(boolean enabled) {
      damHeightSlider.setEnabled( enabled );
      modelStepSizeSlider.setEnabled( enabled );
      infiltrationSlider.setEnabled( enabled );
      evaporationSlider.setEnabled( enabled );
      storeStepGridsCheckBox.setEnabled( enabled );
    }

    /**
     * Liefert die in der GUI eingestellte Staumauer-Hoehe.
     */
    public double getDamnHeight() {
      return damHeightSlider.getValue();
    }

    /**
     * Setzt die in der GUI eingestellte Staumauer-Hoehe.
     * @param value einzustellender Wert
     */
    public void setDamnHeight(double value) {
      damHeightSlider.setValue(value);
    }

    /**
     * Liefert die in der GUI eingestellte Schrittweite des Modells.
     */
    public int getModelStepSize() {
      return (int)modelStepSizeSlider.getValue();
    }

    /**
     * Setzt die in der GUI eingestellte Schrittweite des Modells.
     * @param value einzustellender Wert
     */
    public void setModelStepSize(int value) {
      modelStepSizeSlider.setValue(value);
    }

    /**
     * Liefert die in der GUI eingestellte Evaporation.
     */
    public double getEvaporation() {
      return evaporationSlider.getValue();
    }

    /**
     * Setzt die in der GUI eingestellte Evaporation.
     * @param value einzustellender Wert
     */
    public void setEvaporation(double value) {
      evaporationSlider.setValue(value);
    }

    /**
     * Liefert die in der GUI eingestellte Infiltration.
     */
    public double getInfiltration() {
      return infiltrationSlider.getValue();
    }

    /**
     * Setzt die in der GUI eingestellte Infiltration.
     * @param value einzustellender Wert
     */
    public void setInfiltration(double value) {
      infiltrationSlider.setValue(value);
    }

    /**
     * Prueft, ob in der GUI eingestellt ist, dass die Zwischenergebnisse
     * als Raster gespeichert werden.
     */
    public boolean checkStoreStepGrids() {
      return storeStepGridsCheckBox.isSelected();
    }

    /**
     * Bestimmt, ob in der GUI eingestellt ist, dass die Zwischenergebnisse
     * als Raster gespeichert werden.
     * @param storeStepGrids einzustellender Wert
     */
    public void setStoreStepGrids(boolean storeStepGrids) {
      storeStepGridsCheckBox.setSelected( storeStepGrids );
    }

    /**
     * Setzt die in der GUI angezeigte (berechnete) "Stau-Flaeche".
     * @param area anzuzeigender Wert
     */
    public void setCalculatedArea(double area) {
      dammedAreaValue.setText(
          DECIMAL_FORMAT.format(area) );
    }

    /**
     * Setzt das in der GUI angezeigte (berechnete) "Stau-Volumen".
     * @param volume anzuzeigender Wert
     */
    public void setCalculatedVolume(double volume) {
      dammedVolumeValue.setText(
          DECIMAL_FORMAT.format(volume) );
    }

  }

}
