/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.ca.reservoir;

import java.util.Map;

import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.AbstractXuluModel;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.PropertiesResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.plugin.data.feature.SingleFeatureCollection;

import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;

import schmitzm.data.WritableGrid;
import schmitzm.data.AbstractWritableGrid;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.lang.ResourceProvider;
import edu.bonn.xulu.model.DefaultModelResource;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import edu.bonn.xulu.plugin.model.ca.fire.ImpetusFireCA;


/**
 * Diese Klasse definiert und verwaltet die 7 Ressourcen fuer das Modell
 * {@linkplain SmallReservoirModel} und prueft diese auf Korrektheit.<br>
 * Die Art der benoetigten Ressourcen ist der {@linkplain SmallReservoirModel}-Doku
 * zu entnehmen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SmallReservoirModelContentManager extends AbstractModelContentManager {
  public static final int    DHM_NO             = 0;
  private             String DHM_DESC           = "";
  public static final int    DAM_LOC_NO         = 1;
  private             String DAM_LOC_DESC       = "";
  public static final int    DAM_HEIGHT_NO      = 2;
  private             String DAM_HEIGHT_DESC    = "";
  public static final int    INFILT_NO          = 3;
  private             String INFILT_DESC        = "";
  public static final int    EVAPO_NO           = 4;
  private             String EVAPO_DESC         = "";
  public static final int    STEP_SIZE_NO       = 5;
  private             String STEP_SIZE_DESC     = "";
  public static final int    OUT_AREA_SIZE_NO   = 6;
  private             String OUT_AREA_SIZE_DESC = "";
  public static final int    OUT_AREA_VOL_NO    = 7;
  private             String OUT_AREA_VOL_DESC  = "";
  public static final int    OUT_AREA_NO        = 8;
  private             String OUT_AREA_DESC      = "";
  public static final int    TMP_DHM_NO         = 9;
  private             String TMP_DHM_DESC       = "";

  /**
   * Erzeugt einen neuen ContentManager fuer das Modell
   * {@linkplain SmallReservoirModel}.
   */
  public SmallReservoirModelContentManager() {
    super(10);
    // ===== Ressourcen festlegen =====
    // DHM
    resource[DHM_NO] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,DHM_DESC,ScalarProperty.class,WritableGrid.class,false);
    // Staumauer
    resource[DAM_LOC_NO] = new PropertiesResource(ModelResource.CATEGORY_INPUT,DAM_LOC_DESC,SingleFeatureCollection.class,false);
    // Staumauerhoehe
    resource[DAM_HEIGHT_NO] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,DAM_HEIGHT_DESC,ScalarProperty.class,Double.class,false);
    // Versickerungsrate
    resource[INFILT_NO] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,INFILT_DESC,ScalarProperty.class,Double.class,false);
    // Verdunstungsrate
    resource[EVAPO_NO] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,EVAPO_DESC,ScalarProperty.class,Double.class,false);
    // Modell-Schrittweite
    resource[STEP_SIZE_NO] = new ValuePropertyResource(ModelResource.CATEGORY_INPUT,STEP_SIZE_DESC,ScalarProperty.class,Integer.class,false);
    // Ausgabe: Ueberflutete Flaeche
    resource[OUT_AREA_SIZE_NO] = new ValuePropertyResource(ModelResource.CATEGORY_OUTPUT,OUT_AREA_SIZE_DESC,ListProperty.class,Double.class,false);
    // Ausgabe: Ueberflutetes Volumen
    resource[OUT_AREA_VOL_NO] = new ValuePropertyResource(ModelResource.CATEGORY_OUTPUT,OUT_AREA_VOL_DESC,ListProperty.class,Double.class,false);
    // Ausgabe: Ueberflutete Flaeche (Raster)
    resource[OUT_AREA_NO] = new DefaultModelResource(ModelResource.CATEGORY_OUTPUT,OUT_AREA_DESC,MultiGrid.class,false);
    // Temp: DHM fuer Staumauer (Raster)
    resource[TMP_DHM_NO] = new ValuePropertyResource(ModelResource.CATEGORY_TEMP,TMP_DHM_DESC,ScalarProperty.class,WritableGrid.class,false);

    // ===== Ressourcen Bezeichnungen lokalisieren =====
    resetCaptions(null);
}

  /**
   * Belegt die Ressourcen-Beschreibungen neu. Sind in der uebergebenen Map fuer die Keys
   * keine Werte hinterlegt (oder wenn {@code captionMap = null}), wird auf den
   * {@linkplain SmallReservoirModel#RESOURCE Resource-Provider} des Modells zurueckgegriffen.
   * @param captionMap Map in der nach den Ressourcen-Beschreibungen gesucht wird (kann
   *                   {@code null} sein)
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    ResourceProvider rp = SmallReservoirModel.RESOURCE;
    DHM_DESC            = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.DHM");
    DAM_LOC_DESC        = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.Dam.Loc");
    DAM_HEIGHT_DESC     = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.Dam.Height");
    INFILT_DESC         = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.Infilt");
    EVAPO_DESC          = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.Evapo");
    STEP_SIZE_DESC      = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.In.StepSize");
    OUT_AREA_SIZE_DESC  = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.Out.Area.Size");
    OUT_AREA_VOL_DESC   = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.Out.Area.Vol");
    OUT_AREA_DESC       = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.Out.Area.Grid");
    TMP_DHM_DESC        = AbstractXuluModel.determineCaption(captionMap, rp, "MRDesc.Temp.DHM");
    
    resource[DHM_NO].setDesc(DHM_DESC);
    resource[DAM_LOC_NO].setDesc(DAM_LOC_DESC);
    resource[DAM_HEIGHT_NO].setDesc(DAM_HEIGHT_DESC);
    resource[INFILT_NO].setDesc(INFILT_DESC);
    resource[EVAPO_NO].setDesc(EVAPO_DESC);
    resource[STEP_SIZE_NO].setDesc(STEP_SIZE_DESC);
    resource[OUT_AREA_SIZE_NO].setDesc(OUT_AREA_SIZE_DESC);
    resource[OUT_AREA_VOL_NO].setDesc(OUT_AREA_VOL_DESC);
    resource[OUT_AREA_NO].setDesc(OUT_AREA_DESC);
    resource[TMP_DHM_NO].setDesc(TMP_DHM_DESC);
  }

  public void checkAndError() {
    super.checkAndError();

    // Pruefen, ob FeatureCollection "eine Linie" enthaelt
    SingleFeatureCollection sfc = (SingleFeatureCollection)getResource(1).getData();
    if ( sfc != null && sfc.getFeatureCollection().size() > 0 ) {
      FeatureCollection fc = sfc.getFeatureCollection();
      FeatureIterator   fi = fc.features();
      Geometry          g  = null;
      if ( (g = fi.next().getDefaultGeometry()) == null ||
          !( g instanceof LineString || g instanceof MultiLineString ) )
       throwResourceException(2, SmallReservoirModel.RESOURCE.getString("Error.In.Dam.Loc"));
    } else
    	throwResourceException(2, SmallReservoirModel.RESOURCE.getString("Error.In.Dam.Loc"));

    // Ausgabe-Raster muss mit DHM uebereinstimmen
    WritableGrid dhm = (WritableGrid)((ScalarProperty)getResource(0).getData()).getOneTimeReadAccess().getValue();
    WritableGrid floodedArea = ((MultiGrid)getResource(8).getData()).getTemplateGrid();
    AbstractWritableGrid.compareStructure(
    		dhm, SmallReservoirModel.RESOURCE.getString("Error.In.DHM.Grid"),
    		floodedArea, SmallReservoirModel.RESOURCE.getString("Error.Out.Area.Grid"),
    		false
    );

    // Temp-Raster fuer Staumauer-DHM muss mit DHM uebereinstimmen
    WritableGrid tempDHM = (WritableGrid)((ScalarProperty)getResource(9).getData()).getOneTimeReadAccess().getValue();
    AbstractWritableGrid.compareStructure(
    		dhm, SmallReservoirModel.RESOURCE.getString("Error.In.DHM.Grid"),
    		tempDHM, SmallReservoirModel.RESOURCE.getString("Error.Temp.DHM.Grid"),
    		false
    );
  }

  private void throwResourceException(int resIdx, String errDesc) {
    throw new XuluDataException(
    		SmallReservoirModel.RESOURCE.getString(
    				"Error.Global.Mess",
            getResource(resIdx).getDesc(),
            errDesc
        )
    );
  }
}
