/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.clue;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.MatrixProperty;

import schmitzm.temp.BaseTypeUtil;
import schmitzm.lang.ResourceProvider;

import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.AbstractModelResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.data.XuluDataException;

import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import edu.bonn.xulu.plugin.model.ca.fire.ImpetusFireCA;

import java.util.Locale;
import java.util.Map;

import edu.bonn.xulu.appl.XuluConstants;
import edu.bonn.xulu.XuluModellingPlatform;

/**
 * Diese Klasse verwaltet die Ressourcen fuer das {@linkplain ClueModel Clue-Modell}
 * und prueft diese auf Korrektheit.<br>
 * Die Art der benoetigten Ressourcen ist der {@linkplain ClueModel Model-Doku}
 * zu entnehmen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ClueModelContentManager extends AbstractModelContentManager {
  /** Ressource: Simulation step count */
  public  static final int    MODELSTEPS  = 0; // Nummer der Ressource
  private static final int    MODELSTEPS_MIN  = 1;
  private static final int    MODELSTEPS_MAX  = 50;
  private              String MODELSTEPS_DESC = "";
  /** Ressource: Type of tolerance expression for demand compliance */
  public  static final int    TOLERANCETYPE      = 1; // Nummer der Ressource
  private              String TOLERANCETYPE_DESC = "";
  /** Ressource: Tolerance for demand compliance for each lu type */
  public  static final int    TOLERANCESINGLE      = 2; // Nummer der Ressource
  private              String TOLERANCESINGLE_DESC = "";
  /** Ressource: Average tolerance for demand compliance */
  public  static final int    TOLERANCEAVERAGE      = 3; // Nummer der Ressource
  private              String TOLERANCEAVERAGE_DESC = "";
  /** Ressource: Toleranz fuer Bedarfe */
  public  static final int    CONVFACTQM      = 4; // Nummer der Ressource
  private              String CONVFACTQM_DESC = "";
  /** Ressource: Base scenario */
  public  static final int    BASESCENARIO      = 5; // Nummer der Ressource
  private              String BASESCENARIO_DESC = "";
  /** Ressource: Land use history of base scenario */
  public  static final int    LUHISTORY      = 6; // Nummer der Ressource
  private              String LUHISTORY_DESC = "";
  /** Ressource: LU-Conversion-Matrix */
  public  static final int    LUCONV      = 7; // Nummer der Ressource
  private              String LUCONV_DESC = "";
  /** Ressource: LU-Conversion-Matrix */
  public  static final int    LUCONVDEV   = 8; // Nummer der Ressource
  private              String LUCONVDEV_DESC = "";
  /** Ressource: Conversion elasticity */
  public  static final int    ELASTICITY      = 9; // Nummer der Ressource
  private              String ELASTICITY_DESC = "";
  // Ressource: Area restrictions
  public  static final int    AREARESTR      = 10; // Nummer der Ressource
  private              String AREARESTR_DESC = "";
  /** Ressource: Demand scenario */
  public  static final int    DEMAND      = 11; // Nummer der Ressource
  private              String DEMAND_DESC = "";
  /** Ressource: Static driving forces */
  public  static final int    STATDF      = 12; // Nummer der Ressource
  private              String STATDF_DESC = "";
  /** Ressource: List of the dynamic driving force types */
  public  static final int    DYNDFTYPES      = 13; // Nummer der Ressource
  private              String DYNDFTYPES_DESC = "";
  /** Ressource: Dynamic driving forces */
  public  static final int    DYNDF      = 14; // Nummer der Ressource
  private              String DYNDF_DESC = "";
  /** Ressource: Regression results */
  public  static final int    REGRESSION      = 15; // Nummer der Ressource
  private              String REGRESSION_DESC = "";
  /** Ressource: Gewichtung der Neighborhood-Regression results */
  public  static final int    NBHWEIGHTS     = 16; // Nummer der Ressource
  private              String NBHWEIGHTS_DESC = "";
  /** Ressource: Neighborhood-Settings */
  public  static final int    NBHSETTINGS      = 17; // Nummer der Ressource
  private              String NBHSETTINGS_DESC = "";
  /** Ressource: Neighborhood-Regression results */
  public  static final int    NBHREGRESSION      = 18; // Nummer der Ressource
  private              String NBHREGRESSION_DESC = "";
  /** Ressource: RasterList for "Location specific preference addition" */
  public  static final int    LSPAPROB = 19; // Nummer der Ressource
  private              String LSPAPROB_DESC = "";
  /** Ressource: Factors for "Location specific preference addition" */
  public  static final int    LSPAFACT = 20; // Nummer der Ressource
  private              String LSPAFACT_DESC = "";

  /** Ressource: Temp-Raster for actual LU History */
  public  static final int    ACTLUHISTORY      = 21; // Nummer der Ressource
  private              String ACTLUHISTORY_DESC = "";
  /** Ressource: Temp-Raster for LU Probabilities */
  public  static final int    LUPROB      = 22; // Nummer der Ressource
  private              String LUPROB_DESC = "";
  /** Ressource: Temp-Raster for LU Probabilities */
  public  static final int    NBHPROB      = 23; // Nummer der Ressource
  private              String NBHPROB_DESC = "";
  /** Ressource: Temp-List for iteration variables */
  public  static final int    ITERVAR      = 24; // Nummer der Ressource
  private              String ITERVAR_DESC = "";

  /** Ressource: Temp-Raster for actual LU */
  public  static final int    ACTLU      = 25; // Nummer der Ressource
  private              String ACTLU_DESC = "";
  /** Ressource: Out-Rasters for time step results */
  public  static final int    STEPRESULTS      = 26; // Nummer der Ressource
  private              String STEPRESULTS_DESC = "";
  /** Ressource: Temp-RasterList for total probabilities */
  public  static final int    TOTALPROB      = 27; // Nummer der Ressource
  private              String TOTALPROB_DESC = "";

  /**
   * Erzeugt einen neuen ContentManager fuer das {@linkplain ClueModel ClueModel}.
   */
  public ClueModelContentManager() {
    super(28);
    // Ressourcen festlegen
    resource[MODELSTEPS]       = new ModelStepResource(ModelResource.CATEGORY_INPUT,MODELSTEPS_DESC,MODELSTEPS_MIN,MODELSTEPS_MAX);
    resource[TOLERANCETYPE]    = new ToleranceTypeResource(ModelResource.CATEGORY_INPUT,TOLERANCETYPE_DESC);
    resource[TOLERANCESINGLE]  = new SingleToleranceResource(ModelResource.CATEGORY_INPUT,TOLERANCESINGLE_DESC);
    resource[TOLERANCEAVERAGE] = new AverageToleranceResource(ModelResource.CATEGORY_INPUT,TOLERANCEAVERAGE_DESC);
    resource[CONVFACTQM]       = new ConversionFactorQMResource(ModelResource.CATEGORY_INPUT,CONVFACTQM_DESC);
    resource[BASESCENARIO]     = new InitialScenarioResource(ModelResource.CATEGORY_INPUT,BASESCENARIO_DESC);
    resource[LUHISTORY]        = new LUHistoryResource(ModelResource.CATEGORY_INPUT,LUHISTORY_DESC);
    resource[LUCONV]           = new LUConversionMatrixResource(ModelResource.CATEGORY_INPUT,LUCONV_DESC,false);
    resource[LUCONVDEV]        = new LUConversionMatrixResource(ModelResource.CATEGORY_INPUT,LUCONVDEV_DESC,true);
    resource[ELASTICITY]       = new ConversionElasticityResource(ModelResource.CATEGORY_INPUT,ELASTICITY_DESC);
    resource[AREARESTR]        = new AreaRestrictionsResource(ModelResource.CATEGORY_INPUT,AREARESTR_DESC);
    resource[DEMAND]           = new DemandScenarioResource(ModelResource.CATEGORY_INPUT,DEMAND_DESC);
    resource[STATDF]           = new StaticDrivingForcesResource(ModelResource.CATEGORY_INPUT,STATDF_DESC);
    resource[DYNDFTYPES]       = new DynamicDrivingForceTypesResource(ModelResource.CATEGORY_INPUT,DYNDFTYPES_DESC);
    resource[DYNDF]            = new DynamicDrivingForcesResource(ModelResource.CATEGORY_INPUT,DYNDF_DESC);
    resource[REGRESSION]       = new RegressionResultsResource(ModelResource.CATEGORY_INPUT,REGRESSION_DESC);
    resource[NBHWEIGHTS]       = new NeighborhoodWeightsResource(ModelResource.CATEGORY_INPUT,NBHWEIGHTS_DESC);
    resource[NBHSETTINGS]      = new NeighborhoodSettingsResource(ModelResource.CATEGORY_INPUT,NBHSETTINGS_DESC);
    resource[NBHREGRESSION]    = new NeighborhoodRegressionResultsResource(ModelResource.CATEGORY_INPUT,NBHREGRESSION_DESC);
    resource[ACTLU]            = new ActualLUResource(ModelResource.CATEGORY_OUTPUT,ACTLU_DESC);
    resource[ACTLUHISTORY]     = new ActualLUHistoryResource(ModelResource.CATEGORY_TEMP,ACTLUHISTORY_DESC);
    resource[LUPROB]           = new LUProbabilitiesResource(ModelResource.CATEGORY_TEMP,LUPROB_DESC);
    resource[NBHPROB]          = new NeighborhoodProbabilitiesResource(ModelResource.CATEGORY_TEMP,NBHPROB_DESC);
    resource[LSPAPROB]         = new LSPAProbabilitiesResource(ModelResource.CATEGORY_INPUT,LSPAPROB_DESC);
    resource[LSPAFACT]         = new LSPAFactorsResource(ModelResource.CATEGORY_INPUT,LSPAFACT_DESC);
    resource[TOTALPROB]        = new TotalProbabilitiesResource(ModelResource.CATEGORY_TEMP,TOTALPROB_DESC);
    resource[ITERVAR]          = new IterationVariablesResource(ModelResource.CATEGORY_TEMP,ITERVAR_DESC);
    resource[STEPRESULTS]      = new StepResultsResource(ModelResource.CATEGORY_OUTPUT,STEPRESULTS_DESC);
    // Ressource-Beschreibungen lokalisieren
    resetCaptions(null);
  }

  /**
   * Belegt die Ressourcen-Beschreibungen neu. Sind in der uebergebenen Map fuer die Keys
   * keine Werte hinterlegt (oder wenn {@code captionMap = null}), wird auf den
   * {@linkplain ClueModel#CLUE_RES Resource-Provider} des Modells zurueckgegriffen.
   * @param captionMap Map in der nach den Ressourcen-Beschreibungen gesucht wird (kann
   *                   {@code null} sein)
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    MODELSTEPS_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.ModelSteps",MODELSTEPS_MIN,MODELSTEPS_MAX);
    TOLERANCETYPE_DESC    = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Tol.Type",ClueModel.TOLTYPE_PCT,ClueModel.TOLTYPE_ABS);
    TOLERANCESINGLE_DESC  = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Tol.Single");
    TOLERANCEAVERAGE_DESC = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Tol.Avg");
    CONVFACTQM_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.qmConvFactor");
    BASESCENARIO_DESC     = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LUC.Scenario");
    LUHISTORY_DESC        = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LUC.History");
    LUCONV_DESC           = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LUC.ConvMatrix");
    LUCONVDEV_DESC        = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LUC.ConvDev");
    ELASTICITY_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LUC.ConvElas");
    AREARESTR_DESC        = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.AreaRestr");
    DEMAND_DESC           = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Demand.Scenario");
    STATDF_DESC           = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.DF.Static.Grids");
    DYNDFTYPES_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.DF.Dynamic.Types");
    DYNDF_DESC            = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.DF.Dynamic.Grids");
    REGRESSION_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Regression");
    NBHWEIGHTS_DESC       = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Nbh.Weight");
    NBHSETTINGS_DESC      = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Nbh.Setting");
    NBHREGRESSION_DESC    = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Nbh.Regr");
    LSPAPROB_DESC         = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LSPA.Prob");
    LSPAFACT_DESC         = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.LSPA.Fact");
    ACTLUHISTORY_DESC     = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.LucHist");
    LUPROB_DESC           = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.LucProb");
    NBHPROB_DESC          = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.NbhProb");
    ITERVAR_DESC          = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.IterVar");
    ACTLU_DESC            = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.Luc");
    STEPRESULTS_DESC      = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Out.StepRes");
    TOTALPROB_DESC        = ClueModel.determineCaption(captionMap, ClueModel.CLUE_RES, "MRDesc.Temp.TotalProb");

    resource[MODELSTEPS].setDesc(MODELSTEPS_DESC);
    resource[TOLERANCETYPE].setDesc(TOLERANCETYPE_DESC);
    resource[TOLERANCESINGLE].setDesc(TOLERANCESINGLE_DESC);
    resource[TOLERANCEAVERAGE].setDesc(TOLERANCEAVERAGE_DESC);
    resource[CONVFACTQM].setDesc(CONVFACTQM_DESC);
    resource[BASESCENARIO].setDesc(BASESCENARIO_DESC);
    resource[LUHISTORY].setDesc(LUHISTORY_DESC);
    resource[LUCONV].setDesc(LUCONV_DESC);
    resource[LUCONVDEV].setDesc(LUCONVDEV_DESC);
    resource[ELASTICITY].setDesc(ELASTICITY_DESC);
    resource[AREARESTR].setDesc(AREARESTR_DESC);
    resource[DEMAND].setDesc(DEMAND_DESC);
    resource[STATDF].setDesc(STATDF_DESC);
    resource[DYNDFTYPES].setDesc(DYNDFTYPES_DESC);
    resource[DYNDF].setDesc(DYNDF_DESC);
    resource[REGRESSION].setDesc(REGRESSION_DESC);
    resource[NBHWEIGHTS].setDesc(NBHWEIGHTS_DESC);
    resource[NBHSETTINGS].setDesc(NBHSETTINGS_DESC);
    resource[NBHREGRESSION].setDesc(NBHREGRESSION_DESC);
    resource[ACTLU].setDesc(ACTLU_DESC);
    resource[ACTLUHISTORY].setDesc(ACTLUHISTORY_DESC);
    resource[LUPROB].setDesc(LUPROB_DESC);
    resource[NBHPROB].setDesc(NBHPROB_DESC);
    resource[LSPAPROB].setDesc(LSPAPROB_DESC);
    resource[LSPAFACT].setDesc(LSPAFACT_DESC);
    resource[TOTALPROB].setDesc(TOTALPROB_DESC);
    resource[ITERVAR].setDesc(ITERVAR_DESC);
    resource[STEPRESULTS].setDesc(STEPRESULTS_DESC);
  }

  /**
   * Prueft alle Modell-Ressourcen auf Korrektheit.
   * Neben den lokalen Bedingungen an die Ressourcen (die bereits durch
   * die Methode der Oberklasse gechecked werden), prueft diese Methode, ob
   * <ul>
   * <li>ob die Raster alle dieselbe Groesse haben, wie das Base Scenario</li>
   * <li>die MultiGrids genug Raster enthalten</li>
   * <li>die Matrizen die passenden Groessen haben</li>
   * </ul>
   */
  public void checkAndError() {
    super.checkAndError();

    // Parameter ermitteln, die fuer die Konsistenz relevant sind
    int n = getLUConversionMatrix().getSize(0);
    int m = getStaticDrivingForces().getGridCount();
    int t = getStepCount().getOneTimeReadAccess().getValueAsInt();
    WritableGrid baseGrid = (WritableGrid)getBaseScenario().getOneTimeReadAccess().getValue();
    WritableGrid testGrid = null;

    // Wenn angegeben, muss Abweichungsmatrix die selbe groesse habem, wie
    // die LU-Conversion-Matrix
    MatrixProperty luConvDev = getLUConversionDeviationMatrix();
    if ( luConvDev != null && luConvDev.getSize(0) < n )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.LUC.ConvDev.Col",n) );
    if ( luConvDev != null && luConvDev.getSize(1) < n )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.LUC.ConvDev.Row",n) );

    // Conversion Elasticity muss n Eintraege haben
    int ceCount = getConversionElasticity().getOneTimeReadAccess().getCount();
    if ( ceCount < n )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.LUC.ConvElas.Size",n) );
    // Area Restriction Raster muss dem Base Scenario entsprechen
    testGrid = (WritableGrid)getAreaRestrictions().getOneTimeReadAccess().getValue();
    checkRasterDim(baseGrid,BASESCENARIO_DESC,testGrid,AREARESTR_DESC);

    // Demand Scenario muss (n,t)-Matrix sein
    int[] dsSize = getDemandScenario().getSize();
    if ( dsSize[0] < n )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Demand.Scenario.Col",n) );
    if ( dsSize[1] < t )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Demand.Scenario.Row",t) );

    // Driving Forces muessen m Raster, die dem Base Scenario entsprechen,
    MultiGrid sdfGrid = getStaticDrivingForces();
    if ( sdfGrid.getGridCount() < m )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Static.Grids",m) );
    // erstes befuelltes Raster checken
    checkRasterDim(baseGrid,BASESCENARIO_DESC,sdfGrid.getFirstNotNullGrid(),STATDF_DESC);

    // Liste der Dynamischen DF darf nur m Eintraege haben und Werte zwischen 0 und m enthalten
    int ddfCount = 0;
    if ( getDynamicDrivingForceTypes() != null ) {
      ListPropertyReadAccess ddf = getDynamicDrivingForceTypes().getReadAccess(this);
      ddfCount = ddf.getCount();
      if ( ddfCount > m )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Dynamic.Types.Size",m) );
      for (int i=0; i<ddfCount; i++)
        if ( ddf.getValueAsInt(i) < 0 || ddf.getValueAsInt(i) >= m )
          throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Dynamic.Types.Value",0,m-1) );
      ddf.release();
    }

    // Werden dyn. DF verwendet, muessen auch Raster angegeben werden
    MultiGrid ddfGrids = getDynamicDrivingForces();
    if ( ddfCount > 0 && ddfGrids == null )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Dynamic.Grids.Null",ClueModel.CLUE_RES.getString("MRDesc.DF.Dynamic.Grids")) );
    // Es muessen mind. t Raster fuer jeden dyn. DF angegeben werden
    if ( ddfCount > 0 && ddfGrids.getGridCount() < ddfCount*t )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Dynamic.Grids.Size1",ddfCount*t) );
    // Fuer jeden DF muss dieselbe Anzahl an Grids enthalten sein
    if ( ddfCount > 0 && ddfGrids.getGridCount() % ddfCount != 0 )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.DF.Dynamic.Grids.Size2") );

    // Regression Results muss (m+1,n)-Matrix sein
    int[] rrSize = getRegressionResults().getSize();
    if ( rrSize[0] < m+1 )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Regression.Col",m+1) );
    if ( rrSize[1] < n )
      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Regression.Row",n) );

    // Wenn angegeben, muss die Liste der Neighborhood Weights mind. n Eintraege haben
    ListProperty nbhWeights = getNeighborhoodWeights();
    if ( nbhWeights != null ) {
      if ( nbhWeights.getOneTimeReadAccess().getCount() < n )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Weight.Size",n) );
    }

    // Wenn angegeben, muss die Matrix der Neighborhood Settings in der 3. Dimension
    // die Groessen n haben
    if ( nbhWeights != null ) {
      if ( getNeighborhoodSettings() == null )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Setting.Null") );
      int[] nbSize = getNeighborhoodSettings().getSize();
      if ( nbSize[0]%2 == 0 || nbSize[1]%2 == 0 )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Setting.Size1") );
      if ( getNeighborhoodSettings().getSize(2) < n )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Setting.Size2",n) );

    }

    // Wenn angegeben, muss die Matrix der Neighborhood Regression eine
    // (n+1,n)-Matrix sein
    if ( nbhWeights != null ) {
      if ( getNeighborhoodRegression() == null )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Regr.Null") );
      int[] nbSize = getNeighborhoodRegression().getSize();
      if ( nbSize[0] < n+1 )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Regr.Col",n+1) );
      if ( nbSize[1] < n )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Nbh.Regr.Row",n) );
    }

    // Raster fuer die aktuelle LU muss dem Base Scenario entsprechen
    testGrid = (WritableGrid)getActLUResult().getOneTimeReadAccess().getValue();
    checkRasterDim(baseGrid,BASESCENARIO_DESC,testGrid,ACTLU_DESC);

    // Raster fuer die aktuelle LU History muss dem Base Scenario entsprechen
    testGrid = (WritableGrid)getActLUHistory().getOneTimeReadAccess().getValue();
    checkRasterDim(baseGrid,BASESCENARIO_DESC,testGrid,ACTLUHISTORY_DESC);

//    // MultiGrid fuer die LU-Wahrscheinlichkeiten muss n Raster enthalten,
//    // die dem Base Scenario entsprechen
//    MultiGrid luProbGrids = this.getLUProbabilities();
//    if ( luProbGrids.getGridCount() < n )
//      throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Temp.LucProb.Size",n) );
//    checkRasterDim(baseGrid,BASESCENARIO_DESC,luProbGrids.getGrid(0),LUPROB_DESC);
    // MultiGrid fuer die LU-Wahrscheinlichkeiten muss dem Base Scenario entsprechen
    // Die n Inhalt-Grids werden bei der Modell-Initialisierung dynamisch erzeugt!
    MultiGrid luProbGrids = this.getLUProbabilities();
    checkRasterDim(baseGrid,BASESCENARIO_DESC,luProbGrids.getTemplateGrid(),LUPROB_DESC);

//    // MultiGrid fuer die Nachbarschafts-Wahrscheinlichkeiten muss n Raster enthalten,
//    // die dem Base Scenario entsprechen
//    if ( nbhWeights != null ) {
//      MultiGrid nbhProbGrids = this.getNeighborhoodProbabilities();
//      if ( nbhProbGrids == null )
//        throw new XuluDataException("Temp. raster for neighborhood probabilities must be specified...");
//      if ( nbhProbGrids.getGridCount() < n )
//        throw new XuluDataException("MultiGrid for neighborhood probabilities must contain at leat "+n+" grids (one for each land use type)...");
//      checkRasterDim(baseGrid,BASESCENARIO_DESC,nbhProbGrids.getGrid(0),LUPROB_DESC);
//    }
    // MultiGrid fuer die Nachbarschafts-Wahrscheinlichkeiten muss dem Base Scenario entsprechen
    // Die n Inhalt-Grids werden bei der Modell-Initialisierung dynamisch erzeugt!
    if ( nbhWeights != null ) {
      MultiGrid nbhProbGrids = this.getNeighborhoodProbabilities();
      if ( nbhProbGrids == null )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Temp.NbhProb.Null") );
      checkRasterDim(baseGrid,BASESCENARIO_DESC,nbhProbGrids.getTemplateGrid(),LUPROB_DESC);
    }

    // Werden die LSPA-Raster verwendet, muss fuer jeden LUC-Typ ein Raster
    // hinterlegt sein
    MultiGrid lspaGrids = getLSPAProbabilities();
    if ( lspaGrids != null ) {
      if ( lspaGrids.getGridCount() < n )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.LSPA.Prob.Size",n) );
      checkRasterDim(baseGrid,BASESCENARIO_DESC,lspaGrids.getTemplateGrid(),LSPAPROB_DESC);
    }
    // Werden die LSPA-Faktoren verwendet, muss fuer jeden LUC-Typ ein Faktor
    // hinterlegt sein
    ListProperty lspaFact = getLSPAFactors();
    if ( lspaGrids != null && lspaFact != null ) {
      if ( lspaFact.getOneTimeReadAccess().getCount() < n )
        throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.LSPA.Fact.Size",n) );
    }

    // MultiGrid fuer die Gesamt-Wahrscheinlichkeiten muss dem Base Scenario entsprechen
    // Die n Inhalt-Grids werden bei der Modell-Initialisierung dynamisch erzeugt!
    MultiGrid totProbGrids = this.getNeighborhoodProbabilities();
    if ( totProbGrids != null ) {
      checkRasterDim(baseGrid,BASESCENARIO_DESC,totProbGrids.getTemplateGrid(),TOTALPROB_DESC);
    }

//    // Wenn angegeben, muss das MultiGrid fuer die Zeitschritt-Ergebnisse
//    // t Raster enthalten, die dem Base Scenario entsprechen
//    MultiGrid resultGrids = this.getStepLUResults();
//    if (resultGrids != null) {
//      if ( resultGrids.getGridCount() < t )
//        throw new XuluDataException("MultiGrid for time step results must contain at leat "+t+" grids (one for each time step)...");
//      checkRasterDim(baseGrid, BASESCENARIO_DESC, resultGrids.getGrid(0),STEPRESULTS_DESC);
//    }
    // Das MultiGrid fuer die Zeitschritt-Ergebnisse dem Base Scenario entsprechen
    // Die t Inhalt-Grids werden waehrend des Modellablaufs dynamisch erzeugt!
    MultiGrid resultGrids = this.getStepLUResults();
    checkRasterDim(baseGrid, BASESCENARIO_DESC, resultGrids.getTemplateGrid(),STEPRESULTS_DESC);
  }

  /**
   * Vergleicht zwei Raster auf gleiche Struktur (Hoehe, Breite, Zell-Hoehe,
   * Zell-Breite, Sample-Type).
   */
  private static void checkRasterDim(WritableGrid grid1, String desc1, WritableGrid grid2, String desc2) {
    if ( grid1.getWidth() != grid2.getWidth() )
      throw new XuluDataException(createCompareMessage(ClueModel.CLUE_RES.getString("Error.RasterDim.GridWidth"),desc1,desc2));
    if ( grid1.getHeight() != grid2.getHeight() )
      throw new XuluDataException(createCompareMessage(ClueModel.CLUE_RES.getString("Error.RasterDim.GridHeight"),desc1,desc2));
    if ( grid1.getCellWidth() != grid2.getCellWidth() )
      throw new XuluDataException(createCompareMessage(ClueModel.CLUE_RES.getString("Error.RasterDim.CellWidth"),desc1,desc2));
    if ( grid1.getCellHeight() != grid2.getCellHeight() )
      throw new XuluDataException(createCompareMessage(ClueModel.CLUE_RES.getString("Error.RasterDim.CellHeight"),desc1,desc2));
    if ( grid1.getSampleType() != grid2.getSampleType() )
      throw new XuluDataException(createCompareMessage(ClueModel.CLUE_RES.getString("Error.RasterDim.SampleType"),desc1,desc2));
  }

  /**
   * Erzeugt eine Fehlermeldung fuer eine bestimmte inkompatible Raster-Struktur.
   */
  private static String createCompareMessage(String compType, String objDesc1, String objDesc2) {
    return ClueModel.CLUE_RES.getString("Error.RasterDim",compType,objDesc1,objDesc2);
  }

  //////////////////////////////////////////////////////////////////////
  ///////   vereinfachende Zugriffsmethoden auf Ressourcen   ///////////
  //////////////////////////////////////////////////////////////////////
  /** Liefert Zugriff auf die Ressource "Time Steps". */
  public ScalarProperty getStepCount() {
    return (ScalarProperty)getResourceData(this.MODELSTEPS);
  }

  /** Liefert Zugriff auf die Ressource "Art der Abweichungsangaben". */
  public ScalarProperty getDemandToleranceType() {
    return (ScalarProperty)getResourceData(this.TOLERANCETYPE);
  }

  /** Liefert Zugriff auf die Ressource "Einzel-Abweichung von Bedarfen". */
  public ScalarProperty getSingleDemandTolerance() {
    return (ScalarProperty)getResourceData(this.TOLERANCESINGLE);
  }

  /** Liefert Zugriff auf die Ressource "Durchschnitts-Abweichung von Bedarfen". */
  public ScalarProperty getAverageDemandTolerance() {
    return (ScalarProperty)getResourceData(this.TOLERANCEAVERAGE);
  }


  /** Liefert Zugriff auf die Ressource "Umrechnungsfaktor Bedarfseinheit -> qm". */
  public ScalarProperty getConversionFactor() {
    return (ScalarProperty)getResourceData(this.CONVFACTQM);
  }

  /** Liefert Zugriff auf die Ressource "Base Scenario". */
  public ScalarProperty getBaseScenario() {
    return (ScalarProperty)getResourceData(this.BASESCENARIO);
  }

  /** Liefert Zugriff auf die Ressource "Land use History". */
  public ScalarProperty getLUHistory() {
    return (ScalarProperty)getResourceData(this.LUHISTORY);
  }

  /** Liefert Zugriff auf die Ressource "LU Conversion Matrix". */
  public MatrixProperty getLUConversionMatrix() {
    return (MatrixProperty)getResourceData(this.LUCONV);
  }

  /** Liefert Zugriff auf die Ressource "Deviation of LU Conversion Matrix". */
  public MatrixProperty getLUConversionDeviationMatrix() {
    return (MatrixProperty)getResourceData(this.LUCONVDEV);
  }

  /** Liefert Zugriff auf die Ressource "Conversion Elasticity". */
 public ListProperty getConversionElasticity() {
    return (ListProperty)getResourceData(this.ELASTICITY);
  }

  /** Liefert Zugriff auf die Ressource "Area Restrictions". */
  public ScalarProperty getAreaRestrictions() {
    return (ScalarProperty)getResourceData(this.AREARESTR);
  }

  /** Liefert Zugriff auf die Ressource "Demand Scenario". */
  public MatrixProperty getDemandScenario() {
    return (MatrixProperty)getResourceData(this.DEMAND);
  }

  /** Liefert Zugriff auf die Ressource "Static Driving Forces". */
  public MultiGrid getStaticDrivingForces() {
    return (MultiGrid)getResourceData(this.STATDF);
  }

  /** Liefert Zugriff auf die Ressource "Dynamic Driving Force Types". */
 public ListProperty getDynamicDrivingForceTypes() {
    return (ListProperty)getResourceData(this.DYNDFTYPES);
  }

  /** Liefert Zugriff auf die Ressource "Dynamic Driving Forces". */
 public MultiGrid getDynamicDrivingForces() {
    return (MultiGrid)getResourceData(this.DYNDF);
  }

  /** Liefert Zugriff auf die Ressource "Regression Results". */
  public MatrixProperty getRegressionResults() {
    return (MatrixProperty)getResourceData(this.REGRESSION);
  }

  /** Liefert Zugriff auf die Ressource "Neighborhood Weights". */
  public ListProperty getNeighborhoodWeights() {
    return (ListProperty)getResourceData(this.NBHWEIGHTS);
  }

  /** Liefert Zugriff auf die Ressource "Neighborhood Settings". */
  public MatrixProperty getNeighborhoodSettings() {
    return (MatrixProperty)getResourceData(this.NBHSETTINGS);
  }

  /** Liefert Zugriff auf die Ressource "Neighborhood Regression Results". */
  public MatrixProperty getNeighborhoodRegression() {
    return (MatrixProperty)getResourceData(this.NBHREGRESSION);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "Aktuelle Landnutzung". */
  public ScalarProperty getActLUResult() {
    return (ScalarProperty)getResourceData(this.ACTLU);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "LU-Wahrscheinlichkeiten". */
  public MultiGrid getLUProbabilities() {
    return (MultiGrid)getResourceData(this.LUPROB);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "Nachbarschafts-Wahrscheinlichkeiten". */
  public MultiGrid getNeighborhoodProbabilities() {
    return (MultiGrid)getResourceData(this.NBHPROB);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "Iterationsvariablen". */

  public ListProperty getIterationVariables() {
//    throw new UnsupportedOperationException("Iteration variables are no longer stored in a resource!");
    return (ListProperty)getResourceData(this.ITERVAR);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "Gesamt-Wahrscheinlichkeiten". */
  public MultiGrid getTotalProbabilities() {
    return (MultiGrid)getResourceData(this.TOTALPROB);
  }

  /** Liefert Zugriff auf die Ressource "Location specific preferece addition". */
  public MultiGrid getLSPAProbabilities() {
    return (MultiGrid)getResourceData(this.LSPAPROB);
  }

  /** Liefert Zugriff auf die Ressource "Factors for the location specific preferece addition". */
  public ListProperty getLSPAFactors() {
    return (ListProperty)getResourceData(this.LSPAFACT);
  }

  /** Liefert Zugriff auf die Ressource "Aktuelle Land use History". */
  public ScalarProperty getActLUHistory() {
    return (ScalarProperty)getResourceData(this.ACTLUHISTORY);
  }

  /** Liefert Zugriff auf die Ausgabe-Ressource "LU zu jedem Zeitschritt". */
  public MultiGrid getStepLUResults() {
    return (MultiGrid)getResourceData(this.STEPRESULTS);
  }

  //////////////////////////////////////////////////////////////////////
  ///////////   Klassen fuer die benoetigten Ressourcen   //////////////
  //////////////////////////////////////////////////////////////////////
  /**
   * Ressource fuer Simulationsschritte.
   */
  private class ModelStepResource extends ValuePropertyResource {
    private int stepMin = 0;
    private int stepMax = 0;

    public ModelStepResource(int cat,String desc, int stepMin, int stepMax) {
      super(cat,desc,ScalarProperty.class,Number.class,false);
      this.stepMin = stepMin;
      this.stepMax = stepMax;
    }

    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // muss als Datentyp Long oder Integer haben
      if ( !BaseTypeUtil.isNumeric(((ScalarProperty)data).getType()) ||
            BaseTypeUtil.isDecimal(((ScalarProperty)data).getType()) )
        throwDataException( ClueModel.CLUE_RES.getString("Error.ModelSteps.Type") );
    }

    public void checkAndError(Object data) {
      super.checkAndError(data);
      // muss zwischen 1 und 20 liegen
      checkRangeAndError(this,((ScalarProperty)data).getOneTimeReadAccess().getValueAsLong(),stepMin,stepMax);
    }
  }

  /**
   * Ressource fuer Art der Toleranz-Angaben (prozentual, absolut).
   */
  private class ToleranceTypeResource extends ValuePropertyResource {
    public ToleranceTypeResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,Integer.class,false);
    }
    public void checkAndError(Object data) {
      super.checkAndError(data);
      // Flag muss 0 (prozentual) oder 1 (absolut) sein
      checkRangeAndError(this,((ScalarProperty)data).getOneTimeReadAccess().getValueAsInt(),0,1);
    }
  }

  /**
   * Ressource fuer Toleranz der einzelnen Bedarfe.
   */
  private class SingleToleranceResource extends ValuePropertyResource {
    public SingleToleranceResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,Number.class,false);
    }
    public void checkAndError(Object data) {
      super.checkAndError(data);
      // muss groesser 0 sein
      checkRangeAndError(this,((ScalarProperty)data).getOneTimeReadAccess().getValueAsDouble(),0,Double.MAX_VALUE);
    }
  }

  /**
   * Ressource fuer Durchschnitts-Toleranz der einzelnen Bedarfe.
   */
  private class AverageToleranceResource extends ValuePropertyResource {
    public AverageToleranceResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,Number.class,false);
    }
    public void checkAndError(Object data) {
      super.checkAndError(data);
      // muss groesser 0 sein
      checkRangeAndError(this,((ScalarProperty)data).getOneTimeReadAccess().getValueAsDouble(),0,Double.MAX_VALUE);
    }
  }

  /**
   * Ressource fuer Umrechnungsfaktor Bedarfseinheit >> Quadratmeter.
   */
  private class ConversionFactorQMResource extends ValuePropertyResource {
    public ConversionFactorQMResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,Double.class,false);
    }
  }

  /**
   * Ressource fuer Ausgangsszenario.
   */
  private class InitialScenarioResource extends ValuePropertyResource {
    public InitialScenarioResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,WritableGrid.class,false);
    }
  }

  /**
   * Ressource fuer LU-History.
   */
  private class LUHistoryResource extends ValuePropertyResource {
    public LUHistoryResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,WritableGrid.class,false);
    }
  }

  /**
   * Ressource fuer LU-Conversion-Matrix.
   */
  private class LUConversionMatrixResource extends ValuePropertyResource {
    public LUConversionMatrixResource(int cat,String desc,boolean nullable) {
      super(cat,desc,MatrixProperty.class,Integer.class,nullable);
    }
    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // Matrix muss mindestens 2-dimensional sein
      if ( data != null && ((MatrixProperty)data).getDimension() < 2 )
        throwDataException( ClueModel.CLUE_RES.getString("Error.Expected.Matrix2") );
    }
  }

  /**
   * Ressource fuer Conversion Elasticity.
   */
  private class ConversionElasticityResource extends ValuePropertyResource {
    public ConversionElasticityResource(int cat,String desc) {
      super(cat,desc,ListProperty.class,Double.class,false);
    }

    public void checkAndError(Object data) {
      super.checkAndError(data);
      // Alle Werte der Liste muessen zwischen 0 und 1 liegen
      ListPropertyReadAccess a = ((ListProperty)data).getReadAccess(this);
      for (int i=0; i<a.getCount(); i++) {
        double v = a.getValueAsDouble(i);
        if ( v < 0 || 1 < v ) {
          a.release();
          throwDataException( ClueModel.CLUE_RES.getString("Error.LUC.ConvElas.Value") );
        }
      }
      a.release();
    }
  }

  /**
   * Ressource fuer Area Restrictions.
   */
  private class AreaRestrictionsResource extends ValuePropertyResource {
    public AreaRestrictionsResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,WritableGrid.class,false);
    }
  }

  /**
   * Ressource fuer Demand Scenario.
   */
  private class DemandScenarioResource extends ValuePropertyResource {
    public DemandScenarioResource(int cat,String desc) {
      super(cat,desc,MatrixProperty.class,Integer.class,false);
    }

    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // Matrix muss mindestens 2-dimensional sein
      if ( ((MatrixProperty)data).getDimension() < 2 )
        throwDataException( ClueModel.CLUE_RES.getString("Error.Expected.Matrix2") );
    }
  }

  /**
   * Ressource fuer Static Driving Forces.
   */
  private class StaticDrivingForcesResource extends AbstractModelResource {
    public StaticDrivingForcesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,false);
    }

    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable)
        throwDataException("MultiGrid");
    }

    public void checkAndError(Object data) {
      checkTypeAndError(data);
    }
  }

  /**
   * Ressource fuer Dynamic Driving Force Types.
   */
  private class DynamicDrivingForceTypesResource extends ValuePropertyResource {
    public DynamicDrivingForceTypesResource(int cat,String desc) {
      super(cat,desc,ListProperty.class,Integer.class,true);
    }
  }

  /**
   * Ressource fuer Dynamic Driving Forces.
   */
  private class DynamicDrivingForcesResource extends AbstractModelResource {
    public DynamicDrivingForcesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,true);
    }

    public void checkTypeAndError(Object data) {
      if ( data != null && !(data instanceof MultiGrid) ||
           data == null && !nullable)
        throwDataException("MultiGrid");
    }

    public void checkAndError(Object data) {
      checkTypeAndError(data);
    }
  }

  /**
   * Ressource fuer Regression Results.
   */
  private class RegressionResultsResource extends ValuePropertyResource {
    public RegressionResultsResource(int cat,String desc) {
      super(cat,desc,MatrixProperty.class,Double.class,false);
    }
    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // Matrix muss mindestens 2-dimensional sein
      if ( ((MatrixProperty)data).getDimension() < 2 )
        throwDataException( ClueModel.CLUE_RES.getString("Error.Expected.Matrix2") );
    }
  }

  /**
   * Ressource fuer Neighborhood Weight.
   */
  private class NeighborhoodWeightsResource extends ValuePropertyResource {
    public NeighborhoodWeightsResource(int cat,String desc) {
      super(cat,desc,ListProperty.class,Double.class,true);
    }
  }

  /**
   * Ressource fuer Neighborhood Settings.
   */
  private class NeighborhoodSettingsResource extends ValuePropertyResource {
    public NeighborhoodSettingsResource(int cat,String desc) {
      super(cat,desc,MatrixProperty.class,Double.class,true);
    }
    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // Matrix muss mindestens 3-dimensional sein
      if ( data!=null && ((MatrixProperty)data).getDimension() < 3 )
        throwDataException( ClueModel.CLUE_RES.getString("Error.Expected.Matrix3") );
    }
  }

  /**
   * Ressource fuer Neighborhood Regression Results.
   */
  private class NeighborhoodRegressionResultsResource extends ValuePropertyResource {
    public NeighborhoodRegressionResultsResource(int cat,String desc) {
      super(cat,desc,MatrixProperty.class,Double.class,true);
    }
    public void checkTypeAndError(Object data) {
      super.checkTypeAndError(data);
      // Matrix muss mindestens 2-dimensional sein
      if ( data!=null && ((MatrixProperty)data).getDimension() < 2 )
        throwDataException( ClueModel.CLUE_RES.getString("Error.Expected.Matrix2") );
    }
  }

  /**
   * Ressource fuer die simulierte LU.
   */
  private class ActualLUResource extends ValuePropertyResource {
    public ActualLUResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,WritableGrid.class,false);
    }
  }

  /**
   * Ressource fuer aktuelle LU-History.
   */
  private class ActualLUHistoryResource extends ValuePropertyResource {
    public ActualLUHistoryResource(int cat,String desc) {
      super(cat,desc,ScalarProperty.class,WritableGrid.class,false);
    }
  }

  /**
   * Ressource fuer die errechneten LU-Wahrscheinlichkeiten.
   */
  private class LUProbabilitiesResource extends AbstractModelResource {
    public LUProbabilitiesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,false);
    }
    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable )
        throwDataException("MultiGrid");
    }
    public void checkAndError(Object object) {
      checkTypeAndError(object);
    }
  }

  /**
   * Ressource fuer die errechneten Nachbarschafts-Wahrscheinlichkeiten.
   */
  private class NeighborhoodProbabilitiesResource extends AbstractModelResource {
    public NeighborhoodProbabilitiesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,true);
    }
    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable )
        throwDataException("MultiGrid");
    }
    public void checkAndError(Object object) {
      checkTypeAndError(object);
    }
  }

  /**
   * Optionale Ressource fuer die "Location specific preference addition".
   */
  private class LSPAProbabilitiesResource extends AbstractModelResource {
    public LSPAProbabilitiesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,true);
    }
    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable )
        throwDataException("MultiGrid");
    }
    public void checkAndError(Object object) {
      checkTypeAndError(object);
    }
  }

  /**
   * Optionale Ressource fuer die Faktoren der "Location specific preference addition".
   */
  private class LSPAFactorsResource extends ValuePropertyResource {
    public LSPAFactorsResource(int cat,String desc) {
      super(cat,desc,ListProperty.class,Double.class,true);
    }
  }

  /**
   * Ressource fuer die errechneten Gesamt-Wahrscheinlichkeiten pro LUC-Typ.
   */
  private class TotalProbabilitiesResource extends AbstractModelResource {
    public TotalProbabilitiesResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,true);
    }
    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable )
        throwDataException("MultiGrid");
    }
    public void checkAndError(Object object) {
      checkTypeAndError(object);
    }
  }

  /**
   * Ressource fuer Iterationsvariablen.
   */
  private class IterationVariablesResource extends ValuePropertyResource {
    public IterationVariablesResource(int cat,String desc) {
      super(cat,desc,ListProperty.class,Float.class,false);
    }
  }

  /**
   * Ressource fuer die einzelnen Zeit-Schritt-Ergebnisse.
   */
  private class StepResultsResource extends AbstractModelResource {
    public StepResultsResource(int cat,String desc) {
      super(cat,desc,MultiGrid.class,false);
    }

    public void checkTypeAndError(Object data) {
      if ( data!=null && !(data instanceof MultiGrid) ||
           data==null && !nullable )
        throwDataException("MultiGrid");
    }

    public void checkAndError(Object object) {
      checkTypeAndError(object);
    }
  }

  /**
   * Prueft, ob ein Wert in einem bestimmten Bereich liegt. Wenn nicht, wird
   * fuer die angegebene Ressource eine Exception geworfen.
   */
  private static void checkRangeAndError(ModelResource r, double value, double min, double max) {
    if ( value < min ||  max < value )
     throw new XuluDataException( ClueModel.CLUE_RES.getString("Error.Resource.Not.Allowed",r.getDesc(),value));
  }

}
