/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.clue;

import java.util.Random;

import org.apache.log4j.Logger;
import org.apache.log4j.LogManager;

import schmitzm.lang.LangUtil;

import schmitzm.data.property.Access;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ValuePropertyType;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.WritableGrid;

import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.AbstractModelResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;

import edu.bonn.xulu.plugin.data.grid.MultiGrid;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt den Versuch einer Laufzeit-verbesserten Implementierung der
 * ersten <a href="http://www.dow.wau.nl/clue/" target="_blank">CLUE</a>-Modell-Implementierung
 * {@link ClueModel_Original} dar (CLUE = The <b>C</b>onversion of <b>L</b>and
 * <b>U</b>se and its <b>E</b>ffects).<br>
 * Der Algorithmus ist der gleiche, jedoch wurde versucht, insbesondere im Ablauf
 * der Iteration auf alle Operationen zu verzichten, die sich negative auf die
 * Laufzeit auswirken koennten:
 * <ul>
 *   <li>ueberfluessige (temporaere) Objekt-Erzeugungen</li>
 *   <li>Status-Ausgaben</li>
 *   <li>GUI</li>
 *   <li>Debugging in Dateien</li>
 *   <li>usw.</li>
 * </ul>
 * @see ClueModelContentManager
 * @see ClueModel
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ClueModel_Optimized extends AbstractStepModel {
  /** Konstante fuer die Toleranz-Art "prozentual". */
  public static final int TOLTYPE_PCT = 0;
  /** Konstante fuer die Toleranz-Art "absolut in Bedarf-Einheit". */
  public static final int TOLTYPE_ABS = 1;

  /** Speichert den ContentManager fuer das Modell.
   *  @see ClueModelContentManager */
  protected ClueModelContentManager contManager;

  private final Logger LOG = LogManager.getLogger(this.getClass().getName());

  //********** Lese/Schreibrechte, die fuer den gesamten ***********
  //********** Modellanlauf gehalten werden muessen      ***********
  private PropertyReadAccess  RA_baseLU       = null; // Base Scenario
  private PropertyReadAccess  RA_baseLUHist   = null; // Base LU History
  private PropertyReadAccess  RA_luConv       = null; // LU-Conversion-Matrix
  private PropertyReadAccess  RA_luElas       = null; // LU-Conversion-Elasticity
  private PropertyReadAccess  RA_areaRestr    = null; // Area Restrictions
  private PropertyReadAccess  RA_luDemand     = null; // LU-Bedarfe
  private PropertyReadAccess  RA_sdfGrids     = null; // Static Driving Forces
  private PropertyReadAccess  RA_ddfList      = null; // List of dynmic Driving Forces
  private PropertyReadAccess  RA_ddfGrids     = null; // Dynamic Driving Forces
  private PropertyReadAccess  RA_regrRes      = null; // Regression Results
  private PropertyReadAccess  RA_nbhWeights   = null; // Gewichtung der Nachbarschaft pro LU-Typ
  private PropertyReadAccess  RA_nbhSettings  = null; // Definition der Umgebung pro LU-Typ
  private PropertyReadAccess  RA_nbhRegrRes   = null; // Regression Results fuer Nachbarschaft
  private PropertyWriteAccess WA_outActLU     = null; // speichert die aktuell sim. LU
  private PropertyWriteAccess WA_outActLUHist = null; // speichert die aktuell sim. LU Historie
  private PropertyWriteAccess WA_outLUProb    = null; // speichert die LU-Wahrscheinlichkeiten
  private PropertyWriteAccess WA_outNbhProb   = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private PropertyWriteAccess WA_outIter      = null; // speichert die Iterationsvariablen
  private PropertyWriteAccess WA_outStepLU    = null; // speichert die LU fuer jeden Zeitpunkt

  //**************** Variablen mit denen gearbeitet wird *******************
  // Modell-Parameter
  private int    luCount    = 0;    // Anzahl an LU-Typen
  private int    dfCount    = 0;    // Anzahl an Driving Forces
  private int    ddfCount   = 0;    // Anzahl an dynamischen Driving Forces
  private int    areaWidth  = 0;    // Breite des Untersuchungsgebiets (Rasterzellen)
  private int    areaHeight = 0;    // Hoehe des Untersuchungsgebiets (Rasterzellen)
  private int    areaMinX   = 0;    // Start-Index des Rasters in X-Richtung
  private int    areaMinY   = 0;    // Start-Index des Rasters in Y-Richtung
  private int    areaMaxX   = 0;    // End-Index des Rasters in X-Richtung (EXKLUSIVE!)
  private int    areaMaxY   = 0;    // End-Index des Rasters in Y-Richtung (EXKLUSIVE!)
  private int    demandTolType    = 0;     // Art der Abweichungsangaben
  private float  demandTolSingle  = 0.0f;  // erlaubte Abweichung fuer jeden einzelnen LU-Typ vom Bedarf
  private float  demandTolAverage = 0.0f;  // erlaubte Durchschnitts-Abweichung fuer das Untersuchungsgebiet
  private float  demandConv = 0.0f;  // Umrechnungsfaktor Bedarfseingeit -> qm
  private float  cellSize   = 0f;    // Flaeche einer Rasterzelle (in Bedarfseinheit)
  // Eingabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            baseLU       = null; // Base Scenario
  private WritableGrid            baseLUHist   = null; // Base LU History
  // ---
  // Versuch der Performance-Steigerung (D):
  //       Zugriff auf LU-Elastizitaet nicht ueber Property-Liste,
  //       sondern ueber lokalen Array
  // -> Steigerung der Iteration um 11,4%
//  private ListPropertyReadAccess  luElas       = null; // LU-Conversion-Elasticity
  private float[]                 luElas       = null;
  // ---
  private WritableGrid            areaRestr    = null; // Area Restrictions
  // ---
  // Versuch der Performance-Steigerung (F):
  //       Zugriff auf Demand und LUC-Conversion nicht ueber Propertys,
  //       sondern ueber lokales Array-Matrizen
  // -> Steigerung der Iteration um 65% (Conv-Matrix ~11%, Demand ~52%)
//  private PropertyReadAccess      luDemand     = null; // LU-Bedarfe
//  private PropertyReadAccess      luConv       = null; // LU-Conversion-Matrix
  private float[][]               luDemand     = null;
  private int[][]                 luConv       = null;
  // ---
  // Versuch der Performance-Steigerung (C):
  //       Zugriff auf statische DF und dynamische DF nicht ueber Property-Liste,
  //       sondern ueber in lokalem Array hinterlegte Referenzen
  // -> Steigerung der LUC-W'keitsberechnung um 6,5 %
//  private MultiGrid               sdfGrids     = null; // Static Driving Forces
  private WritableGrid[]          sdfGrids     = null;
//  private ListPropertyReadAccess  ddfList      = null; // Dynamic Driving Force Types
//  private MultiGrid               ddfGrids     = null; // Dynamic Driving Forces
  private WritableGrid[]          ddfGrids     = null;
  // ---
  private int[]                   ddfStartIdx  = null; // Listenindex an dem ein dyn. DF beginnt
  // Versuch der Performance-Steigerung (G):
  //       Zugriff auf Statistik-Parameter nicht ueber Property-Matrix,
  //       sondern ueber lokalen Array
  // -> Steigerung der LUC-W'keitsberechnung um 6,2%
//  private PropertyReadAccess      regrRes      = null; // Statistik
  private float[][]               regrRes      = null;
  // ---

  // Versuch der Performance-Steigerung (H):
  //       Zugriff auf Statistik-Parameter nicht ueber Property-Matrix,
  //       sondern ueber lokale Arrays
  // -> Steigerung der LUC-W'keitsberechnung um 84%
  //    Steigerung der Iteration um 30%
//  private ListPropertyReadAccess  nbhWeights   = null; // Gewichtung der Nachbarschaft pro LU-Typ
//  private PropertyReadAccess      nbhSettings  = null; // Definition der Umgebung pro LU-Typ
//  private PropertyReadAccess      nbhRegrRes   = null; // Regression Results fuer Nachbarschaft
  private float[]                 nbhWeights   = null;
  private float[][][]             nbhSettings  = null;
  private float[][]               nbhRegrRes   = null;
  // ---
  private int                     nbhRadiusX   = 0;    // horizontaler Radius der betrachteten Nachbarschaft
  private int                     nbhRadiusY   = 0;    // vertikaler Radius der betrachteten Nachbarschaft
  private float[][]               luCover      = null; // LU-Bedeckung der Typen ueber die Zeit in Bedarfseinheit
  private float                   totCover     = 0;    // Gesamte Landbedeckungsflaeche (ohne NoData!)
  private Random                  rand         = null; // Zufallsgenerator
  private float[]                 iterSpeed    = null; // Geschwindigkeit fuer die ITER-Variablen
  // Ausgabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            outActLU     = null; // speichert die aktuell sim. LU
  private WritableGrid            outActLUHist = null; // speichert die aktuell sim. LU History
  // ---
  // Versuch der Performance-Steigerung (A):
  //       Zugriff auf LUC-Probs- und NBH-Probs-Raster nicht ueber Property-Liste,
  //       sondern ueber in lokalem Array hinterlegte Referenzen
  // -> Steigerung der Iteration um 43%
//  private MultiGrid               outLUProb    = null; // speichert die LU-Wahrscheinlichkeiten
//  private MultiGrid               outNbhProb   = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private WritableGrid[]            outLUProbGrid = null;
  private WritableGrid[]            outNbhProbGrid = null;
  //---
  // Versuch der Performance-Steigerung (E):
  //       Zugriff auf Iterations-Variablen ueber Property-Liste,
  //       sondern ueber lokalen Array
  // -> Steigerung der Iteration um 5,6%
//  private ListPropertyWriteAccess outIter      = null; // speichert die Iterationsvariablen
  private float[]                 outIter      = null;
  // ---
  private MultiGrid               outStepLU    = null; // speichert die LU fuer jeden Zeitpunkt

//  // Versuch der Performance-Steigerung (B):
//  //       LUC-Probs und NBH-Probs nicht in Raster-Objekten speichern,
//  //       sondern komplett in lokalem Array
//  // -> Steigerung der Iteration um 47%
//  private float[][][]             outLUProb  = null;
//  private float[][][]             outNbhProb = null;

  /**
   * Erzeugt eine neue Instanz des CLUE-Modells.
   */
  public ClueModel_Optimized() {
    super( new ClueModelContentManager() );
    contManager = (ClueModelContentManager)super.contManager;
    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
  }


  ////////////////////////////////////////////////////////////////////////
  ///////////////   Implementierung von AbstractXuluModel   //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
   * werden Referenzen auf einen Teil der Ressourcen in lokalen Variablen
   * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
   * zu koennen.
   */
  public void performModelInit() {
    if ( contManager == null )
      return;
    // Zugriffsrechte aus Ressourcen/Propertys holen
    RA_baseLU       = contManager.getBaseScenario().getReadAccess(this);
    RA_baseLUHist   = contManager.getLUHistory().getReadAccess(this);
    RA_luConv       = contManager.getLUConversionMatrix().getReadAccess(this);
    RA_luElas       = contManager.getConversionElasticity().getReadAccess(this);
    RA_areaRestr    = contManager.getAreaRestrictions().getReadAccess(this);
    RA_luDemand     = contManager.getDemandScenario().getReadAccess(this);
    RA_sdfGrids     = ((ListProperty)contManager.getStaticDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    if ( contManager.getDynamicDrivingForceTypes() != null ) {
      RA_ddfList      = contManager.getDynamicDrivingForceTypes().getReadAccess(this);
      RA_ddfGrids     = ((ListProperty)contManager.getDynamicDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    } else {
      RA_ddfList      = null;
      RA_ddfGrids     = null;
    }
    RA_regrRes      = contManager.getRegressionResults().getReadAccess(this);
    if ( contManager.getNeighborhoodWeights() != null ) {
      RA_nbhWeights  = contManager.getNeighborhoodWeights().getReadAccess(this);
      RA_nbhSettings = contManager.getNeighborhoodSettings().getReadAccess(this);
      RA_nbhRegrRes  = contManager.getNeighborhoodRegression().getReadAccess(this);
      WA_outNbhProb  = ((ListProperty)contManager.getNeighborhoodProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    } else {
      RA_nbhWeights  = null;
      RA_nbhSettings = null;
      RA_nbhRegrRes  = null;
      WA_outNbhProb  = null;
    }
    WA_outActLU     = contManager.getActLUResult().getWriteAccess(this);
    WA_outActLUHist = contManager.getActLUHistory().getWriteAccess(this);
    WA_outLUProb    = ((ListProperty)contManager.getLUProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    WA_outIter      = contManager.getIterationVariables().getWriteAccess(this);
    WA_outStepLU    = ((ListProperty)contManager.getStepLUResults().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);

    // Variablen belegen mit denen gearbeitet wird
    this.stepCount   = contManager.getStepCount().getOneTimeReadAccess().getValueAsInt(); // WICHTIG!!
    this.dfCount     = contManager.getStaticDrivingForces().getGridCount();
    this.demandTolType    = contManager.getDemandToleranceType().getOneTimeReadAccess().getValueAsInt();
    this.demandTolSingle  = contManager.getSingleDemandTolerance().getOneTimeReadAccess().getValueAsFloat();
    this.demandTolAverage = contManager.getAverageDemandTolerance().getOneTimeReadAccess().getValueAsFloat();
    this.demandConv  = contManager.getConversionFactor().getOneTimeReadAccess().getValueAsFloat();
    this.luCount     = contManager.getLUConversionMatrix().getSize(0);
    this.baseLU      = (WritableGrid)RA_baseLU.getValue();
    this.areaWidth   = baseLU.getWidth();
    this.areaHeight  = baseLU.getHeight();
    this.areaMinX    = baseLU.getMinX();
    this.areaMinY    = baseLU.getMinY();
    this.areaMaxX    = baseLU.getMinX() + areaWidth;
    this.areaMaxY    = baseLU.getMinY() + areaHeight;
    this.cellSize    = ((float)(baseLU.getCellWidth() * baseLU.getCellHeight())) / demandConv;
    this.baseLUHist  = (WritableGrid)RA_baseLUHist.getValue();
    // PERFORMANZ D: Elastizitaet in lokalem Array
//    this.luElas      = (ListPropertyReadAccess)RA_luElas;
    this.luElas      = new float[((ListPropertyReadAccess)RA_luElas).getCount()];
    for (int i=0; i<luElas.length; i++)
      this.luElas[i] = RA_luElas.getValueAsFloat(i);

    this.areaRestr   = (WritableGrid)RA_areaRestr.getValue();
    // PERFORMANZ F: Demand und Conversion-Matrix in lokalen Array
//    this.luConv      = RA_luConv;
//    this.luDemand    = RA_luDemand;
    this.luConv      = new int[contManager.getLUConversionMatrix().getSize(0)][contManager.getLUConversionMatrix().getSize(1)];
    for (int i=0; i<luConv.length; i++)
      for (int j=0; j<luConv[i].length; j++)
        luConv[i][j] = RA_luConv.getValueAsInt(i,j);
    this.luDemand    = new float[contManager.getDemandScenario().getSize(0)][contManager.getDemandScenario().getSize(1)];
    for (int i=0; i<luDemand.length; i++)
      for (int j=0; j<luDemand[i].length; j++)
        luDemand[i][j] = RA_luDemand.getValueAsFloat(i,j);

    ListPropertyReadAccess ddfList     = (RA_ddfList != null) ? (ListPropertyReadAccess)RA_ddfList : null;
    this.ddfCount    = (ddfList != null)    ? ddfList.getCount() : 0;
    // PERFORMANZ C: Raster-Listen in lokale Arrays kopieren
//    this.ddfGrids    = (ddfList != null)    ? contManager.getDynamicDrivingForces() : null;
//    this.sdfGrids    = contManager.getStaticDrivingForces();
    this.ddfGrids    = (ddfList != null)    ? new WritableGrid[contManager.getDynamicDrivingForces().getGridCount()] : null;
    for (int i=0; i<ddfGrids.length; i++)
      ddfGrids[i] = contManager.getDynamicDrivingForces().getGrid(i);
    this.sdfGrids    = new WritableGrid[contManager.getStaticDrivingForces().getGridCount()];
    for (int i=0; i<sdfGrids.length; i++)
      sdfGrids[i] = contManager.getStaticDrivingForces().getGrid(i);

    // PERFORMANZ G
//    this.regrRes     = RA_regrRes;
    this.regrRes = new float[contManager.getRegressionResults().getSize(0)][contManager.getRegressionResults().getSize(1)];
    for (int i=0; i<regrRes.length; i++)
      for (int j=0; j<regrRes[i].length; j++)
        this.regrRes[i][j] = RA_regrRes.getValueAsFloat(i,j);
    // PERFORMANZ H
//    this.nbhWeights  = (RA_nbhWeights != null) ? (ListPropertyReadAccess)RA_nbhWeights : null;
    this.nbhWeights  = (RA_nbhWeights != null) ? new float[((ListPropertyReadAccess)RA_nbhWeights).getCount()] : null;
    for (int i=0; i<nbhWeights.length; i++)
      this.nbhWeights[i] = RA_nbhWeights.getValueAsFloat(i);
//    this.nbhSettings = (nbhWeights != null)    ? RA_nbhSettings : null;
    this.nbhSettings = (nbhWeights != null)    ? new float[contManager.getNeighborhoodSettings().getSize(0)][contManager.getNeighborhoodSettings().getSize(1)][contManager.getNeighborhoodSettings().getSize(2)] : null;
    for (int i=0; i<nbhSettings.length; i++)
      for (int j=0; j<nbhSettings[i].length; j++)
        for (int k=0; k<nbhSettings[i][j].length; k++)
        this.nbhSettings[i][j][k] = RA_nbhSettings.getValueAsFloat(i,j,k);
//    this.nbhRegrRes  = (nbhWeights != null)    ? RA_nbhRegrRes : null;
    this.nbhRegrRes = (nbhWeights != null)    ? new float[contManager.getNeighborhoodRegression().getSize(0)][contManager.getNeighborhoodRegression().getSize(1)] : null;
    for (int i=0; i<nbhRegrRes.length; i++)
      for (int j=0; j<nbhRegrRes[i].length; j++)
        this.nbhRegrRes[i][j] = RA_nbhRegrRes.getValueAsFloat(i,j);

    this.nbhRadiusX  = (nbhWeights != null)    ? (contManager.getNeighborhoodSettings().getSize(0)-1)/2 : 0;
    this.nbhRadiusY  = (nbhWeights != null)    ? (contManager.getNeighborhoodSettings().getSize(1)-1)/2 : 0;

    MultiGrid outNbhProb  = (nbhWeights != null)    ? contManager.getNeighborhoodProbabilities() : null;
    for (int i=0; outNbhProb!=null && outNbhProb.getGridCount()<luCount; i++)
      outNbhProb.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten
    this.outActLU     = (WritableGrid)WA_outActLU.getValue();
    this.outActLUHist = (WritableGrid)WA_outActLUHist.getValue();
    MultiGrid outLUProb    = contManager.getLUProbabilities();
    for (int i=0; outLUProb.getGridCount()<luCount; i++)
      outLUProb.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten

    // PERFORMANZ A: Raster-Liste in lokale Arrays kopieren
    this.outLUProbGrid  = new WritableGrid[luCount];
    this.outNbhProbGrid = new WritableGrid[luCount];
    for (int i=0; i<luCount; i++) {
      outLUProbGrid[i]  = outLUProb.getGrid(i);
      outNbhProbGrid[i] = outNbhProb.getGrid(i);
    }
//    // PERFORMANZ B
//    this.outNbhProb = new float[luCount][areaWidth][areaHeight];
//    this.outLUProb  = new float[luCount][areaWidth][areaHeight];

    // PERFORMANZ E: Iterationsvariablen in lokalem Array
//    this.outIter      = (ListPropertyWriteAccess)WA_outIter;
//    // Lister der Iter-Variablen auf benoetigte Groesse auffuellen
//    for (;outIter.getCount()<luCount;)
//      outIter.addValue(0.0f);
    this.outIter      = new float[luCount];

    this.outStepLU    = contManager.getStepLUResults();

    // Bei Toleranz-Art "prozentual" werden die Werte in Prozent
    // angegeben (intern wird aber mit Abweichungen zwischen 0 und 1 gerechnet!)
    if ( demandTolType == TOLTYPE_PCT ) {
      this.demandTolSingle  /= 100.0f;
      this.demandTolAverage /= 100.0f;
    }

    // Die Liste der dyn. DF ist blockweise nach DF grupperiert
    // Es koennen mehr Zeitschritt-Raster angegeben werden, als simuliert
    // werden.
    // Im Array 'ddfStartIdx' wird fuer die dynamischen DF gespeichert, an
    // welcher Stelle in der Liste der Block beginnt.
    // -1 bedeutet, dass der DF ein statische DF ist!
    this.ddfStartIdx = new int[dfCount];
    for (int i=0; i<dfCount; i++)
      ddfStartIdx[i] = - 1;
    for (int i=0; i<ddfCount; i++)
//      ddfStartIdx[i] = i*ddfGrids.getGridCount() / ddfCount; // PERFORMANZ C
      ddfStartIdx[i] = i*ddfGrids.length / ddfCount;

    // "Geschwindigkeit" fuer die Iterationsvariablen
    this.iterSpeed = new float[luCount];

    // Speichert die Flaechendeckung (in der Bedarfseinheit, z.B. ha)
    // fuer die einzelnen LU-Typen zu jedem Zeitpunkt
    // --> Die initialen Werte stehen in der 0-Zeile der Demand-Matrix
    this.luCover   = new float[luCount][stepCount+1];
    this.totCover = 0.0f;
    statusOut.print("InitCover: ");
    for (int lu=0; lu<luCount; lu++) {
//      luCover[lu][0] = luDemand.getValueAsFloat(lu, 0); // ORIGINAL
      luCover[lu][0] = luDemand[lu][0]; // PERFORMANZ F
      totCover += luCover[lu][0]; // Gesamt-Flaeche des Untersuchungsgebiets
      statusOut.print(luCover[lu][0]+"   ");
    }
    statusOut.println("   >> TotalCover "+totCover);

    // Ausgabe-Raster leeren
    for (int y=outActLU.getMinY(); y<outActLU.getMinY()+outActLU.getHeight(); y++)
      for (int x=outActLU.getMinX(); x<outActLU.getMinX()+outActLU.getWidth(); x++)
        outActLU.setRasterSample((float)Float.NaN,x,y);

    LOG.info("========================================");
    LOG.info("MODEL INITIALISED (Log-Level "+LOG.getLevel()+")");
  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    // Ressourcen wieder freigeben
    releaseAccess(RA_baseLU);
    releaseAccess(RA_baseLUHist);
    releaseAccess(RA_luConv);
    releaseAccess(RA_luElas);
    releaseAccess(RA_areaRestr);
    releaseAccess(RA_luDemand);
    releaseAccess(RA_sdfGrids);
    releaseAccess(RA_ddfList);
    releaseAccess(RA_ddfGrids);
    releaseAccess(RA_regrRes);
    releaseAccess(RA_nbhWeights);
    releaseAccess(RA_nbhSettings);
    releaseAccess(RA_nbhRegrRes);
    releaseAccess(WA_outActLU);
    releaseAccess(WA_outActLUHist);
    releaseAccess(WA_outLUProb);
    releaseAccess(WA_outNbhProb);
    releaseAccess(WA_outIter);
    releaseAccess(WA_outStepLU);
    LOG.info("MODEL DISPOSED");
    LOG.info("========================================");
  }

  /**
   * Fuehrt einen Schritt des Modellablaufs durch.
   * @param stepNo zu modellierender Schritt (beginnend bei 1!)
   */
  public void performModelStep(int stepNo) {
    LOG.debug("Step "+stepNo+" started");
    statusOut.println("Started step ".concat(String.valueOf(stepNo)));

    // Raster festlegen, in dem die LU des vorangegangenen
    // Schritts zu finden ist
    WritableGrid lastLUGrid = (stepNo==1) ? baseLU : outStepLU.getGrid(stepNo-2);
    // Raster festlegen, in dem die aktuelle LU History
    // zu finden ist
    WritableGrid luHistGrid = (stepNo==1) ? baseLUHist : outActLUHist;

    // Merken, ob an einer Stelle bereits eine U-Zelle in einen anderen
    // Typ gewechselt hat
    boolean[] alreadyLUCCFromType = new boolean[luCount];
    // Merken, ob an einer Stelle bereits eine U-Zelle neu entstanden ist
    boolean[] alreadyLUCCToType = new boolean[luCount];

    // LU-Wahrscheinlichkeiten berechnen
    calculateLUProbabilities(stepNo);

    // Nachbarschafts-Wahrscheinlichkeiten berechnen
    calulateNeighborhoodProbabilities(stepNo,lastLUGrid);

    // Randomizer fuer diesen Zeitschritt erzeugen
    rand = new Random(System.currentTimeMillis());

    int   iterNo     = 0;
    int   x          = 0;
    int   y          = 0;
    int   lu         = 0;
    int   lastLU     = 0;
    int   bestLU     = 0;
    float bestLUProb = 0.0f;
    float baseProb   = 0.0f;
    float nbhProb    = 0.0f;
    float nbhWeight  = 0.0f;
    float elas       = 0.0f;
    float totalProb  = 0.0f;
    Float lastLUC    = null;
    Float actLUC     = null;

    for (iterNo=1; !checkDemandCompliedAndCalculateIterVar(stepNo,iterNo); iterNo++) {
      LOG.debug("Iter "+iterNo+" started");
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,stepNo,iterNo) );
      // Arrays fuer Iterationsschritt initialisieren
      for (lu = 0; lu < luCount; lu++) {
        // noch kein Wechsel von einem LU-Typ
        alreadyLUCCFromType[lu] = false;
        // noch kein Wechsel zu einem LU-Typ
        alreadyLUCCToType[lu] = false;
        // Flaechen der LU-Typen fuer diesen Zeitschritt
        luCover[lu][stepNo] = 0.0f;
        // Speed fuer die Iterationsvariablen initialisieren
        // --> liegt die Nachfrage bei weniger als 5% des U'gebiets
        //     wird langsamer begonnen
//        if ( luDemand.getValueAsFloat(lu,stepNo)/totCover < 0.05f ) // ORIGINAL
        if ( luDemand[lu][stepNo]/totCover < 0.05f ) // PERFORMANZ F
          iterSpeed[lu] = 0.01f;
        else
          iterSpeed[lu] = 0.05f;
      }

      // Fuer alle Zellen des Untersuchungsgebiets
      for (y = areaMinY; y < areaMaxY; y++)
        for (x = areaMinX; x < areaMaxX; x++) {
          checkBreakingCommands();

          // Rasterzellen ausserhalb des Untersuchungsgebiets komplett
          // ignorieren
          // Grund: NoData-Zellen (Float.NaN) werden beim Cast in int zu 0,
          //        also zur Landnutzung 0!!
          if ( checkNoData(x,y,baseLU) )
            continue;

          // Nutzungsdauern erhoehen
          if (iterNo==1)
            outActLUHist.setRasterSample(luHistGrid.getRasterSampleAsFloat(x, y) + 1, x, y);

          // vorangegangene LU merken
          lastLU = lastLUGrid.getRasterSampleAsInt(x, y);

          // Beste gefundene LU wird mit aktueller LU initialisiert
          bestLU = lastLU;
          bestLUProb = 0.0f;

          // Pruefen, ob Zelle die LU wechseln darf
          if (checkLUCCGeneralAllowed(x, y, stepNo, lastLU, alreadyLUCCToType[lastLU])) {
            // Alle LU-Typen auf Wechsel pruefen
            for (lu = 0; lu < luCount; lu++) {
              // Pruefen, ob Zelle in den betrachteten Typ wechseln darf
              if (checkLUCCAllowed(x, y, stepNo, lastLU, lu, alreadyLUCCFromType[lu])) {
                // Gesamt-Wahrscheinlichkeit berechnen
                // a) DF-W'keit und Nachbarschaft-W'keit gehen anteilig ein
                // b) Elastizitaet geht nur ein, wenn Zelle bereits fuer den
                //    LU-Typ genutzt wird
                // c) ITER geht negativ ein, da es auf Basis der Abweichung
                //    luCover-luDemand berechnet wird. Ein negatives Iter
                //    kennzeichnet also einen Mangel und muss die Gesamtw'keit
                //    erhoehen
//                baseProb  = outLUProb.getGrid(lu).getRasterSampleAsFloat(x, y); // ORIGINAL-VERSION
//                nbhProb   = (nbhWeights == null) ? 0.0f : outNbhProb.getGrid(lu).getRasterSampleAsFloat(x, y); // ORIGINAL-VERSION
                baseProb  = outLUProbGrid[lu].getRasterSampleAsFloat(x, y); // PERFORMANZ A
                nbhProb   = (nbhWeights == null) ? 0.0f : outNbhProbGrid[lu].getRasterSampleAsFloat(x, y); // PERFORMANZ A
//                baseProb  = outLUProb[lu][x][y]; // PERFORMANZ B
//                nbhProb   = (nbhWeights == null) ? 0.0f : outNbhProb[lu][x][y]; // PERFORMANZ B
//                nbhWeight = (nbhWeights == null) ? 0.0f : nbhWeights.getValueAsInt(lu); // ORIGINAL
                nbhWeight = (nbhWeights == null) ? 0.0f : nbhWeights[lu]; // PERFORMANZ H
//                elas      = (lu == lastLU && luElas.getValueAsFloat(lu) < 1) ? luElas.getValueAsFloat(lu) : 0.0f; // ORIGINAL-VERSION
                elas      = (lu == lastLU && luElas[lu] < 1) ? luElas[lu] : 0.0f; // PERFORMANZ D
                // LU-Wahrscheinlichkeit zusammensetzen
                totalProb = baseProb * (1-nbhWeight)
                          + nbhProb  * nbhWeight
                          + elas
//                          - outIter.getValueAsFloat(lu); // ORIGINAL-VERSION
                          - outIter[lu]; // PERFORMANZ E

                // betrachteter Typ ist geeigneter als der bisher beste
                if (totalProb > bestLUProb) {
                  bestLU = lu;
                  bestLUProb = totalProb;
                }
              }
            }
            // Merken wenn wirklich ein Wechsel stattfindet
            if (bestLU != lastLU) {
              alreadyLUCCFromType[lastLU] = true;
              alreadyLUCCToType[bestLU] = true;
            }
          }

          // LU fuer diesen Iterationschritt setzen
          outActLU.setRasterSample((float)bestLU, x, y);
          // Flaechendeckung fuer LU-Typ erhoehen
          luCover[bestLU][stepNo] += cellSize;
        }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,stepNo,iterNo) );
      LOG.debug("Iter "+iterNo+" finished");
    }

    // Liste der Zeitschritt-Ausgaberaster vergroessern
    outStepLU.addGrid();

    // LU in Ausgaberaster des Zeitschritts uebernehmen und ggf.
    // Nutzungsdauer initialisieren
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++) {
        checkBreakingCommands();
        lastLUC = (Float)lastLUGrid.getRasterSample(x,y);
        actLUC  = (Float)outActLU.getRasterSample(x,y);
        // Wenn die Landnutzung gewechselt hat, wird die Historie zurueck
        // gesetzt
        if ( !actLUC.equals(lastLU) )
          outActLUHist.setRasterSample((float)0,x,y);

        // LU in Ausgabe-Raster des simulierten Schritts uebernehmen
        outStepLU.getGrid(stepNo-1).setRasterSample(actLUC,x,y);
      }
    statusOut.println("Ended step ".concat(String.valueOf(stepNo)));
    LOG.debug("Step "+stepNo+" finished");
  }

  //////////////////////////////////////////////////////////////////////
  ////////////////////////   Clue-Methoden   ///////////////////////////
  //////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob eine Rasterzelle mit <code>NoData</code> belegt ist
   * @param x     Raster-Koordinate
   * @param y     Raster-Koordinate
   * @param grid  zu pruefendes Raster
   */
  protected static boolean checkNoData(int x, int y, WritableGrid grid) {
    return Float.isNaN(grid.getRasterSampleAsFloat(x,y));
  }

  /**
   * Berechnet die auf den Driving Forces basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die LU-Wahrscheinlichkeiten".<br>
   * Werden keine dynamischen Driving Forces verwendet, geschieht die Berechnung
   * nur beim ersten Modellierungsschritt.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   */
  protected void calculateLUProbabilities(int stepNo) {
    LOG.debug("LUC-Prop-Calc. started");
    // Wenn keine dynamischen Driving Forces verwendet werden
    // braucht die Berechnung nur fuer den ersten Schritt
    // durchgefuehrt werden
    if ( ddfCount == 0 && stepNo > 1 )
      return;

    statusOut.print("Calculate LU-Probabilities for step "+stepNo+"...");
    int x            = 0;
    int y            = 0;
    int lu           = 0;
    int df           = 0;
    float betaSum    = 0.0f;
    float expBetaSum = 0.0f;
    float prob       = 0.0f;
    // Fuer Zellen des Untersuchungsgebiets
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++)
        // NotData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 ) {
          // Fuer alle LU-Typen
          for (lu = 0; lu < luCount; lu++) {
            betaSum = 0.0f;
            // Fuer alle Driving Forces die Summe der Betas berechnen
            for (df = 0; df < dfCount; df++) {
              checkBreakingCommands();
              // DF-Grid festlegen:
              // Bei dynamischem DF das jew. Zeitschritt-Raster aus der Liste
              // ansonsten das statische Raster
              WritableGrid dfGrid = null;
              if ( ddfStartIdx[df] >= 0 )
//                dfGrid = this.ddfGrids.getGrid(ddfStartIdx[df]+stepNo-1); // ORIGINAL
                  dfGrid = this.ddfGrids[ddfStartIdx[df]+stepNo-1]; // PERFORMANZ C
              else
//                dfGrid = this.sdfGrids.getGrid(df); // ORIGINAL
                 dfGrid = this.sdfGrids[df]; // PERFORMANZ C
              // Beim ersten Driving Force wird die Summe mit der
              // Regressionskonstante initialisiert
              if (df == 0)
//                betaSum = regrRes.getValueAsFloat(0, lu); // ORIGINAL
                betaSum = regrRes[0][lu]; // PERFORMANZ G
              // betaSum = BETA0(U) + Sum( BETA(df+1,U) * DF(df,x,y) )
//              betaSum += regrRes.getValueAsFloat(df + 1, lu) // ORIGINAL
              betaSum += regrRes[df + 1][lu] // PERFORMANZ G
                        * dfGrid.getRasterSampleAsFloat(x, y);
            }
            expBetaSum = (float)Math.exp(betaSum);
            prob = expBetaSum / (1 + expBetaSum);
            // errechnete W'keit im Raster speichern
//            this.outLUProb.getGrid(lu).setRasterSample(prob, x, y); // ORIGINAL
            this.outLUProbGrid[lu].setRasterSample(prob, x, y); // PERFORMANZ A
//            this.outLUProb[lu][x][y] = prob; // PERFORMANZ B
          }
        }
    statusOut.println(" done.");
    LOG.debug("LUC-Prop-Calc. finished");
  }

  /**
   * Berechnet die auf den Nachbarschaftsbeziehungen basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die Nachbarschafts-Wahrscheinlichkeiten".<br>
   * Da sich die Landnutzung mit jedem Zeitschritt aendert geschieht dies
   * vor jedem Zeitschritt neu.<br>
   * Diese Methode macht nichts, wenn ohne Nachbarschafts-W'keiten simuliert wird,
   * also wenn keine Angabe fuer die "Neighborhood-Weights" gemacht wird.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param luGrid LU-Raster aus dem die Umgebung der jeweiligen Zellen
   *               ermittelt werden
   */
  protected void calulateNeighborhoodProbabilities(int stepNo, WritableGrid luGrid) {
    LOG.debug("NBH-Prop-Calc. started");
    // Wenn ohne Nachbarschaftsbeziehungen simuliert wird
    if ( nbhWeights == null )
      return;

    statusOut.print("Calculate Neighborhood-Probabilities for step "+stepNo+"...");

    int      x            = 0;
    int      y            = 0;
    int      lu           = 0;
    int      i            = 0;
    int      dx           = 0;
    int      dy           = 0;
    int      totCellCount = 0;
    float[]  luCellCount  = new float[luCount];
    float    betaSum      = 0.0f;
    float    beta         = 0.0f;
    float    nbhFact      = 0.0f;
    float    expBetaSum   = 0.0f;

    // Fuer Zellen des Untersuchungsgebiets
    for (y=areaMinY; y<areaMaxY; y++)
      for (x=areaMinX; x<areaMaxX; x++)
        // NoData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 )
          // Fuer alle LU-Typen
          for (lu = 0; lu < luCount; lu++) {
            checkBreakingCommands();

            // Anzahl der Zellen in der Umgebung initialisieren
            totCellCount = 0;
            for (i=0; i<luCount; i++)
              luCellCount[i] = 0.0f;

            // Umgebung der Zelle betrachten
            for (dy=-nbhRadiusY; dy<=nbhRadiusY; dy++)
              if ( areaMinY <= y+dy && y+dy<areaMaxY )
                for (dx=-nbhRadiusX; dx<=nbhRadiusX; dx++)
                  if ( areaMinX <= x+dx && x+dx<areaMaxX ) {
                    float luc = luGrid.getRasterSampleAsFloat(x+dx,y+dy);
                    if ( !Float.isNaN(luc) /*!checkNoData(x+dx,y+dy,baseLU)*/ ) {
                        // Alle Zellen in Umgebung zaehlen (jedoch nur diejenigen,
                        // die potentiell auch in die Bewertung eingehen koennen!)
//                        if ( nbhSettings.getValueAsFloat(dx+nbhRadiusX,dy+nbhRadiusY,lu) != 0.0f ) // ORIGINAL
                        float setting = nbhSettings[dx+nbhRadiusX][dy+nbhRadiusY][lu];
                        if ( setting != 0.0f ) { // PERFORMANZ H
                          totCellCount++;
                          // Anzahl der Zellen eines bestimmten Typs aufsummieren
                          // (relativ zum in der Umgebung angegebenen Faktor)
//                          luCellCount[ luGrid.getRasterSampleAsInt(x+dx,y+dy) ] += nbhSettings.getValueAsFloat(dx+nbhRadiusX,dy+nbhRadiusY,lu); // ORIGINAL
//                          luCellCount[ luGrid.getRasterSampleAsInt(x+dx,y+dy) ] += setting ; // PERFORMANZ H
                          luCellCount[ (int)luc ] += setting ; // PERFORMANZ H
                        }
                      }
                    }
              // Summe der Betas berechnen
//              betaSum = nbhRegrRes.getValueAsFloat(0,lu); // Regr.Konstante // ORIGINAL
              betaSum = nbhRegrRes[0][lu]; // Regr.Konstante // PERFORMANZ H
              for (i=0; i<luCount; i++) {
//                beta    = nbhRegrRes.getValueAsFloat(i + 1, lu); // ORIGINAL
                beta    = nbhRegrRes[i + 1][lu]; // PERFORMANZ H
                nbhFact = (luCellCount[i]/totCellCount) / (luCover[i][stepNo-1]/totCover);
                betaSum += beta * nbhFact;
              }
              // Wahrscheinlichkeit berechnen
              expBetaSum = (float)Math.exp(betaSum);
//              outNbhProb.getGrid(lu).setRasterSample((float)(expBetaSum/(1 + expBetaSum)),x,y); // ORIGINAL
              outNbhProbGrid[lu].setRasterSample((float)(expBetaSum/(1 + expBetaSum)),x,y); // PERFORMANZ A
//              this.outNbhProb[lu][x][y] = (float)(expBetaSum/(1 + expBetaSum)); // PERFORMANZ B
          }
    statusOut.println(" done.");
    LOG.debug("NBH-Prop-Calc. finished");
  }


  /**
   * Prueft, ob ein genereller LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliessen die Area Restrictions einen Wechsel aus?<br>
   *     > Rasterwert(x,y) != 0
   *     </li>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet 1
   *     und steigende Bedarf)?<br>
   *     > ELAS(lastLU) == 1 und DEMAND(lastLU,t) >= DEMAND(lastLU,t-1)
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param alreadyLUCCToActualType Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel in den den aktuellen
   *                                Typ der Zelle (lastLU) vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCGeneralAllowed(int x, int y, int stepNo, int lastLU, boolean alreadyLUCCToActualType) {
    // Area Restrictions checken
    if ( areaRestr.getRasterSampleAsFloat(x,y) != 0 )
      return false;
// DER ERSTE TEIL IST KEINE ORIGINAL-CLUE-REGEL!
//    // Conversion Elasticity checken
//    // --> wenn Elasticity 1 und bereits an anderer Stelle eine Zelle des
//    //     bisherigen Typs entstanden ist, darf sie nicht wechseln
//    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
//    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
//    //     an anderer Stelle neue entstehen
//    if (  luElas.getValueAsFloat(lastLU) == 1.0f
//       && ( alreadyLUCCToActualType ||
//           luDemand.getValueAsFloat(lastLU,stepNo) >= luCover[lastLU][stepNo-1] ) )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
    //     an anderer Stelle neue entstehen
//    if (  luElas.getValueAsFloat(lastLU) == 1.0f && // PERFORMANZ D
    if (  luElas[lastLU] == 1.0f &&
//          luDemand.getValueAsFloat(lastLU,stepNo) >= luCover[lastLU][stepNo-1] ) // ALT
//          luDemand.getValueAsFloat(lastLU,stepNo) >= luDemand.getValueAsFloat(lastLU,stepNo-1) ) // ORIGINAL
          luDemand[lastLU][stepNo] >= luDemand[lastLU][stepNo-1] ) // PERFORMANZ F
      return false;

    return true;
  }

  /**
   * Prueft, ob ein bestimmter LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet des
   *     potentiellen Wechseltyps 1 und sinkender Bedarf)?<br>
   *     > lastLU != newLU und ELAS(newLU) == 1 und DEMAND(newLU,t) <= DEMAND(newLU,t-1)
   *     </li>
   * <li>Schliesst die Conversion Matrix einen Wechsel aus (generell oder ueber
   *     Historie)?<br>
   *     > CONV(lastLU,newLU) == 0 oder<br>
   *     > CONV(lastLU,newLU) > 100 und HIST(x,y) < CONV(lastLU,newLU)-101 oder<br>
   *     > CONV(lastLU,newLU) < -100 und HIST(x,y) > -CONV(lastLU,newLU)-101
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param newLU  potentielle neue Bedeckung von (x,y)
   * @param alreadyLUCCFromNewType  Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel vom potentiellen neuen
   *                                Typ in einen anderen Typ vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCAllowed(int x, int y, int stepNo, int lastLU, int newLU, boolean alreadyLUCCFromNewType) {
// IST KEINE ORIGINAL-CLUE-REGEL!
//    // LU-Conversion-Elasticity des neuen Typs checken
//    // --> wenn Elasticity des neuen Typs 1 ist und an anderer Stelle bereits
//    //     eine Zelle des dieses Typs "verloren" ging, darf er hier
//    //     nicht entstehen
//    if (  alreadyLUCCFromNewType && luElas.getValueAsFloat(newLU) == 1.0f )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity des neuen Typs "newLU" 1 und dessen Bedarf sinkt,
    //     darf kein Wechsel in diesen Typ stattfinden, denn dann muessten (um
    //     den Bedarf fuer "newLU" zu erreichen) an anderer Stelle newLU-Zellen
    //     entfernt werden
    if (  lastLU != newLU &&
//          luElas.getValueAsFloat(newLU) == 1.0f && // ORIGINAL
          luElas[newLU] == 1.0f && // PERFORMANZ D
//          luDemand.getValueAsFloat(newLU,stepNo) <= luCover[newLU][stepNo-1] ) // ORIGINAL
          luDemand[newLU][stepNo] <= luCover[newLU][stepNo-1] ) // PERFORMANZ F
//          luDemand.getValueAsFloat(newLU,stepNo) <= luDemand.getValueAsFloat(newLU,stepNo-1) )
      return false;

    // Wert aus LU-Conversion-Matrix ermitteln
    // Bemerke: "von" steht in den Zeilen (also Y, also Dim1)
    //          "nach" steht in den Spalten (also X, also Dim2)
//    int conv = luConv.getValueAsInt(newLU,lastLU); // ORIGINAL
    int conv = luConv[newLU][lastLU]; // PERFORMANZ F
    // 1 --> LUCC generell erlaubt
    if ( conv == 1 )
      return true;
    // 0 --> LUCC generell verboten
    if ( conv == 0 )
      return false;

    // 100+t --> Wechsel erst nach t Schritten der vorangegangenen
    //           LU moeglich
    if ( conv > 100 ) {
//      int minLUHist = conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist >= minLUHist;
      int minLUHist = conv-101;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist >= minLUHist;
    }

    // -100-t --> Wechsel nicht mehr moeglich, wenn Zelle bereits
    //            t Schritte unter der vorangegangenen LU
    if ( conv < -100 ) {
//      int maxLUHist = -conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist < maxLUHist;
      int maxLUHist = -conv-101;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist <= maxLUHist;
    }

    return true;
  }

  private float[] singleDevAbs = null;
  private float[] singleDevPct = null;
  /**
   * Prueft, ob die aktuelle LU-Konfiguration die einzelenen Bedarfsanforderungen
   * (bis auf eine Abweichung) erfuellt und berechnet die Iterationsvariablen
   * neu.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param iterNo aktuelle Iteration (beginnend bei 1)
   */
  protected boolean checkDemandCompliedAndCalculateIterVar(int stepNo, int iterNo) {
    LOG.debug("started\tIter-Recalc");
    if ( singleDevAbs == null ) {
      singleDevAbs = new float[luCount];
      singleDevPct = new float[luCount];
    }
    boolean demandComplied = true;
    // Abweichnungen berechnen
    float   averageDevAbs = 0.0f;
    float   averageDevPct = 0.0f;
    float   totalDevAbs = 0.0f;
    float   totalDevPct = 0.0f;
    float   demand      = 0.0f;
    float   singleDev   = 0.0f;
    for (int lu = 0; lu < luCount; lu++) {
      // ##### Einzelabweichungen #####
      // Absolute Abweichung der Landbedeckung vom Bedarf berechnen
//      demand = luDemand.getValueAsFloat(lu, stepNo); // ORIGINAL
      demand = luDemand[lu][stepNo]; // PERFORMANZ F
      singleDevAbs[lu]  = luCover[lu][stepNo] - demand;
      // Prozentuale Einzel-Abweichnung relativ zum Bedarf
      if (demand == 0)
        // ist der Bedarf 0, aber auch die Bedeckung, ist die Abw. 0%
        singleDevPct[lu] = luCover[lu][stepNo] == 0.0f ? 0.0f : 1.0f;
      else
        singleDevPct[lu] = singleDevAbs[lu] / demand;

      // Abweichung ueber erlaubter Toleranz?
      singleDev = (demandTolType == TOLTYPE_PCT) ? singleDevPct[lu] : singleDevAbs[lu];
      if (Math.abs(singleDev) > demandTolSingle)
        demandComplied = false;

      // ##### Gesamtabweichungen aufsummieren #####
      totalDevAbs += Math.abs(singleDevAbs[lu]);
    }

    // Prozentuale Gesamt-Abweichnung relativ zum Gesamt-Untersuchungsgebiet
    totalDevPct = totalDevAbs / totCover;

    // Durchschnittliche Abweichnung pro LU-Typ
    averageDevAbs = totalDevAbs / luCount;

    // Durchschnittliche Abweichnung pro LU-Typ relativ zum Gesamt-Untersuchungsgebiet
    averageDevPct = averageDevAbs / totCover;

    // Abweichung ueber erlaubter Toleranz?
    float avgDev = (demandTolType == TOLTYPE_PCT) ? averageDevPct : averageDevAbs;
    if ( Math.abs(avgDev) > demandTolAverage )
      demandComplied = false;

    // Iterationsvariablen neu berechnen
    for (int lu = 0; lu < luCount; lu++) {
      // Um Alternieren zwischen 2 Konfiguration zu verhindern
      // eine kleine Zufallszahl hinzuaddieren (zwischen -0,005 und 0,005)
      float randomValue = (rand.nextFloat()-0.5f)/100;

      // --> Im ersten Schritt erhalten alle Iterationsvariablen den Wert 0,5
//      float iter = outIter.getValueAsFloat(lu); // ORIGINAL-VERSION
      float iter = outIter[lu]; // PERFORMANZ E
      if (iterNo == 1)
//        outIter.setValue(0.5f+randomValue, lu); // ORIGINAL-VERSION
        iter = 0.5f+randomValue; // PERFORMANZ E
//        iter = 0.5 + new float[] {0.002f,0,0.002f,0,0,-0.003f}[lu];
//        iter = 0.5 + new float[] {0.002f,0,0.001f,0,0,-0.002f}[lu];
      else {
//        // Anteil der Einzelabweichnung an der Gesamtabweichnung
////        float dev = (totalDevAbs!=0) ? singleDevAbs[lu] / totalDevAbs : 0.0f;
//        float dev  = singleDevPct[lu];
//        iter = outIter.getValueAsDouble(lu) // alter Iter-Wert
//             + dev // Anteil der Abweichung zum Bedarf
//             + randomValue;
        /////////////////////////////////////////
        // Original-Iter-Berechnung aus CLUE
        iterSpeed[lu] += 0.00005f;
        iter += (singleDevAbs[lu]/totCover)/(iterSpeed[lu]*(rand.nextInt(101)+1));

        if (iter > 2)
          iter = 0.5f;
        if ( Math.abs(iter) < 0.001f )
          iter = 0.01f * Math.signum(iter);
        if (iter < -0.5f)
          iter = -0.4f;
        /////////////////////////////////////////
      }
      // Iter speichern
//      outIter.setValue(iter, lu); // ORIGINAL-VERSION
      outIter[lu] = iter; // PERFORMANZ E
    }


//    StringBuffer buffer = new StringBuffer();
//    for (float v : outIter)
//      buffer.append(v).append(" ");
//    LOG.debug("finished\tIter-Recalc "+stepNo+"/"+iterNo+"   "+buffer.toString());
    LOG.debug("finished\tIter-Recalc");
    return demandComplied;
  }
}
