/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.clue;

import java.awt.image.DataBuffer;
import java.awt.Window;
import java.awt.Component;
import java.awt.Container;
import java.awt.Color;
import java.awt.Font;
import java.awt.geom.AffineTransform;
import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JProgressBar;
import javax.swing.BoundedRangeModel;
import javax.swing.DefaultBoundedRangeModel;
import javax.swing.JSlider;
import javax.swing.SpringLayout;
import javax.swing.BorderFactory;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;
import java.util.Random;
import java.text.NumberFormat;
import java.text.DecimalFormat;

import schmitzm.lang.LangUtil;

import schmitzm.data.property.Access;
import schmitzm.data.property.ValueProperty;
import schmitzm.data.property.ValuePropertyType;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ListPropertyReadAccess;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.MatrixProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.WritableGrid;

import schmitzm.swing.SwingUtil;
import schmitzm.swing.SpringUtilities;
import schmitzm.swing.ExpansionBar;

import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.ModelResource;
import edu.bonn.xulu.model.AbstractModelResource;
import edu.bonn.xulu.model.ValuePropertyResource;
import edu.bonn.xulu.model.event.ModelIterationStepStartedEvent;
import edu.bonn.xulu.model.event.ModelIterationStepFinishedEvent;

import edu.bonn.xulu.plugin.data.grid.MultiGrid;

import schmitzm.temp.BaseTypeUtil;

/**
 * <b>BEI DIESER KLASSE HANDELT ES SICH UM DIE URSPRUENGLICHE XULU-IMPLEMENTIERUG
 *    DES CLUE-MODELLS. DARIN KOMMEN NOCH KEINERLEI LAUFZEIT-OPTIMIERENDE TECHNIKEN
 *    ZUM EINSATZ. FUER REALE MODELL-LAEUFE - INSBESONDERE AUF GROSSEN UNTERSUCHUNGSGEBIETEN -
 *    WIRD EMPFOHLEN, DIE UM ETWA EINEN FAKTOR 10 SCHNELLERE IMPLEMENTIERUNG
 *    {@link ClueModel} zu verwenden.</b>
 * <br><br>
 * Diese Klasse stellt eine Implementierung des
 * <a href="http://www.dow.wau.nl/clue/" target="_blank">CLUE</a>-Modells dar
 * (CLUE = The <b>C</b>onversion of <b>L</b>and <b>U</b>se and its <b>E</b>ffects).
 * Dieses arbeitet mit Rasterdaten und berechnet die zeitliche
 * Landnutzungsveraenderung (LUCC) in einem Untersuchungsgebiet iterativ auf
 * Basis von statistischen Eingabeparametern und vorgegebenen Landnutzungsbedarfen.<br>
 * Das CLUE-Modell benoetigt folgende Ressourcen, welche durch den
 * {@link ClueModelContentManager} spezifiziert werden:
 * <ol>
 * <li><b>Modellparameter:</b>
 *     <ul>
 *     <li><i>n</i> = Anzahl modellierter LU-Typen (<b>implizit</b> durch die LU-Conversion-Matrix gegeben!)</li>
 *     <li><i>m</i> = Anzahl an Driving Forces (<b>implizit</b> durch die Anzahl der Driving-Force-Raster gegeben!))</li>
 *     <li><i>o</i> = Anzahl an dynamischen Driving Forces (<b>implizit</b> durch die Liste der dynamischen DF gegeben!))</li>
 *     <li><i>t</i> = Anzahl zu simulierende Zeitschritte (Integer)</li>
 *     <li><i>x</i> = Breite des Untersuchungsgebiets (<b>implizit</b> durch das Ausgangsszenario gegeben!)</li>
 *     <li><i>y</i> = Anzahl zu simulierende Zeitschritte (<b>implizit</b> durch das Ausgangsszenario gegeben!)</li>
 *     <li><i>a</i> = Erlaubte Durchschnitts-Abweichung zu den vorgegebenen LU-Bedarfen</li>
 *     <li><i>s</i> = Erlaubte Abweichung eines jeden LU-Typs zum vorgegebenen LU-Bedarf</li>
 *     <li><i>d</i> = Art der Abweichungsangaben (0=prozentual, 1=absolut in Bedarf-Einheit)</li>
 *     <li><i>u</i> = Umrechnungsfaktor (z.B. 10000 bei ha), um von der (Flaechen-)Einheit der LU-Bedarfe in qm umzurechnen (Double)</li>
 *     </ul></li>
 * <li><b>Ein Ausgangsszenario ({@link WritableGrid})</b><br>
 *     Stellt die Ausgangs-Landnutzung dar und muss Werte zwischen 0 und <i>n</i>
 *     (sowie {@link Float#NaN} = NoData) enthalten.<br>
 *     Dieses Raster bestimmt die Groesse (<i>x</i>, <i>y</i>) des
 *     betrachteten Untersuchungsgebiets. Alle anderen Ein/Ausgabe-Raster muessen
 *     dieser Dimension entsprechen.</li>
 * <li><b>LU-History ({@link WritableGrid})</b><br>
 *     Stellt die Vergangenheit des Ausgangsszenarios dar. Fuer jede Zelle
 *     wird angegeben, wie lange die aktuelle Landnutzung bereits besteht.<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Eine LU-Conversion-Matrix ({@link MatrixProperty MatrixProperty[Integer]})</b><br>
 *     Beschreibt ob und wie ein direkterWechsel zwischen den LU-Typen erlaubt
 *     ist. Stellt eine <i>n</i>�<i>n</i>-Matrix dar und bestimmt somit,
 *     wie viele LU-Typen modelliert werden.</li>
 * <li><b>Conversion Elasticity ({@link ListProperty ListProperty[Double]})</b><br>
 *     Beschreibt die Bestaendigkeit der LU-Typen und stellt so eine weitere
 *     Einschraenkung an den LU-Wechsel dar. Die Liste muss fuer jeden LU-Typ
 *     einen Wert zwischen 0 und 1 enthalten (also <i>n</i> Double-Werte).<br>
 *     <ul>
 *     <li><i>ELAS[U] = 1</i> bedeutet, dass wenn an einer Stelle eine U-Zelle "verschwindet",
 *     an anderer Stelle im gleichen Zeitschritt keine Zelle vom Typ U entstehen darf.</li>
 *     <li><i>ELAS[U] = 0</i> bedeutet, dass uneingeschraenkt an einer Stelle ein Wechsel von U
 *         nach A und gleichzeitig an anderer Stelle von B nach U moeglich ist.</li>
 *     <li><i>0 < ELAS[U] < 1</i> drueckt eine Tendenz zu einem der beiden Extremwerte 0 oder
 *         1 aus.</li>
 *     </ul></li>
 * <li><b>Area Restrictions, Spatial policies ({@link WritableGrid})</b><br>
 *     Beschreiben, an welchen Stellen des Untersuchungsgebiets ein LU-Wechsel generell
 *     verboten ist (z.B. Schutzgebiete oder Bereiche, die nicht zum Untersuchungsgebiet
 *     gehoeren).<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.
 *     Der Rasterwert NoData ({@link Float#NaN}) bedeutet, dass ein LU-Wechsel
 *     verboten ist, jeder andere Wert erlaubt den Wechsel.</li>
 * <li><b>Demand Scenario ({@link MatrixProperty MatrixProperty[Integer]})</b><br>
 *     Beschreibt den LU-Bedarf (in Zellen!) fuer jeden LU-Typ zu jedem
 *     modellierten Zeitschritt. Muss eine <i>n</i>�(<i>t</i>+1)-Matrix darstellen,
 *     die Integer-Werte enthaelt.<br>
 *     <u>Bemerke:</u><br>
 *     Der "Bedarf" in der Zeile 0 stellt die Situation
 *     des Ausgangsszenarios dar. Deshalb sind <i>t</i>+1 Zeilen anzugeben!</li>
 * <li><b>Static Driving Forces ({@link MultiGrid})</b><br>
 *     Beschreiben die Auspraegung der statischen Einflussfaktoren an jeder Stelle des
 *     Untersuchungsgebiets. Zusammen mit den Regressionsparametern werden
 *     hieraus die LU-Wahrscheinlichkeiten errechnet.<br>
 *     Das MultiGrid muss mind. <i>m</i> numerische Raster enthalten, die alle dieselbe
 *     Dimension besitzen, wie das Ausgangsszenario. Werden dynamische DF verwendet
 *     werden die Elemente an deren Index <b>nicht</b> verwendet, sie muessen
 *     jedoch vorhanden sein!</li>
 * <li><b>Opt: Dynamic Driving Force Types ({@link ListProperty ListProperty[Integer]})</b><br>
 *     Diese Liste enthaelt die Nummern der dynamischen Driving Forces. Die
 *     Angabe ist optional, muss aber - wenn angegeben - Werte zwischen
 *     0 und <i>m</i> enthalten.</li>
 * <li><b>Opt: Dynamic Driving Forces ({@link MultiGrid})</b><br>
 *     Beschreiben die Auspraegung der dynamischen Einflussfaktoren an jeder Stelle des
 *     Untersuchungsgebiets ueber die Zeit. Zusammen mit den Regressionsparametern werden
 *     hieraus die LU-Wahrscheinlichkeiten errechnet.<br>
 *     Das MultiGrid muss mind. <i>t</i>*<i>o</i> numerische Raster enthalten,
 *     die alle dieselbe Dimension besitzen, wie das Ausgangsszenario. Es koennen
 *     aber auch mehr Raster angegeben werden. Bedingung ist aber, dass fuer alle
 *     dyn. DF gleich viele Raster (=Zeitpunkte) angegeben werden! Die Grids
 *     sind in der Liste gruppiert nach DF anzugeben, d.h. erst alle Grids
 *     fuer den ersten dyn. DF, dann alle Grids fuer den zweiten, usw.</li>
 * <li><b>Regression Results ({@link MatrixProperty MatrixProperty[Double]})</b><br>
 *     Beschreiben die Einfluss-Staerke (Betas) der Driving Forces auf jeden einzelnen
 *     LU-Typ. Muss eine (<i>m</i>+1)�<i>n</i>-Matrix darstellen,
 *     die Double-Werte enthaelt.</li>
 * <li><b>Opt: Neigborhood Weights ({@link ListProperty ListProperty[Double]})</b><br>
 *     Diese (optionale) Liste enthaelt die Gewichte, wie stark der Nachbarschaftsfaktor
 *     in die Berechnung der LU-Wahrscheinlichkeit fuer jeden LU-Typ eingeht.
 *     Sie muss <i>n</i> Werte enthalten.</li>
 * <li><b>Opt: Neighborhood Settings ({@link MatrixProperty MatrixProperty[Double]}})</b><br>
 *     Beschreibt die Nachbarschaft, die fuer jeden LU-Typ betrachtet wird.
 *     Muss eine <i>w</i>�<i>h</i>�<i>n</i>-Matrix darstellen, die Double-Werte enthaelt.
 *     Dabei bestimmten <i>w</i> und <i>h</i> die Hoehe und Breit der betrachteten
 *     Nachbarschaft. Werden Neighborhodd Weights angegeben, so werden auch die
 *     Settings zu einer Pflicht-Angabe.
 * <li><b>Opt: Neighborhood Regression Results ({@link MatrixProperty MatrixProperty[Double]}})</b><br>
 *     Beschreibt die Einfluss-Staerke (Betas) eines jeden LU-Typs fuer den
 *     Fall, dass er in der Nachbarschaft eines anderen LU-Typs auftritt.
 *     Muss eine (<i>n</i>+1)�<i>n</i>-Matrix darstellen, die Double-Werte enthaelt.
 *     Werden Neighborhodd Weights angegeben, so werden auch die Regression Results
 *     zu einer Pflicht-Angabe.
 * <li><b>Ein Ausgaberaster fuer die aktuelle LU ({@link WritableGrid})</b><br>
 *     Hierin wird die aktuell modellierte LU-Konfiguration gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein Ausgaberaster fuer die aktuelle LU Historie ({@link WritableGrid})</b><br>
 *     Hierin wird die aktuelle Dauer gespeichert, fuer die eine Zelle
 *     unter der aktuellen LU befindet.<br>
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein temp. Ausgaberaster fuer die LU-Wahrscheinlichkeiten ({@link MultiGrid})</b><br>
 *     Hierin werden die ueber die Driving Forces und Regression Results errechneten
 *     LU-Wahrscheinlichkeiten gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ein temp. Ausgaberaster fuer die Nachbarschafts-Wahrscheinlichkeiten ({@link MultiGrid})</b><br>
 *     Hierin werden die ueber die Nachbarschaftszellen und Nachbarschaftsregression errechneten
 *     Wahrscheinlichkeiten gespeichert.
 *     Das Raster muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * <li><b>Ausgabeliste fuer die Iterations-Variablen ({@link ListProperty ListProperty[Double]})</b><br>
 *     Speichert die aktuellen Werte der Iterationsvariablen ITER(U).</li>
 * <li><b>Ausgaberaster fuer die LU zu jedem Zeitpunkt ({@link MultiGrid})</b><br>
 *     Hierin wird die fuer jeden Zeitschritt modellierte LU-Konfiguration gespeichert.<br>
 *     Das MultiGrid muss dieselbe Dimension besitzen, wie das Ausgangsszenario.</li>
 * </ol>
 * Die LU-Wahrscheinlichkeiten P1 (fuer Driving Forces) und P2 (fuer Nachbarschaft)
 * werden jeweils ueber ein <i>Binary Logit Model</i> errechnet und gehen anteilig
 * in die Gesamtwahrscheinlichkeit TPROB ein:
 * <center>
 *    TPROB(i,U) = P1(i,U) * (1-NBH_WEIGHT(U)) + P2(i,U) * NBH_WEIGHT + ELAS(U) - ITER(U)
 * </center><br>
 * Die Elastizitaet ELAS(U) wird nur eingerechnet, wenn sich die Zelle i im vorangegebenen
 * Zeitschritt bereits unter der Landnutzung U befindet.
 * In jedem Zeitschritt wird solgange iteriert, bis die vorgegebenen Bedarfe fuer
 * jeden LU-Typ (bis auf die Abweichungen <i>a</i> und <i>s</i>) erfuellt sind.
 * Nach jedem Iterationsschritt werden die ITER(U)-Variablen (einzige variable
 * Groesse innerhalb einer Zeitschritt-Iteration!) auf Basis der
 * Bedarfsabweichnung <i>DEMAND(U) - COVER(U)</i> angepasst, um die
 * Gesamtwahrscheinlicheiten "in eine geeignete Richtung" zu lenken.
 * <br><br>
 * <b><u>BEACHTE</u>:</b><br>
 * Zur Zeit werden folgende in CLUE integriert Regeln noch nicht unterstuetzt:
 * <ul>
 * <li>Area Restrictions fuer bestimmte Wechsel LU1 -> LU2 (Angabe ueber LU-Conversion
 *     Matrix zusammen mit speziellen Eingabe-Rastern)</li>
 * <li>Aufteilung des Untersuchungsgebiets in mehrere Regionen</li>
 * </ul>
 * @see ClueModelContentManager
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ClueModel_Original extends AbstractStepModel {
  /** Konstante fuer die Toleranz-Art "prozentual". */
  public static final int TOLTYPE_PCT = 0;
  /** Konstante fuer die Toleranz-Art "absolut in Bedarf-Einheit". */
  public static final int TOLTYPE_ABS = 1;

  /** Speichert den ContentManager fuer das Modell.
   *  @see ClueModelContentManager */
  protected ClueModelContentManager contManager;

  private ClueModelGUI  modelGUI = new ClueModelGUI();

  //********** Lese/Schreibrechte, die fuer den gesamten ***********
  //********** Modellanlauf gehalten werden muessen      ***********
  private PropertyReadAccess  RA_baseLU       = null; // Base Scenario
  private PropertyReadAccess  RA_baseLUHist   = null; // Base LU History
  private PropertyReadAccess  RA_luConv       = null; // LU-Conversion-Matrix
  private PropertyReadAccess  RA_luElas       = null; // LU-Conversion-Elasticity
  private PropertyReadAccess  RA_areaRestr    = null; // Area Restrictions
  private PropertyReadAccess  RA_luDemand     = null; // LU-Bedarfe
  private PropertyReadAccess  RA_sdfGrids     = null; // Static Driving Forces
  private PropertyReadAccess  RA_ddfList      = null; // List of dynmic Driving Forces
  private PropertyReadAccess  RA_ddfGrids     = null; // Dynamic Driving Forces
  private PropertyReadAccess  RA_regrRes      = null; // Regression Results
  private PropertyReadAccess  RA_nbhWeights   = null; // Gewichtung der Nachbarschaft pro LU-Typ
  private PropertyReadAccess  RA_nbhSettings  = null; // Definition der Umgebung pro LU-Typ
  private PropertyReadAccess  RA_nbhRegrRes   = null; // Regression Results fuer Nachbarschaft
  private PropertyWriteAccess WA_outActLU     = null; // speichert die aktuell sim. LU
  private PropertyWriteAccess WA_outActLUHist = null; // speichert die aktuell sim. LU Historie
  private PropertyWriteAccess WA_outLUProb    = null; // speichert die LU-Wahrscheinlichkeiten
  private PropertyWriteAccess WA_outNbhProb   = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private PropertyWriteAccess WA_outIter      = null; // speichert die Iterationsvariablen
  private PropertyWriteAccess WA_outStepLU    = null; // speichert die LU fuer jeden Zeitpunkt

  //**************** Variablen mit denen gearbeitet wird *******************
  // Modell-Parameter
  private int    luCount    = 0;    // Anzahl an LU-Typen
  private int    dfCount    = 0;    // Anzahl an Driving Forces
  private int    ddfCount   = 0;    // Anzahl an dynamischen Driving Forces
  private int    areaWidth  = 0;    // Breite des Untersuchungsgebiets (Rasterzellen)
  private int    areaHeight = 0;    // Hoehe des Untersuchungsgebiets (Rasterzellen)
  private int    areaMinX   = 0;    // Start-Index des Rasters in X-Richtung
  private int    areaMinY   = 0;    // Start-Index des Rasters in Y-Richtung
  private int    areaMaxX   = 0;    // End-Index des Rasters in X-Richtung (EXKLUSIVE!)
  private int    areaMaxY   = 0;    // End-Index des Rasters in Y-Richtung (EXKLUSIVE!)
  private int    demandTolType    = 0;    // Art der Abweichungsangaben
  private double demandTolSingle  = 0.0;  // erlaubte Abweichung fuer jeden einzelnen LU-Typ vom Bedarf
  private double demandTolAverage = 0.0;  // erlaubte Durchschnitts-Abweichung fuer das Untersuchungsgebiet
  private double demandConv = 0.0;  // Umrechnungsfaktor Bedarfseingeit -> qm
  private double cellSize   = 0;    // Flaeche einer Rasterzelle (in Bedarfseinheit)
  // Eingabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            baseLU       = null; // Base Scenario
  private WritableGrid            baseLUHist   = null; // Base LU History
  private PropertyReadAccess      luConv       = null; // LU-Conversion-Matrix
  private ListPropertyReadAccess  luElas       = null; // LU-Conversion-Elasticity
  private WritableGrid            areaRestr    = null; // Area Restrictions
  private PropertyReadAccess      luDemand     = null; // LU-Bedarfe
  private MultiGrid               sdfGrids     = null; // Static Driving Forces
  private ListPropertyReadAccess  ddfList      = null; // Dynamic Driving Force Types
  private MultiGrid               ddfGrids     = null; // Dynamic Driving Forces
  private int[]                   ddfStartIdx  = null; // Listenindex an dem ein dyn. DF beginnt
  private PropertyReadAccess      regrRes      = null; // LU-Bedarfe
  private ListPropertyReadAccess  nbhWeights   = null; // Gewichtung der Nachbarschaft pro LU-Typ
  private PropertyReadAccess      nbhSettings  = null; // Definition der Umgebung pro LU-Typ
  private PropertyReadAccess      nbhRegrRes   = null; // Regression Results fuer Nachbarschaft
  private int                     nbhRadiusX   = 0;    // horizontaler Radius der betrachteten Nachbarschaft
  private int                     nbhRadiusY   = 0;    // vertikaler Radius der betrachteten Nachbarschaft
  private double[][]              luCover      = null; // LU-Bedeckung der Typen ueber die Zeit in Bedarfseinheit
  private double                  totCover     = 0;    // Gesamte Landbedeckungsflaeche (ohne NoData!)
  private Random                  rand         = null; // Zufallsgenerator
  private double[]                iterSpeed    = null; // Geschwindigkeit fuer die ITER-Variablen
  // Ausgabe-Daten, die ueber den gesamten Modellablauf benoetigt werden
  private WritableGrid            outActLU     = null; // speichert die aktuell sim. LU
  private WritableGrid            outActLUHist = null; // speichert die aktuell sim. LU History
  private MultiGrid               outLUProb    = null; // speichert die LU-Wahrscheinlichkeiten
  private MultiGrid               outNbhProb   = null; // speichert die Nachbarschafts-Wahrscheinlichkeiten
  private ListPropertyWriteAccess outIter      = null; // speichert die Iterationsvariablen
  private MultiGrid               outStepLU    = null; // speichert die LU fuer jeden Zeitpunkt

  /**
   * Erzeugt eine neue Instanz des CLUE-Modells.
   */
  public ClueModel_Original() {
    super( new ClueModelContentManager() );
    contManager = (ClueModelContentManager)super.contManager;
    // Iterations-Ereignisse der Event-Liste hinzufuegen
    events.add(ModelIterationStepStartedEvent.class);
    events.add(ModelIterationStepFinishedEvent.class);
  }


  ////////////////////////////////////////////////////////////////////////
  ///////////////   Implementierung von AbstractXuluModel   //////////////
  ////////////////////////////////////////////////////////////////////////
  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
   * werden Referenzen auf einen Teil der Ressourcen in lokalen Variablen
   * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
   * zu koennen.
   */
  public void performModelInit() {
    if ( contManager == null )
      return;
    // Zugriffsrechte aus Ressourcen/Propertys holen
    RA_baseLU       = contManager.getBaseScenario().getReadAccess(this);
    RA_baseLUHist   = contManager.getLUHistory().getReadAccess(this);
    RA_luConv       = contManager.getLUConversionMatrix().getReadAccess(this);
    RA_luElas       = contManager.getConversionElasticity().getReadAccess(this);
    RA_areaRestr    = contManager.getAreaRestrictions().getReadAccess(this);
    RA_luDemand     = contManager.getDemandScenario().getReadAccess(this);
    RA_sdfGrids     = ((ListProperty)contManager.getStaticDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    if ( contManager.getDynamicDrivingForceTypes() != null ) {
      RA_ddfList      = contManager.getDynamicDrivingForceTypes().getReadAccess(this);
      RA_ddfGrids     = ((ListProperty)contManager.getDynamicDrivingForces().getProperty(MultiGrid.PROP_GRIDS)).getReadAccess(this);
    } else {
      RA_ddfList      = null;
      RA_ddfGrids     = null;
    }
    RA_regrRes      = contManager.getRegressionResults().getReadAccess(this);
    if ( contManager.getNeighborhoodWeights() != null ) {
      RA_nbhWeights  = contManager.getNeighborhoodWeights().getReadAccess(this);
      RA_nbhSettings = contManager.getNeighborhoodSettings().getReadAccess(this);
      RA_nbhRegrRes  = contManager.getNeighborhoodRegression().getReadAccess(this);
      WA_outNbhProb  = ((ListProperty)contManager.getNeighborhoodProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    } else {
      RA_nbhWeights  = null;
      RA_nbhSettings = null;
      RA_nbhRegrRes  = null;
      WA_outNbhProb  = null;
    }
    WA_outActLU     = contManager.getActLUResult().getWriteAccess(this);
    WA_outActLUHist = contManager.getActLUHistory().getWriteAccess(this);
    WA_outLUProb    = ((ListProperty)contManager.getLUProbabilities().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);
    WA_outIter      = contManager.getIterationVariables().getWriteAccess(this);
    WA_outStepLU    = ((ListProperty)contManager.getStepLUResults().getProperty(MultiGrid.PROP_GRIDS)).getWriteAccess(this);

    // Variablen belegen mit denen gearbeitet wird
    this.stepCount   = contManager.getStepCount().getOneTimeReadAccess().getValueAsInt(); // WICHTIG!!
    this.dfCount     = contManager.getStaticDrivingForces().getGridCount();
    this.demandTolType    = contManager.getDemandToleranceType().getOneTimeReadAccess().getValueAsInt();
    this.demandTolSingle  = contManager.getSingleDemandTolerance().getOneTimeReadAccess().getValueAsDouble();
    this.demandTolAverage = contManager.getAverageDemandTolerance().getOneTimeReadAccess().getValueAsDouble();
    this.demandConv  = contManager.getConversionFactor().getOneTimeReadAccess().getValueAsDouble();
    this.luCount     = contManager.getLUConversionMatrix().getSize(0);
    this.baseLU      = (WritableGrid)RA_baseLU.getValue();
    this.areaWidth   = baseLU.getWidth();
    this.areaHeight  = baseLU.getHeight();
    this.areaMinX    = baseLU.getMinX();
    this.areaMinY    = baseLU.getMinY();
    this.areaMaxX    = baseLU.getMinX() + areaWidth;
    this.areaMaxY    = baseLU.getMinY() + areaHeight;
    this.cellSize    = baseLU.getCellWidth() * baseLU.getCellHeight() / demandConv;
    this.baseLUHist  = (WritableGrid)RA_baseLUHist.getValue();
    this.luConv      = RA_luConv;
    this.luElas      = (ListPropertyReadAccess)RA_luElas;
    this.areaRestr   = (WritableGrid)RA_areaRestr.getValue();
    this.luDemand    = RA_luDemand;
    this.sdfGrids    = contManager.getStaticDrivingForces();
    this.ddfList     = (RA_ddfList != null) ? (ListPropertyReadAccess)RA_ddfList : null;
    this.ddfCount    = (ddfList != null)    ? ddfList.getCount() : 0;
    this.ddfGrids    = (ddfList != null)    ? contManager.getDynamicDrivingForces() : null;
    this.regrRes     = RA_regrRes;
    this.nbhWeights  = (RA_nbhWeights != null) ? (ListPropertyReadAccess)RA_nbhWeights : null;
    this.nbhSettings = (nbhWeights != null)    ? RA_nbhSettings : null;
    this.nbhRadiusX  = (nbhWeights != null)    ? (contManager.getNeighborhoodSettings().getSize(0)-1)/2 : 0;
    this.nbhRadiusY  = (nbhWeights != null)    ? (contManager.getNeighborhoodSettings().getSize(1)-1)/2 : 0;
    this.nbhRegrRes  = (nbhWeights != null)    ? RA_nbhRegrRes : null;
    this.outNbhProb  = (nbhWeights != null)    ? contManager.getNeighborhoodProbabilities() : null;
    for (int i=0; outNbhProb!=null && outNbhProb.getGridCount()<luCount; i++)
      outNbhProb.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten
    this.outActLU     = (WritableGrid)WA_outActLU.getValue();
    this.outActLUHist = (WritableGrid)WA_outActLUHist.getValue();
    this.outLUProb    = contManager.getLUProbabilities();
    for (int i=0; outLUProb.getGridCount()<luCount; i++)
      outLUProb.addGrid(); // Liste muss ein Raster fuer jeden LU-Typ enthalten

    this.outIter      = (ListPropertyWriteAccess)WA_outIter;
    // Lister der Iter-Variablen auf benoetigte Groesse auffuellen
    for (;outIter.getCount()<luCount;)
      outIter.addValue(0.0f);
    this.outStepLU    = contManager.getStepLUResults();

    // Bei Toleranz-Art "prozentual" werden die Werte in Prozent
    // angegeben (intern wird aber mit Abweichungen zwischen 0 und 1 gerechnet!)
    if ( demandTolType == TOLTYPE_PCT ) {
      this.demandTolSingle  /= 100.0;
      this.demandTolAverage /= 100.0;
    }

    // Die Liste der dyn. DF ist blockweise nach DF grupperiert
    // Es koennen mehr Zeitschritt-Raster angegeben werden, als simuliert
    // werden.
    // Im Array 'ddfStartIdx' wird fuer die dynamischen DF gespeichert, an
    // welcher Stelle in der Liste der Block beginnt.
    // -1 bedeutet, dass der DF ein statische DF ist!
    this.ddfStartIdx = new int[dfCount];
    for (int i=0; i<dfCount; i++)
      ddfStartIdx[i] = - 1;
    for (int i=0; i<ddfCount; i++)
      ddfStartIdx[i] = i*ddfGrids.getGridCount() / ddfCount;

    // "Geschwindigkeit" fuer die Iterationsvariablen
    this.iterSpeed = new double[luCount];

    // Speichert die Flaechendeckung (in der Bedarfseinheit, z.B. ha)
    // fuer die einzelnen LU-Typen zu jedem Zeitpunkt
    // --> Die initialen Werte stehen in der 0-Zeile der Demand-Matrix
    this.luCover   = new double[luCount][stepCount+1];
    this.totCover = 0.0;
    statusOut.print("InitCover: ");
    for (int lu=0; lu<luCount; lu++) {
      luCover[lu][0] = luDemand.getValueAsDouble(lu, 0);
      totCover += luCover[lu][0]; // Gesamt-Flaeche des Untersuchungsgebiets
      statusOut.print(luCover[lu][0]+"   ");
    }
    statusOut.println("   >> TotalCover "+totCover);

    // Ausgabe-Raster leeren
    for (int y=outActLU.getMinY(); y<outActLU.getMinY()+outActLU.getHeight(); y++)
      for (int x=outActLU.getMinX(); x<outActLU.getMinX()+outActLU.getWidth(); x++)
        outActLU.setRasterSample(Float.NaN,x,y);

    // spezielle GUI initialisieren
    modelGUI.init(this);
  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    // Ressourcen wieder freigeben
    releaseAccess(RA_baseLU);
    releaseAccess(RA_baseLUHist);
    releaseAccess(RA_luConv);
    releaseAccess(RA_luElas);
    releaseAccess(RA_areaRestr);
    releaseAccess(RA_luDemand);
    releaseAccess(RA_sdfGrids);
    releaseAccess(RA_ddfList);
    releaseAccess(RA_ddfGrids);
    releaseAccess(RA_regrRes);
    releaseAccess(RA_nbhWeights);
    releaseAccess(RA_nbhSettings);
    releaseAccess(RA_nbhRegrRes);
    releaseAccess(WA_outActLU);
    releaseAccess(WA_outActLUHist);
    releaseAccess(WA_outLUProb);
    releaseAccess(WA_outNbhProb);
    releaseAccess(WA_outIter);
    releaseAccess(WA_outStepLU);
    closeLog();
  }

  /**
   * Fuehrt einen Schritt des Modellablaufs durch.
   * @param stepNo zu modellierender Schritt (beginnend bei 1!)
   */
  public void performModelStep(int stepNo) {
    statusOut.println("Started step ".concat(String.valueOf(stepNo)));
//    openLog(stepNo);

    // Raster festlegen, in dem die LU des vorangegangenen
    // Schritts zu finden ist
    WritableGrid lastLUGrid = (stepNo==1) ? baseLU : outStepLU.getGrid(stepNo-2);
    // Raster festlegen, in dem die aktuelle LU History
    // zu finden ist
    WritableGrid luHistGrid = (stepNo==1) ? baseLUHist : outActLUHist;

    // Merken, ob an einer Stelle bereits eine U-Zelle in einen anderen
    // Typ gewechselt hat
    boolean[] alreadyLUCCFromType = new boolean[luCount];
    // Merken, ob an einer Stelle bereits eine U-Zelle neu entstanden ist
    boolean[] alreadyLUCCToType = new boolean[luCount];

    // LU-Wahrscheinlichkeiten berechnen
    calculateLUProbabilities(stepNo);

    // Nachbarschafts-Wahrscheinlichkeiten berechnen
    calulateNeighborhoodProbabilities(stepNo,lastLUGrid);

    // Randomizer fuer diesen Zeitschritt erzeugen
    rand = new Random(System.currentTimeMillis());

    for (int iterNo=1; !checkDemandCompliedAndCalculateIterVar(stepNo,iterNo); iterNo++) {
      this.fireModelEvent( new ModelIterationStepStartedEvent(this,stepNo,iterNo) );
      if ( modelGUI != null )
        modelGUI.setIterationNumber(stepNo+"/"+iterNo);
      // Arrays fuer Iterationsschritt initialisieren
      for (int lu = 0; lu < luCount; lu++) {
        // noch kein Wechsel von einem LU-Typ
        alreadyLUCCFromType[lu] = false;
        // noch kein Wechsel zu einem LU-Typ
        alreadyLUCCToType[lu] = false;
        // Flaechen der LU-Typen fuer diesen Zeitschritt
        luCover[lu][stepNo] = 0.0;
        // Speed fuer die Iterationsvariablen initialisieren
        // --> liegt die Nachfrage bei weniger als 5% des U'gebiets
        //     wird langsamer begonnen
        if ( luDemand.getValueAsDouble(lu,stepNo)/totCover < 0.05 )
          iterSpeed[lu] = 0.01;
        else
          iterSpeed[lu] = 0.05;
      }

      // Fuer alle Zellen des Untersuchungsgebiets
      for (int y = areaMinY; y < areaMaxY; y++)
        for (int x = areaMinX; x < areaMaxX; x++) {
          checkBreakingCommands();

          // Rasterzellen ausserhalb des Untersuchungsgebiets komplett
          // ignorieren
          // Grund: NoData-Zellen (Float.NaN) werden beim Cast in int zu 0,
          //        also zur Landnutzung 0!!
          if ( checkNoData(x,y,areaRestr) || checkNoData(x,y,baseLU) )
            continue;

          // Nutzungsdauern erhoehen
          if (iterNo==1)
            outActLUHist.setRasterSample(luHistGrid.getRasterSampleAsFloat(x, y) + 1, x, y);

          // vorangegangene LU merken
          int lastLU = lastLUGrid.getRasterSampleAsInt(x, y);

          // Beste gefundene LU wird mit aktueller LU initialisiert
          int bestLU = lastLU;
          double bestLUProb = 0.0;

          // Pruefen, ob Zelle die LU wechseln darf
          if (checkLUCCGeneralAllowed(x, y, stepNo, lastLU, alreadyLUCCToType[lastLU])) {
            // Alle LU-Typen auf Wechsel pruefen
            for (int lu = 0; lu < luCount; lu++) {
              // Pruefen, ob Zelle in den betrachteten Typ wechseln darf
              if (checkLUCCAllowed(x, y, stepNo, lastLU, lu, alreadyLUCCFromType[lu])) {
                // Gesamt-Wahrscheinlichkeit berechnen
                // a) DF-W'keit und Nachbarschaft-W'keit gehen anteilig ein
                // b) Elastizitaet geht nur ein, wenn Zelle bereits fuer den
                //    LU-Typ genutzt wird
                // c) ITER geht negativ ein, da es auf Basis der Abweichung
                //    luCover-luDemand berechnet wird. Ein negatives Iter
                //    kennzeichnet also einen Mangel und muss die Gesamtw'keit
                //    erhoehen
                double baseProb  = outLUProb.getGrid(lu).getRasterSampleAsDouble(x, y);
                double nbhProb   = (nbhWeights == null) ? 0.0 : outNbhProb.getGrid(lu).getRasterSampleAsDouble(x, y);
                double nbhWeight = (nbhWeights == null) ? 0.0 : nbhWeights.getValueAsInt(lu);
                double elas      = (lu == lastLU && luElas.getValueAsDouble(lu) < 1) ? luElas.getValueAsDouble(lu) : 0.0;
                // LU-Wahrscheinlichkeit zusammensetzen
                double totalProb = baseProb * (1-nbhWeight)
                                 + nbhProb  * nbhWeight
                                 + elas
                                 - outIter.getValueAsFloat(lu);

                // betrachteter Typ ist geeigneter als der bisher beste
                if (totalProb > bestLUProb) {
                  bestLU = lu;
                  bestLUProb = totalProb;
                }
              }
            }
            // Merken wenn wirklich ein Wechsel stattfindet
            if (bestLU != lastLU) {
              alreadyLUCCFromType[lastLU] = true;
              alreadyLUCCToType[bestLU] = true;
            }
          }

          // LU fuer diesen Iterationschritt setzen
          outActLU.setRasterSample( (float) bestLU, x, y);
          // Flaechendeckung fuer LU-Typ erhoehen
          luCover[bestLU][stepNo] += cellSize;
        }
      this.fireModelEvent( new ModelIterationStepFinishedEvent(this,stepNo,iterNo) );
    }

    // Liste der Zeitschritt-Ausgaberaster vergroessern
    outStepLU.addGrid();

    // LU in Ausgaberaster des Zeitschritts uebernehmen und ggf.
    // Nutzungsdauer initialisieren
    for (int y=areaMinY; y<areaMaxY; y++)
      for (int x=areaMinX; x<areaMaxX; x++) {
        checkBreakingCommands();
        Float lastLU = (Float)lastLUGrid.getRasterSample(x,y);
        Float actLU  = (Float)outActLU.getRasterSample(x,y);
        // Wenn die Landnutzung gewechselt hat, wird die Historie zurueck
        // gesetzt
        if ( actLU != lastLU )
          outActLUHist.setRasterSample((float)0,x,y);

        // LU in Ausgabe-Raster des simulierten Schritts uebernehmen
        outStepLU.getGrid(stepNo-1).setRasterSample(actLU,x,y);
      }

    statusOut.println("Ended step ".concat(String.valueOf(stepNo)));
    closeLog();
  }

  /**
   * Liefert eine CLUE-spezifische GUI, in der die Abweichungen waehrend der
   * Iteration angezeigt werden und ueber die die Toleranzen eingestellt
   * werden koennen
   * @return Instannz von {@link ClueModelGUI}
   */
  public Component getModelSpecificGUI() {
    return modelGUI;
  }

  //////////////////////////////////////////////////////////////////////
  ////////////////////////   Clue-Methoden   ///////////////////////////
  //////////////////////////////////////////////////////////////////////
  /**
   * Prueft, ob eine Rasterzelle mit <code>NoData</code> belegt ist
   * @param x     Raster-Koordinate
   * @param y     Raster-Koordinate
   * @param grid  zu pruefendes Raster
   */
  protected static boolean checkNoData(int x, int y, WritableGrid grid) {
    return Double.isNaN(grid.getRasterSampleAsDouble(x,y));
  }

  /**
   * Berechnet die auf den Driving Forces basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die LU-Wahrscheinlichkeiten".<br>
   * Werden keine dynamischen Driving Forces verwendet, geschieht die Berechnung
   * nur beim ersten Modellierungsschritt.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   */
  protected void calculateLUProbabilities(int stepNo) {
    // Wenn keine dynamischen Driving Forces verwendet werden
    // braucht die Berechnung nur fuer den ersten Schritt
    // durchgefuehrt werden
    if ( ddfCount == 0 && stepNo > 1 )
      return;

    statusOut.print("Calculate LU-Probabilities for step "+stepNo+"...");
    // Fuer Zellen des Untersuchungsgebiets
    for (int y=areaMinY; y<areaMaxY; y++)
      for (int x=areaMinX; x<areaMaxX; x++)
        // NotData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 ) {
          // Fuer alle LU-Typen
          for (int lu = 0; lu < luCount; lu++) {
            double betaSum = 0.0;
            // Fuer alle Driving Forces die Summe der Betas berechnen
            for (int df = 0; df < dfCount; df++) {
              checkBreakingCommands();
              // DF-Grid festlegen:
              // Bei dynamischem DF das jew. Zeitschritt-Raster aus der Liste
              // ansonsten das statische Raster
              WritableGrid dfGrid = null;
              if ( ddfStartIdx[df] >= 0 )
                dfGrid = this.ddfGrids.getGrid(ddfStartIdx[df]+stepNo-1);
              else
                dfGrid = this.sdfGrids.getGrid(df);
              // Beim ersten Driving Force wird die Summe mit der
              // Regressionskonstante initialisiert
              if (df == 0)
                betaSum = regrRes.getValueAsDouble(0, lu);
              // betaSum = BETA0(U) + Sum( BETA(df+1,U) * DF(df,x,y) )
              betaSum += regrRes.getValueAsDouble(df + 1, lu)
                        * dfGrid.getRasterSampleAsDouble(x, y);
            }
            double expBetaSum = Math.exp(betaSum);
            double prob = expBetaSum / (1 + expBetaSum);
            // errechnete W'keit im Raster speichern
            this.outLUProb.getGrid(lu).setRasterSample( (float) prob, x, y);
          }
        }
    statusOut.println(" done.");
  }

  /**
   * Berechnet die auf den Nachbarschaftsbeziehungen basierenden
   * Landnutzungs-Wahrscheinlichkeiten fuer einen Modell-Schritt und speichert
   * diese in der Ressource "temp. Ausgaberaster fuer die Nachbarschafts-Wahrscheinlichkeiten".<br>
   * Da sich die Landnutzung mit jedem Zeitschritt aendert geschieht dies
   * vor jedem Zeitschritt neu.<br>
   * Diese Methode macht nichts, wenn ohne Nachbarschafts-W'keiten simuliert wird,
   * also wenn keine Angabe fuer die "Neighborhood-Weights" gemacht wird.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param luGrid LU-Raster aus dem die Umgebung der jeweiligen Zellen
   *               ermittelt werden
   */
  protected void calulateNeighborhoodProbabilities(int stepNo, WritableGrid luGrid) {
    // Wenn ohne Nachbarschaftsbeziehungen simuliert wird
    if ( nbhWeights == null )
      return;

    statusOut.print("Calculate Neighborhood-Probabilities for step "+stepNo+"...");
    // Fuer Zellen des Untersuchungsgebiets
    for (int y=areaMinY; y<areaMaxY; y++)
      for (int x=areaMinX; x<areaMaxX; x++)
        // NoData-Bereich und Area Restrictions ignorieren
        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 )
          // Fuer alle LU-Typen
          for (int lu = 0; lu < luCount; lu++) {
            checkBreakingCommands();

            // Anzahl der Zellen in der Umgebung initialisieren
            int      totCellCount = 0;
            double[] luCellCount  = new double[luCount];

            // Umgebung der Zelle betrachten
            for (int dy=-nbhRadiusY; dy<=nbhRadiusY; dy++)
              if ( areaMinY <= y+dy && y+dy<areaMaxY )
                for (int dx=-nbhRadiusX; dx<=nbhRadiusX; dx++)
                  if ( areaMinX <= x+dx && x+dx<areaMaxX ) {
                    double lucc = luGrid.getRasterSampleAsDouble(x+dx,y+dy); // PERFORMANZ > nur einmal den Zellen-Wert ermitteln
//                    if ( !checkNoData(x+dx,y+dy,baseLU) ) {
                    if ( !Double.isNaN(lucc) ) {
                        // Alle Zellen in Umgebung zaehlen (jedoch nur diejenigen,
                        // die potentiell auch in die Bewertung eingehen koennen!)
                        double setting = nbhSettings.getValueAsDouble(dx+nbhRadiusX,dy+nbhRadiusY,lu);
                        if ( setting != 0.0 ) {
                          totCellCount++;
                          // Anzahl der Zellen eines bestimmten Typs aufsummieren
                          // (relativ zum in der Umgebung angegebenen Faktor)
//                          luCellCount[ luGrid.getRasterSampleAsInt(x+dx,y+dy) ] += setting;
                          luCellCount[ (int)lucc ] += setting;
                        }
                      }
                    }
              // Summe der Betas berechnen
              double betaSum = nbhRegrRes.getValueAsDouble(0,lu); // Regr.Konstante
              for (int i=0; i<luCount; i++) {
                double beta    = nbhRegrRes.getValueAsDouble(i + 1, lu);
                double nbhFact = (luCellCount[i]/totCellCount) / (luCover[i][stepNo-1]/totCover);
                betaSum += beta * nbhFact;
              }
              // Wahrscheinlichkeit berechnen
              double expBetaSum = Math.exp(betaSum);
              outNbhProb.getGrid(lu).setRasterSample((float)(expBetaSum/(1 + expBetaSum)),x,y);
          }
    statusOut.println(" done.");
  }


  /**
   * Prueft, ob ein genereller LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliessen die Area Restrictions einen Wechsel aus?<br>
   *     > Rasterwert(x,y) != 0
   *     </li>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet 1
   *     und steigende Bedarf)?<br>
   *     > ELAS(lastLU) == 1 und DEMAND(lastLU,t) >= DEMAND(lastLU,t-1)
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param alreadyLUCCToActualType Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel in den den aktuellen
   *                                Typ der Zelle (lastLU) vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCGeneralAllowed(int x, int y, int stepNo, int lastLU, boolean alreadyLUCCToActualType) {
    // Area Restrictions checken
    if ( areaRestr.getRasterSampleAsFloat(x,y) != 0 )
      return false;
// DER ERSTE TEIL IST KEINE ORIGINAL-CLUE-REGEL!
//    // Conversion Elasticity checken
//    // --> wenn Elasticity 1 und bereits an anderer Stelle eine Zelle des
//    //     bisherigen Typs entstanden ist, darf sie nicht wechseln
//    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
//    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
//    //     an anderer Stelle neue entstehen
//    if (  luElas.getValueAsDouble(lastLU) == 1.0
//       && ( alreadyLUCCToActualType ||
//           luDemand.getValueAsDouble(lastLU,stepNo) >= luCover[lastLU][stepNo-1] ) )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
    //     an anderer Stelle neue entstehen
    if (  luElas.getValueAsDouble(lastLU) == 1.0 &&
//          luDemand.getValueAsDouble(lastLU,stepNo) >= luCover[lastLU][stepNo-1] )
          luDemand.getValueAsDouble(lastLU,stepNo) >= luDemand.getValueAsDouble(lastLU,stepNo-1) )
      return false;

    return true;
  }

  /**
   * Prueft, ob ein bestimmter LU-Wechsel fuer eine Zelle erlaubt ist.
   * <ul>
   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet des
   *     potentiellen Wechseltyps 1 und sinkender Bedarf)?<br>
   *     > lastLU != newLU und ELAS(newLU) == 1 und DEMAND(newLU,t) <= DEMAND(newLU,t-1)
   *     </li>
   * <li>Schliesst die Conversion Matrix einen Wechsel aus (generell oder ueber
   *     Historie)?<br>
   *     > CONV(lastLU,newLU) == 0 oder<br>
   *     > CONV(lastLU,newLU) > 100 und HIST(x,y) < CONV(lastLU,newLU)-101 oder<br>
   *     > CONV(lastLU,newLU) < -100 und HIST(x,y) > -CONV(lastLU,newLU)-101
   *     </li>
   * </ul>
   * @param x      Raster-Koordinate
   * @param y      Raster-Koordinate
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
   * @param newLU  potentielle neue Bedeckung von (x,y)
   * @param alreadyLUCCFromNewType  Flag gibt an, ob in der laufenden Iteration
   *                                bereits ein LU-Wechsel vom potentiellen neuen
   *                                Typ in einen anderen Typ vorgenommen wurde.<br>
   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
   */
  protected boolean checkLUCCAllowed(int x, int y, int stepNo, int lastLU, int newLU, boolean alreadyLUCCFromNewType) {
// IST KEINE ORIGINAL-CLUE-REGEL!
//    // LU-Conversion-Elasticity des neuen Typs checken
//    // --> wenn Elasticity des neuen Typs 1 ist und an anderer Stelle bereits
//    //     eine Zelle des dieses Typs "verloren" ging, darf er hier
//    //     nicht entstehen
//    if (  alreadyLUCCFromNewType && luElas.getValueAsDouble(newLU) == 1.0 )
//      return false;

    // Conversion Elasticity checken
    // --> wenn Elasticity des neuen Typs "newLU" 1 und dessen Bedarf sinkt,
    //     darf kein Wechsel in diesen Typ stattfinden, denn dann muessten (um
    //     den Bedarf fuer "newLU" zu erreichen) an anderer Stelle newLU-Zellen
    //     entfernt werden
    if (  lastLU != newLU &&
          luElas.getValueAsDouble(newLU) == 1.0 &&
          luDemand.getValueAsDouble(newLU,stepNo) <= luCover[newLU][stepNo-1] )
//          luDemand.getValueAsDouble(newLU,stepNo) <= luDemand.getValueAsDouble(newLU,stepNo-1) )
      return false;

    // Wert aus LU-Conversion-Matrix ermitteln
    // Bemerke: "von" steht in den Zeilen (also Y, also Dim1)
    //          "nach" steht in den Spalten (also X, also Dim2)
    int conv = luConv.getValueAsInt(newLU,lastLU);
    // 1 --> LUCC generell erlaubt
    if ( conv == 1 )
      return true;
    // 0 --> LUCC generell verboten
    if ( conv == 0 )
      return false;

    // 100+t --> Wechsel erst nach t Schritten der vorangegangenen
    //           LU moeglich
    if ( conv > 100 ) {
//      int minLUHist = conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist >= minLUHist;
      int minLUHist = conv-101;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist >= minLUHist;
    }

    // -100-t --> Wechsel nicht mehr moeglich, wenn Zelle bereits
    //            t Schritte unter der vorangegangenen LU
    if ( conv < -100 ) {
//      int maxLUHist = -conv-100;
//      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//      return luHist < maxLUHist;
      int maxLUHist = -conv-101;
      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
      return luHist <= maxLUHist;
    }

    return true;
  }

  /**
   * Prueft, ob die aktuelle LU-Konfiguration die einzelenen Bedarfsanforderungen
   * (bis auf eine Abweichung) erfuellt und berechnet die Iterationsvariablen
   * neu.
   * @param stepNo modellierter Schritt (beginnend bei 1)
   * @param iterNo aktuelle Iteration (beginnend bei 1)
   */
  protected boolean checkDemandCompliedAndCalculateIterVar(int stepNo, int iterNo) {
    appendLog(iterNo);

    boolean demandComplied = true;
    // Abweichnungen berechnen
    double[] singleDevAbs = new double[luCount];
    double[] singleDevPct = new double[luCount];
    double   averageDevAbs = 0.0;
    double   averageDevPct = 0.0;
    double   totalDevAbs = 0.0;
    double   totalDevPct = 0.0;
    for (int lu = 0; lu < luCount; lu++) {
      // ##### Einzelabweichungen #####
      // Absolute Abweichung der Landbedeckung vom Bedarf berechnen
      double demand = luDemand.getValueAsDouble(lu, stepNo);
      singleDevAbs[lu]  = luCover[lu][stepNo] - demand;
      // Prozentuale Einzel-Abweichnung relativ zum Bedarf
      if (demand == 0)
        // ist der Bedarf 0, aber auch die Bedeckung, ist die Abw. 0%
        singleDevPct[lu] = luCover[lu][stepNo] == 0.0 ? 0.0 : 1.0;
      else
        singleDevPct[lu] = singleDevAbs[lu] / demand;

      // Abweichung ueber erlaubter Toleranz?
      double singleDev = (demandTolType == TOLTYPE_PCT) ? singleDevPct[lu] : singleDevAbs[lu];
      if (modelGUI!=null)
        modelGUI.setDeviation(lu,singleDev);
      if (Math.abs(singleDev) > demandTolSingle)
        demandComplied = false;

      // Abweichung des LU-Typs ausgeben
      appendLog(singleDevAbs[lu],singleDevPct[lu],luDemand.getValueAsDouble(lu, stepNo),luCover[lu][stepNo]);

      // ##### Gesamtabweichungen aufsummieren #####
      totalDevAbs += Math.abs(singleDevAbs[lu]);
    }

    // Prozentuale Gesamt-Abweichnung relativ zum Gesamt-Untersuchungsgebiet
    totalDevPct = totalDevAbs / totCover;
    if (modelGUI!=null)
      modelGUI.setDeviation(luCount,(demandTolType == TOLTYPE_PCT) ? totalDevPct : totalDevAbs);

    // Durchschnittliche Abweichnung pro LU-Typ
    averageDevAbs = totalDevAbs / luCount;

    // Durchschnittliche Abweichnung pro LU-Typ relativ zum Gesamt-Untersuchungsgebiet
    averageDevPct = averageDevAbs / totCover;

    // Abweichung ueber erlaubter Toleranz?
    double avgDev = (demandTolType == TOLTYPE_PCT) ? averageDevPct : averageDevAbs;
    if (modelGUI!=null)
      modelGUI.setDeviation(luCount+1,avgDev);
    if ( Math.abs(avgDev) > demandTolAverage )
      demandComplied = false;

    appendLog(totalDevAbs,totalDevPct);

    // Iterationsvariablen neu berechnen
    for (int lu = 0; lu < luCount; lu++) {
      // Um Alternieren zwischen 2 Konfiguration zu verhindern
      // eine kleine Zufallszahl hinzuaddieren (zwischen -0,005 und 0,005)
      float randomValue = (rand.nextFloat()-0.5f)/100;

      // --> Im ersten Schritt erhalten alle Iterationsvariablen den Wert 0,5
      double iter = outIter.getValueAsFloat(lu);
      if (iterNo == 1)
        iter = 0.5f+randomValue;
//        iter = 0.5 + new float[] {0.002f,0f,0.002f,0f,0f,-0.003f}[lu];
//        iter = 0.5 + new float[] {0.002f,0f,0.001f,0f,0f,-0.002f}[lu];
      else {
//        // Anteil der Einzelabweichnung an der Gesamtabweichnung
////        double dev = (totalDevAbs!=0) ? singleDevAbs[lu] / totalDevAbs : 0.0;
//        double dev  = singleDevPct[lu];
//        iter = outIter.getValueAsFloat(lu) // alter Iter-Wert
//             + dev // Anteil der Abweichung zum Bedarf
//             + randomValue;
        /////////////////////////////////////////
        // Original-Iter-Berechnung aus CLUE
        iterSpeed[lu] += 0.00005;
        iter += (singleDevAbs[lu]/totCover)/(iterSpeed[lu]*(rand.nextInt(101)+1));

        if (iter > 2f)
          iter = 0.5f;
        if ( Math.abs(iter) < 0.001 )
          iter = 0.01f * Math.signum(iter);
        if (iter < -0.5f)
          iter = -0.4f;
        /////////////////////////////////////////
      }
      // Iter speichern
      outIter.setValue((float)iter, lu);
      appendLog(outIter.getValueAsFloat(lu));
    }

    appendLog(" > Demand "+(demandComplied ? "" : "not ")+"complied");
    appendLog("\n");
    return demandComplied;
  }

  //////////////////////////////////////////////////////////////////////
  /////////////////////   Clue-spezifische GUI   ///////////////////////
  //////////////////////////////////////////////////////////////////////

  /**
   * Diese Klasse stellt eine modellspezifische GUI-Komponente fuer das
   * {@link ClueModel CLUE-Modell} dar.
   * In dieser wird fuer jeden modellierten LU-Typ ein {@linkplain ExpansionBar Status-Balken}
   * fuer die Abweichnung vom geforderten Bedarf angezeigt. Sowie je ein Balken
   * fuer die aktuelle Gesamt- und Durchschittsabweichung. Je nach gewaehlter
   * Abweichungsart (prozentual oder absolute Angaben), werden Abweichungen
   * von +/- 100% oder +/- 10000 Bedarfseinheiten angezeigt.<br>
   * Darueber hinaus enthaelt sie je einen Regler, ueber den die maximal
   * erlaubte Abweichung von den Bedarfen (Einzelabweichung der LU-Typen und
   * Durchschnittsabweichung) eingestellt werden kann.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ClueModelGUI extends JComponent {
    private       double MIN = -1.0;
    private       double MAX = 1.0;
    private final double TOL = 0.2;
    private       double SLIDER_FACT = 1000;
    // Format fuer Wert-Anzeige des Sliders
    private DecimalFormat sliderPctFormat = new DecimalFormat("0.0%");
    private DecimalFormat sliderDecFormat = new DecimalFormat("0.0");
    // Format fuer Wert-Anzeige im Balken
    private DecimalFormat statusPctFormat = new DecimalFormat("0.0000%");
    private DecimalFormat statusDecFormat = new DecimalFormat("0.0");
    // Format fuer Wert-Anzeige unterhalb der Balken (Max/Min/Middle)
    private DecimalFormat infoPctFormat   = new DecimalFormat("0.0%");
    private DecimalFormat infoDecFormat   = new DecimalFormat("0.0");

    private DeviationContainer  divCont          = null;
    private JSlider             singleDivSlider  = null;
    private JSlider             averageDivSlider = null;
    private ClueModel_Original  model            = null;
    private DeviationLabel      singleDivLabel   = new DeviationLabel("Erlaubte Bedarfsabweichung pro LU-Typ:",sliderPctFormat);
    private DeviationLabel      averageDivLabel  = new DeviationLabel("Erlaubte durchschnittliche Bedarfsabweichung:",sliderPctFormat);
    private Font                labelFont        = singleDivLabel.getFont().deriveFont(Font.BOLD | Font.CENTER_BASELINE);
    private boolean             alreadyPacked    = false;
    private JLabel              iterationLabel   = null;

    /**
     * Erzeugt eine neue Clue-GUI. Diese wird jedoch erst nach Aufruf
     * der {@link #init(ClueModel_Original)}-Methode komplett erzeugt, da erst dann
     * die Anzahl der angezeigten Balken (= Anzahl LU-Typen) feststeht.
     */
    public ClueModelGUI() {
      super();
      this.setLayout( new SpringLayout() );
      // Label in dem die Iterationsnummer angezeigt wird
      iterationLabel = new JLabel(" ");
      iterationLabel.setBorder( BorderFactory.createLoweredBevelBorder() );
      // Slider/Label fuer Abweichnug der einzelnen Typen
      singleDivSlider  = new JSlider(JSlider.HORIZONTAL,0,(int)(SLIDER_FACT*MAX),(int)(SLIDER_FACT*TOL));
      singleDivSlider.setPaintTicks(true);
      singleDivSlider.setPaintLabels(true);
      singleDivSlider.setPaintTrack(true);
      singleDivSlider.setEnabled(false);
      singleDivSlider.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          double tol = singleDivSlider.getValue() / SLIDER_FACT;
          if ( model!=null )
            model.demandTolSingle = tol;
          if ( singleDivLabel!=null )
            singleDivLabel.setValue(tol);
          if ( divCont!=null )
            divCont.setTolerance(tol);
        }
      });
      singleDivLabel.setFont(labelFont);
      singleDivLabel.setValue(singleDivSlider.getValue()/SLIDER_FACT);

      // Slider/Label fuer Abweichnug der einzelnen Typen
      averageDivSlider  = new JSlider(JSlider.HORIZONTAL,0,(int)(SLIDER_FACT*MAX),(int)(SLIDER_FACT*TOL));
      averageDivSlider.setPaintTicks(true);
      averageDivSlider.setPaintLabels(true);
      averageDivSlider.setPaintTrack(true);
      averageDivSlider.setEnabled(false);
      averageDivSlider.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          double tol = averageDivSlider.getValue() / SLIDER_FACT;
          if ( model!=null )
            model.demandTolAverage = tol;
          if ( averageDivLabel!=null )
            averageDivLabel.setValue(tol);
        }
      });
      averageDivLabel.setFont(labelFont);
      averageDivLabel.setValue(averageDivSlider.getValue()/SLIDER_FACT);

      this.add( averageDivLabel );
      this.add( averageDivSlider );
      this.add( singleDivLabel );
      this.add( singleDivSlider );
      this.add( iterationLabel );
      SpringUtilities.makeCompactGrid(this,5,1,0,0,0,5);
    }

    /**
     * Initialisiert die GUI fuer eine <b>initialisierte</b> Modell-Instanz.
     * Wird zum Ende {@link ClueModel#performModelInit()} aufgerufen.
     */
    public void init(ClueModel_Original model) {
      this.model = model;
      if ( divCont != null )
        remove(divCont);
      SLIDER_FACT = (model.demandTolType == TOLTYPE_PCT) ? 1000.0 : 1.0;
      MIN         = (model.demandTolType == TOLTYPE_PCT) ? -1.0 : -Math.pow(10,(int)Math.log10(model.totCover))/1000;
      MAX         = (model.demandTolType == TOLTYPE_PCT) ?  1.0 : -MIN;
      averageDivSlider.setMaximum((int)(MAX*SLIDER_FACT));
      averageDivLabel.setNumberFormat( model.demandTolType == TOLTYPE_PCT ? sliderPctFormat : sliderDecFormat );
      singleDivSlider.setMaximum((int)(MAX*SLIDER_FACT));
      singleDivLabel.setNumberFormat( model.demandTolType == TOLTYPE_PCT ? sliderPctFormat : sliderDecFormat );

      String[] divDesc = new String[model.luCount+2];
      for (int i=0; i<model.luCount; i++)
        divDesc[i] = String.valueOf(i);
      divDesc[model.luCount]   = "Total";
      divDesc[model.luCount+1] = "Average";

      divCont = new DeviationContainer(
                         divDesc.length,
                         divDesc,
                         MIN,
                         MAX,
                         0,
                         model.demandTolSingle,
                         model.demandTolType == TOLTYPE_PCT ? statusPctFormat : statusDecFormat,
                         model.demandTolType == TOLTYPE_PCT ? infoPctFormat   : infoDecFormat
      );
      divCont.setBarDescHeader("LU-Typ");
      divCont.setBarHeader("Aktuelle Bedarfsabweichung");
      this.add(divCont,4);

      // Wert des Balkens darf erst hier (nach Neu-Erstellung des
      // Abweichungs-Containers) neu gesetzt werden, da beim Setzen
      // des neuen Wertes automatisch die Toleranzen der Abweichungsbalken
      // aktualisiert werden
      setSingleTolerance(model.demandTolSingle);
      setAverageTolerance(model.demandTolAverage);

      SpringUtilities.makeCompactGrid(this,6,1,0,0,0,5);
      if ( !alreadyPacked )
        SwingUtil.packParentWindow(this);
      else
        this.validate();
      alreadyPacked = true;
      singleDivSlider.setEnabled( true );
      averageDivSlider.setEnabled( true );
    }

    /**
     * Setzt die in der GUI angezeigte Interationsnummer. Diese wird
     * durch einen String repraesentiert, damit auch berliebige andere
     * Informationen (anstelle der Nummer) dargestellt werden koennen (z.B.
     * zusaetzlich noch die Schrittnummer).
     * @param iterString dargestellte Iterationsnummer
     */
    public void setIterationNumber(String iterString) {
      iterationLabel.setText("Iteration: "+iterString);
    }

    /**
     * Stellt den Slider fuer die Einzelabweichung auf einen bestimmten Wert
     * ein. Hierdurch wird automatisch auch der Wert fuer das Modell neu
     * gesetzt.
     */
    public void setSingleTolerance(double tol) {
      singleDivSlider.setValue((int)(tol*SLIDER_FACT));
    }

    /**
     * Stellt den Slider fuer die Durchschnittsabweichung auf einen bestimmten Wert
     * ein. Hierdurch wird automatisch auch der Wert fuer das Modell neu
     * gesetzt.
     */
    public void setAverageTolerance(double tol) {
      averageDivSlider.setValue((int)(tol*SLIDER_FACT));
    }

    /**
     * Setzt die Abweichnungswerte fuer alle Statusbalken.
     * @param value Array mit den Abweichungen
     */
    public void setDeviation(double[] value) {
      if ( divCont!= null )
        divCont.setExpansion(value);
    }

    /**
     * Setzt den Abweichnungswert fuer einen Statusbalken.
     * @param idx   Index eines Balkens
     * @param value Array mit den Abweichungen
     */
    public void setDeviation(int idx, double value) {
      if ( divCont!= null )
        divCont.setExpansion(idx,value);
    }

    /**
     * Diese Klasse stellt ein Label dar, in dem (hinter einem Basistext)
     * ein numerischer Abweichungswert (z.B. erlaubte Maximalabweichung) angezeigt wird.
     * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
     * @version 1.0
     */
    private static class DeviationLabel extends JLabel {
      private String baseText        = "";
      private double dev             = 0.0;
      private NumberFormat numFormat = null;

      /**
       * Erzeugt ein neues leeres Label.
       */
      public DeviationLabel() {
        this("");
      }

      /**
       * Erzeugt ein neues Label. Der dargestellte Wert ist initial 0%.
       * @param text      Basistext
       */
      public DeviationLabel(String text) {
        this(text,new DecimalFormat("0.0%"));
      }

      /**
       * Erzeugt ein neues Label. Der dargestellte Wert ist initial 0.
       * @param text      Basistext
       * @param numFormat Anzeigeformat fuer den dargestellten Abweichungswert
       */
      public DeviationLabel(String text, NumberFormat numFormat) {
        this(text,numFormat,0.0);
      }

      /**
       * Erzeugt ein neues Label.
       * @param text      Basistext
       * @param numFormat Anzeigeformat fuer den dargestellten Abweichungswert
       * @param dev       initialer Wert fuer den Abweichungswert
       */
      public DeviationLabel(String text, NumberFormat numFormat, double dev) {
        super();
        this.baseText  = text;
        this.dev    = dev;
        setNumberFormat(numFormat);
        setText(baseText);
      }

      /**
       * Setzt den darzustellenden numerischen Abweichungswert.
       */
      public void setValue(double dev) {
        this.dev = dev;
        setText(baseText);
      }

      /**
       * Setzt den Basistext des Labels.
       */
      public void setText(String text) {
        this.baseText = text;
        String displayText = baseText;
        if ( numFormat != null )
          displayText = displayText+" "+numFormat.format(this.dev);
        super.setText(displayText);
      }

      /**
       * Setzt das Darstellungsformat fuer den Abweichungswert.
       */
      public void setNumberFormat(NumberFormat format) {
        this.numFormat = format;
      }
    }

    /**
     * Diese Klasse stellt einen Container dar, in dem vertikal eine Reihe von
     * {@linkplain ExpansionBar Status-Balken} angezeigt werden. Alle Balken
     * haben dieselben Extremwerte (Min, Max, Middle). Deshalb erhaelt nur
     * der unterste Balken eine Info-Leiste mit diesen Extremwerten.
     * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
     * @version 1.0
     */
    private static class DeviationContainer extends JComponent {
      private static final NumberFormat NUM_FORMAT = new DecimalFormat("0.0%");
      /** Konstante fuer eine vertikale Ausrichtung der Status-Balken */
      public static final int VERTICAL   = ExpansionBar.VERTICAL;
      /** Konstante fuer eine horizontale Ausrichtung der Status-Balken */
      public static final int HORIZONTAL = ExpansionBar.HORIZONTAL;

      private ExpansionBar[] expansionBar   = null;
      private JLabel[]       expansionLabel = null;
      private JLabel         typeLabel      = null;
      private JLabel         divLabel       = null;
      private Font           sampleFont     = new JLabel().getFont();

      /**
       * Erzeugt einen neuen Container. Die Balken stellen eine prozentuale
       * Abweichung von +/- 100% um 0% dar. Die Balken erhalten keine Beschreibung,
       * sondern werden durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       */
      public DeviationContainer(int count, double tol) {
        this(count,-1,1,0,tol,NUM_FORMAT);
      }

      /**
       * Erzeugt einen neuen Container. Als Mitte-Wert fuer die Balken wird
       * 0 verwendet. Die Balken erhalten keine Beschreibung, sondern werden
       * durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param form     Darstellungsformat fuer die Zahlen im Balken und in der
       *                 Info-Leiste des untersten Balken
       */
      public DeviationContainer(int count, double min, double max, double tol, NumberFormat form) {
        this(count,min,max,0,tol,form);
      }

      /**
       * Erzeugt einen neuen Container. Die Balken werden durchnummeriert.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param form     Darstellungsformat fuer die Zahlen im Balken und in der
       *                 Info-Leiste des untersten Balken
       */
      public DeviationContainer(int count, double min, double max, double mid, double tol, NumberFormat form) {
        this(count,null,min,max,mid,tol,form,form);
      }

      /**
       * Erzeugt einen neuen Container.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param desc     Beschreibungen fuer die Balken (kann <code>null</code> sein!)
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer alle Balken (siehe {@link ExpansionBar})
       * @param barForm  Darstellungsformat fuer die Zahlen im Balken
       * @param infoForm Darstellungsformat fuer die Zahlen in der Info-Leiste
       *                 des untersten Balken
       */
      public DeviationContainer(int count, String[] desc, double min, double max, double mid, double tol, NumberFormat barForm, NumberFormat infoForm) {
        this(count,desc,min,max,mid,LangUtil.createArray(count,tol),barForm,infoForm);
      }

      /**
       * Erzeugt einen neuen Container.
       * @param count    Anzahl an Statusbalken im Conatiner
       * @param desc     Beschreibungen fuer die Balken (kann <code>null</code> sein!)
       * @param min      Minimum-Wert fuer die Balken
       * @param max      Maximum-Wert fuer die Balken
       * @param mid      Mitte-Wert fuer die Balken
       * @param tol      Toleranzwert fuer jeden einzelnen Balken (siehe {@link ExpansionBar})
       * @param barForm  Darstellungsformat fuer die Zahlen im Balken
       * @param infoForm Darstellungsformat fuer die Zahlen in der Info-Leiste
       *                 des untersten Balken
       */
      public DeviationContainer(int count, String[] desc, double min, double max, double mid, double[] tol, NumberFormat barForm, NumberFormat infoForm) {
        super();
        if ( count <= 0 )
          throw new IllegalArgumentException("At least one divergence bar must specified!");
        if ( desc!=null && desc.length < count )
          throw new IllegalArgumentException("Less describtions specified!");

        this.setLayout( new SpringLayout() );

        // PROBLEME mit den Labels bei VERTICAL
        // --> Label-Breite wird wohl zu 0
        int orient = HORIZONTAL;

        sampleFont = sampleFont.deriveFont(Font.BOLD);
        if (orient == VERTICAL)
          sampleFont = sampleFont.deriveFont( AffineTransform.getRotateInstance(Math.PI/2) );

        // Label erzeugen
        typeLabel = new JLabel("Type");
        divLabel = new JLabel("Divergence");
        typeLabel.setFont(sampleFont);
        divLabel.setFont(sampleFont);

        // Statusbalken erzeugen
        this.expansionBar   = new ExpansionBar[count];
        this.expansionLabel = new JLabel[count];
        for (int i=0; i<count; i++) {
          expansionBar[i]   = new ExpansionBar(orient,ExpansionBar.LINEAR,min,max,mid,-tol[i],tol[i],0,barForm,infoForm);
          expansionLabel[i] = new JLabel( desc!=null && i<desc.length ? desc[i] : String.valueOf(i));
          if ( orient == HORIZONTAL ) {
            expansionLabel[i].setHorizontalAlignment(JLabel.CENTER);
            expansionLabel[i].setVerticalAlignment(JLabel.TOP);
          } else {
            expansionLabel[i].setFont( expansionLabel[i].getFont().deriveFont( AffineTransform.getRotateInstance(Math.PI/2) ) );
            expansionLabel[i].setHorizontalAlignment(JLabel.LEFT);
            expansionLabel[i].setVerticalAlignment(JLabel.CENTER);
          }
        }
        // beim untersten Balken sollen die Labels erscheinen
        expansionBar[count-1].setValueLabelsPainted(true);

        // Komponenten einfuegen und ausrichten
        if ( orient == HORIZONTAL ) {
          typeLabel.setHorizontalAlignment(JLabel.CENTER);
          typeLabel.setVerticalAlignment(JLabel.BOTTOM);
          divLabel.setHorizontalAlignment(JLabel.CENTER);
          divLabel.setVerticalAlignment(JLabel.BOTTOM);
          add( typeLabel );
          add( divLabel );
          for (int i=0; i<count; i++) {
            add( expansionLabel[i] );
            add( expansionBar[i] );
          }
          // Layout ausrichten
          SpringUtilities.makeCompactGrid(this,getComponentCount()/2,2,0,0,10,0);
        } else {
          typeLabel.setHorizontalAlignment(JLabel.CENTER);
          typeLabel.setVerticalAlignment(JLabel.CENTER);
          divLabel.setHorizontalAlignment(JLabel.CENTER);
          divLabel.setVerticalAlignment(JLabel.CENTER);
          add( divLabel );
          for (int i=0; i<count; i++)
            add( expansionBar[i] );
          add( typeLabel );
          for (int i=0; i<count; i++)
            add( expansionLabel[i] );
          // Layout ausrichten
          SpringUtilities.makeCompactGrid(this,2,getComponentCount()/2,0,0,10,0);
        }
      }

      /**
       * Setzt den Toleranzwert fuer alle Balken.
       */
      public void setTolerance(double tol) {
        for (int i=0; i<expansionBar.length; i++) {
          expansionBar[i].setMinimumTolerance( -tol );
          expansionBar[i].setMaximumTolerance(  tol );
        }
      }

      /**
       * Setzt den Wert fuer alle Balken.
       */
      public void setExpansion(double[] value) {
        for (int i=0; i<value.length; i++)
          setExpansion(i,value[i]);
      }

      /**
       * Setzt die Beschreibung fuer einen Balken.
       */
      public void setBarDesc(int idx, String text) {
        expansionLabel[idx].setText(text);
      }

      /**
       * Setzt den Wert fuer einen Balken.
       */
      public void setExpansion(int idx, double value) {
        expansionBar[idx].setValue(value);
      }

      /**
       * Setzt die Beschreibung die oberhalb der Balken-Beschreibungen steht.
       */
      public void setBarDescHeader(String text) {
        typeLabel.setText(text);
      }

      /**
       * Setzt die Beschreibung die oberhalb der Balken steht.
       */
      public void setBarHeader(String text) {
        divLabel.setText(text);
      }

    }
  }

  ////////////////////////////////////////////////////////////////////////
  ////////////////////   Nur zu temporaeren Zwecken   ////////////////////
  ////////////////////////////////////////////////////////////////////////
  private java.io.BufferedWriter logWriter = null;
  private void openLog(int stepNo) {
    if ( logWriter != null )
      closeLog();
    try {
      logWriter = new java.io.BufferedWriter( new java.io.OutputStreamWriter( new java.io.FileOutputStream("Clue_"+stepNo+".log") ) );
    } catch (Exception err) {
      statusOut.println("Fehler beim Erzeugen des Log-Files!");
    }
  }

  private void appendLog(Object... text) {
    if ( logWriter!=null )
      for (int i=0; i<text.length; i++)
        try {
          logWriter.write(text[i].toString());
          if ( !text[i].equals("\n") )
            logWriter.write("\t");
        } catch (Exception err) {
          statusOut.println("Fehler beim Schreiben in das Log-File!");
        }
  }

  private void closeLog() {
    try {
      if ( logWriter != null ) {
        logWriter.flush();
        logWriter.close();
        logWriter = null;
      }
    } catch (Exception err) {
      statusOut.println("Fehler beim Schliessen des Log-Files!");
    }
  }
  ////////////////////////////////////////////////////////////////////////

}
