package edu.bonn.xulu.plugin.model.parallel;

import java.util.HashMap;
import java.util.Map;

import edu.bonn.xulu.plugin.model.parallel.ClueTaskDefinition.Task;

import schmitzm.data.WritableGrid;
import appl.parallel.spmd.AbstractSPMDTask;
import appl.parallel.spmd.AdvancedSPMDServerInterface;
import appl.parallel.spmd.MultiDataObject;
import appl.parallel.spmd.SPMDServerInterface;
import appl.parallel.spmd.SyncPoint;
import appl.parallel.spmd.SyncPoint.Priority;

/**
 * This is a tuned version of the {@link ClueTask}. Preloading and multithreading are
 * used for a better resource utilization.
 * 
 * @author Dominik Appl
 */
public class ClueTaskTuned extends AbstractSPMDTask {

	private WritableGrid baseLU;
	private AdvancedSPMDServerInterface controller;
	private int luCount;
	private int dfCount;
	private int[] ddfStartIdx;
	private float[][] regrRes;
	private int areaMinY;
	private int areaMinX;
	private int areaMaxY;
	private int areaMaxX;
	private WritableGrid[] outLUProb;
	private WritableGrid areaRestr;
	private int nbhRadiusX;
	private int nbhRadiusY;
	private float[][][] nbhSettings;
	private float[][] nbhRegrRes;
	private float totCover;
	private int localMinX;
	private int localMinY;
	private int localMaxX;
	private int localMaxY;
	private WritableGrid[] outNbhProb;
	private WritableGrid outActLUHist;
	private float[][] luDemand;
	private float[] luElas;
	private int[][]	luConv;
	private float[] oldLUCover;
	private WritableGrid outActLU;
	private boolean CALC_WITH_NBH;
	private float[] nbhWeights;
	private float cellSize;
	private WritableGrid baseLUHist;
	private MultiDataObject outStepLU;
	private SyncPoint drivingForcesSync;
	
	private static Object monitor = new Object();
	private Map<Float,Float> 		  luMap	        = new HashMap<Float,Float>(); // speichert das Mapping von LUCC-Typen im BAseScenario zu 0,1,2,...
	private Map<Float,Float> 	  	  luMapReverse  = new HashMap<Float,Float>(); // speichert das Mapping von LUCC-Typen im BAseScenario zu 0,1,2,...
	  
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.AbstractSPMDTask#supportsMultiThreading()
	 */
	@Override
	public boolean supportsMultiThreading() {
		return true;
	}
	
	/* (non-Javadoc)
	 * @see appl.parallel.spmd.AbstractSPMDTask#init()
	 */
	@Override
	public void init() {
		controller = getAdvancedSPMDServerController();
        
		luCount = (Integer) controller.getBaseParameter("luCount");
		dfCount = (Integer) controller.getBaseParameter("dfCount");
		ddfStartIdx = (int[]) controller.getBaseParameter("ddfStartIdx");
		
		long time = System.currentTimeMillis();
        SyncPoint loadPoint = new SyncPoint(0,Priority.HIGH);
        
        /**  
         * load all data at the same time (uses heavy multithreading!)
         * might cause problems with many participating servers, because
         * the client might be overloaded with get-threads
         */
		controller.preloadPartition("outActLU",loadPoint);
		controller.preloadPartition("outActLUHist",loadPoint);
		controller.preloadPartition("baseLUHist",loadPoint);
		controller.preloadPartition("baseLU",loadPoint);
		controller.preloadPartition("areaRestr",loadPoint);
				
		
		for (int i = 0; i < luCount; i++) {
			controller.preloadMultiPartition("outLUProb",i, loadPoint);
			controller.preloadMultiPartition("outNbhProb",i, loadPoint);
		}
		//wait on the loading to finish
		controller.synchronizeToSyncPoint(loadPoint);
		//after loading the basedata we can start preloading the static or dynamic 
		//DrivingForces which will be needed when calculating the LUProbabilities
		preloadData(1);
		
		/******************* end of preloading section ***********************/
		
		// initalize partitions
		outActLU = (WritableGrid) controller.getPartition("outActLU");
		outActLUHist = (WritableGrid) controller.getPartition("outActLUHist");
		baseLUHist = (WritableGrid) controller.getPartition("baseLUHist");
		baseLU = (WritableGrid) controller.getPartition("baseLU");
		areaRestr = (WritableGrid) controller.getPartition("areaRestr");
		
		//initialize base Parameters
		nbhRadiusX = (Integer) controller.getBaseParameter("nbhRadiusX");
		nbhRadiusY = (Integer) controller.getBaseParameter("nbhRadiusY");
		nbhSettings = (float[][][]) controller.getBaseParameter("nbhSettings");
		nbhRegrRes = (float[][]) controller.getBaseParameter("nbhRegrRes");
		totCover = (Float) controller.getBaseParameter("totCover");
		
		regrRes = (float[][]) controller.getBaseParameter("regrRes");
		CALC_WITH_NBH = (Boolean)    controller.getBaseParameter("CALC_WITH_NBH");
		nbhWeights = (float[])    controller.getBaseParameter("nbhWeights");
		cellSize = (Float)      controller.getBaseParameter("cellSize");
		luConv			         = (int[][])    controller.getBaseParameter("luConv");
		luDemand 				 = (float[][])  controller.getBaseParameter("luDemand");
		luElas  				 = (float[])    controller.getBaseParameter("luElas"); 
		luMap = (Map<Float, Float>) controller.getBaseParameter("luMap");
		luMapReverse = (Map<Float, Float>) controller.getBaseParameter("luMapReverse");
		
		
		areaMinY = controller.getLocalCalcMinY();
		areaMinX = controller.getLocalCalcMinX();
//		+1 because Clue is programmed that area-MaxCoordinates are exclusive!
		areaMaxY = controller.getLocalCalcMaxY()+1;
		areaMaxX = controller.getLocalCalcMaxX()+1;
		
//		the bounds of the available data (including neighborhood)

		localMinX = (int)controller.getLocalBounds().getMinX();
		localMinY = (int)controller.getLocalBounds().getMinY();
		localMaxX = (int)controller.getLocalBounds().getMaxX();
		localMaxY = (int)controller.getLocalBounds().getMaxY();
		
				outLUProb = new WritableGrid[luCount];
		for (int i = 0; i < outLUProb.length; i++) {
			outLUProb[i] = (WritableGrid) controller.getMultiPartition("outLUProb",i);
		}

		//initialize outNbhProb
		outNbhProb = new WritableGrid[luCount];
		for (int i = 0; i < outNbhProb.length; i++) {
			outNbhProb[i] = (WritableGrid) controller.getMultiPartition("outNbhProb",i);
		}
		outStepLU = controller.getMultiData("outStepLU");
		
		// clear output data (only done one for primary thread
		if (controller.isMasterThread())
			for (int y = localMinY; y < localMaxY; y++)
				for (int x = localMinX; x < localMaxX; x++)
					outActLU.setRasterSample(Float.NaN, x, y);
		
		System.out.println("Finished init of CLUE_TASK_TUNED in "
				+ (System.currentTimeMillis() - time) + " ms");
	}

	
	/**
	 * Preloads the static or dynamic driving forces which will be needed for the given step no
	 * 
	 * @param stepNo the stepNo
	 */
	private void preloadData(int stepNo) {
		//do not forget to reinitialize the syncPoint
		drivingForcesSync = new SyncPoint(stepNo,Priority.HIGH);
		for (int df = 0; df < dfCount; df++) { 
		if ( ddfStartIdx[df] >= 0 )
             controller.preloadMultiPartition("ddfGrids",ddfStartIdx[df]+(stepNo-1),drivingForcesSync);
           else
             controller.preloadMultiPartition("sdfGrids",df,drivingForcesSync);
		 }
	}


	/* (non-Javadoc)
	 * @see appl.parallel.spmd.AbstractSPMDTask#run()
	 */
	@Override
	public Object run(Object... parameters) {
		controller = this.getAdvancedSPMDServerController();
		
		Task task = (Task) parameters[0]; 
		switch(task){
		case LU_PROB:
			calculateLUProbabilities(parameters);
			break;
		case NEIGHBORHOOD_PROB:
			calulateNeighborhoodProbabilities(parameters);
			break;
		case CLUE_ITER:
			return clue_iter(parameters);
		case STEP_FINALIZATION:
			finalizeStep(parameters);
			break;
		}
		return null;
	}
	
	private void finalizeStep(Object... parameters){
		
		WritableGrid outStepLUGrid, lastLUGrid;
		//it is import to synchronize the #addElement because this method is not
		//thread-safe. Two threads Thread access the method at the same time
		//two new grids would be created, each thread would use its own grid
		//and at last only one of the two grids would have been added to the multi grid.
		synchronized (monitor) {
			// initialisiere StepCount
			int stepNo = (Integer) parameters[1];

			// Liste der Zeitschritt-Ausgaberaster vergroessern
			if (outStepLU.getCount() <= stepNo - 1)
				outStepLU.addElement();
			// Raster festlegen, in dem die LU des vorangegangenen
			// Schritts zu finden ist
			if (stepNo == 1)
				lastLUGrid = baseLU;
			else
				lastLUGrid = (WritableGrid) controller.getMultiPartition(
						"outStepLU", stepNo - 2);

			// LU in Ausgaberaster des Zeitschritts uebernehmen und ggf.
			// Nutzungsdauer initialisieren
			outStepLUGrid = (WritableGrid) outStepLU.getElement(stepNo - 1);
		}
	    float lastLUC;
	    float actLUC;
	    for (int y=areaMinY; y<areaMaxY; y++)
	      for (int x=areaMinX; x<areaMaxX; x++) {
	        //checkBreakingCommands();
	    	  lastLUC = luMap.get( lastLUGrid.getRasterSampleAsFloat(x,y) );
	          actLUC  = luMap.get( outActLU.getRasterSampleAsFloat(x,y) );
	        // Wenn die Landnutzung gewechselt hat, wird die Historie zurueck
	        // gesetzt
	        if ( actLUC != lastLUC )
	          outActLUHist.setRasterSample(0f,x,y);

	        // LU in Ausgabe-Raster des simulierten Schritts uebernehmen
	        // Die 0-basierten LU-Nummern werden wieder zurueckgemapped. (SK)
	        outStepLUGrid.setRasterSample( luMapReverse.get( actLUC) ,x,y);
	      }
	  }
	
	
	 /**
	 * 
	 */
	private void calculateLUProbabilities(Object... parameters) {
		//wait on the preloading to finish
/*P*/	controller.synchronizeToSyncPoint(drivingForcesSync);
		//		initialisiere Parameter
		int	stepNo = (Integer) parameters[1];
		
				
		///************************* START CALCULATION *********************************/
//	    System.out.println("Local Area:" + controller.getLocalBounds());
		int x            = 0;
	    int y            = 0;
	    int lu           = 0;
	    int df           = 0;
	    float betaSum    = 0;
	    float expBetaSum = 0;
	    float prob       = 0;
	    // Fuer Zellen des Untersuchungsgebiets
	    for (y=areaMinY; y<areaMaxY; y++)
	        for (x=areaMinX; x<areaMaxX; x++)
	        // NotData-Bereich und Area Restrictions ignorieren
	        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 ) {
	          // Fuer alle LU-Typen
	          for (lu = 0; lu < luCount; lu++) {
	            betaSum = 0;
	            // Fuer alle Driving Forces die Summe der Betas berechnen
	            for (df = 0; df < dfCount; df++) {
	              ///****checkBreakingCommands();
	              // DF-Grid festlegen:
	              // Bei dynamischem DF das jew. Zeitschritt-Raster aus der Liste
	              // ansonsten das statische Raster
	              WritableGrid dfGrid = null;
	              if ( ddfStartIdx[df] >= 0 )
	                dfGrid = (WritableGrid) controller.getMultiPartition("ddfGrids",ddfStartIdx[df]+(stepNo-1));
	              else
	                dfGrid = (WritableGrid) controller.getMultiPartition("sdfGrids",df);
	              // Beim ersten Driving Force wird die Summe mit der
	              // Regressionskonstante initialisiert
	              if (df == 0)
	                betaSum = regrRes[0][lu];
	              // betaSum = BETA0(U) + Sum( BETA(df+1,U) * DF(df,x,y) )
	              betaSum += regrRes[df + 1][lu]
	                       * dfGrid.getRasterSampleAsFloat(x, y);
	            }
	            expBetaSum = (float)Math.exp(betaSum);
	            prob = expBetaSum / (1 + expBetaSum);
	            // errechnete W'keit im Raster speichern
	            outLUProb[lu].setRasterSample( (float) prob, x, y);
	          }
	        }
	    	//start preloading the values for the next step
/*P*/	    preloadData(stepNo+1);
	    
			// destroy no longer used ddfs to save memory
			for (df = 0; df < dfCount; df++) {
			if (ddfStartIdx[df] >= 0)
				controller.destroyMultiPartition("ddfGrids", ddfStartIdx[df]
						+ stepNo - 1);
		}
		
	}

	
	public void calulateNeighborhoodProbabilities(Object... parameters) {
		//initialize parameters
		int	stepNo = (Integer) parameters[1]; 
		float[][] luCover =  (float[][]) parameters[2];
		
				
		/***************************** Parallel Computation *********************************
		 * The following code is near a copy&paste from Martin Schmitz Clue - code. The 2 lines 
		 * changed are indicated with '*P*'.
		 * 
		 * About the adjustment:
		 * The change was necessary because of the neighborhood region the algorithm
		 * looks at. It is wasted computing time to try to calculate the  probabilities 
		 * for this overlapping region, but on the other hand the algorithm must be aware
		 * of the additional data for the calculation of the inner bounds.  
		 **********************************************************************************/
		
		// Raster festlegen, in dem die LU des vorangegangenen
	    // Schritts zu finden ist
		WritableGrid luGrid;
		if(stepNo==1)
			luGrid = (WritableGrid) controller.getPartition("baseLU");
		else 
			luGrid = (WritableGrid) controller.getMultiPartition("outStepLU", stepNo-2);

		int      x            = 0;
	    int      y            = 0;
	    int      lu           = 0;
	    int      i            = 0;
	    int      dx           = 0;
	    int      dy           = 0;
	    int      totCellCount = 0;
	    float[]  luCellCount  = new float[luCount];
	    float    betaSum      = 0;
	    float    beta         = 0;
	    float    nbhFact      = 0;
	    float    expBetaSum   = 0;
	    float    lucc         = 0;
	    float    setting      = 0;
	    System.out.println("ClueTask - Calc Area min/max X: " + areaMinX + "/" + areaMaxX + "  Y: " + areaMinY + "/" + areaMaxY);
	    // Fuer Zellen des Untersuchungsgebiets
	    for (y=areaMinY; y<areaMaxY; y++)
	      for (x=areaMinX; x<areaMaxX; x++)
	        // NoData-Bereich und Area Restrictions ignorieren
	        if ( !checkNoData(x,y,baseLU) && areaRestr.getRasterSampleAsFloat(x,y)==0 )
	          // Fuer alle LU-Typen
	          for (lu = 0; lu < luCount; lu++) {
	            //checkBreakingCommands();

	            // Anzahl der Zellen in der Umgebung initialisieren
	            totCellCount = 0;
	            for (i=0; i<luCount; i++)
	              luCellCount[i] = 0;
	           // Umgebung der Zelle betrachten
	            for (dy=-nbhRadiusY; dy<=nbhRadiusY; dy++)
/*P*/ 	           if ( localMinY <= y+dy && y+dy<localMaxY )
	                for (dx=-nbhRadiusX; dx<=nbhRadiusX; dx++)
/*P*/                  if ( localMinX <= x+dx && x+dx<localMaxX ) {
	                    lucc = luGrid.getRasterSampleAsFloat(x+dx,y+dy); // PERFORMANZ > nur einmal den Zellen-Wert ermitteln
	                    if ( !Float.isNaN(lucc) /*!checkNoData(x+dx,y+dy,baseLU)*/ ) {
//	                    	 Die LU-Typen stehen im Raster nicht 0-basiert, deshalb mapping
	                   		lucc = luMap.get( lucc );
	                   		
	                    	// Alle Zellen in Umgebung zaehlen (jedoch nur diejenigen,
	                        // die potentiell auch in die Bewertung eingehen koennen!)
	                        setting = nbhSettings[dx+nbhRadiusX][dy+nbhRadiusY][lu];
	                        if ( setting != 0.0 ) {
	                          totCellCount++;
	                          // Anzahl der Zellen eines bestimmten Typs aufsummieren
	                          // (relativ zum in der Umgebung angegebenen Faktor)
	                          // luCellCount[ luGrid.getRasterSampleAsInt(x+dx,y+dy) ] += setting;
//	                        TODO Bricht hier ab, wenn LUCC-Klassen nicht 0,1,2 ... zaehlen...
//	                         luCellCount[ (int)lucc ] += setting;
	                          
	                          luCellCount[ (int)lucc ] += setting;
	                        }
	                      }
	                    }
	              // Summe der Betas berechnen
	              betaSum = nbhRegrRes[0][lu]; // Regr.Konstante
	              for (i=0; i<luCount; i++) {
	                beta    = nbhRegrRes[i + 1][lu];
	                nbhFact = (luCellCount[i]/totCellCount) / (luCover[i][stepNo-1]/totCover);
	                betaSum += beta * nbhFact;
	              }
	              // Wahrscheinlichkeit berechnen
	              expBetaSum = (float)Math.exp(betaSum);
	              outNbhProb[lu].setRasterSample((float)(expBetaSum/(1 + expBetaSum)),x,y);
	          }
	}
	
	
	public Object clue_iter(Object... parameters) {
		// initialize parameters
		int stepNo        = (Integer) parameters[1];
		int iterNo        = (Integer) parameters[2];
		float[] outIter   = (float[]) parameters[3];
		oldLUCover = (float[]) parameters[4];
		
		/** saves the new, additional lu for each step (needed for synchronisation */
		float[] newLUCover = new float[luCount]; 

		WritableGrid luHistGrid;
		if(stepNo==1)
			luHistGrid = baseLUHist;
		else 
			luHistGrid = outActLUHist;
		
		// Raster festlegen, in dem die LU des vorangegangenen
	    // Schritts zu finden ist
		WritableGrid lastLUGrid;
		if(stepNo==1)
			lastLUGrid = baseLU;
		else 
			lastLUGrid = (WritableGrid) controller.getMultiPartition("outStepLU", stepNo-2);
		int x = 0;
		int y = 0;
		int lu = 0;
		int lastLU = 0;
		int bestLU = 0;
		float bestLUProb = 0;
		float baseProb = 0;
		float nbhProb = 0;
		float nbhWeight = 0;
		float elas = 0;
		float totalProb = 0;
//		float lastLUC = 0;
//		float actLUC = 0;
		 // Merken, ob an einer Stelle bereits eine U-Zelle in einen anderen
	    // Typ gewechselt hat
	    boolean[] alreadyLUCCFromType = new boolean[luCount];
	    // Merken, ob an einer Stelle bereits eine U-Zelle neu entstanden ist
	    boolean[] alreadyLUCCToType = new boolean[luCount];
	    float[] iterSpeed = new float[luCount];
		// Arrays fuer Iterationsschritt initialisieren
	      for (lu = 0; lu < luCount; lu++) {
	        // noch kein Wechsel von einem LU-Typ
	        alreadyLUCCFromType[lu] = false;
	        // noch kein Wechsel zu einem LU-Typ
	        alreadyLUCCToType[lu] = false;
	        // Flaechen der LU-Typen fuer diesen Zeitschritt
	        newLUCover[lu] = 0;
	       
			// Speed fuer die Iterationsvariablen initialisieren
	        // --> liegt die Nachfrage bei weniger als 5% des U'gebiets
	        //     wird langsamer begonnen
	        if ( luDemand[lu][stepNo]/totCover < 0.05f )
	          iterSpeed[lu] = 0.01f;
	        else
	          iterSpeed[lu] = 0.05f;
	      }
		  // Fuer alle Zellen des Untersuchungsgebiets
	      for (y = areaMinY; y < areaMaxY; y++)
	        for (x = areaMinX; x < areaMaxX; x++) {
	          //******checkBreakingCommands();

	          // Rasterzellen ausserhalb des Untersuchungsgebiets komplett
	          // ignorieren
	          // Grund: NoData-Zellen (Float.NaN) werden beim Cast in int zu 0,
	          //        also zur Landnutzung 0!!
	          if ( checkNoData(x,y,areaRestr) || checkNoData(x,y,baseLU) )
	            continue;

	          // Nutzungsdauern erhoehen
	          if (iterNo==1)
	            outActLUHist.setRasterSample(luHistGrid.getRasterSampleAsFloat(x, y) + 1, x, y);

	          // vorangegangene LU merken. 
	          // Diese steht im Raster nicht unbedingt 0-basiert
	          lastLU = luMap.get( new Float (lastLUGrid.getRasterSampleAsInt(x, y)) ).intValue();

	          // Beste gefundene LU wird mit aktueller LU initialisiert
	          bestLU = lastLU;
	          bestLUProb = 0;

	          // Pruefen, ob Zelle die LU wechseln darf
	          if (checkLUCCGeneralAllowed(x, y, stepNo, lastLU, alreadyLUCCToType[lastLU])) {
	            // Alle LU-Typen auf Wechsel pruefen
	            for (lu = 0; lu < luCount; lu++) {
	              // Pruefen, ob Zelle in den betrachteten Typ wechseln darf
	              if (checkLUCCAllowed(x, y, stepNo, lastLU, lu, alreadyLUCCFromType[lu])) {
	                // Gesamt-Wahrscheinlichkeit berechnen
	                // a) DF-W'keit und Nachbarschaft-W'keit gehen anteilig ein
	                // b) Elastizitaet geht nur ein, wenn Zelle bereits fuer den
	                //    LU-Typ genutzt wird
	                // c) ITER geht negativ ein, da es auf Basis der Abweichung
	                //    luCover-luDemand berechnet wird. Ein negatives Iter
	                //    kennzeichnet also einen Mangel und muss die Gesamtw'keit
	                //    erhoehen
	                 	baseProb  =  outLUProb[lu].getRasterSampleAsFloat(x, y);
	                nbhProb   = !CALC_WITH_NBH ? 0 : outNbhProb[lu].getRasterSampleAsFloat(x, y);
	                nbhWeight = !CALC_WITH_NBH ? 0 : nbhWeights[lu];
	                elas      = (lu == lastLU && luElas[lu] < 1) ? luElas[lu] : 0;
	                // LU-Wahrscheinlichkeit zusammensetzen
	                totalProb = baseProb * (1-nbhWeight)
	                          + nbhProb  * nbhWeight
	                          + elas
	                          - outIter[lu];

	                // betrachteter Typ ist geeigneter als der bisher beste
	                if (totalProb > bestLUProb) {
	                  bestLU = lu;
	                  bestLUProb = totalProb;
	                }
	              }
	            }
	            // Merken wenn wirklich ein Wechsel stattfindet
	            if (bestLU != lastLU) {
	              alreadyLUCCFromType[lastLU] = true;
	              alreadyLUCCToType[bestLU] = true;
	            }
	          }
	          // LU fuer diesen Iterationschritt setzen
	          outActLU.setRasterSample( luMapReverse.get ( (float) bestLU) , x, y);
	          // Flaechendeckung fuer LU-Typ erhoehen
/*P*/	          newLUCover[bestLU] += cellSize;

	        }
/*P*/	      return newLUCover;
	 }
	
	
	  /**
	   * Prueft, ob ein genereller LU-Wechsel fuer eine Zelle erlaubt ist.
	   * <ul>
	   * <li>Schliessen die Area Restrictions einen Wechsel aus?<br>
	   *     > Rasterwert(x,y) != 0
	   *     </li>
	   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet 1
	   *     und steigende Bedarf)?<br>
	   *     > ELAS(lastLU) == 1 und DEMAND(lastLU,t) >= DEMAND(lastLU,t-1)
	   *     </li>
	   * </ul>
	   * @param x      Raster-Koordinate
	   * @param y      Raster-Koordinate
	   * @param stepNo modellierter Schritt (beginnend bei 1)
	   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
	   * @param alreadyLUCCToActualType Flag gibt an, ob in der laufenden Iteration
	   *                                bereits ein LU-Wechsel in den den aktuellen
	   *                                Typ der Zelle (lastLU) vorgenommen wurde.<br>
	   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
	   */
	  protected boolean checkLUCCGeneralAllowed(int x, int y, int stepNo, int lastLU, boolean alreadyLUCCToActualType) {
	    // Area Restrictions checken
	    if ( areaRestr.getRasterSampleAsFloat(x,y) != 0 )
	      return false;
//	 DER ERSTE TEIL IST KEINE ORIGINAL-CLUE-REGEL!
//	    // Conversion Elasticity checken
//	    // --> wenn Elasticity 1 und bereits an anderer Stelle eine Zelle des
//	    //     bisherigen Typs entstanden ist, darf sie nicht wechseln
//	    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
//	    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
//	    //     an anderer Stelle neue entstehen
//	    if (  luElas.getValueAsFloat(lastLU) == 1.0
//	       && ( alreadyLUCCToActualType ||
//	           luDemand.getValueAsFloat(lastLU,stepNo) >= luCover[lastLU][stepNo-1] ) )
//	      return false;

	    // Conversion Elasticity checken
	    // --> wenn Elasticity 1 und Bedarf steigt, darf kein Wechsel
	    //     stattfinden, denn dann muessten (um den Bedarf zu decken)
	    //     an anderer Stelle neue entstehen
	    if (  luElas[lastLU] == 1.0 &&
//	          luDemand[lastLU][stepNo] >= luCover[lastLU][stepNo-1] )
	          luDemand[lastLU][stepNo] >= luDemand[lastLU][stepNo-1] )
	      return false;

	    return true;
	  }
	  
	  /**
	   * Prueft, ob ein bestimmter LU-Wechsel fuer eine Zelle erlaubt ist.
	   * <ul>
	   * <li>Schliesst die Conversion Elasticity einen Wechsel aus (Elastizitaet des
	   *     potentiellen Wechseltyps 1 und sinkender Bedarf)?<br>
	   *     > lastLU != newLU und ELAS(newLU) == 1 und DEMAND(newLU,t) <= DEMAND(newLU,t-1)
	   *     </li>
	   * <li>Schliesst die Conversion Matrix einen Wechsel aus (generell oder ueber
	   *     Historie)?<br>
	   *     > CONV(lastLU,newLU) == 0 oder<br>
	   *     > CONV(lastLU,newLU) > 100 und HIST(x,y) < CONV(lastLU,newLU)-101 oder<br>
	   *     > CONV(lastLU,newLU) < -100 und HIST(x,y) > -CONV(lastLU,newLU)-101
	   *     </li>
	   * </ul>
	   * @param x      Raster-Koordinate
	   * @param y      Raster-Koordinate
	   * @param stepNo modellierter Schritt (beginnend bei 1)
	   * @param lastLU Bedeckung von (x,y) zum Zeitpunkt stepNo-1
	   * @param newLU  potentielle neue Bedeckung von (x,y)
	   * @param alreadyLUCCFromNewType  Flag gibt an, ob in der laufenden Iteration
	   *                                bereits ein LU-Wechsel vom potentiellen neuen
	   *                                Typ in einen anderen Typ vorgenommen wurde.<br>
	   *                                <b>Dieses Flag wird zur Zeit nicht verwendet!</b>
	   */
	  protected boolean checkLUCCAllowed(int x, int y, int stepNo, int lastLU, int newLU, boolean alreadyLUCCFromNewType) {
//	 IST KEINE ORIGINAL-CLUE-REGEL!
//	    // LU-Conversion-Elasticity des neuen Typs checken
//	    // --> wenn Elasticity des neuen Typs 1 ist und an anderer Stelle bereits
//	    //     eine Zelle des dieses Typs "verloren" ging, darf er hier
//	    //     nicht entstehen
//	    if (  alreadyLUCCFromNewType && luElas.getValueAsFloat(newLU) == 1.0 )
//	      return false;

	    // Conversion Elasticity checken
	    // --> wenn Elasticity des neuen Typs "newLU" 1 und dessen Bedarf sinkt,
	    //     darf kein Wechsel in diesen Typ stattfinden, denn dann muessten (um
	    //     den Bedarf fuer "newLU" zu erreichen) an anderer Stelle newLU-Zellen
	    //     entfernt werden
	    if (  lastLU != newLU &&
	          luElas[newLU] == 1.0 &&
	          luDemand[newLU][stepNo] <= oldLUCover[newLU] )
//	          luDemand.getValueAsFloat(newLU,stepNo) <= luDemand.getValueAsFloat(newLU,stepNo-1) )
	      return false;

	    // Wert aus LU-Conversion-Matrix ermitteln
	    // Bemerke: "von" steht in den Zeilen (also Y, also Dim1)
	    //          "nach" steht in den Spalten (also X, also Dim2)
	    int conv = luConv[newLU][lastLU];
	    // 1 --> LUCC generell erlaubt
	    if ( conv == 1 )
	      return true;
	    // 0 --> LUCC generell verboten
	    if ( conv == 0 )
	      return false;

	    // 100+t --> Wechsel erst nach t Schritten der vorangegangenen
	    //           LU moeglich
	    if ( conv > 100 ) {
//	      int minLUHist = conv-100;
//	      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//	      return luHist >= minLUHist;
	      int minLUHist = conv-101;
	      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
	      return luHist >= minLUHist;
	    }

	    // -100-t --> Wechsel nicht mehr moeglich, wenn Zelle bereits
	    //            t Schritte unter der vorangegangenen LU
	    if ( conv < -100 ) {
//	      int maxLUHist = -conv-100;
//	      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
//	      return luHist < maxLUHist;
	      int maxLUHist = -conv-101;
	      int luHist    = outActLUHist.getRasterSampleAsInt(x,y);
	      return luHist <= maxLUHist;
	    }

	    return true;
	  }
	
		
	/**
	   * Prueft, ob eine Rasterzelle mit <code>NoData</code> belegt ist
	   * @param x     Raster-Koordinate
	   * @param y     Raster-Koordinate
	   * @param grid  zu pruefendes Raster
	   */
	  protected static boolean checkNoData(int x, int y, WritableGrid grid) {
	    return Float.isNaN(grid.getRasterSampleAsFloat(x,y));
	  }
	  
}
