package edu.bonn.xulu.plugin.model.parallel.demo;

import java.awt.Rectangle;

import appl.parallel.model.AbstractParallelStepModel;
import appl.parallel.spmd.AdvancedSPMDClientInterface;
import appl.parallel.spmd.SPMDClientInterface;
import appl.parallel.spmd.SyncPoint;
import appl.util.RasterUtil;
import edu.bonn.xulu.model.AbstractStepModel;
import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;

/**
 *This is a tuned version of the {@link AverageNeighborhoodParallelDemoModel}, which uses
 *multithreading and preloading. 
 * <ol>
 * <li><b>Input Grid ({@code inputGrid}):</b> A grid with input values </li>
 * <li><b>Output Grid ({@code outputGrid}):</b> A grid where the output is
 * stored </li>
 * <li><b>Number of steps ({@code steps}):</b> the number of steps to
 * compute. </li>
 * </ol>
 * @see AverageNeighborhoodContentManager
 * @author Dominik Appl
 * @version 1.0
 */
public class AVNTuned extends
		AbstractParallelStepModel {

	private int neighborhoodRange = 6;

	/** Speichert den ContentManager fuer das Modell.
	 *  @see AverageNeighborhoodContentManager */
	protected AverageNeighborhoodContentManager contManager;

	//********** Lese/Schreibrechte, die fuer den gesamten ***********
	//********** Modellanlauf gehalten werden muessen      ***********
	private PropertyReadAccess RA_inputGrid = null; // Input Grid

	private PropertyWriteAccess WA_outputGrid = null; // Output Grid

	private PropertyReadAccess RA_steps = null; // Number of steps

	//**************** Variablen mit denen gearbeitet wird *******************
	private WritableGrid inputGrid = null; // Input Grid

	private WritableGrid outputGrid = null; // Output Grid

	private PropertyReadAccess steps = null; // Number of steps

	AdvancedSPMDClientInterface clientController = null;

	/**
	 * Erzeugt eine neue Instanz des Modells.
	 */
	public AVNTuned() {
		super(new AverageNeighborhoodContentManager());
		this.contManager = (AverageNeighborhoodContentManager) super.contManager;
	}

	////////////////////////////////////////////////////////////////////////
	///////////////   Implementierung von AbstractXuluModel   //////////////
	////////////////////////////////////////////////////////////////////////
	/**
	 * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
	 * geladen werden und mit Lese/Schreib-Rechten versehen werden. Zudem
	 * werden Referenzen auf die Ressourcen in lokalen Variablen
	 * gespeichert, um waehrend des Modellablaufs effizienter darauf zugreifen
	 * zu koennen.
	 */
	public void performModelInit() {
		clientController = getAdvancedSPMDController();
		if (contManager == null)
			return;
		// Zugriffsrechte aus Ressourcen/Propertys holen
		RA_inputGrid = null;
		if (contManager.getResource(0).getData() != null)
			RA_inputGrid = ((ScalarProperty) contManager.getResource(0)
					.getData()).getReadAccess(this);
		WA_outputGrid = null;
		if (contManager.getResource(1).getData() != null)
			WA_outputGrid = ((ScalarProperty) contManager.getResource(1)
					.getData()).getWriteAccess(this);
		RA_steps = null;
		if (contManager.getResource(2).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(2).getData())
					.getReadAccess(this);

		// Variablen belegen mit denen gearbeitet wird
		inputGrid = (WritableGrid) RA_inputGrid.getValue();
		outputGrid = (WritableGrid) WA_outputGrid.getValue();
		steps = RA_steps;

		this.stepCount = steps.getValueAsInt();

		//****  Initialize parallelization ********////
		long remoteStartTime = System.currentTimeMillis();
		clientController.setNeighborhoodRange(neighborhoodRange);
		clientController.addToSplitControl(inputGrid, "inputGrid");
		clientController.addToSplitControl(outputGrid, "outputGrid");
		System.out.println("After "
				+ ((System.currentTimeMillis() - remoteStartTime))
				+ "ms: finished with adding to splitcontrol");
	}

	/**
	 * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
	 * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
	 */
	public void performModelDispose() {

		// Ressourcen wieder freigeben
		releaseAccess(RA_inputGrid);
		releaseAccess(WA_outputGrid);
		releaseAccess(RA_steps);
	}

	/**
	 * Fuehrt einen Schritt des Modellablaufs durch.
	 * @param stepNo zu modellierender Schritt (beginnend bei 1!)
	 */
	public void performModelStep(int stepNo) {
		long remoteStartTime = System.currentTimeMillis();
		
		statusOut.println("Starting step " + stepNo + "...");
		/***********************************************************************
		 * 						START REMOTE ALGORITHM
		 //***********************************************************************/
		if (stepNo > 1)
			clientController.updateNeighborhood(inputGrid);
		try {
			Object[] sum = clientController.runSPMDModelTask(new AVNTunedTask(),
					neighborhoodRange);
			float overallSum = 0;
			for (int i = 0; i < sum.length; i++)
				overallSum += (Float) sum[i];
			System.out.println("finished step " + (stepNo) + " in "
					+ ((System.currentTimeMillis() - remoteStartTime)) + "ms "
					+ " with sum " + +overallSum);
			remoteStartTime = System.currentTimeMillis();
		} catch (Throwable e) {
			e.printStackTrace();
		}

		//***************** END STEP CONTROL *****************************
		//After last step: Merge the grids back to the client data pool
		if (stepNo == stepCount) {
			//After execution:
			// merge partitions and check if the local and remotely calculated
			// versions are equal
			remoteStartTime = System.currentTimeMillis();
			
			//in this algorithm we tune the merging by allowing the merge in a separate Thread
			//this normally makes sense especially if you have to merge after every step!
			
			clientController.mergePartition(outputGrid,new SyncPoint(),statusOut,"Finished merging!");
			System.out.println("Merging is done in a separate thread, so this took only "
					+ ((System.currentTimeMillis() - remoteStartTime)) + "ms");
			// Thread.currentThread().sleep(10000);
			if (inputGrid.getWidth() < 20)
				RasterUtil.printGrid(outputGrid, 5, 1, "Remote Output");
		}
	}
}
