package edu.bonn.xulu.plugin.model.parallel.demo;

import java.awt.Rectangle;

import appl.parallel.model.AbstractParallelStepModel;
import appl.parallel.spmd.SPMDClientInterface;
import appl.util.RasterUtil;
import edu.bonn.xulu.model.AbstractStepModel;
import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;

/**
 * This class is a simple demo algorithm for the usage of Xulu / V. It
 * implements exactly the same functionality as the
 * {@link AverageNeighborhoodSerialDemoModel serial counterpart}, but makes
 * heavy use of parallel computations. See
 * {@link AverageNeighborhoodSerialDemoModel} for details about the model
 * semantics. The Model needs three resources, which are specified using the
 * {@linkplain AverageNeighborhoodContentManager}:
 * <ol>
 * <li><b>Input Grid ({@code inputGrid}):</b> A grid with input values </li>
 * <li><b>Output Grid ({@code outputGrid}):</b> A grid where the output is
 * stored </li>
 * <li><b>Number of steps ({@code steps}):</b> the number of steps to
 * compute. </li>
 * </ol>
 * 
 * @see AverageNeighborhoodContentManager
 * @author Dominik Appl
 * @version 1.0
 */
public class AverageNeighborhoodParallelDemoModel extends
		AbstractParallelStepModel {

	/** specifies which neighborhood range to use */
	private int neighborhoodRange = 6;

	/**Speichert den ContentManager fuer das Modell.
	 * @see AverageNeighborhoodContentManager*/
	protected AverageNeighborhoodContentManager contManager;

	// ********** Read and Write access ***********
	private PropertyReadAccess RA_inputGrid = null; // Input Grid
	private PropertyWriteAccess WA_outputGrid = null; // Output Grid
	private PropertyReadAccess RA_steps = null; // Number of steps

	// **************** working variables  *******************
	private WritableGrid inputGrid = null; // Input Grid
	private WritableGrid outputGrid = null; // Output Grid
	private PropertyReadAccess steps = null; // Number of steps

	SPMDClientInterface clientController = null;

	/** Creates a new instance. */
	public AverageNeighborhoodParallelDemoModel() {
		super(new AverageNeighborhoodContentManager());
		this.contManager = (AverageNeighborhoodContentManager) super.contManager;
	}

	/**
	 * Performes the initalization. In addition to the work in the serial version
	 * here the {@link SPMDClientInterface} is retrieved. The neighborhood range is 
	 * set and the two grids are submitted to Xulu / V - control. 
	 */
	public void performModelInit() {
		clientController = getSPMDController();
		if (contManager == null)
			return;
		// Zugriffsrechte aus Ressourcen/Propertys holen
		RA_inputGrid = null;
		if (contManager.getResource(0).getData() != null)
			RA_inputGrid = ((ScalarProperty) contManager.getResource(0)
					.getData()).getReadAccess(this);
		WA_outputGrid = null;
		if (contManager.getResource(1).getData() != null)
			WA_outputGrid = ((ScalarProperty) contManager.getResource(1)
					.getData()).getWriteAccess(this);
		RA_steps = null;
		if (contManager.getResource(2).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(2).getData())
					.getReadAccess(this);

		// Variablen belegen mit denen gearbeitet wird
		inputGrid = (WritableGrid) RA_inputGrid.getValue();
		outputGrid = (WritableGrid) WA_outputGrid.getValue();
		steps = RA_steps;

		this.stepCount = steps.getValueAsInt();

		// **** Initialize parallelization ********////
		long remoteStartTime = System.currentTimeMillis();
		clientController.setNeighborhoodRange(neighborhoodRange);
		clientController.addToSplitControl(inputGrid, "inputGrid");
		clientController.addToSplitControl(outputGrid, "outputGrid");
		System.out.println("After "
				+ ((System.currentTimeMillis() - remoteStartTime))
				+ "ms: finished with adding to splitcontrol");
	}

	/**
	 * frees read / write access
	 */
	public void performModelDispose() {

		// Ressourcen wieder freigeben
		releaseAccess(RA_inputGrid);
		releaseAccess(WA_outputGrid);
		releaseAccess(RA_steps);
	}

	/** Executes one model step:	 */
	public void performModelStep(int stepNo) {
		long remoteStartTime = System.currentTimeMillis();
		statusOut.println("Starting step " + stepNo + "...");

		/***********************************************************************
		 * START REMOTE ALGORITHM //********************** START STEP CONTROL
		 **********************************************************************/

		if (stepNo > 1)
			clientController.updateNeighborhood(inputGrid);

		try {
			Object[] sum = clientController.runSPMDModelTask(new AVNTask(),
					neighborhoodRange);
			float overallSum = 0;
			for (int i = 0; i < sum.length; i++)
				overallSum += (Float) sum[i];
			System.out.println("finished step " + (stepNo) + " in "
					+ ((System.currentTimeMillis() - remoteStartTime)) + "ms "
					+ " with sum " + +overallSum);
			remoteStartTime = System.currentTimeMillis();
		} catch (Throwable e) {
			e.printStackTrace();
		}

		// ***************** END STEP CONTROL *****************************
		if (stepNo == stepCount) {
			// After execution:
			// merge partitions and check if the local and remotely calculated
			// versions are equal
			remoteStartTime = System.currentTimeMillis();
			clientController.mergePartition(outputGrid);
			System.out.println("Merging took another "
					+ ((System.currentTimeMillis() - remoteStartTime)) + "ms");
			// Thread.currentThread().sleep(10000);
			if (inputGrid.getWidth() < 20)
				RasterUtil.printGrid(outputGrid, 5, 1, "Remote Output");
		}
	}
}
