package edu.bonn.xulu.plugin.model.parallel.demo;

import java.awt.Rectangle;

import appl.util.RasterUtil;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.XuluModel;
import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;

/**
 *  This very simple model looks at every grid cell and takes the average 
 *  over a the surrounding cells in a specified neighborhood. It was implemented
 *  as demonstration of the Xulu / V functionality. You can see the parallel version
 *  of this model in {@link AverageNeighborhoodParallelDemoModel} and also a tuned
 *  version using late-merging support in {@link AVNTuned}. 
 * <ol>
 * <li><b>Input Grid ({@code inputGrid}):</b> A grid with input values </li>
 * <li><b>Output Grid ({@code outputGrid}):</b> A grid where the output is
 * stored </li>
 * <li><b>Number of steps ({@code steps}):</b> the number of steps to
 * compute. </li>
 * </ol>
 * 
 * @see AverageNeighborhoodContentManager
 * @author Dominik Appl
 * @version 1.0
 */
public class AverageNeighborhoodSerialDemoModel extends AbstractStepModel {

	/**
	 * the width/height of the neighborhood region over which the algorithm takes 
	 * the average
	 */
	private int neighborhoodRange = 6;

	/**
	 * Speichert den ContentManager fuer das Modell.
	 * 
	 * @see AverageNeighborhoodContentManager
	 */
	protected AverageNeighborhoodContentManager contManager;

	// ********** Lese/Schreibrechte, die fuer den gesamten ***********
	// ********** Modellanlauf gehalten werden muessen ***********
	private PropertyReadAccess RA_inputGrid = null; // Input Grid

	private PropertyWriteAccess WA_outputGrid = null; // Output Grid

	private PropertyReadAccess RA_steps = null; // Number of steps

	// **************** Variablen mit denen gearbeitet wird *******************
	private WritableGrid inputGrid = null; // Input Grid

	private WritableGrid outputGrid = null; // Output Grid

	private PropertyReadAccess steps = null; // Number of steps

	/**
	 * Erzeugt eine neue Instanz des Modells.
	 */
	public AverageNeighborhoodSerialDemoModel() {
		super(new AverageNeighborhoodContentManager());
		this.contManager = (AverageNeighborhoodContentManager) super.contManager;
	}

	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		if (contManager == null)
			return;
		// Zugriffsrechte aus Ressourcen/Propertys holen
		RA_inputGrid = null;
		if (contManager.getResource(0).getData() != null)
			RA_inputGrid = ((ScalarProperty) contManager.getResource(0)
					.getData()).getReadAccess(this);
		WA_outputGrid = null;
		if (contManager.getResource(1).getData() != null)
			WA_outputGrid = ((ScalarProperty) contManager.getResource(1)
					.getData()).getWriteAccess(this);
		RA_steps = null;
		if (contManager.getResource(2).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(2).getData())
					.getReadAccess(this);

		// Variablen belegen mit denen gearbeitet wird
		inputGrid = (WritableGrid) RA_inputGrid.getValue();
		outputGrid = (WritableGrid) WA_outputGrid.getValue();
		steps = RA_steps;

		this.stepCount = steps.getValueAsInt();

	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {

		// Ressourcen wieder freigeben
		releaseAccess(RA_inputGrid);
		releaseAccess(WA_outputGrid);
		releaseAccess(RA_steps);
	}

	/**
	 * This very simple model looks at every grid cell and takes the average
	 * over a the surrounding cells in a specified neighborhood.
	 * 
	 * @param stepNo
	 *            zu modellierender Schritt (beginnend bei 1!)
	 */
	public void performModelStep(int stepNo) {
		
		statusOut.println("Starting step " + stepNo + "...");
		//this rectangle is only used to check if values are outside the grid (see below)
		Rectangle gridBounds = new Rectangle(0, 0, inputGrid.getWidth(),
				inputGrid.getHeight());

		/***********************************************************************
		 * the overallsum is simply the sum over ALL grid elements. it can be used to
		 * compare the result of this model with e.g. the parallel version of
		 * this algorithm.
		 **********************************************************************/ 
		float overallSum = 0f;
		long localStartTime = System.currentTimeMillis();
		//iterate over all cells
		for (int y = 0; y < inputGrid.getHeight(); y++)
			for (int x = 0; x < inputGrid.getWidth(); x++) {
				float tmp = inputGrid.getRasterSampleAsFloat(x, y);
				if (!Float.isNaN(tmp))
					overallSum += tmp;
				// the local sum is simply the sum over all elements in the
				// neighborhood
				float localSum = 0;
				// number of cells over which the sum is calculated
				int noOfCells = 0;
				// for each cell: calculate the sum of all neighbors
				for (int y2 = y - neighborhoodRange; y2 <= y
						+ neighborhoodRange; y2++)
					for (int x2 = x - neighborhoodRange; x2 <= x
							+ neighborhoodRange; x2++)
						// check if the coordinates are valid (inside the
						// grid and not NaN)
						if (gridBounds.contains(x2, y2))
							if (!Float.isNaN(inputGrid.getRasterSampleAsFloat(
									x2, y2))) {
								localSum += inputGrid.getRasterSampleAsFloat(
										x2, y2);
								noOfCells++;
							}
				//after calculation: Set the grid sample
				outputGrid
						.setRasterSample(localSum / (float) (noOfCells), x, y);
			}

		System.out.print("Finished step " + (stepNo) + " in "
				+ ((System.currentTimeMillis() - localStartTime)) + " ms");
		localStartTime = System.currentTimeMillis();
		System.out.println(" with sum (local calculated): " + overallSum);
		// copy values into results from output to input (for next step)
		RasterUtil.copyInto(outputGrid, inputGrid);
	}
}
