package edu.bonn.xulu.plugin.model.sleuth;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;

/**
 * <b>MULTIPLE RESOLUTION COMPARISON</b><br>
 * The Multiple Resolution Comparison is a Xulu-plugin that can be used in order to validate modelling results.
 * In general, the Multiple Resolution Comparison is a method to quantify the agreement between two maps, 
 * whereby it averages neighbouring pixels of both maps into coarser pixels in order to quantify the agreement between maps at coarser resolutions.
 * The percent correct tends to rise at coarser resolutions, because it reduces the effect of error due to location. By comparing the results with 
 * a Multiple Resolution Comparison of a NULL-Model (no change has taken place between t1 and t2), the NULL-Resolution can be identified,
 * which indicates the resolution at which the modelling result is more accurate than the NULL-Model.<br>
 * For details refer to:<br>
 * <i>Costanza,R. (1989)</i>: Model Goodness of Fit - A Multiple Resolution Procedure. In: Ecological Modelling, vol. 47, no. 3-4, pp. 199-215.<br>
 * <i>Pontius Jr, R.G. (2002)</i>: Statistical Methods to Partition Effects of Quantity and Location During Comparison of Categorical Maps at Multiple Resolutions. 
 * In: Photogrammetric Engineering &amp; Remote Sensing, vol. 68, no. 10,  pp. 1041-1049.<br>
 * <i>Pontius Jr, R.G., Huffaker, D. &amp; K. Denman (2004)</i>: Useful Techniques of Validation for Spatially Explicit Land-Change Models. 
 * In: Ecological Modelling, vol. 179, no. 4, pp. 445-461.<br>
 * 
 * The following input parameters/images are required:<br>
 * <ul>
 *  <li><b>Number of Steps ({@code steps})</b>: Because this plugin is built as a step model, the number of (time) steps always has to be 1.</li>
 *  <li><b>Reference Map ({@code referenceMap})</b>: A raster containing the reference (observed) land cover.</li>
 *  <li><b>Simulation Map ({@code simulationMap})</b>: A raster containing the simulated land cover.</li>
 * </ul>
 * 
 * Furthermore the following outputs are created:<br>
 * <ul>
 *  <li><b>Actual Validation (@code actualValidation)</b>: A raster containing the actual resolution grid with the fraction correct in every cell</li>
 *  <li><b>Out Step ({@code outStep})</b>: A Grid List containing a grid (Actual Validation) for every resolution.</li>
 *  <li><b>Out Results ({@code outResults})</b>: A list containing the total agreement for every resolution.</li>
 * </ul>
 * 
 * Other important values are written to the status output window!<br>
 * 
 * @see MultipleResolutionValidationContentManager
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 *
 */
public class MultipleResolutionValidation extends AbstractStepModel {

	/**
	 * Contains the actual pixel size (from 1 to n, with n being one coarse cell that contains the whole study area).
	 * During the MRV it is subsequently aggregated in a geometric progression as a multiple of 2.
	 */
	private int neighborhoodRange = 0;
	
	/** Array containing the fraction correct for every resolution. */
	private float[][] resultArray;
	
	/**
	 * Saves the ContentManager for the model.
	 * 
	 * @see MultipleResolutionValidationContentManager
	 */
	protected MultipleResolutionValidationContentManager contManager;

	// ********** Read/Write access, that has to be maintained ***
	// ********** through the whole model run ********************
	private PropertyReadAccess RA_steps = null;	//Number of Steps
	private PropertyReadAccess RA_referenceMap = null; // Reference Map
	private PropertyReadAccess RA_simulationMap = null;	//Simulation Map
	private PropertyWriteAccess WA_actualValidation = null; // Output Grid
	private PropertyWriteAccess WA_outStep = null;	// Step Results
	private ListPropertyWriteAccess WA_outResults = null;	//Output Validation Results

	// **************** Variables to work with *******************
	private PropertyReadAccess steps = null; // Number of steps
	private WritableGrid referenceMap = null; // Reference Map
	private WritableGrid simulationMap = null;	// Simulation Map
	private WritableGrid actualValidation = null; // Output Grid
	private MultiGrid outStep = null;	//Step Results
	private ListPropertyWriteAccess outResults = null;	//Output Validation Results


	/**
	 * Creates a new model instance.
	 */
	public MultipleResolutionValidation() {
		super(new MultipleResolutionValidationContentManager());
		this.contManager = (MultipleResolutionValidationContentManager) super.contManager;
	}

	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		if (contManager == null)
			return;
		// Get access authorisation of resource/properties
		RA_steps = null;
		if (contManager.getResource(0).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(0).getData())
					.getReadAccess(this);
		RA_referenceMap = null;
		if (contManager.getResource(1).getData() != null)
			RA_referenceMap = ((ScalarProperty) contManager.getResource(1)
					.getData()).getReadAccess(this);
		RA_simulationMap = null;
		if (contManager.getResource(2).getData() != null)
			RA_simulationMap = ((ScalarProperty) contManager.getResource(2)
					.getData()).getReadAccess(this);
		WA_actualValidation = null;
		if (contManager.getResource(3).getData() != null)
			WA_actualValidation = ((ScalarProperty) contManager.getResource(3)
					.getData()).getWriteAccess(this);
		WA_outResults = null;
		if (contManager.getResource(5).getData() != null)
			WA_outResults = ((ListProperty) contManager.getResource(5)
					.getData()).getWriteAccess(this);

		//Assign variables to work with
		steps = RA_steps;
		referenceMap = (WritableGrid) RA_referenceMap.getValue();
		simulationMap = (WritableGrid) RA_simulationMap.getValue();
		actualValidation = (WritableGrid) WA_actualValidation.getValue();

		this.stepCount = steps.getValueAsInt();
		outStep = (MultiGrid)contManager.getResource(4).getData();
		outResults = WA_outResults;
	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {

		// free resources
		releaseAccess(RA_steps);
		releaseAccess(RA_referenceMap);
		releaseAccess(RA_simulationMap);
		releaseAccess(WA_actualValidation);
		releaseAccess(WA_outStep);
		releaseAccess(WA_outResults);
	}

	/**
	 * Performs the MRV. The function is implemented as a Xulu Step Model, but actually does not need any time steps. So
	 * the stepNo MUST always be 1. The steps of the coarsening procedure are defined inside of the function. When using a 
	 * MRV parameter file simply use 1 for the number of steps.
	 */
	public void performModelStep(int stepNo) {
		
		statusOut.println("Starting step " + stepNo + "...");
		
		int RoverallSum = 0; //Total number of pixels in Reference Map
		int RnumClasses = 0; //Total number of classes in Reference Map
		int SoverallSum = 0; //Total number of pixels in Simulation Map
		int SnumClasses = 0; //Total number of classes in Simulation Map
		int correct = 0;	//Number of pixels counted correct
		double fractionCorrect = 0f;	//Fraction correct
		long localStartTime = System.currentTimeMillis(); //current Start Time
		int numSteps = 0;
		
		neighborhoodRange = (int) Math.sqrt(referenceMap.getWidth()*referenceMap.getHeight()); //maximum possible pixel size
		numSteps = (int) ((int)1+ Math.floor(Math.log(neighborhoodRange)/Math.log(2)));	//calculates how many steps are needed by doubling the pixel size until neighborhoodRange is reached. It is the reverse of an exponential function, which is the logarithm of neighborhoodRange to the basis 2. 2 is added because the calculation starts with 1 and not with 2 (1,2,4,8,16,32,...) and an array starts with 0.
		statusOut.println("Maximum pixel size (as multiple of original resolution): "+neighborhoodRange);
		statusOut.println("Number of steps: "+numSteps);
		resultArray = new float[2][1+numSteps];
		
		//Calculation of the fraction correct for the original Resolution
		for (int x = 0; x < referenceMap.getWidth(); x++)
			for (int y = 0; y < referenceMap.getHeight(); y++) {
				float Rtmp = referenceMap.getRasterSampleAsFloat(x, y); //get value of each raster cell in reference map
				float Stmp = simulationMap.getRasterSampleAsFloat(x, y);//get value of each raster cell in simulation map
				if (!Float.isNaN(Rtmp) && !Float.isNaN(Stmp)){	//if the value is a number, continue ...
					RoverallSum ++;		//count all cells in reference map
					SoverallSum ++;		//count all cells in simulation map
					if (Rtmp > RnumClasses)
						RnumClasses = (int) Rtmp;	//Number of classes in reference map. If a higher cell value is detected than the actual RnumClasses, the number of classes is increased to that value
					if (Stmp > SnumClasses)
						SnumClasses = (int) Stmp;	//see RnumClasses
					if (Rtmp == Stmp)
						correct++;		//if the same cell value is detected in the simulation and the reference map, increase the number of correct pixels
				}
		}
		fractionCorrect = (double) correct / (double) RoverallSum;	//divide the number of correct pixels by the number of total pixels

		//Counts pixels per class
		for (int i = 1; i <= RnumClasses; i++){		//Loop, until you reach the total number of classes
			int Rj = 0;		//Number of classes in Reference Map
			int Sj = 0;		//Number of classes in Simulation Map
			for (int x = 0; x < referenceMap.getWidth(); x++)
				for (int y = 0; y < referenceMap.getHeight(); y++){
					float Rtmp = referenceMap.getRasterSampleAsFloat(x, y);	//get value of each raster cell in reference map
					float Stmp = simulationMap.getRasterSampleAsFloat(x, y);//get value of each raster cell in simulation map
					if (!Float.isNaN(Rtmp))
						if ((int) Rtmp == i)
							Rj++;	//if cell belongs to category i, count it (for Reference map)
					if (!Float.isNaN(Stmp))
						if ((int) Stmp == i)
							Sj++;	//if cell belongs to category i, count it (for Simulation map)
				}
			statusOut.println("Klasse "+i+": Referenz("+Rj+"), Simulation("+Sj+")");
		}
		
		int ngrids = 0;		//number of resolution grids
		float Vg = 0f;	//sum of the individual cell weights for the calculation of the average weighted fit after Pontius (2002)
		float VgWeights = 0f;	//sum of the cell weights
		float WeightedAgreement = 0f;	//total weighted agreement over all resolutions after Pontius (2002)
		for (int r = 1; r <= neighborhoodRange; r = r+r){	//for pixel size 1 up to the maximum pixel size, where one large cell covers the whole area
			statusOut.println(r);
			ngrids++;		//count the number of grids
			outStep.addGrid();	//add a grid to the outStep Output GridList
			float SumWn = 0f;	//Sum of the weighted means 
			int SumCells = 0;	//Number of the cells
			float totalAgreement = 0f;	//total agreement is the weighted mean of the fraction correct over all resolution grids
			for (int x = 0; x < referenceMap.getWidth(); x = x+r)	//for every pixel size...
				for (int y = 0; y < referenceMap.getHeight(); y = y+r){
					int nCells = 0;		//number of cells making up a coarse cell
					float SumMinRS = 0f;	//when two maps are compared, the minimum value is taken (MIN), this value is the sum of all these MINs
					float Wn = 0;	//weighted cell
					for(int x2 = x; x2 < x+r; x2++)	//for every cell within a coarse cell...
						for(int y2 = y; y2 < y+r; y2++){
							if ( x2 < referenceMap.getMinX() || y2 < referenceMap.getMinY() || x2 > referenceMap.getMinX()+referenceMap.getWidth()-1 || y2 > referenceMap.getMinY()+referenceMap.getHeight()-1 ||
									Float.isNaN(referenceMap.getRasterSampleAsFloat(x2,y2)))
								continue;
							nCells++;	//count all cells that have a value
						}
					for (int i = 0; i <= RnumClasses; i++){	//for every class i
						int Rn = 0;		//Number of cells in the reference map
						int Sn = 0;		//Number of cells in the simulation map
						float Rnj = 0f;	//Portion of cells in the reference map
						float Snj = 0f;	//Portion of cells in the simulation map
						float MinRS = 0f;	//Both maps are compared and the Minimum (MIN) for each cell taken
						for(int x3 = x; x3 < x+r; x3++)	//for every cell within a coarse cell...
							for(int y3 = y; y3 < y+r; y3++){
								if ( x3 < referenceMap.getMinX() || y3 < referenceMap.getMinY() || x3 > referenceMap.getMinX()+referenceMap.getWidth()-1 || y3 > referenceMap.getMinY()+referenceMap.getHeight()-1 ||
										Float.isNaN(referenceMap.getRasterSampleAsFloat(x3,y3)))
									continue;
								float Rtmp = referenceMap.getRasterSampleAsFloat(x3, y3);	//get value of each cell in the reference map
								float Stmp = simulationMap.getRasterSampleAsFloat(x3, y3);	//get value of each cell in the simulation map
								if ((int) Rtmp == i)
									Rn++;	//count all cells that belong to class i in reference map 
								if ((int) Stmp == i)
									Sn++;	//count all cells that belong to class i in simulation map
							}
						if(nCells > 0){	//if positive cells are available...
							Rnj = (float)Rn / (float)nCells;	//compute the portion of these cells within a coarse cell in the reference map
							Snj = (float)Sn / (float)nCells;	//compute the portion of these cells within a coarse cell in the simulation map
						}
						MinRS = Math.min(Rnj, Snj);	//take for every coarse cell the MIN between R and S of the portion of class i
						SumMinRS += MinRS;	//Sum all the MIN values
						
					}
					Wn = (float)nCells * SumMinRS;	//all coarse cells are weighted by multiplying the Sum of all MIN values with the number of cells that make up a coarse cell
					 
					//Compute output maps
					float WnCell = 0f;
					for(int x4 = x; x4 < x+r; x4++)	//for every cell within a coarse cell...
						for(int y4 = y; y4 < y+r; y4++){
							if ( x4 < referenceMap.getMinX() || y4 < referenceMap.getMinY() || x4 > referenceMap.getMinX()+referenceMap.getWidth()-1 || y4 > referenceMap.getMinY()+referenceMap.getHeight()-1 ||
									Float.isNaN(referenceMap.getRasterSampleAsFloat(x4,y4)))
								continue;
							WnCell = Wn / (float)nCells; //divide the weighted sum of the MIN values by the number of positive cells
							actualValidation.setRasterSample( WnCell ,x4 ,y4);	//for every pixel in the Actual Comparison Grid take this value
							outStep.getGrid(ngrids-1).setRasterSample(WnCell,x4,y4);	//and write the same grid into the Output Grid List
						}
				
					SumCells += (float)nCells;	//Sum all positive cells in a resolution grid
					SumWn += Wn;	//Sum all weighted portions in a resolution grid
				}
			
			//Total Agreement:
			statusOut.println(SumWn+" / "+SumCells);
			totalAgreement = SumWn / (float)SumCells;	//Main Results
			statusOut.println(totalAgreement);	
			outResults.addValue(totalAgreement);
			resultArray[0][ngrids] = r;
			resultArray[1][ngrids] = totalAgreement;	//write the results in an array 
			//the result array looks like this:
			//1	   |2	 |4	   |8    |...
			//0.876|0.892|0.913|0.948|...
			Vg += SumWn * totalAgreement;
			VgWeights += SumWn;
		}
		
		//From here on the results of the Multiple Resolution Comparison are written on the screen
		
		//those variables are needed to compute the weighted average of all fits after Costanza (1989)
		float w = 0f;	//weight depending on resolution
		float Fw = 0f;	//weighted fit at certain resolution
		float SumW = 0f;	//Sum of all weights
		float SumFw = 0f;	//Sum of all weighted agreements
		float Ft = 0f;
		
		for(int i = 1; i <= ngrids; i++){
			//print every element of the result array comma separated in one line in the status output
			System.out.print(resultArray[1][i]+",");
			//calculations for weighted average after Costanza (1989):
			w = (float) Math.exp(-0.1*(resultArray[0][i] - 1));
			Fw = resultArray[1][i]*w;
			SumW += w;
			SumFw += Fw;
		}
		
		Ft = SumFw / SumW;	//Final calculation of the weighted average of all fits after Costanza (1989)
		WeightedAgreement = Vg / VgWeights;	//Final calculation of the weighted average of all fits after Pontius (2002)
		System.out.print("Ft:"+Ft+",TA:"+WeightedAgreement);	//the result is written directly behind the single fits for every resolution
		
		System.out.println("");
		System.out.print("Finished step " + (stepNo) + " in "
				+ ((System.currentTimeMillis() - localStartTime)) + " ms");
		localStartTime = System.currentTimeMillis();
		System.out.println(" with sum (local calculated): " + RoverallSum);
		
	}
}
