package edu.bonn.xulu.plugin.model.sleuth;

import java.util.HashMap;

import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;

import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;

/**
 *  <b>Urban Growth Model</b><br>
 *  This Model simulates urban growth. It is based on the algorithm of the SLEUTH-Model, which has originally been written in C code and could
 *  be run under UNIX. The SLEUTH-Model has been developed by Keith Clarke at the UC Santa Barbara, CA, within the Gigalopolis Project, funded by the
 *  USGS. Therefore it is under public domain. The name SLEUTH has been derived from the image input requirements of the model:
 *  Slope, Land cover, Exclusion, Urbanization, Transportation, and Hillshade.<br>
 *  In the original SLEUTH-Model the core Urban Growth Model (UGM) calls and drives an additional land cover deltatron model (LCD). Both models are
 *  tightly coupled, but the UGM can be run separately. Here, only the UGM part of the SLEUTH-Model has been translated into JAVA. Minor changes
 *  in the code have been necessary. Those changes are explained within the code. In general, the UGM is a Cellular Automaton (CA). All CA elements 
 *  in the code have been untouched and are the same as in the original UGM.<br>
 *  This is the basic UGM code, a plain Cellular Automaton. Additional functions are implemented in other models, like:<br>
 *  <ul>
 *   <li><b>UGM MonteCarlo ({@link UrbanGrowthModel_MC UrbanGrowthModel_MC})</b>: The same model algorithm like this, but with Monte Carlo Iterations. The resulting map shows the probabilities for new urbanization (@link MultipleResolutionValidation MultipleResolutionValidation)</li>
 *   <li><b>UGM Calibration ({@link UrbanGrowthModelCalibration UrbanGrowthModelCalibration})</b>: The same model algorithm like this, but in order to calibrate the model by finding the correct parameter values with the method of Multiple Resolution Comparison (@link MultipleResolutionValidation)</li>
 *   <li><b>UGM SelfModifying({@link UrbanGrowthModelSelfModification UrbanGrowthModelSelfModifying})</b>: The same model algorithm like this, but with the possibility of self-modification of the parameter values.</li>
 *   <li><b>UGM CalibrationSelfModifying({@link UrbanGrowthModelCalibrationSelfModification UrbanGrowthModelCalibrationSelfModification})</b>: The same model algorithm like this, but in order to calibrate the model by finding the correct parameter values with the method of Multiple Resolution Comparison (@link MultipleResolutionValidation)
 *    While calibrating the model modifies the parameter values.</li>
 *  </ul>
 *  <br>
 *  The original SLEUTH-Model can be downloaded via the <a href="http://www.ncgia.ucsb.edu/projects/gig/">Gigalopolis Website</a>. To run it, a
 *  UNIX environment with the GNU C compiler (gcc) are required.<br>
 *  For details refer to:<br>
 *  <i>Clarke, K.C., Hoppen, S. & L.J. Gaydos (1997)</i>: A self-modifying cellular automaton model of historical urbanization in the San Francisco Bay area.
 *  In: Environment and Planning B: Planning and Design, vol. 24, pp. 247-261.<br>
 *  <i>Clarke, K.C. & L.J. Gaydos (1998)</i>: Loose-coupling a cellular automaton model and GIS: long-term urban growth prediction for San Francisco
 *  and Washington/Baltimore. In: International Journal of Geographical Information Science, vol. 12, no. 7, pp. 699-714.<br>
 *  <i>Jantz, C.A., Goetz, S.J. & M.K. Shelley (2003)</i>: Using the SLEUTH urban growth model to simulate the impacts of future policy scenarios
 *  on urban land use in the Baltimore-Washington metropolitan area. In: Environment and Planning B: Planning and Design, vol: 30, pp. 251-271.<br>
 *  
 *  <br>
 *  The following input parameters/images are required:<br>
 * <ul>
 *  <li><b>Input Grid ({@code inputGrid})</b>: A grid containing the input (initial) urban land cover.</li>
 *  <li><b>Area Restriction ({@code areaRestr})</b>: A grid containing the area restrictions (e.g. water, natural reserves, areas with lower probability, etc.).</li>
 *  <li><b>Transportation Grid ({@code roadGrid})</b>: A grid containing the road network.</li>
 *  <li><b>Slope Grid ({@code slopeGrid})</b>: A grid containing the slope.</li>
 *  <li><b>Number of steps ({@code steps})</b>: The number of time steps to compute.</li>
 *  <li><b>Spread ({@code spread})</b>: The Spread-Coefficient.</li>
 *  <li><b>Disp ({@code disp})</b>: The Dispersion(Diffusion)-Coefficient.</li>
 *  <li><b>Breed ({@code breed})</b>: The Breed-Coefficient.</li>
 *  <li><b>Road Gravity ({@code roadGravity})</b>: The Road Gravity-Coefficient.</li>
 *  <li><b>Slope ({@code slope})</b>: The Slope-Coefficient.</li>
 *  <li><b>Critical Slope ({@code criticalSlope})</b>: The Percentage at which urbanization on steep slopes is not possible.</li>
 * </ul>
 * 
 * <br>
 * Furthermore the following outputs are created:<br>
 * <ul>
 *  <li><b>Output Grid ({@code outputGrid})</b>: A grid containing the output, that is actualized in every time step.</li>
 *  <li><b>outStep ({@code outStep})</b>: A Grid List containing the results for every time step.</li>
 * </ul>
 * 
 * @see UrbanGrowthModelContentManager
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 */
public class UrbanGrowthModel extends AbstractStepModel {

	/**
	 * Saves the ContentManager for the model.
	 * 
	 * @see UrbanGrowthModelContentManager
	 */
	//protected UrbanGrowthModelContentManager contManager;

	// ********** Read/Write access, that has to be maintained ***
	// ********** through the whole model run ********************
	protected PropertyReadAccess RA_steps = null; // Number of steps
	protected PropertyReadAccess RA_inputGrid = null; // Input Grid
	protected PropertyWriteAccess WA_outputGrid = null; // Output Grid
	protected PropertyReadAccess RA_areaRestr = null; //Area Restrictions
	protected PropertyReadAccess RA_roadGrid = null;	//Road Grid
	protected PropertyReadAccess RA_slopeGrid = null;	//Slope Grid
	protected PropertyReadAccess RA_spread = null;	//Spread-Koeffizient
	protected PropertyReadAccess RA_disp = null;		//Dispersions-Koeffizient
	protected PropertyReadAccess RA_breed = null;		//Breed-Koeffizient
	protected PropertyReadAccess RA_roadGravity = null;	//Road-Gravity-Koeffizient
	protected PropertyReadAccess RA_slope = null;	//Slope-Koeffizient
	protected PropertyReadAccess RA_criticalSlope = null;	//Critical Slope Value
	protected PropertyWriteAccess WA_outStep = null;	// Step Results

	// **************** Variables to work with *******************
	protected PropertyReadAccess steps = null; // Number of steps
	protected WritableGrid inputGrid = null; // Input Grid
	protected WritableGrid outputGrid = null; // Output Grid
	protected WritableGrid areaRestr = null; //Area Restrictions
	protected WritableGrid roadGrid = null;	//Road Grid
	protected WritableGrid slopeGrid = null;	//Slope Grid
	protected double spread = 0;	//Spread-Koeffizient
	protected double disp = 0;	//Dispersions-Koeffizient
	protected double breed = 0;	//Breed-Koeffizient
	protected double roadGravity = 0;	//Road-Gravity-Koeffizient
	protected double slope = 0;	//Slope-Koeffizient
	protected double criticalSlope = 0;	//Critical Slope Value
	protected MultiGrid outStep = null;	//Step Results
		
	/**
	 * Moore Neighborhood as a two dimensional array. It's the central element of a Cellular Automaton.
	 */
	protected  static final int[][] NEIGHBOR = new int[][] {
	      {-1,-1}, {0,-1}, {1,-1},
	      {-1, 0},         {1, 0},
	      {-1, 1}, {0, 1}, {1, 1}
	  };
	
	/**
	 * A Lookup-Table which relates actual slope values to slope coefficient influenced probabilities.
	 */
	protected  HashMap<Number,Number> slopeLUT;
	
	protected int areaMinX = 0;
	protected int areaMinY = 0;
	protected int areaMaxX = 0;
	protected int areaMaxY = 0;
	protected int areaWidth = 0;
	protected int areaHeight = 0;
	
	protected double disp_value = 0;	//Dispersion Value
	protected double max_RG_value = 0;	//Maximum RoadGravity Value
	protected double rg_value = 0;		//RoadGravity Value
	protected double max_search_index = 0;	//Maximum search index for road
	
	/**
	 * Creates a new model instance.
	 */
	public UrbanGrowthModel(UrbanGrowthModelContentManager contManager) {
		super(contManager);
		this.contManager = (UrbanGrowthModelContentManager) super.contManager;
	}
	
	public UrbanGrowthModel(){
		super(new UrbanGrowthModelContentManager());
	}

	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		
		if (contManager == null)
			return;
		// Get access authorisation of resource/properties
		RA_steps = null;
		if (contManager.getResource(0).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(0).getData())
					.getReadAccess(this);
		RA_inputGrid = null;
		if (contManager.getResource(1).getData() != null)
			RA_inputGrid = ((ScalarProperty) contManager.getResource(1)
					.getData()).getReadAccess(this);
		WA_outputGrid = null;
		if (contManager.getResource(2).getData() != null)
			WA_outputGrid = ((ScalarProperty) contManager.getResource(2)
					.getData()).getWriteAccess(this);
		RA_areaRestr = null;
		if (contManager.getResource(3).getData() != null)
			RA_areaRestr = ((ScalarProperty) contManager.getResource(3)
					.getData()).getReadAccess(this);
		RA_roadGrid = null;
		if (contManager.getResource(4).getData() != null)
			RA_roadGrid = ((ScalarProperty) contManager.getResource(4)
					.getData()).getReadAccess(this);
		RA_slopeGrid = null;
		if (contManager.getResource(5).getData() != null)
			RA_slopeGrid = ((ScalarProperty) contManager.getResource(5)
					.getData()).getReadAccess(this);
		RA_spread = null;
		if (contManager.getResource(6).getData() != null)
			RA_spread = ((ScalarProperty) contManager.getResource(6).getData())
					.getReadAccess(this);
		RA_disp = null;
		if (contManager.getResource(7).getData() != null)
			RA_disp = ((ScalarProperty) contManager.getResource(7).getData())
					.getReadAccess(this);
		RA_breed = null;
		if (contManager.getResource(8).getData() != null)
			RA_breed = ((ScalarProperty) contManager.getResource(8).getData())
					.getReadAccess(this);
		RA_roadGravity = null;
		if (contManager.getResource(9).getData() != null)
			RA_roadGravity = ((ScalarProperty) contManager.getResource(9).getData())
					.getReadAccess(this);
		RA_slope = null;
		if (contManager.getResource(10).getData() != null)
			RA_slope = ((ScalarProperty) contManager.getResource(10).getData())
					.getReadAccess(this);
		RA_criticalSlope = null;
		if (contManager.getResource(11).getData() != null)
			RA_criticalSlope = ((ScalarProperty) contManager.getResource(11).getData())
					.getReadAccess(this);

		// Assign variables to work with
		inputGrid = (WritableGrid) RA_inputGrid.getValue();
		outputGrid = (WritableGrid) WA_outputGrid.getValue();
		steps = RA_steps;
		this.stepCount = steps.getValueAsInt();
		areaRestr = (WritableGrid) RA_areaRestr.getValue();
	    spread  = RA_spread.getValueAsDouble();
	    disp	= RA_disp.getValueAsDouble();
	    breed	= RA_breed.getValueAsDouble();
	    roadGravity = RA_roadGravity.getValueAsDouble();
	    slope	= RA_slope.getValueAsDouble();
	    criticalSlope = RA_criticalSlope.getValueAsDouble();
	    roadGrid = (WritableGrid) RA_roadGrid.getValue();
	    slopeGrid = (WritableGrid) RA_slopeGrid.getValue();
	    outStep = (MultiGrid)contManager.getResource(12).getData();
	    
	    //count the number of starting individuals
	    int startAnz = 0;
		startAnz = countSetIndividuals();
		
		statusOut.println("Number of starting cells: "+startAnz);
		statusOut.println("Neighborhood size: "+NEIGHBOR.length);
		
		//calculate the maximum slope
		int slopeMax = 0;
		slopeMax = getSlopeMax();
		
		statusOut.println("Max. Slope: "+slopeMax+" %");
				
		//Creates a new Lookup-table
		//slopeLUT = new HashMap<Number,Number>();
		
		//assigns for every slope value between 0 and the maximum slope that is below the critical slope the value calculated for the variable "lookup".
		//all values higher than the critical slope become 1

		int MaxSlopeResistance = 0;
		MaxSlopeResistance = 100;	//Max slope resistance is 100%
		
		double exp = getSlopeExp(slope,MaxSlopeResistance);
		slopeLUT = getSlopeLookup(slopeMax,exp,criticalSlope);
		
		//raster dimensions
		areaMinX = inputGrid.getMinX();
		areaMinY = inputGrid.getMinY();
		areaMaxX = inputGrid.getMinX() + inputGrid.getWidth() - 1;
		areaMaxY = inputGrid.getMinY() + inputGrid.getHeight() - 1;
		areaWidth = inputGrid.getWidth();
		areaHeight = inputGrid.getHeight();
		
		//coefficient calculations of for growth rules
		disp_value = Math.round((disp*0.005)* Math.sqrt(Math.pow(inputGrid.getWidth(), 2)+(Math.pow(inputGrid.getHeight(), 2))));	//Dispersion value (Number of cells to select spontaneously)
		statusOut.println("Dispersion-Value: "+disp_value);
		max_RG_value = 100.0;			//Maximum possible road gravity value
		rg_value = Math.round((roadGravity/max_RG_value)*((areaWidth + areaHeight)/16)); 	//calculation of the road gravity value
		statusOut.println("RoadGravity-Value: "+rg_value);
		max_search_index = Math.round(4*((int)rg_value*(1+(int)rg_value)));	//area in which to look for a road cell
		statusOut.println("MaxSearchIndex: "+max_search_index);
		
		//outputGrid = inputGrid;
	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {

		// free resources
		releaseAccess(RA_steps);
		releaseAccess(RA_inputGrid);
		releaseAccess(WA_outputGrid);
		releaseAccess(RA_areaRestr);
		releaseAccess(RA_spread);
		releaseAccess(RA_disp);
		releaseAccess(RA_breed);
		releaseAccess(RA_roadGravity);
		releaseAccess(RA_slope);
		releaseAccess(RA_criticalSlope);
		releaseAccess(RA_roadGrid);
		releaseAccess(RA_slopeGrid);
		releaseAccess(WA_outStep);
	}
	
	/**
	 * This is the main run of the Urban Growth Model. Depending on the input parameters the cellular automaton looks in the neighborhood of individual cells for specific conditions
	 * and turns them from non-urban into urban, following certain growth rules. Those growth rules are:<br>
	 *  <ol>
	 *   <li><b>spontaneous growth</b>: controlled by dispersion (@code disp): Randomly selects potential new growth cells.</li>
	 *   <li><b>new spreading centers</b>: controlled by breed (@code breed): Growing urban centers from spontaneous growth.</li>
	 *   <li><b>edge</b>: controlled by spread (@code spread): Old or new urban centers spawn additional growth.</li>
	 *   <li><b>road influenced</b>: controlled by road-gravity(@code roadGravity), dispersion (@code disp), breed (@code breed): Newly urbanized cell spawns growth along transportation network.</li>
	 *  </ol>
	 * @exception IllegalArgumentException 
	 * @param stepNo
	 */
	public void performModelStep(int stepNo) {
		
		statusOut.println("Starting step " + stepNo + "...");
		
		//time measure
		long localStartTime = System.currentTimeMillis();
		
		//cells to be created
		boolean[][] tmpGrid = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		Float actUrb = null;	//urban cell in the raster

	
		/*****************************************************************
		 * DISPERSION and BREED (Growth Phase 1 and 2)
		 *****************************************************************/
		boolean[][] tmpGridDisp = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridDisp = DispersionGrowth(disp_value,breed);	//calls the method "DispersionBreed" to perform the Dispersion and Breed Growth Phase
		int anzDispBreed = 0;	//number of cells urbanized by Dispersion and Breed
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridDisp[X][Y]==true){
					anzDispBreed++;
					tmpGrid[X][Y] = true;	//write the temporary Dispersion and Breed Cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Spontaneous and New Spreading Center Cells: "+anzDispBreed);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
		
				
		/*********************************************************************
		 * EDGE
		 *********************************************************************/
		boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridSpread = EdgeGrowth(spread);	//calls the method "EdgeGrowth" to perform the Spread (Edge Growth) Phase
		int anzSpread = 0;	//number of cells urbanized by Spread
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridSpread[X][Y]==true){
					anzSpread++;
					tmpGrid[X][Y] = true;	//write the temporary Spread cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Edge Growth Cells: "+anzSpread);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
	
		/*********************************************************************
		 * ROAD
		 *********************************************************************/
		boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridRoad = RoadGrowth(tmpGrid,rg_value,breed,disp);	//calls the method "RoadGrowth" to perform the Road weighted growth
		int anzRoad = 0;	//number of cells urbanized by Road Growth
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridRoad[X][Y]==true){
					anzRoad++;
					tmpGrid[X][Y] = true;	//write the temporary Road Growth cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Road Growth Cells: "+anzRoad);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
		
		
		outStep.addGrid();
		int anzGesamt = 0;
		
		// Write the temporary Grid to the Output Grid
		for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
			for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
				
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				
				if (tmpGrid[x][y] == true)
					anzGesamt++;
				
				boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive, when it is 1
				if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
					tmpGrid[x][y] = lebt;	//take the values from the current output raster
				

				outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the results in the output raster
				actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
				outStep.getGrid(stepNo-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
			}
		}
		
		statusOut.println("Urban pixels at step "+stepNo+": "+anzGesamt);
	
		System.out.println("Finished step " + (stepNo) + " in "
				+ ((System.currentTimeMillis() - localStartTime)) + " ms\n");
		localStartTime = System.currentTimeMillis();
	}
	
	/**
	 * This function counts the starting individual cells for the urban growth model run and sets the outputGrid
	 * to the base inputGrid.<br>
	 * It only works for a binary classification with urban cells having the value 1 and non-urban
	 * cells the value 0.
	 * @return startAnz
	 */
	public int countSetIndividuals(){
		int startAnz = 0;
		for (int i = 0; i < inputGrid.getWidth();i++ ) {
			for (int ii = 0; ii < inputGrid.getHeight(); ii++ ) {
				float val = inputGrid.getRasterSampleAsFloat(i,ii);
				if (val == 1.0 ) {
					val = 1f;
					startAnz++;		//count if value is 1
				}
			else val = 0f;
			inputGrid.setRasterSample(val, inputGrid.getMinX() + i, inputGrid.getMinY() + ii);
			outputGrid.setRasterSample(val, inputGrid.getMinX() + i, inputGrid.getMinY() + ii);
			}
		}
		return startAnz;
	}
	
	/**
	 * Calculates the maximum slope from the Slope Raster.
	 * @return slopeMax
	 */
	public int getSlopeMax(){
		int slopeMax = 0;
		for (int s = 0; s < inputGrid.getWidth();s++){
			for(int ss = 0; ss < inputGrid.getHeight();ss++){
				float slopeVal = slopeGrid.getRasterSampleAsFloat(s,ss);
				if (slopeVal > slopeMax){
					slopeMax = (int) slopeVal;	//if the actual slope is larger than the highest measured slope up to now, increase the maximum value to that point
				} 
			}
		}
		return slopeMax;
		
	}
	
	/**
	 * Takes a slope value from the slope lookup-table depending on the given cell in the slope raster.
	 * @return slopeDoubleNb
	 */
	public double getSlopeValue(int X,int Y, HashMap<Number,Number> slopeLUT){
		int getSlopeNb = (int) slopeGrid.getRasterSampleAsFloat(X, Y);	//takes a slope value from the slope raster
		Double doubleObjNb = (Double)slopeLUT.get(getSlopeNb);	//looks for that slope value in the Lookup-Table and takes it as a double number
		double slopeDoubleNb = doubleObjNb.doubleValue();	//converts the double object into a double value.
		return slopeDoubleNb;		//returns the slope value
	}
	
	/**
	 * Calculates the exp value that is needed to generate the slope lookup-table 
	 * @param slope
	 * @param MaxSlopeResistance
	 * @return exp
	 */
	public double getSlopeExp(double slope, int MaxSlopeResistance){
		double exp = (slope / (MaxSlopeResistance / 2.0));
		return exp;
	}
	
	public HashMap<Number,Number> getSlopeLookup(int slopeMax,double exp,double criticalSlope){
		//Creates a new Lookup-table
		HashMap<Number, Number> newLUT = new HashMap<Number,Number>();
		double[] lookup = new double[slopeMax+1];
		Number key = 0;
		for (int i = 0;i <= slopeMax;i++){
			if (i <= criticalSlope){
				double val = (criticalSlope - i) / criticalSlope;
				lookup[i] = 1 - Math.pow(val,exp);
				key = i;
			} else{
				lookup[i] = 1;
				key = i;
			}
			newLUT.put(key,lookup[i]);
			//statusOut.println("Lookup: "+key+", "+lookup[i]);
		}	
		return newLUT;
	}
	
	/**
	 * This is phase 1 of the growth cycle, where spontaneous growth occurs (DISPERSION). Those new cells can turn out to be
	 * new spreading centers by urbanizing neighboring pixels (BREED {@link BreedGrowth}).
	 * @return tmpGridDispBreed
	 */
	public boolean[][] DispersionGrowth(double disp_value,double breed){
		//DISPERSION
		int dispRandX = 0;	//x-coordinate for randomly created cell
		int dispRandY = 0;	//y-coordinate for randomly created cell
		float slopeRandDisp = 0;	//random value in order to compare the selected dispersion cell with the slope in that cell
		int anzDispTreffer = 0;		//Dispersion
		int anzBreedTreffer = 0;
		boolean[][] tmpGridDispBreed = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		int p = 0;	//Distribute so many new cells randomly in suitable places, until disp_value is reached.
		do{
			dispRandX = Math.round((float) Math.random()*areaWidth);	//Random position in x-direction
			dispRandY = Math.round((float) Math.random()*areaHeight);	//Random position in y-direction
			if ( dispRandX < areaMinX || dispRandY < areaMinY || dispRandX > areaMaxX || dispRandY > areaMaxY ||
					Float.isNaN(outputGrid.getRasterSampleAsFloat(dispRandX,dispRandY)) ) 
				continue;		//if a cell is outside the raster (NaN), break and continue loop from the beginning
			if (areaRestr.getRasterSampleAsFloat(dispRandX,dispRandY) != 0 || outputGrid.getRasterSampleAsFloat(dispRandX,dispRandY) == 1)
	        	continue;		//if a cell is inside an excluded area or already exists, break and continue loop from the beginning
	        slopeRandDisp = (float) Math.random();	//a random slope value between 0.0 and 1.0
	        double slopeDouble = getSlopeValue(dispRandX,dispRandY,slopeLUT);	//returns the slope value for the cell
	        if(slopeDouble < slopeRandDisp){	//if the slope value of the pixel is smaller than the slope value
	        	tmpGridDispBreed[dispRandX][dispRandY] = true;	//true is written in the temporary raster
	        	anzDispTreffer++;		//count number of dispersion cells
	        	
	        	//BREED
	        	float breedRnd = (float) Math.random()*100;		//Breed Random value
	        	if (tmpGridDispBreed[dispRandX][dispRandY] == true && breedRnd < breed){	//if there is a new dispersion cell and you pass the Breed-Test
	        		int[][] tmpBreedArray = new int[2][2];	//an array containing the two x and y coordinates of two urbanized cells by breed
	        		tmpBreedArray = BreedGrowth(dispRandX,dispRandY);	//perform a breed for this cell
	        		int x1 = tmpBreedArray[0][0];	//x-coordinate of first cell
	        		int y1 = tmpBreedArray[0][1];	//y-coordinate of first cell
	        		int x2 = tmpBreedArray[1][0];	//x-coordinate of second cell
	        		int y2 = tmpBreedArray[1][1];	//y-coordinate of second cell
	        		tmpGridDispBreed[x1][y1] = true;	//make the temporary grid true at that position
	        		tmpGridDispBreed[x2][y2] = true;
	        		anzBreedTreffer += 2;	//count the hits
	        	}
			} p++;
		} while (p < disp_value);	//while instead of for-loop, because it can only be checked in the end
									//if all cells could be distributed, which is not known at the beginning of the loop.
									//a part of the cells falls in excluded areas, etc.
		statusOut.println("Dispersion-Treffer: "+anzDispTreffer);
		statusOut.println("Breed-Treffer: "+anzBreedTreffer);
		return tmpGridDispBreed;	//returns a temporary Array containing the newly urbanized cells by the dispersion and breed growth rules
	}
	
	/**
	 * This is phase 2 of the growth cycle. New spreading centers can induce new growth right in the neighborhood,
	 * which is calles "breed". This method is fired by new dispersion cells (@link DispersionGrowth), as well as
	 * by new road cells ({@link RoadGrowth}).
	 * @param X
	 * @param Y
	 * @return tmpBreedArray
	 */
	public int[][] BreedGrowth(int X,int Y){
		boolean[][] tmpGridBreed = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		int[][] tmpBreedArray = new int[2][2];	//an array containing the x and y coordinates of two cells
		int anzBreedTreffer = 0;	//Breed
    	int breedNb = calculateNeighbors(X,Y,0f);	//calculate number of neighbors of a new spreading center cell
    	int nbBreedX  = 0;	//x-coordinate of neighbouring cells of new spreading center cell
    	int nbBreedY  = 0;	//y-coordinate of neighbouring cells of new spreading center cell
    	int RNBreed1 = 0;	//1st Random number for breed
    	int RNBreed2 = 0;	//2nd Random number for breed
    	float RNBreedSlope = 0f;	//Slope value of the randomly selected breed cell
    	if (breedNb >= 2){		//if 2 or more neighbor cells are available around a new dispersion cell
    		int anzBreedTemp = 0;	//number of temporary breed cells
    		int[] nbPossibleBreedUrban = null;	//Array of possible breed cells
    		RNBreed1 = 1 + Math.round((float)Math.random()*(breedNb-1));	//take the first neighbor cell by random (between 1 and the Sum of neighbor cells)
    		RNBreed2 = 1 + Math.round((float)Math.random()*(breedNb-1));	//take the second neighbor cell by random (between 1 and the Sum of neighbor cells)
    		do{
    			RNBreed2 = 1 + Math.round((float)Math.random()*(breedNb-1));
    		} while (RNBreed2 == RNBreed1);//if the second cell is the same as the first cell try again until the second random number is different from the first
    		
    		for (int cellBreed = 0; cellBreed < NEIGHBOR.length; cellBreed++){	//for every cell in the neighborhood...
    			nbBreedX = X+NEIGHBOR[cellBreed][0];
    			nbBreedY = Y+NEIGHBOR[cellBreed][1];
    			if ( nbBreedX < areaMinX || nbBreedY < areaMinY || nbBreedX > areaMaxX || nbBreedY > areaMaxY ||
    					Float.isNaN(outputGrid.getRasterSampleAsFloat(nbBreedX,nbBreedY)) ) 
    				continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
    			if (areaRestr.getRasterSampleAsFloat(nbBreedX,nbBreedY) != 0)
    				continue;	//if a cell is inside an excluded area, break and continue loop from the beginning
    			RNBreedSlope = (float)Math.random();	//a random slope value between 0.0 and 1.0
    			double slopeDoubleNb = getSlopeValue(nbBreedX,nbBreedY,slopeLUT);	//returns the slope value for the neighboring cells which are available for breed
    			float breedSample = outputGrid.getRasterSampleAsFloat(nbBreedX,nbBreedY);	//takes the value from the input grid
    			if (breedSample == 0f){		//if it is free for urbanization...
    				anzBreedTemp++;		//increase the number of temporary breed cells
    				nbPossibleBreedUrban = new int[] {nbBreedX,nbBreedY,anzBreedTemp};	//create a new array containing the x and y coordinates and the number of the breed pixel
    				if (nbPossibleBreedUrban[2] == RNBreed1 || nbPossibleBreedUrban[2] == RNBreed2 && slopeDoubleNb < RNBreedSlope){	//if the cell has the same value as the first or second random value and its slope value is smaller than the random slope value...
    					tmpBreedArray[anzBreedTreffer][0] = nbBreedX;	//x-coordinate
    					tmpBreedArray[anzBreedTreffer][1] = nbBreedY;	//y-coordinate
    					tmpGridBreed[nbBreedX][nbBreedY] = true;	//urbanize the cell in the temporary grid
    					anzBreedTreffer++;		//and increase the number of breed cells
    				}
    			}
    		}
    	}
		return tmpBreedArray; 
	}
	
	
	/**
	 * This is phase 3 of the growth cycle, where edge growth appears based on the urban pixels that exist at 
	 * this time (including the new spontaneous and spreading center pixels).
	 * @return tmpGridSpread
	 */
	public boolean[][] EdgeGrowth(double spread){
		
		boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		//definitions for neighborhood cells
		int nbX = 0; // X-Coordinate of a neighboring cell
	    int nbY = 0; // Y-Coordinate of a neighboring cell
	    
		int anzSpreadTreffer = 0;	//Number of Spread cells
		
		//iterate over all cells
		for (int x = 0; x < inputGrid.getWidth(); x++)
			for (int y = 0; y < inputGrid.getHeight(); y++) {		//for every cell in x/y direction...
				
				int areaMinXadd = areaMinX + x;		//x-coordinate cell in the raster
				int areaMinYadd = areaMinY + y;		//y-coordinate cell in the raster
				
				int anzNachbarn = calculateNeighbors(x,y,1f);	//number of urban neighbors
				
				boolean lebt = outputGrid.getRasterSampleAsFloat(areaMinXadd, areaMinYadd ) > 0f;	//a cell is alive if it is 1
				
				int SpR = Math.round((float) Math.random()*100);				//a random number to compare with the spread coefficient
				if (lebt && (anzNachbarn>=2) && (SpR < spread) && (anzNachbarn < 8)) {	//if a cell is alive, it has more or equal than two urban neighbors and less than 8 urban neighbors (at least 1 non-urban neighbor) and we pass the spread coefficient test...
					int anzTemp = 0;		//temporary possible cells for urbanization
					int[] nbPossibleUrban = null;	//array with coordinates of possible cells for urbanization
					//int RN = 1 + Math.round((float)Math.random()*(8-anzNachbarn-1));	//this is an option to distribute the ranom number only between the possible cells for urbanization
					int RN = (int) Math.round(8*Math.random());	//a random number between 0 and 8
					float RNEdgeSlope = 0;
					for (int nbCellChange = 0; nbCellChange < NEIGHBOR.length; nbCellChange++){	//loop over all neighbor cells
						checkBreakingCommands();
						nbX = x+NEIGHBOR[nbCellChange][0];
						nbY = y+NEIGHBOR[nbCellChange][1];
						if ( nbX < areaMinX || nbY < areaMinY || nbX > areaMaxX || nbY > areaMaxY ||
					               Float.isNaN(outputGrid.getRasterSampleAsFloat(nbX,nbY)))
					             continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
						if (areaRestr.getRasterSampleAsFloat(nbX,nbY) != 0)
				        	continue;	//if a cell is inside an excluded area, break and continue loop from the beginning
						RNEdgeSlope = (float)Math.random();		//a random number to be compared with the slope value
						double slopeDoubleNb = getSlopeValue(nbX,nbY,slopeLUT);	//returns the slope value for the neighboring cells which are available for breed
	        			float rasterSample = outputGrid.getRasterSampleAsFloat(nbX,nbY);	//takes the value from the input gridn
						if (rasterSample == 0f){	//if the value is 0 and thus is possible for urbanization
							anzTemp++;				//count it
							nbPossibleUrban = new int[] {nbX,nbY,anzTemp};	//write it with its coordinates in an array
							//float NbValueEdge = 1f;
							if (nbPossibleUrban[2] == RN){	//if the number in the "possible for urbanization" array is the same as the random number...
								//int urbanNbTemp = calculateNeighbors(nbX,nbY,NbValueEdge);	//to reduce the effect of edge growth here could be included an option that allows only growth when a new urban cell has at least 2 urban neighbors
								if (tmpGridSpread[nbX][nbY]==false && slopeDoubleNb < RNEdgeSlope){	// if the cell is non-urban and it passes the slope test //urbanNbTemp >= 2 && 
									tmpGridSpread[nbX][nbY] = true;	//write it in the temporary raster
									anzSpreadTreffer++;	//count the hits
								}
							} 
						}
					}
				}
			}
		return tmpGridSpread;	//returns the temporary array
	}
	
	/**
	 * This is phase 4 of the growth cycle, where growth along roads appears. Every newly developed cell can induce road growth.
	 * Therefore for a defined number of new cells (by the breed-coeffient) the algorithm looks for a road in the neighborhood.
	 * If the new cell is located directly on a road, the algorithm starts a "road walk" ({@link roadWalk}), and moves a new cell along the road
	 * to a defined distance. If the new cell is not on a road, it looks for a road in a defined neighborhood (by {@code rg_value}). 
	 * If a road is found, it starts a "road walk" from there.
	 * @param tmpGrid
	 * @return tmpGridRoad
	 */
	public boolean[][] RoadGrowth(boolean[][] tmpGrid,double rg_value, double breed, double disp){
		boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		int anzGesamt = countTempUrbanCells(tmpGrid);	//number of urban cells in the temporary cell array
		int anzRoadTreffer = 0;
		int[] cellValue = null;	//array of cell values
		int RSearchRoad = 0;	//Random Road Search Index
		int r = 1;	//number between 1 and the disp-Value, whih is increased during the follwing while-loop. It determines how many pixels are selected randomly for a possible road trip
		int anzTempRoadNb = 0;
		
		do{
			RSearchRoad = Math.round((float)Math.random()*anzGesamt);	//a random value between 1 and the total amount of newly urbanized pixels
			int anzTruePixels = 0;	
			int roadNeighborhoodRange = 0;
			int RSearchNbRoad = 0;
			for (int x = 0; x < inputGrid.getWidth();x++){		//for every cell in the raster...
				for(int y = 0; y < inputGrid.getHeight();y++){
					int X = inputGrid.getMinX() + x;	//x-coordinate cell in the raster
					int Y = inputGrid.getMinY() + y;	//y-coordinate cell in the raster
					if (tmpGrid[X][Y]==true){	//if a new urbanized cell is found
						anzTruePixels++;		//count the number of urban pixels, and...(and this is different to "countTempUrbanCells")
						cellValue = new int[] {X,Y,anzTruePixels};	//write them with their coordinates in an array
						int tempRoadCell[] = null;		//an array of temporary road cells
						int tempNbRoadCell[] = null;	//an array of temporary neighboring cells of these road cells
						if (cellValue[2] == RSearchRoad){	//randomly select a growth pixel to start search for road (by comparing the cell number with the generated random number between 1 and total amount of new cells)
						//	statusOut.println("Randomly selected road pixel found...");
							r++;	//increase r, because a potential road trip pixel has been selected
							float roadValue = roadGrid.getRasterSampleAsFloat(X,Y);	//take the value from the road map for that pixel
							if (roadValue > 0.0){		//if the value is greater than 0 and thus is a road...
								tempRoadCell = new int[] {X,Y};		//add its coordinates to the temporary road grid
							//	statusOut.println("Strasse in Entfernung 0: "+tempRoadCell[0]+","+tempRoadCell[1]);
								int nbCellNew[] = roadWalk(tempRoadCell[0],tempRoadCell[1],disp);	//perform a road walk
								if(nbCellNew != null){	//if the result of the road walk is positive,...
									tmpGridRoad[nbCellNew[0]][nbCellNew[1]] = true;		//place a new urban pixel in the temporary Grid
									anzRoadTreffer++;	//?!?
									int[][] tmpBreedArray = new int[2][2];	//an array containing the x and y coordinates of two breed cells
									tmpBreedArray = BreedGrowth(nbCellNew[0],nbCellNew[1]);	//this new urban cell can act as a new spreading center:
									int x1 = tmpBreedArray[0][0];	//x-coordinate of first cell
					        		int y1 = tmpBreedArray[0][1];	//y-coordinate of first cell
					        		int x2 = tmpBreedArray[1][0];	//x-coordinate of second cell
					        		int y2 = tmpBreedArray[1][1];	//y-coordinate of second cell
					        		tmpGridRoad[x1][y1] = true;	//make the temporary grid true at that position
					        		tmpGridRoad[x2][y2] = true;
									anzRoadTreffer += 2;
								} else{
							//		statusOut.println("No new RoadBreedCells found directly");
								}
							} else{		//if the selected pixel has not been a road pixel search in the neighborhood for a road
								int anzRoadNb = 0;
						//		statusOut.println("no road cell on first run...");
							
								//With a loop look in the neighborhood for a road in an increasing radius until you reach the road gravity value
								for (int roadSearch = 0; roadSearch <= rg_value; roadSearch++){
								//	statusOut.println("RG_VALUE: "+rg_value+"; ROAD-SEARCH-VALUE = "+roadSearch);
									if (anzRoadNb > 0)	//if a road is found break and continue from the beginning (the next loops will end here as well until it stops when rg_value ist reached)
										continue;
									for (int nbRoadCellsX = X - roadNeighborhoodRange; nbRoadCellsX <= X + roadNeighborhoodRange; nbRoadCellsX++){
										for (int nbRoadCellsY = Y - roadNeighborhoodRange; nbRoadCellsY <= Y + roadNeighborhoodRange; nbRoadCellsY++){	//increase the neighborhood in every loop
											if ( nbRoadCellsX < areaMinX || nbRoadCellsY < areaMinY || nbRoadCellsX > areaMaxX || nbRoadCellsY > areaMaxY ||
													Float.isNaN(outputGrid.getRasterSampleAsFloat(nbRoadCellsX,nbRoadCellsY)) )
												continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
											float roadValueNb = roadGrid.getRasterSampleAsFloat(nbRoadCellsX,nbRoadCellsY);		//take the value from the road map for that pixel
											if (roadValueNb > 0f){	//if the value is greater than 0 and thus is a road...
												anzRoadNb++;	//increase the number of road neighbors
									//			statusOut.println("Number of road neighbors: "+anzRoadNb);
											}
										}
									}
									roadNeighborhoodRange++;	//increase the road neighborhood range
									RSearchNbRoad = Math.round((float)Math.random()*anzRoadNb);		//a random value between 0 and the number of road neighbors
									anzTempRoadNb = 0;	//number of temporary road neighbors
									for (int nbRoadCellsX = X - roadNeighborhoodRange; nbRoadCellsX <= X + roadNeighborhoodRange; nbRoadCellsX++){
										for (int nbRoadCellsY = Y - roadNeighborhoodRange; nbRoadCellsY <= Y + roadNeighborhoodRange; nbRoadCellsY++){	//again increase the neighborhood in every loop
						        			//statusOut.println("TEST-ROAD: "+nbRoadCellsX+","+nbRoadCellsY);
											if ( nbRoadCellsX < areaMinX || nbRoadCellsY < areaMinY || nbRoadCellsX > areaMaxX || nbRoadCellsY > areaMaxY ||
													Float.isNaN(outputGrid.getRasterSampleAsFloat(nbRoadCellsX,nbRoadCellsY)) ) //wenn Nachbarzelle auerhalb des Rasters (NaN) --> ignorieren
												continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
											float roadValueNbTemp = roadGrid.getRasterSampleAsFloat(nbRoadCellsX,nbRoadCellsY);	//take the road value from that pixel
											if (roadValueNbTemp > 0f){	//if the value is greater than 0 and thus is a road... 
												anzTempRoadNb++;	//increase the number of temporary road neighbors
												tempNbRoadCell = new int[] {nbRoadCellsX,nbRoadCellsY,anzTempRoadNb};	//build an array of neighboring road cells with its coordinates
											//	statusOut.println("in Array: "+tempNbRoadCell[2]+"Random: "+RSearchNbRoad+"Anzahl temporrer Zellen: "+anzTempRoadNb);
												if (tempNbRoadCell[2]==RSearchNbRoad){	//check the position in that array with the random number
													tempRoadCell = tempNbRoadCell;	//the main temporary road cell is moved to this position
													//statusOut.println("Strasse in Entfernung "+roadSearch+": "+tempRoadCell[0]+","+tempRoadCell[1]+","+tempRoadCell[2]);
													int nbCellNew[] = roadWalk(tempRoadCell[0],tempRoadCell[1],disp);	//perform a road walk
													if(nbCellNew != null){	//if the result of the road walk is positive,...
														tmpGridRoad[nbCellNew[0]][nbCellNew[1]] = true;	//place a new urban cell in the temporary grid
														anzRoadTreffer++;
														int[][] tmpBreedArray = new int[2][2];
														tmpBreedArray = BreedGrowth(nbCellNew[0],nbCellNew[1]);
														int x1 = tmpBreedArray[0][0];
										        		int y1 = tmpBreedArray[0][1];
										        		int x2 = tmpBreedArray[1][0];
										        		int y2 = tmpBreedArray[1][1];
										        		tmpGridRoad[x1][y1] = true;
										        		tmpGridRoad[x2][y2] = true;
														anzRoadTreffer += 2;
													} else{
												//		statusOut.println("No new road cell found after road search.");
													}
												}
											} else{
											//	statusOut.println("Nothing to find...");
											}
										}
									}
								}
							}
						}
					}	
				}
			}
			
		} while (r <= breed);
		statusOut.println("Road-Treffer: "+anzRoadTreffer);
		return tmpGridRoad;
	}
	
	/**
	 * This method performs a walk along a road starting from an initial cell with the coordinates X, Y.
	 * It returns the position of the final cell after the road walk, which than can start a new urbanization.
	 * @param X
	 * @param Y
	 * @return nbCellNew
	 */
	public int[] roadWalk(int X,int Y,double disp){
		boolean end_of_road = false;	//start with the assumption that the end of the road trip is not completed yet	
		int run = 0;	//distance a cell is moving
		int run_value = 0;	//maximum moving distance
		int[] nbCellNew = null;	 //Array containing the coordinates of the new urbanized cells after the road walk
		int[] nbCellTemp = null;	//Array containing the neighbors of the end cell
		while(!end_of_road){	//until the end of the road is not reached...
			end_of_road = true;		//say that this is so
			int nbRoadX = 0;	//neighborhood of the cell
			int nbRoadY = 0;
			int areaMinX = inputGrid.getMinX();								//raster dimensions
			int areaMinY = inputGrid.getMinY();
			int areaMaxX = inputGrid.getMinX() + inputGrid.getWidth() - 1;
			int areaMaxY = inputGrid.getMinY() + inputGrid.getHeight() - 1;
			int roadNb = 0;		//number of neighboring cells
			int roadNbTemp = 0;	//number of neighboring cells
			int RN = 0;	//a random number
						
			int[] nbCell = null;	//Array containing a neighboring cell
			for (int nbRoadCells = 0; nbRoadCells < NEIGHBOR.length; nbRoadCells++){	//for every neighbor of the starting cell...
				nbRoadX = X+NEIGHBOR[nbRoadCells][0];	
				nbRoadY = Y+NEIGHBOR[nbRoadCells][1];
				if ( nbRoadX < areaMinX || nbRoadY < areaMinY || nbRoadX > areaMaxX || nbRoadY > areaMaxY ||
			               Float.isNaN(outputGrid.getRasterSampleAsFloat(nbRoadX,nbRoadY)) ) 
			             continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
				float roadSample = roadGrid.getRasterSampleAsFloat(nbRoadX,nbRoadY);	//take the value
				if (roadSample > 0f)	//if a road is present (>1),
					roadNb++;			//count it
			}
			
			if (roadNb > 0){	//if neighbors are available
				RN = 1 + Math.round((float)Math.random()*(roadNb-1));	//select a random one
			
				for (int nbRoadCells = 0; nbRoadCells < NEIGHBOR.length; nbRoadCells++){	//for every neighbor of the starting cell...
					nbRoadX = X+NEIGHBOR[nbRoadCells][0];
					nbRoadY = Y+NEIGHBOR[nbRoadCells][1];
					if ( nbRoadX < areaMinX || nbRoadY < areaMinY || nbRoadX > areaMaxX || nbRoadY > areaMaxY ||
							Float.isNaN(outputGrid.getRasterSampleAsFloat(nbRoadX,nbRoadY)) ) 
			             	continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
					float roadSample = roadGrid.getRasterSampleAsFloat(nbRoadX,nbRoadY);	//take the value
					if (roadSample > 0f){	//if a road is present (>1),
						roadNbTemp++;		//count it
						nbCell = new int[] {nbRoadX,nbRoadY,roadNbTemp};	//write it in an array
						if (nbCell[2] == RN){	//if the number in the array is the same as the randomly selected number
							end_of_road = false;	//the end of the road is not reached
							run++;	//increase the run value
							X = nbRoadX;
							Y = nbRoadY;
							float roadSample2 = roadGrid.getRasterSampleAsFloat(X,Y);	//take the value of the road
							run_value = (int) (roadSample2 / 100 * disp);	//calculate how far the cell can be moved along the road
							if (run > run_value){	//if the run-value exceeds the maximum value...
								end_of_road = true;	//end the road walk
								int Nb = calculateNeighbors(X,Y,0f);	//find the neighbors of the cell found
								RN = 1 + Math.round((float)Math.random()*(Nb-1));	//select a random number between 1 and the number of neighbors
								if (Nb > 0){	//if a neighbor is found...
									int nbUrbanizeX = 0;
									int nbUrbanizeY = 0;
									int NbTemp = 0;
									float RNRoadSlope = 0f;
									for (int urbanizeCells = 0; urbanizeCells < NEIGHBOR.length; urbanizeCells++){	//look at every neighbor cell
										nbUrbanizeX = X+NEIGHBOR[urbanizeCells][0];
										nbUrbanizeY = Y+NEIGHBOR[urbanizeCells][1];
										if ( nbUrbanizeX < areaMinX || nbUrbanizeY < areaMinY || nbUrbanizeX > areaMaxX || nbUrbanizeY > areaMaxY ||
												Float.isNaN(outputGrid.getRasterSampleAsFloat(nbUrbanizeX,nbUrbanizeY)) ) 
											continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
										if (areaRestr.getRasterSampleAsFloat(nbUrbanizeX,nbUrbanizeY) != 0)
											continue;	//if a cell is inside an excluded area, break and continue loop from the beginning
										float rasterSample = outputGrid.getRasterSampleAsFloat(nbUrbanizeX,nbUrbanizeY);	//take the raster value
										RNRoadSlope = (float)Math.random();	//generate a random number in order to compare it with the slope
										double slopeDoubleNb = getSlopeValue(nbUrbanizeX,nbUrbanizeY,slopeLUT);	//returns the slope value for the neighboring cells which are available for breed
										if (rasterSample == 0f){	//if the raster sample is available for urbanization...
											NbTemp++;	//count it
											nbCellTemp = new int[] {nbUrbanizeX,nbUrbanizeY,NbTemp};	//create an array holding the coordinates and the number of that neighbor cell
											if (nbCellTemp[2] == RN && slopeDoubleNb < RNRoadSlope){	//if its number is the same as the random number and it passes the slope-test..
												nbCellNew = new int[] {nbCellTemp[0],nbCellTemp[1],nbCellTemp[2]};	//write its coordinates in the array to be returned
											}
										}	
									}
								
							} else{
								nbCellNew = null;	
							}	break;
						}
						
					}
					
				}
			}
			
			}
		}
		return nbCellNew;	//return the result array
		}
	
	/**
	 * This method returns the number of neighbors of a cell X/Y. The input for this method are the coordinates of the center cell
	 * and the value that the neighboring cells must have in order to be counted. In general this is 0 for non-urban cells and 1 
	 * for urban cells.
	 * @param X
	 * @param Y
	 * @param NbValue
	 * @return Nb
	 */
	public int calculateNeighbors(int X, int Y, float NbValue){
		int nbX = 0;
		int nbY = 0;
		int areaMinX = inputGrid.getMinX();								//raster dimensions
		int areaMinY = inputGrid.getMinY();
		int areaMaxX = inputGrid.getMinX() + inputGrid.getWidth() - 1;
		int areaMaxY = inputGrid.getMinY() + inputGrid.getHeight() - 1;
		int Nb = 0;
		for (int nbCells = 0; nbCells < NEIGHBOR.length; nbCells++){
			nbX = X+NEIGHBOR[nbCells][0];
			nbY = Y+NEIGHBOR[nbCells][1];
			if ( nbX < areaMinX || nbY < areaMinY || nbX > areaMaxX || nbY > areaMaxY ||
		               Float.isNaN(outputGrid.getRasterSampleAsFloat(nbX,nbY)))
		             continue;	//if a cell is outside the raster (NaN), break and continue loop from the beginning
			if (areaRestr.getRasterSampleAsFloat(nbX,nbY) != 0)
	        	continue;	//if a cell is inside an excluded area, break and continue loop from the beginning
			float rasterSample = outputGrid.getRasterSampleAsFloat(nbX,nbY);	//take the raster value
			if (rasterSample == NbValue)	//if it is the same as the value we look for...
				Nb++;	//count it
		}
		return Nb;
	}
	
	/**
	 * Returns the number of urban cells in the temporary cell array.
	 * @return anzGesamt
	 */
	public int countTempUrbanCells(boolean[][] tmpGrid){
		int anzGesamt = 0;
		for (int x = 0; x < outputGrid.getWidth();x++)
			for(int y = 0; y < outputGrid.getHeight();y++){
				if (tmpGrid[x][y]==true){
					anzGesamt++;
				}		
			}
		return anzGesamt;
	}
	
	/**
	 * Writes a temporary grid array with boolean values to the outputGrid.
	 * @param tmpGrid
	 */
	public void writeToGrid(boolean tmpGrid[][]){
		int t = 0;
		for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
			for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
			
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
			
				boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive when it is 1
				if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
					tmpGrid[x][y] = lebt;	//take the information from the current output grid
				if(tmpGrid[x][y] == true)
					t++;
				outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the result to the output grid
			}
		}
		statusOut.println("Temporary Urban Cells: "+t);
	}
}
