package edu.bonn.xulu.plugin.model.sleuth;

import edu.bonn.xulu.model.XuluModel;
import edu.bonn.xulu.plugin.data.grid.MultiGrid;
import schmitzm.data.WritableGrid;
import schmitzm.data.property.ListProperty;
import schmitzm.data.property.ListPropertyWriteAccess;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.PropertyWriteAccess;
import schmitzm.data.property.ScalarProperty;

/**
 * <b>Calibration of the Urban Growth Model</b><br>
 *  This model is based on the {@link UrbanGrowthModel Urban Growth Model}, developed by Keith Clarke at the UC Santa Barbara, CA, within the Gigalopolis Project, funded by the
 *  USGS. It has exactly the same growth rules as its parent class, but runs with n Monte-Carlo Iterations. So the resulting map does not contain the urban extent after a model
 *  run, but the probability for new urbanization. The more often a cell has been selected for urbanization during the Monte-Carlo Iterations, the higher is that value in the 
 *  output map. The model performs this procedure for every combination of the different coefficients. There is not a single coefficient value as in the original model, but a starting
 *  value, a step value and an end value. So, when starting with a coefficient of 5, a step value of 10 and an end value of 25, the model will perform 3 steps (for 5, 15, and 25). Be 
 *  aware that entering many steps for many coefficients will result in a huge amount of possible combinations. Therefore the calibration should be performed sequential from a coarse
 *  calibration over a fine calibration to a final calibration. The perforamce of the different model runs is measured by the {@link edu.bonn.xulu.plugin.model.mrv.MultipleResolutionValidation Multiple Resolution Comparison}.
 *  This validation procedure has been adopted for this class and is integrated in a separate method. The original SLEUTH Urban Growth Model is calibrated by a different procedure.
 *  For a detailed description refer to {@link UrbanGrowthModel Urban Growth Model}, {@link UrbanGrowthModel_MC Urban Growth Model MC} or 
 *  {@link edu.bonn.xulu.plugin.model.mrv.MultipleResolutionValidation Multiple Resolution Comparison}
 *  
 *  <br>
 *  In addition to the UGM parameters, these parameters are needed:<br>
 * <ul>
 *  <li><b>Spread Start</b>: Spread Starting Coefficient</li>
 *  <li><b>Spread Step</b>: Spread Step Coefficient</li>
 *  <li><b>Spread End</b>: Spread End Coefficient</li>
 *  <li><b>Dispersion Start</b>: Dispersion Starting Coefficient</li>
 *  <li><b>Dispersion Step</b>: Dispersion Step Coefficient</li>
 *  <li><b>Dispersion End</b>: Dispersion End Coefficient</li>
 *  <li><b>Breed Start</b>: Breed Starting Coefficient</li>
 *  <li><b>Breed Step</b>: Breed Step Coefficient</li>
 *  <li><b>Breed End</b>: Breed End Coefficient</li>
 *  <li><b>RoadGravity Start</b>: RoadGravity Starting Coefficient</li>
 *  <li><b>RoadGravity Step</b>: RoadGravity Step Coefficient</li>
 *  <li><b>RoadGravity End</b>: RoadGravity End Coefficient</li>
 *  <li><b>Slope Start</b>: Slope Starting Coefficient</li>
 *  <li><b>Slope Step</b>: Slope Step Coefficient</li>
 *  <li><b>Slope End</b>: Slope End Coefficient</li>
 *  <li><b>Monte Carlo Iterations</b>: The number of Monte Carlo Iterations.</li>
 *  <li><b>Sum Step Results</b>: List of resulting maps containing the sums over all Monte Carlo Iterations for a single parameter combination.</li>
 * </ul>
 * 
 * @see UrbanGrowthModelCalibrationContentManager
 * @see UrbanGrowthModel
 * @see UrbanGrowthModel_MC
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 */
public class UrbanGrowthModelCalibration extends UrbanGrowthModel {

	// ********** Read/Write access, that has to be maintained ***
	// ********** through the whole model run ********************
	protected PropertyReadAccess RA_referenceMap = null;	//Observed Validation Grid
	protected PropertyReadAccess RA_spreadStart = null;	//Spread-Coefficient
	protected PropertyReadAccess RA_spreadStep = null;	//Spread-Coefficient
	protected PropertyReadAccess RA_spreadEnd = null;	//Spread-Coefficient
	protected PropertyReadAccess RA_dispStart = null;		//Dispersion-Coefficient
	protected PropertyReadAccess RA_dispStep = null;		//Dispersion-Coefficient
	protected PropertyReadAccess RA_dispEnd = null;		//Dispersion-Coefficient
	protected PropertyReadAccess RA_breedStart = null;		//Breed-Coefficient
	protected PropertyReadAccess RA_breedStep = null;		//Breed-Coefficient
	protected PropertyReadAccess RA_breedEnd = null;		//Breed-Coefficient
	protected PropertyReadAccess RA_roadGravityStart = null;	//Road-Gravity-Coefficient
	protected PropertyReadAccess RA_roadGravityStep = null;	//Road-Gravity-Coefficient
	protected PropertyReadAccess RA_roadGravityEnd = null;	//Road-Gravity-Coefficient
	protected PropertyReadAccess RA_slopeStart = null;	//Slope-Coefficient
	protected PropertyReadAccess RA_slopeStep = null;	//Slope-Coefficient
	protected PropertyReadAccess RA_slopeEnd = null;	//Slope-Coefficient
	protected PropertyReadAccess RA_criticalSlope = null;	//Critical Slope Value
	protected PropertyReadAccess RA_MonteCarlo = null;	//Number of MonteCarlo Iterations
	protected ListPropertyWriteAccess WA_outResults = null;	//Output Validation Results
	protected PropertyWriteAccess WA_SumOutStep = null;	// Step Results
	

	// **************** Variablen to work with *******************
	protected WritableGrid referenceMap = null;	//Observed Validation Grid 
	protected double spreadStart = 0;	//Spread-Coefficient
	protected double spreadStep = 0;	//Spread-Coefficient
	protected double spreadEnd = 0;	//Spread-Coefficient
	protected double dispStart = 0;	//Dispersion-Coefficient
	protected double dispStep = 0;	//Dispersion-Coefficient
	protected double dispEnd = 0;	//Dispersion-Coefficient
	protected double breedStart = 0;	//Breed-Coefficient
	protected double breedStep = 0;	//Breed-Coefficient
	protected double breedEnd = 0;	//Breed-Coefficient
	protected double roadGravityStart = 0;	//Road-Gravity-Coefficient
	protected double roadGravityStep = 0;	//Road-Gravity-Coefficient
	protected double roadGravityEnd = 0;	//Road-Gravity-Coefficient
	protected double slopeStart = 0;	//Slope-Coefficient
	protected double slopeStep = 0;	//Slope-Coefficient
	protected double slopeEnd = 0;	//Slope-Coefficient
	protected double criticalSlope = 0;	//Critical Slope Value
	protected int MonteCarlo = 0;	//Number of Monte Carlo Iterations
	protected ListPropertyWriteAccess outResults = null;	//Output Validation Results
	protected MultiGrid SumOutStep = null;	//Step Results
		
	protected int years = 0;
	protected int neighborhoodRange = 0;
	protected int numSteps = 0;
	protected float[] resultArray;
	protected float[] mrvTempResultArray;
	protected float[] finalArray;
	
	/**
	 * First Constructor
	 * @param contManager
	 */
	public UrbanGrowthModelCalibration(UrbanGrowthModelCalibrationContentManager contManager) {
		super(contManager);
		this.contManager = (UrbanGrowthModelCalibrationContentManager) super.contManager;
	}
	
	/**
	 * Second Constructor. The Content Manager is the parent content manager as well as the child content manager
	 */
	public UrbanGrowthModelCalibration(){
		super(new UrbanGrowthModelCalibrationContentManager());
	}
	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {

		if (contManager == null)
			return;
		// Get access authorisation of resource/properties
		RA_steps = null;
		if (contManager.getResource(0).getData() != null)
			RA_steps = ((ScalarProperty) contManager.getResource(0).getData())
					.getReadAccess(this);
		RA_inputGrid = null;
		if (contManager.getResource(1).getData() != null)
			RA_inputGrid = ((ScalarProperty) contManager.getResource(1)
					.getData()).getReadAccess(this);
		WA_outputGrid = null;
		if (contManager.getResource(2).getData() != null)
			WA_outputGrid = ((ScalarProperty) contManager.getResource(2)
					.getData()).getWriteAccess(this);
		RA_areaRestr = null;
		if (contManager.getResource(3).getData() != null)
			RA_areaRestr = ((ScalarProperty) contManager.getResource(3)
					.getData()).getReadAccess(this);
		RA_roadGrid = null;
		if (contManager.getResource(4).getData() != null)
			RA_roadGrid = ((ScalarProperty) contManager.getResource(4)
					.getData()).getReadAccess(this);
		RA_slopeGrid = null;
		if (contManager.getResource(5).getData() != null)
			RA_slopeGrid = ((ScalarProperty) contManager.getResource(5)
					.getData()).getReadAccess(this);
		RA_referenceMap = null;
		if(contManager.getResource(6).getData() != null)
			RA_referenceMap = ((ScalarProperty) contManager.getResource(6)
					.getData()).getReadAccess(this);
		RA_spreadStart = null;
		if (contManager.getResource(7).getData() != null)
			RA_spreadStart = ((ScalarProperty) contManager.getResource(7).getData())
					.getReadAccess(this);
		RA_spreadStep = null;
		if (contManager.getResource(8).getData() != null)
			RA_spreadStep = ((ScalarProperty) contManager.getResource(8).getData())
					.getReadAccess(this);
		RA_spreadEnd = null;
		if (contManager.getResource(9).getData() != null)
			RA_spreadEnd = ((ScalarProperty) contManager.getResource(9).getData())
					.getReadAccess(this);
		RA_dispStart = null;
		if (contManager.getResource(10).getData() != null)
			RA_dispStart = ((ScalarProperty) contManager.getResource(10).getData())
					.getReadAccess(this);
		RA_dispStep = null;
		if (contManager.getResource(11).getData() != null)
			RA_dispStep = ((ScalarProperty) contManager.getResource(11).getData())
					.getReadAccess(this);
		RA_dispEnd = null;
		if (contManager.getResource(12).getData() != null)
			RA_dispEnd = ((ScalarProperty) contManager.getResource(12).getData())
					.getReadAccess(this);
		RA_breedStart = null;
		if (contManager.getResource(13).getData() != null)
			RA_breedStart = ((ScalarProperty) contManager.getResource(13).getData())
					.getReadAccess(this);
		RA_breedStep = null;
		if (contManager.getResource(14).getData() != null)
			RA_breedStep = ((ScalarProperty) contManager.getResource(14).getData())
					.getReadAccess(this);
		RA_breedEnd = null;
		if (contManager.getResource(15).getData() != null)
			RA_breedEnd = ((ScalarProperty) contManager.getResource(15).getData())
					.getReadAccess(this);
		RA_roadGravityStart = null;
		if (contManager.getResource(16).getData() != null)
			RA_roadGravityStart = ((ScalarProperty) contManager.getResource(16).getData())
					.getReadAccess(this);
		RA_roadGravityStep = null;
		if (contManager.getResource(17).getData() != null)
			RA_roadGravityStep = ((ScalarProperty) contManager.getResource(17).getData())
					.getReadAccess(this);
		RA_roadGravityEnd = null;
		if (contManager.getResource(18).getData() != null)
			RA_roadGravityEnd = ((ScalarProperty) contManager.getResource(18).getData())
					.getReadAccess(this);
		RA_slopeStart = null;
		if (contManager.getResource(19).getData() != null)
			RA_slopeStart = ((ScalarProperty) contManager.getResource(19).getData())
					.getReadAccess(this);
		RA_slopeStep = null;
		if (contManager.getResource(20).getData() != null)
			RA_slopeStep = ((ScalarProperty) contManager.getResource(20).getData())
					.getReadAccess(this);
		RA_slopeEnd = null;
		if (contManager.getResource(21).getData() != null)
			RA_slopeEnd = ((ScalarProperty) contManager.getResource(21).getData())
					.getReadAccess(this);
		RA_criticalSlope = null;
		if (contManager.getResource(22).getData() != null)
			RA_criticalSlope = ((ScalarProperty) contManager.getResource(22).getData())
					.getReadAccess(this);
		RA_MonteCarlo = null;
		if (contManager.getResource(23).getData() != null)
			RA_MonteCarlo = ((ScalarProperty) contManager.getResource(23).getData())
					.getReadAccess(this);
		WA_outResults = null;
		if (contManager.getResource(24).getData() != null)
			WA_outResults = ((ListProperty) contManager.getResource(24)
					.getData()).getWriteAccess(this);
		
		
		// Assign variables to work with
		inputGrid = (WritableGrid) RA_inputGrid.getValue();
		outputGrid = (WritableGrid) WA_outputGrid.getValue();
		steps = RA_steps;
		areaRestr = (WritableGrid) RA_areaRestr.getValue();
		roadGrid = (WritableGrid) RA_roadGrid.getValue();
	    slopeGrid = (WritableGrid) RA_slopeGrid.getValue();
		referenceMap 	= (WritableGrid) RA_referenceMap.getValue();	
	    spreadStart  	= RA_spreadStart.getValueAsDouble();
	    spreadStep  	= RA_spreadStep.getValueAsDouble();
	    spreadEnd  		= RA_spreadEnd.getValueAsDouble();
	    dispStart		= RA_dispStart.getValueAsDouble();
	    dispStep		= RA_dispStep.getValueAsDouble();
	    dispEnd			= RA_dispEnd.getValueAsDouble();
	    breedStart		= RA_breedStart.getValueAsDouble();
	    breedStep		= RA_breedStep.getValueAsDouble();
	    breedEnd		= RA_breedEnd.getValueAsDouble();
	    roadGravityStart= RA_roadGravityStart.getValueAsDouble();
	    roadGravityStep = RA_roadGravityStep.getValueAsDouble();
	    roadGravityEnd 	= RA_roadGravityEnd.getValueAsDouble();
	    slopeStart		= RA_slopeStart.getValueAsDouble();
	    slopeStep		= RA_slopeStep.getValueAsDouble();
	    slopeEnd		= RA_slopeEnd.getValueAsDouble();
	    criticalSlope = RA_criticalSlope.getValueAsDouble();
	    MonteCarlo 		= RA_MonteCarlo.getValueAsInt();
		outResults 		= WA_outResults;
		outStep 		= (MultiGrid)contManager.getResource(25).getData();
		SumOutStep		= (MultiGrid)contManager.getResource(26).getData();
		stepCount 		= 1;
		years 			= steps.getValueAsInt();

		//count the number of starting individuals
	    int startAnz = 0;
		startAnz = countSetIndividuals();
		
		statusOut.println("Number of starting cells: "+startAnz);
		statusOut.println("Neighborhood size: "+NEIGHBOR.length);
		
		neighborhoodRange = (int) Math.sqrt(inputGrid.getWidth()*inputGrid.getHeight()); //maximum possible pixel size
		numSteps = (int) ((int)1+ Math.floor(Math.log(neighborhoodRange)/Math.log(2)));	//calculates how many steps are needed by doubling the pixel size until neighborhoodRange is reached. It is the reverse of an exponential function, which is the logarithm of neighborhoodRange to the basis 2. 2 is added because the calculation starts with 1 and not with 2 (1,2,4,8,16,32,...) and an array starts with 0.
		statusOut.println("Maximum pixel size (as multiple of original resolution): "+neighborhoodRange);
		statusOut.println("Number of steps: "+numSteps);
		
	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {
		super.performModelDispose();
		
		releaseAccess(RA_referenceMap);
		releaseAccess(RA_spreadStart);
		releaseAccess(RA_spreadStep);
		releaseAccess(RA_spreadEnd);
		releaseAccess(RA_dispStart);
		releaseAccess(RA_dispStep);
		releaseAccess(RA_dispEnd);
		releaseAccess(RA_breedStart);
		releaseAccess(RA_breedStep);
		releaseAccess(RA_breedEnd);
		releaseAccess(RA_roadGravityStart);
		releaseAccess(RA_roadGravityStep);
		releaseAccess(RA_roadGravityEnd);
		releaseAccess(RA_slopeStart);
		releaseAccess(RA_slopeStep);
		releaseAccess(RA_slopeEnd);
		releaseAccess(RA_MonteCarlo);
		releaseAccess(WA_outResults);
		releaseAccess(WA_SumOutStep);
	}

	/**
	 * The main method of the Urban Growth Model. For a detailed description refer to the performModelStep in 
	 * {@link UrbanGrowthModel UrbanGrowthModel}. The difference is, that the number of steps is not based on
	 * the years that have to be modelled. The number of steps is always 1, all parameter combinations, Monte Carlo
	 * Iterations, and time steps are performed within this method. Those combinations are included in different
	 * for-loops. 
	 */
	public void performModelStep(int stepNo) {
		
		//Definition of variables needed in the whole model run
		long localStartTime = System.currentTimeMillis(); //time measure
		Float actUrb = null;	//urban cell in the raster
		int totalSteps = 0;
		int runs = 0;		
	
		//raster dimensions
		areaMinX = inputGrid.getMinX();
		areaMinY = inputGrid.getMinY();
		areaMaxX = inputGrid.getMinX() + inputGrid.getWidth() - 1;
		areaMaxY = inputGrid.getMinY() + inputGrid.getHeight() - 1;
		areaWidth = inputGrid.getWidth();
		areaHeight = inputGrid.getHeight();
			
		// Iterate the calibration for:
		// 1. every slope value between start and end...
		// 2. every spread value between start und end...
		// 3. every dispersion value between start und end...
		// 4. every breed value between start and end...
		
		for(int slope = (int)slopeStart;slope <= (int)slopeEnd; slope = slope+(int)slopeStep){
			
			//calculate the maximum slope
			int slopeMax = 0;
			slopeMax = getSlopeMax();
			
			statusOut.println("Max. Slope: "+slopeMax+" %");
					
			//assigns for every slope value between 0 and the maximum slope that is below the critical slope the value calculated for the variable "lookup".
			//all values higher than the critical slope become 1

			int MaxSlopeResistance = 100;	//Max slope resistance is 100%
			double exp = getSlopeExp(slope,MaxSlopeResistance);	//calculate the exp value based on the actual slope
			slopeLUT = getSlopeLookup(slopeMax,exp,criticalSlope);	//generate the lookup-table based on the actual exp-value
			
			for(int spread = (int)spreadStart; spread <= (int)spreadEnd; spread = spread+(int)spreadStep){
				for(int disp = (int)dispStart; disp <= (int)dispEnd; disp = disp+(int)dispStep){
					for(int breed = (int)breedStart; breed <= (int)breedEnd+1; breed = breed+(int)breedStep){
						for(int roadGravity = (int)roadGravityStart; roadGravity <=(int)roadGravityEnd; roadGravity = roadGravity+(int)roadGravityStep){	
									
							//Here the original model algorithm starts. By calibrating the model each parameter combination
							//is being iterated, so the algorithm is performed n-times. Because a lot of random values are
							// in the model, each step should be performed a Monte Carlo number of times. The results are 
							// averaged at the end of the algorithm. 
							//
							runs++;
							mrvTempResultArray = new float[numSteps+2];
							finalArray = new float[numSteps+2];
														
							//coefficient calculations of for growth rules
							max_RG_value = 100.0;			//Maximum possible road gravity value
							double disp_value = Math.round((disp*0.005)* Math.sqrt(Math.pow(inputGrid.getWidth(), 2)+(Math.pow(inputGrid.getHeight(), 2))));
							double rg_value = Math.round((roadGravity/max_RG_value)*((areaWidth + areaHeight)/16));
							statusOut.println("RG"+roadGravity+" / MaxRG:"+max_RG_value+" * (Width"+areaWidth+" + areaHeight"+areaHeight+") / 16");
							
							max_search_index = Math.round(4*((int)rg_value*(1+(int)rg_value)));	//area in which to look for a road cell
							statusOut.println("MaxSearchIndex: "+max_search_index);
														
							for(int MC = 1; MC <= MonteCarlo; MC++){	//perform the model algoith 1-MC times
								statusOut.println("Performing "+MonteCarlo+" Monte Carlo Iterations (years: "+stepNo);
								totalSteps++;
								/**
								 * The model run is being performed for every time step. For the final evaluation 
								 * only the image of the final step is needed.
								 */
								for(int timeSteps = 1;timeSteps <=years;timeSteps++){	
									statusOut.println("Year "+timeSteps);
									boolean[][] tmpGrid = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
									
									/*****************************************************************
									 * DISPERSION and BREED (Growth Phase 1 and 2)
									 *****************************************************************/
									boolean[][] tmpGridDisp = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
									tmpGridDisp = DispersionGrowth(disp_value,breed);	//calls the method "DispersionBreed" to perform the Dispersion and Breed Growth Phase
									int anzDispBreed = 0;	//number of cells urbanized by Dispersion and Breed
									for (int x = 0; x < inputGrid.getWidth();x++)
										for(int y = 0; y < inputGrid.getHeight();y++){
											int X = inputGrid.getMinX() + x;
											int Y = inputGrid.getMinY() + y;
											if (tmpGridDisp[X][Y]==true){
												anzDispBreed++;
												tmpGrid[X][Y] = true;	//write the temporary Dispersion and Breed Cells to the overall temporary Cell array
											}		
										}
									statusOut.println("Number of Spontaneous and New Spreading Center Cells: "+anzDispBreed);
									writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
									
									/*********************************************************************
									 * EDGE
									 *********************************************************************/
									boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
									tmpGridSpread = EdgeGrowth(spread);	//calls the method "EdgeGrowth" to perform the Spread (Edge Growth) Phase
									int anzSpread = 0;	//number of cells urbanized by Spread
									for (int x = 0; x < inputGrid.getWidth();x++)
										for(int y = 0; y < inputGrid.getHeight();y++){
											int X = inputGrid.getMinX() + x;
											int Y = inputGrid.getMinY() + y;
											if (tmpGridSpread[X][Y]==true){
												anzSpread++;
												tmpGrid[X][Y] = true;	//write the temporary Spread cells to the overall temporary Cell array
											}		
										}
									statusOut.println("Number of Edge Growth Cells: "+anzSpread);
									writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
								
									/*********************************************************************
									 * ROAD
									 *********************************************************************/
									boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
									statusOut.println("Start Road Growth...");
									tmpGridRoad = RoadGrowth(tmpGrid,rg_value,breed,disp);	//calls the method "RoadGrowth" to perform the Road weighted growth
									statusOut.println("Road things returned...");
									int anzRoad = 0;	//number of cells urbanized by Road Growth
									for (int x = 0; x < inputGrid.getWidth();x++)
										for(int y = 0; y < inputGrid.getHeight();y++){
											int X = inputGrid.getMinX() + x;
											int Y = inputGrid.getMinY() + y;
											if (tmpGridRoad[X][Y]==true){
												anzRoad++;
												tmpGrid[X][Y] = true;	//write the temporary Road Growth cells to the overall temporary Cell array
											}		
										}
									statusOut.println("Number of Road Growth Cells: "+anzRoad);
									writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid

									// Here the main modeling algorithm ends. Next the temporary raster is written in the output raster. 
									//With the outputGrid the next time step is calculated until the maximum number of time steps
									//is reached. After that the next Monte Carlo iteration is performed for the specific parameter combination. 									
									
									// Write the temporary Grid to the Output Grid
									int anzGesamt = 0;
									for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
										for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
											
											int X = inputGrid.getMinX() + x;
											int Y = inputGrid.getMinY() + y;
										
											if (tmpGrid[x][y] == true)
												anzGesamt++;
											
											boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive, when it is 1
											if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
												tmpGrid[x][y] = lebt;	//take the values from the current output raster
										
											outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the results in the output raster
											actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
										}
									}
									statusOut.println("Urban pixels at step "+timeSteps+": "+anzGesamt);
									
									if (timeSteps == years){
										outStep.addGrid();	//add a raster to the GridList
										
										// Write the temporary Grid to the Output Grid
										for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
											for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
												actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
												outStep.getGrid(MC-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
											}
										}
										WritableGrid simulationMap = outStep.getGrid(MC-1);		//write the final step result to the grid list
										float[] mrv = MultipleResolutionComparison(simulationMap);	//perform the MRV
										for(int i = 0;i < mrv.length;i++){
											statusOut.println(mrv[i]);
											mrvTempResultArray[i] += mrv[i];
										}
									}
								}	//END OF THE TIME STEPS
								
								if (MC == MonteCarlo){
									SumOutStep.addGrid();	//add a final raster to the GridList (contains the sum of all iterations)
									
									//statusOut.println("Step Grids: "+outStep.getGridCount());
									
									for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
										for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {	//for every pixel...
											float sumUrb = 0; 	//sum of pixel values
											for (int gridNo = 0; gridNo < MC; gridNo++){	//look in every grid of the GridList
												float tempUrb = (Float) outStep.getGrid(gridNo).getRasterSample(x,y);	//take the pixel value
												sumUrb  += tempUrb;		//and add its value
											}
											SumOutStep.getGrid(runs-1).setRasterSample(sumUrb,x,y);	//put the layer in the raster list containing the step results
										}
									}
									
									System.out.print("Slope: "+slope+" Spread: "+spread+" Disp: "+disp+" Breed: "+breed+" Road: "+roadGravity+" : ");
									
									for(int i = 0;i < mrvTempResultArray.length-2;i++){
										//print every element of the result array comma separated in one line in the status output
										finalArray[i] = mrvTempResultArray[i]/MonteCarlo;
										System.out.print(finalArray[i]+",");
									}
									System.out.print("//Weighted Averages (FT, WA): ");
									for(int j = 0;j < 2;j++){
										//print every element of the result array comma separated in one line in the status output
										finalArray[numSteps+j] = mrvTempResultArray[j]/MonteCarlo;
										System.out.print(finalArray[j]+",");
									}
									for (int grids = MC-1; grids >= 0; grids--){
										outStep.removeGrid(grids); 				//empty GridList
									}
									//statusOut.println("Step Grids nach leeren: "+outStep.getGridCount());
							
								} 
								countSetIndividuals();
								
							}	System.out.println("");
							
							//END OF THE MONTE CARLO ITERATIONS
						}	//END OF ROADGRAVITY COMBINATION
					}	//END OF BREED COMBINATION
				}	//END OF DISP COMBINATION
			}	//END OF SPREAD COMBINATION
		}	//END OF SLOPE COMBINATION
		System.out.println("Finished "  + " in " + ((System.currentTimeMillis() - localStartTime)) + " ms\n");
	}	//END OF MODEL STEP
	
	/**
	 * <b>MULTIPLE RESOLUTION VALIDATION</b><br>
	 * When all time steps are run, the result is being evaluated. Therefore the 
	 * Multiple Resolution Comparison is chosen (refer to ostanza(1989), Pontius (2002), 
	 * Pontius et al. (2005), Lesschen et al. (2005), Pontius et al. (2008a), Pontius et al. (2008b))
	 * For the original SLEUTH model a number of landscape measures are used for validation
	 * (refer to Clarke et al. (1997), Dietzel et al. (2007)), like e.g. the LeeSallee
	 * or OSM. Because the same method (MRV) is used within XULU to compare the results of different models,
	 * it was also chosen to evaluate the performance of the different calibration steps of the Urban Growth Model.
	 */
	public float[] MultipleResolutionComparison(WritableGrid simulationMap){
		//Variablendefinitionen
		int RoverallSum = 0; //Total number of pixels in Reference Map
		int RnumClasses = 0; //Total number of classes in Reference Map
		int SoverallSum = 0; //Total number of pixels in Simulation Map
		int SnumClasses = 0; //Total number of classes in Simulation Map
		int correct = 0;	//Number of pixels counted correct
		double fractionCorrect = 0f;	//Fraction correct
//		int numSteps = 0;
							
	//	neighborhoodRange = (int) Math.sqrt(referenceMap.getWidth()*referenceMap.getHeight()); //maximum possible pixel size
	//	numSteps = (int) ((int)1+ Math.floor(Math.log(neighborhoodRange)/Math.log(2)));	//calculates how many steps are needed by doubling the pixel size until neighborhoodRange is reached. It is the reverse of an exponential function, which is the logarithm of neighborhoodRange to the basis 2. 2 is added because the calculation starts with 1 and not with 2 (1,2,4,8,16,32,...) and an array starts with 0.
	//	statusOut.println("Maximum pixel size (as multiple of original resolution): "+neighborhoodRange);
	//	statusOut.println("Number of steps: "+numSteps);
		resultArray = new float[numSteps+2];
		
		for (int x = 0; x < referenceMap.getWidth(); x++)
			for (int y = 0; y < referenceMap.getHeight(); y++) {
				float Rtmp = referenceMap.getRasterSampleAsFloat(x, y); //get value of each raster cell in reference map
				float Stmp = simulationMap.getRasterSampleAsFloat(x, y);//get value of each raster cell in simulation map
				if (!Float.isNaN(Rtmp) && !Float.isNaN(Stmp)){	//if the value is a number, continue ...
					RoverallSum ++;		//count all cells in reference map
					SoverallSum ++;		//count all cells in simulation map
					if (Rtmp > RnumClasses)
						RnumClasses = (int) Rtmp;	//Number of classes in reference map. If a higher cell value is detected than the actual RnumClasses, the number of classes is increased to that value
					if (Stmp > SnumClasses)
						SnumClasses = (int) Stmp;	//see RnumClasses
					if (Rtmp == Stmp)
						correct++;		//if the same cell value is detected in the simulation and the reference map, increase the number of correct pixels
				}
		}
		fractionCorrect = (double) correct / (double) RoverallSum;	//divide the number of correct pixels by the number of total pixels
		
		//Counts pixels per class
		for (int i = 1; i <= RnumClasses; i++){		//Loop, until you reach the total number of classes
			int Rj = 0;		//Number of classes in Reference Map
			int Sj = 0;		//Number of classes in Simulation Map
			for (int x = 0; x < referenceMap.getWidth(); x++)
				for (int y = 0; y < referenceMap.getHeight(); y++){
					float Rtmp = referenceMap.getRasterSampleAsFloat(x, y);	//get value of each raster cell in reference map
					float Stmp = simulationMap.getRasterSampleAsFloat(x, y);//get value of each raster cell in simulation map
					if (!Float.isNaN(Rtmp))
						if ((int) Rtmp == i)
							Rj++;	//if cell belongs to category i, count it (for Reference map)
					if (!Float.isNaN(Stmp))
						if ((int) Stmp == i)
							Sj++;	//if cell belongs to category i, count it (for Simulation map)
				}
			statusOut.println("Class "+i+": Reference("+Rj+"), Simulation("+Sj+")");
		}
		
		int ngrids = 0;		//number of resolution grids
		float Vg = 0f;	//sum of the individual cell weights for the calculation of the average weighted fit after Pontius (2002)
		float VgWeights = 0f;	//sum of the cell weights
		float WeightedAgreement = 0f;	//total weighted agreement over all resolutions after Pontius (2002)
		for (int r = 1; r <= neighborhoodRange; r = r+r){	//for pixel size 1 up to the maximum pixel size, where one large cell covers the whole area
			//statusOut.println(r);
			ngrids++;		//count the number of grids
			float SumWn = 0f;	//Sum of the weighted means 
			int SumCells = 0;	//Number of the cells
			float totalAgreement = 0f;	//total agreement is the weighted mean of the fraction correct over all resolution grids
			for (int x = 0; x < referenceMap.getWidth(); x = x+r)	//for every pixel size...
				for (int y = 0; y < referenceMap.getHeight(); y = y+r){
					int nCells = 0;		//number of cells making up a coarse cell
					float SumMinRS = 0f;	//when two maps are compared, the minimum value is taken (MIN), this value is the sum of all these MINs
					float Wn = 0;	//weighted cell
					for(int x2 = x; x2 < x+r; x2++)	//for every cell within a coarse cell...
						for(int y2 = y; y2 < y+r; y2++){
							if ( x2 < referenceMap.getMinX() || y2 < referenceMap.getMinY() || x2 > referenceMap.getMinX()+referenceMap.getWidth()-1 || y2 > referenceMap.getMinY()+referenceMap.getHeight()-1 ||
									Float.isNaN(referenceMap.getRasterSampleAsFloat(x2,y2)))
								continue;
							nCells++;	//count all cells that have a value
						}
					for (int i = 0; i <= RnumClasses; i++){	//for every class i
						int Rn = 0;		//Number of cells in the reference map
						int Sn = 0;		//Number of cells in the simulation map
						float Rnj = 0f;	//Portion of cells in the reference map
						float Snj = 0f;	//Portion of cells in the simulation map
						float MinRS = 0f;	//Both maps are compared and the Minimum (MIN) for each cell taken
						for(int x3 = x; x3 < x+r; x3++)	//for every cell within a coarse cell...
							for(int y3 = y; y3 < y+r; y3++){
								if ( x3 < referenceMap.getMinX() || y3 < referenceMap.getMinY() || x3 > referenceMap.getMinX()+referenceMap.getWidth()-1 || y3 > referenceMap.getMinY()+referenceMap.getHeight()-1 ||
										Float.isNaN(referenceMap.getRasterSampleAsFloat(x3,y3)))
									continue;
								float Rtmp = referenceMap.getRasterSampleAsFloat(x3, y3);	//get value of each cell in the reference map
								float Stmp = simulationMap.getRasterSampleAsFloat(x3, y3);	//get value of each cell in the simulation map
								if ((int) Rtmp == i)
									Rn++;	//count all cells that belong to class i in reference map 
								if ((int) Stmp == i)
									Sn++;	//count all cells that belong to class i in simulation map
							}
						if(nCells > 0){	//if positive cells are available...
							Rnj = (float)Rn / (float)nCells;	//compute the portion of these cells within a coarse cell in the reference map
							Snj = (float)Sn / (float)nCells;	//compute the portion of these cells within a coarse cell in the simulation map
						}
						MinRS = Math.min(Rnj, Snj);	//take for every coarse cell the MIN between R and S of the portion of class i
						SumMinRS += MinRS;	//Sum all the MIN values
						
					}
					Wn = (float)nCells * SumMinRS;	//all coarse cells are weighted by multiplying the Sum of all MIN values with the number of cells that make up a coarse cell
										
					SumCells += (float)nCells;	//Sum all positive cells in a resolution grid
					SumWn += Wn;	//Sum all weighted portions in a resolution grid
				}
			
			//Total Agreement:
			//statusOut.println(SumWn+" / "+SumCells);
			totalAgreement = SumWn / (float)SumCells;	//Main Results
			//statusOut.println(totalAgreement);	
			outResults.addValue(totalAgreement);
			//resultArray[0][ngrids-1] = r;
			resultArray[ngrids-1] = totalAgreement;	//write the results in an array 
			//the result array looks like this:
			//1	   |2	 |4	   |8    |...
			//0.876|0.892|0.913|0.948|...
			Vg += SumWn * totalAgreement;
			VgWeights += SumWn;
		}
		
		//From here on the results of the Multiple Resolution Comparison are written on the screen
		
		//those variables are needed to compute the weighted average of all fits after Costanza (1989)
		float w = 0f;	//weight depending on resolution
		float Fw = 0f;	//weighted fit at certain resolution
		float SumW = 0f;	//Sum of all weights
		float SumFw = 0f;	//Sum of all weighted agreements
		float Ft = 0f;
		
		for(int i = 0; i <= ngrids-1; i++){
//			System.out.print(resultArray[i]+",");
			//calculations for weighted average after Costanza (1989):
			w = (float) Math.exp(-0.1*(resultArray[i] - 1));
			Fw = resultArray[i]*w;
			SumW += w;
			SumFw += Fw;
		}
		
		Ft = SumFw / SumW;	//Final calculation of the weighted average of all fits after Costanza (1989)
		WeightedAgreement = Vg / VgWeights;	//Final calculation of the weighted average of all fits after Pontius (2002)
		
//		System.out.print("Ft:"+Ft+",TA:"+WeightedAgreement);	//the result is written directly behind the single fits for every resolution
		resultArray[numSteps] = Ft;
//		System.out.println("");
//		System.out.println(resultArray[numSteps+1]);
		resultArray[numSteps+1] = WeightedAgreement;
//		System.out.println(resultArray[numSteps+2]);
		return resultArray;	
	}
}
