package edu.bonn.xulu.plugin.model.sleuth;

import edu.bonn.xulu.model.XuluModel;
import schmitzm.data.WritableGrid;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ScalarProperty;

/**
 * <b>Calibration of the Urban Growth Model with Self Modification</b><br>
 *  This model is based on the {@link UrbanGrowthModelCalibration Urban Growth Model Calibration}, which is a deviation of the {@link UrbanGrowthModel Urban Growth Model} developed by Keith Clarke at the UC Santa Barbara, CA, 
 *  within the Gigalopolis Project, funded by the USGS. In addition to the calibration routine a number of main input parameters is modified during each time step. When the growth rate exceeds or falls below certain
 *  critical values the coefficients are modified. For a detailed description refer to the <a href="http://www.ncgia.ucsb.edu/projects/gig/v2/About/gwSelfMod.htm">SLEUTH Homepage</a>.
 *  
 *  <br>
 *  In addition to the UGM Calibration Coefficients, these parameters are needed:<br>
 * <ul>
 *  <li><b>Road Gravity Sensitivity</b>: Influences the roadGravity coefficient</li>
 *  <li><b>Slope Sensitivity</b>: Influences the slope coefficient</li>
 *  <li><b>Critical Low</b>: When growth rate falls below this value, the coefficients are altered</li>
 *  <li><b>Critical High</b>: When growth rate exceeds this value, the coefficients are altered</li>
 *  <li><b>Boom</b>: Boom Phase. The Coefficients are increased by this value.</li>
 *  <li><b>Bust</b>: Bust Phase. The Coefficients are decreased by this value.</li>
 * </ul>
 * 
 * @see UrbanGrowthModelCalibrationSelfModificationContentManager
 * @see UrbanGrowthModelCalibration
 * @see UrbanGrowthModelSelfModification
 * @see UrbanGrowthModel
 * @see UrbanGrowthModel_MC
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 */
public class UrbanGrowthModelCalibrationSelfModification extends UrbanGrowthModelCalibration {

	// ********** Read/Write access, that has to be maintained ***
	// ********** through the whole model run ********************
	protected PropertyReadAccess RA_roadGravSens = null;	//RoadGravity Sensitivity
	protected PropertyReadAccess RA_slopeSens = null;	//Slope Sensitivity
	protected PropertyReadAccess RA_criticalLow = null;	//Critical Low
	protected PropertyReadAccess RA_criticalHigh = null;	//Critical High
	protected PropertyReadAccess RA_Boom = null;	//Boom
	protected PropertyReadAccess RA_Bust = null;	//Bust
	
	// **************** Variablen to work with *******************
	protected double roadGravSens = 0;	//RoadGravity Sensitivity
	protected double slopeSens = 0;	//Slope Sensitivity
	protected double criticalLow = 0;	//critical Low
	protected double criticalHigh = 0;	//critical High
	protected double boom = 0;	//Boom
	protected double bust = 0;	//Bust
	
	protected int slopeMax = 0;
	protected int totalPixels = 0;
	protected int excldCount = 0;
	
	/**
	 * First Constructor
	 * @param contManager
	 */
	public UrbanGrowthModelCalibrationSelfModification(UrbanGrowthModelCalibrationSelfModificationContentManager contManager) {
		super(contManager);
		//this.contManager = (UrbanGrowthModelCalibrationSelfModificationContentManager) super.contManager;
	}
	
	/**
	 * Second Constructor. The Content Manager is the parent content manager as well as the child content manager
	 */
	public UrbanGrowthModelCalibrationSelfModification(){
		super(new UrbanGrowthModelCalibrationSelfModificationContentManager());
	}

	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		super.performModelInit();
		this.contManager = (UrbanGrowthModelCalibrationSelfModificationContentManager) super.contManager;
	
		RA_roadGravSens = null;
		if (contManager.getResource(27).getData() != null)
			RA_roadGravSens = ((ScalarProperty) contManager.getResource(27).getData())
					.getReadAccess(this);
		RA_slopeSens = null;
		if (contManager.getResource(28).getData() != null)
			RA_slopeSens = ((ScalarProperty) contManager.getResource(28).getData())
					.getReadAccess(this);
		RA_criticalLow = null;
		if (contManager.getResource(29).getData() != null)
			RA_criticalLow = ((ScalarProperty) contManager.getResource(29).getData())
					.getReadAccess(this);
		RA_criticalHigh = null;
		if (contManager.getResource(30).getData() != null)
			RA_criticalHigh = ((ScalarProperty) contManager.getResource(30).getData())
					.getReadAccess(this);
		RA_Boom = null;
		if (contManager.getResource(31).getData() != null)
			RA_Boom = ((ScalarProperty) contManager.getResource(31).getData())
					.getReadAccess(this);
		RA_Bust = null;
		if (contManager.getResource(32).getData() != null)
			RA_Bust = ((ScalarProperty) contManager.getResource(32).getData())
					.getReadAccess(this);
		
		// Assign variables to work with
		roadGravSens = RA_roadGravSens.getValueAsDouble();
	    slopeSens = RA_slopeSens.getValueAsDouble();
	    criticalLow = RA_criticalLow.getValueAsDouble();
	    criticalHigh = RA_criticalHigh.getValueAsDouble();
	    boom = RA_Boom.getValueAsDouble();
	    bust = RA_Bust.getValueAsDouble();
	    
		//Calculate total amount of pixels and number of excluded pixels
		for (int x = 0; x < inputGrid.getWidth();x++ ) 
			for (int y = 0; y < inputGrid.getHeight(); y++ ) {
				if ( x < inputGrid.getMinX() || y < inputGrid.getMinY() || x > inputGrid.getMinX()+inputGrid.getWidth()-1 || y > inputGrid.getMinY()+inputGrid.getHeight()-1 ||
			            Float.isNaN(outputGrid.getRasterSampleAsFloat(x,y)) ) //ignore neighbor, if outside the raster
			             continue;
				totalPixels++;
				if (areaRestr.getRasterSampleAsFloat(x,y) != 0 && !Float.isNaN(areaRestr.getRasterSampleAsFloat(x,y)))
					excldCount++;
			}
	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {
		super.performModelDispose();
		//free resources
		releaseAccess(RA_roadGravSens);
		releaseAccess(RA_slopeSens);
		releaseAccess(RA_criticalLow);
		releaseAccess(RA_criticalHigh);
		releaseAccess(RA_Boom);
		releaseAccess(RA_Bust);
	}

	/**
	 * The main method of the Urban Growth Model. For a detailed description refer to the performModelStep in 
	 * {@link UrbanGrowthModel UrbanGrowthModel} and {@link UrbanGrowthModelCalibration Urban Growth Model Calibration}. 
	 * The difference to the plain UGM Calibration model is the Self Modification of several coefficients.
	 */
	public void performModelStep(int stepNo) {

		//Definition of variables needed in the whole model run
		double slopeF = 0;
		double spreadF = 0;
		double dispF = 0;
		double breedF = 0;
		double roadGravityF = 0;
		double growthRate = 0;
		int thisYear = 0;
		int roadCount = 0;
		
		//Definition of variables needed in the whole model run
		long localStartTime = System.currentTimeMillis(); //time measure
		Float actUrb = null;	//urban cell in the raster
		int totalSteps = 0;
		int runs = 0;		
	
		//raster dimensions
		areaMinX = inputGrid.getMinX();
		areaMinY = inputGrid.getMinY();
		areaMaxX = inputGrid.getMinX() + inputGrid.getWidth() - 1;
		areaMaxY = inputGrid.getMinY() + inputGrid.getHeight() - 1;
		areaWidth = inputGrid.getWidth();
		areaHeight = inputGrid.getHeight();
			
		// Iterate the calibration for:
		// 1. every slope value between start and end...
		// 2. every spread value between start und end...
		// 3. every dispersion value between start und end...
		// 4. every breed value between start and end...
		
		for(int slope = (int)slopeStart;slope <= (int)slopeEnd; slope = slope+(int)slopeStep){
			for(int spread = (int)spreadStart; spread <= (int)spreadEnd; spread = spread+(int)spreadStep){
				for(int disp = (int)dispStart; disp <= (int)dispEnd; disp = disp+(int)dispStep){
					for(int breed = (int)breedStart; breed <= (int)breedEnd+1; breed = breed+(int)breedStep){
						for(int roadGravity = (int)roadGravityStart; roadGravity <=(int)roadGravityEnd; roadGravity = roadGravity+(int)roadGravityStep){
														
							//Here the original model algorithm starts. By calibrating the model each parameter combination
							//is being iterated, so the algorithm is performed n-times. Because a lot of random values are
							// in the model, each step should be performed a Monte Carlo number of times. The results are 
							// averaged at the end of the algorithm. 
							//
							runs++;
							mrvTempResultArray = new float[numSteps+2];
							finalArray = new float[numSteps+2];

							
														
							for(int MC = 1; MC <= MonteCarlo; MC++){	//perform the model algoith 1-MC times
								statusOut.println("Performing "+MonteCarlo+" Monte Carlo Iterations (MC: "+MC+")");
								totalSteps++;
								
								//In every Iteration the coefficients are set to their defined values. Within an iteration those coefficients can be modified in every step. 
								slopeF = (double)slope;		//slopeF is slope at the beginning of the loop and is modified within the loop.
								spreadF = (double)spread;		//spreadF is spread at the beginning of the loop and is modified within the loop.
								dispF = (double) disp;		//dispF is disp at the beginning of the loop and is modified within the loop.
								breedF = (double)breed;	//breedF is breed at the beginning of the loop and is modified within the loop.
								roadGravityF = (double) roadGravity;	//roadGravityF is roadGravity at the beginning of the loop and is modified within the loop.
								
								/**
								 * The model run is being performed for every time step. For the final evaluation 
								 * only the image of the final step is needed.
								 */
									for(int timeSteps = 1;timeSteps <=years;timeSteps++){
									
										statusOut.println("Year "+timeSteps);
										boolean[][] tmpGrid = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster

										//recalculate the slope values
										slopeMax = getSlopeMax();
										int MaxSlopeResistance = 100;
										double exp = getSlopeExp(slopeF,MaxSlopeResistance);	//calculate the exp value based on the actual slope
										slopeLUT = getSlopeLookup(slopeMax,exp,criticalSlope);	//generate the lookup-table based on the actual exp-value
										//statusOut.println("SlopeMax, slope exp: "+slopeMax+", "+exp);
										
										double disp_value = Math.round((dispF*0.005)* Math.sqrt(Math.pow(inputGrid.getWidth(), 2)+(Math.pow(inputGrid.getHeight(), 2))));
										max_RG_value = 100.0;			//Maximum possible road gravity value
										double rg_value = Math.round((roadGravityF/max_RG_value)*((areaWidth + areaHeight)/16));
										//statusOut.println("RG"+roadGravityF+" / MaxRG:"+max_RG_value+" * (Width"+areaWidth+" + areaHeight"+areaHeight+") / 16");
										
										max_search_index = Math.round(4*((int)rg_value*(1+(int)rg_value)));	//area in which to look for a road cell
										//statusOut.println("MaxSearchIndex: "+max_search_index);
										
										/*****************************************************************
										 * DISPERSION and BREED (Growth Phase 1 and 2)
										 *****************************************************************/
										boolean[][] tmpGridDisp = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
										tmpGridDisp = DispersionGrowth(disp_value,breedF);	//calls the method "DispersionBreed" to perform the Dispersion and Breed Growth Phase
										int anzDispBreed = 0;	//number of cells urbanized by Dispersion and Breed
										for (int x = 0; x < inputGrid.getWidth();x++)
											for(int y = 0; y < inputGrid.getHeight();y++){
												int X = inputGrid.getMinX() + x;
												int Y = inputGrid.getMinY() + y;
												if (tmpGridDisp[X][Y]==true){
													anzDispBreed++;
													tmpGrid[X][Y] = true;	//write the temporary Dispersion and Breed Cells to the overall temporary Cell array
												}		
											}
										statusOut.println("Number of Spontaneous and New Spreading Center Cells: "+anzDispBreed);
										writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
										
										/*********************************************************************
										 * EDGE
										 *********************************************************************/
										boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
										tmpGridSpread = EdgeGrowth(spread);	//calls the method "EdgeGrowth" to perform the Spread (Edge Growth) Phase
										int anzSpread = 0;	//number of cells urbanized by Spread
										for (int x = 0; x < inputGrid.getWidth();x++)
											for(int y = 0; y < inputGrid.getHeight();y++){
												int X = inputGrid.getMinX() + x;
												int Y = inputGrid.getMinY() + y;
												if (tmpGridSpread[X][Y]==true){
													anzSpread++;
													tmpGrid[X][Y] = true;	//write the temporary Spread cells to the overall temporary Cell array
												}		
											}
										statusOut.println("Number of Edge Growth Cells: "+anzSpread);
										writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
									
										/*********************************************************************
										 * ROAD
										 *********************************************************************/
										boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
										statusOut.println("Start Road Growth...");
										tmpGridRoad = RoadGrowth(tmpGrid,rg_value,breed,disp);	//calls the method "RoadGrowth" to perform the Road weighted growth
										statusOut.println("Road things returned...");
										int anzRoad = 0;	//number of cells urbanized by Road Growth
										for (int x = 0; x < inputGrid.getWidth();x++)
											for(int y = 0; y < inputGrid.getHeight();y++){
												int X = inputGrid.getMinX() + x;
												int Y = inputGrid.getMinY() + y;
												if (tmpGridRoad[X][Y]==true){
													anzRoad++;
													tmpGrid[X][Y] = true;	//write the temporary Road Growth cells to the overall temporary Cell array
												}		
											}
										statusOut.println("Number of Road Growth Cells: "+anzRoad);
										writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid

										// Here the main modeling algorithm ends. Next the temporary raster is written in the output raster. 
										//With the outputGrid the next time step is calculated until the maximum number of time steps
										//is reached. After that the next Monte Carlo iteration is performed for the specific parameter combination. 									
										
										// Write the temporary Grid to the Output Grid
										int anzGesamt = 0;
										for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
											for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
												
												int X = inputGrid.getMinX() + x;
												int Y = inputGrid.getMinY() + y;
											
												if (tmpGrid[x][y] == true)
													anzGesamt++;
												
												boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive, when it is 1
												boolean road = roadGrid.getRasterSampleAsFloat(X, Y) > 0f;	//eine Zelle beinhaltet eine Strae...
												if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
													tmpGrid[x][y] = lebt;	//take the values from the current output raster
												if(lebt == true)
													thisYear++;		//all urban cells in this year
												if(road == true)
													roadCount++;	//road cells
												
												outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the results in the output raster
												actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
											}
										}
										statusOut.println("Urban pixels at step "+timeSteps+": "+anzGesamt);
										
										//SELF-MODIFICATION
										growthRate = ((float)anzGesamt / (float)thisYear)*100;	//growth rate
										statusOut.println("Wachstumsrate: "+growthRate+", Urban: "+anzGesamt);
										double percent_urban = (100.0 *(thisYear+roadCount) / (totalPixels-roadCount-excldCount));
										
										//Boom-Year
										if(growthRate > criticalHigh){	//if the growth rate exceeds the critical high value
											slopeF -= (float)percent_urban * (float)slopeSens;	//decrease the slope coefficient
											if(slopeF <= 0.01)	//if it is already very low, make it 1
												slopeF = 1.0;
											roadGravityF += (float)percent_urban * roadGravSens;	//increase the road Gravity
											if(roadGravityF > 100.0)
												roadGravityF = 100.0;
											if(dispF < 100.0){	//increase the other coefficients by multiplying with the boom parameter
												dispF *= boom;
												if(dispF > 100.0)
													dispF = 100.0;
												breedF *= boom;
												if(breedF > 100.0)
													breedF = 100.0;
												spreadF *= boom;
												if(spreadF > 100.0)
													spreadF = 100.0;
											}
										}
										
										//Bust-Year
										if(growthRate < criticalLow){	//if the growth rate falls below the critical low value
											slopeF += (float)percent_urban * (float)slopeSens;	//increase the slope coefficient
											if(slopeF > MaxSlopeResistance)
												slopeF = MaxSlopeResistance;
											roadGravityF -= (float)percent_urban * roadGravSens;	//decrease the road Gravity
											if(roadGravityF <= 0.01)
												roadGravityF = 1.0;
											if(growthRate < criticalLow && dispF > 0){	//decrease the other coefficients by multiplying with the bust parameter
												dispF *= bust;
												if(dispF <= 0.01)
													dispF = 1.0;
												spreadF *= bust;
												if(spreadF <= 0.01)
													spreadF = 1.0;
												breedF *= bust;
												if(breedF <= 0.01)
													breedF = 1.0;
											}
										}
										
										
										if (timeSteps == years){
											outStep.addGrid();	//add a raster to the GridList
											
											// Write the temporary Grid to the Output Grid
											for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
												for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
													actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
													outStep.getGrid(MC-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
												}
											}
											WritableGrid simulationMap = outStep.getGrid(MC-1);		//write the final step result to the grid list
											float[] mrv = MultipleResolutionComparison(simulationMap);	//perform the MRV
											for(int i = 0;i < mrv.length;i++){
												statusOut.println(mrv[i]);
												mrvTempResultArray[i] += mrv[i];
											}
										}
									}	//END OF THE TIME STEPS
									
									if (MC == MonteCarlo){
										SumOutStep.addGrid();	//add a final raster to the GridList (contains the sum of all iterations)
										
										//statusOut.println("Step Grids: "+outStep.getGridCount());
										
										for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
											for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {	//for every pixel...
												float sumUrb = 0; 	//sum of pixel values
												for (int gridNo = 0; gridNo < MC; gridNo++){	//look in every grid of the GridList
													float tempUrb = (Float) outStep.getGrid(gridNo).getRasterSample(x,y);	//take the pixel value
													sumUrb  += tempUrb;		//and add its value
												}
												SumOutStep.getGrid(runs-1).setRasterSample(sumUrb,x,y);	//put the layer in the raster list containing the step results
											}
										}
										
										System.out.print("Slope: "+slope+" Spread: "+spread+" Disp: "+disp+" Breed: "+breed+" Road: "+roadGravity+" : ");
										
										for(int i = 0;i < mrvTempResultArray.length-2;i++){
											//print every element of the result array comma separated in one line in the status output
											finalArray[i] = mrvTempResultArray[i]/MonteCarlo;
											System.out.print(finalArray[i]+",");
										}
										System.out.print("//Weighted Averages (FT, WA): ");
										for(int j = 0;j < 2;j++){
											//print every element of the result array comma separated in one line in the status output
											finalArray[numSteps+j] = mrvTempResultArray[j]/MonteCarlo;
											System.out.print(finalArray[j]+",");
										}
										for (int grids = MC-1; grids >= 0; grids--){
											outStep.removeGrid(grids); 				//empty GridList
										}
										//statusOut.println("Step Grids nach leeren: "+outStep.getGridCount());
								
									} 
									countSetIndividuals();
									
								}	System.out.println("");
								
								//END OF THE MONTE CARLO ITERATIONS
							}	//END OF ROADGRAVITY COMBINATION
						}	//END OF BREED COMBINATION
					}	//END OF DISP COMBINATION
				}	//END OF SPREAD COMBINATION
			}	//END OF SLOPE COMBINATION
			System.out.println("Finished "  + " in " + ((System.currentTimeMillis() - localStartTime)) + " ms\n");
		}	//END OF MODEL STEP
}