package edu.bonn.xulu.plugin.model.sleuth;

import edu.bonn.xulu.model.XuluModel;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ScalarProperty;

/**
 *  <b>Urban Growth Model with Self Modification</b><br>
 *  This model is based on the {@link UrbanGrowthModel Urban Growth Model}, developed by Keith Clarke at the UC Santa Barbara, CA, within the Gigalopolis Project, funded by the
 *  USGS. It has exactly the same growth rules as its parent class, with Self Modification of several coefficients enabled.
 *  For a detailed description refer to {@link UrbanGrowthModel Urban Growth Model}
 *  
 * <br>
 * In addition to the UGM Coefficients, these parameters are needed:<br>
 * <ul>
 *  <li><b>Road Gravity Sensitivity</b>: Influences the roadGravity coefficient</li>
 *  <li><b>Slope Sensitivity</b>: Influences the slope coefficient</li>
 *  <li><b>Critical Low</b>: When growth rate falls below this value, the coefficients are altered</li>
 *  <li><b>Critical High</b>: When growth rate exceeds this value, the coefficients are altered</li>
 *  <li><b>Boom</b>: Boom Phase. The Coefficients are increased by this value.</li>
 *  <li><b>Bust</b>: Bust Phase. The Coefficients are decreased by this value.</li>
 * </ul>
 * 
 * @see UrbanGrowthModelSelfModificationContentManager
 * @see UrbanGrowthModel
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 */
public class UrbanGrowthModelSelfModification extends UrbanGrowthModel {

	//protected UrbanGrowthModelSelfModificationContentManager contManager;

	// ********** Read/Write access, that has to be maintained ***
	// ********** through the whole model run ********************
	private PropertyReadAccess RA_roadGravSens = null;	//RoadGravity Sensitivity
	private PropertyReadAccess RA_slopeSens = null;	//Slope Sensitivity
	private PropertyReadAccess RA_criticalLow = null;	//Critical Low
	private PropertyReadAccess RA_criticalHigh = null;	//Critical High
	private PropertyReadAccess RA_Boom = null;	//Boom
	private PropertyReadAccess RA_Bust = null;	//Bust

	//**************** Variables to work with *******************
	protected double roadGravSens = 0;	//RoadGravity Sensitivity
	protected double slopeSens = 0;	//Slope Sensitivity
	protected double criticalLow = 0;	//critical Low
	protected double criticalHigh = 0;	//critical High
	protected double boom = 0;	//Boom
	protected double bust = 0;	//Bust
		
	private int slopeMax = 0;
	private int totalPixels = 0;
	private int excldCount = 0;
	
	/**
	 * First Constructor
	 * @param contManager
	 */
	public UrbanGrowthModelSelfModification(UrbanGrowthModelSelfModificationContentManager contManager) {
		super(contManager);
	//	this.contManager = (UrbanGrowthModelSelfModificationContentManager) super.contManager;
	}
	/**
	 * Second Constructor. The Content Manager is the parent content manager as well as the child content manager
	 */
	public UrbanGrowthModelSelfModification(){
		super(new UrbanGrowthModelSelfModificationContentManager());
	}

	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		super.performModelInit();
		
		RA_roadGravSens = null;
		if (contManager.getResource(13).getData() != null)
			RA_roadGravSens = ((ScalarProperty) contManager.getResource(13).getData())
					.getReadAccess(this);
		RA_slopeSens = null;
		if (contManager.getResource(14).getData() != null)
			RA_slopeSens = ((ScalarProperty) contManager.getResource(14).getData())
					.getReadAccess(this);
		RA_criticalLow = null;
		if (contManager.getResource(15).getData() != null)
			RA_criticalLow = ((ScalarProperty) contManager.getResource(15).getData())
					.getReadAccess(this);
		RA_criticalHigh = null;
		if (contManager.getResource(16).getData() != null)
			RA_criticalHigh = ((ScalarProperty) contManager.getResource(16).getData())
					.getReadAccess(this);
		RA_Boom = null;
		if (contManager.getResource(17).getData() != null)
			RA_Boom = ((ScalarProperty) contManager.getResource(17).getData())
					.getReadAccess(this);
		RA_Bust = null;
		if (contManager.getResource(18).getData() != null)
			RA_Bust = ((ScalarProperty) contManager.getResource(18).getData())
					.getReadAccess(this);

		// Variables to work with
	    roadGravSens = RA_roadGravSens.getValueAsDouble();
	    slopeSens = RA_slopeSens.getValueAsDouble();
	    criticalLow = RA_criticalLow.getValueAsDouble();
	    criticalHigh = RA_criticalHigh.getValueAsDouble();
	    boom = RA_Boom.getValueAsDouble();
	    bust = RA_Bust.getValueAsDouble();

	  //Calculate total amount of pixels and number of excluded pixels
		for (int x = 0; x < inputGrid.getWidth();x++ ) 
			for (int y = 0; y < inputGrid.getHeight(); y++ ) {
				if ( x < inputGrid.getMinX() || y < inputGrid.getMinY() || x > inputGrid.getMinX()+inputGrid.getWidth()-1 || y > inputGrid.getMinY()+inputGrid.getHeight()-1 ||
			            Float.isNaN(outputGrid.getRasterSampleAsFloat(x,y)) ) //ignore neighbor, if outside the raster
			             continue;
				totalPixels++;
				if (areaRestr.getRasterSampleAsFloat(x,y) != 0 && !Float.isNaN(areaRestr.getRasterSampleAsFloat(x,y)))
					excldCount++;
			}
		
		statusOut.println("Total: "+totalPixels+", Excluded: "+excldCount);
		
	}

	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {
		super.performModelDispose();
		// free resources
		releaseAccess(RA_roadGravSens);
		releaseAccess(RA_slopeSens);
		releaseAccess(RA_criticalLow);
		releaseAccess(RA_criticalHigh);
		releaseAccess(RA_Boom);
		releaseAccess(RA_Bust);
	}

	/**
	 * The main method of the Urban Growth Model. For a detailed description refer to the performModelStep in 
	 * {@link UrbanGrowthModel UrbanGrowthModel}. The difference to the plain UGM Calibration model is the Self Modification 
	 * of several coefficients at the end of each modelling step.
	 */
	public void performModelStep(int stepNo) {
		
		long localStartTime = System.currentTimeMillis();
		statusOut.println("Starting step " + stepNo + "...");
		
		int MaxSlopeResistance = 100;
		slopeMax = getSlopeMax();
		double exp = getSlopeExp(slope,MaxSlopeResistance);	//calculate the exp value based on the actual slope
		statusOut.println("EXP: "+exp);
		slopeLUT = getSlopeLookup(slopeMax,exp,criticalSlope);	//generate the lookup-table based on the actual exp-value

		boolean[][] tmpGrid = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];
		Float actUrb = null;

		double growthRate = 0;
		int thisYear = 0;
		int roadCount = 0;
		
		//DISPERSION
		double disp_value = Math.round((disp*0.005)* Math.sqrt(Math.pow(inputGrid.getWidth(), 2)+(Math.pow(inputGrid.getHeight(), 2))));
		
		//ROAD GRAVITY
		double max_RG_value = 100.0;
		double rg_value = Math.round((roadGravity/max_RG_value)*((areaWidth + areaHeight)/16)); 
		

		/*****************************************************************
		 * DISPERSION and BREED (Growth Phase 1 and 2)
		 *****************************************************************/
		boolean[][] tmpGridDisp = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridDisp = DispersionGrowth(disp_value,breed);	//calls the method "DispersionBreed" to perform the Dispersion and Breed Growth Phase
		int anzDispBreed = 0;	//number of cells urbanized by Dispersion and Breed
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridDisp[X][Y]==true){
					anzDispBreed++;
					tmpGrid[X][Y] = true;	//write the temporary Dispersion and Breed Cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Spontaneous and New Spreading Center Cells: "+anzDispBreed);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
		
				
		/*********************************************************************
		 * EDGE
		 *********************************************************************/
		boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridSpread = EdgeGrowth(spread);	//calls the method "EdgeGrowth" to perform the Spread (Edge Growth) Phase
		int anzSpread = 0;	//number of cells urbanized by Spread
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridSpread[X][Y]==true){
					anzSpread++;
					tmpGrid[X][Y] = true;	//write the temporary Spread cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Edge Growth Cells: "+anzSpread);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
	
		/*********************************************************************
		 * ROAD
		 *********************************************************************/
		boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
		tmpGridRoad = RoadGrowth(tmpGrid,rg_value,breed,disp);	//calls the method "RoadGrowth" to perform the Road weighted growth
		int anzRoad = 0;	//number of cells urbanized by Road Growth
		for (int x = 0; x < inputGrid.getWidth();x++)
			for(int y = 0; y < inputGrid.getHeight();y++){
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				if (tmpGridRoad[X][Y]==true){
					anzRoad++;
					tmpGrid[X][Y] = true;	//write the temporary Road Growth cells to the overall temporary Cell array
				}		
			}
		statusOut.println("Number of Road Growth Cells: "+anzRoad);
		writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
		
		
		outStep.addGrid();
		int anzGesamt = 0;
		
		// Write the temporary Grid to the Output Grid
		for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
			for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
				
				int X = inputGrid.getMinX() + x;
				int Y = inputGrid.getMinY() + y;
				
				if (tmpGrid[x][y] == true)
					anzGesamt++;
				
				boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive, when it is 1
				boolean road = roadGrid.getRasterSampleAsFloat(X, Y) > 0f;	//a cell contains a road...
				if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
					tmpGrid[x][y] = lebt;	//take the values from the current output raster
				if(lebt == true)
					thisYear++;		//all urban cells in this year
				if(road == true)
					roadCount++;	//road cells

				outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the results in the output raster
				actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
				outStep.getGrid(stepNo-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
			}
		}
		
		
		//SELF-MODIFICATION
		growthRate = ((float)anzGesamt / (float)thisYear)*100;	//growth rate
		statusOut.println("Wachstumsrate: "+growthRate+", Urban: "+anzGesamt);
		double percent_urban = (100.0 *(thisYear+roadCount) / (totalPixels-roadCount-excldCount));
		
		//Boom-Year
		if(growthRate > criticalHigh){		//if the growth rate exceeds the critical high value
			slope -= (float)percent_urban * (float)slopeSens;	//decrease the slope coefficient
			if(slope <= 0.01)		//if it is already very low, make it 1
				slope = 1.0;
			roadGravity += (float)percent_urban * roadGravSens;	//increase the road Gravity
			if(roadGravity > 100.0)
				roadGravity = 100.0;
			if(disp < 100.0){		//increase the other coefficients by multiplying with the boom parameter
				disp *= boom;
				if(disp > 100.0)
					disp = 100.0;
				breed *= boom;
				if(breed > 100.0)
					breed = 100.0;
				spread *= boom;
				if(spread > 100.0)
					spread = 100.0;
			}
		}
		//Bust-Year
		if(growthRate < criticalLow){		//if the growth rate falls below the critical low value
			slope += (float)percent_urban * (float)slopeSens;	//increase the slope coefficient
			if(slope > MaxSlopeResistance)
				slope = MaxSlopeResistance;
			roadGravity -= (float)percent_urban * roadGravSens;	//decrease the road Gravity
			if(roadGravity <= 0.01)
				roadGravity = 1.0;
			if(growthRate < criticalLow && disp > 0){	//decrease the other coefficients by multiplying with the bust parameter
				disp *= bust;
				if(disp <= 0.01)
					disp = 1.0;
				spread *= bust;
				if(spread <= 0.01)
					spread = 1.0;
				breed *= bust;
				if(breed <= 0.01)
					breed = 1.0;
			}
		}
		System.out.println("thisYear: "+thisYear+", roadCount: "+roadCount+", excldCount: "+excldCount+", percent_urban: "+percent_urban);
		System.out.println("Slope-Sensitivity: "+slope+", RoadGravity-Sensitivity: "+roadGravity+", Diffusion-Sensitivity: "+disp+", Breed-Sensitivity: "+breed+", Spread-Sensitivity: "+spread);
		
		System.out.println("Finished step " + (stepNo) + " in "
				+ ((System.currentTimeMillis() - localStartTime)) + " ms\n");
		localStartTime = System.currentTimeMillis();
	}

}
