package edu.bonn.xulu.plugin.model.sleuth;


import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.property.ScalarProperty;
import edu.bonn.xulu.model.XuluModel;

/**
 *  <b>Urban Growth Model with Monte-Carlo Iterations</b><br>
 *  This model is based on the {@link UrbanGrowthModel Urban Growth Model}, developed by Keith Clarke at the UC Santa Barbara, CA, within the Gigalopolis Project, funded by the
 *  USGS. It has exactly the same growth rules as its parent class, but runs with n Monte-Carlo Iterations. So the resulting map does not contain the urban extent after a model
 *  run, but the probability for new urbanization. The more often a cell has been selected for urbanization during the Monte-Carlo Iterations, the higher is that value in the 
 *  output map.
 *  For a detailed description refer to {@link UrbanGrowthModel Urban Growth Model}
 *  
 *  <br>
 *  In addition to the UGM parameters, this parameter is needed:<br>
 * <ul>
 *  <li><b>Monte Carlo Iterations({@code slopeGrid})</b>: The number of Monte Carlo Iterations.</li>
 * </ul>
 * 
 * @see UrbanGrowthModel_MC_ContentManager
 * @see UrbanGrowthModel
 * @author <a href="mailto:goetzke@uni-bonn.de">Roland Goetzke</a>
 * @version 1.0
 */
public class UrbanGrowthModel_MC extends UrbanGrowthModel{
	
	protected PropertyReadAccess RA_MonteCarlo = null; // Number of MonteCarlo Iterations
	
	protected int MonteCarlo = 0; // Number of MonteCarlo Iterations
	protected int years = 0;

	/**
	 * First Constructor
	 * @param contManager
	 */
	public UrbanGrowthModel_MC(UrbanGrowthModel_MC_ContentManager contManager) {
		super(contManager);
		this.contManager = (UrbanGrowthModel_MC_ContentManager) super.contManager;
	}
	/**
	 * Second Constructor. The Content Manager is the parent content manager as well as the child content manager
	 */
	public UrbanGrowthModel_MC(){
		super(new UrbanGrowthModel_MC_ContentManager());
	}
	
	/**
	 * Initializes the model. Like in the init method of every {@link XuluModel}
	 * the resources are initalized. 
	 */
	public void performModelInit() {
		super.performModelInit();
		
		
		RA_MonteCarlo = null;
		if (contManager.getResource(13).getData() != null)
			RA_MonteCarlo = ((ScalarProperty) contManager.getResource(13).getData())
					.getReadAccess(this);

		MonteCarlo  = RA_MonteCarlo.getValueAsInt();
		this.stepCount = MonteCarlo;
		years = steps.getValueAsInt();
	}
		
	/**
	* like in every model: frees the resources
	 */
	public void performModelDispose() {
		super.performModelDispose();
		releaseAccess(RA_MonteCarlo);
	}
	
	/**
	 * The main method of the Urban Growth Model. For a detailed description refer to the performModelStep in 
	 * {@link UrbanGrowthModel UrbanGrowthModel}. The difference is, that the number of steps is not based on
	 * the years that have to be modelled, but on the number of Monte-Carlo iterations the model has to perform.
	 * In every iteration the full {@link UrbanGrowthModel UrbanGrowthModel} is included in a for-loop. In the
	 * StepResults GridList are the results for the final year for every Monte-Carlo iteration and the sum of
	 * all iterations, which indicates high values for pixels that have been selected several times for urbanization. 
	 */
	public void performModelStep(int stepNo) {
		
		statusOut.println("Performing "+MonteCarlo+" Monte Carlo Iterations (years: "+stepNo);
		long localStartTime = System.currentTimeMillis(); //time measure
		Float actUrb = null;	//urban cell in the raster
		for(int timeSteps = 1;timeSteps <=years;timeSteps++){	//for every time step...
			statusOut.println("Year "+timeSteps);
			boolean[][] tmpGrid = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
							
			/*****************************************************************
			 * DISPERSION and BREED (Growth Phase 1 and 2)
			 *****************************************************************/
			boolean[][] tmpGridDisp = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
			tmpGridDisp = DispersionGrowth(disp_value,breed);	//calls the method "DispersionBreed" to perform the Dispersion and Breed Growth Phase
			int anzDispBreed = 0;	//number of cells urbanized by Dispersion and Breed
			for (int x = 0; x < inputGrid.getWidth();x++)
				for(int y = 0; y < inputGrid.getHeight();y++){
					int X = inputGrid.getMinX() + x;
					int Y = inputGrid.getMinY() + y;
					if (tmpGridDisp[X][Y]==true){
						anzDispBreed++;
						tmpGrid[X][Y] = true;	//write the temporary Dispersion and Breed Cells to the overall temporary Cell array
					}		
				}
			statusOut.println("Number of Spontaneous and New Spreading Center Cells: "+anzDispBreed);
			writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
				
			/*********************************************************************
			 * EDGE
			 *********************************************************************/
			boolean[][] tmpGridSpread = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
			tmpGridSpread = EdgeGrowth(spread);	//calls the method "EdgeGrowth" to perform the Spread (Edge Growth) Phase
			int anzSpread = 0;	//number of cells urbanized by Spread
			for (int x = 0; x < inputGrid.getWidth();x++)
				for(int y = 0; y < inputGrid.getHeight();y++){
					int X = inputGrid.getMinX() + x;
					int Y = inputGrid.getMinY() + y;
					if (tmpGridSpread[X][Y]==true){
						anzSpread++;
						tmpGrid[X][Y] = true;	//write the temporary Spread cells to the overall temporary Cell array
					}		
				}
			statusOut.println("Number of Edge Growth Cells: "+anzSpread);
			writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
			
			/*********************************************************************
			 * ROAD
			 *********************************************************************/
			
			boolean[][] tmpGridRoad = new boolean[inputGrid.getWidth()][inputGrid.getHeight()];	//a temporary raster
			tmpGridRoad = RoadGrowth(tmpGrid,rg_value,breed,disp);	//calls the method "RoadGrowth" to perform the Road weighted growth
			int anzRoad = 0;	//number of cells urbanized by Road Growth
			for (int x = 0; x < inputGrid.getWidth();x++)
				for(int y = 0; y < inputGrid.getHeight();y++){
					int X = inputGrid.getMinX() + x;
					int Y = inputGrid.getMinY() + y;
					if (tmpGridRoad[X][Y]==true){
						anzRoad++;
						tmpGrid[X][Y] = true;	//write the temporary Road Growth cells to the overall temporary Cell array
					}		
				}
			statusOut.println("Number of Road Growth Cells: "+anzRoad);
			writeToGrid(tmpGrid);	//calls the method "writeToGrid" to write the new urbanized cells to the output grid
			
			int anzGesamt = 0;
			
			// Write the temporary Grid to the Output Grid
			for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
				for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
					
					int X = inputGrid.getMinX() + x;
					int Y = inputGrid.getMinY() + y;
					
					if (tmpGrid[x][y] == true)
						anzGesamt++;
					
					boolean lebt = outputGrid.getRasterSampleAsFloat(X , Y ) > 0f;	//a cell is alive, when it is 1
					if (tmpGrid[x][y] == false)	//if there is nothing in the temporary raster
						tmpGrid[x][y] = lebt;	//take the values from the current output raster
					
					outputGrid.setRasterSample( tmpGrid[x][y] ? 1f : 0f ,X ,Y);	//write the results in the output raster
					actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
					//	outStep.getGrid(stepNo-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
				}
			}
			statusOut.println("Urban pixels at step "+timeSteps+": "+anzGesamt);	
		}
		outStep.addGrid();	//add a raster to the GridList
		
		// Write the temporary Grid to the Output Grid
		for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
			for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {
				actUrb  = (Float)outputGrid.getRasterSample(x,y);	//the actual urban extent
				outStep.getGrid(stepNo-1).setRasterSample(actUrb,x,y);	//put the layer in the raster list containing the step results
			}
		}	
		
		countSetIndividuals();	//sets the outputGrid to the initial inputGrid in order to start the next modelling step with the same initial map
		System.out.println("Finished step " + (stepNo) + " in "
				+ ((System.currentTimeMillis() - localStartTime)) + " ms\n");
		if (stepNo == MonteCarlo){
			outStep.addGrid();	//add a final raster to the GridList (contains the sum of all iterations)
			
			for (int y = 1; y < inputGrid.getHeight()-1;y++ ) {
				for (int x = 1; x < inputGrid.getWidth()-1; x++ ) {	//for every pixel...
					float sumUrb = 0; 	//sum of pixel values
					for (int gridNo = 0; gridNo <= stepNo; gridNo++){	//look in every grid of the GridList
						float tempUrb = (Float) outStep.getGrid(gridNo).getRasterSample(x,y);	//take the pixel value
						sumUrb  += tempUrb;		//and add its value
					}
					outStep.getGrid(stepNo).setRasterSample(sumUrb,x,y);	//put the layer in the raster list containing the step results
				}
			}	
		}
	}
}
