/** XULU - This file is part of the eXtendable Unified Land Use Modelling Platform (XULU)

    This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
    This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
    You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

    Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
    Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
    Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package edu.bonn.xulu.plugin.model.test;

import java.awt.image.DataBuffer;

import schmitzm.data.property.ScalarProperty;
import schmitzm.data.property.PropertyReadAccess;
import schmitzm.data.WritableGrid;
import edu.bonn.xulu.data.XuluDataException;
import edu.bonn.xulu.model.AbstractStepModel;
import edu.bonn.xulu.model.ModelContentManager;
import edu.bonn.xulu.model.AbstractModelContentManager;
import edu.bonn.xulu.model.PropertiesResource;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt ein Test-Modell dar, dass eigentlich gar nichts macht.
 * Es benoetigt 3 Ressourcen:
 * <ol>
 * <li>Eine Ganzzahl, welche die Anzahl der durchzufuehrenden Modellierungsschritte
 *     festlegt</li>
 * <li>Ein diskretes Double-Raster als Eingabe ({@link WritableGrid})</li>
 * <li>Ein entsprechendes diskretes Double-Raster als Ausgabe ({@link WritableGrid})</li>
 * </ol>
 * In jedem Zeitschritt t werden alle Rasterzellen r des Eingabe-Rasters und deren
 * t-Umgebung betrachtet. Ist der Wert von r groesser als der Durchschnitt der
 * Umgebungswerte, wird die entsprechende Zelle im Ausgabe-Raster auf 1 gesetzt,
 * ansonsten auf 0.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class TestModel extends AbstractStepModel {

  /** Speichert den ContentManager fuer das Modell.
   *  @see TestModel.ContentManager */
  protected ContentManager contManager;

  private WritableGrid inGrid    = null;
  private WritableGrid outGrid   = null;


  // ZUM TESTEN: Referenz auf Property des Ausgaberasters
  ScalarProperty outRasterProp = null;

  /**
   * Erzeugt ein neues Test-Modell.
   */
  public TestModel() {
    super( new ContentManager() );
    contManager = (ContentManager)super.contManager;
  }

  /**
   * Initialisiert das Model, indem die Ressourcen aus dem ContentManager
   * geladen werden (Lese/Schreib-Rechte).
   */
  public void performModelInit() {
    if ( contManager == null )
      return;

//    /////////////////////////
//    // ZUM TESTEN: die VisualisationUpdateListener des Ausgaberaster auf ein
//    //             ChangeEvent reagieren lassen
//    outRasterProp = (ScalarProperty)contManager.getResource(2).getData();
//    schmitzm.data.event.ObjectListener[] listener = outRasterProp.getObjectListener(edu.bonn.xulu.appl.VisualisationUpdateListener.class);
//    for (int i=0; i<listener.length; i++)
//      ((edu.bonn.xulu.appl.VisualisationUpdateListener)listener[i]).setAutoRefresh(true);
//    /////////////////////////

    // Ressourcen aus dem ContentManager referenzieren (Lese/Schreib-Rechte
    // erzeugen)
    this.stepCount = contManager.getStepCount();
    inGrid         = contManager.getInRaster();
    outGrid        = contManager.getOutRaster();
    // Ausgabe-Raster leeren
    for (int y=outGrid.getMinY(); y<outGrid.getHeight(); y++)
      for (int x=outGrid.getMinX(); x<outGrid.getWidth(); x++)
        outGrid.setRasterSample(Float.NaN,x,y);

//    // ZUM TESTEN: Aenderung des Ausgaberasters propagieren
//    outRasterProp.fireGeneralEvent();
  }

  /**
   * Fuehrt einen Schritt des Modellablaufs durch.
   */
  public void performModelStep(int stepNo) {
    statusOut.println("Started step ".concat(String.valueOf(stepNo)));
    for (int y=inGrid.getMinY(); y<inGrid.getHeight(); y++)
      for (int x=inGrid.getMinX(); x<inGrid.getWidth(); x++) {
        // Wert des Eingabe-Rasters
        float  inValue = (Float)inGrid.getRasterSample(x,y);
        // Durchschnitt der Umgebung berechnen
        int    cellCount = 0;
        float  cellSum   = 0.0f;
        for (int r=0; r<stepNo; r++)
          for (int uy=y-r; uy<=y+r; uy++)
            if ( uy >= inGrid.getMinY() && uy < inGrid.getMinY()+inGrid.getHeight() )
              for (int ux=x-r; ux<=x+r; ux++)
                if ( ux >= inGrid.getMinX() && ux < inGrid.getMinX()+inGrid.getWidth() ) {
                  cellCount++;
                  cellSum += (Float)inGrid.getRasterSample(ux,uy);
                  checkBreakingCommands();
                }
        // Wenn Wert groesser dem Durchschnitt, Ausgabe auf 1 setzten, sonst
        // auf 0
        int outValue = ( inValue > cellSum/cellCount ) ? 1 : 0;
        outGrid.setRasterSample((float)outValue,x,y);
      }
//    // ZUM TESTEN: Aenderung des Ausgaberasters propagieren
//    outRasterProp.fireGeneralEvent();

    statusOut.println("Ended step ".concat(String.valueOf(stepNo)));
  }

  /**
   * Gibt die Lese/Schreibrechte auf das Ein- und Ausgaberaster wieder frei.
   * Interne Ressourcen brauchen keine aus dem Speicher entfernt werden.
   */
  public void performModelDispose() {
    contManager.releaseInRaster();
    contManager.releaseOutRaster();
  }


  /**
   * Diese Klasse verwaltet die Ressourcen fuer das {@linkplain TestModel TestModel}
   * und prueft diese auf Korrektheit.<br>
   * Die Art der benoetigten Ressourcen ist der {@linkplain TestModel Model-Doku}
   * zu entnehmen.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ContentManager extends AbstractModelContentManager {
    // Ressource 0: Anzahl an Modellschritten
    private static final int    STEP_MIN = 1;
    private static final int    STEP_MAX = 100;
    private static final String DESC0 = "Anzahl an Modellschritten (".concat(String.valueOf(STEP_MIN)).concat("-").concat(String.valueOf(STEP_MAX)).concat(")");
    private PropertyReadAccess stepAccess = null;
    // Ressource 1: Eingabe-Raster
    private static final String DESC1 = "Eingabe-Raster (double-Raster)";
    private PropertyReadAccess inRasterAccess = null;
    // Ressource 2: Ausgabe-Raster
    private static final String DESC2 = "Ausgabe-Raster (double-Raster)";
    private PropertyReadAccess outRasterAccess = null;

    /**
     * Erzeugt einen neuen ContentManager fuer das {@linkplain TestModel TestModel}.
     */
    public ContentManager() {
      super(3);
      // Ressource 0: Anzahl an Modellschritten
      resource[0] = new PropertiesResource(DESC0,ScalarProperty.class,false) {
        public void checkTypeAndError(Object data) {
          super.checkTypeAndError(data);
          // muss als Datentyp Long oder Integer haben
          if ( !BaseTypeUtil.isNumeric(((ScalarProperty)data).getType()) ||
                BaseTypeUtil.isDecimal(((ScalarProperty)data).getType()) )
            throw new XuluDataException("Long- oder Integer-Property erforderlich fuer '".concat(this.getDesc()).concat("'"));
        }
        public void checkAndError(Object data) {
          super.checkAndError(data);
          // muss zwischen 1 und 20 liegen
          int value = ((ScalarProperty)data).getOneTimeReadAccess().getValueAsInt();
          if ( value < STEP_MIN ||  STEP_MAX < value )
            throw new XuluDataException("'".concat(this.getDesc()).concat("' nicht zul�ssig!"));
        }
      };

      // Ressource 1: Eingabe-Raster
      resource[1] = new PropertiesResource(DESC1,ScalarProperty.class,false) {
        public void checkTypeAndError(Object data) {
          super.checkTypeAndError(data);
          // muss als Datentyp ein WritableGrid haben
          if ( !WritableGrid.class.isAssignableFrom(((ScalarProperty)data).getType()) )
            throw new XuluDataException("WritableGrid-Property erforderlich fuer '".concat(this.getDesc()).concat("'"));
        }
        public void checkAndError(Object data) {
          super.checkAndError(data);
          // das Raster muss Double-Werte liefern
          WritableGrid value = (WritableGrid)((ScalarProperty)data).getOneTimeReadAccess().getValue();
          if ( value.getSampleType() != DataBuffer.TYPE_FLOAT )
            throw new XuluDataException("Das Eingabe-Raster muss double-Werte liefern!");
        }
      };

      // Ressource 2: Ausgabe-Raster
      resource[2] = new PropertiesResource(DESC2,ScalarProperty.class,false) {
        public void checkTypeAndError(Object data) {
          super.checkTypeAndError(data);
          // muss als Datentyp ein WritableGrid haben
          if ( !WritableGrid.class.isAssignableFrom(((ScalarProperty)data).getType()) )
            throw new XuluDataException("WritableGrid-Property erforderlich fuer '".concat(this.getDesc()).concat("'"));
        }
        public void checkAndError(Object data) {
          super.checkAndError(data);
          // das Raster muss Integer aufnehmen koennen
          WritableGrid value = (WritableGrid)((ScalarProperty)data).getOneTimeReadAccess().getValue();
          if ( value.getSampleType() != DataBuffer.TYPE_FLOAT )
            throw new XuluDataException("Das Ausgabe-Raster muss double-Werte aufnehmen koennen!");
        }
      };
    }

    /**
     * Prueft alle Modell-Ressourcen auf Korrektheit.
     * Neben den lokalen Bedingungen an die Ressourcen (die bereits durch
     * die Methode der Oberklasse gechecked werden), prueft diese Methode,
     * ob das Ausgabe-Raster dieselbe Groesse hat, wie das Eingabe-Raster.
     */
    public void checkAndError() {
      super.checkAndError();
      WritableGrid in = this.getInRaster();
      WritableGrid out = this.getOutRaster();
      if ( in.getWidth() != out.getWidth() || in.getHeight() != out.getHeight() ) {
        releaseInRaster();
        releaseOutRaster();
        throw new XuluDataException("Das Ausgabe-Raster muss dieselbe Groesse haben, wie das Eingabe-Raster");
      }
      releaseInRaster();
      releaseOutRaster();
    }

    /**
     * Liefert die Anzahl an Schritten, die das Modell laufen soll.
     */
    public int getStepCount() {
      return ((ScalarProperty)resource[0].getData()).getOneTimeReadAccess().getValueAsInt();
    }

    /**
     * Liefert das Raster, das die Modell-Ausgaben aufnimmt.
     */
    public WritableGrid getInRaster() {
      releaseInRaster();
      inRasterAccess = ((ScalarProperty)resource[1].getData()).getReadAccess(this);
      return (WritableGrid)inRasterAccess.getValue();
    }

    /**
     * Gibt das Leserecht auf das Eingabe-Raster wieder frei.
     */
    public void releaseInRaster() {
      if ( inRasterAccess != null )
        inRasterAccess.release();
    }

    /**
     * Liefert das Raster, das dem Modell als Eingabe dient.
     */
    public WritableGrid getOutRaster() {
      releaseOutRaster();
      outRasterAccess = ((ScalarProperty)resource[2].getData()).getReadAccess(this);
      return (WritableGrid)outRasterAccess.getValue();
    }

    /**
     * Gibt das Leserecht auf das Eingabe-Raster wieder frei.
     */
    public void releaseOutRaster() {
      if ( outRasterAccess != null )
        outRasterAccess.release();
    }

  }
}
